unit Channel;
{Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

uses Expr, Classes, TypInfo, IniFiles, MathSup, PersistentVariables, DataTypes;

type

  TChannel = class;
  TChannels = class;

  TChannelArrayVariable = class (TArray)
  private
    Channel: TChannel;
  public
    constructor Create(AChannel: TChannel);
    function Value(var a:array of Double):Double; override;
    function LowerBound(Bound:Integer): Integer; override;
    function UpperBound(Bound:Integer): Integer; override;
    procedure Store(var a:array of Double); override;
  end;

  TSeriesVariable = class (TArray)
  private
    Channel: TChannel;
  public
    constructor Create(AChannel: TChannel);
    function Value(var a:array of Double):Double; override;
    function LowerBound(Bound:Integer): Integer; override;
    function UpperBound(Bound:Integer): Integer; override;
    procedure Store(var a:array of Double); override;
  end;

  TPeakArray = array [0..3] of TDoubleArray;

  TChannel = class(TVariabledPersistent)
  private
    FSeriesMemberSize: Integer;
    FChannels: TChannels;
    FNumber: Integer;
    FChannelArrayVariable: TChannelArrayVariable;
    FSeriesVariable: TSeriesVariable;
    FLeftMark: Integer;
    FMark: Integer;
    FRightMark: Integer;
    procedure SetSeriesMemberSize(const Value: Integer);
    function GetSeriesValue(I,J:Integer):Integer;
    procedure SetSeriesValue(I, J: Integer; const Value: Integer);
    function GetValue(i: Integer): Extended;
    function GetText(i: Integer): string;
    procedure SetText(i: Integer; const Value: string);
    function GetChannelArrayVariable: TChannelArrayVariable;
    function GetSeriesVariable: TSeriesVariable;
  protected
    constructor Create(AChannels:TChannels);
    procedure SetName(const Value: string); override;
  public
    DataType: TDataType;
    DataSourceName: string;
    DataSourceInfo: string; { Extra info written by DataSource after collection }
    Count: Integer; { Number of values filled }
    Data32: array of Integer; { Note that only one of Data16, Data32, Data64 is ever used }
    Data16: array of Smallint;
    Data64: array of Double;
    Series: array of TShortintArray;
    Visible: Boolean;
    Invert: Boolean;
    Display: TComponent;
    Peaks: array [0..1] of TPeakArray;

    destructor Destroy; override;
    procedure Change(NewDataType: TDataType; AMaxPoints: Integer; Preserve: Boolean);
    procedure SetExtendedValue(i: Integer; const Value: Extended);
    procedure SetInt16Value(i:Integer; Value:Smallint);
    procedure SetInt64Value(i:Integer; Value:Int64);
    procedure SetIntegerValue(i:Integer; Value:Integer);
    function Int64Value(i:Integer):Int64;
    procedure ReadConfig(AMaxPoints: Integer; ConfigFile:TCustomIniFile);
    procedure Changed(CountChanged:Boolean);
    function Limits(out Mi, Mx: Double; First, Last: Integer):Boolean;
    function NormalizeNpeaks(Index:Integer):Integer;
    function LastPeak(Index:Integer):Integer;
    procedure WriteRLE(var F: Text; First, Last: Integer);
    function ReadRle(Stream: TStream): string;
    function GetnPeaks(Index: Integer): Integer;
    procedure SetnPeaks(Index: Integer; const Value: Integer);

    procedure Compact(n:Integer; Average: Boolean);
    procedure InvertValues;
    function Average(First,Last:Integer):Double;
    procedure DeletePoints(First, nPoints:Integer);
    procedure InsertPoints(First, nPoints:Integer);
    procedure Paste(Source: TChannel; First, Last: Integer);
    procedure Reverse(ReversePeaks:Boolean);
    procedure Deglitch(Point: Integer);
    procedure AddVariables(Lex: TLex; Prefix: string); override;

    function MaxPoints: Integer;
    function GetSeriesMemberSize(I: Integer):Integer;
    property Number: Integer read FNumber;
    property Text[i:Integer]:string read GetText write SetText;
    property Value[i:Integer]:Extended read GetValue write SetExtendedValue;
    property SeriesMemberSize:Integer read FSeriesMemberSize write SetSeriesMemberSize;
    property SeriesValue[I:Integer; J:Integer]:Integer read GetSeriesValue write SetSeriesValue;
    property Channels: TChannels read FChannels;
    property ChannelArrayVariable: TChannelArrayVariable read GetChannelArrayVariable;
    property SeriesVariable: TSeriesVariable read GetSeriesVariable;
  published
    property m:Integer read FMark write FMark;
    property l:Integer read FLeftMark write FLeftMark;
    property r:Integer read FRightMark write FRightMark;
    property n:Integer read FSeriesMemberSize write SetSeriesMemberSize;
    property nPeaks: Integer       index 0 read GetnPeaks write SetnPeaks;
    property nSeriesPeaks: Integer index 1 read GetnPeaks write SetnPeaks;
  end;

  TChannelClass = class of TChannel;

  TChannels = class(TVariabledPersistent)
  private
    FChannel: array of TChannel;
    FMaxPoints: Integer;
    FLeftMark: Integer;
    FMark: Integer;
    FRightMark: Integer;
    FStart: Double;
    FFinish: Double;
    FFileName: string;
    function GetChannel(Index: Integer): TChannel;
    function GetVisible(Index: Integer): Boolean;
    procedure SetVisible(Index: Integer; const Value: Boolean);
    function GetnChannels: Integer;
    function GetCount(Index: Integer): Integer;
    procedure SetMaxPoints(const Value: Integer);
    function GetDataType(Index: Integer): TDataType;
    function GetDataSourceInfo(Index: Integer): string;
    procedure SetDataSourceInfo(Index: Integer; const Value: string);
    function GetDataSourceName(Index: Integer): string;
    procedure SetDataSourceName(Index: Integer; const Value: string);
    procedure SetCount(Index: Integer; const Value: Integer);
    function GetInvert(Index: Integer): Boolean;
    procedure SetInvert(Index: Integer; const Value: Boolean);
    procedure SetnCollected(const Value: Integer);
    procedure FindUniqueName(c: Integer);
    procedure SetFileName(const Value: string);
  public
    PlotChanged: Boolean; { Cleared on next plot }
    DataChanged: Boolean; { Cleared on save }
    nCollected: Integer;
    StartTime, EndTime: TDateTime;
    Comments: string;
    ChannelClass: TChannelClass;
    AbortOffEnd: Boolean; { Set to raise EAbort if accessing array element above higher bound }

    constructor Create; overload; virtual;
    constructor Create(ConfigFile:TCustomIniFile); overload;
    destructor Destroy; override;
    procedure Changed(CountChanged:Boolean);

    procedure SaveToFile(AFileName:string; ItemSeparator: Char = ' ');
    procedure SaveToTextFile(var F: Text; FileName: string; ItemSeparator: Char = ' '; All: Boolean = True; First: Integer = 0; Last: Integer = High(Integer); RLE: Boolean = True);
    function LoadFromStream(Stream: TStream; AFileName: string; LoadSeries:Boolean = True):string;
    function LoadFromDriveStream(Stream: TStream; AFileName: string; LoadSeries:Boolean = True):string;
    function LoadFromFile(FileName: string; LoadSeries:Boolean = True):string;
    procedure Empty;
    procedure SetnChannels(Value: Integer; ConfigFile: TCustomIniFile);
    function ChannelWithName(AName: string):Integer;
    procedure Compact(n:Integer; Average: Boolean);
    function DeleteChannel(c:Integer; Keep:Boolean = False):TChannel;
    procedure NormalizeNcollected;
    procedure Reverse(ReversePeaks:Boolean);
    procedure AddChannel(AChannel:TChannel; After: Integer = High(Integer));
    function AddDoubleChannel: TChannel;
    procedure AddVariables(Lex: TLex; Prefix: string); override;

    procedure ReadConfig(ConfigFile:TCustomIniFile);
    procedure WriteConfig(ConfigFile:TCustomIniFile);

    property nChannels: Integer read GetnChannels;
    property Visible[Index:Integer]:Boolean read GetVisible write SetVisible;
    property Invert[Index:Integer]:Boolean read GetInvert write SetInvert;
    property Count[Index:Integer]:Integer read GetCount write SetCount;
    property DataType[Index:Integer]:TDataType read GetDataType;
    property DataSourceName[Index:Integer]:string read GetDataSourceName write SetDataSourceName;
    property DataSourceInfo[Index:Integer]:string read GetDataSourceInfo write SetDataSourceInfo;
    property MaxPoints: Integer read FMaxPoints write SetMaxPoints;
    property Channel[Index:Integer]:TChannel read GetChannel; default;
    property FileName: string read FFileName write SetFileName;
  published
    property m:Integer read FMark write FMark;
    property l:Integer read FLeftMark write FLeftMark;
    property r:Integer read FRightMark write FRightMark;
    property Start:Double read FStart write FStart;
    property Finish:Double read FFinish write FFinish;
    property n:Integer read nCollected write SetnCollected;
  end;

function InvertSeries(var Buffer; DataType: TDataType; Count: Integer):Int64;
function IsDriveStream(Stream: TStream):Boolean;

type
  TExternalLoader = function (Channels: TChannels; Stream: TStream; AFileName: string):string;
var
  ExternalLoader: TExternalLoader = nil;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Math, Huge, TextAssign, TextStrm,
     Units, Sysinfo, UCompat;

function QuoteForCSV(Value: string):string;
var
  P: Integer;
begin
  for p := 1 to Length(Value) do
    if Value[p] in [#9,',','"'] then begin
      Result := QuotedStr(Value);
      Exit;
    end;
  Result := Value;
end;

function ReadDataType(ConfigFile:TCustomIniFile; const Section, Ident: string; Default: TDataType):TDataType;
var
  s: String;
begin
  s := ConfigFile.ReadString(Section,Ident,'');
  if (s = '') or not TryStrToDataType(s,Result) then
    Result := Default;
end;

procedure WriteBytes(var F: TextFile; var Buffer; Count: Longint);
var
  P, PB: PAnsiChar;
  nBuffer: Longint;
  Temp: Integer;
begin
  PB := @Buffer;
  with TTextRec(F) do begin
    repeat
      nBuffer := Min(BufSize-BufPos,Count);
      P := PAnsiChar(BufPtr) + BufPos;
      Move(PB^, P^, nBuffer);
      Inc(BufPos, nBuffer);
      Dec(Count, nBuffer);
      Inc(PB, nBuffer);
      if Count > 0 then begin
        { Not enough space in buffer, so empty }
        Temp := TTextIOFunc(InOutFunc)(TTextRec(F));
        if Temp <> 0 then begin
          SetInOutRes(Temp);
          Exit;
        end;
      end;
    until Count = 0;
  end;
end;

{ TChannels }

function TChannels.GetChannel(Index: Integer): TChannel;
begin
  Result := FChannel[Index];
end;

function TChannels.GetCount(Index: Integer): Integer;
begin
  Result := FChannel[Index].Count;
end;

function TChannels.GetDataSourceInfo(Index: Integer): string;
begin
  Result := FChannel[Index].DataSourceInfo;
end;

function TChannels.GetDataSourceName(Index: Integer): string;
begin
  Result := FChannel[Index].DataSourceName;
end;

function TChannels.GetDataType(Index: Integer): TDataType;
begin
  Result := FChannel[Index].DataType;
end;

function TChannels.GetInvert(Index: Integer): Boolean;
begin
  Result := FChannel[Index].Invert;
end;

function TChannels.GetnChannels: Integer;
begin
  Result := Length(FChannel);
end;

function TChannels.GetVisible(Index: Integer): Boolean;
begin
  Result := FChannel[Index].Visible;
end;

procedure TChannels.SetCount(Index: Integer; const Value: Integer);
begin
  FChannel[Index].Count := Value;
end;

procedure TChannels.SetDataSourceInfo(Index: Integer; const Value: string);
begin
  FChannel[Index].DataSourceInfo := Value;
end;

procedure TChannels.SetDataSourceName(Index: Integer; const Value: string);
begin
  FChannel[Index].DataSourceName := Value;
end;

procedure TChannels.SetFileName(const Value: string);
var
  S, E: string;
begin
  FFileName := Value;
  S := ExtractFileName(FileName);
  E := ExtractFileExt(FileName);
  if Length(E) > 0 then
    Delete(S,Length(S)-Length(E)+1,Length(E));
  Name := S;
end;

procedure TChannels.SetInvert(Index: Integer; const Value: Boolean);
begin
  FChannel[Index].Invert := Value;
end;

procedure TChannels.SetMaxPoints(const Value: Integer);
var
  i: Integer;
begin
  FMaxPoints := Value;
  for i := 0 to nChannels - 1 do
    FChannel[i].Change(DataType[i],Value,False);
  if nCollected > MaxPoints then
    nCollected := MaxPoints;
end;

procedure TChannels.FindUniqueName(c: Integer);
  function DefName(c:Integer):string;
  begin
    if c >= 8 then
      Result := 'c' + IntToStr(c+1)
    else
      Result := char(Ord('a')+c);
  end;
var
  nName: Integer;
  Duplicate: Boolean;
begin
  if Channel[c].Name = '' then
    Channel[c].Name := DefName(c);
  nName := c;
  repeat
    Duplicate := ChannelWithName(Channel[c].Name) <> c;
    Inc(nName);
    if Duplicate then
      Channel[c].Name := DefName(nName);
  until not Duplicate;
end;

procedure TChannels.SetnChannels(Value: Integer; ConfigFile: TCustomIniFile);
var
  OldnChannels, i: Integer;
begin
  OldnChannels := nChannels;
  if Value > nChannels then begin
    SetLength(FChannel,Value);
    for i := OldnChannels to nChannels - 1 do begin
      FChannel[i] := ChannelClass.Create(Self);
      FChannel[i].FNumber := i;
      if Assigned(ConfigFile) then
        FChannel[i].ReadConfig(MaxPoints,ConfigFile)
      else begin
        FChannel[i].Change(tyNone,0,False);
        Visible[i] := True;
        Invert[i] := False;
      end;
      FindUniqueName(i);
    end;
  end else begin
    for i := Value to nChannels-1 do
      FChannel[i].Free;
    SetLength(FChannel,Value);
  end;
end;

procedure TChannels.SetnCollected(const Value: Integer);
var
  i, v: Integer;
begin
  v := Max(Value,0);
  MaxPoints := Max(MaxPoints,v);
  for i := 0 to nChannels - 1 do
    Channel[i].Count := v;
  nCollected := v;
end;

procedure TChannels.SetVisible(Index: Integer; const Value: Boolean);
begin
  FChannel[Index].Visible := Value;
end;

procedure TChannels.AddChannel(AChannel: TChannel; After: Integer);
var
  i, c: Integer;
begin
  if (After >= -1) and (After < nChannels) then
    c := After+1
  else
    c := nChannels;
  SetLength(FChannel, nChannels+1);
  for i := nChannels - 1 downto c+1 do
    FChannel[i] := FChannel[i-1];
  FChannel[c] := AChannel;
  AChannel.FChannels := Self;
  for i := c to nChannels - 1 do
    FChannel[i].FNumber := i;
  FindUniqueName(c);
  Changed(True);
end;

function TChannels.AddDoubleChannel: TChannel;
begin
  SetnChannels(nChannels + 1,nil);
  Result := Channel[nChannels-1];
  Result.Change(tyDouble,MaxPoints,False);
  Result.Count := nCollected;
end;

procedure TChannels.AddVariables(Lex: TLex; Prefix: string);
var
  i: Integer;
begin
  inherited;
  for i := 0 to nChannels - 1 do
    with Channel[i] do
      AddVariables(Lex,Prefix+Name+'.');
end;

procedure TChannels.Changed(CountChanged: Boolean);
var
  c: Integer;
begin
  DataChanged := True;
  PlotChanged := True;
  if CountChanged then begin
    nCollected := 0;
    for c := 0 to nChannels - 1 do
      nCollected := Max(nCollected,Channel[c].Count);
  end;
end;

function TChannels.ChannelWithName(AName: string): Integer;
var
  i: Integer;
begin
  for i := 0 to nChannels - 1 do
    if CompareText(Channel[i].Name,AName) = 0 then begin
      Result := i;
      Exit;
    end;
  Result := -1;
end;

procedure TChannels.Compact(n: Integer; Average: Boolean);
var
  i: Integer;
begin
  for i := 0 to nChannels - 1 do
    Channel[i].Compact(n, Average);
  MaxPoints := nCollected;
end;

constructor TChannels.Create(ConfigFile: TCustomIniFile);
begin
  Create;
  ReadConfig(ConfigFile);
  DataChanged := False;
end;

constructor TChannels.Create;
begin
  ChannelClass := TChannel;
  inherited;
end;

function TChannels.DeleteChannel(c:Integer; Keep:Boolean = False):TChannel;
var
  i: Integer;
begin
  if Keep then begin
    Result := Channel[c];
    Result.FChannels := nil;
    Result.FNumber := -1;
  end else begin
    Channel[c].Free;
    Result := nil;
  end;
  for i := c to nChannels - 2 do begin
    FChannel[i] := FChannel[i+1];
    Channel[i].FNumber := i;
  end;
  SetLength(FChannel, nChannels-1);
end;

destructor TChannels.Destroy;
begin
  SetnChannels(0,nil);
  inherited;
end;

procedure TChannels.Empty;
var
  I: Integer;
begin
  for I := 0 to nChannels - 1 do
    with FChannel[I] do begin
      Count := 0;
      DataSourceInfo := '';
    end;
  nCollected := 0;
  DataChanged := False;
end;

function TChannels.LoadFromFile(FileName: string; LoadSeries:Boolean):string;
var
  Stream: TStream;
begin
  Stream := nil;
  try
    if FileName = '' then begin
      Stream := THugeMemoryStream.Create;
      THugeMemoryStream(Stream).SetFromClipboard({$IFNDEF MSWINDOWS}1{$ELSE}CF_TEXT{$ENDIF}); { TODO: Remove bodge }
    end else
      Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
    Result := LoadFromStream(Stream,FileName,LoadSeries);
  finally
    Stream.Free;
  end;
end;

function IsDriveStream(Stream: TStream):Boolean;
const
  DriveMarker = 'Written by Drive ';
var
  Buf: string;
  Position: Int64;
  n: Integer;
begin
  SetLength(Buf,Length(DriveMarker));
  Position := Stream.Position;
  n := Stream.Read(Buf[1],Length(DriveMarker));
  Stream.Position := Position;
  Result := (n = Length(DriveMarker)) and (Buf = DriveMarker);
end;

function TChannels.LoadFromStream(Stream: TStream; AFileName: string; LoadSeries:Boolean):string;
begin
  if Assigned(ExternalLoader) and not IsDriveStream(Stream) then
    Result := ExternalLoader(Self, Stream, AFileName)
  else
    Result := LoadFromDriveStream(Stream,AFileName, LoadSeries)
end;

procedure TChannels.NormalizeNcollected;
var
  c, MxC: Integer;
begin
  MxC := 0;
  for c := 0 to nChannels - 1 do
    MxC := Max(MxC,Channel[c].Count);
  nCollected := MxC;
end;

procedure TChannels.ReadConfig(ConfigFile: TCustomIniFile);
var
  i, nc: Integer;
begin
  MaxPoints := ConfigFile.ReadInteger('Channels','MaxPoints',32000);
  nc := Max(ConfigFile.ReadInteger('Channels','Number',4),1);
  for i := 0 to Min(nChannels,nc)-1 do
    Channel[i].ReadConfig(MaxPoints,ConfigFile);
  SetnChannels(nc,ConfigFile);
  for i := 0 to nChannels - 1 do
    FindUniqueName(i);
end;

procedure TChannels.Reverse(ReversePeaks:Boolean);
var
  c: Integer;
begin
  for c := 0 to nChannels - 1 do
    Channel[c].Reverse(ReversePeaks);
end;

procedure TChannels.WriteConfig(ConfigFile: TCustomIniFile);
var
  i: Integer;
  s: string;
begin
  ConfigFile.WriteInteger('Channels','Number',nChannels);
  ConfigFile.WriteInteger('Channels','MaxPoints',MaxPoints);
  for i := 0 to nChannels-1 do
    with Channel[i] do begin
      s := 'Channel'+IntToStr(i+1);
      ConfigFile.WriteString(s,'Name',Name);
      ConfigFile.WriteBool(s,'Visible',Visible);
      ConfigFile.WriteBool(s,'Invert',Invert);
      ConfigFile.WriteString(s,'Type',DataTypeName(DataType));
      ConfigFile.WriteString(s,'Source',DataSourceName);
    end;
end;

function TChannels.LoadFromDriveStream(Stream: TStream; AFileName: string; LoadSeries:Boolean): string;
var
  S: TTextStream;
  FirstLine, Ty, Item: string;
  AItem: AnsiString;
  DT: TDateTime;
  Da: TDataType;
  nPoints: Integer;
  ColDest: array of Integer;
  i, Ch, Col, Line, MxP, nSeries, PeakIndex: Integer;
  D: Double;
  P: PAnsiChar;
  Binary, Minus, RLE: Boolean;
  Card: Cardinal;
  procedure Cancel(ErrMsg: string);
  begin
    SetnChannels(0,nil);
    Result := ErrMsg;
  end;
begin
  Result := '';
  S := TTextStream.Create(Stream);
  try
    S.ItemSeparators := [isTab,isComma];
    S.FlowToNextLine := False;
    S.FlowToNextRecord := False;
    S.AllowDefaults := False;
    S.Comments := False;
    S.IncludeSpaces := True;
    { Header info }
    S.NextStringVar(FirstLine);
    if AFileName = '' then begin
      i := Pos(' - to - ',FirstLine);
      if i > 0 then
        AFileName := Copy(FirstLine, i+8, Length(FirstLine));
    end;
    nPoints := 1000; { Default if missing from header; not essential anyway }
    while not S.AtEndOfFile do begin
      S.NextLine;
      if not S.GetFreeItem then
        Continue;
      Item := S.Token;
      if Item = 'Start Time' then begin
        S.NextStringVar(Item);
        if TryStrToDateTime(Item,DT) then
          StartTime := DT;
      end else if Item = 'End Time' then begin
        S.NextStringVar(Item);
        if TryStrToDateTime(Item,DT) then
          EndTime := DT;
      end else if Item = 'Start' then begin
        S.NextDoubleVar(D);
        Start := D;
      end else if Item = 'Finish' then begin
        S.NextDoubleVar(D);
        Finish := D;
      end else if Item = 'nPoints' then
        nPoints := S.NextInteger(1000)
      else if Item = '*Channels*' then
        Break
      else begin
        Cancel('Unknown header item:'+S.Token);
        Exit;
      end;
    end;
    { Read Channel info }
    SetnChannels(0,nil);
    nSeries := 0;
    while not S.AtEndOfFile do begin
      S.NextLine;
      if not S.GetFreeItem then
        Continue;
      Item := S.Token;
      if (Item <> '') and (Item[1] = '*') then
        Break;
      S.NextStringVar(Ty);
      if not TryStrToDataType(Ty,Da) then begin
        Cancel('Unknown data type:'+Ty);
        Exit;
      end;
      SetnChannels(nChannels+1,nil);
      with Channel[nChannels-1] do begin
        Change(Da,nPoints,False);
        if IsSeries(Da) then
          Inc(nSeries);
        if Item <> '' then
          Name := Item;
        DataSourceName := S.NextString('');
        DataSourceInfo := S.NextString('');
        Item := S.NextString('');
        Invert := CompareText(Item,'Inverted') = 0;
      end;
    end;
    Comments := '';
    if Item = '*Comments*' then
      repeat
        S.NextLine;
        Item := S.RestOfLine;
        if Item = '*Start of Data*' then
          Break
        else begin
          if (Item <> '') and (Item[1] = '#') then
            Delete(Item,1,1);
          if Comments <> '' then
            Comments := Comments + sLineBreak;
          Comments := Comments + Item;
        end;
      until S.AtEndOfFile;
    { allow for columns out of order }
    SetLength(ColDest,nChannels);
    for i := 0 to nChannels - 1 do
      ColDest[i] := i;
    { Read Column names }
    S.NextLine;
    Col := 0;
    S.GetFreeItem;
    Binary := S.Token = '*Binary*';
    if not Binary then
      repeat
        Item := S.Token;
        i := ChannelWithName(Item);
        if i >= 0 then
          ColDest[Col] := i
        else begin
          Cancel('Unknown channel name:'+Item);
          Exit;
        end;
        Inc(Col);
      until not S.GetFreeItem;
    { Now the actual data }
    Line := 0;
    if Binary then begin
      S.NextLine;
      S.NextStringVar(Item);
      repeat
        Ch := ChannelWithName(Item);
        with Channel[Ch] do begin
          S.NextIntegerVar(i);
          Count := i;
          S.NextLine;
          Change(DataType,Count,False);
          case DataType of
            tyInt16:
              S.ReadBuffer(Data16[0],Count*SizeOf(SmallInt));
            tyInteger, tySingle, tyShorts, tySmalls:
              S.ReadBuffer(Data32[0],Count*SizeOf(Integer));
            tyDouble, tyInt64,tyIntegers:
              S.ReadBuffer(Data64[0],Count*SizeOf(Int64));
          end;
        end;
        Item := S.NextString('');
      until (Item = '') or (Item = '*Series*') or (Item = '*BinaryPeaks*') or (Item = '*Peaks*');
    end else
      while not S.AtEndOfFile do begin
        S.NextLine;
        Col := 0;
        if not S.GetFreeItem then
          Break;
        if (S.Token <> '') and not (S.Token[1] in ['0'..'9','-','+','.']) then
          Break;
        while not S.AtEndOfFile do begin
          with Channel[ColDest[Col]] do begin
            if Line >= MaxPoints then
              Change(DataType,MaxPoints+10000,False);
            if Trim(S.Token) <> '' then
              Text[Line] := S.Token;
            Count := Line+1;
          end;
          Inc(Col);
          if not S.GetFreeItem then
            Break;
        end;
        Inc(Line);
      end;
    { Read peak lists, if any }
    while not S.AtEndOfFile do begin
      Binary := S.Token = '*BinaryPeaks*';
      if not (Binary or (S.Token = '*Peaks*')) then
        Break;
      S.GetFreeItem;
      PeakIndex := 0;
      if S.Token = '*Series*' then begin
        PeakIndex := 1;
        S.GetFreeItem;
      end;
      Ch := ChannelWithName(S.Token);
      if Ch < 0 then begin
        Cancel('Unknown channel name:'+S.Token);
        Exit;
      end;
      S.NextIntegerVar(Line);
      with Channel[Ch] do begin
        SetnPeaks(PeakIndex,Line);
        if Binary then begin
          S.NextLine;
          for i := 0 to 3 do
            S.ReadBuffer(Peaks[PeakIndex][i][0],Line*SizeOf(Double));
          S.GetFreeItem;
        end else begin
          Line := 0;
          while not S.AtEndOfFile do begin
            S.NextLine;
            if not S.GetFreeItem then
              Continue;
            if not ((S.Token = '') or S.IsRealToken(D)) then
              Break;
            for i := 0 to 3 do begin
              if S.Token = '' then
                SetNan(Peaks[PeakIndex][i][Line])
              else
                Peaks[PeakIndex][i][Line] := S.RealToken;
              S.GetFreeItem;
            end;
            Inc(Line)
          end;
          SetnPeaks(PeakIndex,Line);
        end;
      end;
    end;
    { Now read series data }
    if LoadSeries and (nSeries > 0) then begin
      while not S.AtEndOfFile do begin
        Binary := S.Token = '*Series*';
        if Binary then
          S.GetFreeItem;
        RLE := S.Token = '*RLE*';
        if RLE then
          S.GetFreeItem;
        Ch := ChannelWithName(S.Token);
        if Ch < 0 then begin
          Cancel('Unknown channel name:'+S.Token);
          Exit;
        end;
        if not IsSeries(Channel[Ch].DataType) then begin
          Cancel('Not a series:'+S.Token);
          Exit;
        end;
        S.GetFreeItem;
        with Channel[Ch] do
          if Binary then begin
            S.NextIntegerVar(i);
            Count := i;
            S.NextIntegerVar(i);
            SeriesMemberSize := i;
            S.NextLine;
            if RLE then begin
              Item := ReadRLE(S);
              if Item <> '' then begin
                Cancel(Item);
                Break;
              end;
            end else
              for i := 0 to Count - 1 do
                S.ReadBuffer(Series[i][0],SeriesMemberSize*DataTypeSize[DataType]);
            if not S.GetFreeItem then
              Break;
          end else begin
            Line := 0;
            Mxp := 0;
            while not S.AtEndOfFile do begin
              Col := 0;
              S.NextLine;
              if not S.GetFreeItem then
                Continue; { Skip blank lines }
              if not S.IsIntegerToken(i) then
                Break; { End of series }
              AItem := S.RestOfLine;
              P := @AItem[1];
              repeat
                if Length(Series) <= Line then
                  SetLength(Series,Line+1000);
                if Length(Series[Line]) <= Col*DataTypeSize[DataType] then
                  SetLength(Series[Line],(Col+100)*DataTypeSize[DataType]);
                SeriesValue[Line,Col] := i;
                Inc(Col);
                if P^ = #0 then
                  Break;
                if P^ = ',' then
                  Inc(P);
                if P^ = '-' then begin
                  Minus := True;
                  Inc(P);
                end else
                  Minus := False;
                if not (P^ in ['0'..'9']) then begin
                  Cancel(Format('Unexpected character in Line %d:%c',[S.Line,P^]));
                  Exit;
                end;
                Card := 0;
                repeat
                  Card := Card*10 + Ord(P^) - Ord('0');
                  Inc(P);
                until not (P^ in ['0'..'9']);
                if Minus then
                  i := -Card
                else
                  i := Card;
              until False;
              if Col > 0 then begin
                if Col > SeriesMemberSize then
                  SeriesMemberSize := Col;
                Mxp := Max(MxP,Col);
                Inc(Line);
              end;
            end;
            SetLength(Series,Line);
            Count := Max(Line,Count);
            FSeriesMemberSize := MxP;
          end;
      end;
    end;
    { Fixup Count, Maxpoints }
    MxP := 0;
    for i := 0 to nChannels - 1 do
      with Channel[i] do begin
        Change(DataType,Count,False);
        MxP := Max(MxP,MaxPoints);
      end;
    FMaxPoints := Mxp;
    nCollected := Mxp;
    FileName := AFileName;
  finally
    S.Free;
  end;
  DataChanged := False;
  PlotChanged := True;
end;

procedure TChannels.SaveToTextFile(var F: Text; FileName: string; ItemSeparator: Char; All: Boolean; First, Last: Integer; RLE: Boolean);
var
  i, j, c, np, PeakIndex: Integer;
  WriteBuffer: array [0..8191] of Byte;
  Binary: Boolean;
  nBuf: string[16];
  s: string;
  P: PAnsiChar;
  Abuf: AnsiString;
  function Convert(v: Integer):PAnsiChar;
  var
    i: Cardinal;
  begin
    nBuf[16] := #0;
    Result := @nBuf[16];
    if v < 0 then
      i := -v
    else
      i := v;
    repeat
      Dec(Result);
      Result^ := AnsiChar(Ord('0') + (i mod 10));
      i := i div 10;
    until i = 0;
    if v < 0 then begin
      Dec(Result);
      Result^ := '-';
    end;
  end;
  function CountFor(i:Integer):Integer;
  begin
    Result := Min(Last,Channel[i].Count-1)-First+1;
  end;
  function Writing(i:Integer):Boolean;
  begin
    with Channel[i] do
      Result := (All or Visible) and (DataType <> tyNone) and (CountFor(i) > 0);
  end;
  function RLESmaller(c: Integer):Boolean;
  var
    Mi, Mx: Double;
  begin
    { TODO: Could make this more general, but limit it to event counted data style series for now }
    with Channel[c] do
      Result := RLE and (SeriesMemberSize < 32767) and Limits(Mi,Mx,First,Last) and (Mi >= 0) and (Mx / SeriesMemberSize < 10);
  end;
begin
  Binary := ItemSeparator = #0;
  if Binary then
    ItemSeparator := ',';
  First := Max(First,0);
  try
    np := 0;
    for c := 0 to nChannels-1 do
      if Writing(c) then
        np := Max(CountFor(c),np);
    Last := First + np-1;
    if np = 0 then
      Exit;
    SetTextBuf(F,WriteBuffer);
    Write(F,'Written by ',VersionInfo,' on ',ComputerInfo, ' at ', DateTimeToStr(Now));
    if FileName <> '' then
      Write(F,' - to - ',ExtractFileName(FileName));
    WriteLn(F);
    if StartTime <> 0 then
      WriteLn(F,'Start Time',ItemSeparator,QuoteForCSV(DateTimeToStr(StartTime)));
    if EndTime <> 0 then
      WriteLn(F,'End Time',ItemSeparator,QuoteForCSV(DateTimeToStr(EndTime)));
    if Start <> 0 then
      WriteLn(F,'Start',ItemSeparator,FloatToStr(Start));
    if Finish <> 0 then
      WriteLn(F,'Finish',ItemSeparator,FloatToStr(Finish));
    WriteLn(F,'nPoints',ItemSeparator,np);
    WriteLn(F,'*Channels*');
    for c := 0 to nChannels-1 do
      with Channel[c] do
        if Writing(c) then begin
          if Invert then
            s := 'Inverted'
          else
            s := '';
          WriteLn(F,QuoteForCSV(Name),ItemSeparator,DataTypeName(DataType),ItemSeparator,
                  QuoteForCSV(DataSourceName),ItemSeparator,QuoteForCSV(DataSourceInfo),ItemSeparator,s);
        end;
    { Remove trailing newline }
    if (Comments <> '') and (Comments[Length(Comments)] = #10) then
        Delete(Comments,Length(Comments),1);
    if (Comments <> '') and (Comments[Length(Comments)] = #13) then
        Delete(Comments,Length(Comments),1);
    if Comments <> '' then begin
      WriteLn(F,'*Comments*');
      { add a # add the start of each line, and replace end of comments marker }
      WriteLn(F,'#',StringReplace(StringReplace(Comments,'*Start of Data*','>Start of Data<',[rfReplaceAll]),#10,#10'#',[rfReplaceAll]));
    end;
    WriteLn(F,'*Start of Data*');
    if Binary then begin
      WriteLn(F,'*Binary*');
      for c := 0 to nChannels-1 do begin
        with Channel[c] do
          if Writing(c) then begin
            WriteLn(F,QuoteForCSV(Name),ItemSeparator,CountFor(c));
            case DataType of
              tyInt16:
                WriteBytes(F,Data16[First],CountFor(c)*SizeOf(SmallInt));
              tyInteger, tySingle, tyShorts, tySmalls:
                WriteBytes(F,Data32[First],CountFor(c)*SizeOf(Integer));
              tyDouble, tyInt64,tyIntegers:
                WriteBytes(F,Data64[First],CountFor(c)*SizeOf(Int64));
            end;
          end
      end
    end else begin
      { Column Header line }
      s := '';
      for c := 0 to nChannels-1 do
        with Channel[c] do
          if Writing(c) then begin
            if s <> '' then
              s := s + ItemSeparator;
            s := s + QuoteForCSV(Name);
          end;
      WriteLn(F,s);
      for i := First to Last do begin
        s := '';
        for c := 0 to nChannels-1 do
          with Channel[c] do
            if Writing(c) then begin
              if s <> '' then
                s := s + ItemSeparator;
              if i < Count then
                s := s + Text[i]
              else
                s := s + ' ';
            end;
        Writeln(F,s);
      end;
    end;
    { Write Peaks }
    for c := 0 to nChannels-1 do
      with Channel[c] do
        if Writing(c) then
          for PeakIndex := 0 to 1 do
            if NormalizeNpeaks(PeakIndex) > 0 then begin
              if PeakIndex > 0 then
                s := ItemSeparator+'*Series*'
              else
                s := '';
              np := GetnPeaks(PeakIndex);
              if Binary then begin
                WriteLn(F,'*BinaryPeaks*',s,ItemSeparator,QuoteForCSV(Name),ItemSeparator,np);
                for i := 0 to 3 do
                  WriteBytes(F,Peaks[PeakIndex][i][0],np*SizeOf(Double));
              end else begin
                WriteLn(F,'*Peaks*',s,ItemSeparator,QuoteForCSV(Name),ItemSeparator,np);
                for i := 0 to np-1 do begin
                  s := '';
                  for j := 0 to 3 do begin
                    if j <> 0 then
                      s := s + ',';
                    if not IsNan(Peaks[PeakIndex][j][i]) then
                      s := s + FloatToStr(Peaks[PeakIndex][j][i]);
                  end;
                  WriteLn(F,s);
                end;
              end;
            end;
    { Write series data }
    for c := 0 to nChannels-1 do
      with Channel[c] do
        if Writing(c) and IsSeries(DataType) then
          if Binary then
            if RLESmaller(c) then begin
              WriteLn(F,'*Series*',ItemSeparator,'*RLE*',ItemSeparator,QuoteForCSV(Name),ItemSeparator,DataTypeName(DataType),ItemSeparator,Count,ItemSeparator,SeriesMemberSize);
              Channel[c].WriteRLE(F,First,Last);
            end else begin
              WriteLn(F,'*Series*',ItemSeparator,QuoteForCSV(Name),ItemSeparator,DataTypeName(DataType),ItemSeparator,Count,ItemSeparator,SeriesMemberSize);
              for i := First to Last do
                WriteBytes(F,Series[i][0],SeriesMemberSize*DataTypeSize[DataType])
            end
          else begin
            WriteLn(F,QuoteForCSV(Name),ItemSeparator,DataTypeName(DataType));
            SetLength(Abuf,SeriesMemberSize*20); { TODO: too big }
            for i := First to Last do begin
              P := @Abuf[1];
              for j := 0 to SeriesMemberSize - 1 do begin
                if j > 0 then begin
                  P^ := ',';
                  Inc(P);
                end;
                P := StrECopy(P,Convert(SeriesValue[i,j]))
              end;
              WriteBytes(F,Abuf[1],P-@Abuf[1]);
              WriteLn(F);
            end;
          end;
  finally
    CloseFile(F);
  end;
end;

procedure TChannels.SaveToFile(AFileName: string; ItemSeparator: Char);
var
  F: Text;
  i: Integer;
begin
  if ItemSeparator = ' ' then begin
    ItemSeparator := ',';
    if CompareText(ExtractFileExt(AFileName),'.txt') = 0 then
      ItemSeparator := #9
    else if CompareText(ExtractFileExt(AFileName),'.dr') = 0 then
      ItemSeparator := #0
  end;
  {$IFDEF FPC}
    AssignFile(F,AFileName);
    Rewrite(F);
  {$ELSE}
    {$I-}
    AssignFile(F,AFileName);
    Rewrite(F);
    i := IOResult;
    if i <> 0 then
      RaiseLastOSError(i); { More usefull error message }
    {$I+}
  {$ENDIF}
  SaveToTextFile(F,FileName,ItemSeparator);
  FileName := AFileName;
end;

{ TChannel }

procedure TChannel.Compact(n: Integer; Average: Boolean);
var
  row, w, c, i, j: Integer;
  S: Extended;
begin
  row := 0;
  w := 0;
  while row < Count do begin
    c := Min(row+n,Count);
    S := Value[row];
    for i := row+1 to c - 1 do
      S := S + Value[i];
    if IsSeries(DataType) then begin
      Series[w] := Series[row];
      for i := row+1 to c - 1 do
        for j := 0 to SeriesMemberSize - 1 do
          SeriesValue[w,j] := SeriesValue[w,j] + SeriesValue[i,j];
      if Average then
        for j := 0 to SeriesMemberSize - 1 do
          SeriesValue[w,j] := Round(SeriesValue[w,j]/(c-row));
    end;
    Dec(c,row);
    Inc(row,n);
    if Average then
      Value[w] := S/c
    else if c <> n then
      Value[w] := S*n/c { Last point - conserve scaling }
    else
      Value[w] := S;
    Inc(w);
  end;
  Change(DataType,w,False);
end;

procedure TChannel.AddVariables(Lex: TLex; Prefix: string);
begin
  inherited;
  if Prefix = '' then
    Prefix := Name
  else
    Delete(Prefix,Length(Prefix),1); { Remove trailing '.' } 
  Lex.AddFunction(Prefix,ChannelArrayVariable);
  if IsSeries(DataType) then
    Lex.AddFunction(Prefix+'.series',SeriesVariable);
end;

function TChannel.Average(First, Last: Integer): Double;
var
  i, np: Integer;
begin
  Result := 0;
  np := Min(Last,Count-1) - Max(First,0) + 1;
  if np <= 0 then
    raise Exception.CreateFmt('Empty range for %s:%d %d',[Name,First,Last]);
  for i := Max(First,0) to Min(Last,Count-1) do
    Result := Result + Value[i];
  Result := Result / np;
end;

procedure TChannel.Change(NewDataType: TDataType; AMaxPoints: Integer; Preserve: Boolean);
var
  Modified, CountChanged: Boolean;
  SaveSeriesMemberSize, i, j, dir: Integer;
  OldDataType: TDataType;
begin
  Modified := DataType <> NewDataType;
  Preserve := Preserve and Modified; { No action needed if data type not changing }
  CountChanged := Count > AMaxPoints;
  Modified := Modified or CountChanged;
  Count := Min(Count,AMaxPoints);
  if Modified or (MaxPoints <> AMaxPoints) then
    case NewDataType of
      tyNone:
        begin
          Series := nil;
          Data16 := nil;
          Data32 := nil;
          Data64 := nil;
          FreeAndNil(Display);
        end;
      tyInt16:
        begin
          if Preserve then begin
            SetLength(Data16, AMaxPoints);
            for i := 0 to Count - 1 do
              Data16[i] := SmallInt(Int64Value(i));
          end;
          Data32 := nil;
          Data64 := nil;
          Series := nil;
          SetLength(Data16, AMaxPoints); { No-op if preserving }
          FreeAndNil(Display);
        end;
      tyShorts, tySmalls, tyInteger, tySingle:
        begin
          if Preserve then begin
            SetLength(Data32, AMaxPoints);
            if NewDataType = tySingle then
              for i := 0 to Count - 1 do
                PSingle(@Data32[i])^ := Value[i]
            else
              for i := 0 to Count - 1 do
                Data32[i] := Integer(Int64Value(i));
          end;
          Data16 := nil;
          Data64 := nil;
          if not IsSeries(NewDataType) then
            Series := nil;
          SetLength(Data32, AMaxPoints);
          FreeAndNil(Display);
        end;
      tyDouble, tyInt64, tyIntegers:
        begin
          if Preserve then begin
            SetLength(Data64, AMaxPoints);
            if NewDataType <> tyDouble then
              for i := 0 to Count - 1 do
                PInt64(@Data64[i])^ := Int64Value(i)
            else
              for i := 0 to Count - 1 do
                Data64[i] := Value[i];
          end;
          Data16 := nil;
          Data32 := nil;
          if not IsSeries(NewDataType) then
            Series := nil;
          SetLength(Data64, AMaxPoints);
          FreeAndNil(Display);
        end;
    end;
  OldDataType := DataType;
  DataType := NewDataType;
  if IsSeries(NewDataType) then begin
    SetLength(Series, AMaxPoints);
    SaveSeriesMemberSize := SeriesMemberSize;
    FSeriesMemberSize := -1; { Trigger refresh of array sizes }
    if Preserve and IsSeries(DataType) then begin
      if DataTypeSize[NewDataType] > DataTypeSize[OldDataType] then begin
        { New larger than old; copy starts from last point }
        SetSeriesMemberSize(SaveSeriesMemberSize); { And need size adjusting now }
        dir := -1;
      end else
        dir := 1;
      for i := 0 to Count - 1 do begin
        if dir < 0 then
          j := SaveSeriesMemberSize-1
        else
          j := 0;
        while (j >= 0) and (j < SaveSeriesMemberSize) do begin
          case OldDataType of
            tyShorts:   SeriesValue[i,j] := Series[i][J];
            tySmalls:   SeriesValue[i,j] := PSmallInt(@Series[i][J*2])^;
            tyIntegers: SeriesValue[i,j] := PInteger(@Series[i][J*4])^;
          end;
          Inc(j,dir);
        end;
      end;
    end;
    SetSeriesMemberSize(SaveSeriesMemberSize); { Shrink array if necessary;
                                                 No effect if size increasing, because size already set above }
  end;
  if Modified then
    Changed(CountChanged);
end;

procedure TChannel.Changed(CountChanged: Boolean);
begin
  if Assigned(Channels) then begin
    Channels.nCollected := Max(Count,Channels.nCollected);
    Channels.Changed(CountChanged);
  end;
end;

procedure TChannel.Deglitch(Point: Integer);
var
  i, P1, P2: Integer;
begin
  if (Point < 0) or (Point >= Count) or (Count = 1) then
    Exit;
  P1 := Point-1;
  P2 := Point+1;
  { For points at either end, just copy }
  if P2 >= Count then
    P2 := P1;
  if P1 < 0 then
    P1 := P2;
  Value[Point] := (Value[P1]+Value[P2])/2;
  if IsSeries(DataType) then
    for i := 0 to SeriesMemberSize - 1 do
      SeriesValue[Point,i] := (SeriesValue[P1,i]+SeriesValue[P2,i]) div 2;
end;

procedure TChannel.DeletePoints(First, nPoints: Integer);
var
  nItems, i: Integer;
begin
  if First < 0 then begin
    Dec(nPoints,First);
    First := 0;
  end;
  nPoints := Min(Count-First,nPoints);
  nItems := Count - (First+nPoints); { Number of items after deleted region }
  if nItems > 0 then begin
    case DataType of
      tyInt16:             Move(Data16[First+nPoints],Data16[First],nItems*SizeOf(Data16[0]));
      tyShorts, tySmalls,
      tyInteger, tySingle: Move(Data32[First+nPoints],Data32[First],nItems*SizeOf(Data32[0]));
      tyIntegers,
      tyDouble, tyInt64:   Move(Data64[First+nPoints],Data64[First],nItems*SizeOf(Data64[0]));
    end;
    if IsSeries(DataType) then
      for i := First to Count - nPoints - 1 do
        Series[i] := Series[i+nPoints];
  end;
  Change(DataType,Count-nPoints,False);
end;

procedure TChannel.InsertPoints(First, nPoints: Integer);
var
  nItems, i: Integer;
begin
  Change(DataType,Count+nPoints,False); { Note that this does not change count }
  nItems := Count - First;
  if nItems > 0 then begin
    case DataType of
      tyInt16:             Move(Data16[First],Data16[First+nPoints],nItems*SizeOf(Data16[0]));
      tyShorts, tySmalls,
      tyInteger, tySingle: Move(Data32[First],Data32[First+nPoints],nItems*SizeOf(Data32[0]));
      tyIntegers,
      tyDouble, tyInt64:   Move(Data64[First],Data64[First+nPoints],nItems*SizeOf(Data64[0]));
    end;
    if IsSeries(DataType) then
      for i := Count - 1 downto First do
        Series[i+nPoints] := Series[i];
  end;
  Count := Count + nPoints;
  if nPoints > 0 then
    Changed(True);
end;

constructor TChannel.Create(AChannels: TChannels);
begin
  inherited Create;
  FChannels := AChannels;
  SetLength(PersistentVariables,nVariables+8);
  PersistentVariables[nVariables-8] := TVectorVariable.Create('x',0,False);
  PersistentVariables[nVariables-7] := TVectorVariable.Create('y',0,False);
  PersistentVariables[nVariables-6] := TVectorVariable.Create('w',0,False);
  PersistentVariables[nVariables-5] := TVectorVariable.Create('oc',0,False);
  PersistentVariables[nVariables-4] := TVectorVariable.Create('sx',0,False);
  PersistentVariables[nVariables-3] := TVectorVariable.Create('sy',0,False);
  PersistentVariables[nVariables-2] := TVectorVariable.Create('sw',0,False);
  PersistentVariables[nVariables-1] := TVectorVariable.Create('soc',0,False);
  PrefixVariables := True;
end;

destructor TChannel.Destroy;
begin
  Display.Free;
  FChannelArrayVariable.Free;
  FSeriesVariable.Free;
  inherited;
end;

function TChannel.GetSeriesMemberSize(I: Integer): Integer;
begin
  Result := Length(Series[I]) div DataTypeSize[DataType];
end;

{$IFOPT R+}
  {$DEFINE RangeCheck}
{$ENDIF}

function InvertSeries(var Buffer; DataType: TDataType; Count: Integer):Int64;
var
  i: Integer;
  PS: PShortInt;
  PM: PSmallInt;
  PI: PInteger;
begin
  Result := 0;
  case DataType of
    tyShorts:
      begin
        PS:= @Buffer;
        for i := 0 to Count-1 do begin
          PS^ := not PS^;
          {$R-}
          Inc(Result,PS^);
          {$IFDEF RangeCheck}{$R+}{$ENDIF}
          Inc(PS);
        end;
      end;
    tySmalls:
      begin
        PM := @Buffer;
        for i := 0 to Count-1 do begin
          PM^ := not PM^;
          {$R-}
          Inc(Result,PM^);
          {$IFDEF RangeCheck}{$R+}{$ENDIF}
          Inc(PM);
        end;
      end;
    tyIntegers:
      begin
        PI := @Buffer;
        for i := 0 to Count-1 do begin
          PI^ := not PI^;
          {$R-}
          Inc(Result,PI^);
          {$IFDEF RangeCheck}{$R+}{$ENDIF}
          Inc(PI);
        end;
      end;
  end;
end;

function TChannel.GetSeriesValue(I, J: Integer): Integer;
begin
  case DataType of
    tyShorts:  Result := Series[i][J];
    tySmalls:  Result := PSmallInt(@Series[i][J*2])^ ;
    tyIntegers: Result := PInteger(@Series[i][J*4])^;
  else
    raise Exception.Create('Not series');
  end;
end;

function TChannel.GetSeriesVariable: TSeriesVariable;
begin
  if not Assigned(FSeriesVariable) then
    FSeriesVariable := TSeriesVariable.Create(Self);
  Result := FSeriesVariable;
end;

procedure TChannel.SetSeriesValue(I, J: Integer; const Value: Integer);
begin
  case DataType of
    tyShorts:  Series[i][J] := ShortInt(Value);
    tySmalls:  PSmallInt(@Series[i][J*2])^ := SmallInt(Value);
    tyIntegers: PInteger(@Series[i][J*4])^ := Value;
  else
    raise Exception.Create('Not series');
  end;
  Changed(False);
end;

function TChannel.GetText(i: Integer): string;
var
  Buffer: array [0..22] of Char;
  v: Extended;
begin
  case DataType of
    tyNone:     Result := '';
    tyInt16:    Result := IntToStr(Data16[i]);
    tyShorts, tySmalls, tyInteger:
                Result := IntToStr(Data32[i]);
    tySingle:   begin
                  v := PSingle(@Data32[i])^;
                  {$IFNDEF FPC}
                    {$IFDEF UNICODE}
                      Result := FloatToStr(v); { TODO -cUNICODE : Original code gives ambiguous overload }
                    {$ELSE}
                      SetString(Result, PChar(@Buffer[0]), FloatToText(@Buffer[0],v,fvExtended,ffGeneral,7,9999));
                    {$ENDIF}
                  {$ELSE}
                  { TODO: Check following for FPC }
                  Result := FloatToStr(v);
                  {$ENDIF}
                end;
    tyDouble:   Result := FloatToStr(Data64[i]);
    tyInt64, tyIntegers:
                Result := IntToStr(PInt64(@Data64[i])^);
  end;
end;

procedure TChannel.SetText(i: Integer; const Value: string);
begin
  SetExtendedValue(i,StrToFloat(Value));
end;

function TChannel.ReadRle(Stream: TStream):string;
var
  Buf: array of Byte;
  i, j, k, p, v, nv, mx: Integer;
begin
  Result := '';
  SetLength(Buf,SeriesMemberSize*DataTypeSize[DataType]+3); { + 1 is space for marker byte }
  mx := 0;
  for i := 0 to Count - 1 do begin
    Stream.ReadBuffer(Buf[0],3); { Count and first indicator }
    if Buf[2] = 0 then begin
      { Not RLE }
      mx := High(Integer);
      Stream.ReadBuffer(Series[i][0],SeriesMemberSize*DataTypeSize[DataType]);
    end else if Buf[2] < 253 then begin
      Stream.ReadBuffer(Buf[3],PWord(@Buf[0])^-1); { Count and first indicator }
      j := 0;
      p := 2;
      while j < SeriesMemberSize do begin
        nv := Buf[p];
        v := Buf[p+1];
        Inc(p,2);
        mx := Max(mx,v);
        for k := 1 to nv do begin
          SeriesValue[i,j] := v;
          Inc(j);
        end;
      end;
    end else begin
      Result := 'Unsupported format';
      Exit;
    end;
  end;
  if mx < High(ShortInt) then
    Change(tyShorts,MaxPoints,True)
  else if mx < High(SmallInt) then
    Change(tySmalls,MaxPoints,True);
end;

procedure TChannel.WriteRLE(var F: Text; First, Last: Integer);
var
  Buf: array of Byte;
  i, j, p, v, nv: Integer;
begin
  { Layout: 2 bytes - length of this series in file (excluding length)
    Then repeat: 1 byte count, 1 byte value
    First byte has special significance:
      0 => entire line not RLE; >=253 reserved
  }
  SetLength(Buf,SeriesMemberSize*DataTypeSize[DataType]+3); { + 1 is space for marker byte }
  for i := First to Last do begin
    p := 2;
    j := 0;
    while j < SeriesMemberSize do begin
      v := SeriesValue[i,j];
      if (v < 0) or (v > High(Byte)) then
        Break;
      Inc(j);
      nv := 1;
      while (nv < 252) and (j < SeriesMemberSize) and (SeriesValue[i,j] = v) do begin
        Inc(j);
        Inc(nv);
      end;
      Buf[p] := nv;
      Buf[p+1] := v;
      Inc(p,2);
      if p >= High(Buf)-1 then
        Break;
    end;
    if j < SeriesMemberSize then begin
      { Smaller to just write as is }
      p := Length(Buf)-2;
      Buf[2] := 0; { Non-rle marker }
      Move(Series[i][0],Buf[3],p-1);
    end;
    PWord(@Buf[0])^ := p;
    WriteBytes(F, Buf[0], p+2);
  end;
end;

procedure TChannel.SetSeriesMemberSize(const Value: Integer);
var
  i: Integer;
begin
  if SeriesMemberSize <> Value then begin
    FSeriesMemberSize := Value;
    if IsSeries(DataType) then
      for i := 0 to High(Series) do
        SetLength(Series[i],SeriesMemberSize*DataTypeSize[DataType]);
  end;
end;

function TChannel.GetValue(i: Integer): Extended;
begin
  case DataType of
    tyNone:     SetNan(Result);
    tyInt16:    Result := Data16[i];
    tyShorts, tySmalls, tyInteger:
                Result := Data32[i];
    tySingle:   Result := PSingle(@Data32[i])^;
    tyDouble:   Result := Data64[i];
    tyInt64, tyIntegers:
                Result := PInt64(@Data64[i])^;
  end;
end;

function TChannel.Int64Value(i: Integer): Int64;
begin
  case DataType of
    tyInt16:    Result := Data16[i];
    tyShorts, tySmalls, tyInteger:
                Result := Data32[i];
    tySingle:   Result := Round(PSingle(@Data32[i])^);
    tyDouble:   Result := Round(Data64[i]);
    tyInt64, tyIntegers:
                Result := PInt64(@Data64[i])^;
  end;
end;

procedure TChannel.InvertValues;
var
  i: Integer;
  PS: PSingle;
  PI: PInt64;
begin
  case DataType of
    tyInt16:   for i := 0 to Count - 1 do Data16[i] := not Data16[i];
    tyInteger: for i := 0 to Count - 1 do Data32[i] := not Data32[i];
    tySingle:  for i := 0 to Count - 1 do begin PS := @Data32[i]; PS^ := -PS^; end;
    tyDouble:  for i := 0 to Count - 1 do Data64[i] := -Data64[i];
    tyInt64:   for i := 0 to Count - 1 do begin PI := @Data64[i]; PI^ := not PI^; end;
    tyShorts, tySmalls, tyIntegers:
      for i := 0 to  Count - 1 do
        SetInt64Value(i,InvertSeries(Series[i][0],DataType,SeriesMemberSize));
  end;
  Changed(False);
end;

function TChannel.LastPeak(Index:Integer): Integer;
var
  i, j: Integer;
begin
  for i := GetnPeaks(Index)-1 downto 0 do
    for j := 0 to 2 do
      if not IsNan(Peaks[Index][j][i]) then begin
        Result := i;
        Exit;
      end;
  Result := -1;
end;

function TChannel.Limits(out Mi, Mx: Double; First, Last: Integer): Boolean;
var
  i: Integer;
  v: Double;
begin
  Mi := 0;
  Mx := 0;
  Result := Count > First;
  if not Result then
    Exit;
  Result := False;
  for i := First to Min(Last,Count-1) do begin
    v := Value[i];
    if IsNanOrInf(v) then
      Continue;
    if not Result then begin
      Mi := v;
      Mx := v;
      Result := True;
    end else begin
      Mi := Min(v,Mi);
      Mx := Max(v,Mx);
    end;
  end;
end;

function TChannel.MaxPoints: Integer;
begin
  case DataType of
    tyNone:   Result := 0;
    tyInt16:  Result := Length(Data16);
    tyInteger, tySingle:  Result := Length(Data32);
    tyDouble, tyInt64: Result := Length(Data64);
    tyShorts, tySmalls, tyIntegers:  Result := Length(Series);
  end;
end;

function TChannel.NormalizeNpeaks(Index:Integer): Integer;
var
  i, j: Integer;
  Empty: Boolean;
begin
  i := GetnPeaks(Index)-1;
  Empty := True;
  while (i >= 0) and Empty do begin
    for j := 0 to 2 do
      Empty := Empty and IsNan(Peaks[Index][j][i]);
    if Empty then
      Dec(i);
  end;
  SetnPeaks(Index, i+1);
  Result := GetnPeaks(Index);
end;

procedure TChannel.Paste(Source: TChannel; First, Last: Integer);
var
  PointsLost, i, j: Integer;
begin
  First := Max(First,0);
  Last := Max(Min(Last,Count-1),First-1);
  PointsLost := Last-First+1;
  if PointsLost > Source.Count then
    DeletePoints(First,PointsLost-Count)
  else if Source.Count > PointsLost then
    InsertPoints(First,Source.Count-PointsLost);
  for i := First to First+Source.Count - 1 do
    Value[i] := Source.Value[i-First];
  if IsSeries(DataType) and IsSeries(Source.DataType) then
  SeriesMemberSize := Max(SeriesMemberSize,Source.SeriesMemberSize);
  for i := First to First+Source.Count - 1 do
    for j := 0 to Source.SeriesMemberSize - 1 do
      SeriesValue[i,j] := Source.SeriesValue[i-First,j]
end;

procedure TChannel.ReadConfig(AMaxPoints: Integer; ConfigFile: TCustomIniFile);
var
  s, Na: string;
begin
  s := 'Channel'+IntToStr(Number+1);
  if Number > 7 then
    Na := 'c'+IntToStr(Number+1)
  else
    Na := char(Ord('a')+Number);
  Name := ConfigFile.ReadString(s,'Name',Na);
  Visible := ConfigFile.ReadBool(s,'Visible',True);
  Invert := ConfigFile.ReadBool(s,'Invert',False);
  if Number < 6 then
    Change(ReadDataType(ConfigFile,s,'Type',tyInteger),AMaxPoints,False)
  else
    Change(ReadDataType(ConfigFile,s,'Type',tyDouble),AMaxPoints,False);
  DataSourceName := ConfigFile.ReadString(s,'Source','');
end;

procedure TChannel.Reverse(ReversePeaks:Boolean);
var
  S: Smallint;
  I, V: Integer;
  B: Int64;
  A: TShortIntArray;
begin
  case DataType of
    tyInt16:
      for I := 0 to (Count-1) div 2 do begin
        S := Data16[I];
        Data16[I] := Data16[Count-1-I];
        Data16[Count-1-I] := S;
      end;
    tyInteger, tySingle, tyShorts, tySmalls:
      for I := 0 to (Count-1) div 2 do begin
        V := Data32[I];
        Data32[I] := Data32[Count-1-I];
        Data32[Count-1-I] := V;
      end;
    tyDouble, tyInt64, tyIntegers:
      for I := 0 to (Count-1) div 2 do begin
        B := PInt64(@Data64[I])^;
        PInt64(@Data64[I])^ := PInt64(@Data64[Count-1-I])^;
        PInt64(@Data64[Count-1-I])^ := B;
      end;
  end;
  if IsSeries(DataType) then
    for I := 0 to (Count-1) div 2 do begin
      A := Series[I];
      Series[I] := Series[Count-1-I];
      Series[Count-1-I] := A;
    end;
  if ReversePeaks then
    for i := 0 to nPeaks - 1 do
      if not IsNan(Peaks[0][0][i]) then
        Peaks[0][0][i] := Count - 1 - Peaks[0][0][i];
  Changed(False);
end;

procedure TChannel.SetName(const Value: string);
begin
  inherited;
  if Assigned(FChannelArrayVariable) then
    ChannelArrayVariable.Name := Name;
  if Assigned(FSeriesVariable) then
    SeriesVariable.Name := Value + '.series';
end;

procedure TChannel.SetnPeaks(Index: Integer; const Value: Integer);
var
  i: Integer;
begin
  for i := 0 to High(Peaks[Index]) do begin
    SetLength(Peaks[Index][i],Value);
    (PersistentVariables[nVariables-8+4*Index+i] as TVectorVariable).Values := Peaks[Index][i];
  end;
end;

procedure TChannel.SetInt64Value(i: Integer; Value: Int64);
begin
  case DataType of
    tyInt16:    Data16[i] := SmallInt(Value);
    tyInteger, tyShorts, tySmalls:  Data32[i] := Integer(Value);
    tySingle:   PSingle(@Data32[i])^ := Value;
    tyDouble:   Data64[i] := Value;
    tyInt64, tyIntegers:
                PInt64(@Data64[i])^ := Value;
  end;
  Changed(False);
end;

procedure TChannel.SetInt16Value(i: Integer; Value: Smallint);
begin
  case DataType of
    tyInt16:    Data16[i] := Value;
    tyInteger, tyShorts, tySmalls:  Data32[i] := Value;
    tySingle:   PSingle(@Data32[i])^ := Value;
    tyDouble:   Data64[i] := Value;
    tyInt64, tyIntegers:
                PInt64(@Data64[i])^ := Value;
  end;
  Changed(False);
end;

procedure TChannel.SetIntegerValue(i, Value: Integer);
begin
  case DataType of
    tyInt16:    Data16[i] := SmallInt(Value);
    tyInteger, tyShorts, tySmalls:
                Data32[i] := Value;
    tySingle:   PSingle(@Data32[i])^ := Value;
    tyDouble:   Data64[i] := Value;
    tyInt64, tyIntegers:
                PInt64(@Data64[i])^ := Value;
  end;
  Changed(False);
end;

procedure TChannel.SetExtendedValue(i: Integer; const Value: Extended);
begin
  case DataType of
    tyInt16:    Data16[i] := SmallInt(Round(Value));
    tyInteger, tyShorts, tySmalls:
                Data32[i] := Integer(Round(Value));
    tySingle:   PSingle(@Data32[i])^ := Value;
    tyDouble:   Data64[i] := Value;
    tyInt64, tyIntegers:
                PInt64(@Data64[i])^ := Round(Value); { Note Round returns an Int64 result }
  end;
  Changed(False);
end;

function TChannel.GetChannelArrayVariable: TChannelArrayVariable;
begin
  if not Assigned(FChannelArrayVariable) then
    FChannelArrayVariable := TChannelArrayVariable.Create(Self);
  Result := FChannelArrayVariable;
end;

function TChannel.GetnPeaks(Index:Integer): Integer;
begin
  Result := Length(Peaks[Index][0]);
end;

{ TChannelArrayVariable }

constructor TChannelArrayVariable.Create(AChannel: TChannel);
begin
  Channel := AChannel;
  inherited Create(Channel.Name, 1);
end;

function TChannelArrayVariable.LowerBound(Bound:Integer): Integer;
begin
  Result := 0;
end;

procedure TChannelArrayVariable.Store(var a:array of Double);
begin
  with Channel do
    if a[1] > -0.5 then
      if a[1] < Count+0.5 then
        Value[Round(a[1])] := a[0]
      else if Channels.AbortOffEnd then
        Abort;
end;

function TChannelArrayVariable.UpperBound(Bound:Integer): Integer;
begin
  with Channel do
    if Channels.AbortOffEnd then
      Result := Channels.MaxPoints-1
    else
      Result := Count-1;
end;

function TChannelArrayVariable.Value(var a: array of Double): Double;
begin
  Result := 0;
  with Channel do
    if a[0] > -0.5 then
      if a[0] < Count-0.5 then
        Result := Channel.Value[Round(a[0])]
      else if Channels.AbortOffEnd then
        Abort;
end;

{ TSeriesVariable }

constructor TSeriesVariable.Create(AChannel: TChannel);
begin
  Channel := AChannel;
  inherited Create(Channel.Name+'.series', 2);
end;

function TSeriesVariable.LowerBound(Bound:Integer): Integer;
begin
  Result := 0;
end;

procedure TSeriesVariable.Store(var a: array of Double);
var
  i: Integer;
begin
  with Channel do
    if (a[2] > -0.5) and (a[2] < Count-0.5) then begin
      i := Round(a[2]);
      if (a[1] > -0.5) and (a[1] < SeriesMemberSize-0.5) then
        Channel.SeriesValue[i,Round(a[1])] := Integer(Round(a[0]));
    end;
end;

function TSeriesVariable.UpperBound(Bound:Integer): Integer;
begin
  if Bound = 0 then
    Result := Channel.Count - 1
  else
    Result := Channel.SeriesMemberSize - 1;
end;

function TSeriesVariable.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  with Channel do
    if (a[1] > -0.5) and (a[1] < Count-0.5) then begin
      i := Round(a[1]);
      if (a[0] > -0.5) and (a[0] < SeriesMemberSize-0.5) then begin
        Result := Channel.SeriesValue[i,Round(a[0])];
        Exit;
      end;
    end;
  Result := 0;
end;

end.

