unit CheckImport;
{ Check an imported file yields the same .pgo file }

{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}

interface

uses Classes, TextStrm, Mixture;

function CheckImported(IMix, Mixture: TMixture; var nTried: Integer; Tolerance: Double; var Failures: Integer; Script: TStrings):Double;
function CheckCalculatedParameters(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer;
                                   Flags: Integer { 1: Allow parameters to be equal in magnitude but opposite sign }
                                   ):Double;


procedure ApplyScript(Mixture: TMixture; Script: TStrings);

implementation

uses SysUtils, Math, MathSup, UCompat, Worker, MNode, MNodeExpr;

function DiffNos(n1, n2, Tolerance: Double; var MaxErr: Double):Boolean;
var
  RelTol: Double;
begin
  if IsNanOrInf(n1) then begin
    Result := not IsNanOrInf(n2);
    Exit;
  end;
  if Abs(n1-n2) < Tolerance then begin
    MaxErr := Max(MaxErr, Abs(n1-n2));
    Result := False;
    Exit;
  end;
  { Try again on fractional difference }
  RelTol := Max(Abs(n1),Abs(n2))*Tolerance;
  Result := Abs(n1-n2) >= RelTol;
  if RelTol > Tolerance then
    MaxErr := Max(MaxErr, Abs(n1-n2)/Max(Abs(n1),Abs(n2)))
  else
    MaxErr := Max(MaxErr, Abs(n1-n2));
end;

function CheckImported(IMix, Mixture: TMixture;
                       var nTried: Integer; Tolerance: Double; var Failures: Integer; Script: TStrings):Double;
  procedure CompareNode(A, B: TMNode);
    procedure Fail(Name: string);
    begin
      Inc(Failures);
      if Failures > 0 then begin
        WriteLn(Log, Name, ' in ', A.FullHeirarchicName, ':', A.ClassName);
      end;
    end;
    function CheckStr(Name, sA, sB: string):Boolean;
    begin
      Result := sA = sB;
      if not Result then
        Fail(Format('%s difference: %s %s', [Name, sA, sB]));
    end;
    function CheckInt(Name: string; iA, iB: Integer):Boolean;
    begin
      Result := iA = iB;
      if not Result then
        Fail(Format('%s difference: %d %d', [Name, iA, iB]));
    end;
    procedure CheckNo(Name: string; n1, n2, tol: Double);
    var
      Msg: string;
      Bad: Boolean;
    begin
      if DiffNos(n1, n2, tol, Result) then
        Fail(Format('%s out of tolerance: %g %g %g (%g)', [Name, n1, n2, n1-n2, n1/n2]))
    end;
    procedure CheckArray(Name: string; A, B: TStringArrayArray);
    var
      i, j: Integer;
      n1, n2, Tol: Double;
    begin
      if not CheckInt(Name+' count', Length(A), Length(B)) then
        Exit;
      for i := 0 to High(A) do begin
        if not CheckInt(Name+' count '+IntToStr(i), Length(A[i]), Length(B[i])) then
          Exit;
        for j := 0 to High(A[i]) do
          if A[i][j] <> B[i][j] then begin
            if TryStrToFloat(A[i][j],n1) and TryStrToFloat(B[i][j],n2) then begin
              Tol := Max(Abs(n1)*Tolerance, Tolerance);
              if not DiffNos(n1,n2,Tol,Result) then
                Exit;
            end;
            Fail(Format('%s[%d][%d] difference: %s %s', [Name, i, j, A[i][j], B[i][j]]));
          end;
      end;
    end;
  var
    i: Integer;
    Tol: Double;
  begin
    if not CheckStr('Class', A.ClassName, B.ClassName) then
      Exit;
    CheckStr('Name', A.Name, B.Name);
    CheckStr('Comment', A.Comment, B.Comment);
    if CheckInt('nSettings', A.nSettings, B.nSettings) then
      for i := 1 to A.nSettings do begin
        CheckInt(A.SettingNames[i], A.Settings[i], B.Settings[i]);
        CheckStr('Setting name', A.SettingNames[i], B.SettingNames[i]);
      end;
    if CheckInt('nParameters', A.nParameters, B.nParameters) then
      for i := 1 to A.nParameters do begin
        Tol := Max(Abs(A.Parameters[i])*Tolerance, Tolerance);
        CheckNo(A.ParameterNames[i], A.Parameters[i], B.Parameters[i], Tol);
        CheckStr('Parameter name', A.ParameterNames[i], B.ParameterNames[i]);
      end;
    if CheckInt('Component Count', A.ComponentCount, B.ComponentCount) then
      for i := 0 to A.ComponentCount - 1 do
        CompareNode(A.Components[i] as TMNode, B.Components[i] as TMNode);
    if A is TMNodeWithArray then begin
      CheckArray('A', TMNodeWithArray(A).A, TMNodeWithArray(B).A);
      CheckArray('AF', TMNodeWithArray(A).AF, TMNodeWithArray(B).AF);
    end;
  end;
var
  E: Exception;
begin
  Result := 0;
  ApplyScript(IMix, Script);
  E := IMix.FixupProperties;
  if Assigned(E) then begin
    WriteLn(Log, 'Imxix:',E.Message);
    E.Free;
  end;
  E := Mixture.FixupProperties;
  if Assigned(E) then begin
    WriteLn(Log, 'Mixture:',E.Message);
    E.Free;
  end;
  { Copy some values across, as these will never be identical }
  IMix.Name := Mixture.Name;
  IMix.ShowEstUnc := Mixture.ShowEstUnc; { so .var files with correlations will compare correctly }
  CompareNode(Mixture, IMix);
  IMix.Free;
end;

procedure ApplyScript(Mixture: TMixture; Script: TStrings);
var
  i: Integer;
  s: string;
begin
  if Script = nil then
    Exit;
  for i := 0 to Script.Count-1 do begin
    s := Trim(Script[i]);
    if S = '' then
      Continue;
    if s[1] in [';', '/', '#'] then
      Continue;
    try
      Mixture.HandleExpression(s);
    except
      on E: Exception do
        raise EReportException.Create('Error on line '+IntToStr(i+1)+' of script:'+E.Message);
    end;
  end;
end;

function CheckCalculatedParameters(Mixture: TMixture; S: TTextStream; var nTried: Integer; Tolerance: Double; var Failures: Integer; Flags: Integer):Double;
  procedure Fail(Name: string);
  begin
    Inc(Failures);
    if Failures > 0 then begin
      WriteLn(Log, 'Line ', S.Line, ':', Name);
    end;
  end;
var
  m, ErrMsg: string;
  Molecule: TMolecule;
  Manifold: TManifold;
  MNode: TMNode;
  State: TState;
  viInfo: TValueInfo;
  V, Vc: Double;
  Tol: Double;
begin
  Result := 0;
  while not S.AtEndOfFile do begin
    if S.GetFreeItem then
      if S.Token <> 'Parameter' then
        Fail(S.Token)
      else begin
        Inc(nTried);
        S.NextStringVar(m);
        Molecule := Mixture.FindMolecule(m);
        if not Assigned(Molecule) then
          Molecule := Mixture.Species[0].Isotopomer[0]
        else
          S.NextStringVar(m);
        Manifold := Molecule.ManifoldWithName(m);
        if Assigned(Manifold) then
          S.NextStringVar(m)
        else
          Manifold := Molecule.Manifold[0];
        State := Manifold.StateWithName(m);
        if Assigned(State) then begin
          S.NextStringVar(m);
          viInfo := State.SubPropInfo(m,ErrMsg);
        end else begin
          viInfo := Manifold.SubPropInfo(m,ErrMsg);
          if not viInfo.IsSingleValue then
            viInfo := Mixture.SubPropInfo(m,ErrMsg);
          if not viInfo.IsSingleValue and (Manifold.nComponentsWithName(m,MNode) = 1) then begin
            S.NextStringVar(m);
            if m <> '=' then
              viInfo := MNode.SubPropInfo(m,ErrMsg);
          end;
        end;
        if not viInfo.IsSingleValue then
          Fail(m+' '+ErrMsg)
        else begin
          S.NextStringVar(m);
          if m <> '=' then
            Fail('Expecting =, found '+m);
          S.NextStringVar(m);
          V := S.RealToken;
          Tol := Max(Abs(V)*Tolerance,Tolerance);
          Vc := viInfo.GetValue;
          if ((Flags and 1) = 1) and ((Vc < 0) xor (V < 0)) then
            Vc := -Vc; {Ignore sign differences }
          if DiffNos(V, VC, Tol, Result) then
            Fail(Format('%s out of tolerance %g %g diff=%g',[viInfo.FullName,V,Vc,V-Vc]));
        end;
      end;
    S.NextLine;
  end;
end;

end.
