
unit ExoMol;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}
interface

uses SysUtils, TextStrm, MathSup, QuantumNumber, pgDOM, MNode, Mixture, Obs, DRintensity, ExternalList;

function ImportExoMolFile(S: TTextStream): TMixture;

type

  TExoMolBaseReader = class(TObservationsReader)
    StateCols, QnoCols: TIntegerArray;
    constructor Create; override;
    function HandleDirective(var Handled: Boolean):Boolean; override;
    procedure AddStates(AFileName: string);
    function ReadQnos(J: TQuantumNumber; var Q: array of Integer; const Quanta: string):Integer;
    procedure ReadState(var Manifold: TManifold; var State: TState; const Quanta: string);
  end;

  TExoMolReader = class(TExoMolBaseReader)
    function DoReadOneLine: Boolean; override;
  end;

  TExoMolState = record
    J: TQuantumNumber;
    Quanta: string;
  end;

  TExoMolTransBaseReader = class(TExoMolBaseReader)
    ExoMolTransReader: TExoMolTransBaseReader;
    States: array of TExoMolState;
    constructor Create; override;
  end;

  TExoMolTransReader = class(TExoMolTransBaseReader)
    UnitsChecked: Boolean;
    function DoReadOneLine: Boolean; override;
  end;

  TExoMolMap = record
    J: SmallInt;
    Index: SmallInt;
  end;

  TExoMolTrans = record
    UpperID, LowerID: Integer;
    A: Double;
  end;

  TExoMolManifold = class(TExternalManifold)
  public
    Map: array of TExoMolMap; { So that Eigenvalue can be worked out from State ID }
    procedure MapEv(StateID: Integer; out AIndex: Integer; out Ev: TEigenvalues);
    procedure LoadTransFile(const FileName: string);
    function ReadStatesFile(S: TTextStream):TQuantumNumber;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromTransFileClick(Sender: TObject);
    {$ENDIF}
  end;

  TExoMolTransitions = class(TTransitionMoments)
  public
    Trans: array of TExoMolTrans;
    procedure LoadTransFile(const FileName: string);
    procedure ReadTransStream(S: TTextStream);
    procedure AddToSpectrum(Spectrum: TSpectrum; Simulation: TSimulation; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean);
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromTransFileClick(Sender: TObject);
    {$ENDIF}
  end;

  TExoMolMolecule = class(TExternalMolecule)
  public
    procedure LoadTransFile(const FileName: string);
    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception;
                            Simulation: TSimulation;
                            J1Sel: Integer = Low(Integer); Sym1Sel: Integer = -1; J2Sel: Integer = Low(Integer); Sym2Sel: Integer = -1;
                            Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                            DeltaJ: TQuantumNumber = -10000;
                            TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                            State1Sel: TState = nil; State2Sel: TState = nil; IncludePerturbations: Boolean = True; IncludeZero: Boolean = False; RemoveSpins: Byte = 0):Boolean; override;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromTransFileClick(Sender: TObject);
    {$ENDIF}
  end;


implementation

uses {$IFNDEF CONSOLE} Forms, Dialogs, {$ENDIF}
     Math, Classes, Huge, UCompat, Units, Global, SimpleQnos, Worker;

function ImportExoMolFile(S: TTextStream): TMixture;
var
  Species: TSpecies;
  Molecule: TExoMolMolecule;
  Manifold: TExoMolManifold;
  Jmax: TQuantumNumber;
begin
  Result := TMixture.Create(nil);
  Result.Name := S.StreamName;
  Species := TSpecies.Create(Result);
  Species.Name := 'Species';
  Molecule := TExoMolMolecule.Create(Species);
  Molecule.Name := 'Molecule';
  Manifold := TExoMolManifold.Create(Molecule);
  Manifold.Name := 'Manifold';
  Manifold.Initial := True;
  Jmax := Manifold.ReadStatesFile(S);
  if Jmax > 0 then
    Species.Jmax := Jmax;
end;

{ TExoMolMolecule }

procedure TExoMolMolecule.LoadTransFile(const FileName: string);
var
  T: TExoMolTransitions;
  M: TExoMolManifold;
begin
  if FileName = '' then
    Exit;
  T := ChildWithClass(TExoMolTransitions) as TExoMolTransitions;
  if Assigned(T) then
    T.LoadTransFile(FileName)
  else begin
    M := ChildWithClass(TExoMolManifold) as TExoMolManifold;
    if Assigned(M) then
      M.LoadTransFile(FileName)
    else
      raise EReportException.Create('Must be single manifold or transitions object');
  end;
end;

{$IFNDEF CONSOLE}
function PromptForTransFile: string;
begin
  Result := '';
  with TOpenDialog.Create(Application) do
    try
      Filter := 'Transitions File (*.trans)|*.trans|All Files|'+AllFilesMask;
      Title := 'Load .trans file';
      if Execute then
        Result := FileName;
    finally
      Free;
    end;
end;

procedure TExoMolMolecule.AddHandlers;
begin
  inherited;
  if (nManifolds = 1) or (nTransitionMoments = 1) then
    AddHandler('Add .trans file...', 'Load transitions from .trans file', {$IFDEF FPC}@{$ENDIF}LoadFromTransFileClick);
end;

procedure TExoMolMolecule.LoadFromTransFileClick(Sender: TObject);
begin
  LoadTransFile(PromptForTransFile);
end;
{$ENDIF}

function TExoMolMolecule.AddToSpectrum(Spectrum: TSpectrum;
  PrintLevel: TPrintLevel; const Excitation: TExcitation;
  var Problem: Exception; Simulation: TSimulation; J1Sel, Sym1Sel, J2Sel,
  Sym2Sel: Integer; Manifold1Sel, Manifold2Sel: TManifold;
  DeltaJ: TQuantumNumber; TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment; State1Sel, State2Sel: TState;
  IncludePerturbations, IncludeZero: Boolean; RemoveSpins: Byte): Boolean;
var
  nInitial, nConnections, i, j: Integer;
  Manifold1, Manifold2: TManifold;
  Connection: TTransitionMoments;
begin
  Result := False;
  if Abundance = 0 then Exit;
  nInitial := 0;
  nConnections := 0;
  if J1Sel < 0 then
    J1Sel := Low(Integer);
  if J2Sel < 0 then
    J2Sel := Low(Integer);
  for i := 0 to nManifolds-1 do
    if (Manifold[i] = Manifold1Sel)
       or (not Assigned(Manifold1Sel) and Manifold[i].Initial ) then begin
      { For each manifold marked as initial }
      Manifold1 := Manifold[i];
      Inc(nInitial);
      for j := 0 to nManifolds-1 do
        if (Manifold[j] = Manifold2Sel)
           or (not Assigned(Manifold2Sel) and ((j >= i) or not Manifold[j].Initial)) then begin
          { For every other manifold, except those we have already started from }
          Manifold2 := Manifold[j];
          if Spectrum.IsDistribution then
            raise NotYet
          else
            Connection := TransitionMoments[i, j];
          if Assigned(Connection)
            and ((Connection = TransitionMomentsSel) or not Assigned(TransitionMomentsSel)) then begin
            Inc(nConnections);
            { TODO: Implement J selection, including DeltaJ  ; could use original logic with more work }
            if Connection is TExoMolTransitions then
               TExoMolTransitions(Connection).AddToSpectrum(Spectrum, Simulation, PrintLevel, Excitation, Problem, IncludeZero)
            else
              raise NotYet;
          end;
      end;
    end;
  if (nManifolds > 0) and (nInitial = 0) then
      raise EReportException.Create('No Initial States set for '+ Name);
  Result := nConnections > 0;
end;

{ TExoMolTransitions }

{$IFNDEF CONSOLE}
procedure TExoMolTransitions.AddHandlers;
begin
  inherited;
  AddHandler('Add .trans file...', 'Load transitions from .trans file', {$IFDEF FPC}@{$ENDIF}LoadFromTransFileClick);
end;

procedure TExoMolTransitions.LoadFromTransFileClick(Sender: TObject);
begin
  LoadTransFile(PromptForTransFile);
end;
{$ENDIF}

procedure TExoMolTransitions.AddToSpectrum(Spectrum: TSpectrum;
  Simulation: TSimulation; PrintLevel: TPrintLevel;
  const Excitation: TExcitation; var Problem: Exception;
  IncludeZero: Boolean);
var
  Line: TLine;
  wt: Double;
  i, t: Integer;
  ET: TEigenvalues;
  ManifoldU, ManifoldL: TExoMolManifold;
begin
  Line.Width := 0;
  Line.Colour := Colour;
  wt := Molecule.Abundance*(Molecule.Species).Concentration;
  ManifoldU := Bra as TExoMolManifold;
  ManifoldL := Ket as TExoMolManifold;
  for i := 0 to High(Trans) do
    with Trans[i] do begin
      ManifoldU.MapEv(UpperID, Line.IndexU, Line.EigenvaluesU);
      ManifoldL.MapEv(LowerID, Line.IndexL, Line.EigenvaluesL);
      Line.Frequency := Line.EigenvaluesU.Energy[Line.IndexU] - Line.EigenvaluesL.Energy[Line.IndexL];
      if Line.Frequency < 0 then begin
        { Swap upper and lower }
        t := Line.IndexU; Line.IndexU := Line.IndexL; Line.IndexL := t;
        ET := Line.EigenvaluesU; Line.EigenvaluesU := Line.EigenvaluesL; Line.EigenvaluesL := ET;
        Line.Frequency := -Line.Frequency;
      end;
      Line.Strength := wt * A * Line.EigenvaluesU.Degeneracy(Line.IndexU) / IntensityFactor(Line.Frequency, cm1, HzperMolecule);
      Spectrum.AddLine(Line, Simulation, cm1, Problem);
    end;
end;

procedure TExoMolTransitions.ReadTransStream(S: TTextStream);
var
  ManifoldU, ManifoldL: TExoMolManifold;
  nSkipped, nTrans: Integer;
begin
  FixupProperties.Free; { So next two lines work }
  ManifoldU := Bra as TExoMolManifold;
  ManifoldL := Ket as TExoMolManifold;
  nSkipped := 0;
  nTrans := Length(Trans);
  while not S.AtEndOfFile do begin
    if S.GetFreeItem then begin
      if nTrans > High(Trans) then
        SetLength(Trans, nTrans+10000);
      with Trans[nTrans] do begin
        UpperID := S.IntegerToken;
        S.NextIntegerVar(LowerID);
        S.NextDoubleVar(A);
        if (UpperID < Length(ManifoldU.Map)) and (LowerID < Length(ManifoldL.Map))
           and (ManifoldU.Map[UpperID].Index >0) and (ManifoldL.Map[LowerID].Index > 0) then
          Inc(nTrans)
        else
          Inc(nSkipped);
      end;
    end;
    S.NextLine;
  end;
  SetLength(Trans,nTrans);
  if nSkipped > 0 then
    ShowInfo(IntToStr(nSkipped)+' transitions discarded as state ID''s not found');
end;

procedure TExoMolTransitions.LoadTransFile(const FileName: string);
var
  FileStream: TFileStream;
  S: TFreeFormatStream;
begin
  if FileName = '' then
    Exit;
  S := nil;
  FileStream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    S := TFreeFormatStream.Create(FileStream);
    ReadTransStream(S);
  finally
    S.Free;
    FileStream.Free;
  end;
end;

function TExoMolTransitions.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child: TDOMNode;
  M: THugeMemoryStream;
  S: TFreeFormatStream;
begin
  Result := inherited ReadProperties(Node, False);
  Child := Node.FirstChild;
  if Assigned(Child) and (Child is TDOMCDATASection) then begin
    M := THugeMemoryStream.Create(TDOMCDATASection(Child).Data);
    S := TFreeFormatStream.Create(M);
    try
      ReadTransStream(S)
    finally
      S.Free;
      M.Free;
    end;
    Node.RemoveChild(Child);
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TExoMolTransitions.WriteProperties(Parent: TDOMNode; AllParameters, WithChildren: Boolean): TDOMElement;
var
  Strings: TStringList;
  I: Integer;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Trans)+1;
    Strings.Add('');
    for I := 0 to High(Trans) do
      with Trans[I] do
        Strings.Add(Format('%d %d %g',[UpperID, LowerID, A]));
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

{ TExoMolManifold }

{$IFNDEF CONSOLE}
procedure TExoMolManifold.AddHandlers;
begin
  inherited;
  AddHandler('Add .trans file...', 'Load transitions from .trans file', {$IFDEF FPC}@{$ENDIF}LoadFromTransFileClick);
end;

procedure TExoMolManifold.LoadFromTransFileClick(Sender: TObject);
begin
  LoadTransFile(PromptForTransFile);
end;
{$ENDIF}

procedure TExoMolManifold.LoadTransFile(const FileName: string);
var
  T: TExoMolTransitions;
begin
  if FileName = '' then
    Exit;
  T := (Owner as TExoMolMolecule).FindTransitionMoments(Self,Self) as TExoMolTransitions;
  if T = nil then begin
    T := TExoMolTransitions.Create(Owner);
    T.SetBraKetName(Name,Name);
  end;
  T.LoadTransFile(FileName);
end;

procedure TExoMolManifold.MapEv(StateID: Integer; out AIndex: Integer; out Ev: TEigenvalues);
begin
  with Map[StateID] do begin
    Ev := Eigenvalues(J);
    AIndex := Index;
  end;
end;

function TExoMolManifold.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child: TDOMNode;
  M: THugeMemoryStream;
  S: TFreeFormatStream;
begin
  Result := inherited ReadProperties(Node, False);
  Child := Node.FirstChild;
  if Assigned(Child) and (Child is TDOMCDATASection) then begin
    M := THugeMemoryStream.Create(TDOMCDATASection(Child).Data);
    S := TFreeFormatStream.Create(M);
    try
      ReadStatesFile(S)
    finally
      S.Free;
      M.Free;
    end;
    Node.RemoveChild(Child);
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TExoMolManifold.ReadStatesFile(S: TTextStream): TQuantumNumber;
var
  StateID, LargestID: Integer;
  Js: TQuantumNumber;
  Energy: Double;
  gTot: Integer;
  Quanta: string;
begin
  Result := -1;
  LargestID := 0;
  while not S.AtEndOfFile do begin
    if S.GetFreeItem then begin
      StateID := S.IntegerToken;
      S.NextDoubleVar(Energy);
      S.NextIntegerVar(gTot);
      S.NextStringVar(Quanta);
      if not TryStrToQuantumNumber(Quanta, Js, qftJ) then
        raise EReportException.Create('Bad J value:'+Quanta);
      Quanta := Trim(S.RestOfLine);
      LargestID := Max(LargestID, StateID);
      if StateID > High(Map) then
        SetLength(Map, LargestID+10000);
      Result := Max(Js, Result);
      with Map[StateID] do begin
        Index := AddLevel(Energy, gTot, Js, Quanta);
        J := Js;
      end;
    end;
    S.NextLine;
  end;
  SetLength(Map, LargestID+1);
end;

function TExoMolManifold.WriteProperties(Parent: TDOMNode; AllParameters, WithChildren: Boolean): TDOMElement;
var
  Strings: TStringList;
  id: Integer;
  Ev: TExternalEigenvalues;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Map)+1;
    Strings.Add('');
    for id := 0 to High(Map) do
      with Map[id] do
        if Index > 0 then begin
          Ev := Eigenvalues(J);
          Strings.Add(Format('%d %g %d %s %s',[id,Ev.Energy[Index],Ev.Degeneracy(Index),QuantumNumber.QuantumNumberToStr(J,qftJ),Ev.Quanta(Index)]));
        end;
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

{ TExoMolReader }

function TExoMolReader.DoReadOneLine: Boolean;
var
  StateID, gTot, i, NQN: Integer;
  AStateU, AStateL: TState;
  Qnos: TSimpleQnos;
  Molecule: TMolecule;
begin
  Result := False;
  if DefaultStateL <> nil then
    Molecule := DefaultStateL.Molecule
  else
    Molecule := DefaultMolecule;
  ReadInt(StateID,s);
  f.NextDoubleVar(Obs.Frequency);
  ReadInt(gTot,NextItem);
  ReadQuantumNumber(Qnos.QN[0], NextItem);
  Obs.ObsType := obELower;
  Obs.Weight := 1;
  Obs.ManifoldL := DefaultManifoldL;
  Obs.Name := f.RestOfLine;
  NQN := ReadQnos(Qnos.QN[0], Qnos.QN, Obs.Name);
  for i := 0 to NQN - 1 do
    Qnos.QNP[i] := -1;
  AStateU := nil;
  AStateL := DefaultStateL;
  ReadState(Obs.ManifoldL, AStateL, Obs.Name);
  if ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU,
                           Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name, Molecule,
                           ComputeQ(DefaultStateU, DefaultStateL, Molecule, NQN, RemoveSpins, -1), NQN, Qnos,
                           ExField, UpperMap, LowerMap, QuantumNumberFormat, RemoveSpins, BranchRank, DefaultSymmetry, True) then begin
    ReportError('Bad Quantum numbers');
  end else
    Result := True;
end;

{ TExoMolTransStateReader }

type

  TExoMolTransStateReader = class(TExoMolTransBaseReader)
    constructor Create; override;
    function DoReadOneLine: Boolean; override;
  end;

constructor TExoMolTransStateReader.Create;
begin
  inherited;
  ExoMolTransReader := AncestorWithClass(TExoMolTransBaseReader) as TExoMolTransBaseReader;
end;

function TExoMolTransStateReader.DoReadOneLine: Boolean;
var
  StateID: Integer;
begin
  Result := False; { No observations in this mode }
  ReadInt(StateID,s);
  NextItem; { Discard energy }
  NextItem; { Discard gTot }
  if StateID > High(ExoMolTransReader.States) then
    SetLength(ExoMolTransReader.States,StateID+10000);
  with ExoMolTransReader.States[StateID] do begin
    if Quanta <> '' then
      ReportError('Duplicate StateID:'+IntToStr(StateID));
    ReadQuantumNumber(J, NextItem);
    Quanta := f.RestOfLine;
    if Quanta = '' then
      Quanta := ' '; { Null string flags missing entry }
  end;
end;

{ TExoMolTransReader }

function TExoMolTransReader.DoReadOneLine: Boolean;
var
  Molecule: TMolecule;
  UpperID, LowerID, NQN: Integer;
  AStateU, AStateL: TState;
  Qnos: TSimpleQnos;
begin
  Result := False;
  if DefaultStateU <> nil then
    Molecule := DefaultStateU.Molecule
  else if DefaultStateL <> nil then
    Molecule := DefaultStateL.Molecule
  else
    Molecule := DefaultMolecule;
  ReadInt(UpperID,s);
  ReadInteger(LowerID);
  f.NextDoubleVar(Obs.Frequency);

  if Length(ExoMolTransReader.States) = 0 then
    AddStates(ChangeFileExt(FileName,'.states'));
  if not UnitsChecked and (Mixture.Simulations[0].IntensityUnits <> EinsteinA) then
    ReportError('Simulation IntensityUnits must be EinsteinA');
  UnitsChecked := True;

  if UpperID <= High(ExoMolTransReader.States) then
    with ExoMolTransReader.States[UpperID] do begin
      if Quanta = '' then
        Exit; { State not present }
      NQN := ReadQnos(J, Qnos.QNP, Quanta);
      Obs.ManifoldU := DefaultManifoldU;
      AStateU := DefaultStateU;
      ReadState(Obs.ManifoldU, AStateU, Quanta);
    end
  else
    Exit;
  if LowerID <= High(ExoMolTransReader.States) then
    with ExoMolTransReader.States[LowerID] do begin
      if Quanta = '' then
        Exit; { State not present }
      NQN := ReadQnos(J, Qnos.QN, Quanta);
      Obs.ManifoldL := DefaultManifoldL;
      AStateL := DefaultStateL;
      ReadState(Obs.ManifoldL, AStateL, Quanta);
    end
  else
    Exit;

  Obs.ObsType := obIntensity;
  Obs.Weight := 1;
  if ConvertQuantumNumbers(Obs.ManifoldU, AStateU, Obs.JU, Obs.SymU, Obs.IndexU,
                           Obs.ManifoldL, AStateL, Obs.JL, Obs.SymL, Obs.IndexL, Obs.Name, Molecule,
                           ComputeQ(DefaultStateU, DefaultStateL, Molecule, NQN, RemoveSpins, -1), NQN, Qnos,
                           ExField, UpperMap, LowerMap, QuantumNumberFormat, RemoveSpins, BranchRank, DefaultSymmetry, True) then begin
    ReportError('Bad Quantum numbers');
  end else
    Result := True;
end;

{ TExoMolBaseReader }

constructor TExoMolBaseReader.Create;
var
  R: TExoMolBaseReader;
begin
  inherited;
  Units := cm1;
  R := AncestorWithClass(TExoMolBaseReader) as TExoMolBaseReader;
  if Assigned(R) then begin
    StateCols := Copy(R.StateCols);
    QnoCols := Copy(R.QnoCols);
  end;
end;

procedure TExoMolBaseReader.AddStates(AFileName: string);
begin
  AddToObservations(Mixture, AFileName, nil, ObsType, Take, TExoMolTransStateReader, Self);
end;

function TExoMolBaseReader.HandleDirective(var Handled: Boolean): Boolean;
  procedure LoadArray(var A: TIntegerArray);
  var
    n: Integer;
  begin
    n := 0;
    while f.GetFreeItem do begin
      if n > High(A) then
        SetLength(A, n+4);
      A[n] := f.IntegerToken;
      Inc(n);
    end;
    SetLength(A,n);
  end;
begin
  Result := inherited HandleDirective(Handled);
  if Handled then
    Exit;
  Handled := True;
  if LowerCase(s) = 'addstates' then
    AddStates(NextItem)
  else if LowerCase(s) = 'statecols' then
    LoadArray(StateCols)
  else if LowerCase(s) = 'qnocols' then
    LoadArray(QnoCols)
  else
    Handled := False;
end;

function TExoMolBaseReader.ReadQnos(J: TQuantumNumber; var Q: array of Integer; const Quanta: string): Integer;
var
  i, e: Integer;
  procedure AddQnos(s,e: Integer);
  var
    p, sp: Integer;
  begin
    p := s;
    e := Min(e,Length(Quanta));
    while (p <= e) do begin
      while (p <= e) and (Quanta[p] = ' ') do
        Inc(p);
      sp := p;
      while (p <= e) and (Quanta[p] <> ' ') do
        Inc(p);
      if p > sp then begin
        if not TryStrToQuantumNumber(Copy(Quanta,sp,p-sp), Q[Result], qftJ) then
          Q[Result] := -1;
        Inc(Result);
      end;
    end;
  end;
begin
  Q[0] := J;
  Result := 1;
  i := 0;
  while i < Length(QnoCols) do begin
    if i < High(QnoCols) then
      e := QnoCols[i+1]
    else
      e := High(Integer); { Take all of string }
    AddQnos(QnoCols[i], e);
    Inc(i,2);
  end;
end;

procedure TExoMolBaseReader.ReadState(var Manifold: TManifold; var State: TState; const Quanta: string);
var
  i, e: Integer;
  Name: string;
  MNode: TMNode;
  procedure AddQnos(s,e: Integer);
  var
    p, sp: Integer;
  begin
    p := s;
    e := Min(e,Length(Quanta));
    while (p <= e) do begin
      while (p <= e) and (Quanta[p] = ' ') do
        Inc(p);
      sp := p;
      while (p <= e) and (Quanta[p] <> ' ') do
        Inc(p);
      if p > sp then begin
        if Name <> '' then
          Name := Name + '_';
        Name := Name + Copy(Quanta,sp,p-sp);
      end;
    end;
  end;
begin
  Name := '';
  i := 0;
  while i < Length(StateCols) do begin
    if i < High(StateCols) then
      e := StateCols[i+1]
    else
      e := High(Integer); { Take all of string }
    AddQnos(StateCols[i], e);
    Inc(i,2);
  end;
  if Name <> '' then begin
    State := Manifold.StateWithName(Name);
    if (State = nil) and (Manifold.Molecule.nComponentsWithName(Name,MNode) = 1) then
      if MNode is TState then begin
        State := MNode as TState;
        Manifold := State.Manifold;
      end else if (MNode is TManifold) and (TManifold(MNode).nStates = 1) then begin
        Manifold := MNode as TManifold;
        State := Manifold.State[0];
      end;
    if State = nil then
      ReportError('State not found:'+Name);
  end;
end;

{ TExoMolTransBaseReader }

constructor TExoMolTransBaseReader.Create;
var
  Reader: TObservationsReader;
begin
  inherited;
  Reader := Self;
  while Reader <> nil do begin
    if Reader is TExoMolTransReader then
      ExoMolTransReader := TExoMolTransReader(Reader);
    Reader := Reader.IncludedFrom;
  end;
end;

initialization
  TExoMolManifold.RegisterClass(TExternalMolecule);
  TExoMolTransitions.RegisterClass(TExternalMolecule);
  TExoMolMolecule.RegisterClass(TSpecies);
end.
