unit Fitter;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, MatrixVec, MNode, Mixture, Overlay, Obs, MathSup;

type
  OnShowContour = procedure (Spectra: TOverlays; Simulation: TColumnVector; ObsMinusCalc: Boolean) of object;

function FitLines(var Parameters: TParameters; Obs: TObservationArray; Precision: Integer; Mixture: TMixture; CheckDerivatives: Double):string;
function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;
function CalculateObs(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType; Wait: Boolean = True): Boolean;
function FindObservationColour(const Obs: TObservation; Upper: Boolean):TColor;

type

  TFitter = class
  public
    nObs, nPar: Integer;
    a: TMatrix;   { d(obs)/d(par)ZZzz }
    omc: TColumnVector; { normally obs - calc }
    par: TColumnVector; { Parameters - changed by this routine }
    weights: TColumnVector; { nil or the relative standard deviation of each value in obs }
    OldSD, NewSD: Double; { Previous and estimated next standard deviations }
    UnWtOldSD, UnWtNewSD: Double; { Previous and estimated next standard deviations ignoring weights }
    b: TColumnVector;  { Standard Deviations of the parameters}
    ada: TMatrix; { Correlation Matrix }
    SVD: TMatrix; SVDw: TColumnVector; { Details of Singular Value Decomposition }
    Correlations: TDoubleArrayArray; { Correlation of observation with following observations }
    ParameterScales: TColumnVector; { Scaling factors for parameters; can be nil }
    CheckDerivatives: Double;
    AlternateFitter: TFitter;
    Mixture: TMixture;
    SubtractDerivs: Boolean;
    constructor Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double; MaxCorrelatedObservations: Integer = 0);
    destructor Destroy; override;
    function DoLeastSQ(var Parameters: TParameters; Precision: Integer; Apply: Boolean = True):string;
    procedure CalculateDerivatives(var Parameters: TParameters);
    function CalculateAllObs(v: TColumnVector): Boolean; virtual; abstract;
    procedure AddAlternateFitter;
  end;

implementation

uses Progress, SysUtils, NLFit, ComplexMatrix, Vib, StaticField,
     SpectrumPlot, ExpPlot, Units, DRintensity, StrUtils, UCompat,
     Worker, PGWork, Math, QuantumNumber, Constrain, Profile;

var
  ResidualOutput: string = 'Residual Output';
  LeastSquaresTime: string = 'Least Squares';
  LineFitTime: string = 'Line Fit';
  ContourFitTime: string = 'Contour Fit';

{ TFitter }

procedure TFitter.AddAlternateFitter;
begin
  AlternateFitter := TFitter.Create(0, 0, Mixture, 0);
  AlternateFitter.nPar := nPar;
  AlternateFitter.nObs := nObs;
  AlternateFitter.a.Assign(a);
  AlternateFitter.omc.Assign(omc);
  AlternateFitter.weights.Assign(weights);
  AlternateFitter.Correlations := Copy(Correlations);
end;

constructor TFitter.Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double; MaxCorrelatedObservations: Integer);
begin
  nObs := AnObs;
  nPar := AnPar;
  a := TMatrix.Create(nObs, nPar);
  omc := TColumnVector.Create(nObs);
  weights := TColumnVector.Create(nObs); { Also used as workspace }
  if MaxCorrelatedObservations > 0 then
    SetLength(Correlations,nObs-1);
  Mixture := AMixture;
  CheckDerivatives := ACheckDerivatives;
  SubtractDerivs := True;
end;

destructor TFitter.Destroy;
begin
  FreeAndNil(AlternateFitter);
  a.Free;
  omc.Free;
  par.Free;
  b.Free;
  ada.Free;
  weights.Free;
  SVD.Free;
  SVDw.Free;
  ParameterScales.Free;
  inherited;
end;

function TFitter.DoLeastSQ(var Parameters: TParameters; Precision: Integer; Apply: Boolean):string;
var
  i, j, p, l, nErrDig, nDiscard, nDerivs: Integer;
  r1: string;
  Sum, C, v, x: Double;
  Sorter: TEigenvectorSorter;
  OldParameters: array of Double;
  MaxDerivative: array of Double;
  ObsMax: array of Integer;
  ShowScale: Boolean;
begin
  TStart(LeastSquaresTime);
  if (Mixture.FitDetails or Assigned(AlternateFitter)) and (nPar > 0) then begin
    if Assigned(AlternateFitter) then
      nDerivs := 2
    else
      nDerivs := 1;

    if Mixture.FitDetails then begin
      WriteLn(Log);
      Write(Log,' ':4);
      if nDerivs > 1 then
        WriteLn(Log, 'Derivative Matrix; alternate derivatives listed second')
      else
        WriteLn(Log, 'Derivative Matrix');
      for i := 0 to nPar-1 do
        Write(Log, ' ', Parameters[i].Name:13*nDerivs);
      WriteLn(Log);
      for i := 1 to nObs do begin
        Write(Log, i:6);
        for j := 1 to nPar do begin
          Write(Log, DTosMA(a[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
          if nDerivs = 2 then begin
            Write(Log, DTosMA(AlternateFitter.a[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
          end;
        end;
        WriteLn(Log);
      end;
    end;

    if Assigned(AlternateFitter) then begin
      SetLength(MaxDerivative, nPar);
      SetLength(ObsMax, nPar);
      for j := 1 to nPar do begin
        C := 0;
        { First pass: find biggest derivative }
        for i := 1 to nObs do
          C := Max(C, Max(Abs(a[i,j]), Abs(AlternateFitter.a[i,j])));
        { Second pass: find biggest difference in derivatives }
        x := 0;
        ObsMax[j-1] := 1;
        if C > 0 then
          for i := 1 to nObs do begin
            v := Abs(a[i,j]-AlternateFitter.a[i,j])/C;
            if v > x then begin
              x := v;
              ObsMax[j-1] := i;
            end;
          end;
        MaxDerivative[j-1] := x;
      end;

      WriteLn(Log);
      WriteLn(Log, 'Alternate fit with scaled step size for derivatives');
      AlternateFitter.omc.Assign(omc);
      if Assigned(weights) then
        AlternateFitter.weights.Assign(weights)
      else
        FreeAndNil(AlternateFitter.weights);
      AlternateFitter.DoLeastSQ(Parameters, Precision, False);
      WriteLn(Log);
    end;
  end;

  { Fill in parameter info }
  par := TColumnVector.Create(nPar);
  for i := 0 to nPar-1 do
    par[i+1] := Parameters[i].Value;
  b := TColumnVector.Create(nPar);
  ada := TMatrix.Create(nPar, nPar);
  if Mixture.SVDThresh >= 0 then begin
    SVD := TMatrix.Create(nPar, nPar);
    SVDw := TColumnVector.Create(nPar);
  end else begin
    FreeAndNil(SVD);
    FreeAndNil(SVDw);
  end;
  if Mixture.ScaledParameters then
    ParameterScales := TColumnVector.Create
  else
    FreeAndNil(ParameterScales);
  nDiscard := LeastSQ(a, omc, par, weights, OldSD, NewSD, UnWtOldSD, UnWtNewSD, b, ada, SVDw, SVD,
                      Mixture.SVDThresh, Mixture.FitDetails, Mixture.ScaleChanges, Correlations, ParameterScales);
  WriteLn(Log);
  if nPar > 0 then
    if SVD <> nil then
      Write(Log, 'SVD fit: ')
    else
      Write(Log, 'QR fit: ');
  Write(Log, nObs, ' Observations,  ', nPar, ' Parameters');
  if ParameterScales <> nil then
    WriteLn(Log, ' (scaled)')
  else
    WriteLn(Log);
  if nPar > 0 then
    Result := 'Initial '
  else
    Result := '';
  Result := Result + 'Average Error: ' + FloatToStr(OldSD);
  if OldSD <> UnWtOldSD then
    Result := Result + '  (Unweighted: ' +  FloatToStr(UnWtOldSD) + ')';
  WriteLn(Log, Result);
  if nPar > 0 then begin
    r1 := 'Predicted New Error:   ' + FloatToStr(NewSD);
    if NewSD <> UnWtNewSD then
      r1 := r1 + '  (Unweighted: ' +  FloatToStr(UnWtNewSD) + ')';
    WriteLn(Log, r1);
    Result := Result + ' - ' + r1;
    if nDiscard > 0 then
      WriteLn(Log, '**** ', nDiscard, ' combinations of parameters not floated **** (SVD threshold = ', FloatToStr(Mixture.SVDThresh),')');

    { For constants (such as the observed spectrum scale factor) where the residual is
      inversly proportional to the constant, the estimated change is particularly poor if
      the change is large. Adjust the change by converting to a change in the inverse,
      and then inverting. [The ratio of derivatives of a parameter p and it's inverse is -1/p^2
      so a change d in p corresponds to a change -d/p^2 in 1/p so the new p is
      1/(1/p-d/p^2) = p/(1-d/p) }
    for i := 0 to nPar - 1 do
      if Parameters[i].Inverse then begin
        WriteLn(Log, 'Linear New Value for ',Parameters[i].Name,': ',par[i+1]);
        C := Parameters[i].Value/(1-(par[i+1]-Parameters[i].Value)/Parameters[i].Value);
        { Check that parameter does not change sign; fall back to original method if it does }
        if C * Parameters[i].Value > 0 then
          par[i+1] := C;
      end;

    WriteLn(Log, 'Parameters:');
    Write(Log, '  # Old                 New                 Std Dev   Change/Std Sens    ');
    if Assigned(AlternateFitter) then
      Write(Log, 'Deriv Diff  Obs# ');
    ShowScale := (Mixture.FitDetails or (nDiscard > 0)) and (ParameterScales <> nil);
    if ShowScale then
      Write(Log, 'Scale   ');
    WriteLn(Log, 'Summary              Name');
    for i := 0 to nPar-1 do begin
      { Calculate parameter sensitivity; defined in JKG Watson, J Mol Spec 66 500 (1977)
        See also RJ LeRoy J Mol Spec 191 223 (1998) for additional ways of dealing with this issue.
        Choose number of significant figures such that using the rounded values allows the
        calculated standard deviation to rise by no more then 10% }
      Sum := 0;
      for j := 1 to nObs do
        { Note - don't need Sqr(a[j,i+1]/weights[j]) as a has been adjusted to give observations with standard deviation 1 }
        Sum := Sum + Sqr(a[j,i+1]);
      { and use it to determine the number of uncertainty digits to show }
      Sum := 0.1*NewSD/nPar/Sqrt(Sum/nObs);
      nErrDig := Max(Trunc(Log10(b[i+1]/Sum))+1,2);
      Write(Log, Format('%3d%s%s%s%s %s', [i+1, DtosMA(Parameters[i].Value,20,-1),
                 DtosMA(par[i+1],20,-1), DtosMA(b[i+1],10,-1), DtosMA((par[i+1]-Parameters[i].Value)/b[i+1],10,4),
                 DtosMA(Sum,8,-1)]));
      if Assigned(AlternateFitter) then
        Write(Log, DtosMA(MaxDerivative[i],11,-1), ' ', ObsMax[i]:5);
      if ShowScale then
        Write(Log, DtosMA(ParameterScales[i+1],8,-1));
      WriteLn(Log, Format(' %-20s %s', [DtosErr(par[i+1],b[i+1],nErrDig), Parameters[i].Name]));
      if Apply then begin
        Parameters[i].Value := par[i+1];
        Parameters[i].StdDev := b[i+1];
      end;
    end;
    if Apply then
      WriteConstraints(Log);
    if nPar > 1 then begin
      WriteLn(Log);
      WriteLn(Log, 'Correlation Matrix');
      for i := 1 to nPar do
        Write(Log, i:8);
      WriteLn(Log);
      for i := 1 to nPar do begin
        Write(Log, i:3);
        for j := 1 to i do
          Write(Log, ada[i,j]:8:3);
        WriteLn(Log);
      end;

      if (Mixture.FitDetails or (nDiscard > 0)) and (SVD <> nil) then begin
        WriteLn(Log);
        WriteLn(Log, 'Singular Value Decomposition');
        if (Mixture.SVDlimit <= 0) or (Mixture.SVDlimit > nPar) then begin
          Write(Log,' ':17);
          for i := 0 to nPar-1 do
            Write(Log, ' ', Parameters[i].Name:13);
          WriteLn(Log);
          for i := 1 to nPar do begin
            Write(Log, i:3, DTosMA(SVDW[i],14,-1));
            for j := 1 to nPar do
              Write(Log, DTosMA(SVD[i,j],Max(14,Length(Parameters[j-1].Name)+1),-1));
            WriteLn(Log);
          end;
        end else begin
          l := 0;
          for i := 0 to nPar-1 do
            l := Max(l, Length(Parameters[i].Name));
          Sorter := TEigenvectorSorter.Create(SVD, False);
          for i := 1 to nPar do begin
            Write(Log, i:3, DTosMA(SVDW[i],14,-1));
            Sorter.Process(i, Mixture.SVDlimit);
            for j := 1 to Mixture.SVDlimit do begin
              p := Sorter.Indices[j-1];
              Write(Log, DTosMA(SVD[i,p],Mixture.Precision+6,-1), ' ', Parameters[p-1].Name, ' ':l-Length(Parameters[p-1].Name)+1);
              if j < Mixture.SVDlimit then
                Write(Log, '+');
            end;
            WriteLn(Log);
          end;
          Sorter.Free;
        end;
      end;
    end;
    if not Apply then
      Exit;
    { Copy Back Parameters }
    Mixture.Correlations.FitDetails.Parameters := Copy(Parameters);
    with Mixture.Correlations do begin
      SetLength(FitDetails.Correlation, ada.RowCount);
      for i := 0 to ada.RowCount - 1 do begin
        SetLength(FitDetails.Correlation[i],i);
        for j := 0 to i - 1 do
          FitDetails.Correlation[i][j] := ada[i+1,j+1];
      end;
    end;
    SetLength(OldParameters, nPar);
    try
      for i := 0 to nPar-1 do
        with Parameters[i] do begin
          OldParameters[i] := MNode.Parameters[Index];
          MNode.SetParameter(Index, par[i+1]);
          MNode.StdDev[Index] := b[i+1];
        end;
      ApplyConstraints;
      Mixture.CheckProperties(True);
    except
      on E:Exception do begin
        WriteLn(Log,'**** Failure at end of fit:'+E.Message+' - reverting parameters');
          for i := 0 to nPar-1 do
            with Parameters[i] do
              MNode.SetParameter(Index, OldParameters[i]);
        ApplyConstraints;
      end;
    end;
  end;
  TEnd(LeastSquaresTime);
end;

procedure TFitter.CalculateDerivatives(var Parameters: TParameters);
var
  i, j, deriv, nDerivs: Integer;
  vo, vn, dv: Double;
  Msg: string;
  AA: TMatrix;
begin
  if CheckDerivatives <> 0 then begin
    AddAlternateFitter;
    nDerivs := 2;
  end else
    nDerivs := 1;
  {  Calculate Derivatives wrt each parameter }
  try
    PushProgress('Derivatives', High(Parameters));
    for i := 0 to High(Parameters) do begin
      SetProgress(i);
      vo := Parameters[i].Value;
      try
        if Parameters[i].Increment <= 0 then begin
          vn := vo * 1.0001;
          if vn = 0 then
            vn := 0.001; { Random Guess }
        end else
          vn := vo + Parameters[i].Increment;
        for deriv := 1 to nDerivs do begin
          if deriv = 2 then
            vn := (vn-vo)*CheckDerivatives + vo;
          try
            Parameters[i].MNode.SetParameter(Parameters[i].Index, vn);
            ApplyConstraints;
            Mixture.CheckProperties(True); { TODO: Was Parameters[i].MNode.CheckProperties; inefficient as is }
            if not CalculateAllObs(weights) then
              if not CheckForAbort then
                raise EReportException.Create('Transitions not found - see log for details');
          except
            on EAbort do
              Abort;
            on E:Exception do begin
              Msg := 'Error while stepping '+Parameters[i].MNode.FullParameterName(Parameters[i].Index)+' to '+FloatToStr(vn)
                           +': '+E.Message;
              WriteLn(Log, '**** ',Msg);
              raise;
            end;
          end;
          if deriv = 1 then
            AA := a
          else
            AA := AlternateFitter.a;
          for j := 1 to nObs do begin
            if SubtractDerivs then
              dv := omc[j]
            else
              dv := 0;
            AA[j,i+1] := (dv - weights[j])/(vn-vo);
          end
        end;
      finally
        Parameters[i].MNode.SetParameter(Parameters[i].Index, vo);
        ApplyConstraints;
        { TODO: See above Parameters[i].MNode.CheckProperties; }
      end;
    end;
  finally
    PopProgress;
  end;
end;

type

  { TContourFitter }

  TContourFitter = class(TFitter)
    OriginalCalcI, OriginalObsI: array of Double;
    Original: Boolean;
    Spectra: TOverlays;
    Simulation: TFitPlot;
    constructor Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double; MaxCorrelatedObservations: Integer = 0);
    destructor Destroy; override;
    function CalculateAllObs(v: TColumnVector): Boolean; override;
  end;

constructor TContourFitter.Create(AnPar, AnObs: Integer; AMixture: TMixture; ACheckDerivatives: Double; MaxCorrelatedObservations: Integer);
begin
  inherited Create(AnPar, AnObs, AMixture, ACheckDerivatives, MaxCorrelatedObservations);
  SubtractDerivs := False;
end;

destructor TContourFitter.Destroy;
begin
  FreeAndNil(Simulation);
  inherited Destroy;
end;

function TContourFitter.CalculateAllObs(v: TColumnVector): Boolean;
var
  i, j, l: Integer;
  Problem: Exception;
  CalcI, ObsI, df: Double;
  ConditionSimulation: TSimulation;
  procedure PlaceObsCalc;
  begin
    if IsNanOrInf(CalcI) then
      raise EReportException.CreateFmt('Calculated value of observation %d is %g',[l+1, CalcI]);
    if Original then begin
      OriginalCalcI[l] := CalcI;
      OriginalObsI[l] := ObsI;
    end else begin
      { Slightly odd logic here as "observed" values can actually change if baseline/scaling/correction changed }
      { Order of calculation here to minimize rounding errors }
      CalcI := CalcI - OriginalCalcI[l];
      ObsI := ObsI - OriginalObsI[l];
    end;
    v[l+1] := ObsI - CalcI;
    Inc(l);
  end;
begin
  Result := True;
  ConditionSimulation := Mixture.Simulations[0];
  { Copy in required frequencies }
  l := 0;
  for i := 0 to Spectra.nPlots-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      if not Hide then begin
        if Count > 0 then
          CheckProperties(True); { Calculate corrected intensity }
        if ConvolutionNeeded then begin
          PrepareConvolution(ConditionSimulation, ConditionSimulation.IntensityUnits, 0, True);
          with ConvolutedSpectrum do begin
            df := (FMax-FMin)/(nDF-1);
            for j := 0 to nDF-1 do
              Simulation.Frequency[l+j] := FMin+j*df;
            Inc(l, nDF);
          end;
        end else
          for j := StartPoint to EndPoint do
            if not IsNanOrInf(Intensity[j]) then begin
              Simulation.Frequency[l] := Convert(AdjustFrequency(Frequency[j]),Units,ConditionSimulation.PlotUnits);
              Inc(l);
            end;
      end;
  Simulation.MakeFromSimulation(ConditionSimulation, not Mixture.NoLineList, Problem);
  if Assigned(Problem) then
    if Problem is EWarningException then
      Problem.Free
    else
      raise Problem;
  { On first cycle, populate Original Calclated and "observed" inteities }
  if Original then begin
    SetLength(OriginalCalcI, l);
    SetLength(OriginalObsI, l);
  end;
  l := 0; { Copy simulation, doing vertical scaling as we go }
  for i := 0 to Spectra.nPlots-1 do
   if Spectra[i] is TExperimentalPlot then
    with TExperimentalPlot(Spectra[i]) do
      if not Hide then begin
        if ConvolutionNeeded then
          with ConvolutedSpectrum do
            for j := 0 to nDF-1 do begin
              CalcI := Simulation.Plot[l];
              ObsI := Plot[j];
              PlaceObsCalc;
            end
        else
          for j := StartPoint to EndPoint do
            if not IsNanOrInf(Intensity[j]) then begin
              CalcI := ConvertIntensity(Simulation.Plot[l],Simulation.Frequency[l],ConditionSimulation.PlotUnits,ConditionSimulation.IntensityUnits,Units,IntensityUnits);
              ObsI := AdjustIntensity(Intensity[j]);
              PlaceObsCalc;
            end;
      end;
end;

function FitContour(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour;
                    var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;
var
  f: TContourFitter;
  nObs: Integer;
  i: Integer;
  vo, vn: Double;
  ConditionSimulation: TSimulation;
begin
  TStart(ContourFitTime);
  ConditionSimulation := Mixture.Simulations[0];
  WriteLn(Log);
  { Find total number of observed points }
  nObs := 0;
  for i := 0 to Spectra.nPlots-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if not Hide then begin
          Limits(Frequency, vo, vn, StartPoint, EndPoint);
          if ConvolutionNeeded then begin
            Inc(nObs, ConditionSimulation.nDF);
            Writeln(log, 'Contour fit to ', FloatToStr(AdjustFrequency(vo)), ' - ', FloatToStr(AdjustFrequency(vn)), ' ', NameOfUnit(Units),
                         ' convoluted to ', ConditionSimulation.nDF: 7, ' points from ', Name);
          end else begin
            Inc(nObs, EndPoint-StartPoint+1);
            Writeln(log, 'Contour fit to ', FloatToStr(AdjustFrequency(vo)), ' - ', FloatToStr(AdjustFrequency(vn)), ' ', NameOfUnit(Units),
                         ' (Points ', StartPoint:7, ' - ', EndPoint:7, ') of ', Name);
          end;
        end;
  ApplyConstraints;
  CheckConstraints(Parameters);
  f := TContourFitter.Create(Length(Parameters), nObs, Mixture, CheckDerivatives);
  f.Spectra := Spectra;
  try
    f.Simulation := TFitPlot.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
    f.Simulation.ndf := nObs;
    f.Original := True;
    f.CalculateAllObs(f.omc);
    if Mixture.FitDetails then begin
      WriteLn(Log, 'Observations');
      WriteLn(Log, '#':6, ' Frequency           Observed       Calculated');
      for i := 0 to nObs-1 do
        WriteLn(Log, i:6, DtosMA(f.Simulation.Frequency[i],20,-1), DtosMA(f.OriginalObsI[i],15,-1), DtosMA(f.OriginalCalcI[i],15,-1));
    end;
    if Assigned(ShowContour) then
      ShowContour(Spectra, f.omc, True);
    if JustResiduals then Exit;
    {  Calculate Derivatives wrt each parameter }
    { TODO: Optimize for simple derivatives }
    { Note that f.weights is only used as workspace, and all weights are taken as 1 }
    f.Original := False;
    f.CalculateDerivatives(Parameters);
    FreeAndNil(f.Weights);
    Spectra.ClearStdDev;
    Result := f.DoLeastSQ(Parameters, Mixture.Precision);
  finally
    f.Free;
    TEnd(ContourFitTime);
  end;
end;

function CalculateObs(var Observation: TObservation; out Value: Double; out RegeneratedLabel: string; RegenerateLabel:Boolean; AObsType: TObsType; Wait: Boolean): Boolean;
var
  EigL,EigU: TEigenvalues;
  LineList: TLineList;
  i: Integer;
  Mixture: TMixture;
  Molecule: TMolecule;
  TransitionMoments: TTransitionMoments;
  Intensity: Double;
  Problem: Exception;
  CalcLine: TLine;
  ConditionSimulation: TSimulation;
  viInfo: TValueInfo;
begin
  Result := False;
  if RegenerateLabel or not Wait then
    RegeneratedLabel := 'l';
  with Observation do begin
    if (AObsType = obIntensity) and not Wait then
      Exit
    else if AObsType = obParameter then begin
      viInfo.MNode := MNode;
      viInfo.viType := TValueInfoType(IndexL);
      viInfo.Index := IndexU;
      Value := viInfo.GetValue;
    end else if AObsType = obExpression then
      Value := (MNode as TExpression).Expr.Value
    else if AObsType <> obEUpper then begin
      AIndexL := FindEigenvaluesAndIndex(EigL, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField, True, Wait);
      if not Assigned(EigL) or (AIndexL <= 0) then
        Exit; { not waiting or error }
      Value := EigL.E[AIndexL];
      if AObsType <> obELower then
        Value := -Value;
      if RegenerateLabel then
        RegeneratedLabel := EigL.DominantBasis[AIndexL].StateLabel(ManifoldL.Mixture.QuantumNumberFormat);
    end else
      Value := 0;
    Result := True;
    if not (AObsType in [obELower, obParameter, obExpression]) then begin
      AIndexU := FindEigenvaluesAndIndex(EigU, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField, True, Wait);
      Result := Assigned(EigU) and (AIndexU > 0);
      if not Result then begin
        { not waiting or error }
        if not Wait then
          RegeneratedLabel := 'l';
        Exit;
      end;
      Value := Value + EigU.E[AIndexU];
      if RegenerateLabel then begin
        if AObsType <> obEUpper then
          RegeneratedLabel := ' - ' + RegeneratedLabel;
        RegeneratedLabel := EigU.DominantBasis[AIndexU].StateLabel(ManifoldU.Mixture.QuantumNumberFormat) + RegeneratedLabel;
        if AObsType <> obEUpper then
          RegeneratedLabel := ' : ' + EigU.DominantBasis[AIndexU].TxName(EigU.DominantBasis[AIndexU],EigL.DominantBasis[AIndexL]) + ' : ' + RegeneratedLabel;
      end;
    end;
    if RegenerateLabel and (AObsType = obLine) and (ManifoldU.Initial or ManifoldL.Initial) then begin
      { Consider regenerating intensity calculation }
      Molecule := ManifoldU.Molecule;
      TransitionMoments := Molecule.FindTransitionMoments(ManifoldU, ManifoldL);
      if Assigned(TransitionMoments) and TransitionMoments.Allowed(JU, SymU, JL, SymL) then begin
        TransitionMoments.MakeLine(CalcLine, EigL, EigU, AIndexL, AIndexU, ManifoldL.Mixture.Simulations[0]);
        Problem := nil;
        Mixture := ManifoldL.Mixture;
        Intensity := Mixture.Simulations[0].LineIntensity(CalcLine, Mixture.Units, Problem);
        if not Assigned(Problem) then
          RegeneratedLabel := FloatToStrF(Intensity, ffExponent, 5, 2) + RegeneratedLabel;
      end;
    end;
    if AObsType = obIntensity then begin
      { Slightly inefficient - calculate all lines between these pairs of Manifold, J and Sym.
        Could be optimized by sorting observations by Manifold, J and sym }
      Mixture := ManifoldL.Mixture;
      ConditionSimulation := Mixture.Simulations[0];
      if ConditionSimulation.IntensityUnits <> Observation.IntensityUnits then begin
        ConditionSimulation := TSimulation.Create(nil);
        ConditionSimulation.UpdateBaseNode(Mixture.Simulations[0]);
        ConditionSimulation.IntensityUnits := Observation.IntensityUnits;
      end;
      LineList := TLineList.Create(ConditionSimulation.PlotUnits,IsPopDist(ConditionSimulation.IntensityUnits));
      { TODO: IsPopDist(Mixture.IntensityUnits) may need treating as a special case }
      try
        Problem := nil;
        ManifoldU.Molecule.AddToSpectrum(LineList, Mixture.PrintLevel, RandomExcitation, Problem, ConditionSimulation,
                                         JL, ManifoldL.Molecule.JDeAdjustedSym(JL,SymL),
                                         JU, ManifoldU.Molecule.JDeAdjustedSym(JU,SymU), ManifoldL, ManifoldU,
                                         -10000, nil, nil, nil, nil, true, true, Observation.RemoveSpins);
        for i := 0 to LineList.nLines-1 do
          with LineList.Line[i] do
            if ((EigenvaluesU.Manifold = ManifoldU) and (IndexU = AindexU) and (IndexL = AindexL) and (EigenvaluesU.J = JU) and (EigenvaluesU.Sym = SymU) ) { TODO: -cField: Sym test fails with Field }
               or ((EigenvaluesU.Manifold = ManifoldL) and (IndexL = AindexU) and (IndexU = AindexL) and (EigenvaluesL.J = JU) and (EigenvaluesL.Sym = SymU)) then begin
              Value := ConditionSimulation.LineIntensity(LineList.Line[i], LineList.Units, Problem, Observation.RemoveSpins);
              if Assigned(Problem) then begin
                Result := False;
                WriteLn(Log, Problem.Message);
                Problem.Free;
              end else if IsNanOrInf(Value) then begin
                LineList.ShowLine(i, 0, LineList.Units, LineList.Units, True, Problem);
                Problem.Free;
                Result := False;
              end;
              Exit;
            end;
        if Mixture.AllowZeroI then
          Value := 0
        else begin
          WriteLn(Log, 'Transition not found during fitting - line number ', LineNumber, ' of ', FileName, ':', Name);
          for i := 0 to LineList.nLines-1 do
            LineList.ShowLine(i, 0, ConditionSimulation.PlotUnits, ConditionSimulation.PlotUnits, False, Problem);
          Problem.Free;
          Result := False;
        end;
      finally
        LineList.Free;
        if ConditionSimulation <> Mixture.Simulations[0] then
          ConditionSimulation.Free;
      end;
    end;
  end;
end;

function FindObservationColour(const Obs: TObservation; Upper: Boolean):TColor;
  procedure FindColour(Manifold: TManifold; J: TQuantumNumber; Sym: Shortint; Ind: Integer; IsUpper: Boolean);
  var
    Ev: TEigenvalues;
  begin
    Ev := nil;
    with Obs do begin
      if not Assigned(Manifold) or (ObsType in [obParameter,obExpression]) or (Ind <= 0) then
        Exit;
      if (Ind = FindEigenvaluesAndIndex(Ev, Manifold, J, Sym, Ind, Name,
                                       not IsUpper, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField))
         and Assigned(Ev) then
      Result := Ev.DominantBasis[Ind].State.FindColour;
    end;
  end;
begin
  Result := Obs.Colour;
  if Result <> clNone then
    Exit;
  with Obs do begin
    if Upper or (ObsType = obEUpper) then
      FindColour(ManifoldU, JU, SymU, AIndexU, True)
    else
      FindColour(ManifoldL, JL, SymL, AIndexL, False);
    if (Result = clNone) and (ObsType in [obLine, obIntensity, obCDLower, obCDUpper]) then
      if not Upper then { Try other state for colour }
        FindColour(ManifoldU, JU, SymU, AIndexU, True)
      else
        FindColour(ManifoldL, JL, SymL, AIndexL, False);
  end;
  if Result = clNone then
    Result := clBlack;
end;

type

  { TLineFitter }

  TLineFitter = class(TFitter)
  public
    Obs: TObservationArray;
    PreCalcPartitionFunction: Boolean;
    function CalculateAllObs(v: TColumnVector): Boolean; override;
  end;


function TLineFitter.CalculateAllObs(v: TColumnVector): Boolean;
var
  i, Step, Block, BlockSize, iWorker, Pass: Integer;
  val: Double;
  Worker: TPgopherWorker;
  RLab: string;
  Problem: Exception;
begin
  Result := True;
  nObs := Length(Obs);
  Problem := nil;
  if PreCalcPartitionFunction then
    Mixture.PreCalculatePartitionFunctions(Mixture.Simulations[0], Problem);
  if Assigned(Problem) then begin
    Result := False;
    WriteLn(Log, Problem.Message);
    Problem.Free;
    Exit;
  end;
  PushProgress('Calculating Observations',nObs);
  try
    if nWorkers > 0 then begin
      { Parallel algorithm }
      { Heuristic for maximise independent calculations:
        make large steps through observations array }
      {$IFDEF ALG1}
      for Block := 0 to BlockSize-1 do begin
        SetProgress(Block, BlockSize-1);
        i := Block;
        while (i < nObs) and Result do begin
          Worker := FindWorkerWait as TPgopherWorker;
          if Assigned(Worker) and Result then
            Worker.CalculateObs(@Obs[i], v.Item[i+1], @Result)
          else begin
            { Something has failed }
            Result := False;
            if Assigned(Worker) then
              Worker.Skip;
          end;
          Inc(i, BlockSize);
        end;
        if not Result then
          Break;
      end;
      {$ENDIF}
      {$IFDEF ALG2}
      DivMod(nObs, nWorkers, BlockSize, i);
      //Slower on test
      if i > 0 then
        Inc(BlockSize);
      Block := 0;
      for i := 1 to nWorkers do begin
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) and Result then
          Worker.CalculateObsBlock(Block, BlockSize, v, Obs, @Result)
        else begin
          { Something has failed }
          Result := False;
          if Assigned(Worker) then
            Worker.Skip;
        end;
        Inc(Block, BlockSize);
      end;
      {$ENDIF}
      { Not entirely clear why, but for CN test case, this gives the fastest results on test system
        May well depend on specific machine, though }
      BlockSize := nObs div nWorkers;
      if BlockSize = 0 then
        BlockSize := 1;
      for Block := 0 to BlockSize-1 do begin
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) and Result then
          Worker.CalculateObsBlock(Block, -BlockSize, v, Obs, @Result)
        else begin
          { Something has failed }
          Result := False;
          if Assigned(Worker) then
            Worker.Skip;
        end;
      end;
      WaitNoBusyWorkers;
      if not Result then
        for Block := 0 to BlockSize-1 do
          for iWorker := 0 to nWorkers do begin
            i := Block + iWorker*BlockSize;
            if (i < nObs) and IsNanOrInf(v[i+1]) then
              raise EReportException.CreateFmt('Calculated value of observation %d is %g',[i+1, v[i+1]]);
          end;
    end else { Serial algorithm }
      for i := 0 to nObs-1 do begin
        SetProgress(i);
        Result := Result and CalculateObs(Obs[i], val, RLab, False, Obs[i].ObsType);
        if IsNanOrInf(val) then
          raise EReportException.CreateFmt('Calculated value of observation %d is %g',[i+1, val]);
        v[i+1] := val;
      end;
  finally
    PopProgress;
  end;
end;

function FitLines(var Parameters: TParameters; Obs: TObservationArray; Precision: Integer;
                  Mixture: TMixture; CheckDerivatives: Double):string;
var
  f: TLineFitter;
  nObs: Integer;
  i, j, wl: Integer;
  vo: Double;
  { For determining printout format }
  LargestMolecule, MolName, LargestLower, LargestUpper, LargestName: string;  { For Determining field lengths }
  SymWidth: Integer;
  MoleculeChanged, LowerChanged, UpperChanged: Boolean; { For Determining if field needs printing }
  RLab: string;
  JHeader: string;
  HasEField, HasBField, HasJ, HasF, PreCalcPartitionFunction: Boolean;
  MaxCorrelation: Integer;
  //Node: TMNode;
  FullLabel: string;
  viInfo: TValueInfo;
  Header: string;
  Manifold: TManifold;
begin
  nObs := Length(Obs);
  if nObs <= 0 then
    raise EReportException.Create('No observations');
  if nObs < Length(Parameters) then
    raise EReportException.CreateFmt('More parameters (%d) than observations (%d)', [Length(Parameters), nObs]);
  TStart(LineFitTime);
  { Work out widths of fields for printout }
  MoleculeChanged := False;
  LowerChanged := False;
  UpperChanged := False;
  LargestLower := '';
  LargestUpper := '';
  LargestMolecule := '';
  LargestName := '';
  SymWidth := 1;
  HasEField := False;
  HasBField := False;
  HasJ := False;
  HasF := False;
  MaxCorrelation := 0;
  PreCalcPartitionFunction := False;
  for i := 0 to nObs-1 do
    with Obs[i] do begin
      if ObsType = obParameter then begin
        viInfo.MNode := MNode;
        viInfo.viType := TValueInfoType(IndexL);
        viInfo.Index := IndexU;
        FullLabel := viInfo.MNode.UniqueName+'.'+viInfo.ValueName;
        if Length(FullLabel) > Length(LargestName) then
          LargestName := FullLabel;
        Continue;
      end;
      if ObsType = obExpression then begin
        FullLabel := (MNode as TExpression).Text;
        if Length(FullLabel) > Length(LargestName) then
          LargestName := FullLabel;
        Continue;
      end;
      if ObsType <> obEUpper then
        Manifold := ManifoldL
      else
        Manifold := ManifoldU;
      if (ObsType = obIntensity)  then
        PreCalcPartitionFunction := PreCalcPartitionFunction or IsNormalized(Mixture.Simulations[0].IntensityUnits);
      MolName := Manifold.Molecule.Name;
      SymWidth := Max(SymWidth, Manifold.Molecule.SymLength);
      if LargestMolecule <> MolName then begin
        if Length(LargestMolecule) > 0 then
          MoleculeChanged := True;
        if Length(LargestMolecule) < Length(MolName) then
          LargestMolecule := MolName;
      end;
      if (ObsType <> obEUpper) and (LargestLower <> ManifoldL.Name) then begin
        if Length(LargestLower) > 0 then
          LowerChanged := True;
        if Length(LargestLower) < Length(ManifoldL.Name) then
          LargestLower := ManifoldL.Name
      end;
      if (ObsType <> obELower) and (LargestUpper <> ManifoldU.Name) then begin
        if Length(LargestUpper) > 0 then
          UpperChanged := True;
        if Length(LargestUpper) < Length(ManifoldU.Name) then
          LargestUpper := ManifoldU.Name
      end;
      if IsNonZero(ExField) then begin
        HasEField := HasEField or (ExField.E <> 0);
        HasBField := HasBField or (ExField.B <> 0);
      end else if not (Manifold is TVibrationalManifold) then
        if Length(Manifold.Molecule.HyperfineBasisStates)-RemoveSpins > 1 then
          HasF := True
        else
          HasJ := True;
      MaxCorrelation := Max(MaxCorrelation, Length(Correlation));
    end;

  ApplyConstraints;
  CheckConstraints(Parameters);
  f := TLineFitter.Create(Length(Parameters), nObs, Mixture, CheckDerivatives, MaxCorrelation);
  f.Obs := Obs;
  f.PreCalcPartitionFunction := PreCalcPartitionFunction;
  try
    { Determine obs - calc }
    if not f.CalculateAllObs(f.omc) then
      if not CheckForAbort then
        raise EReportException.Create('Transitions not found - see log for details');
    for i := 0 to nObs-1 do
      with Obs[i] do begin
        Calculated := f.omc[i+1];
        Flags := Flags + [ofCalculatedSet];
        if Length(Correlation) > 0 then
          f.Correlations[i] := Correlation;
      end;
    f.CalculateDerivatives(Parameters);
    { Write out headers }
    PushProgress('Residual Output');
    TStart(ResidualOutput);
    WriteLn(Log);
    Write(Log, 'Residuals before fit');
    if MoleculeChanged then begin
      if Length(LargestMolecule) < 3 then
        LargestMolecule := 'Mol';
    end else begin
      if LargestMolecule <> '' then
        Write(Log, ' for ', LargestMolecule);
      LargestMolecule := '';
    end;
    if not (LowerChanged or UpperChanged) then
      if (LargestLower = '') or (LargestUpper = LargestLower) then begin
        if LargestUpper <> '' then
          Write(Log, ' for the ', LargestUpper, ' manifold')
      end else if LargestUpper = '' then
        Write(Log, ' for the ', LargestLower, ' manifold')
      else
        Write(Log, ' for the ', LargestUpper, ' - ', LargestLower, ' Transition');
    WriteLn(Log);
    Header := '';
    if Length(LargestMolecule) > 0 then
      Header := Copy('Molecule', 1, Length(LargestMolecule)) + ' ';
    if UpperChanged then
      Header := Header + StringOfChar(' ', Length(LargestUpper)+1);
    JHeader := '';
    if HasJ then
      JHeader := JHeader + 'J';
    if HasF then
      JHeader := JHeader + 'F';
    if HasBField or HasEField then
      JHeader := JHeader + 'M';
    if JHeader = '' then
      JHeader := '-'; { Vibrational only }
    if Mixture.QuantumNumberFormat = qftJ then
      JHeader := '  '+JHeader;
    Header := Header + '  ' + JHeader + '''' + 'S''' + StringOfChar(' ', SymWidth+1) + '#''';
    if LowerChanged then
      Header := Header + StringOfChar(' ', Length(LargestLower)+1);
    Header := Header + '  ' + JHeader + '"' + 'S"' + StringOfChar(' ', SymWidth+1) + '#"';
    if HasEField then
      Header := Header + ' E        ';
    if HasBField then
      Header := Header + '  B       ';
    if Length(LargestName)+2 > Length(Header) then
      Header := Header + StringOfChar(' ',Length(LargestName) - Length(Header) + 2);
    Write(Log, Header, '     Observed    Calculated      Obs-Calc StdDev ');
    for i := 1 to MaxCorrelation do
      Write(Log, ' Correl');
    WriteLn(Log);
    { Work out obs-calc, weights }
    for i := 0 to nObs-1 do begin
      SetProgress(i, nObs);
      with Obs[i] do begin
        if Length(LargestMolecule) > 0 then
          Write(Log, ManifoldU.Molecule.Name:Length(LargestMolecule), ' ');
        if ObsType = obParameter then begin
          viInfo.MNode := MNode;
          viInfo.viType := TValueInfoType(IndexL);
          viInfo.Index := IndexU;
          FullLabel := viInfo.MNode.UniqueName+'.'+viInfo.ValueName;
          Write(Log, FullLabel, ' ':Length(Header)-Length(FullLabel)-1);
        end else if ObsType = obExpression then begin
          FullLabel := (MNode as TExpression).Text;
          Write(Log, FullLabel, ' ':Length(Header)-Length(FullLabel)-1);
        end else begin
          if ObsType <> obELower then begin
            if UpperChanged then
              Write(Log, ManifoldU.Name:Length(LargestUpper), ' ');
            Write(Log, ManifoldU.Mixture.QuantumNumberToStr(JU):2+Length(JHeader),
                       ' ', PadRight(ManifoldU.Molecule.JAdjustedSymName(JU,SymU),SymWidth),
                       ' ', AIndexU:3, ' ')
          end else begin
            wl := 8+SymWidth+Length(JHeader);
            if UpperChanged then
              wl := wl + Length(LargestUpper)+1;
            Write(Log, ' ':wl);
          end;
          if ObsType <> obEUpper then begin
            if LowerChanged then
                Write(Log, ManifoldL.Name:Length(LargestLower), ' ');
            Write(Log, ManifoldL.Mixture.QuantumNumberToStr(JL):2+Length(JHeader),
                       ' ', PadRight(ManifoldL.Molecule.JAdjustedSymName(JL,SymL),SymWidth),
                       ' ', AIndexL:3)
          end else begin
            wl := 7+SymWidth+Length(JHeader);
            if LowerChanged then
              wl := wl + Length(LargestLower)+1;
            Write(Log, ' ':wl)
          end;
        end;
        if HasEField then
          Write(Log, DtosMA(ExField.E,10,-1));
        if HasBField then
          Write(Log, DtosMA(ExField.B,10,5));
        if ObsType = obIntensity then
          Write(Log, Frequency:14, f.omc[i+1]:14,
                     Frequency-f.omc[i+1]:14,
                     DtosMA(Weight,8,-1))
        else
          Write(Log, AtLeast(Frequency,14,Precision), AtLeast(f.omc[i+1],14,Precision),
                     AtLeast(Frequency-f.omc[i+1],14,Precision),
                     DtosMA(Weight,8,-1));
        for j := 1 to MaxCorrelation do
          if (i < nObs-1) and (j <= Length(f.Correlations[i])) and (f.Correlations[i][j-1] <> 0) then
            Write(Log, DtosMA(f.Correlations[i][j-1],7,4))
          else
            Write(Log, ' ':7);
        if ofRegenerateLabel in Flags then begin
          CalculateObs(Obs[i], vo, RLab, True, Obs[i].ObsType);
          Write(Log, ' ', RLab, ' :':(LongestObsLabel+2-Length(Rlab)));
        end else begin
          Write(Log, ' ', Name, ' :':(LongestObsLabel+2-Length(Name)));
        end;
        Writeln(Log, LineNumber:4, ':', FileName);
      end;
      f.omc[i+1] := f.omc[i+1] - Obs[i].Frequency;
      f.weights[i+1] := Obs[i].Weight;
    end;
    TEnd(ResidualOutput);
    PopProgress;
    Result := f.DoLeastSQ(Parameters, Precision);
  finally
    f.Free;
    TEnd(LineFitTime);
  end;
end;

end.
