unit ImportACDD;
{ Import of data files from
  ACDD
  ASYTOP, A PROGRAM FOR DETAILED ANALYSIS OF GAS PHASE
  MAGNETIC RESONANCE SPECTRA OF ASYMMETRIC TOP MOLECULES.
  T.J. SEARS. REF. IN COMP. PHYS. COMMUN. 34 (1984) 123
}

{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

interface

uses
  Classes, SysUtils, Mixture, TextStrm, Obs;

function ImportACDDFile(S: TTextStream): TMixture;

type
  { Import of fort.9 eigenvalue dump file }
  { Requires ACDD to be modified to dump eigenvalues after diagonalization }
  TACDD9Reader = class(TObservationsReader)
    function DoReadOneLine: Boolean; override;
  end;

implementation

uses CODATA, OAsymT, QuantumNumber, Units, MathSup;

function ImportACDDFile(S: TTextStream): TMixture;
var
  nManifolds: Integer;
  Species: TSpecies;
  Molecule: TAsymmetricMolecule;
  ManifoldL, ManifoldU: TAsymmetricManifold;
  Line: string;
  Origin: Double;
  Spin, NucSpin: TQuantumNumber;
  TransitionMoments: TTransitionMoments;
  procedure ReadLine;
  begin
    Line := S.RestOfLine;
    S.NextLine;
  end;
  function ipar(n:Integer):Integer;
  var
    s: string;
  begin
    s := Trim(Copy(line,(n-1)*2+1,2));
    s := StringReplace(s, ' ', '', [rfReplaceAll]);
    if s = '' then
      Result := 0
    else
      Result := StrToInt(s);
  end;
  function par(n:Integer):Double;
  var
    s: string;
  begin
    s := Trim(Copy(line,(n-1)*10+1,10));
    s := StringReplace(s, ' ', '', [rfReplaceAll]);
    if s = '' then
      Result := 0
    else
      Result := StrToFloatEx(s);
  end;
  procedure ReadManifold(Manifold: TManifold);
    function NonZero(n:Integer):Boolean;
    var
      i: Integer;
    begin
      Result := False;
      for i := 1 to n do
        if Par(i) <> 0 then begin
          Result := True;
          Exit;
        end;
    end;
  var
    State: TAsymmetricTop;
    Nuc: TAsymmetricTopNucleus;
  begin
    State := TAsymmetricTop.Create(Manifold);
    State.Name := 'X';
    State.S := Spin;
    ReadLine; { A1,B1,C1 - Upper }
    State.A := Par(1);
    State.B := Par(2);
    State.C := Par(3);
    ReadLine; { DN1,DNK1,DK1,DELN1,DELK1 }
    State.DJ  := Par(1);
    State.DJK := Par(2);
    State.DK  := Par(3);
    State.deltaJ := Par(4);
    State.deltaK := Par(5);
    ReadLine; { HN1,HNK1,HKN1,HK1,SHN1,SHNK1,SHK1 }
    State.HJ := Par(1);
    State.HJK := Par(2);
    State.HKJ := Par(3);
    State.HK := Par(4);
    State.phiJ := Par(5);
    State.phiJK := Par(6);
    State.phiK := Par(7);
    ReadLine; { LK1 }
    State.LK := -Par(1);
    ReadLine; { ALPHA1,BETA1 }
    State.alpha := Par(1);
    State.beta := Par(2);
    ReadLine; { EAA1,EBB1,ECC1,EAB1 }
    State.eaa := Par(1);
    State.ebb := Par(2);
    State.ecc := Par(3);
    State.eabbar := Par(4);
    ReadLine; { DKS1,DNPKS1,DNKS1,DNS1,DLKS1,DLNS1 }
    State.DsK := Par(1);
    State.DsKN := Par(2); { Read KN + NK }
    State.DsNK := Par(3);
    State.DsKN := State.DsKN - State.DsNK;
    State.DsN := Par(4);
    State.ds2 := Par(5);
    State.ds2 := Par(6);
    ReadLine; { HKS1 }
    if Par(1) <> 0 then
      with TAsymmetricTopPerturbation.Create(Manifold) do begin
        Value := Par(1)/2;
        NSPower := 1;
        aPower := 4;
      end;
    ReadLine; { AF,AI,BI,ABI }
    if NucSpin > 0 then begin
      State.AdjustNuclei(True);
      Nuc := State.Nucleus(0) as TAsymmetricTopNucleus;
      Nuc.Spin := NucSpin;
      Nuc.aF := Par(1);
      Nuc.Tzz := Par(2);
      Nuc.Txxmyy := Par(2)+2*Par(3);
    end else
      Nuc :=nil;
    ReadLine; { AQ,BQ }
    if Assigned(Nuc) then begin
      Nuc.CHIzz := Par(2);
      Nuc.CHIxxmyy := Par(2)+2*Par(3);
    end;
    ReadLine; { GAA1,GBB1,GCC1,GAB1,GN1 }
    if NonZero(5) then begin
      TransitionMoments := TTransitionMoments.Create(Molecule, Manifold.Name, Manifold.Name);
      with TAsymmetricTopOrbitalTransitionMoment.Create(TransitionMoments,State.Name,State.Name) do begin
        gsaa := Par(1);
        gsbb := Par(2);
        gscc := Par(3);
        gsabbar := Par(4);
        { TODO: GN1 }
      end;
    end;
    ReadLine; { GRA1,GRB1,GRC1 }
    if NonZero(3) then begin
      if not Assigned(TransitionMoments) then
        TransitionMoments := TTransitionMoments.Create(Molecule, Manifold.Name, Manifold.Name);
      with TAsymmetricTopRotationalTransitionMoment.Create(TransitionMoments,State.Name,State.Name) do begin
        graa := Par(1)/MeOverMp;
        grbb := Par(2)/MeOverMp;
        grcc := Par(3)/MeOverMp;
      end;
    end;
  end;
begin
  Result := TMixture.Create(nil);
  try
    Result.Name := 'Mixture';
    Species := TSpecies.Create(Result);
    Species.Name := 'S';
    Molecule := TAsymmetricMolecule.Create(Species);
    Molecule.Name := 'M';
    Molecule.SReduction := False;
    ReadLine; { Skip IDN,IDK,IDDK,I2S,I2I }
    Spin := Ipar(4);
    NucSpin := Ipar(5);
    if NucSpin > 0 then begin
      Molecule.nNuclei := 1;
    end;
    ReadLine; { Skip IFLD }
    ReadLine; { XNU }
    ManifoldL := TAsymmetricManifold.Create(Molecule);
    ManifoldL.Name := 'Lower';
    ManifoldL.Initial := True;
    Origin := Par(1);
    if Origin <> 0 then begin
      ManifoldU := TAsymmetricManifold.Create(Molecule);
      ManifoldU.Name := 'Upper';
    end else
      ManifoldU := ManifoldL;
    ReadLine; { TEMP }
    Result.Simulations[0].Temperature := Par(1);
    TransitionMoments := nil;
    ReadManifold(ManifoldU);
    ManifoldU.State[0].Origin := Origin;
    if ManifoldL <> ManifoldU then begin
      TransitionMoments := nil;
      ReadManifold(ManifoldL);
      ManifoldU.State[0].Name := 'A';
      TransitionMoments := nil; { No diagonal transition moment }
    end;
    { ILSQS }
    { IPRED }
    { IEIGS }
    { NMAX,NMIN }
    { Fake a transition moment }
    if not Assigned(TransitionMoments) then
      TransitionMoments := TTransitionMoments.Create(Molecule, ManifoldL.Name, ManifoldU.Name);
    TCartesianTransitionMoment.Create(TransitionMoments, ManifoldL.State[0].Name, ManifoldU.State[0].Name).Strength := 1;
  except
    Result.Free;
    Raise;
  end;
end;

{ TACDD9Reader }

function TACDD9Reader.DoReadOneLine: Boolean;
var
  i: Integer;
begin
  Obs.JL := f.IntegerToken;
  f.NextIntegerVar(i);
//  if i = 0 then
//    Obs.SymL := i
//  else
//    Obs.SymL := i mod 3 + 1;
  if i >= 2 then
    Obs.SymL := i xor 1
  else
    Obs.SymL := i;
  f.NextIntegerVar(i); Obs.IndexL := i;
  f.NextDoubleVar(Obs.Frequency);
  Obs.ObsType := obELower;
  Obs.LineNumber := f.Line;
  Obs.FileName := FileName;
  Obs.Weight := 1;
  Result := True;
end;

end.
