unit LoomisWood;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
{ Support for Loomis Wood style plots }
interface

uses Classes, pgDOM, Expr, DataFitter, DataSource, MNode, MNodeExpr,
     {$IFNDEF CONSOLE} PlotOC, {$ENDIF} Peaks;

type

  TLoomisWood = class(TMNodeExprVarsNoConstants)
  private
    mVar, mLowestVar: TVariable;
    FPeaks: TPeaks;
    FDataSource: TDataSource;
    {$IFNDEF CONSOLE}
    ResidualPlotter: TResidualPlotter;
    {$ENDIF}
    function GetDataSource: TDataSource;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
  protected
    procedure SetUpVariables; override;
  public
    LastyOffset: Double;
    Shift0: Double;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AddHandlers; override;
    procedure ShowLoomisWood(Sender: TObject);
    procedure AddNodeParameters(var FittingParameters: TParameters); override;
    procedure FindZeroShift;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function xShift(m: Double):Double;
    function Peaks: TPeaks;
    function HasDataSource: Boolean;
    property DataSource: TDataSource read GetDataSource write FDataSource;
  published
    property yOffset: Double index 1 read GetParameter write SetParameter;
    property mLowest: Double index 2 read GetParameter write SetParameter;

    property nPlots:           Integer index 3 read GetSetting write SetSetting;
    property mOffset:          Integer index 4 read GetSetting write SetSetting;
    property ShrinkSimulation: Boolean index 5 read GetFlag write SetFlag;
  end;

implementation

uses {$IFNDEF CONSOLE} Gendlg, LoomisWoodFrm, ExpBitmap, ResidFrm, Main, {$ENDIF}
     SysUtils, Math, Peak, UCompat, MathSup, Overlay, ExpPlot;

{$IFNDEF CONSOLE}
type
  TLWResidualPlotter = class(TResidualPlotter)
  protected
    LoomisWood: TLoomisWood;
    function GetX(Index: Integer): Double; override;
    function SkipY(var O, C, W: Double; Index: Integer):Boolean; override;
    function nObs:Integer; override;
    procedure SetXType(const Value: TOCXStyle); override;
  public
    constructor Create(ALoomisWood: TLoomisWood);
    destructor Destroy; override;
    function ResidualAvailable(var Res, Wt: Double; Index: Integer):Boolean; override;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string; override;
    procedure EditLastPoint; override;
    procedure ShowLastPoint; override;
  end;

{ TLWResidualPlotter }

constructor TLWResidualPlotter.Create(ALoomisWood: TLoomisWood);
begin
  LoomisWood := ALoomisWood;
  LoomisWood.ResidualPlotter := Self;
  inherited Create('');
end;

destructor TLWResidualPlotter.Destroy;
begin
  if LoomisWood.ResidualPlotter = Self then
    LoomisWood.ResidualPlotter := nil;
  inherited;
end;

procedure TLWResidualPlotter.EditLastPoint;
begin
  if (LastLineNumber >= 0) and (LastLineNumber < LoomisWoodForm.Grid.RowCount) then
    LoomisWoodForm.Grid.Row := LastLineNumber + 1;
end;

function TLWResidualPlotter.GetX(Index: Integer): Double;
begin
  with LoomisWood.Peaks[Index] do begin
    SkipPoint := SkipPoint or IsNan(Weight) or IsNan(Actual) or (Weight <= 0);
    case Xtype of
      ocxNumber: Result := Actual;
      ocxFrequency, ocxEnergy: Result := Position;
      ocxCalculated: Result := Position-Residual;
      ocxObserved: Result := Position;
      ocxWeight: Result := Weight;
    else
      SetNan(Result);
    end;
  end;
end;

function TLWResidualPlotter.nObs: Integer;
begin
  Result := LoomisWood.Peaks.nPeaks;
end;

function TLWResidualPlotter.OnGetPointInfo(Index: Integer; Extra: PtrInt): string;
begin
  if Index < LoomisWood.Peaks.nPeaks then begin
    Result := ' ' + IntToStr(Index+1);
    if LastLineNumber < 0 then
      LastLineNumber := Index;
  end else
    Result := '';
end;

function TLWResidualPlotter.ResidualAvailable(var Res, Wt: Double; Index: Integer): Boolean;
begin
  Result := True;
  with LoomisWood.Peaks[Index] do begin
    Res := Residual;
    Wt := Weight;
  end;
end;

procedure TLWResidualPlotter.SetXType(const Value: TOCXStyle);
begin
  inherited SetXType(Value);
  if XLabel = '' then
    SetXType(ocxNumber); { Force sensible default }
end;

procedure TLWResidualPlotter.ShowLastPoint;
begin
  if (LastLineNumber >= 0) and (LastLineNumber < LoomisWood.Peaks.nPeaks) then
    MainForm.CentrePlot(LoomisWood.Peaks[LastLineNumber].Position);
end;

function TLWResidualPlotter.SkipY(var O, C, W: Double; Index: Integer): Boolean;
begin
  with LoomisWood.Peaks[Index] do begin
    Result := IsNan(Weight) or IsNan(Actual) or (Weight <= 0);
    O := Position;
    C := Position-Residual;
    W := Weight;
  end;
end;

{$ENDIF}

type

  TLWDataSource = class ({$IFNDEF CONSOLE} TInteractiveDataSource {$ELSE} TFitterDataSource {$ENDIF})
  private
    LoomisWood: TLoomisWood;
  public
    FitExpr: TExpr;
    AvgErr: Double;
    destructor Destroy; override;
    procedure AddIndependentVariables(Lex:TLex); override;
    function nRows: Integer; override;
    function GetData(r: Longint; var x, y, w: Double; SetVariableControlsState: Boolean = False; ExtraAxis: Integer = -1): Boolean; override;
    function Weighted: Boolean; override;
    function Caption: string; override;
    procedure SetFitExpr(Expr: TExpr; nPar: Integer); override;
  end;

{ TLWDataSource }

procedure TLWDataSource.AddIndependentVariables(Lex: TLex);
begin
  Lex.AddVariable(LoomisWood.mVar);
  Lex.AddVariable(LoomisWood.mLowestVar);
end;

function TLWDataSource.Caption: string;
begin
  Result := LoomisWood.FullHeirarchicName;
end;

destructor TLWDataSource.Destroy;
begin
  LoomisWood.FDataSource := nil;
  inherited;
end;

function TLWDataSource.GetData(r: Integer; var x, y, w: Double;
  SetVariableControlsState: Boolean; ExtraAxis: Integer): Boolean;
var
  Peak: TPeak;
begin
  if r = 1 then
    LoomisWood.FindZeroShift;
  Peak := LoomisWood.Peaks[r-1];
  y := Peak.Position - LoomisWood.Shift0;
  x := Peak.Actual;
  w := Peak.Weight;
  LoomisWood.mVar.Value := x;
  DataSourceFitter.ObservedVariable.Value := y;
  DataSourceFitter.weightVariable.Value := w;
  Result := LoomisWood.Peaks.Valid(r-1);
end;

function TLWDataSource.nRows: Integer;
begin
  Result := LoomisWood.Peaks.nPeaks;
end;

procedure TLWDataSource.SetFitExpr(Expr: TExpr; nPar: Integer);
var
  i, nData: Integer;
  x, y, w, Sum: Double;
  Peak: TPeak;
begin
  FitExpr := Expr;
  if not Assigned(FitExpr) or not Assigned(LoomisWood) then
    Exit;
  { Fill in residuals }
  Sum := 0;
  nData := 0;
  for i := 0 to nRows-1 do
    if GetData(i+1,x,y,w) then begin
      Peak := LoomisWood.Peaks[i];
      Peak.Residual := Peak.Position - LoomisWood.Shift0 - FitExpr.Value ;
      Sum := Sum + Sqr(Peak.Residual);
      Inc(nData);
      LoomisWood.Peaks[i] := Peak;
    end;
  if nData > nPar then
    AvgErr := Sqrt(Sum/(nData-nPar))
  else
    SetNan(AvgErr);
  {$IFNDEF CONSOLE}
  LoomisWoodForm.AvgErr := AvgErr;
  LoomisWoodForm.UnLoadExpressionAndVariables;
  LoomisWoodForm.UpdateStatus;
  ResidualForm.ResidualPlotter := TLWResidualPlotter.Create(LoomisWood);
  MainForm.ConsiderReplot(False,False);
  {$ENDIF}
end;

function TLWDataSource.Weighted: Boolean;
begin
  Result := True;
end;

{ TLoomisWood }

procedure TLoomisWood.AddHandlers;
begin
  inherited;
  AddHandler('View...', 'View, edit and fit peaks', {$IFDEF FPC}@{$ENDIF}ShowLoomisWood);
end;

procedure TLoomisWood.AddNodeParameters(var FittingParameters: TParameters);
begin
  { Parameters can only be floated in the Loomis Wood form }
end;

constructor TLoomisWood.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',nPlots=5,yOffset=1';
  inherited;
  Comment := 'b*(m-mLowest)+d*(m-mLowest)^2';
  LastyOffset := -99;
  FPeaks := TPeaks.Create(Self);
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
end;

destructor TLoomisWood.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  FreeAndNil(FPeaks);
  {$IFNDEF CONSOLE}
  FreeAndNil(ResidualPlotter);
  {$ENDIF}
  FreeAndNil(FDataSource);
  inherited;
end;

procedure TLoomisWood.FindZeroShift;
var
  i: Integer;
  m: Double;
begin
  m := MaxDouble;
  for i := 0 to Peaks.nPeaks - 1 do
    if Peaks.Valid(i) and (Abs(Peaks[i].Actual) < Abs(m)) then begin
      m := Round(Peaks[i].Actual);
      Shift0 := Peaks[i].Position;
    end;
  if m = MaxDouble then
    mLowest := 0
  else
    mLowest := m;
  mLowestVar.Value := m;
end;

function TLoomisWood.GetDataSource: TDataSource;
begin
  if not Assigned(FDataSource) then begin
    FDataSource := TLWDataSource.Create(nil);
    TLWDataSource(FDataSource).LoomisWood := Self;
  end;
  Result := FDataSource;
end;

function TLoomisWood.HasDataSource: Boolean;
begin
  Result := Assigned(FDataSource);
end;

procedure TLoomisWood.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  if (Sender = Self) and (Change = mnPropNameChange) and (Index = nParameters) then
    Floating[nParameters] := True;  { Default parameters to floating }
end;

function TLoomisWood.Peaks: TPeaks;
begin
  Result := FPeaks;
end;

function TLoomisWood.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child: TDOMNode;
begin
  Result := inherited ReadProperties(Node, False);
  Peaks.ReadPeaks(Node.FirstChild);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

procedure TLoomisWood.SetUpVariables;
begin
  inherited SetUpVariables;
  V('m', mVar);
  V('mLowest', mLowestVar);
end;

procedure TLoomisWood.ShowLoomisWood(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  ConnectLoomisWoodForm(Self);
  {$ENDIF}
end;

function TLoomisWood.WriteProperties(Parent: TDOMNode; AllParameters, WithChildren: Boolean): TDOMElement;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Peaks.WritePeaks(Result);
end;

function TLoomisWood.xShift(m: Double): Double;
begin
  mVar.Value := m + mOffset;
  mLowestVar.Value := mLowest;
  Result := Evaluate;
end;

initialization
  TLoomisWood.RegisterClass(TExperimentalPlot);
  TLoomisWood.RegisterClass(TOverlays);
  {$IFNDEF CONSOLE}
  TLoomisWood.RegisterClass(TBitmapPlot);
  {$ENDIF}
end.
