unit MNodeExpr;
{ Unit to handle Custom Population, Linestrength etc where a general expression is given
  in terms of variables }
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses SysUtils, Classes, MNode, Expr, Units, pgDOM;

type

  { TMNodeWithExprs }

  TMNodeWithExprs = class(TMNodeWithParameters)
  { Base for allowing MNode allowing general expressions with variables }
  protected
    procedure AddVariables(ALex: TLex); virtual;
    function MakeLex: TLex; virtual;
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function ParseString(const s: string; var Expr: TExpr):Exception; overload;
    function ParseString(const s: string; var Expr: TExpr; const SearchPath: array of TMNode; CreateUnknown: Boolean):Exception; overload;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ClearExpr; virtual;
    procedure PromptForParameter(Sender: TObject);
    procedure AddHandlers; override;
  end;

  TVariables = class (TMNodeWithExprs)
    { Used to hold variables used in constraints and elsewhere }
  public
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Global: Boolean index 1 read GetFlag write SetFlag default True;
  end;

  TMNodeExpr = class(TMNodeWithExprs)
  { MNode allowing general expression with variables }
  { TODO: put expression in separate field rather than comment field }
  protected
    FExpr: TExpr;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function CheckVariables: Exception;
    procedure DoCheckVariables(Sender: TObject);
    procedure AddHandlers; override;
    procedure ClearExpr; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function FullyActive: Boolean; { Active and non-blank expression }
    property Expr: TExpr read FExpr;
  published
    property Active: Boolean index 1 read GetFlag write SetFlag default True;
  end;

  TMNodeExprClass = class of TMNodeExpr;

  { TMNodeExprVars }

  TMNodeExprVars = class(TMNodeExpr)
  { Add standard way of handling a variable list}
  protected
    VariableList: TVariableList;
    nDebugOutput: Integer;
    DebugTag: array of string; { 1 for each thread }
    TriggerVariable: TVariable;
    procedure SetUpVariables; virtual;
    procedure UpdateVariables; virtual;
    procedure AddVariables(ALex: TLex); override;
  public
    procedure V(const VName: string; var vr: TVariable);
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure ClearCache(ClearChildren: Boolean); override;
    destructor Destroy; override;
    function Evaluate:Double;
  published
    property nDebug: Integer index 2 read GetSetting write SetSetting;
  end;

  TMNodeExprVarsNoConstants = class(TMNodeExprVars)
  protected
    function MakeLex: TLex; override;
  end;

  { TMNodeWithArray }

  TMNodeWithArray = class(TMNodeWithExprs)
  private
    nErrors: Integer;
  protected
    function MakeLex: TLex; override;
    procedure InternalAddRow(const NewRow: string; var A:TStringArrayArray);
  public
    A, AF: TStringArrayArray;
    LastError: string;
    Rows: array of Integer; { Index into string array A }
    procedure AddVariables(ALex: TLex); override;
    procedure HandleError(Row, Col: Integer; Msg: string);
    function HandleValue(Index, Col: Integer; UseAF: Boolean = False; BlankIsZero: Boolean = False):Double;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function Percent(n:Integer):string; virtual;
  published
    procedure AddRow(const NewRow: string);
    procedure AddFRow(const NewRow: string);
    procedure ClearArray;
    procedure ClearFArray;
  end;

  TActiveMNodeWithArray = class(TMNodeWithArray)
  private
    FEffectiveOwner: TMNode;
  public
    class function MakeDummy(AOwner: TMNode):TActiveMNodeWithArray;
    constructor Create(AOwner: TComponent); override;
    procedure ClearArrays; virtual; abstract;
    destructor Destroy; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function Apply(Verbose: Boolean):string; virtual; abstract;
    function ApplyOK(Verbose: Boolean):Boolean;
    function ReallyActive: Boolean; virtual; { As Active, but requires some extra information to be present }
    property EffectiveOwner: TMNode read FEffectiveOwner;
  published
    property Active: Boolean index 1 read GetFlag write SetFlag;
  end;

  TActiveMNodeWithArrayClass = class of TActiveMNodeWithArray;

  TTable = class(TMNode)
  public
    ColumnNames: array of string;
    Values: array of array of Double;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromFileClick(Sender: TObject);
    procedure PasteClick(Sender: TObject);
    {$ENDIF}
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure LoadFromStream(Stream: TStream);
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
  published
    property Debug: Boolean index 1 read GetFlag write SetFlag;
  end;

  TTableFunction = class (TFunction)
  public
    Table: TTable;
    constructor Create(ATable: TTable);
    function Value(var a:array of Double):Double; override;
  end;

function FindSubProp(Text: string; const SearchPath: array of TMNode; SingleValuesOnly: Boolean):TValueInfo;
function ActiveMnodeExpr(MNode: TMnode; MNodeExprClass: TMNodeExprClass): TMNodeExpr;

implementation

uses {$IFNDEF CONSOLE }
       {$IFDEF MSWINDOWS} Windows, {$ENDIF}
       Clipbrd, Forms, ConstFrm, Dialogs, ExpPlot, Huge, WCompat,
     {$ENDIF}
     Math, MathSup, UCompat, Worker, TextStrm, Constrain, Mixture, Overlay;

function FindSubProp(Text: string; const SearchPath: array of TMNode; SingleValuesOnly: Boolean):TValueInfo;
var
  Ignored: string;
  procedure Search(MNode: TMNode; ImplicitDot: Boolean);
  begin
    if Assigned(MNode) then
      Result := MNode.SubPropInfo(Text, Ignored, ImplicitDot);
    if Assigned(Result.MNode) and SingleValuesOnly and not Result.IsSingleValue then
      Result.MNode := nil;
  end;
var
  i: Integer;
  Variables: TVariables;
begin
  Result.MNode := nil;
  { Look in first node always }
  Search(SearchPath[0], True);
  if Assigned(Result.MNode) then
    Exit;
  for i := 0 to High(SearchPath) do
    { then look for global variables first }
    if SearchPath[i] is TRootMNode then begin
      Variables := TVariables(SearchPath[i].ChildWithClass(TVariables));
      if Assigned(Variables) and Variables.Global then
        Search(Variables, True);
      if Assigned(Result.MNode) then
        Exit;
    end;
  for i := 1 to High(SearchPath) do begin
    Search(SearchPath[i], False);
    if Assigned(Result.MNode) then
      Exit;
  end;
end;

{ TMNodeExprVars }

procedure TMNodeExprVars.V(const VName: string; var vr: TVariable);
begin
  vr := TVariable.Create(VName);
  VariableList.Add(vr);
end;

procedure TMNodeExprVars.SetUpVariables;
begin
  TriggerVariable := nil;
  VariableList := TVariableList.Create;
end;

procedure TMNodeExprVars.UpdateVariables;
begin
  {}
end;

procedure TMNodeExprVars.AddVariables(ALex: TLex);
var
  i: Integer;
begin
  SetLength(DebugTag, nWorkers+1);
  inherited;
  if not Assigned(VariableList) then
    SetUpVariables;
  UpdateVariables;
  for i := 0 to VariableList.Count - 1 do begin
    VariableList.Objects[i].SetAllValues;
    ALex.AddVariable(VariableList.Objects[i]);
  end;
end;

procedure TMNodeExprVars.ClearCache(ClearChildren: Boolean);
begin
  inherited;
  nDebugOutput := 0;
end;

destructor TMNodeExprVars.Destroy;
begin
  FreeAndNil(VariableList);
  inherited Destroy;
end;

function TMNodeExprVars.Evaluate: Double;
var
  i: Integer;
begin
  Result := Expr.Value;
  case nDebug of
     0: Exit; { No Debugging }
    -2: if Result = 0 then Exit;
    -3: if not Assigned(TriggerVariable) or (Result = TriggerVariable.Value) then Exit;
  else
    if (nDebug > 0) and (nDebugOutput > nDebug) then
      Exit;
  end;
  Inc(nDebugOutput);
  Write(Log, FullHeirarchicName, DebugTag[ThreadIndex-1], ' ');
  for i := 0 to VariableList.Count - 1 do
    with VariableList.Objects[i] do
      Write(Log, Name,'=',FloatToStr(Value),' ');
  WriteLn(Log,'->',FloatToStr(Result));
end;

function TMNodeExprVars.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 2 then
    Result := 'Set non zero to write information to the log window on each use'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMNodeExprVars.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  s: string;
  Debug: Boolean;
begin
  s := ReadValue(Node, 'Debug', '');
  if (s <> '') and TryStrToBool(s, Debug) and Debug then
    nDebug := -1;
  Result := inherited ReadProperties(Node, CheckUnread);
end;

{ TMNodeExpr }

procedure TMNodeExpr.AddHandlers;
begin
  AddHandler('Check Variables', 'Parse expression, and add undefined variables', {$IFDEF FPC}@{$ENDIF}DoCheckVariables);
  inherited;
end;

function TMNodeExpr.CheckVariables: Exception;
begin
  { Parse comment field as expression, and create any undefined variables as parameters }
  Result := ParseString(Comment, FExpr);
end;

procedure TMNodeExpr.ClearExpr;
begin
  inherited;
  FreeAndNil(FExpr);
end;

constructor TMNodeExpr.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Active=True';
  inherited Create(AOwner);
end;

destructor TMNodeExpr.Destroy;
begin
  ClearExpr;
  inherited;
end;

procedure TMNodeExpr.DoCheckVariables(Sender: TObject);
var
  E: Exception;
begin
  E := CheckVariables;
  if Assigned(E) then
    raise E;
end;

function TMNodeExpr.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  if FullyActive and not Assigned(Result) then
    Result := CheckVariables;
end;

function TMNodeExpr.FullyActive: Boolean;
begin
  Result := Active and (Trim(Comment) <> '')
end;

function TMNodeExpr.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'True if object is to be used'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TMNodeWithExprs }

procedure TMNodeWithExprs.AddHandlers;
begin
  inherited;
  AddHandler('Add Parameter...', 'Add a custom parameter', {$IFDEF FPC}@{$ENDIF}PromptForParameter);
end;

procedure TMNodeWithExprs.PromptForParameter(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
    ConstantsForm.AddParameterItemClick(Self);
  {$ENDIF}
end;

procedure TMNodeWithExprs.ClearExpr;
begin
  {}
end;

constructor TMNodeWithExprs.Create(AOwner: TComponent);
begin
  inherited;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
end;

destructor TMNodeWithExprs.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  inherited;
end;

procedure TMNodeWithExprs.AddVariables(ALex: TLex);
  procedure AddTables(MNode: TMNode);
  var
    i: Integer;
  begin
    for i := 0 to ComponentCount - 1 do
      if Components[i] is TTable then
        ALex.AddFunction(TTableFunction.Create(Components[i] as TTable), True)
  end;
begin
  { Actually adding functions }
  AddTables(RootNode); { Global tables first }
  AddTables(Self); { Potentially overridden by local tables }
end;

function TMNodeWithExprs.MakeLex: TLex;
begin
  Result := TLex.Create;
  Result.AddFunction(ConstantPi);
  AddVariables(Result);
end;

procedure TMNodeWithExprs.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  if (Change = mnPropNameChange) and (Sender = Self) then
    ClearExpr;
end;

function TMNodeWithExprs.ParseString(const s: string; var Expr: TExpr):Exception;
var
  SearchPath: array [0..1] of TMNode;
begin
  SearchPath[0] := Self;
  SearchPath[1] := RootNode;
  Result := ParseString(s, Expr, SearchPath, True);
end;

function TMNodeWithExprs.ParseString(const s: string; var Expr: TExpr; const SearchPath: array of TMNode; CreateUnknown: Boolean): Exception;
var
  VariableList: TVariableList;
  i, p: Integer;
  NewExpr: TExpr; { Note: don't ues Expr var as AddParameter() call below can free it via MNodeChange above }
  Lex: TLex;
  viInfo: TValueInfo;
begin
  { Parse string as expression, and create any undefined variables as parameters }
  Result := nil;
  ClearExpr;
  try
    Lex := MakeLex;
    NewExpr := TExpr.Create(Lex, True);
    Lex.ReturnUnknown := True;
    VariableList := NewExpr.MakeVariables(s);
    try
      for i := 0 to VariableList.Count - 1 do begin
        viInfo := FindSubProp(VariableList[i], SearchPath, True);
        if viInfo.IsSingleValue then
          Lex.AddFunction(TMNodeVariable.Create(VariableList[i],viInfo),True)
        else if not CreateUnknown then
          raise Exception.Create('Unknown variable :'+VariableList[i])
        else if Pos('.', VariableList[i]) > 0 then
          raise Exception.Create('Variable names must not contain a dot:'+VariableList[i])
        else begin
          p := AddParameter(VariableList[i]);
          Lex.AddFunction(TMNodeVariable.Create(VariableList[i],Self,p),True);
        end;
        Lex.RemoveVariable(VariableList.Objects[i]);
      end;
    finally
      VariableList.Free;
    end;
    Lex.ReturnUnknown := False;
    NewExpr.Text := s;
    with Lex.Functions do
      for i := 0 to Count - 1 do
        if Objects[i] is TMNodeVariable then
          with TMNodeVariable(Objects[i]) do
            Name := viInfo.MNode.FullHeirarchicName;
  except
    on E: Exception do begin
      Result := EReportException.Create(E.Message);
      FreeAndNil(NewExpr);
    end;
  end;
  Expr := NewExpr;
end;

function TMNodeWithExprs.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, False) + ReadArbitraryParameters(Self, Node);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

procedure TMNodeWithExprs.ScaleParameters(UFrom, UTo: TResonanceUnits);
begin
  { Disable all conversions }
end;

{ TMNodeWithArray }

procedure TMNodeWithArray.AddVariables(ALex: TLex);
begin
  { TODO }
end;

procedure TMNodeWithArray.ClearArray;
begin
  SetLength(A,0);
  Modified := True;
end;

procedure TMNodeWithArray.ClearFArray;
begin
  SetLength(AF,0);
  Modified := True;
end;

function TMNodeWithArray.MakeLex: TLex;
begin
  Result := inherited MakeLex;
  Result.IdentifierChars := Result.IdentifierChars + ['[',']'];
  Result.Options := Result.Options - [xFundamentalConstants];
end;

procedure TMNodeWithArray.InternalAddRow(const NewRow: string; var A: TStringArrayArray);
var
  n, p, c: Integer;
  Separator: Char;
  s: string;
begin
  TrimArray(A);
  n := Length(A);
  SetLength(A,n+1);
  s := NewRow;
  p := Pos(#9, s);
  if p > 0 then
    Separator := #9  { Assume tab separated list of values }
  else
    Separator := ',';  { Assume comma separated list of values }
  c := 0;
  repeat
    p := Pos(Separator, s);
    SetLength(A[n],c+1);
    if p > 0 then begin
      A[n][c] := Copy(s,1,p-1);
      Delete(s,1,p)
    end else
      A[n][c] := s;
    Inc(c);
  until p <= 0;
  TrimArray(A);
  Modified := True;
end;

procedure TMNodeWithArray.AddFRow(const NewRow: string);
begin
  InternalAddRow(NewRow, AF);
end;

procedure TMNodeWithArray.AddRow(const NewRow: string);
begin
  InternalAddRow(NewRow, A);
end;

procedure TMNodeWithArray.HandleError(Row, Col: Integer; Msg: string);
var
  p, s, n: Integer;
begin
  { Handle %%n substitutions (for atom name, for example) }
  p := Pos('%%', Msg);
  while p > 0 do begin
    s := p+2;
    while (s <= length(Msg)) and (Msg[s] in ['0'..'9']) do
      Inc(s);
    n := StrToInt(Copy(Msg,p+2,s-p-2));
    Msg := Copy(Msg,1,p-1) + Percent(n) + Copy(Msg,s,Length(Msg));
    p := Pos('%%', Msg);
  end;
  if LastError <> '' then begin
    Inc(nErrors);
    if nErrors > 1 then
      Exit; { Limit length }
    LastError := LastError + '; '
  end else
    nErrors := 0;
  if Row >= 0 then
    LastError := LastError + 'Row '+ IntToStr(Row+1)+':';
  if Col >= 0 then
    LastError := LastError + 'Col ' + IntToStr(Col+1)+':';
  LastError := LastError + Msg;
end;

function TMNodeWithArray.HandleValue(Index, Col: Integer; UseAF: Boolean; BlankIsZero: Boolean): Double;
var
  r: Integer;
  S: TStringArrayArray;
  c: string;
  Expr: TExpr;
  E: Exception;
begin
  SetNan(Result);
  if UseAF then begin
    r := Index;
    S := AF;
  end else begin
    r := Rows[Index];
    S := A;
  end;
  if Col > High(S[r]) then
    c := ''
  else
    c := S[r][Col];
  if c = '' then
    if UseAf or BlankIsZero then
      Result := 0
    else
      HandleError(r,Col,'Missing Value')
  else if TryStrToFloat(c,Result) then
    Exit { Simple number }
  else begin
    try
      Expr := nil;
      E := ParseString(c, Expr);
      if Assigned(E) then
        raise E;
      Result := Expr.Value;
    except
      on E: Exception do begin
        SetNan(Result);
        HandleError(r,Col,'Bad expression:'+c);
      end;
    end;
    Expr.Free;
  end;
end;

function TMNodeWithArray.Percent(n: Integer): string;
begin
  Result := IntToStr(n);
end;

function TMNodeWithArray.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMNode;
  s: string;
  DoneA: Boolean;
begin
  Result := inherited ReadProperties(Node, False);
  SetLength(A,0);
  SetLength(AF,0);
  Child := Node.FirstChild;
  DoneA := False;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if Child is TDOMCDATASection then begin
      s := TDOMCDATASection(Child).Data;
      Node.RemoveChild(Child);
      if not DoneA then begin
        ReadArray(A, s);
        DoneA := True;
      end else begin
        ReadArray(AF,s);
        Break;
      end;
    end;
    Child := NextChild;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TMNodeWithArray.WriteProperties(Parent: TDOMNode;
  AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  s: string;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  TrimArray(A);
  Result.AppendChild(Result.OwnerDocument.CreateCDATASection(StringRep(A)));
  s := StringRep(AF);
  if s <> '' then
    Result.AppendChild(Result.OwnerDocument.CreateCDATASection(s));
end;

{ TActiveMNodeWithArray }

function TActiveMNodeWithArray.ApplyOK(Verbose: Boolean): Boolean;
begin
  LastError := '';
  try
    Apply(Verbose);
  except
    on E:Exception do
      LastError := E.Message;
  end;
  Result := LastError = '';
end;

constructor TActiveMNodeWithArray.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Active=True';
  inherited;
  FEffectiveOwner := AOwner as TMNode;
end;

destructor TActiveMNodeWithArray.Destroy;
begin
  ClearArrays;
  inherited;
end;

function TActiveMNodeWithArray.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If this object is to be used.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

class function TActiveMNodeWithArray.MakeDummy(AOwner: TMNode): TActiveMNodeWithArray;
begin
  Result := Create(nil);
  Result.FEffectiveOwner := AOwner;
end;

function TActiveMNodeWithArray.ReallyActive: Boolean;
begin
  Result := Active and (Length(A) + Length(AF) > 0);
end;

{ TVariables }

function TVariables.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If variables are to be available in all expressions without the name as prefix.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function ActiveMnodeExpr(MNode: TMnode; MNodeExprClass: TMNodeExprClass): TMNodeExpr;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to MNode.Count-1 do
    if (MNode.Components[i] is MnodeExprClass) and TMNodeExpr(MNode.Components[i]).Active then
      if Assigned(Result) then
        raise EReportException.Create('Multiple Active '+MNodeExprClass.DisplayClassName+' in '+MNode.FullHeirarchicName)
      else
        Result := TMNodeExpr(MNode.Components[i]);
end;

{ TMNodeExprVarsNoConstants }

function TMNodeExprVarsNoConstants.MakeLex: TLex;
begin
  Result := inherited MakeLex;
  Result.Options := Result.Options - [xFundamentalConstants];
end;

{ TTable }

procedure TTable.LoadFromStream(Stream: TStream);
var
  S: TTextStream;
  procedure ReadColumnNames;
  var
    n: Integer;
  begin
    n := 0;
    repeat
      SetLength(ColumnNames,n+1);
      ColumnNames[n] := S.Token;
      Inc(n);
    until not S.GetFreeItem;
  end;
var
  nCol, nCols, nRows: Integer;
  Value: Double;
begin
  { Based on DatLoader }
  S := TTextStream.Create(Stream);
  S.ItemSeparators := [isTab, isComma];
  { First pass, determining number of columns }
  nCols := 0;
  nRows := 0;
  while not S.AtEndOfFile do begin
    nCol := 0;
    while not S.AtEndOfLine do
      if S.GetFreeItem then begin
        if not S.IsRealToken(Value) then begin
          if (nCol = 0) and (nRows = 0) then begin
            SetLength(ColumnNames,0);
            ReadColumnNames;
          end;
          Break;
        end;
        Inc(nCol);
        nCols := Max(nCol, nCols);
      end;
    S.NextLine;
    if nCol > 0 then
      Inc(nRows);
  end;
  if nRows <= 0 then
    Exit;
  SetLength(ColumnNames, nCols);
  for nCol := 0 to nCols - 1 do
    if ColumnNames[nCol] = '' then
      ColumnNames[nCol] := IntToStr(nCol+1);
  { Repeat read logic, this time setting arrays }
  SetLength(Values, nRows, nCols);
  S.Position := 0;
  nRows := 0;
  while not S.AtEndOfFile do begin
    nCol := 0;
    while not S.AtEndOfLine do
      if S.GetFreeItem then begin
        if not S.IsRealToken(Value) then
          Break;
        Values[nRows][nCol] := Value;
        Inc(nCol);
      end;
    S.NextLine;
    if nCol = nCols then
      Inc(nRows);
  end;
  SetLength(Values, nRows);
end;

function TTable.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  s: TStringStream;
begin
  Result := inherited ReadProperties(Node, False);
  if Assigned(Node.FirstChild) and (Node.FirstChild is TDOMCDATASection) then begin
    S := TStringStream.Create(TDOMCDATASection(Node.FirstChild).Data);
    try
      LoadFromStream(S);
    finally
      S.Free;
    end;
    Node.RemoveChild(Node.FirstChild);
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TTable.WriteProperties(Parent: TDOMNode; AllParameters, WithChildren: Boolean): TDOMElement;
var
  Strings: TStringList;
  s: string;
  i, j: Integer;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  if Length(Values) = 0 then
    Exit;
  Strings := TStringList.Create;
  try
    Strings.Capacity := Length(Values)+1;
    s := '';
    for i := 0 to High(ColumnNames) - 1 do
      s := s + ColumnNames[i] + #9;
    s := s + ColumnNames[High(ColumnNames)];
    Strings.Add(s);
    for i := 0 to High(Values) do begin
      s := '';
      for j := 0 to High(Values[i])-1 do
        s := s + FloatToStr(Values[i][j]) + #9;
      s := s + FloatToStr(Values[i][High(Values[i])]);
      Strings.Add(s);
    end;
    Result.AppendChild(Parent.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

function TTable.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'True to print out details of each table lookup.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

procedure TTable.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i, j: Integer;
begin
  inherited;
  if Length(ColumnNames) > 0 then begin
    Write(F, Indent, '  ');
    for i := 0 to High(ColumnNames) do
      Write(F, ColumnNames[i]:20);
    WriteLn(F);
  end;
  for j := 0 to High(Values) do begin
    Write(F, Indent, '  ');
    for i := 0 to High(Values[j]) do
      Write(F, FloatToStr(Values[j][i]):20);
    WriteLn(F);
  end;
end;

{$IFNDEF CONSOLE}
procedure TTable.AddHandlers;
begin
  inherited;
  AddHandler('Load...', 'Load from file', {$IFDEF FPC}@{$ENDIF}LoadFromFileClick);
  AddHandler('Paste', 'Paste from clipboard', {$IFDEF FPC}@{$ENDIF}PasteClick);
end;

procedure TTable.LoadFromFileClick(Sender: TObject);
var
  M: TMemoryStream;
begin
  M := nil;
  with TOpenDialog.Create(Application) do
    try
      Filter := 'Text File (*.txt;*.dat;*.csv)|*.txt;*.dat;*.csv|All Files|'+AllFilesMask;
      Title := 'Load Table';
      if Execute then begin
        M := TMemoryStream.Create;
        M.LoadFromFile(FileName);
        LoadFromStream(M);
        Modified := True;
        PerformMNodeNotification(Self, mnPropNameChange, -1);
      end;
    finally
      Free;
      M.Free;
    end;
end;

procedure TTable.PasteClick(Sender: TObject);
var
  M: THugeMemoryStream;
begin
  if ClipboardHasText then begin
    M := THugeMemoryStream.Create;
    try
      M.SetFromClipboard(CF_TEXT);
      LoadFromStream(M);
      Modified := True;
      PerformMNodeNotification(Self, mnPropNameChange, -1);
    finally
      M.Free;
    end;
  end;
end;
{$ENDIF}

{ TTableFunction }

constructor TTableFunction.Create(ATable: TTable);
var
  n: Integer;
begin
  Table := ATable;
  if Length(Table.Values) = 0 then
    n := 0
  else
    n := High(Table.Values[0]);
  inherited Create(Table.Name, n);
end;

function TTableFunction.Value(var a: array of Double): Double;
  procedure ShowArgs;
  var
    i: Integer;
  begin
    Write(Log, Table.FullHeirarchicName, ' ');
    for i := 0 to High(a) do
      Write(Log, FloatToStr(a[i]), ' ');
  end;
var
  i, j: Integer;
  Found: Boolean;
begin
  if Table.Debug then
    ShowArgs;
  for i := 0 to High(Table.Values) do begin
    Found := True;
    for j := 0 to High(a) do begin
      Found := a[High(a)-j] = Table.Values[i][j];
      if not Found then
        Break;
    end;
    if Found then begin
      Result := Table.Values[i][High(Table.Values[i])];
      if Table.Debug then
        WriteLn(Log, 'Entry ', i, ' = ', FloatToStr(Result));
      Exit;
    end;
  end;
  Result := 0; { Not Found }
  if Table.Debug then
    WriteLn(Log, ' *** Not Found');
end;

initialization
  TVariables.RegisterClass(TOverlays);
  TVariables.RegisterClass(TMixture);
  TTable.RegisterClass(TOverlays);
  TTable.RegisterClass(TMixture);
end.
