unit MNodeFunc;
{ Unit to handle scripting of TMNode functions; in separate file as rather compiler dependent. }
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses TypInfo, Classes, MNode;

procedure CheckMNodeFunction(MNode: TMNodeClass; const AName: string; TypeInfo: PTypeInfo; AltTypeInfo: PTypeInfo);

procedure CallMNodeFunction(MNode: TMNode; MethodAddress: Pointer; TypeData: PTypeData; Arguments: string);

implementation

uses SysUtils, UCompat, PropHelp, MNodeExpr, zMatrix, InternalCoords;

type
// From typinfo.pas FPC compiler source file TTypeData record
// As the short strings are simply concatenated, a proper structure definition is not possible
//  RTTIParameter = Record
//    Flags: TParamFlags;
//    ParamName: ShortString;
//    TypeName: ShortString;
//   end;
//  PRTTIParameter= ^RTTIParameter;
  PParamFlags = ^TParamFlags;

  TMethodString = procedure (const Arg: string) of object;
  TMethodBoolBool = procedure (B1, B2: Boolean) of object;
  TMethodNone = procedure of object;

function BumpParamPointer(var P: PByte; out TypeName: ShortString):TParamFlags;
var
  S: PShortString;
begin
  Result := PParamFlags(P)^;
  Inc(P, SizeOf(TParamFlags));
  S := PShortString(P); { Argument name }
  Inc(P, Length(S^)+1);
  S := PShortString(P); { Type name }
  Inc(P, Length(S^)+1);
  TypeName := S^;
end;

procedure CheckMNodeFunction(MNode: TMNodeClass; const AName: string; TypeInfo: PTypeInfo; AltTypeInfo: PTypeInfo);
var
  Data, AltData: PTypeData;
  Flags: TParamFlags;
  D, AD: PByte;
  i: Integer;
  S, S1: ShortString;
  OK: Boolean;
begin
  if (MNode.MethodAddress(AName) = nil) or (TypeInfo^.Kind <> tkMethod) then
    raise Exception.Create(AName);
  Data := GetTypeData(TypeInfo);
  D := @Data^.ParamList;
  AD := nil;
  if AltTypeInfo <> nil then begin
    AltData := GetTypeData(AltTypeInfo);
    if (AltData^.ParamCount <> Data^.ParamCount) or (AltTypeInfo^.Kind <> tkMethod) then
      raise Exception.Create(AName);
    AD := @AltData^.ParamList;
  end;
  { Check parameters - require simple value for each }
  for i := 0 to Data^.ParamCount-1 do begin
    Flags := BumpParamPointer(D, S);
    if S = {$IFDEF FPC} 'AnsiString' {$ELSE} 'string' {$ENDIF} then
      OK := Flags = [pfConst]
    else if S = 'Boolean' then
      OK := (Flags = [pfConst]) or (Flags = [])
    else
      OK := False;
    if OK and (AD <> nil) then begin
      OK := BumpParamPointer(AD, S1) = Flags;
      OK := OK and (S = S1);
    end;
    if not OK then
      raise Exception.Create(AName);
  end;
end;

procedure CallMNodeFunction(MNode: TMNode; MethodAddress: Pointer; TypeData: PTypeData; Arguments: string);
var
  Method: TMethod;
  Args: array of string;
  i, P: Integer;
  function StringArg(n: Integer):string;
  begin
    Result := Args[n];
    if (Result <> '') and (Result[1] in ['''', '"']) and (Result[Length(Result)] = Result[1]) then
      Result := Copy(Result,2,Length(Result)-2);
  end;
  function BoolArg(n: Integer):Boolean;
  begin
    if not TryStrToBoolean(Args[n],Result) then
      raise Exception.Create('Boolean not recognized:'+Args[n]);
  end;
begin
  Method.Data := MNode;
  Method.Code := MethodAddress;
  SetLength(Args, TypeData^.ParamCount);
  for i := 0 to High(Args) do begin
    if Arguments = '' then
      raise EReportException.Create('Argument missing for '+MNode.UniqueName+'.'+MNode.MethodName(MethodAddress));
    if Arguments[1] in ['''', '"'] then begin
      { Quoted string }
      P := 2;
      repeat
        while (P <= Length(Arguments)) and (Arguments[P] <> Arguments[1]) do
          Inc(P);
        Inc(P);
        { Merge repeated quote marks }
        if (P <= Length(Arguments)) and (Arguments[P] = Arguments[1]) then
          Delete(Arguments,P,1)
        else
          Break;
      until False;
      while (P <= Length(Arguments)) and (Arguments[P] = ' ') do
        Inc(P);
      if (P <= Length(Arguments)) and (Arguments[P] <> ',') then
        raise EReportException.Create('Expected comma in call to  '+MNode.UniqueName+'.'+MNode.MethodName(MethodAddress));
      if P > Length(Arguments) then
        P := -1;
    end else
      P := Pos(',',Arguments);
    if P > 0 then begin
      Args[i] := Trim(Copy(Arguments,1,P-1));
      Delete(Arguments,1,P);
      Arguments := Trim(Arguments);
    end else begin
      Args[i] := Trim(Arguments);
      Arguments := '';
    end;
  end;
  if Arguments <> '' then
    raise EReportException.Create('Too many arguments in call to '+MNode.UniqueName+'.'+MNode.MethodName(MethodAddress));
  if TypeData = GetTypeData(TypeInfo(TMethodString)) then
    TMethodString(Method)(StringArg(0))
  else if TypeData = GetTypeData(TypeInfo(TMethodBoolBool)) then
    TMethodBoolBool(Method)(BoolArg(0), BoolArg(1))
  else if TypeData = GetTypeData(TypeInfo(TMethodNone)) then
    TMethodNone(Method)()
  else
    raise Exception.Create(MNode.MethodName(MethodAddress));
end;

initialization
  { Placed here to avoid problems with @ operator in InternalCoords and TypeInfo built-in }
  TMNode.RegisterExternalFunction('AddNew', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNode.AddNew){$ENDIF});
  TMNode.RegisterExternalFunction('DeleteNode', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TMNode.DeleteNode){$ENDIF});
  TMNode.RegisterExternalFunction('DeleteDynamicParameter', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNode.DeleteDynamicParameter){$ENDIF});
  TCartesianDisplacements.RegisterExternalFunction('CalculateFromL', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TCartesianDisplacements.CalculateFromL){$ENDIF});
  TSymmetryCoords.RegisterExternalFunction('CalculateF', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TSymmetryCoords.CalculateF){$ENDIF});
  TzMatrix.RegisterExternalFunction('LoadFromState', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TzMatrix.LoadFromState){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('AddRow', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.AddRow){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('AddFRow', TypeInfo(TMethodString) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.AddFRow){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('ClearArray', TypeInfo(TMethodNone) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.ClearArray){$ENDIF});
  TMNodeWithArray.RegisterExternalFunction('ClearFArray', TypeInfo(TMethodNone) {$IFDEF FPC}, TypeInfo(@TMNodeWithArray.ClearFArray){$ENDIF});
  TInternalCoords.RegisterExternalFunction('CalculateFfromF', TypeInfo(TMethodBoolBool) {$IFDEF FPC}, TypeInfo(@TInternalCoords.CalculateFfromF){$ENDIF});
end.
