unit ObsSelect;
{$I utils.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}
interface
{ Mechanism for selecting observations by quantum number }

uses Mixture, Expr, QuantumNumber, Contnrs;

function AddSelection(var Selections: TObjectList; Selection: string; QuantumNumberFormat: TQuantumNumberFormat):string;
function CheckSelections(Obs: TObservation; Selections: TObjectList):Boolean;
function CheckSelection(var Obs: TObservation; Expr: TExpr):Boolean;

implementation

uses SysUtils, Fitter, QuantumMath, MNode, Obs, UCompat;

type
  PObservation = ^TObservation;

var
  CurrentObs: PObservation; { Global variable - not threadsafe }

type

  TComputedVariable = class (TArray)
  private
    ObsType: TObsType;
  public
    constructor Create(AName: String; AObsType: TObsType);
    function Value(var a:array of Double):Double; override;
  end;

  { TComputedQno }

  TComputedQno = class (TArray)
  private
    QnoType: TQnoType;
    Upper: Boolean;
  public
    constructor Create(AName: String; AQnoType: TQnoType; AUpper: Boolean);
    function Value(var a:array of Double):Double; override;
  end;

var
  Parser: TLex;
  Ju, Jl,
  Frequency, Weight, StdDev, SymU, SymL, IndexU, IndexL: TVariable;
  EU, EL, Intensity, Ku, Kl: TFunction;


function AddSelection(var Selections: TObjectList; Selection: string; QuantumNumberFormat: TQuantumNumberFormat):string;
var
  Expr: TExpr;
begin
  Expr := TExpr.Create(Parser);
  Result := '';
  try
    Expr.Text := Selection;
    if not Assigned(Selections) then
      Selections := TObjectList.Create;
    Selections.Add(Expr);
  except
    on E: Exception do begin
      Result := E.Message;
      Expr.Free;
    end;
  end;
end;

function CheckSelection(var Obs: TObservation; Expr: TExpr):Boolean;
begin
  CurrentObs := @Obs;
  Ju.Value := Obs.JU / 2;
  Jl.Value := Obs.JL / 2;
  Frequency.Value := Obs.Frequency;
  Weight.Value := Obs.Weight;
  StdDev.Value := Obs.Weight;
  SymU.Value := Obs.SymU;
  SymL.Value := Obs.SymL;
  IndexU.Value := Abs(Obs.IndexU);
  IndexL.Value := Abs(Obs.IndexL);
  Result := Expr.Value <> 0;
end;

function CheckSelections(Obs: TObservation; Selections: TObjectList):Boolean;
var
  i: Integer;
begin
  Result := False;
  if Assigned(Selections) then
    for i := Selections.Count - 1 downto 0 do
      if not CheckSelection(Obs, Selections[i] as TExpr) then
        Exit;
  Result := True;
end;

{ TComputedQno }

constructor TComputedQno.Create(AName: String; AQnoType: TQnoType; AUpper: Boolean);
begin
  inherited Create(AName, 0);
  QnoType := AQnoType;
  Upper := AUpper;
end;

function TComputedQno.Value(var a: array of Double): Double;
var
  Index: Integer;
  Ev: TEigenvalues;
  QnoInfo: TQnoInfo;
begin
  Index := -1;
  with CurrentObs^ do begin
    if Upper and (ObsType in [obEUpper, obLine, obCDLower, obCDUpper] ) then
      Index := FindEigenvaluesAndIndex(Ev, ManifoldU, JU, SymU, IndexU, Name, False, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
    if not Upper and (ObsType in [obELower, obLine, obCDLower, obCDUpper] ) then
      Index := FindEigenvaluesAndIndex(Ev, ManifoldL, JL, SymL, IndexL, Name, True, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
  end;
  if Index > 0 then begin
    QnoInfo := Ev.Manifold.Molecule.QnoInfo(QnoType);
    Result := Ev.DominantBasis[Index].Qno(QnoType);
    if QnoInfo.StepSize = 2 then
      Result := Result / 2;
  end else
    raise EReportException.Create(Name+' could not be evaluated');
end;

{ TComputedVariable }

constructor TComputedVariable.Create(AName: String; AObsType: TObsType);
begin
  inherited Create(AName,0);
  ObsType := AObsType;
end;

function TComputedVariable.Value(var a: array of Double): Double;
var
  s: string;
begin
  if not CalculateObs(CurrentObs^, Result, s, False, ObsType) then
    Abort;
end;

var
  i: Integer;

initialization
  Parser := TLex.Create;
  Ju := Parser.MakeVariable('J''');
  Jl := Parser.MakeVariable('J"');
  Frequency := Parser.MakeVariable('Frequency');
  Weight := Parser.MakeVariable('Weight');
  StdDev := Parser.MakeVariable('StdDev');
  SymU := Parser.MakeVariable('Sym''');
  SymL := Parser.MakeVariable('Sym"');
  IndexU := Parser.MakeVariable('Index''');
  IndexL := Parser.MakeVariable('Index"');
  EU := Parser.AddFunction(TComputedVariable.Create('E''', obEUpper));
  EL := Parser.AddFunction(TComputedVariable.Create('E"', obELower));
  Intensity := Parser.AddFunction(TComputedVariable.Create('Intensity', obIntensity));
  Ku := Parser.AddFunction(TComputedQno.Create('K''',qtA,True));
  Kl := Parser.AddFunction(TComputedQno.Create('K"',qtA,False));
finalization
  Parser.Variables.OwnsObjects := True;
  with Parser.Functions do
    for i := 0 to Count - 1 do
      Objects[i].Free;
  Parser.Free;
end.
