unit OscFields;
{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}

{ Objects under TSimulation to describe polarization of applied field }

interface

uses Classes, MNode, QuantumNumber, Mixture;

type

  TOscillatingField = class(TMNode)
  public
    constructor Create(AOwner: TComponent); override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure SetModified(const Value: Boolean); override;
  published
    property Strength:  Double index 1 read GetParameter write SetParameter;
  end;

  TDirection = (poRandom, poParallel, poPerpendicular, poCircular);

  TPolarization = class(TOscillatingField)
  private
    function GetDirection(const Index: Integer): TDirection;
    procedure SetDirection(const Index: Integer; const Value: TDirection);
  public
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
  published
    property Direction: TDirection index 1 read GetDirection write SetDirection;
  end;

function Mfactor(Jp, Mp, Rank, J, M: TQuantumNumber; Simulation:TSimulation):Double;

implementation

uses Math, MathSup, Global, nj;

{ TOscillatingField }

constructor TOscillatingField.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=1';
  DimensionlessParameters := DimensionlessParameters + ',Strength';
  inherited;
end;

function TOscillatingField.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative strength of field - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TOscillatingField.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    RootNode.ClearCache(True);
end;

{ TPolarization }

const
  DirectionNames: array [0..4] of PChar = ('Random', 'Parallel', 'Perpendicular', 'Circular', nil);

function TPolarization.GetDirection(const Index: Integer): TDirection;
begin
  Result := TDirection(GetSetting(Index));
end;

function TPolarization.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if SettingNames[AIndex] = 'Direction' then
    Result := @DirectionNames[0];
end;

procedure TPolarization.SetDirection(const Index: Integer; const Value: TDirection);
begin
  SetSetting(Index, Ord(Value));
end;

{ M dependence to transtion moment }
function Mfactor(Jp, Mp, Rank, J, M: TQuantumNumber; Simulation:TSimulation):Double;
  function Tkp(p: TQuantumNumber):Double;
  begin
    Result := Minus(Jp-Mp) * threej(Jp, -Mp, Rank, p, J, M);
  end;
  function PolFactor(Direction: TDirection):Double;
  begin
    { Choose right values of p; scale factors to give same total
      absorption independent of polarization. Scale factors come from the
      fact that the sum over a given component, p, is independent of p.
      The scale factors squared should then sum to one. }
    case Direction of
      poRandom:
        if Abs(Mp-M) <= Rank then begin
          Result := Tkp(Mp-M);
          Result := Result/Sqrt(Rank+1);
        end else
          Result := 0;
      poParallel: Result := Tkp(0);
      poPerpendicular:
        if Mp > M then
          Result := Tkp(Rank)/Sqrt(Two)
        else
          Result := Tkp(-Rank)/Sqrt(Two);
      poCircular: Result := Tkp(Rank);
    end;
  end;
var
  i: Integer;
begin
  if Assigned(Simulation) then
    if Simulation.ComponentCount > 0 then begin
      Result := 0;
      for i := 0 to Simulation.ComponentCount - 1 do
        if Simulation.Components[i] is TPolarization then
          with TPolarization(Simulation.Components[i]) do
            Result := Result + Strength * PolFactor(Direction)
    end else
      Result := PolFactor(poRandom) { Default if polarisation not specified }
  else
    Result := Tkp(0) { Normally simple static field }
end;

initialization
  TPolarization.RegisterClass(TSimulation);
end.
