unit Overlay;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} {$IFDEF MSWINDOWS} Windows, {$ENDIF} {$ENDIF}
     Units, Graphics, GraphM, GraphRange, SysUtils, Classes, MNode, Mixture, pgDOM, Math, MathSup, Peaks;

type

  { TPlot }

  TPlot = class(TStretchableMNode)
  private
    FPeaks: TPeaks;
  protected
    FDefaultName: Boolean;
    FHMin, FHMax: Double;
    function GetHMax: Double; virtual;
    function GetHMin: Double; virtual;
    function GetIMin: Double; virtual; abstract;
    function GetIMax: Double; virtual; abstract;
    procedure SetName(const Value: TComponentName); override;
  public
    Xname, Yname: string;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AdjustFrequency(Freq: Double):Double; virtual;
    procedure AddPeak(Position: Double; DisplayUnits: TResonanceUnits; PlotTag: Integer);
    function DeAdjustFrequency(Freq: Double):Double; virtual;
    function Calibrating: Boolean;
    procedure ClearSelection; virtual; abstract;
    function ConvertPeakPosition(Position: Double; PlotUnits: TResonanceUnits):Double; virtual;
    procedure CycleUnits; override;
    function Empty: Boolean; virtual; abstract;
    function FindPeak(AFmin, AFmax: Double; out Position: Double; PlotUnits: TResonanceUnits): Boolean; virtual; abstract;
    function LimitsInRange(Simulation: TSimulation; var AIMin, AImax: Double; PlotIntensityUnits: TIntensityUnits; xShift: Double):Boolean; virtual;
    procedure LoadFromFile(FileName: string; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits); virtual; abstract;
    procedure Invert; virtual; abstract;
    procedure SaveForUndo; virtual; abstract;
    procedure Undo; virtual; abstract;
    function CanUndo: Boolean; virtual;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    procedure SelectFrequencyRange(AFmin, AFmax: Double); virtual; abstract;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    procedure OffsetBy(Offset: Double); virtual;
    function Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
                  PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean;
                  ShowName: Boolean; ForceAxis: Boolean = False; xShift: Double = 0):Boolean; virtual;
    function ShowPeaks(G: TGraphM; AImin, AImax: Double; Simulation: TSimulation; xShift: Double):Boolean;
    function Resolution: Double; virtual; abstract;
    procedure WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function nPeaks: Integer;
    function Peaks: TPeaks;
    property HMin: Double read GetHMin;
    property HMax: Double read GetHMax;
    property IMin: Double read GetIMin;
    property IMax: Double read GetIMax;
    property DefaultName:Boolean read FDefaultName;
  published
    property FrequencyOffset: Double index 1 read GetParameter write SetParameter; { Frequency Offset }
    property FrequencyScale: Double  index 2 read GetParameter write SetParameter;
    { In Base: property Ymin: Double            index 3 read GetParameter write SetParameter; // Default is unset - auto }
    { In Base: property Ymax: Double            index 4 read GetParameter write SetParameter; // Default is unset - auto }

    property Units:  TResonanceUnits index 1 read GetUnits  write SetUnits;
    property IntensityUnits: TIntensityUnits index 2 read GetIntensityUnits write SetIntensityUnits;
    property Colour: TColor          index 3 read GetColour write SetColour;
    property Hide:   Boolean         index 4 read GetFlag   write SetFlag;
    property ZeroValid: Boolean      index 5 read GetFlag   write SetFlag;
    { In Base: property AutoMin:   Boolean      index 6 read GetFlag write SetFlag; }
    { In Base: property AutoMax:   Boolean      index 7 read GetFlag write SetFlag; }
    property FixCentre: Boolean      index 8 read GetFlag write SetFlag;
  end;

  TPlotClass = class of TPlot;

  { TOverlays }

  TOverlays = class(TRootMNode)
  private
    FHasLimits: Boolean;
    FnPlots: Integer;
    FPlots: TMNodeArray;
    function GetPlot(AIndex: Integer): TPlot;
    procedure SetnComponents(const Index, Value: Integer);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    HMin, HMax: Double; { Units as set below }
    Imin, Imax: Double;
    LastShowOffset: Boolean;
    constructor Create(AOwner: TComponent); override;
    procedure CycleUnits; override;
    procedure Show(G: TGraphM; Simulation: TSimulation; PlotUnits: TResonanceUnits; PlotIntensityUnits: TIntensityUnits;
                   PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean; Offset: Boolean = false);
    {$IFNDEF CONSOLE}
    function FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot; PlotUnits: TResonanceUnits; const PlotRanges: TPlotRanges; WindowTag: Integer): Integer;
    procedure AddImageFromStream(Sender: TObject; Stream: TSTream; const Ext: string);
    {$ENDIF}
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure MergeFromStream(Stream: TStream; AFileName: string = '<Stream>'); override;
    function LimitsInRange(Simulation: TSimulation; PlotUnits: TResonanceUnits; var AImin, AImax: Double; PlotIntensityUnits: TIntensityUnits; out AllArbitrary: Boolean):Boolean;
    procedure Invert;
    procedure ChooseUniqueColourFor(Plot: TPlot);
    function nVisible: Integer; { Number of non-hidden plots }
    property Plots[AIndex: Integer]: TPlot read GetPlot; default;
    procedure SaveToFile(AFileName: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure AdjustPlotFraction(var Frac:Double);
    property HasLimits: Boolean read FHasLimits;
    property nPlots: Integer read FnPlots;
  published
    property Units:          TResonanceUnits index 1 read GetUnits write SetUnits;
    property IntensityUnits: TIntensityUnits index 2 read GetIntensityUnits write SetIntensityUnits;
    property ContourFile:    Boolean index 3 read GetFlag write SetFlag; { if set, read an observation file when contour fitting, mainly for constraints }
    property nComponents:    Integer index 4 read GetSetting write SetnComponents;
    property PeakThreshold:  Double index 1 read GetParameter write SetParameter;
  end;

{ This routine adds to the given overlay, or creates the overlay if necessary;
  It return true if a new overlay hes been created }
function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits):Boolean; overload;

implementation

uses {$IFNDEF CONSOLE} Clipbrd, ExpBitmap, MNodeClip, CalFrm, LineFrm, PDFreader, LoomisWoodFrm, {$ENDIF}
     Peak, MNodeColour, ExpPlot, Global, CalSource, I2, Profile, TabbedLine, UCompat,
     MNodeExpr, LoomisWood;

var
  LoadOverlayTime: string = 'Load Overlay';

function AddToOverlay(var Overlay: TOverlays; FileName: string; Merge: Boolean; DefaultUnits: TResonanceUnits; AlwaysPrompt: Boolean; DefaultIntensityUnits: TIntensityUnits):Boolean;
var
  Plot: TPlot;
  i: Integer;
  Stream: TFileStream;
  NewOverlay: TOverlays;
  function TryLoading(PlotClass: TPlotClass): TPlot;
  begin
    Result := PlotClass.Create(NewOverlay);
    try
      Result.LoadFromFile(FileName, DefaultUnits, AlwaysPrompt, DefaultIntensityUnits);
    except
      FreeAndNil(Result);
    end;
  end;
begin
  Result := not Merge or not Assigned(Overlay);
  Plot := nil;
  try
    if Result then begin
      NewOverlay := TOverlays.Create(nil);
      NewOverlay.Name := 'Overlays';
      // if Assigned(Mixture) then
      //  Overlays.Units := Mixture.Units;
    end else
      NewOverlay := Overlay;
    TStart(LoadOverlayTime);
    i := Pos(' Calibration>',FileName);
    if (i > 0) and (i = Length(FileName)-12) then
      MakeCalibrationSource(Copy(FileName, 2, Pos(' Calibration>',FileName)-2),NewOverlay)
    else if (LowerCase(ExtractFileExt(FileName)) = '.ovr')
            or (LowerCase(ExtractFileExt(FileName)) = '.xml')
            or (LowerCase(ExtractFileExt(FileName)) = '.pgo') then begin
      Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
      try
        if Result then
          NewOverlay.LoadFromStream(Stream, FileName)
        else
          NewOverlay.MergeFromStream(Stream, FileName)
      finally
        Stream.Free;
      end;
    {$IFNDEF CONSOLE}
    end else if LowerCase(ExtractFileExt(FileName)) = '.pdf' then begin
      if NewOverlay <> Overlay then begin
        NewOverlay.Units := DefaultUnits;
        NewOverlay.IntensityUnits := DefaultIntensityUnits;
      end;
      with TPDFReader.Create(FileName) do
        try
          ForAllImages({$IFDEF FPC}@{$ENDIF}NewOverlay.AddImageFromStream);
        finally
          Free;
        end;
    {$ENDIF}
    end else {$IFNDEF CONSOLE} if (Filename <> '') or (PasteXMLNodeOnto(NewOverlay) = nil) then {$ENDIF} begin
      Plot := TryLoading(TExperimentalPlot);
      {$IFNDEF CONSOLE}
      if not Assigned(Plot) then
        Plot := TryLoading(TBitmapPlot);
      {$ENDIF}
      if not Assigned(Plot) then begin
        if FileName = '' then
          FileName := 'clipboard';
        raise EReportException.Create('Overlay Load from ' + FileName + ' failed');
      end;
//      if Result then
//        NewOverlay.IntensityUnits := Plot.IntensityUnits;
      for i := Plot.ComponentIndex to NewOverlay.nPlots-1 do
        NewOverlay.ChooseUniqueColourFor(NewOverlay.Plots[i]);
      Plot.Modified := False;
    end;
    NewOverlay.CheckProperties;
    TEnd(LoadOverlayTime);
    if Result then begin
      Overlay.Free;
      Overlay := NewOverlay;
      Overlay.Modified := False;
      PerformMNodeNotification(NewOverlay, mnNewBaseNode, 0);
    end;
  except
    if Result then
      FreeAndNil(NewOverlay)
    else
      Plot.Free;
    raise;
  end;
end;

{ TPlot }

function TPlot.DeAdjustFrequency(Freq: Double): Double;
begin
  if FixCentre then begin
    Result := (FHMax+FHMin)/2;
    Result := (Freq - Result - FrequencyOffset)/FrequencyScale + Result;
  end else
    Result := (Freq-FrequencyOffset)/FrequencyScale ;
end;

function TPlot.AdjustFrequency(Freq: Double): Double;
begin
  if FixCentre then begin
    { Choose function so that centre of plot is unaffected by FrequencyScale }
    Result := (FHMax+FHMin)/2;
    Result := (Freq-Result)*FrequencyScale + Result + FrequencyOffset;
  end else
    Result := Freq*FrequencyScale + FrequencyOffset;
end;

function TPlot.Show(G: TGraphM; Simulation: TSimulation; PlotIntensityUnits: TIntensityUnits;
                    PlotRange: TPlotRange; var PlotRanges: TPlotRanges; OwnAxis: Boolean;
                    ShowName: Boolean; ForceAxis: Boolean; xShift: Double):Boolean;
var
  xl, xr, gl, gr, x0, x1, y0, y1: Double;
  LeftJustify: Boolean;
  Justification : TAlignment;
begin
  G.MapArg(x0, x1, y0, y1);
  MapFromPlotRange(G, PlotRange);
  if ShowName then begin
    xl := Convert(HMin,Units,Simulation.PlotUnits);
    xr := Convert(HMax,Units,Simulation.PlotUnits);
    InOrder(xl,xr);
    gl := G.Xmin;
    gr := G.Xmax;
    InOrder(gl,gr);
    LeftJustify := G.Xmax > G.Xmin;
    if xl - gl > gr - xr then
      { More space on left }
      LeftJustify := not LeftJustify
    else { More space on right }
      xl := xr;
    if LeftJustify then
      Justification := taLeftJustify
    else
      Justification := taRightJustify;
    G.TextJ(xl, (PlotRange.yi+PlotRange.yx)/2, Justification, Name)
  end;
  Result := ShowPeaks(G, PlotRange.yi, PlotRange.yx, Simulation, xShift);
  G.Map(x0, x1, y0, y1);
end;

function TPlot.ShowPeaks(G: TGraphM; AImin, AImax: Double; Simulation: TSimulation; xShift: Double):Boolean;
var
  nShown: Integer;
  Mi, Mx, Fi, Fx, s, ds: Double;
  ConvertLabel: Boolean;
  LW: TLoomisWood;
begin
  G.MapArg(Fi, Fx, Mi, Mx);
  if Fi > Fx then begin
    s := Fi;
    Fi := Fx;
    Fx := s;
  end;
  Fx := Fx + xShift;
  Fi := Fi + xShift;
  with Owner as TOverlays do
    if LastShowOffset or (Count = 1) then
      G.Style(4);
  ds := (Fx-Fi)/G.Width;
  {$IFNDEF CONSOLE}
    ConvertLabel := CalibratingPlot <> Self;
    LW := ActiveLoomisWood;
    if (LW <> nil) and not ((LW.Owner = Self) or (LW.Owner = Owner)) then
      LW := nil;
  {$ELSE}
    ConvertLabel := True;
    LW := nil;
  {$ENDIF}
  nShown := Peaks.Plot(G, Fi, Fx, AImin, AImax, ds, xShift, Simulation.PlotUnits, ConvertLabel, {$IFDEF FPC}@{$ENDIF}ConvertPeakPosition, False);
  if Assigned(LW) then
    nShown := nShown + LW.Peaks.Plot(G, Fi, Fx, AImin, AImax, ds, xShift, Simulation.PlotUnits, True, {$IFDEF FPC}@{$ENDIF}ConvertPeakPosition, True);
  Result := nShown > 0;
end;

procedure TPlot.AddPeak(Position: Double; DisplayUnits: TResonanceUnits; PlotTag: Integer);
var
  Peak: TPeak;
  Lab1, Lab2, Lab: string;
begin
  Peak := EmptyPeak;
  if Calibrating then
    Peak.Position := Position
  else
    Peak.Position := DeAdjustFrequency(Position);
  Peaks.AddPeak(Peak);
  Lab1 := ExtractFileName(Name);
  if (Lab1 = '1') and ((Owner as TOverlays).nPlots = 1) then
    Lab1 := '';
  Lab2 := ExtractFileName((Owner as TMNode).Name);
  if (Lab1 <> '') and (Lab2 <> '') then
    Lab := Lab1 + ' - ' + Lab2
  else
    Lab := Lab1 + Lab2;
  PeakMeasured(Convert(Position, Units, DisplayUnits), DisplayUnits, Lab, Self, PlotTag);
end;

function TPlot.Calibrating: Boolean;
begin
  {$IFNDEF CONSOLE}
  Result := Self = CalibratingPlot;
  {$ELSE}
  Result := False;
  {$ENDIF}
end;

function TPlot.CanUndo: Boolean;
begin
  Result := False;
end;

function TPlot.ConvertPeakPosition(Position: Double; PlotUnits: TResonanceUnits): Double;
begin
  Result := Convert(AdjustFrequency(Position), Units, PlotUnits);
end;

constructor TPlot.Create(AOwner: TComponent);
begin
  Xname := 'Frequency';
  Yname := 'Intensity';
  SavedProperties := SavedProperties + ',Colour=None,AutoMin=true,AutoMax=true,Ymax=1,FrequencyScale=1';
  DimensionlessParameters := DimensionlessParameters + ',FrequencyScale';
  FPeaks := TPeaks.Create(Self);
  inherited;
  FDefaultName := True;
end;

destructor TPlot.Destroy;
begin
  FreeAndNil(FPeaks);
  inherited;
end;

function TPlot.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TPlot.LimitsInRange(Simulation: TSimulation; var AIMin, AImax: Double; PlotIntensityUnits: TIntensityUnits; xShift: Double): Boolean;
begin
  { Note - must be overriden in subclass; this routine assumes AIMin and AImax have been set }
  if AutoMin then
    ModifyParameter(IndexOfParameter('Ymin'),AIMin,False,False)
  else
    AIMin := Ymin;
  if AutoMax then
    ModifyParameter(IndexOfParameter('Ymax'),AIMax,False,False)
  else
    AImax := Ymax;
  Result := True;
end;

function TPlot.nPeaks: Integer;
begin
  Result := Peaks.nPeaks;
end;

procedure TPlot.OffsetBy(Offset: Double);
begin
  FrequencyOffset := FrequencyOffset + Offset;
end;

function TPlot.Peaks: TPeaks;
begin
  Result := FPeaks;
end;

const
  PlotsettingsNames: array [1..8] of string = (
'Horizontal Axis units.',
'Verical units of the data; select arbitrary to autoscale the plot',
'Colour for plot.',
'Set to hide this plot.',
'Set if zero in data is correct.',
'Set Ymin automatically to minimum point in plot range',
'Set Ymax automatically to maximum point in plot range',
'Apply FrequencyScale so that centre does not move if set'
);

function TPlot.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(PlotsettingsNames))
     and (AIndex <= High(PlotsettingsNames)) then
    Result := PlotsettingsNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
  PlotParameterNames: array [1..4] of string = (
'Frequency offset used when displaying plot.',
'Scaling factor for frequency; see also FixCentre',
'Lower end of current plot vertical range; overwritten if automin set',
'Upper end of current plot vertical range; overwritten if automax set'
);

function TPlot.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(PlotParameterNames))
     and (AIndex <= High(PlotParameterNames)) then
    Result := PlotParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TPlot.WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean);
begin
  { Compatibility - do not write Ymin/Ymax if autorange on }
  if AutoMin and (AIndex = IndexOfParameter('Ymin')) then
    Exit;
  if AutoMax and (AIndex = IndexOfParameter('Ymax')) then
    Exit;
  inherited;
end;

function TPlot.GetHMax: Double;
begin
  if FrequencyScale >= 0 then
    Result := AdjustFrequency(FHMax)
  else
    Result := AdjustFrequency(FHMin);
end;

function TPlot.GetHMin: Double;
begin
  if FrequencyScale < 0 then
    Result := AdjustFrequency(FHMax)
  else
    Result := AdjustFrequency(FHMin);
end;

procedure TPlot.CycleUnits;
var
  NewUnits: TResonanceUnits;
begin
  if Units = High(TResonanceUnits) then
    NewUnits := Low(TResonanceUnits)
  else
    NewUnits := Succ(Units);
  ScaleParameters(Units, NewUnits);
  Units := NewUnits;
end;

procedure TPlot.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  i: Integer;
  Peak: TPeak;
begin
  for i := 0 to nPeaks-1 do begin
    Peak := Peaks[i];
    Peak.Position := Convert(Peak.Position, UFrom, UTo);
    if not IsNan(Peak.Weight) then
      Peak.Weight := Convert(Peak.Weight, UFrom, UTo);
    Peaks[i] := Peak;
  end;
end;

procedure TPlot.SetName(const Value: TComponentName);
begin
  inherited;
  FDefaultName := Value = '';
end;

{ TOverlays }

constructor TOverlays.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',PeakThreshold=0.5';
  DimensionlessParameters := DimensionlessParameters + ',PeakThreshold=0.5';
  inherited;
end;

function TOverlays.FixupProperties:Exception;
var
  yStarted: Boolean;
  vi, vx, ix: Double;
  i: Integer;
  function GetIntensity(Plot: TPlot; Value: Double):Double;
  begin
    { Stick spectra intensities are independent of the frequency units }
    Result := ConvertIntensity(Value, (Plot.HMin+Plot.HMax)/2, Plot.Units, Plot.IntensityUnits, Units, IntensityUnits,
                               (Plot is TExperimentalPlot) and TExperimentalPlot(Plot).IsLineList);
  end;
begin
  FHasLimits := False;
  Result := inherited FixupProperties;
  FnPlots := AdjustNodeIndexes(TPlot, FPlots);
  if Assigned(Result) or (nPlots = 0) then Exit;
  yStarted := False;
  for i := 0 to nPlots-1 do
    if not (Plots[i] is TCalibrationSource) and not Plots[i].Empty and not Plots[i].Hide then begin
      vi := Convert(Plots[i].HMin, Plots[i].Units, Units);
      vx := Convert(Plots[i].HMax, Plots[i].Units, Units);
      InOrder(vi,vx);
      if not FHasLimits or (HMin > vi) then HMin := vi;
      if not FHasLimits or (HMax < vx) then HMax := vx;
      FHasLimits := True;
      if not IsArbitraryStyle(Plots[i].IntensityUnits) then begin
        ix := GetIntensity(Plots[i], Plots[i].IMin);
        if not yStarted or (IMin > ix) then IMin := ix;
        ix := GetIntensity(Plots[i], Plots[i].IMax);
        if not yStarted or (IMax < ix) then IMax := ix;
        yStarted := True;
      end;
    end;
  if not yStarted then begin
    IMin := 0; { All arbitrary units }
    IMax := 1;
  end;
  if FHasLimits and (HMin = HMax) then begin
    HMin := Max(HMin - 1,HMin/10); { Avoid zero }
    HMax := HMax + 1
  end;
end;

function TOverlays.GetPlot(AIndex: Integer): TPlot;
begin
  Result := FPlots[AIndex] as TPlot;
end;

procedure TOverlays.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  FnPlots := AdjustNodeIndexes(TPlot, FPlots, AComponent, Operation);
end;

{$IFNDEF CONSOLE}
function TOverlays.FindPeak(AArea: TDRect; var Position: Double; var Plot: TPlot;
                            PlotUnits: TResonanceUnits; const PlotRanges: TPlotRanges; WindowTag: Integer): Integer;
{ Return number of overlays scanned; < 0 indicates a peak found }
var
  i, j, PlotIndex, nFound, PlotTag: Integer;
  p, Frac, CFrac, CHMin, CHMax, xOffset: Double;
begin
  { Track fraction of plot area covered }
  Result := 0;
  Frac := 0;
  nFound := 0;
  for i := 0 to nPlots - 1 do begin
    PlotIndex := -1;
    { Find matching plot range }
    for j := 0 to High(PlotRanges) do
      if PlotRanges[j].Tag = i+10 then begin
        PlotIndex := j;
        Break;
      end;
    while (PlotIndex >= 0) and (PlotIndex <= High(PlotRanges)) and (PlotRanges[PlotIndex].Tag mod 1000 = i+10) do begin
      if RangeOverlaps(AArea.Bottom, AArea.Top, PlotRanges[PlotIndex]) then begin
        Inc(Result); { Track number of overlays considered }
        xOffset := PlotRanges[PlotIndex].xOffset;
        with Plots[i] do
          if FindPeak(AArea.Left+xOffset, AArea.Right+xOffset, p, PlotUnits) then begin
            PlotTag := PlotRanges[PlotIndex].Tag;
            if PlotTag < 1000 then
              PlotTag := PlotTag + 1000*WindowTag;
            if GetPlot(i) is TCalibrationSource then begin
              Inc(nFound);
              Plot := GetPlot(i);
              Position := p;
            end else begin
              CHMin := Convert(HMax,Units,PlotUnits);
              CHMax := Convert(HMin,Units,PlotUnits);
              InOrder(CHMin, CHMax);
              CFrac := (Min(AArea.Right,CHMax) - Max(CHMin,AArea.Left))/(AArea.Right - AArea.Left);
              if CFrac > Frac then begin
                Position := p;
                Plot := GetPlot(i);
                if CFrac > 1.5*Frac then
                  nFound := 1
                else
                  Inc(nFound);
                Frac := CFrac;
              end;
            end;
          end;
      end;
      Inc(PlotIndex);
    end;
  end;
  if (nFound = 1) or ((nFound > 1) and LastShowOffset) then begin
    Plot.AddPeak(Position, PlotUnits, PlotTag);
    Result := -Result; { Flag a peak actually found }
  end;
end;

procedure TOverlays.AddImageFromStream(Sender: TObject; Stream: TSTream; const Ext: string);
begin
  if nPlots < 50 then
    with TBitmapPlot.Create(Self) do
      try
        LoadImageFromStream(Sender, Stream, Ext);
      except
        Free;
      end;
end;
{$ENDIF}

procedure TOverlays.Show(G: TGraphM; Simulation: TSimulation; PlotUnits: TResonanceUnits;
                         PlotIntensityUnits: TIntensityUnits;
                         PlotRange: TPlotRange;  { Input: y1..y2 is screen plot range to use }
                                                 {        yi..yx is logical plot range }
                                                 { Mainly ignored if OwnAxis }
                         var PlotRanges: TPlotRanges;
                         OwnAxis: Boolean;       { If set, set compute PlotRange from overlay spectra }
                         Offset: Boolean = false { Offset individual plots } );
var
  i, m: Integer;
  yl, dy, xShift, xShift0, y2, dy2, y2O: Double;
  AllArbitrary: Boolean;
  UnitsCaption: string;
  LW: TLoomisWood;
  lPlotRange: TPlotRange;
begin
  LastShowOffset := Offset;
  if nVisible = 0 then Exit;
  yl := 0;
  if Offset then
    dy := 1/nVisible
  else
    dy := 1;
  if OwnAxis then begin
    OwnAxis := OwnAxis and ((nVisible = 1) or Offset);
    if not OwnAxis then
      if IsArbitraryStyle(PlotIntensityUnits) then begin
        { Spectra will all have same vertical appearance on plot, apart from those with AlignZero set }
        LimitsInRange(Simulation, PlotUnits, PlotRange.yi, PlotRange.yx, PlotIntensityUnits, AllArbitrary);
      end else begin
        { Show normal vertical axis, allowing spectra to have different vertical ranges }
        LimitsInRange(Simulation, PlotUnits, PlotRange.yi, PlotRange.yx, PlotIntensityUnits, AllArbitrary);
        // Was PlotRange.yi := Imin;
        // PlotRange.yx := Imax;
        if AllArbitrary then
          UnitsCaption := '' { Normalised but no real units - Show scale, but leave units blank }
        else
          { TODO: Strictly should check for mixture of linelist/non linelist styles }
          UnitsCaption := NameOfUnit(PlotIntensityUnits,False,PlotUnits);
        AxisFromPlotRange(PlotRange, G, G.Xmin, UnitsCaption);
      end;
  end;
  for i := 0 to nPlots - 1 do
    if not Plots[i].Hide then begin
      if OwnAxis then
        PlotRange := GraphRange.PlotRange(G, yl, yl+dy, 0, 1);
      G.StyleP(clBlack, G.LineStyle.Width, G.LineStyle.Pattern);
      LW := ActiveMnodeExpr(Plots[i], TLoomisWood) as TLoomisWood;
      if Assigned(LW) and (LW.nPlots > 1) then begin
        y2 := yl;
        if LW.yOffset >= 0 then
          y2O := Min(LW.yOffset,1.2)
        else
          y2O := Max(LW.yOffset,-1.2);
        dy2 := dy/(1+(LW.nPlots-1)*Abs(y2O));
        if y2O < 0 then
          y2 := yl-dy2*y2O*(LW.nPlots-1);
        xShift0 := LW.xShift(0);
        for m := 0 to LW.nPlots-1 do begin
          xShift := LW.xShift(m)-xShift0;
          lPlotRange := GraphRange.PlotRange(G, y2, y2+dy2, 0, 1);
          if Plots[i].Show(G, Simulation, PlotIntensityUnits, lPlotRange, PlotRanges, OwnAxis, Offset and (nVisible > 1), False, xShift) then begin
            PlotRanges[High(PlotRanges)].Tag := 10+i+(m*1000);
            PlotRanges[High(PlotRanges)].xOffset := xShift;
          end;
          y2 := y2 + dy2*y2O;
        end;
      end else
        if Plots[i].Show(G, Simulation, PlotIntensityUnits, PlotRange, PlotRanges, OwnAxis, Offset and (nVisible > 1)) then
          PlotRanges[High(PlotRanges)].Tag := 10+i;
      if Offset then
        yl := yl + dy;
    end;
end;

function TOverlays.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  WriteValue(Result, 'Version', VersionInfo, '');
end;

procedure TOverlays.AdjustPlotFraction(var Frac: Double);
var
  Shrink: Boolean;
  LW: TLoomisWood;
  i, nTraces: Integer;
begin
  { Shrink simulation if any active LoomisWood objects with ShrinkSimulation set }
  Shrink := False;
  nTraces := 1; { One for the simulation }
  LW := ActiveMnodeExpr(Self, TLoomisWood) as TLoomisWood;
  if Assigned(LW) then
    Shrink := Shrink or LW.ShrinkSimulation;
  for i := 0 to nPlots - 1 do
    if not Plots[i].Hide then begin
      LW := ActiveMnodeExpr(Plots[i], TLoomisWood) as TLoomisWood;
      if Assigned(LW) then begin
        Inc(nTraces, LW.nPlots);
        Shrink := Shrink or LW.ShrinkSimulation;
      end else
        Inc(nTraces);
    end;
  if Shrink then
    if Frac > 0 then
      Frac := One/nTraces
    else
      Frac := -One/nTraces;
end;

procedure TOverlays.ChooseUniqueColourFor(Plot: TPlot);
var
  i, ci: Integer;
  Clash: Boolean;
  NewColour: TColor;
begin
  for ci := 1 to High(MNodeColourValues)-3 do begin
    NewColour := MNodeColourValues[ci];
    Clash := False;
    for i := 0 to nPlots - 1 do
      if (Plots[i] <> Plot)
         and (Plots[i].Colour = NewColour) then begin
        Clash := True;
        Break;
      end;
    if not Clash then Break;
  end;
  { Can't avoid a clash, so choose a random value }
  if Clash then
    NewColour := MNodeColourValues[Random(High(MNodeColourValues)-3)+1];
  Plot.Colour := NewColour;
end;

procedure TOverlays.SaveToFile(AFileName: string);
var
  Stream: TStream;
  i, nP: Integer;
  s: string;
begin
  { Add meaningful names on initial save }
  s := ExtractFileName(AFileName);
  s := Copy(s,1,Length(s)-Length(ExtractFileExt(s)));
  for i := 0 to nPlots - 1 do
    if Plots[i].DefaultName then
      if nPlots = 1 then
        Plots[i].Name := s
      else
        Plots[i].Name := s + '-' + Items[i].Name;
  if LowerCase(ExtractFileExt(AFileName)) = '.dat' then begin
    nP := 0;
    Stream := TFileStream.Create(AFileName, fmCreate);
    try
      for i := 0 to nPlots - 1 do
        if Plots[i] is TExperimentalPlot then begin
          Inc(nP);
          if nP > 1 then
            WriteNewline(Stream);
          TExperimentalPlot(Plots[i]).SimpleSaveToStream(Stream);
        end;
    finally
      Stream.Free;
    end;
  end else
    inherited;
end;

procedure TOverlays.SetnComponents(const Index, Value: Integer);
var
  i: Integer;
begin
  SetSetting(Index,Value);
  for i := 0 to ComponentCount-1 do
    if Components[i] is TExperimentalPlot then
      TExperimentalPlot(Components[i]).AdjustParameterNames;
end;

procedure TOverlays.Invert;
var
  i: Integer;
begin
  for i := 0 to nPlots - 1 do
    Plots[i].Invert;
end;

function TOverlays.LimitsInRange(Simulation: TSimulation;
  PlotUnits: TResonanceUnits; var AImin, AImax: Double;
  PlotIntensityUnits: TIntensityUnits;
  out AllArbitrary: Boolean): Boolean;
var
  i, pass: Integer;
  yi, yx: Double;
begin
  Result := False;
  AllArbitrary := True;
  for pass := 0 to 1 do
    { Double pass arrangement allows non arbitrary unit spectra to be correctly
      scaled with respect to each other, with arbitrary ones autoscaled }
    for i := 0 to nPlots - 1 do
      if not Plots[i].Hide and ((pass = 0) xor IsArbitraryStyle(Plots[i].IntensityUnits))
         and Plots[i].LimitsInRange(Simulation, yi, yx, PlotIntensityUnits, 0) then begin
        if IsArbitraryStyle(Plots[i].IntensityUnits) then begin
          if not Result then begin
            AImin := 0;
            AImax := 1;
          end;
          { Arbitrary autoscales, but limit scaling when zero is flagged as good }
          if Plots[i].ZeroValid then
            yi := Min(yi/yx*AImax,0)
          else
            yi := AImin;
          yx := AImax;
        end else
          AllArbitrary := False;
        if Result then begin
          AImin := Min(AImin, yi);
          AImax := Max(AImax, yx);
        end else begin
          AImin := yi;
          AImax := yx;
          Result := True;
        end;
      end;
end;

var
  LoadCount: Integer = 1;
procedure TOverlays.MergeFromStream(Stream: TStream; AFileName: string);
  procedure PossiblyAddNames(S, E: Integer; const Base: string);
  var
    i: Integer;
  begin
    for i := S to E - 1 do
      if Length(Plots[i].name) < 4 then
        Plots[i].Name := Base + ' ' + Plots[i].Name;
  end;
var
  n0: Integer;
  OriginalName: string;
begin
  { Force unique names }
  n0 := nPlots;
  OriginalName := Name;
  inherited MergeFromStream(Stream, AFileName);
  if n0 = 0 then
    Exit;
  if (AFileName = '') or (AFileName = '<stream>') then begin
    AFileName := 'Plt' + IntToStr(LoadCount);
    Inc(LoadCount);
  end else
    AFileName := ExtractFileName(AFileName);
  PossiblyAddNames(0, n0 - 1, OriginalName);
  PossiblyAddNames(n0, nPlots - 1, AFileName);
end;

const
  OverlaySettingNames: array[1..3] of string = (
'Frequency/Wavelength Units for plot.',
'Intensity unitss for plot',
'Set to read an extrenal observation file when contour fitting, typically to set constraints.'
);

function TOverlays.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(OverlaySettingNames))
     and (AIndex <= High(OverlaySettingNames)) then
    Result := OverlaySettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TOverlays.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Threshold for peak finder.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TOverlays.CycleUnits;
begin
  { Note - just change top level - do not change units of individual plots }
  if Units = High(TResonanceUnits) then
    Units := Low(TResonanceUnits)
  else
    Units := Succ(Units);
end;

function TOverlays.nVisible: Integer;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to nPlots - 1 do
    if not Plots[i].Hide then
      Inc(Result);
end;

function TOverlays.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  { Version info first written in 5.2.860 }
  FReadVersion :=  DecodePgopherVersion(ReadValue(Node, 'Version', ''));
  Result := inherited ReadProperties(Node, False);
end;

initialization
  TOverlays.RegisterClass(nil);
end.

