unit Flate;

{$H+}

interface

uses
  Classes, SysUtils, PDFreader, PDFStream;

const
  flateMaxLitCodes     = 288; // max # literal codes
  flateMaxDistCodes    =  30; // max # distance codes
type

  // Huffman code table entry
  FlateCode = record
    len: Word;                  // code length, in bits
    val: Word;                  // value represented by this code
  end;
  TFlateCodeArray = array of FlateCode;
  FlateHuffmanTab = record
    codes: TFlateCodeArray;
    maxLen: Integer;
  end;

  { TFlateDecompressionStream }

  TFlateDecompressionStream = class(TPDFStream)
  private
    buf: array of Byte;
    remain, index, codeSize, blockLen: Integer;
    codeBuf: Cardinal;
    endOfBlock, compressedBlock, eofPending: Boolean;
    litCodeTab:FlateHuffmanTab;  // literal code table
    distCodeTab:FlateHuffmanTab; // distance code table
    codeLengths: array [0..flateMaxLitCodes + flateMaxDistCodes-1] of Integer; // literal and distance code lengths
    procedure ReadSome;
    procedure startBlock;
    procedure LoadFixedCodes;
    procedure readDynamicCodes;
    function getCodeWord(bits:Integer):Integer;
    function getHuffmanCodeWord(const tab: FlateHuffmanTab):Integer;
    procedure compHuffmanCodes(const lengths:array of Integer; First, n: Integer; var tab :FlateHuffmanTab);
  public
    constructor Create(Source: TStream; Params: TPDFDictionary);
    function NextByte: Byte; override;
  end;

implementation

{$IFNDEF FPC}
uses UCompat;
{$ENDIF}

type

  // Decoding info for length and distance code words
  FlateDecode = record
    bits: Integer;  // # extra bits
    first: Integer; // first length/distance
  end;

const
  flateWindow = 32768;    // buffer size
  flateMask   = flateWindow-1;
  // flateMaxHuffman      =  15; // max Huffman code length
  flateMaxCodeLenCodes =  19; // max # code length codes

  codeLenCodeMap: array [0..flateMaxCodeLenCodes-1] of Integer
    = (16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15);

  lengthDecode: array [0..flateMaxLitCodes-258] of FlateDecode = (
    (bits:0; first:  3),
    (bits:0; first:  4),
    (bits:0; first:  5),
    (bits:0; first:  6),
    (bits:0; first:  7),
    (bits:0; first:  8),
    (bits:0; first:  9),
    (bits:0; first: 10),
    (bits:1; first: 11),
    (bits:1; first: 13),
    (bits:1; first: 15),
    (bits:1; first: 17),
    (bits:2; first: 19),
    (bits:2; first: 23),
    (bits:2; first: 27),
    (bits:2; first: 31),
    (bits:3; first: 35),
    (bits:3; first: 43),
    (bits:3; first: 51),
    (bits:3; first: 59),
    (bits:4; first: 67),
    (bits:4; first: 83),
    (bits:4; first: 99),
    (bits:4; first:115),
    (bits:5; first:131),
    (bits:5; first:163),
    (bits:5; first:195),
    (bits:5; first:227),
    (bits:0; first:258),
    (bits:0; first:258),
    (bits:0; first:258)
  );

  distDecode: array [0..flateMaxDistCodes-1] of FlateDecode = (
    (bits: 0; first:    1),
    (bits: 0; first:    2),
    (bits: 0; first:    3),
    (bits: 0; first:    4),
    (bits: 1; first:    5),
    (bits: 1; first:    7),
    (bits: 2; first:    9),
    (bits: 2; first:   13),
    (bits: 3; first:   17),
    (bits: 3; first:   25),
    (bits: 4; first:   33),
    (bits: 4; first:   49),
    (bits: 5; first:   65),
    (bits: 5; first:   97),
    (bits: 6; first:  129),
    (bits: 6; first:  193),
    (bits: 7; first:  257),
    (bits: 7; first:  385),
    (bits: 8; first:  513),
    (bits: 8; first:  769),
    (bits: 9; first: 1025),
    (bits: 9; first: 1537),
    (bits:10; first: 2049),
    (bits:10; first: 3073),
    (bits:11; first: 4097),
    (bits:11; first: 6145),
    (bits:12; first: 8193),
    (bits:12; first:12289),
    (bits:13; first:16385),
    (bits:13; first:24577)
  );

  flateFixedLitCodeTabCodes: array [0..511] of FlateCode = (
    (len:7; val:$0100),
    (len:8; val:$0050),
    (len:8; val:$0010),
    (len:8; val:$0118),
    (len:7; val:$0110),
    (len:8; val:$0070),
    (len:8; val:$0030),
    (len:9; val:$00c0),
    (len:7; val:$0108),
    (len:8; val:$0060),
    (len:8; val:$0020),
    (len:9; val:$00a0),
    (len:8; val:$0000),
    (len:8; val:$0080),
    (len:8; val:$0040),
    (len:9; val:$00e0),
    (len:7; val:$0104),
    (len:8; val:$0058),
    (len:8; val:$0018),
    (len:9; val:$0090),
    (len:7; val:$0114),
    (len:8; val:$0078),
    (len:8; val:$0038),
    (len:9; val:$00d0),
    (len:7; val:$010c),
    (len:8; val:$0068),
    (len:8; val:$0028),
    (len:9; val:$00b0),
    (len:8; val:$0008),
    (len:8; val:$0088),
    (len:8; val:$0048),
    (len:9; val:$00f0),
    (len:7; val:$0102),
    (len:8; val:$0054),
    (len:8; val:$0014),
    (len:8; val:$011c),
    (len:7; val:$0112),
    (len:8; val:$0074),
    (len:8; val:$0034),
    (len:9; val:$00c8),
    (len:7; val:$010a),
    (len:8; val:$0064),
    (len:8; val:$0024),
    (len:9; val:$00a8),
    (len:8; val:$0004),
    (len:8; val:$0084),
    (len:8; val:$0044),
    (len:9; val:$00e8),
    (len:7; val:$0106),
    (len:8; val:$005c),
    (len:8; val:$001c),
    (len:9; val:$0098),
    (len:7; val:$0116),
    (len:8; val:$007c),
    (len:8; val:$003c),
    (len:9; val:$00d8),
    (len:7; val:$010e),
    (len:8; val:$006c),
    (len:8; val:$002c),
    (len:9; val:$00b8),
    (len:8; val:$000c),
    (len:8; val:$008c),
    (len:8; val:$004c),
    (len:9; val:$00f8),
    (len:7; val:$0101),
    (len:8; val:$0052),
    (len:8; val:$0012),
    (len:8; val:$011a),
    (len:7; val:$0111),
    (len:8; val:$0072),
    (len:8; val:$0032),
    (len:9; val:$00c4),
    (len:7; val:$0109),
    (len:8; val:$0062),
    (len:8; val:$0022),
    (len:9; val:$00a4),
    (len:8; val:$0002),
    (len:8; val:$0082),
    (len:8; val:$0042),
    (len:9; val:$00e4),
    (len:7; val:$0105),
    (len:8; val:$005a),
    (len:8; val:$001a),
    (len:9; val:$0094),
    (len:7; val:$0115),
    (len:8; val:$007a),
    (len:8; val:$003a),
    (len:9; val:$00d4),
    (len:7; val:$010d),
    (len:8; val:$006a),
    (len:8; val:$002a),
    (len:9; val:$00b4),
    (len:8; val:$000a),
    (len:8; val:$008a),
    (len:8; val:$004a),
    (len:9; val:$00f4),
    (len:7; val:$0103),
    (len:8; val:$0056),
    (len:8; val:$0016),
    (len:8; val:$011e),
    (len:7; val:$0113),
    (len:8; val:$0076),
    (len:8; val:$0036),
    (len:9; val:$00cc),
    (len:7; val:$010b),
    (len:8; val:$0066),
    (len:8; val:$0026),
    (len:9; val:$00ac),
    (len:8; val:$0006),
    (len:8; val:$0086),
    (len:8; val:$0046),
    (len:9; val:$00ec),
    (len:7; val:$0107),
    (len:8; val:$005e),
    (len:8; val:$001e),
    (len:9; val:$009c),
    (len:7; val:$0117),
    (len:8; val:$007e),
    (len:8; val:$003e),
    (len:9; val:$00dc),
    (len:7; val:$010f),
    (len:8; val:$006e),
    (len:8; val:$002e),
    (len:9; val:$00bc),
    (len:8; val:$000e),
    (len:8; val:$008e),
    (len:8; val:$004e),
    (len:9; val:$00fc),
    (len:7; val:$0100),
    (len:8; val:$0051),
    (len:8; val:$0011),
    (len:8; val:$0119),
    (len:7; val:$0110),
    (len:8; val:$0071),
    (len:8; val:$0031),
    (len:9; val:$00c2),
    (len:7; val:$0108),
    (len:8; val:$0061),
    (len:8; val:$0021),
    (len:9; val:$00a2),
    (len:8; val:$0001),
    (len:8; val:$0081),
    (len:8; val:$0041),
    (len:9; val:$00e2),
    (len:7; val:$0104),
    (len:8; val:$0059),
    (len:8; val:$0019),
    (len:9; val:$0092),
    (len:7; val:$0114),
    (len:8; val:$0079),
    (len:8; val:$0039),
    (len:9; val:$00d2),
    (len:7; val:$010c),
    (len:8; val:$0069),
    (len:8; val:$0029),
    (len:9; val:$00b2),
    (len:8; val:$0009),
    (len:8; val:$0089),
    (len:8; val:$0049),
    (len:9; val:$00f2),
    (len:7; val:$0102),
    (len:8; val:$0055),
    (len:8; val:$0015),
    (len:8; val:$011d),
    (len:7; val:$0112),
    (len:8; val:$0075),
    (len:8; val:$0035),
    (len:9; val:$00ca),
    (len:7; val:$010a),
    (len:8; val:$0065),
    (len:8; val:$0025),
    (len:9; val:$00aa),
    (len:8; val:$0005),
    (len:8; val:$0085),
    (len:8; val:$0045),
    (len:9; val:$00ea),
    (len:7; val:$0106),
    (len:8; val:$005d),
    (len:8; val:$001d),
    (len:9; val:$009a),
    (len:7; val:$0116),
    (len:8; val:$007d),
    (len:8; val:$003d),
    (len:9; val:$00da),
    (len:7; val:$010e),
    (len:8; val:$006d),
    (len:8; val:$002d),
    (len:9; val:$00ba),
    (len:8; val:$000d),
    (len:8; val:$008d),
    (len:8; val:$004d),
    (len:9; val:$00fa),
    (len:7; val:$0101),
    (len:8; val:$0053),
    (len:8; val:$0013),
    (len:8; val:$011b),
    (len:7; val:$0111),
    (len:8; val:$0073),
    (len:8; val:$0033),
    (len:9; val:$00c6),
    (len:7; val:$0109),
    (len:8; val:$0063),
    (len:8; val:$0023),
    (len:9; val:$00a6),
    (len:8; val:$0003),
    (len:8; val:$0083),
    (len:8; val:$0043),
    (len:9; val:$00e6),
    (len:7; val:$0105),
    (len:8; val:$005b),
    (len:8; val:$001b),
    (len:9; val:$0096),
    (len:7; val:$0115),
    (len:8; val:$007b),
    (len:8; val:$003b),
    (len:9; val:$00d6),
    (len:7; val:$010d),
    (len:8; val:$006b),
    (len:8; val:$002b),
    (len:9; val:$00b6),
    (len:8; val:$000b),
    (len:8; val:$008b),
    (len:8; val:$004b),
    (len:9; val:$00f6),
    (len:7; val:$0103),
    (len:8; val:$0057),
    (len:8; val:$0017),
    (len:8; val:$011f),
    (len:7; val:$0113),
    (len:8; val:$0077),
    (len:8; val:$0037),
    (len:9; val:$00ce),
    (len:7; val:$010b),
    (len:8; val:$0067),
    (len:8; val:$0027),
    (len:9; val:$00ae),
    (len:8; val:$0007),
    (len:8; val:$0087),
    (len:8; val:$0047),
    (len:9; val:$00ee),
    (len:7; val:$0107),
    (len:8; val:$005f),
    (len:8; val:$001f),
    (len:9; val:$009e),
    (len:7; val:$0117),
    (len:8; val:$007f),
    (len:8; val:$003f),
    (len:9; val:$00de),
    (len:7; val:$010f),
    (len:8; val:$006f),
    (len:8; val:$002f),
    (len:9; val:$00be),
    (len:8; val:$000f),
    (len:8; val:$008f),
    (len:8; val:$004f),
    (len:9; val:$00fe),
    (len:7; val:$0100),
    (len:8; val:$0050),
    (len:8; val:$0010),
    (len:8; val:$0118),
    (len:7; val:$0110),
    (len:8; val:$0070),
    (len:8; val:$0030),
    (len:9; val:$00c1),
    (len:7; val:$0108),
    (len:8; val:$0060),
    (len:8; val:$0020),
    (len:9; val:$00a1),
    (len:8; val:$0000),
    (len:8; val:$0080),
    (len:8; val:$0040),
    (len:9; val:$00e1),
    (len:7; val:$0104),
    (len:8; val:$0058),
    (len:8; val:$0018),
    (len:9; val:$0091),
    (len:7; val:$0114),
    (len:8; val:$0078),
    (len:8; val:$0038),
    (len:9; val:$00d1),
    (len:7; val:$010c),
    (len:8; val:$0068),
    (len:8; val:$0028),
    (len:9; val:$00b1),
    (len:8; val:$0008),
    (len:8; val:$0088),
    (len:8; val:$0048),
    (len:9; val:$00f1),
    (len:7; val:$0102),
    (len:8; val:$0054),
    (len:8; val:$0014),
    (len:8; val:$011c),
    (len:7; val:$0112),
    (len:8; val:$0074),
    (len:8; val:$0034),
    (len:9; val:$00c9),
    (len:7; val:$010a),
    (len:8; val:$0064),
    (len:8; val:$0024),
    (len:9; val:$00a9),
    (len:8; val:$0004),
    (len:8; val:$0084),
    (len:8; val:$0044),
    (len:9; val:$00e9),
    (len:7; val:$0106),
    (len:8; val:$005c),
    (len:8; val:$001c),
    (len:9; val:$0099),
    (len:7; val:$0116),
    (len:8; val:$007c),
    (len:8; val:$003c),
    (len:9; val:$00d9),
    (len:7; val:$010e),
    (len:8; val:$006c),
    (len:8; val:$002c),
    (len:9; val:$00b9),
    (len:8; val:$000c),
    (len:8; val:$008c),
    (len:8; val:$004c),
    (len:9; val:$00f9),
    (len:7; val:$0101),
    (len:8; val:$0052),
    (len:8; val:$0012),
    (len:8; val:$011a),
    (len:7; val:$0111),
    (len:8; val:$0072),
    (len:8; val:$0032),
    (len:9; val:$00c5),
    (len:7; val:$0109),
    (len:8; val:$0062),
    (len:8; val:$0022),
    (len:9; val:$00a5),
    (len:8; val:$0002),
    (len:8; val:$0082),
    (len:8; val:$0042),
    (len:9; val:$00e5),
    (len:7; val:$0105),
    (len:8; val:$005a),
    (len:8; val:$001a),
    (len:9; val:$0095),
    (len:7; val:$0115),
    (len:8; val:$007a),
    (len:8; val:$003a),
    (len:9; val:$00d5),
    (len:7; val:$010d),
    (len:8; val:$006a),
    (len:8; val:$002a),
    (len:9; val:$00b5),
    (len:8; val:$000a),
    (len:8; val:$008a),
    (len:8; val:$004a),
    (len:9; val:$00f5),
    (len:7; val:$0103),
    (len:8; val:$0056),
    (len:8; val:$0016),
    (len:8; val:$011e),
    (len:7; val:$0113),
    (len:8; val:$0076),
    (len:8; val:$0036),
    (len:9; val:$00cd),
    (len:7; val:$010b),
    (len:8; val:$0066),
    (len:8; val:$0026),
    (len:9; val:$00ad),
    (len:8; val:$0006),
    (len:8; val:$0086),
    (len:8; val:$0046),
    (len:9; val:$00ed),
    (len:7; val:$0107),
    (len:8; val:$005e),
    (len:8; val:$001e),
    (len:9; val:$009d),
    (len:7; val:$0117),
    (len:8; val:$007e),
    (len:8; val:$003e),
    (len:9; val:$00dd),
    (len:7; val:$010f),
    (len:8; val:$006e),
    (len:8; val:$002e),
    (len:9; val:$00bd),
    (len:8; val:$000e),
    (len:8; val:$008e),
    (len:8; val:$004e),
    (len:9; val:$00fd),
    (len:7; val:$0100),
    (len:8; val:$0051),
    (len:8; val:$0011),
    (len:8; val:$0119),
    (len:7; val:$0110),
    (len:8; val:$0071),
    (len:8; val:$0031),
    (len:9; val:$00c3),
    (len:7; val:$0108),
    (len:8; val:$0061),
    (len:8; val:$0021),
    (len:9; val:$00a3),
    (len:8; val:$0001),
    (len:8; val:$0081),
    (len:8; val:$0041),
    (len:9; val:$00e3),
    (len:7; val:$0104),
    (len:8; val:$0059),
    (len:8; val:$0019),
    (len:9; val:$0093),
    (len:7; val:$0114),
    (len:8; val:$0079),
    (len:8; val:$0039),
    (len:9; val:$00d3),
    (len:7; val:$010c),
    (len:8; val:$0069),
    (len:8; val:$0029),
    (len:9; val:$00b3),
    (len:8; val:$0009),
    (len:8; val:$0089),
    (len:8; val:$0049),
    (len:9; val:$00f3),
    (len:7; val:$0102),
    (len:8; val:$0055),
    (len:8; val:$0015),
    (len:8; val:$011d),
    (len:7; val:$0112),
    (len:8; val:$0075),
    (len:8; val:$0035),
    (len:9; val:$00cb),
    (len:7; val:$010a),
    (len:8; val:$0065),
    (len:8; val:$0025),
    (len:9; val:$00ab),
    (len:8; val:$0005),
    (len:8; val:$0085),
    (len:8; val:$0045),
    (len:9; val:$00eb),
    (len:7; val:$0106),
    (len:8; val:$005d),
    (len:8; val:$001d),
    (len:9; val:$009b),
    (len:7; val:$0116),
    (len:8; val:$007d),
    (len:8; val:$003d),
    (len:9; val:$00db),
    (len:7; val:$010e),
    (len:8; val:$006d),
    (len:8; val:$002d),
    (len:9; val:$00bb),
    (len:8; val:$000d),
    (len:8; val:$008d),
    (len:8; val:$004d),
    (len:9; val:$00fb),
    (len:7; val:$0101),
    (len:8; val:$0053),
    (len:8; val:$0013),
    (len:8; val:$011b),
    (len:7; val:$0111),
    (len:8; val:$0073),
    (len:8; val:$0033),
    (len:9; val:$00c7),
    (len:7; val:$0109),
    (len:8; val:$0063),
    (len:8; val:$0023),
    (len:9; val:$00a7),
    (len:8; val:$0003),
    (len:8; val:$0083),
    (len:8; val:$0043),
    (len:9; val:$00e7),
    (len:7; val:$0105),
    (len:8; val:$005b),
    (len:8; val:$001b),
    (len:9; val:$0097),
    (len:7; val:$0115),
    (len:8; val:$007b),
    (len:8; val:$003b),
    (len:9; val:$00d7),
    (len:7; val:$010d),
    (len:8; val:$006b),
    (len:8; val:$002b),
    (len:9; val:$00b7),
    (len:8; val:$000b),
    (len:8; val:$008b),
    (len:8; val:$004b),
    (len:9; val:$00f7),
    (len:7; val:$0103),
    (len:8; val:$0057),
    (len:8; val:$0017),
    (len:8; val:$011f),
    (len:7; val:$0113),
    (len:8; val:$0077),
    (len:8; val:$0037),
    (len:9; val:$00cf),
    (len:7; val:$010b),
    (len:8; val:$0067),
    (len:8; val:$0027),
    (len:9; val:$00af),
    (len:8; val:$0007),
    (len:8; val:$0087),
    (len:8; val:$0047),
    (len:9; val:$00ef),
    (len:7; val:$0107),
    (len:8; val:$005f),
    (len:8; val:$001f),
    (len:9; val:$009f),
    (len:7; val:$0117),
    (len:8; val:$007f),
    (len:8; val:$003f),
    (len:9; val:$00df),
    (len:7; val:$010f),
    (len:8; val:$006f),
    (len:8; val:$002f),
    (len:9; val:$00bf),
    (len:8; val:$000f),
    (len:8; val:$008f),
    (len:8; val:$004f),
    (len:9; val:$00ff)
  );

  flateFixedDistCodeTabCodes: array [0..31] of FlateCode = (
    (len:5; val:$0000),
    (len:5; val:$0010),
    (len:5; val:$0008),
    (len:5; val:$0018),
    (len:5; val:$0004),
    (len:5; val:$0014),
    (len:5; val:$000c),
    (len:5; val:$001c),
    (len:5; val:$0002),
    (len:5; val:$0012),
    (len:5; val:$000a),
    (len:5; val:$001a),
    (len:5; val:$0006),
    (len:5; val:$0016),
    (len:5; val:$000e),
    (len:0; val:$0000),
    (len:5; val:$0001),
    (len:5; val:$0011),
    (len:5; val:$0009),
    (len:5; val:$0019),
    (len:5; val:$0005),
    (len:5; val:$0015),
    (len:5; val:$000d),
    (len:5; val:$001d),
    (len:5; val:$0003),
    (len:5; val:$0013),
    (len:5; val:$000b),
    (len:5; val:$001b),
    (len:5; val:$0007),
    (len:5; val:$0017),
    (len:5; val:$000f),
    (len:0; val:$0000)
  );


{ TFlateDecompressionStream }

procedure TFlateDecompressionStream.ReadSome;
var
  code1, code2: Integer;
  len, dist: Integer;
  i, j, k: Integer;
  c: Byte;
begin

  if endOfBlock then
    startBlock;

  if compressedBlock then begin
    code1 := getHuffmanCodeWord(litCodeTab);
    //Writeln(code1);
    if code1 < 256 then begin
      buf[index] := Byte(code1);
      remain := 1;
    end else if code1 = 256 then begin
      endOfBlock := True;
      remain := 0;
    end else begin
      Dec(code1,257);
      code2 := lengthDecode[code1].bits;
      if code2 > 0 then
        code2 := getCodeWord(code2);
      len := lengthDecode[code1].first + code2;
      code1 := getHuffmanCodeWord(distCodeTab);
      code2 := distDecode[code1].bits;
      if code2 > 0 then
        code2 := getCodeWord(code2);
      dist := distDecode[code1].first + code2;
      i := index;
      j := (index - dist) and flateMask;
      for k := 0 to len-1 do begin
        buf[i] := buf[j];
        i := (i + 1) and flateMask;
        j := (j + 1) and flateMask;
      end;
      remain := len;
    end
  end else begin
    if blockLen < flateWindow then
      len := blockLen
    else
      len := flateWindow;
    i := 0;
    j := index;
    while i < len do begin
      if Stream.Read(c,1) <> 1 then begin
        endOfBlock := True;
        eofPending := True;
        Break;
      end;
      buf[j] := c;
      Inc(i);
      j := (j + 1) and flateMask
    end;
    remain := i;
    Dec(blockLen,len);
    if blockLen = 0 then
      endOfBlock := True;
  end;
end;

procedure TFlateDecompressionStream.startBlock;
var
  blockHdr, check: Integer;
  function Read2:Word;
  begin
    if Stream.Read(Result,2) <> 2 then
      raise Exception.Create('PDF:Unexpected end of Flate stream');
    Result := NtoLE(Result);
  end;
begin

  // free the code tables from the previous block
  litCodeTab.codes := Nil;
  distCodeTab.codes := Nil;

  // read block header
  blockHdr := getCodeWord(3);
  if (blockHdr and 1) <> 0 then
    eofPending := True;
  blockHdr := blockHdr shr 1;

  if blockHdr = 0 then begin
    // uncompressed block
    compressedBlock := False;
    blockLen := Read2;
    check := Read2;
    if check <> ((not blockLen) and $ffff) then
      raise Exception.Create('PDF:Bad uncompressed block length in flate stream');
    codeBuf := 0;
    codeSize := 0;
  end else if blockHdr = 1 then begin
    // compressed block with fixed codes
    compressedBlock := True;
    loadFixedCodes;
  end else if blockHdr = 2 then begin
    // compressed block with dynamic codes
    compressedBlock := True;
    readDynamicCodes;
  end else
    // unknown block type
    raise Exception.CreateFmt('PDF:Bad block type (%d) in flate stream',[blockHdr]);
  endOfBlock := False;
end;

procedure TFlateDecompressionStream.LoadFixedCodes;
  procedure Copy(var codes: TFlateCodeArray; const Source: array of FlateCode);
  var
    i: Integer;
  begin
    SetLength(codes,Length(Source));
    for i := 0 to High(Source) do
      codes[i] := Source[i];
  end;
begin
  Copy(litCodeTab.codes, flateFixedLitCodeTabCodes);
  litCodeTab.maxLen := 9;
  Copy(distCodeTab.codes, flateFixedDistCodeTabCodes);
  distCodeTab.maxLen := 5;
end;

procedure TFlateDecompressionStream.readDynamicCodes;
var
  numCodeLenCodes: Integer;
  numLitCodes: Integer;
  numDistCodes: Integer;
  codeLenCodeLengths: array [0..flateMaxCodeLenCodes-1] of Integer;
  codeLenCodeTab:FlateHuffmanTab;
  len, rept, code, i: Integer;
  procedure StuffLen;
  begin
    if i + rept > numLitCodes + numDistCodes then
      raise Exception.Create('PDF:Flate: bad repeat');
    while rept > 0 do begin
      codeLengths[i] := len;
      Inc(i);
      Dec(rept);
    end;
  end;
begin

  codeLenCodeTab.codes := Nil;

  // read lengths
  numLitCodes := getCodeWord(5) + 257;
  numDistCodes := getCodeWord(5) + 1;
  numCodeLenCodes := getCodeWord(4) + 4;
  if (numLitCodes > flateMaxLitCodes) or
      (numDistCodes > flateMaxDistCodes) or
      (numCodeLenCodes > flateMaxCodeLenCodes) then
    raise Exception.Create('PDF:Flate: bad codes');

  // build the code length code table
  for i := 0 to flateMaxCodeLenCodes-1 do
    codeLenCodeLengths[i] := 0;

  for i := 0 to numCodeLenCodes-1 do
    codeLenCodeLengths[codeLenCodeMap[i]] := getCodeWord(3);
  compHuffmanCodes(codeLenCodeLengths, 0, flateMaxCodeLenCodes, codeLenCodeTab);

  // build the literal and distance code tables
  len := 0;
  rept := 0;
  i := 0;
  while i < numLitCodes + numDistCodes do begin
    code := getHuffmanCodeWord(codeLenCodeTab);
    if code = 16 then begin
      rept := getCodeWord(2)+3;
      StuffLen;
    end else if code = 17 then begin
      rept := getCodeWord(3)+3;
      len := 0;
      StuffLen;
    end else if code = 18 then begin
      rept := getCodeWord(7)+11;
      len := 0;
      StuffLen;
    end else begin
      len := code;
      codeLengths[i] := code;
      Inc(i);
    end;
  end;
  compHuffmanCodes(codeLengths, 0, numLitCodes, litCodeTab);
  compHuffmanCodes(codeLengths, numLitCodes, numDistCodes, distCodeTab);
end;

function TFlateDecompressionStream.getCodeWord(bits: Integer): Integer;
begin
  while codeSize < bits do begin
    codeBuf := codeBuf or (RawByte shl codeSize);
    Inc(codeSize,8);
  end;
  Result := codeBuf and ((1 shl bits) - 1);
  codeBuf := codeBuf shr bits;
  Dec(codeSize,bits);
end;

function TFlateDecompressionStream.getHuffmanCodeWord(const tab: FlateHuffmanTab): Integer;
var
  Code: FlateCode;
begin
  while codeSize < tab.maxLen do begin
    codeBuf := codeBuf or (RawByte shl codeSize);
    Inc(codeSize,8);
  end;
  code := tab.codes[codeBuf and ((1 shl tab.maxLen) - 1)];
  //WriteLn('H:',codeBuf and ((1 shl tab.maxLen) - 1),' ', code.len,' ',code.val,' ',tab.maxlen);
  if (codeSize = 0) or (codeSize < code.len) or (code.len = 0) then
    raise Exception.Create('PDF:Flate:Bad Huffman');
  codeBuf := codeBuf shr code.len;
  Dec(codeSize,code.len);
  Result := code.val;
end;

procedure TFlateDecompressionStream.compHuffmanCodes(const lengths:array of Integer; First, n: Integer; var tab: FlateHuffmanTab);
var
  tabSize, len, code, code2, skip, val, i, t: Integer;
begin
  // find max code length
  tab.maxLen := 0;
  for val := 0 to n-1 do
    if lengths[First+val] > tab.maxLen then
      tab.maxLen := lengths[First+val];

  // allocate the table
  tabSize := 1 shl tab.maxLen;
  SetLength(tab.codes, tabSize);

  // clear the table
  for i := 0 to tabSize-1 do begin
    tab.codes[i].len := 0;
    tab.codes[i].val := 0;
  end;

  // build the table
  len := 1;
  code := 0;
  skip := 2;
  while len <= tab.maxLen do begin
    for val := 0 to n-1 do begin
      if lengths[First+val] = len then begin
        // bit-reverse the code
        code2 := 0;
        t := code;
        for i := 0 to len-1 do begin
          code2 := (code2 shl 1) or (t and 1);
          t := t shr 1;
        end;
        // fill in the table entries
        i := code2;
        while i < tabSize do begin
          tab.codes[i].len := len;
          tab.codes[i].val := val;
          Inc(i,skip);
        end;
        Inc(code);
      end;
    end;
    Inc(len);
    code := code shl 1;
    skip := skip shl 1;
  end;
end;

constructor TFlateDecompressionStream.Create(Source: TStream; Params: TPDFDictionary);
var
  cmf, flg: Byte;
begin
  inherited Create(Source);
  SetLength(buf, flateWindow);
  FillChar(buf[0],flateWindow,0);
  litCodeTab.codes := Nil;
  distCodeTab.codes := Nil;
  index := 0;
  remain := 0;
  codeBuf := 0;
  codeSize := 0;
  compressedBlock := False;
  endOfBlock := True;
  //eof = gTrue;

  // read header
  //~ need to look at window size?
  //endOfBlock = eof = gTrue;
  Eof := True;
  if (Stream.Read(cmf,1) <> 1) or (Stream.Read(flg,1) <> 1) then
    Exit;
  if (cmf and $0f) <> 8 then
    raise Exception.Create('PDF:Unknown compression method in flate stream');
  if (((cmf shl 8) + flg) mod 31) <> 0 then
    raise Exception.Create('PDF:Bad FCHECK in flate stream');
  if (flg and $20) <> 0 then
    raise Exception.Create('PDF:FDICT bit set in flate stream');
  Eof := False;
end;

function TFlateDecompressionStream.NextByte: Byte;
begin
  while remain = 0 do begin
    if endOfBlock and eofPending then begin
      Eof := True;
      Exit;
    end;
    ReadSome;
  end;
  Result := buf[index];
  index := (index + 1) and flateMask;
  Dec(remain);
  //WriteLn('B:',Result);
end;

end.

