unit PGWork;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Worker, MatrixVec, MNode, Mixture, Interlock, SpectrumPlot, SysUtils, MatrixElements;

type
  TWorkType = (wtSkip, wtCalculateObs, wtAddLine, wtSymbolicMatrixElement, wtCalculateObsBlock);

  PException = ^Exception;

  TAddIn = record
    Plot: TSpectrumPlot;
    LineList: TLineList;
    Simulation: TSimulation;
    FirstLine: Integer;
    Increment: Integer;
  end;

  { TPgopherWorker }

  TPgopherWorker = class(TInterlockedWorker)
  private
    WorkType: TWorkType;
    AProblem: PException;
    {wtCalculateObs}
    APObs: PObservation;
    APValue: PDouble;
    APOK: PBoolean;
    AI: TAddIn;
    {wtSymbolicMatrixElement}
    AMatrixElements: PTMatrixElements;
    AMNode: TMNodeWithVariables;
    AIndex: Integer;
    ARankWanted: Integer;
    {wtCalculateObsBlock}
    AStart, ASize: Integer;
    V: TColumnVector;
    Obs: TObservationArray;
  public
    procedure DoWork; override;
    procedure Skip;
    procedure SetProblem(E: Exception);
    procedure CalculateObs(PObs: PObservation; PValue: PDouble; POK: PBoolean);
    procedure CalculateObsBlock(PStart, PSize: Integer; AV: TColumnVector; AObs: TObservationArray; POK: PBoolean);
    procedure AddIn(Plot: TSpectrumPlot; LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
    procedure SymbolicMatrixElement(MatrixElements: PTMatrixElements; MNode: TMNodeWithVariables; Index: Integer; RankWanted: Integer; var Problem: Exception);
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Math, Fitter, Profile, Clock, MathSup, UCompat;

{ TPgopherWorker }

var
  { One per object might be fastest, but might end up with a lot of unused objects }
  AddInCriticalSection: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

procedure TPgopherWorker.AddIn(Plot: TSpectrumPlot; LineList: TLineList;
  Simulation: TSimulation; var Problem: Exception; FirstLine: Integer;
  Increment: Integer);
begin
  AI.Plot := Plot;
  AI.LineList := LineList;
  AI.Simulation := Simulation;
  AProblem := @Problem;
  AI.FirstLine := FirstLine;
  AI.Increment := Increment;
  WorkType := wtAddLine;
  StartWork;
end;

procedure TPgopherWorker.CalculateObs(PObs: PObservation; PValue: PDouble; POK: PBoolean);
begin
  APObs := PObs;
  APValue := PValue;
  APOK := POK;
  WorkType := wtCalculateObs;
  SetNan(APValue^);
  StartWork;
end;

procedure TPgopherWorker.CalculateObsBlock(PStart, PSize: Integer; AV: TColumnVector; AObs: TObservationArray; POK: PBoolean);
begin
  AStart := PStart;
  ASize := PSize;
  APOK := POK;
  V := AV;
  Obs := AObs;
  WorkType := wtCalculateObsBlock;
  StartWork;
end;

procedure TPgopherWorker.DoWork;
var
  NotUsed: string;
  LocalProblem: Exception;
  i, last, step: Integer;
begin
  NotUsed := '';
  case WorkType of
    wtSkip: { Make worker free again };
    wtCalculateObs: if not Fitter.CalculateObs(APObs^, APValue^, NotUsed, False, APObs^.ObsType) or IsNanOrInf(APValue^) then
                      APOK^ := False;
    wtCalculateObsBlock:
               {$IFDEF TWOPASS}
                 // On test, this two pass algorithm does not help
                 for Pass := 1 to 2 do begin
                   for i := AStart to Min(AStart+ASize-1, Length(Obs)-1) do begin
                     APValue := V.Item[i+1];
                     if Pass = 1 then
                       SetNan(APValue^);
                     if (Pass = 1) or IsNan(APValue^) then begin
                       ObsOK := Fitter.CalculateObs(Obs[i], APValue^, NotUsed, False, Obs[i].ObsType, Pass=2);
                       if not ObsOK and (Pass=1) then begin
                         ObsOK := Length(NotUsed) = 1;
                       end else
                         ObsOK := ObsOk and not IsNanOrInf(APValue^);
                       if not ObsOK then
                         APOK^ := False;
                     end;
                     if not APOK^ then
                       Break;
                   end;
                 end;
               {$ENDIF}
               begin
                 last := Length(Obs)-1;
                 if ASize < 0 then begin
                   step := -ASize
                 end else begin
                   step := 1;
                   last := Min(AStart+ASize-1, last)
                 end;
                 i := AStart;
                 while APOK^ and (i <= last) do begin
                   APValue := V.Item[i+1];
                   SetNan(APValue^);
                   if not Fitter.CalculateObs(Obs[i], APValue^, NotUsed, False, Obs[i].ObsType) or IsNanOrInf(APValue^) then
                       APOK^ := False;
                   Inc(i, step);
                 end;
               end;
    wtAddLine: with AI do begin
                 LocalProblem := nil;
                 Plot.AddIn(LineList, Simulation, LocalProblem, FirstLine, Increment);
                 SetProblem(LocalProblem);
               end;
    wtSymbolicMatrixElement:
               try
                 AMatrixElements^ := AMNode.DoSymbolicRotationalMatrixElement(AIndex, ARankWanted);
               except
                 on E:Exception do begin
                   {$IFDEF CONSOLE}
                     ConsiderException(Log, E);
                   {$ENDIF}
                   SetProblem(Exception(AcquireExceptionObject));
                 end;
               end;
  else
    WorkType := wtSkip;
    raise NotYet;
  end;
  WorkType := wtSkip;
  inherited;
end;

procedure TPgopherWorker.SetProblem(E: Exception);
begin
  if Assigned(E) then begin
    {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(AddInCriticalSection);
    try
      if Assigned(AProblem^) then
        E.Free
     else
       AProblem^:= E;
    finally
     {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(AddInCriticalSection);
    end;
  end;

end;

procedure TPgopherWorker.Skip;
begin
  WorkType := wtSkip;
  StartWork;
end;

procedure TPgopherWorker.SymbolicMatrixElement(MatrixElements: PTMatrixElements;
  MNode: TMNodeWithVariables; Index: Integer; RankWanted: Integer; var Problem: Exception);
begin
  AMatrixElements := MatrixElements;
  AMNode := MNode;
  AIndex := Index;
  AProblem := @Problem;
  ARankWanted := RankWanted;
  WorkType := wtSymbolicMatrixElement;
  StartWork;
end;

initialization
  {$IFDEF FPC}System.{$ENDIF}InitCriticalSection(AddInCriticalSection);
finalization
  {$IFDEF FPC}System.{$ENDIF}DoneCriticalSection(AddInCriticalSection);
end.

