unit PartFunc;
{ Methods for calculating partition functions }

{$I utils.inc}{Copyright (C) 2007-11 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Menus, {$ENDIF}
     Math, MathSup, MNode, Mixture, SysUtils, ExpPlot, pgDOM, Classes;

type

  { TDefaultPartitionFunction }

  TDefaultPartitionFunction = class(TPartitionFunction)
  public
    function CalculatePartitionFunction(Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double; var TheSpinSpeciesAbundance: TDoubleArray;
                                        EMax: Double = MaxDouble; Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false; RemoveSpins: Byte = 0):Double; override;
  end;

  { TInterpolatedPartitionFunction }

  TInterpolatedPartitionFunction = class(TDefaultPartitionFunction)
  private
    FPlot: TExperimentalPlot; { Slight cheat - use frequency to hold T, Intensity to hold Value }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function CalculatePartitionFunction(Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double; var SpinSpeciesAbundance: TDoubleArray;
                                        EMax: Double = MaxDouble; Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false; RemoveSpins: Byte = 0):Double; override;
    function EnergyZero(EstimatedOrigin: Double):Double; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function Plot: TExperimentalPlot;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    {$IFNDEF CONSOLE}
    procedure AddHandlers; override;
    procedure LoadFromFileClick(Sender: TObject);
    {$ENDIF}
  published
    property Zero:    Double index 1 read GetParameter write SetParameter;
    property Scale:   Double index 2 read GetParameter write SetParameter;

    property Order:    Integer index 2 read GetSetting write SetSetting;
    property AutoZero: Boolean index 3 read GetFlag write SetFlag;
    property Multiply: Boolean index 4 read GetFlag write SetFlag;
  end;

implementation

uses {$IFNDEF CONSOLE} Forms, Dialogs, ExpPlotLoader, Main, {$ENDIF}
     UCompat, PropHelp, QuantumNumber, Worker, Progress, StaticField, Units, Profile;

var
  PartitionFunctionTime: string = 'Partition Function';

{ TDefaultPartitionFunction }

function TDefaultPartitionFunction.CalculatePartitionFunction(
  Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double;
  var TheSpinSpeciesAbundance: TDoubleArray;
  EMax: Double;
  Print: Boolean; ShowMolecule: Boolean; ShowHeader: Boolean;
  RemoveSpins: Byte): Double;
var
  Q, T: Double;
  i, l, Pass: Integer;
  J, JLow, JUp: TQuantumNumber;
  Sym: Integer;
  Ev: TEigenvalues;
  nInitialManifolds, nLevels: Integer;
  LastSums: array [0..8] of Double; { Record last 9 partial sums }
  ManifoldSums: array of Double;
  LastIndex, IncludeTspin: Integer;
  Converged: Boolean;
  SpinSpeciesSum, SpinSpeciesPartSum, TspinSum: array of Double;
  MatElemTypes: TMatElemTypes;
  AnyVal: Boolean;
begin
  TStart(PartitionFunctionTime);
  MatElemTypes := [meHrot, meAllDerivatives, meFloatedDerivatives];
  Emin := MaxDouble;
  with Molecule do begin
    SetLength(SpinSpeciesSum,nSpinSpecies);
    SetLength(SpinSpeciesPartSum,nSpinSpecies);
    for Pass := 1 to 2 do begin
      Result := 0;
      nInitialManifolds := 0;
      if Length(TheSpinSpeciesAbundance) = 0 then begin
        IncludeTspin := 0;
        if Pass = 1 then
          Continue; { Two passes only required when using spin temperature }
      end else
        IncludeTspin := Pass;
      for i := 0 to nSpinSpecies - 1 do
        SpinSpeciesSum[i] := 0;
      if Print then begin
        for i := 0 to nManifolds-1 do
          if Manifold[i].Initial then
            Inc(nInitialManifolds);
        if (nInitialManifolds > 0) and ShowHeader then begin
          if Length(TheSpinSpeciesAbundance) > 0 then
            if Pass = 1 then
              WriteLn(Log,'Pass 1 - using Temperature = Tspin')
            else
              WriteLn(Log,'Pass 2 - using Required Temperature');
          Write(Log, 'Partition function sum ');
          if not ShowMolecule then
            Write(Log, 'for ', Name);
          if Length(TheSpinSpeciesAbundance) > 0 then
            if Pass = 1 then
              WriteLn(Log, ' at Temperature = ', Simulation.Tspin:6:3, ' K, Tvib = ', Simulation.Tvib:6:3, ' K')
            else
              WriteLn(Log, ' at Temperature = ', Simulation.Temperature:6:3, ' K, Tvib = ', Simulation.Tvib:6:3, ' K')
          else
            WriteLn(Log, ' at Temperature = ', Simulation.Temperature:6:3, ' K, Tvib = ', Simulation.Tvib:6:3, ' K, Tspin = ', Simulation.Tspin:6:3, ' K');
          if ActualAssumedOrigin <> 0 then
            WriteLn(Log, 'Using origin of ', ActualAssumedOrigin:25:Mixture.Precision);
          if ShowMolecule then
            Write(Log, Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
          if nInitialManifolds > 1 then
            Write(Log, Copy('Manifold',1,Length(Mixture.LongestManifoldName)):Length(Mixture.LongestManifoldName), ' ');
          Write(Log, '   J               Total        Contribution   nLevels');
          if nSpinSpecies > 1 then
            for i := 1 to nSpinSpecies do
              Write(Log, ' Spin Species ', i);
          WriteLn(Log);
        end;
      end;
      Overflow := nil;
      SetLength(ManifoldSums,nManifolds);
      for i := 0 to nManifolds-1 do
        if Manifold[i].Initial then begin
          ManifoldSums[i] := 0;
          JUp := Manifold[i].FindJmax(-1, RemoveSpins);
          if Simulation.HasFields then
            JLow:= -JUp
          else
            JLow := Manifold[i].FindJmin(HalfSpin, RemoveSpins);
          PushProgress('Partition Function for ' + Manifold[i].Name, JUp { don't div 2 as don't know, so be pessimistic });
          try
            LastIndex := 0;
            if (Manifold[i].UsePopParams or (Simulation.Temperature <= 0)) and (Manifold[i].nPopulations = 0) then
              Continue; { No population specifed }
            if (nWorkers > 0) and not Print and not Mixture.ShowDerivatives and (Mixture.PrintLevel <= plLineList) then begin
              { Precalculate (most of) the energy levels required }
              J := JLow;
              while J <= JUp do begin
                Sym := Manifold[i].FirstSym;
                while Sym >= 0 do begin
                  Manifold[i].Eigenvalues(J, Sym, MatElemTypes, Simulation.ExField, False, plDefault, nil, True, RemoveSpins);
                  if Simulation.HasFields then
                    Break;
                  Sym := Manifold[i].NextSym(Sym);
                end;
                SetProgress(J div 2);
                Inc(J, 2);
              end;
            end;
            J := JLow;
            { Turn off automatic J range if HasFields on }
            Converged := not (Manifold[i].AutoQConverge and not Simulation.HasFields);
            while not Converged or (J <= JUp) do begin
              Sym := Manifold[i].FirstSym;
              Q := 0;
              nLevels := 0;
              for l := 0 to nSpinSpecies - 1 do
                SpinSpeciesPartSum[l] := 0;
              while Sym >= 0 do begin
                if J <= Jup then
                  Ev := Manifold[i].Eigenvalues(J, Sym, MatElemTypes, Simulation.ExField, True, plDefault, nil, True, RemoveSpins)
                else { Only energy levels required for higher states, as transitions will not be calculated }
                  Ev := Manifold[i].Eigenvalues(J, Sym, [], Simulation.ExField, True, plDefault, nil, True, RemoveSpins);
                if (Ev.Size > 0) and (Ev.Energy[1] < Emin) then
                  Emin := Ev.Energy[1];
                for l := 1 to Ev.Size do begin
                  if Ev.Energy[l] > Emax then
                    Continue;
                  { Abs required as population function is allowed to return negative values }
                  T := Ev.Degeneracy(l) * Abs(Manifold[i].Population(Ev, l, Simulation, Overflow, IncludeTspin));
                  Q := Q + T;
                  Inc(nLevels);
                  SpinSpeciesPartSum[Ev.SpinSpecies(l)-1] := SpinSpeciesPartSum[Ev.SpinSpecies(l)-1] + T;
                  if not Assigned(Overflow) then
                    if IsNanOrInf(T) then
                      Overflow := EReportException.Create('Overflow in Boltzmann factor at J='+QuantumNumberToStr(J) +', E='+DtosM(Ev.Energy[l],0,Mixture.Precision)
                                                          +' for '+Ev.DominantLabel(l,Mixture.QuantumNumberFormat)+' in '+Manifold[i].UniqueName)
                    else if IsNanOrInf(Q) then
                      Overflow := EReportException.Create('Overflow in partition function at J='+QuantumNumberToStr(J)+' for '+Manifold[i].UniqueName);
                  if Assigned(Overflow) then begin
                    SetNan(Result);
                    Overflow.Message := Overflow.Message + ' - fail in partition function';
                    if Print then
                      WriteLn(Log, Overflow.Message);
                    Exit;
                  end;
                end;
                Sym := Manifold[i].NextSym(Sym);
                if Simulation.HasFields then
                  Break;
              end;
              Result := Result + Q;
              ManifoldSums[i] := ManifoldSums[i] + Q;
              for l := 0 to nSpinSpecies - 1 do
                SpinSpeciesSum[l] :=  SpinSpeciesSum[l] + SpinSpeciesPartSum[l];
              if (Result <> 0) and not Converged then begin
                Converged := (LastIndex > High(LastSums))
                             and not ((Result - LastSums[LastIndex mod Length(LastSums)])/ Result > 1e-4);
                LastSums[LastIndex mod Length(LastSums)] := Result;
                Inc(LastIndex);
              end else if (J > 30) and (Result = 0) and Manifold[i].AutoQConverge then begin
                SetNan(Result);
                if not Assigned(Overflow) then
                  Overflow := EReportException.Create('Failure in partition function calculation - zero sum at J=15 (underflow in Boltzmann factor?)');
                Exit;
              end;
              if Print and (nLevels > 0) then begin
                if ShowMolecule then
                  Write(Log, Manifold[i].Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
                if nInitialManifolds > 1 then
                  Write(Log, Manifold[i].Name:Length(Mixture.LongestManifoldName), ' ');
                Write(Log, QuantumNumberToStr(J):4, DtosMA(Result,20,8), DtosMA(Q,20,8), nLevels:10);
                if nSpinSpecies > 1 then
                  for l := 0 to nSpinSpecies - 1 do
                    Write(Log,DtosMA(SpinSpeciesPartSum[l],15,8));
                WriteLn(Log);
              end;
              SetProgress(J div 2);
              Inc(J, 2);
              if IsNanOrInf(Result) then
                Break;
            end;
          finally
            PopProgress;
          end;
        end;
      if Print and (nInitialManifolds > 1) then begin
        WriteLn(Log);
        WriteLn(Log, 'Contribution from each manifold');
        for i := 0 to nManifolds - 1 do
          if Manifold[i].Initial then
            WriteLn(Log, Manifold[i].Name:Length(Mixture.LongestManifoldName), DtosMA(ManifoldSums[i],20,8));
      end;
      if Print and (nSpinSpecies > 1) then begin
        WriteLn(Log);
        WriteLn(Log, 'Contribution from each spin species');
        for i := 0 to nSpinSpecies - 1 do
          WriteLn(Log, i+1:4, DtosMA(SpinSpeciesSum[i],20,8));
      end;
      if Result = 0 then
        Exit;
      if Pass = 1 then begin
        SetLength(TspinSum,nSpinSpecies);
        for i := 0 to nSpinSpecies - 1 do
          TspinSum[i] := SpinSpeciesSum[i]/Result; { Save fractional abundance }
      end else if Length(TheSpinSpeciesAbundance) > 0 then begin
        Q := 0; { New partition function sum }
        for i := 0 to nSpinSpecies - 1 do begin
          if SpinSpeciesSum[i] <> 0 then
            TheSpinSpeciesAbundance[i] := TspinSum[i]*Result/SpinSpeciesSum[i]
          else if TspinSum[i] <> 0 then begin
            SetNan(Result);
            if not Assigned(Overflow) then
              Overflow := EReportException.Create('Failure in partition function calculation - zero sum for spin species '+IntToStr(i));
            Exit;
          end else
            TheSpinSpeciesAbundance[i] := 0;
          Q := Q + TheSpinSpeciesAbundance[i]*SpinSpeciesSum[i];
        end;
        Result := Q;
        if Print then begin
          WriteLn(Log);
          WriteLn(Log, 'Adjust for Tspin = ', Simulation.Tspin:6:3, ' K, Partition Function = ', DtosMA(Result,20,8));
          WriteLn(Log, 'Contribution from each spin species');
          WriteLn(Log, '   #          Fraction            Weighting           Partition Sum');
          for i := 0 to nSpinSpecies - 1 do
            WriteLn(Log, i+1:4, DtosMA(TspinSum[i],20,8), DtosMA(TheSpinSpeciesAbundance[i],20,8), DtosMA(TheSpinSpeciesAbundance[i]*SpinSpeciesSum[i],20,8));
        end;
      end;
    end;
  end;
  TEnd(PartitionFunctionTime);
end;

{ TInterpolatedPartitionFunction }

function TInterpolatedPartitionFunction.Plot: TExperimentalPlot;
begin
  if not Assigned(FPlot) then begin
    FPlot := TExperimentalPlot.Create(nil);
    FPlot.Xname := 'T';
    FPlot.Yname := 'Q';
  end;
  Result := FPlot;
end;

constructor TInterpolatedPartitionFunction.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Order=4,Scale=1,AutoZero=True';
  inherited Create(AOwner);
end;

destructor TInterpolatedPartitionFunction.Destroy;
begin
  FreeAndNil(FPlot);
  inherited Destroy;
end;

{$IFNDEF CONSOLE}
procedure TInterpolatedPartitionFunction.AddHandlers;
begin
  inherited;
  AddHandler('Load...', 'Load from file', {$IFDEF FPC}@{$ENDIF}LoadFromFileClick);
end;

procedure TInterpolatedPartitionFunction.LoadFromFileClick(Sender: TObject);
begin
  with TOpenDialog.Create(Application) do
    try
      Filter := 'Text File (*.txt;*.dat;*.csv)|*.txt;*.dat;*.csv|All Files|'+AllFilesMask;
      Title := 'Load partition function';
      if Execute then begin
        Plot.LoadFromFile(FileName,cm1,MainForm.PromptonImportItem.Checked,Arbitrary);
        Plot.Sort(True);
        Modified := True;
        PerformMNodeNotification(Self, mnPropNameChange, -1);
      end;
    finally
      Free;
    end;
end;
{$ENDIF}

function TInterpolatedPartitionFunction.CalculatePartitionFunction(
  Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double;
  var SpinSpeciesAbundance: TDoubleArray;
  EMax: Double;
  Print: Boolean; ShowMolecule: Boolean; ShowHeader: Boolean; RemoveSpins: Byte): Double;
var
  i, j, il, iu: Integer;
  Product, Correction, Calculated: Double;
begin
  if RemoveSpins <> 0 then
    raise NotYet;
  { TODO: Handle SpinSpeciesAbundance? }
  Overflow := nil;
  Correction := 1;
  if AutoZero or Multiply then begin
    Calculated := inherited CalculatePartitionFunction(Molecule, Simulation, Overflow, Emin, SpinSpeciesAbundance, Emax, Print, ShowMolecule, ShowHeader);
    if AutoZero then
      Correction := exp(-Emin/Molecule.BoltzmannConstant/Simulation.Temperature);
  end;
  i := 0;
  while (i < Length(Plot.Frequency)) and (Simulation.Temperature > Plot.Frequency[i]) do
    Inc(i);
  { Lagrangian interpolation }
  { Use a symmetrical number of points around the selected region }
  il := Max(i - Order div 2,0);
  iu := Min(i + Order div 2 - 1, High(Plot.Frequency));
  Result := 0;
  for i := il to iu do begin
    Product := 1;
    for j := il to iu do
      if i <> j then
        Product := Product * (Simulation.Temperature - Plot.Frequency[j])/(Plot.Frequency[i]-Plot.Frequency[j]);
    Result := Result + Product*Plot.Intensity[i];
  end;
  if il > iu then
    Result := 1; { Default to 1 if empty }
  Result := Result * Scale * Correction;
  if Multiply then
    Result := Calculated*Result;
  if Print then begin
    if ShowHeader then
      Write(Log, 'Partition function sum ');
    if not ShowMolecule then
      Write(Log, 'for ', Molecule.Name);
    WriteLn(Log, ' at Temperature = ', Simulation.Temperature:6:3, ' K: Q = ', Result:8:5);
    if Zero <> 0 then
      WriteLn(Log, 'Using Energy Zero of ', Zero:25:Molecule.Mixture.Precision);
    if AutoZero then
      WriteLn(Log, 'Using a correction factor of ',FloatToStr(Correction),' to account for Emin = ', Emin:25:Molecule.Mixture.Precision);
    if Scale <> 1 then
      WriteLn(Log, 'Scale Factor ', Scale);
    Write(Log, 'Using ', iu-il+1 ,' point Lagrangian interpolation over: ');
    for i := il to iu do
      Write(Log, Plot.Frequency[i]:6:3, ' K, Q= ', Plot.Intensity[i]:8:5, '   ');
    WriteLn(Log);
  end;
end;

function TInterpolatedPartitionFunction.EnergyZero(EstimatedOrigin: Double): Double;
begin
  Result := Zero;
end;

const
PartitionFunctionParameterNames: array [1..2] of string = (
'Energy zero of the tabulated partition function',
'Scale interpolated partition function by this'
);

function TInterpolatedPartitionFunction.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(PartitionFunctionParameterNames))
     and (AIndex <= High(PartitionFunctionParameterNames)) then
    Result := PartitionFunctionParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TInterpolatedPartitionFunction.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    2: Result := 'Order of Lagrangian Interpolation to use (should be even)';
    3: Result := 'Set to use lowest calculated energy level as energy zero.';
    4: Result := 'Partition function Qcalc*Qinterpolated, rather than simply Qinterpolated';
  else
    Result:=inherited HelpOnSetting(AIndex);
  end;
end;

function TInterpolatedPartitionFunction.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  s: string;
  Fractional: Boolean;
  i: Integer;
begin
  s := ReadValue(Node, 'Fractional', 'False');
  Result := inherited ReadProperties(Node, False);
  if Assigned(Node.FirstChild) and (Node.FirstChild is TDOMCDATASection) then begin
    Plot.ReadData(TDOMCDATASection(Node.FirstChild));
    Node.RemoveChild(Node.FirstChild);
  end;
  if TryStrToBoolean(s, Fractional) and Fractional then begin
    { Backward compatibility }
    Multiply := True;
    with Plot do
      for i := 0 to High(Intensity) do
        Intensity[i] := 1/Intensity[i];
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TInterpolatedPartitionFunction.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  Result.AppendChild(Plot.WriteBasicData(Parent));
end;

initialization
  TInterpolatedPartitionFunction.RegisterClass(TMolecule);
end.

