unit Peaks;
{ Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

uses Math, pgDOM, Peak, GraphM, Units, MNode;

type

  TPeakChanged = procedure (PeakNumber: Integer) of object;
  TConvertPeakPosition = function (Position: Double; PlotUnits: TResonanceUnits): Double of object;

  { TPeaks }

  TPeaks = class
  private
    FPeaks: array of TPeak;
    function GetnPeaks: Integer;
    function GetPeak(AIndex: Integer): TPeak;
    procedure SetnPeaks(Value: Integer);
    procedure SetPeak(AIndex: Integer; const Value: TPeak);
    function GetActual(AIndex: Integer): Double;
  public
    Owner: TMNode;
    OnPeakChanged: TPeakChanged;
    constructor Create(AOwner: TMNode);
    procedure AddPeak(Peak: TPeak);
    procedure Changed(PeakNumber: Integer);
    procedure Clear;
    procedure DeletePeak(AIndex:Integer);
    procedure SortPeaks(Field:Integer; Ascending: Boolean);
    procedure ReadPeaks(Child: TDOMNode);
    procedure WritePeaks(Result: TDOMNode);
    procedure SetActual(AIndex: Integer; const Value: Double);
    function Plot(G: TGraphM; Fi, Fx, Imin, Imax, ds, xShift: Double; PlotUnits: TResonanceUnits;
                  ConvertLabel: Boolean; ConvertPeakPosition: TConvertPeakPosition; ShowActual: Boolean):Integer;
    function OwnerPlot: TMnode;
    function Valid(AIndex:Integer):Boolean;
    property Actual[AIndex:Integer]: Double read GetActual write SetActual;
    property Peaks[AIndex:Integer]: TPeak read GetPeak write SetPeak; default;
    property nPeaks: Integer read GetnPeaks write SetnPeaks;
  end;

var
  DummyPeaks: TPeaks;

implementation

uses SysUtils, Classes, UCompat, TabbedLine, MathSup, ShellSort, Overlay;

{ TPeaks }

procedure TPeaks.AddPeak(Peak: TPeak);
begin
  SetLength(FPeaks, Length(FPeaks)+1);
  SetPeak(High(FPeaks), Peak);
end;

procedure TPeaks.Clear;
begin
  nPeaks := 0;
end;

constructor TPeaks.Create(AOwner: TMNode);
begin
  Owner := AOwner;
  inherited Create;
end;

procedure TPeaks.DeletePeak(AIndex: Integer);
var
  i: Integer;
begin
  if (AIndex < 0) or (AIndex >= nPeaks) then
    Exit;
  for i := AIndex to nPeaks - 2 do
    FPeaks[i] := FPeaks[i+1];
  nPeaks := nPeaks - 1;
end;

procedure TPeaks.Changed(PeakNumber: Integer);
begin
  PerformMNodeNotification(Owner, mnPeakChanged, PeakNumber);
  if Assigned(OnPeakChanged) then
    OnPeakChanged(PeakNumber);
end;

function TPeaks.GetActual(AIndex: Integer): Double;
begin
  Result := FPeaks[AIndex].Actual;
end;

function TPeaks.GetnPeaks: Integer;
begin
  Result := Length(FPeaks);
end;

procedure TPeaks.SetActual(AIndex: Integer; const Value: Double);
begin
  FPeaks[AIndex].Actual := Value;
  Changed(AIndex);
end;

procedure TPeaks.SetnPeaks(Value: Integer);
var
  OldNPeaks, i: Integer;
begin
  OldNPeaks := nPeaks;
  Value := Max(Value,0);
  if Value <> OldNPeaks then begin
    SetLength(FPeaks, Value);
    for i := OldNPeaks to Value - 1 do
      Peaks[i] := EmptyPeak;
    Changed(-1);
  end;
end;

function TPeaks.GetPeak(AIndex: Integer): TPeak;
begin
  Result := FPeaks[AIndex]
end;

function TPeaks.OwnerPlot: TMnode;
begin
  Result := nil;
  if Assigned(Owner) then
    if Owner is TPlot then
      Result := Owner
     else
       Result := Owner.Owner as TPlot;
end;

function TPeaks.Valid(AIndex: Integer): Boolean;
var
  Peak: TPeak;
begin
  Peak := Peaks[AIndex];
  Result := not(IsNanOrInf(Peak.Actual) or IsNanOrInf(Peak.Position) or IsNanOrInf(Peak.Weight)) and (Peak.Weight > 0);
end;

function TPeaks.Plot(G: TGraphM; Fi, Fx, Imin, Imax, ds, xShift: Double; PlotUnits: TResonanceUnits;
                     ConvertLabel: Boolean; ConvertPeakPosition: TConvertPeakPosition; ShowActual: Boolean):Integer;
var
  i: Integer;
  Peak: TPeak;
  s: Double;
  Lab: string;
begin
  Result := 0;
  for i := 0 to nPeaks-1 do begin
    Peak := Peaks[i];
    if IsNanOrInf(Peak.Position) then
      Continue;
    s := Peak.Position;
    if Assigned(ConvertPeakPosition) then
      s := ConvertPeakPosition(Peak.Position, PlotUnits);
    if (s >= Fi) and (s <= Fx) then begin
      if not ConvertLabel then
        Lab := FloatToStr(Peak.Position)
      else
        Lab := RoundedStr(s, ds);
      if ShowActual and not IsNanOrInf(Peak.Actual) then
        Lab := FloatToStr(Peak.Actual) + ' ' + Lab;
      s := s - xShift;
      G.ClipPoint(s, Imin);
      G.ClipJoin(s, Imax);
      G.TextAngle(90);
      G.TextJ(S, Imax, taRightJustify, Lab);
      Inc(Result);
    end;
  end;
  G.TextAngle(0);
end;

procedure TPeaks.SetPeak(AIndex: Integer; const Value: TPeak);
begin
  FPeaks[AIndex] := Value;
  Changed(AIndex);
end;

type
  TPeakSorter = class (TShellSorter)
    Peaks: TPeaks;
    Fld: Integer;
    Asc: Boolean;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

{ TPeakSorter }

function TPeakSorter.InOrder(I, J: Integer): Boolean;
var
  PI, PJ: PDouble;
begin
  PI := PDouble(@Peaks.FPeaks[I]);
  Inc(PI,Fld);
  PJ := PDouble(@Peaks.FPeaks[J]);
  Inc(PJ,Fld);
  { Relies on contiguous allocation; explicit check in SortPeaks }
  if IsNanOrInf(PI^) then
    Result := IsNanOrInf(PJ^)
  else if IsNanOrInf(PJ^) then
    Result := True
  else if Asc then
    Result := PI^ <= PJ^
  else
    Result := PI^ >= PJ^
end;

procedure TPeakSorter.Swap(I, J: Integer);
begin
  SwapValues(Peaks.FPeaks[I],Peaks.FPeaks[J],Sizeof(TPeak));
end;

procedure TPeaks.SortPeaks(Field: Integer; Ascending: Boolean);
begin
  with TPeakSorter.Create do begin
    if SizeOf(TPeak) <> 4*SizeOf(Double) then
      raise Exception.Create('TPeaks.SortPeaks');
    Peaks := Self;
    Fld := Min(Max(Field,0),3);
    Asc := Ascending;
    Sort(nPeaks);
    Free;
    Changed(-1);
  end;
end;

procedure TPeaks.ReadPeaks(Child: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
  Peak: TPeak;
  NextChild: TDOMNode;
begin
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if Child is TDOMCDATASection then begin
      Strings := TStringList.Create;
      try
        Strings.Text := TDOMCDATASection(Child).Data;
        Child.ParentNode.RemoveChild(Child);
        for i := 0 to Strings.Count-1 do
          if TryStrToPeak(Strings[i], Peak) then begin
            if not IsNan(Peak.Position) then
              AddPeak(Peak);
          end else
            raise EReportException.Create('Bad line in peak list:'+Strings[i]);
      finally
        Strings.Free;
      end;
    end;
    Child := NextChild;
  end;
end;

procedure TPeaks.WritePeaks(Result: TDOMNode);
var
  Strings: TStringList;
  i: Integer;
begin
  if nPeaks = 0 then
    Exit;
  Strings := TStringList.Create;
  try
    Strings.Capacity := nPeaks+1;
    Strings.Add('');
    for i := 0 to nPeaks-1 do
      Strings.Add(PeakToString(Peaks[i]));
    Result.AppendChild(Result.OwnerDocument.CreateCDATASection(Strings.Text));
  finally
    Strings.Free;
  end;
end;

initialization
  DummyPeaks := TPeaks.Create(nil);
finalization
  FreeAndNil(DummyPeaks);
end.
