unit PopFunc;
{ Methods for calculating custom populations }

{$I utils.inc}{Copyright (C) 2010 C M Western; see commandline.pas for license}
interface

uses Classes, SysUtils, MNode, Mixture, Constrain, Expr, pgDOM, Units;

type

  { TCustomPopulationFunction }

  TCustomPopulationFunction = class(TPopulationFunction)
  { Custom population, using expression taken from comment field }
  private
    ResultVariable, EVariable, kBVariable, JVariable, SymVariable, IndexVariable,
    AVariable, BVariable, MJVariable, EBVariable, OriginVariable: TVariable;
    AFactor, BFactor: Double;
    ExtraVariables: TVariableArray;
  protected
    procedure UpdateVariables; override;
    procedure SetUpVariables; override;
  public
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure Population(var Result: Double; Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer); override;
  end;

implementation

uses QuantumMath, StrUtils, Worker, QuantumNumber, MNodeExpr;

{ TCustomPopulationFunction }

function TCustomPopulationFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to generate the partition function'
  else
    Result:=inherited HelpOnSetting(AIndex);
end;

procedure TCustomPopulationFunction.UpdateVariables;
var
  QI: TQnoInfo;
begin
  QI := (Owner as TManifold).Molecule.QnoInfo(qtA);
  AVariable.Name := AnsiReplaceStr(QI.Name, '|', '');
  if QI.StepSize = 2 then
    AFactor := 2
  else
    AFactor := 1;
  QI := (Owner as TManifold).Molecule.QnoInfo(qtB);
  BVariable.Name := QI.Name;
  if QI.StepSize = 2 then
    BFactor := 2
  else
    BFactor := 1;
end;

procedure TCustomPopulationFunction.SetUpVariables;
begin
  inherited;
  V('Result', ResultVariable);
  TriggerVariable := ResultVariable;
  V('E',EVariable);
  V('kB',kBVariable);
  kBVariable.Value := (RootNode as TMixtureMNode).BoltzmannConstant;
  V('J',JVariable);
  V('Symmetry',SymVariable);
  V('Index',IndexVariable);
  V('A',AVariable);
  V('B',BVariable);
  V('MJ', MJVariable);
  V('EB', EBVariable);
  V('Origin', OriginVariable);
  with AncestorWithClass(TMolecule) as TMolecule do
    SetUpPopulationFunctionVariables(Self, ExtraVariables);
end;

class function TCustomPopulationFunction.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := MNode is TManifold;
end;

procedure TCustomPopulationFunction.Population(var Result: Double;
  Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation;
  var Overflow: Exception; IncludeTspin: Integer);
var
  State: TBasisState;
begin
  { Preset variables }
  ResultVariable.Value := Result;
  EVariable.Value := Eigenvalues.E[i];
  State := Eigenvalues.DominantBasis[i];
  MJVariable.Value := Eigenvalues.J/2;
  JVariable.Value := State.J/2;
  SymVariable.Value := State.Sym;
  IndexVariable.Value := i;
  AVariable.Value := State.Qno(qtA)/AFactor;
  BVariable.Value := State.Qno(qtB)/BFactor;
  OriginVariable.Value := State.State.EstimatedStateOrigin;
  EBVariable.Value := State.State.Molecule.ActualAssumedOrigin;
  if High(ExtraVariables) > 0 then
    with AncestorWithClass(TMolecule) as TMolecule do
      SetPopulationFunctionVariables(Self, ExtraVariables, State);
  if (nDebug < 0) or (nDebug <= nDebugOutput) then
    DebugTag[ThreadIndex-1] := ' ' + State.StateLabel(qftJ);
  Result := Evaluate;
end;

initialization
  RegisterClass(TCustomPopulationFunction);
  TTable.RegisterClass(TCustomPopulationFunction);
end.
