unit QuantumMath;

interface

uses QuantumNumber;

type

  TNuclearInfo = record
    Name: String;
    Spin: TQuantumNumber;
    AsNext: Boolean;
  end;

function rl(J, K, Kp: Integer): Double;
function rlx(n: Integer; J, K, Kp: TQuantumNumber):Double;
// Not used: function rly(n: Integer; J, K, Kp: TQuantumNumber):Double;
function rlyRe(n: Integer; J, K, Kp: TQuantumNumber):Double; { Matrix element of rly; multiply by i^n for complex value }
function ReducedMatrixElement(Rank, J: TQuantumNumber): Double;

function DeltaJName(DeltaJ: TQuantumNumber): Char;
function IsDeltaJName(C: Char; out ADeltaJ: TQuantumNumber):Boolean;

type
  TQnoType = (qtJ=0, qtN, qtF, qtM, qtA, qtB, qtF1, qtF2, qtF3, qtF4, qtF5, qtF6, qtF7, qtF8, qtF9);
  TQnoInfo = record
    MinValue: TQuantumNumber;
    MaxValue: TQuantumNumber;
    StepSize: Integer;
    Name: string;
  end;

const
  QnoTypeNames: array [0..15] of PChar = ('J', 'N', 'F', 'M', 'A', 'B', 'F1', 'F2', 'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', nil);
  QnoExprNames: array [0..30] of PChar = ('J', 'J(J+1)', 'N', 'N(N+1)', 'F', 'F(F+1)', 'M', 'M^2', 'A', 'A^2', 'B', 'B^2',
                                          'F1', 'F1(F1+1)', 'F2', 'F2(F2+1)', 'F3', 'F3(F3+1)', 'F4', 'F4(F4+1)', 'F5', 'F5(F5+1)',
                                          'F6', 'F6(F6+1)', 'F7', 'F7(F7+1)', 'F8', 'F8(F8+1)', 'F9', 'F9(F9+1)', nil);
  NullQnoInfo: TQnoInfo = (MinValue:0; MaxValue:0; StepSize:2; Name:'');

procedure MergeQno(var Result: TQnoInfo; QnoInfo: TQnoInfo; QnoType: TQnoType);
function MapQnoExpr(ExprName: Integer; Value: Double):Double;
function MapQnoExprName(ExprName: Integer; Name: string):string;

implementation

uses SysUtils, Math, MathSup, nj;

function IsTriangle(J1, J2, J3: TQuantumNumber): Boolean;
{ Do J1, J2, J3 satisfy the triangle rule for angular momentum addition? }
begin
  if Odd(J1+J2+J3) then
    raise Exception.CreateFmt('Triangle: %d %d %d', [J1, J2, J3]);
  Result := (Abs(J1-J2) <= J3) and (J1+J2 >= J3);
end;

procedure MatchOdd(A, B: TQuantumNumber);
begin
  if Odd((A xor B) and 1) then
    raise Exception.CreateFmt('Mix of integral and half integral quantum numbers: %d %d', [A, B]);
end;

function SmallestProjection(A: TQuantumNumber): TQuantumNumber;
begin
  Result := A and 1;
end;

procedure BadMinus(A: TQuantumNumber);
begin
  { Do not inline this function }
  raise Exception.CreateFmt('Minus(%d)', [A]);
end;

function Minus(A: TQuantumNumber): Integer; { (-1)**A }
begin
  if (A and 1) <> 0 then
    BadMinus(A);
  if (A and 2) <> 0 then
    Result := -1
  else
    Result := 1;
end;

function rl(J, K, Kp: Integer): Double;
{ Work out matrix element of arbitrary power of raising/lowering operator }
var
  JSq: Double;
begin
  Result := 1;
  JSq := J*(J+2) / Four;
  while K < Kp do begin
    Result := Result * (JSq - K*(K+2)/Four);
    Inc(K, 2);
  end;
  while K > Kp do begin
    Result := Result * (JSq - K*(K-2)/Four);
    Dec(K, 2);
  end;
  Result := Sqrt(Result);
end;

function rlxy(n: Integer; J, K, Kp: TQuantumNumber; Sign: Integer):Double;
{ Helper function for matrix elements of Jx^n and Jy^n }
{ TODO: This is likely to be inefficient for large values of n }
begin
  Result := 0;
  if 2*n < Abs(K-Kp) then
    Exit;
  if 2*n = Abs(K-Kp) then
    { Easy case - equivalent to all raising/all lowering operator }
    if n = 0 then
      Result := 1
    else begin
      Result := rl(J,K,Kp) / IntPower(2,n);
      if (K < Kp) and Odd(n) then
        Result := Sign * Result;
    end
  else begin
    { Need to consider combinations;
      e.g. Jx^2 can give DeltaK = 0, +/-2 Matrix elements }
    if Abs(Kp+2) <= J then
      Result := Result + rlxy(n-1,J,K,Kp+2,Sign)*rl(J,Kp+2,Kp)/2;
    if Abs(Kp-2) <=J then
      Result := Result + Sign * rlxy(n-1,J,K,Kp-2,Sign)*rl(J,Kp-2,Kp)/2;
  end;
end;

function rlx(n: Integer; J, K, Kp: TQuantumNumber):Double;
{ Matrix elements of Jx^n }
begin
  Result := rlxy(n,J,K,Kp,1);
end;

function rlyRe(n: Integer; J, K, Kp: TQuantumNumber):Double;
{ Matrix element of rly; multiply by i^n for complex value }
begin
  Result := rlxy(n,J,K,Kp,-1);
end;

function rly(n: Integer; J, K, Kp: TQuantumNumber):Double;
{ Matrix elements of Jy^n; for odd powers the result is multiplied by i }
begin
  Result := rlxy(n,J,K,Kp,-1);
  { Multiply by i^n }
  if Odd(n) then
    Result := Minus(n+1)*Result { Multiply by a bonus i to give a real result }
  else
    Result := Minus(n)*Result;
end;

function ReducedMatrixElement(Rank, J: TQuantumNumber): Double;
{ Return the reduced matrix element <J||T^k(J)||J> formed by
  coupling J with J k times. }
var
  i: Integer;
begin
  { Brown and Carrington 5.134 }
  { K!sqrt((2J+K+1)!/(2^K(2K)!(2J-K)!) }
  Result := IntPower(2,-(Rank div 2));
  for I := J - Rank div 2 + 1 to J + Rank div 2 + 1 do
    Result := Result * I;     { (2J+K+1)!/(2J-K)! }
  for I := 2 to Rank do       { 2K! }
    Result := Result / I;
  Result := Sqrt(Result);
  for I := 2 to Rank div 2 do  { K! }
    Result := Result * I;
end;

function DeltaJName(DeltaJ: TQuantumNumber): Char;
begin
  if Abs(DeltaJ) > (Ord('Z')-Ord('Q'))*2 then
    Result := '?'
  else
    Result := Char(Ord('Q') + DeltaJ div 2);
end;

function IsDeltaJName(C: Char; out ADeltaJ: TQuantumNumber):Boolean;
var
 DeltaJ: TQuantumNumber;
begin
  C := LowerCase(C){$IFNDEF FPC}[1]{$ENDIF};
  DeltaJ := (Ord(C) - Ord('q'))*2;
  Result := Abs(DeltaJ) <= (Ord('Z')-Ord('Q'))*2;
  if Result then
    ADeltaJ := DeltaJ;
end;

function BinomialCoefficient(n, k: Integer):Integer;
var
  i: Integer;
begin
  { n!/((n-k)!k!) }
  if k > n-k then  { Make k small }
    k := n-k;
  Result := 1;
  for i := n-k+1 to n do
    Result := Result * i;
  for i := 2 to k do
    Result := Result div i;
end;

function TryStrToQuantumNumber(const str: string; var Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):Boolean;
var
  Code: Integer;
  tail: string;
begin
  case QuantumNumberFormat of
    qft2J: Result := TryStrToInt(str, Value);
    qftJ:
      begin
        Result := True;
        Val(str, Value, Code);
        if Code > 0 then begin
          tail := Copy(str,Code,2);
          if tail = '/2' then
            { Do Nothing }
          else if tail = '.5' then
            Value := Value * 2 + 1
          else
            Result := False;
        end else
          Value := Value * 2
      end;
    qftJph:
      begin
        Result := TryStrToInt(str, Value);
        if Result then
          Value := 2*Value - 1;
      end;
    qftJmh:
      begin
        Result := TryStrToInt(str, Value);
        if Result then
          Value := 2*Value + 1;
      end;
  end;
end;

function QuantumNumberToStr(const Value:TQuantumNumber; QuantumNumberFormat: TQuantumNumberFormat):string;
begin
  case QuantumNumberFormat of
    qft2J: Result := IntToStr(Value);
    qftJ:  begin
             Result := IntToStr(Value div 2);
             if Odd(Value) then
               Result := Result + '.5';
           end;
    qftJph: Result := IntToStr((Value + 1) div 2);
    qftJmh: Result := IntToStr((Value - 1) div 2);
  end;
end;

procedure MergeQno(var Result: TQnoInfo; QnoInfo: TQnoInfo; QnoType: TQnoType);
begin
  if Result.Name = '' then
    Result := QnoInfo
  else if QnoInfo.Name <> '' then begin
    if Result.Name <> QnoInfo.Name then
      Result.Name := QnoTypeNames[Ord(QnoType)];
    if Result.MinValue <> QnoInfo.MinValue then begin
      if Odd(Result.MinValue+QnoInfo.MinValue) then
        Result.StepSize := 1;
      Result.MinValue := Min(Result.MinValue, QnoInfo.MinValue);
    end;
    Result.MaxValue := Max(Result.MaxValue, QnoInfo.MaxValue);
    Result.StepSize := Min(Result.StepSize, QnoInfo.StepSize);
  end;
end;

function MapQnoExpr(ExprName: Integer; {Index into QnoExprNames }
                    Value: Double      {(Doubled) Quantum number value}
                    ):Double;
begin
  if Odd(ExprName) then
    if TQnoType(ExprName div 2) in [qtM, qtA, qtB] then
      Result := Sqr(Value)/4
    else
      Result := Value*(Value+2)/4
  else
    Result := Value/2;
end;

function MapQnoExprName(ExprName: Integer; Name: string):string;
begin
  if Odd(ExprName) then
    if TQnoType(ExprName div 2) in [qtM, qtA, qtB] then
      Result := Name+'^2'
    else
      Result := Name+'('+Name+'+1)'
  else
    Result := Name;
end;

end.
