unit Settings;
{$I utils.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}
interface

uses Classes, Mixture, Forms, pgDOM;

type

  { TStringMNode }

  { Allows any setting to be a string value; GetSetting and SetSetting will thus not necessarily work }

  TSettingType = (stInteger, stDouble, stString);

  TStringMNode = class(TStreamableMNode)
  private
    FSettingsValue: array of string;
    FSettingsType: array of TSettingType;
    FSettingsHints: array of string;
    FParametersHints: array of string;
    function GetSettingsType(AIndex: Integer): TSettingType;
    function GetParametersHints(AIndex: Integer): string;
    function GetSettingsHints(AIndex: Integer): string;
    procedure SetParametersHints(AIndex: Integer; const Value: string);
    procedure SetSettingsHints(AIndex: Integer; const Value: string);
  protected
    procedure SetSettingsType(AIndex: Integer; const AValue: TSettingType);
    procedure SetnParameters(AValue: Integer); override;
    procedure SetnSettings(AValue: Integer); override;
  public
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
    function AddSetting(AName: string; SettingType: TSettingType; Default: string):Integer; overload;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function IsSettingEnum(AIndex: Integer):Boolean; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;

    property SettingsType[AIndex: Integer]: TSettingType read GetSettingsType;
    property SettingsHints[AIndex: Integer]: string read GetSettingsHints write SetSettingsHints;
    property ParametersHints[AIndex: Integer]: string read GetParametersHints write SetParametersHints;
  end;

  { TFormSettings }

  { Provides basic Form geometry settings }

  TFormSettings = class (TStringMNode)
  public
    procedure LoadFromForm(Form: TCustomForm; SavePosition: Boolean);
    function RestoreForm(Form: TCustomForm):Boolean;
  published
    property Left:   Integer index 1 read GetSetting write SetSetting;
    property Top:    Integer index 2 read GetSetting write SetSetting;
    property Width:  Integer index 3 read GetSetting write SetSetting;
    property Height: Integer index 4 read GetSetting write SetSetting;
  end;

implementation

uses Types, SysUtils, WCompat, StdCtrls, Worker, MathSup, MNode, WhatsThis;

{ TFormSettings }

procedure TFormSettings.LoadFromForm(Form: TCustomForm; SavePosition: Boolean);
begin
  if Form.Visible and SavePosition then begin  
    Left := Form.Left;
    Top := Form.Top;
    Width := Form.Width;
    Height := Form.Height;
  end else begin
    Left := 0;
    Top := 0;
    Height := 0;
    Width := 0;
  end;
end;

const
  Mpx: Integer = 50; { Minimum number of pixels that must show }
function TFormSettings.RestoreForm(Form: TCustomForm): Boolean;
var
 W: TRect;
begin
  W := ScreenAreaFor(nil);
  Result := (Width >= Mpx) and (Height >= Mpx)
            and (Left + Mpx < W.Right)
            and (Left + Width - Mpx > W.Left)
            and (Top + Height - Mpx > W.Top)
            and (Top + Mpx < W.Bottom);
  { Only restore position if it is on screen }
  if Result then begin
    Form.SetBounds(Left,Top,Width,Height);
    if Form is THelpForm then
      THelpForm(Form).HasBeenArranged := True;
  end;
end;

{ TStringMNode }

function TStringMNode.GetParametersHints(AIndex: Integer): string;
begin
  Result := FParametersHints[AIndex];
end;

function TStringMNode.GetSettingsHints(AIndex: Integer): string;
begin
  Result := FSettingsHints[AIndex]
end;

function TStringMNode.GetSettingsType(AIndex: Integer): TSettingType;
begin
  Result := FSettingsType[AIndex];
end;

function TStringMNode.HelpOnParameter(AIndex: Integer): string;
begin
  Result := FParametersHints[AIndex];
end;

function TStringMNode.HelpOnSetting(AIndex: Integer): string;
begin
  Result := FSettingsHints[AIndex];
end;

function TStringMNode.IsSettingEnum(AIndex: Integer): Boolean;
begin
  Result := False;
end;

function TStringMNode.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Index, i: Integer;
  Child, NextChild: TDOMElement;
  Attr: TDOMAttr;
begin
  Result := inherited ReadProperties(Node, False);
  { Read Parameters }
  Child := Node.FirstChild as TDOMElement;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling as TDOMElement;
    if CompareText(Child.NodeName, 'Parameter') = 0 then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      if Assigned(Attr) then begin
        Index := IndexOfParameter(Attr.Value);
        if Index > 0 then begin
          Result := Result + ReadFloatableDOMNode(Index,FloatToStr(DefaultParameters[Index]),Child);
          Node.RemoveChild(Child);
        end;
      end;
    end;
    Child := NextChild;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

procedure TStringMNode.SetSettingsHints(AIndex: Integer; const Value: string);
begin
  FSettingsHints[AIndex] := Value;
end;

procedure TStringMNode.SetSettingsType(AIndex: Integer; const AValue: TSettingType);
begin
  FSettingsType[AIndex] := AValue;
end;

procedure TStringMNode.SetnParameters(AValue: Integer);
begin
  inherited;
  SetLength(FParametersHints, nParameters+1);
end;

procedure TStringMNode.SetnSettings(AValue: Integer);
var
  OldValue, i: Integer;
begin
  OldValue := nSettings;
  inherited SetnSettings(AValue);
  SetLength(FSettingsValue, nSettings+1);
  SetLength(FSettingsType, nSettings+1);
  SetLength(FSettingsHints, nSettings+1);
  for i := OldValue+1 to nSettings do
    if OldValue < 0 then
      SetSettingsType(i, stInteger)
    else
      SetSettingsType(i, stString); { Dynamic parameters must default to string }
end;

procedure TStringMNode.SetParameter(AIndex: Integer; const Value: Double);
begin
  inherited ModifyParameter(AIndex, Value, False);
end;

procedure TStringMNode.SetParametersHints(AIndex: Integer; const Value: string);
begin
  FParametersHints[AIndex] := Value;
end;

function TStringMNode.SettingToStr(AIndex: Integer): string;
begin
  if SettingsType[AIndex] = stInteger then
    Result := IntToStr(GetSetting(AIndex))
  else
    Result := FSettingsValue[AIndex];
end;

procedure TStringMNode.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  inherited ModifySetting(AIndex, AValue, False);
end;

procedure TStringMNode.SetSettingFromStr(AIndex: Integer; const s: string);
begin
  if SettingsType[AIndex] = stInteger then begin
    SetSetting(AIndex, StrToInt(s));
    FSettingsValue[AIndex] := SettingToStr(AIndex);
  end else if FSettingsValue[AIndex] <> s then begin
    FSettingsValue[AIndex] := s;
    SetSetting(AIndex, 1-GetSetting(AIndex)); { Trigger updates }
  end;
end;

function TStringMNode.AddSetting(AName: string; SettingType: TSettingType;
  Default: string): Integer;
var
   RValue: Double;
begin
  case SettingType of
    stInteger: Result := AddSetting(AName, Default);
    stDouble:  begin
                if not TryStrToFloatEx(Default, RValue) then
                  SetNan(RValue);
                Result := AddParameter(AName, RValue);
                SetParameter(Result, RValue);
                Exit; { Avoid setting settings stuff }
               end;
    stString:  Result := AddSetting(AName, Default);
  end;
  FSettingsType[Result] := SettingType;
end;

initialization
  TFormSettings.RegisterClass(TMixture);
end.

