unit TranFunc;
{ Methods for calculating custom transition moments }

{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}
interface

uses SysUtils, Expr, MNode, Mixture, MathSup;

type

  { TCustomTransitionFunction }

  TCustomTransitionFunction = class(TTransitionFunction)
  { Custom population, using expression taken from comment field }
  private
    ValueVariable, mVariable,
    JpVariable, JppVariable, NpVariable, NppVariable, FpVariable, FppVariable,
    ApVariable, AppVariable, BpVariable, BppVariable, PhasepVariable, PhaseppVariable: TVariable;
    ExtraVariables: TVariableArray;
    AFactor, BFactor: Double;
  protected
    procedure UpdateVariables; override;
    procedure SetUpVariables; override;
  public
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
    procedure MatrixElement(var Value: Complex; ABra, AKet: TBasisState; SignA, SignB: Integer; Intensity: Boolean); override;
  end;

implementation

uses Classes, StrUtils, UCompat, Worker, QuantumNumber, QuantumMath, MNodeExpr;

{ TCustomTransitionFunction }

procedure TCustomTransitionFunction.UpdateVariables;
var
  Molecule: TMolecule;
  QI: TQnoInfo;
begin
  Molecule := (Owner.Owner as TTransitionMoments).Molecule;
  QI := Molecule.QnoInfo(qtA);
  ApVariable.Name := AnsiReplaceStr(QI.Name, '|', '')+'''';
  AppVariable.Name := AnsiReplaceStr(QI.Name, '|', '')+'"';
  if QI.StepSize = 2 then
    AFactor := 2
  else
    AFactor := 1;
  QI := Molecule.QnoInfo(qtB);
  BpVariable.Name := QI.Name+'''';
  BppVariable.Name := QI.Name+'"';
  if QI.StepSize = 2 then
    BFactor := 2
  else
    BFactor := 1;
end;

procedure TCustomTransitionFunction.SetUpVariables;
begin
  inherited;
  V('Value', ValueVariable);
  TriggerVariable := ValueVariable;
  V('J''', JpVariable);
  V('J"', JppVariable);
  V('N''', NpVariable);
  V('N"', NppVariable);
  V('F''', FpVariable);
  V('F"', FppVariable);
  V('A''',ApVariable);
  V('B''',BpVariable);
  V('A"',AppVariable);
  V('B"',BppVariable);
  V('Phase''', PhasepVariable);
  V('Phase"', PhaseppVariable);
  V('m',mVariable);
  with AncestorWithClass(TMolecule) as TMolecule do
    SetUpTransitionFunctionVariables(Self, ExtraVariables);
end;

class function TCustomTransitionFunction.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := MNode is TTransitionMoment;
end;

procedure TCustomTransitionFunction.MatrixElement(var Value: Complex; ABra, AKet: TBasisState; SignA, SignB: Integer; Intensity: Boolean);
var
  Ratio: Double;
  IsComplex: Boolean;
begin
  { Preset variables }
  JpVariable.Value := ABra.J / 2;
  JppVariable.Value := AKet.J / 2;
  // Don't use: mVariable.Value := (ABra.N-AKet.N)*(ABra.N+AKet.N+2)/8;
  case ABra.N-AKet.N of
     2: mVariable.Value := AKet.N/2 + 1;
    -2: mVariable.Value := -AKet.N/2;
  else
    mVariable.Value := 0;
  end;
  NpVariable.Value := ABra.N / 2;
  NppVariable.Value := AKet.N / 2;
  FpVariable.Value := ABra.F / 2;
  FppVariable.Value := AKet.F / 2;
  ApVariable.Value := ABra.Qno(qtA)/AFactor;
  BpVariable.Value := ABra.Qno(qtB)/BFactor;
  AppVariable.Value := AKet.Qno(qtA)/AFactor;
  BppVariable.Value := AKet.Qno(qtB)/BFactor;
  PhasepVariable.Value := ABra.Phase;
  PhaseppVariable.Value := AKet.Phase;
  if High(ExtraVariables) > 0 then
    with AncestorWithClass(TMolecule) as TMolecule do
      SetTransitionFunctionVariables(Self, ExtraVariables, ABra, AKet, SignA, SignB);
  { As TExpr does not handle complex numbers (yet), only act on real or imaginary part }
  if Abs(Value.im) > Abs(Value.re) then begin
    ValueVariable.Value := Value.im;
    Ratio := Value.re/Value.im;
    IsComplex := True;
  end else begin
    ValueVariable.Value := Value.re;
    if Value.re <> 0 then
      Ratio := Value.im/Value.re
    else
      Ratio := 0;
    IsComplex := False;
  end;
  if Abs(Ratio) > 1e-14  then
    raise EReportException.Create('Complex custom transition matrix element');
  if (nDebug < 0) or (nDebug <= nDebugOutput) then
    DebugTag[ThreadIndex-1] := ' ' + ABra.StateLabel(qftJ) + ' - ' + AKet.StateLabel(qftJ);
  Value.re := Evaluate;
  if IsComplex then begin
    Value.im := Value.re;
    Value.re := 0;
  end;
end;

initialization
  RegisterClass(TCustomTransitionFunction);
  TTable.RegisterClass(TCustomTransitionFunction);
end.
