unit Expr;
{$I utils.inc}
{ TODO: For sorting to work, must not use locale specific sorting }
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF FPC} DynLibs, {$ENDIF}
     MathSup, UCompat, Classes, SysUtils;
{.$DEFINE CHECKSIMPLIFY}
{.$DEFINE SHOWALLCHANGES}

const
  MaxLoopDepth = 3;
  MaxStackDepth = 50;

type


  toType = (toNumber, toVariable, toOperator, toSpecial, toUnknown,
            toMonadicFunction, toDiadicFunction, toPVariable, toThreadVariable, toStore, toThreadStore, toDrop, toDup,
            toIterationStart, toIterationNext,
            toArrayLow, toArrayLow2, toArrayLow3, toArrayLow4, toArrayLow5, toArrayLowMax,
            toArrayHigh, toArrayHigh2, toArrayHigh3, toArrayHigh4, toArrayHigh5, toArrayHighMax,
            toArrayStore, toStartLoop, toEndLoop, toLoopVar,
            toFunction, toFunction1, toFunction2, toFunction3);
            { Implicit toFunction4, .... encodes number of arguments; these must be last }

  TSimplifyAction = (siNoCombineSqrt, siExpand, siNone, siSplitDivide, siMinusOneInteger, siForce);
  TSimplifyActions = set of TSimplifyAction;


  EExprError = class(Exception);
  EExprFormatError = class(EExprError);

  TEvaluator = class;
  TTokenNode = class;

  TEvaluatorOption = (evoZeroOverZero {0/0 => 0}, evoPureImag { sqrt(x) -> sqrt(-x) } );
  TEvaluatorOptions = set of TEvaluatorOption;

  { TExprNode }

  TExprNode = class
  private
    FChildren: array of TExprNode;
    FDependsOnValid, FSortTextHighValid, FSortTextLowValid: Boolean;
    FDependsOn, FSortTextHigh, FSortTextLow: string;
    function GetChildren(AIndex: Integer): TExprNode; inline;
    procedure SetChildren(AIndex: Integer; const Value: TExprNode); inline;
  protected
    function SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions = []):TExprNode; virtual;
  public
    Base: TTokenNode;
    Simplified: TSimplifyActions; { Last simplification method }
    {$IFDEF CHECKSIMPLIFY}
    InitialValue:Double;
    InitialText: string;
    procedure SetInitialText;
    {$ENDIF}
    procedure Changed;
    constructor Create(ABase: TTokenNode);
    destructor Destroy; override;
    procedure AddNode(ANode: TExprNode);
    procedure DeleteNode(AIndex: Integer);
    function MoveChild(Index: Integer):TExprNode;
    function Text(Readable:Boolean = False): string; virtual; abstract;
    procedure CheckUnchanged(Expr: TExprNode); {$IFNDEF CHECKSIMPLIFY} inline; {$ENDIF}
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; overload; virtual;
    function Compile(EvaluatorOptions: TEvaluatorOptions = []): TEvaluator; overload;
    function EvalSortText(ConstantsHigh:Boolean = False): string; virtual;
    function SortText(ConstantsHigh:Boolean = False): string;
    function Simplify(Actions:TSimplifyActions=[]):TExprNode; virtual;
    function IsConstant(out Value: Double):Boolean; virtual;
    function Duplicate:TExprNode; virtual;
    function Compare(Node: TExprNode; ConstantsHigh: Boolean):Integer; virtual; { 1 for Self in greater, 0 for equal }
    function EvalDependsOn: String; virtual;
    function DependsOn: String;
    function nNodes:Integer;
    function RealValue:Double; virtual; abstract;
    function EqualTo(AValue: Integer):Boolean;
    function nChildren:Integer; inline;
    property Children[AIndex:Integer]: TExprNode read GetChildren write SetChildren;
  end;

  { TIntegerNode }

  TIntegerNode = class(TExprNode)
    Value: Integer;
    constructor Create(AValue: Integer);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
    function Compare(Node: TExprNode; ConstantsHigh: Boolean):Integer; override;
    function Text(Readable:Boolean = False): string; override;
    function IsConstant(out AValue: Double):Boolean; override;
    function Duplicate:TExprNode; override;
    function RealValue:Double; override;
  end;

  { TRealNode }

  TRealNode = class(TExprNode)
    Value: Double;
    constructor Create(AValue: Double);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
    function Compare(Node: TExprNode; ConstantsHigh: Boolean):Integer; override;
    function Text(Readable:Boolean = False): string; override;
    function IsConstant(out AValue: Double):Boolean; override;
    function Duplicate:TExprNode; override;
    function RealValue:Double; override;
  end;

  { TVariableNode }

  TVariableNode = class(TExprNode)
  private
    procedure SetValue(const AValue: Double);
  public
    function Text(Readable:Boolean = False): string; override;
    function EvalDependsOn: String; override;
    function EvalSortText(ConstantsHigh:Boolean = False): string; override;
    function RealValue:Double; override;
    procedure SetAllValues;
    property Value: Double read RealValue write SetValue;
  end;

  { TFunctionNode }

  TFunctionNode = class(TExprNode)
  protected
    function SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions):TExprNode; override;
  public
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
    function Text(Readable:Boolean = False): string; override;
    function Simplify(Actions:TSimplifyActions):TExprNode; override;
    function RealValue:Double; override;
  end;

  TOperatorNode = class(TFunctionNode)
    function Text(Readable:Boolean = False): string; override;
  end;

  TTokenNode = class
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; virtual; abstract;
    function CreateExprNode:TExprNode; overload; virtual; abstract;
    function CreateExprNode(Node: TExprNode):TExprNode; overload;
    function CreateExprNode(Node1, Node2: TExprNode):TExprNode; overload;
  end;

  TVariable = class (TTokenNode)
  private
    function GetValue: Double;
    procedure SetValue(AValue: Double);
  public
    Name: string;
    Values: array of Double; { Value can change during GetValue on Expr; Index into array is ThreadIndex-1 }
    constructor Create(s:String);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
    function CreateExprNode:TExprNode; override;
    procedure SetAllValues;
    property Value: Double read GetValue write SetValue;
  end;
  TVariableArray = array of TVariable;

  TOperatorProperty = (opAssociative, opCommutative, opDiscardSingleArg, opNoEvaluate);
  TOperatorProperties = set of TOperatorProperty;

  TFunction = class (TTokenNode)
  public
    Name: string;
    Arguments: Integer;
    Properties: TOperatorProperties;
    function Value(var a:array of Double):Double; virtual; abstract;
    constructor Create(s:String; n:Integer; AProperties: TOperatorProperties = []);
    function CreateExprNode:TExprNode; override;
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TConstant }

  TConstant = class (TFunction)
    FValue: Double; { If Value changes, expression must be re-parsed }
  public
    function Value(var a:array of Double):Double; override;
    constructor Create(v:Double; s:String);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  TFunction1 = function (a: Double):Double;
  TFunction2 = function (a, b: Double):Double;
  TFunctionv = function (var a: array of Double):Double;

  { TMonadicFunction }

  TMonadicFunction = class (TFunction)
  public
    Func: TFunction1;
    function Value(var a:array of Double):Double; override;
    constructor Create(f:TFunction1; s:String);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TDiadicFunction }

  TDiadicFunction = class (TFunction)
  public
    Func: TFunction2;
    function Value(var a:array of Double):Double; override;
    constructor Create(f:TFunction2; s:String; AProperties: TOperatorProperties = []);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TNadicFunction }

  TNadicFunction = class (TFunction)
  public
    Func: TFunctionv;
    function Value(var a:array of Double):Double; override;
    constructor Create(f:TFunctionv; NArgs: Integer; s:String; AProperties: TOperatorProperties = []);
  end;

  TOperator = class (TDiadicFunction)
  public
    Priority:Integer;
    constructor Create(f:TFunction2; s:String; APriority: Integer; AProperties: TOperatorProperties = []);
    function Value(var a:array of Double):Double; override;
  end;

  { TVaradicFunction }

  TVaradicFunction = class(TFunction) { Variable number of arguments }
  public
    Func: TFunctionv;
    function Value(var a:array of Double):Double; override;
    constructor Create(f:TFunctionv; MinArgs: Integer; s:String; AProperties: TOperatorProperties = []);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TArray }
  { An abstract function that can be assigned to; note that the number of arguments can be zero }
  TArray = class (TFunction)
  public
    function LowerBound(Bound: Integer): Integer; virtual; abstract;
    function UpperBound(Bound: Integer): Integer; virtual; abstract;
    procedure Store(var a:array of Double); virtual; abstract;
  end;

  { An concrete implementation using a built in array }
  TVectorVariable = class (TArray)
  private
    CheckBounds: Boolean;
  public
    Values: TDoubleArray;
    constructor Create(s:String; Size: Integer = 0; IsBoundChecked: Boolean = True);
    function Value(var a:array of Double):Double; override;
    function LowerBound(Bound: Integer): Integer; override;
    function UpperBound(Bound: Integer): Integer; override;
    procedure Store(var a:array of Double); override;
  end;

  { TImplicitLoopVariable }
  { loop variables, typically i, j, k ... }
  TImplicitLoopVariable = class(TFunction)
  private
    Offset: Integer;
  public
    constructor Create(s: string; AOffset:Integer);
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TIteratingFunction }
  { An abstract function for iterating over a range, such as Sum(0:3,i) = 6 }
  { Note that Value is only called in special circumstances, e.g. Sum(0,1,2,3) = 6 }
  TIteratingFunction = class (TFunction)
  public
    procedure StartIteration(Evaluator: TEvaluator); virtual;
    function NextIteration(Evaluator: TEvaluator; var a:array of Double): Double; virtual; abstract;
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  TStatusProc = procedure (var aLastErr, Info: ShortString; var Calls, Errs: LongInt; var Extra: Double);

  TLoadedFunction = class(TVaradicFunction)
  public
    StatusFunc: TStatusProc;
    hLibrary: TLibHandle;
    constructor Create(s:String);
    destructor Destroy; override;
  end;

  TFakeArrayStoreProcedure = procedure (a:PDouble; n: Integer) of object;
  TArrayStoreProcedure = procedure (var a:array of Double) of object;
  TFunctionValue = function (var a: array of Double): Double of object;
  TFakeFunctionValue = function (a: PDouble; n: Integer): Double of object;

  TToken = record
            case ty: toType of
              toNumber:   (Value: Double);
              toVariable, toThreadVariable: (Variable: TVariable);
              toFunction: (Func: TFunction);
              toOperator: (Op: TOperator);
              toSpecial:  (c: Char);
              toMonadicFunction: (Func1: TFunction1; MonFunc:TMonadicFunction);
              toDiadicFunction:  (Func2: TFunction2; DiFunc: TDiadicFunction);
              toPVariable, toStore: (PValue: PDouble);
              toFunction1:  (Funcv: TFunctionValue);
              toArrayLow, toArrayHigh: (Arr: TArray);
              toArrayStore: (StoreProc: TFakeArrayStoreProcedure);
              toStartLoop, toEndLoop, toLoopVar, toDup: (Offset: Integer);
              toIterationStart, toIterationNext: (IFunc: TIteratingFunction);
              { Next two lines are a fudge }
              toFunction2: (Fudge: TArrayStoreProcedure);
              toFunction3: (FuncvF: TFakeFunctionValue);
            end;
  PToken = ^TToken;

  TLexOption = ( xGlobalVariables, xGlobalFunctions,
                 xFundamentalConstants, xImplicitLoopVariables,
                 xBinary, xOctal, xCHex, xPascalHex, xLoadedFunctions,
                 xNonLocalAssign);
  { Note that xFundamentalConstants and xImplicitLoopVariables are mutually
    exclusive (k clashes, and l if MaxLoopDepth >= 4 }
  TLexOptions = set of TLexOption;

  { Essentially TSringlist, but with optional case insensitive find on case sensitive list }

  { TCaseStringList }

  TCaseStringList = class(TStringList)
  public
    function Find(const S: string; out Index: Integer; ACaseSensitive: Boolean): Boolean; overload;
    function IndexOf(const S: string; ACaseSensitive: Boolean): Integer; overload;
  end;

  { Essentially TSringlist, with objects optionally owned by the List }

  TOwnedCaseStringList = class(TCaseStringList)
  protected
    procedure SetCapacity(NewCapacity: Integer); override;
  public
    {$IFNDEF FPC}{Introduced 28 Sep 2010; may also be in Delphi 2009?}
    OwnsObjects: Boolean;
    {$ENDIF}
    constructor Create;
    destructor Destroy; override;
    procedure Delete(Index: Integer); override;
  end;

  { TVariableList }

  TVariableList = class(TOwnedCaseStringList)
  private
    function GetObjects(Index: Integer): TVariable;
    procedure SetObjects(Index: Integer; const Value: TVariable);
  public
    function VariableValues: string;
    constructor Create;
    procedure CheckThreadVars;
    procedure Add(Variable: TVariable);
    property Objects[Index: Integer]: TVariable read GetObjects write SetObjects;
  end;

  { TLex }

  TLex = class(TObject)
  private
    FCaseSensitive: Boolean;
    FString: string;
    p:Integer;
    FFunctions: TStringList;
    FVariables: TVariableList;
    FLocalFunctions: TOwnedCaseStringList; { Functions to be destroyed with Lex }
    FPutBack: Boolean;
    FPutBackToken: TToken;
    FOptions: TLexOptions;
    procedure SetCaseSensitive(AValue: Boolean);
  protected
    procedure SetText(s: string); virtual;
  public
    Token: string;
    ReturnUnknown: Boolean;
    IdentifierChars: set of Char;
    CommentChars: set of Char;
    constructor Create;
    destructor Destroy; override;
    property Options: TLexOptions read FOptions write FOptions
             default [ xGlobalVariables, xGlobalFunctions, xFundamentalConstants, xBinary, xOctal,
                       xCHex, xPascalHex ];
    property Functions: TStringList read FFunctions { write FFunctions} ;
    property Variables: TVariableList read FVariables { write FVariables} ;
    property Text:string read FString write SetText;
    function AddFunction(f: TFunction; Local: Boolean = False):TFunction; overload;
    function AddFunction(Name: string; f: TFunction; Local: Boolean = False):TFunction; overload;
    procedure AddVariable(v: TVariable);
    procedure RemoveVariable(v: TVariable);
    procedure RemoveVariables(AVariables: TVariableList);
    function MakeVariable(Name: string): TVariable;
    function NextToken(LocalReturnUnknown: Boolean = False):TToken; virtual;
    procedure PutBackToken(t:TToken);
    function FindVariableFromAddress(PValue: PDouble):TVariable;
    property CaseSensitive: Boolean read FCaseSensitive write SetCaseSensitive;
  end;

  TTokenArray = array of TToken;

  TEvaluatorState = record
    t: ^TToken; { Instruction pointer }
    Stack: array [0..MaxStackDepth] of Double;
    p: PDouble; { Stack pointer }
    LoopEnds: array [1..MaxLoopDepth] of Double;
    LoopDepth: Integer;
    Sums: array [1..MaxLoopDepth] of Double;
    nValues: array [1..MaxLoopDepth] of Integer;
    LoopVars: array [1..MaxLoopDepth] of Double;
  end;
  PEvaluatorState = ^TEvaluatorState;

  { TEvaluator }

  TEvaluator = class(TObject)
  private
    FEvaluatorOptions: TEvaluatorOptions;
    FTokList: TTokenArray;
    FTokSize: Integer;
    FSuspended: Boolean;
  protected
    SavedState: TEvaluatorState;
    ThreadStates: array of PEvaluatorState;
    procedure ClearTokList; virtual;
  public
    nSqrts: Integer; { Count of sqrt() calls; fudge to use in absence of complex evaluator }
    destructor Destroy; override;
    procedure CheckStack;
    function AddToken(Token: TToken):Integer;
    function Value(Resume: Boolean = False):Double;
    property TokList: TTokenArray read FTokList;
    property TokSize: Integer read FTokSize;
    property Suspended: Boolean read FSuspended;
    { Note that the options only apply during compilation }
    property EvaluatorOptions: TEvaluatorOptions read FEvaluatorOptions write FEvaluatorOptions;
  end;

  TExpr = class(TEvaluator)
  private
    FLex: TLex;
    FOwnsLex: Boolean;
    LocalVariables: TVariableList;
    LocalFunctions: TStringList;
    procedure SetText(s: string);
  protected
    procedure ClearTokList; override;
  public
    constructor Create(Lex:TLex; OwnsLex: Boolean = False);
    destructor Destroy; override;
    function MakeVariables(Text: string):TVariableList;
    function ExprTree: TExprNode;
    property Text:string write SetText;
    property Lex: TLex read FLex;
  end;

var
  GlobalFunctions, GlobalVariables, FundamentalConstants, LoadedFunctions: TCaseStringList;
  OperatorPlus, OperatorMinus,
  OperatorPower, OperatorTimes,
  OperatorDivide, OperatorModulus,
  OperatorShiftRight, OperatorShiftLeft,
  OperatorGreaterThan, OperatorLessThan,
  OperatorEquals, OperatorLessThanOrEqual,
  OperatorGreaterThanOrEqual, OperatorNotEqual,
  OperatorNotEqualAlt, OperatorAnd,
  OperatorOr:TOperator;
  VoigtFunction: TNadicFunction;
  MaxFunction, MinFunction, PolyFunction: TVaradicFunction;
  MonadicMinus: TMonadicFunction;
  ModFunction, Atan2Function, LorFunction, GauFunction: TDiadicFunction;
  AbsFunction,
  SinFunction, CosFunction,
  TanFunction, AcosFunction,
  AsinFunction, AtanFunction,
  AcosFunctionAlt, AsinFunctionAlt,
  AtanFunctionAlt,
  ExpFunction, LnFunction,
  Log10Function, SqrtFunction,
  SqrFunction, CeilFunction,
  FloorFunction, RoundFunction,
  IntFunction, FracFunction,
  AirFunction, VacFunction,
  nQuartzFunction, FactorialFunction
  : TMonadicFunction;
  ConstantPi: TConstant;
  SqrtMinusOne: TFunction;

var
  ImplicitLoopVariables: TCaseStringList;

{ Helpers }

procedure FreeStringList(var l:TCaseStringList);
function IsIntegerValue(Value: Double):Boolean;
procedure RegisterDiadicFunction(var o:TDiadicFunction; f: TFunction2; c:string);
procedure RegisterMonadicFunction(var o:TMonadicFunction; f: TFunction1; c:string);
function EB(Args: array of const; Actions: TSimplifyActions=[]; NilValue: Integer = 0; DuplicateObjects: Boolean = True):TExprNode;
function EN(Args: array of const; Actions: TSimplifyActions=[]; NilValue: Integer = 0):TExprNode;
function SimplifiesToInt(out Value: Integer; Args: array of const; NilValue: Integer = 0):Boolean;
function CheckSimplify(Expr:TExprNode; Actions:TSimplifyActions):TExprNode; {$IFNDEF CHECKSIMPLIFY} inline; {$ENDIF}

{$IFOPT R+}
var
   ExprNodesCreated, ExprNodesDestroyed: Integer;
{$ENDIF}
{$IFDEF CHECKSIMPLIFY}
var
  SimplifyCount: Integer;
{$ENDIF}

implementation

uses {$IFNDEF NO_MATH_UNIT} Math, {$ENDIF}
     CODATA, Worker, Wofz, AirVac;

{ TCaseStringList }

function TCaseStringList.Find(const S: string; out Index: Integer; ACaseSensitive: Boolean): Boolean;
begin
  if ACaseSensitive = CaseSensitive then
    Result := Find(S, Index)
  else begin
    Index := IndexOf(S, ACaseSensitive);
    Result := Index >= 0;
  end;
end;

function TCaseStringList.IndexOf(const S: string; ACaseSensitive: Boolean): Integer;
begin
  if ACaseSensitive = CaseSensitive then
    Result := IndexOf(S)
  else begin
    Result := 0;
    while Result < Count do begin
      if ACaseSensitive then begin
        if CompareStr(Strings[Result],S) = 0 then
          Exit
      end else
        if CompareText(Strings[Result],S) = 0 then
          Exit;
      Inc(Result);
    end;
    Result := -1; { Not found }
  end;
end;

procedure TExprNode.Changed;
begin
  Simplified := [siNone];
  FDependsOnValid := False;
  FSortTextLowValid := False;
  FSortTextHighValid := False;
  {$IFDEF CHECKSIMPLIFY}
  //WriteLn('Changed From: ', InitialText);
  //WriteLn('          To: ', Text(True));
  {$ENDIF}
end;

procedure TExprNode.CheckUnchanged(Expr: TExprNode);
{$IFDEF CHECKSIMPLIFY}
var
  V: Double;
{$ENDIF}
begin
{$IFDEF CHECKSIMPLIFY}
  V := Expr.RealValue;
  if not SameValue(InitialValue, V) then begin
    WriteLn(Log, 'Was:', InitialValue, ' Now:', V);
    WriteLn(Log, 'Was:', InitialText);
    WriteLn(Log, ' Now:', Expr.Text(True));
  end;
{$ENDIF}
  //WriteLn(' New:', Expr.Text(True));
end;

function CheckSimplify(Expr:TExprNode; Actions:TSimplifyActions):TExprNode; {$IFNDEF CHECKSIMPLIFY} inline; {$ENDIF}
{$IFDEF CHECKSIMPLIFY}
var
  V, V1: Double;
  OldExpr: TExprNode;
  i, Count: Integer;
  OS, NS: string;
{$ENDIF}
begin
  Result := Expr;
{$IFDEF CHECKSIMPLIFY}
  Inc(SimplifyCount);
  Count := SimplifyCount;
  for i := 0 to Expr.nChildren-1 do
    if not Assigned(Expr.Children[i]) then
      Writeln('Nil node:', Expr.Text);
  {$IFDEF SHOWALLCHANGES}
  OS := Expr.Text(True);
  {$ENDIF}
  OldExpr := Expr.Duplicate;
  try
    V := Expr.RealValue;
  except
    OS := OldExpr.Text(True);
    SetNan(V);
  end;
  if (Expr.Simplified <> Actions) or (siForce in Actions) then
    Result := Expr.Simplify(Actions);
  Result.Simplified := Actions - [siForce];
  V1 := Result.RealValue;
  if not SameValue(V,V1) then begin
    Result.RealValue;
    Writeln(Log, OldExpr.RealValue, ' ', V1);
    Writeln(Log, OldExpr.Text(True));
    Writeln(Log, Result.Text(True));
    WriteLn(Log);
  end;
  {$IFDEF SHOWALLCHANGES}
  try
    NS := Result.Text(True);
  except
    on Exception do
      NS := '???';
  end;
  if NS <> OS then begin
    Writeln(Count:4, ' ', OS);
    Writeln(SimplifyCount:4, ' ', NS);
    WriteLn;
  end;
  {$ENDIF}
  OldExpr.Free;
{$ELSE}
  if (Expr.Simplified <> Actions) or (siForce in Actions) then begin
    Result := Expr.Simplify(Actions);
    Result.Simplified := Actions - [siForce];
  end;
{$ENDIF}
end;

function EB(Args: array of const; Actions: TSimplifyActions; NilValue: Integer; DuplicateObjects: Boolean):TExprNode;
var
  i: Integer;
  procedure AddObject(Node: TExprNode);
  begin
    if Assigned(Result) then
      Result.AddNode(Node)
    else
      Result := Node;
  end;
  procedure AddFunction(const AName: string);
  var
    i: Integer;
    Node: TExprNode;
  begin
    i := GlobalFunctions.IndexOf(AName);
    if i < 0 then
      raise EExprError.Create('EB.AddFunction:"'+AName+'"');
    Node := Result;
    Result := ((GlobalFunctions.Objects[i]) as TFunction).CreateExprNode;
    if Assigned(Node) then
      Result.AddNode(Node);
  end;
begin
  Result := nil;
  for i := 0 to High(Args) do
    with Args[i] do
      case VType of
        vtInteger:    AddObject(TIntegerNode.Create(VInteger));
        vtObject:     if not Assigned(VObject) then
                        AddObject(TIntegerNode.Create(NilValue))
                      else if DuplicateObjects then
                        AddObject((VObject as TExprNode).Duplicate)
                      else
                        AddObject(VObject as TExprNode);
        vtChar:       AddFunction(VChar);
        vtWideChar:   AddFunction(VWideChar);
        vtString:     AddFunction(VString^);
        vtAnsiString: AddFunction(PAnsiChar(VAnsiString));
        {$IFDEF UNICODE}
        vtUnicodeString: AddFunction(PWideChar(VUnicodeString));
        {$ENDIF}
        vtInt64:      if Abs(VInt64^) <= High(Integer) then
                        AddObject(TIntegerNode.Create(VInt64^))
                      else
                        raise EExprError.CreateFmt('EB:value:%d', [VInt64^]);
      else
        raise EExprError.CreateFmt('EB:%d', [VType]);
      end;
  Result := CheckSimplify(Result,Actions);
end;

function EN(Args: array of const; Actions: TSimplifyActions; NilValue: Integer):TExprNode;
begin
  Result := EB(Args, Actions, NilValue, False);
end;

function SimplifiesToInt(out Value: Integer; Args: array of const; NilValue: Integer = 0):Boolean;
var
  Node: TExprNode;
begin
  Node := EB(Args, [], NilValue);
  //Writeln('SimplifiesToInt:',Node.Text(True));
  Result := Node is TIntegerNode;
  if Result then
    Value := TIntegerNode(Node).Value;
  Node.Free;
end;

constructor TExpr.Create(Lex:TLex; OwnsLex: Boolean);
begin
  inherited Create;
  FLex := Lex;
  LocalVariables := TVariableList.Create;
  FOwnsLex := OwnsLex;
end;

procedure TExpr.ClearTokList;
begin
  inherited;
  LocalVariables.Clear;
end;

destructor TExpr.Destroy;
begin
  if FOwnsLex then
    FreeAndNil(Flex);
  inherited Destroy;
  FreeAndNil(LocalVariables);
end;

function TExpr.ExprTree: TExprNode;
var
  pt: ^TToken;
  EStack: array of TExprNode;
  sp: Integer;
  procedure Push(Node: TExprNode);
  begin
    Inc(sp);
    if sp > High(EStack) then
      SetLength(EStack, Length(EStack)+20);
    EStack[sp] := Node;
  end;
  function Pop: TExprNode;
  begin
    Result := EStack[sp];
    Dec(sp);
  end;
  function HandleTFunction(nArgs: Integer; Func: TFunction):TExprNode;
  var
    i: Integer;
  begin
    Result := Func.CreateExprNode;
    for i := 0 to nArgs - 1 do
      Result.AddNode(EStack[sp - nArgs + 1 + i]);
    Dec(sp,nArgs);
    Push(Result);
  end;
  function MakeIntegerOrRealNode(Value: Double): TExprNode;
  begin
    if IsIntegerValue(Value) then
      Result := TIntegerNode.Create(Round(Value))
    else
      Result := TRealNode.Create(Value);
  end;
var
  n: Integer;
begin
  if TokSize = 0 then
    raise EExprError.Create('TExprTree: No Expression');
  pt := @FTokList[0];
  sp := -1;
  SetLength(EStack,20);
  while pt^.ty <> toSpecial do begin
    case pt^.ty of
      toPVariable: Push(FLex.FindVariableFromAddress(pt^.PValue).CreateExprNode);
      toThreadVariable: Push(pt^.Variable.CreateExprNode);
      toNumber:    Push(MakeIntegerOrRealNode(pt^.Value));
      toMonadicFunction: HandleTFunction(1,pt^.MonFunc);
      toDiadicFunction:  HandleTFunction(2,pt^.DiFunc);
    else
      n := Ord(pt^.ty) - Ord(toFunction);
      if n < 0 then
        raise EExprError.Create('ExprTree: NotYet');
      { toFunction, toFunction1, toFunction2 ... }
      HandleTFunction(n, TFunction(TMethod(pt^.Funcv).Data));
    end;
    Inc(pt);
  end;
  Result := Pop;
end;

procedure TExpr.SetText(s: string);
var
  LoopDepth: Integer;
  LoopStarts: array [1..MaxLoopDepth] of Integer;
  procedure PushLoop;
  var
    t: TToken;
  begin
    Inc(LoopDepth);
    if LoopDepth > MaxLoopDepth then
      raise EExprFormatError.Create('Maximum loop depth exceeded');
    t.ty := toStartLoop;
    LoopStarts[LoopDepth] := AddToken(t);
  end;
  function PopLoop(LoopEnd:Integer):Integer;
  var
    t: TToken;
  begin
    t.ty := toEndLoop;
    t.Offset := LoopEnd-LoopStarts[LoopDepth];
    Result := AddToken(t);
    { Fixup loop jumps }
    FTokList[LoopStarts[LoopDepth]].Offset := t.Offset+1;
    Dec(LoopDepth);
  end;
  procedure AddImplicitLoopVariables(First,Last:Integer);
  var
    t: TToken;
    i: Integer;
  begin
    for i := First to Last do begin
      if (i < 1) or (i > LoopDepth) then
        raise EExprError.Create('Implicit loop variable problem');
      t.ty := toLoopVar;
      t.Offset := i;
      AddToken(t);
    end;
  end;
  procedure Primary; forward;
  procedure ExpressionTail(Priority:Integer);
  var
    t: TToken;
  begin
    t := FLex.NextToken(False);
    while (t.ty = toOperator) and (t.Op.Priority > Priority) do begin
      Primary;
      ExpressionTail(t.Op.Priority);
      t.Op.Compile(Self);
//      t.ty := toDiadicFunction; { toFunction; }
//      t.Func2 := t.Op.Func; { t.Func := t.Op; } { t.Op.Func TDiadicFunction(t.Func).Func }
//      AddToken(t);
      t := FLex.NextToken(False);
    end;
    Flex.PutBackToken(t);
  end;
  procedure Expression(Priority:Integer);
  begin
    Primary;
    ExpressionTail(0);
  end;
  function NextIs(c:Char; Peek: Boolean = False): Boolean;
  var
    t: TToken;
  begin
    t := FLex.NextToken(False);
    Result := (t.ty = toSpecial) and (t.C = c);
    if not Result or Peek then
      Flex.PutBackToken(t);
  end;
  procedure CheckFor(c:Char);
  var
    t: TToken;
  begin
    t := FLex.NextToken(False);
    if ((t.ty <> toSpecial) or (t.C <> c)) and (Flex.Token <> c) then
      raise EExprFormatError.CreateFmt('TExpr: Expected %s Found %s', [c, Flex.Token]);
  end;
  procedure IteratingFunctionCall(tf: TToken);
  var
    i: Integer;
  begin
    { Layout:
        Evaluation of range
        toIterationStart
        toStartLoop to BodyEnd;
      BodyStart:
        Evaluation of all arguments, apart from range
        (Note that the number of arguments excludes the range)
        toIterationNext
      BodyEnd:
        toLoopEnd to BodyStart
      Done:
    }
    CheckFor('(');
    Expression(0);
    if NextIs(':') then begin
      Expression(0);
      tf.ty := toIterationStart;
      AddToken(tf);
      PushLoop;
      for i := 1 to tf.Func.Arguments do begin
        CheckFor(',');
        Expression(0);
      end;
      PopLoop(tf.Func.Compile(Self)); { Generate toIterationNext }
    end else if tf.Func.Arguments = 1 then begin
      { Alternate form - simple list of arguments, without iteration }
      i := 1;
      while NextIs(',') do begin
        Inc(i);
        Expression(0);
      end;
      if NextIs(':') then
        raise EExprFormatError.Create('Specify range first for '+tf.Func.Name);
      if i = 1 then
        raise EExprFormatError.Create('Must specify range for '+tf.Func.Name);
      tf.Func.Compile(Self,i) { Generate normal function call }
    end else
      raise EExprFormatError.Create('Expected : for '+tf.Func.Name);
    CheckFor(')');
  end;
  procedure Primary;
  var
    t: TToken;
    n, nArgs: Integer;
  begin
    t := FLex.NextToken(False);
    case t.ty of
      toOperator:
        if t.Op = OperatorPlus
          then Primary
        else if t.Op = OperatorMinus then begin
          Primary;
          MonadicMinus.Compile(Self);
        end else
          raise EExprFormatError.CreateFmt('TExpr: Unexpected Operator (%s)', [FLex.Token]);
      toNumber: AddToken(t);
      toVariable: t.Variable.Compile(Self);
      toFunction:
        if t.Func is TIteratingFunction then
          IteratingFunctionCall(t)
        else if t.Func.Arguments >= 0 then begin
          { Fixed number of arguments }
          if t.Func.Arguments <> 0 then
            if not NextIs('(') then begin
              if not (t.Func is TArray) then
                CheckFor('(')
              else
                AddImplicitLoopVariables(LoopDepth-t.Func.Arguments+1,LoopDepth)
            end else begin
              for n := 1 to t.Func.Arguments do begin
                Expression(0);
                if n = t.Func.Arguments then
                  CheckFor(')')
                else
                  CheckFor(',');
              end;
            end
          else if t.Func is TImplicitLoopVariable then
            if TImplicitLoopVariable(t.Func).Offset > LoopDepth then
              raise EExprError.Create('Inactive loop variable used:'+t.Func.Name);
          t.Func.Compile(Self);
        end else begin
          { Variable number of arguments }
          nArgs := 0;
          CheckFor('(');
          repeat
            Expression(0);
            Inc(nArgs);
          until not NextIs(',');
          CheckFor(')');
          if nArgs < - t.Func.Arguments then
            raise EExprFormatError.CreateFmt('TExpr: Not enough arguments to %s', [t.Func.Name]);
          t.Func.Compile(Self,nArgs);
        end;
      toSpecial:
        if t.C = '(' then begin
          Expression(0);
          CheckFor(')');
        end else if t.c = 'e' then
          raise EExprFormatError.Create('TExpr: Unexpected end of expression')
        else
          raise EExprFormatError.CreateFmt('TExpr: Unexpected Token (%s)', [t.C]);
    end;
  end;
  procedure Statement(CheckEnd: Boolean);
  var
    t: TToken;
    LetSeen: Boolean;
    IsAssignment: Boolean;
    Loops: Integer;
    procedure AssignmentBody(ExtraDrops:Integer);
    var
      LoopEnd, i: Integer;
    begin
      CheckFor('=');
      Statement(False);
      { Note: Local Variable can only be used in following statements }
      if t.ty = toUnknown then begin
        LocalVariables.Add(t.Variable);
        t.ty := toVariable;
      end;
      if t.ty = toVariable then begin
        if nWorkers > 0 then
          t.ty := toThreadStore
        else begin
          t.ty := toStore;
          t.PValue := @t.Variable.Values[0];
        end;
      end;
      LoopEnd := AddToken(t);
      for i := 1 to Loops do
        LoopEnd := PopLoop(LoopEnd);
      for i := 1 to ExtraDrops do begin
        t.ty:= toDrop;
        t.Offset := 1;
        AddToken(t);
      end;
      if CheckEnd and NextIs(';') and not NextIs('e') then begin
        t.ty:= toDrop;
        t.Offset := 0;
        AddToken(t);
      end;
    end;
    procedure LetError;
    begin
      raise EExprFormatError.Create('Expected variable name after let, found ' + FLex.Token);
    end;
    procedure ArrayAssignment;
    var
      CloseToken, ArgTerm: Char;
      i, nSingles, DupOffset, LDepth, ExtraDrops: Integer;
      SingleValue: array of Boolean;
      t1: TToken;
      procedure AddLoopBound(Bound:Integer);
      begin
        t.ty := toType(Ord(toArrayLow)+Bound-1); AddToken(t);
        t.ty := toType(Ord(toArrayHigh)+Bound-1); AddToken(t);
        PushLoop;
        Inc(Loops);
      end;
    begin
      ExtraDrops := 0;
      if NextIs('[') or NextIs('(') then begin
        if Flex.Token = '[' then
          CloseToken := ']'
        else
          CloseToken := ')';
        IsAssignment := LetSeen;
        SetLength(SingleValue,t.Func.Arguments);
        nSingles := 0;
        LDepth := LoopDepth + 1;
        for i := 1 to t.Func.Arguments do begin
          SingleValue[i-1] := True;
          if i <> t.Func.Arguments then
            ArgTerm := ','
          else
            ArgTerm := CloseToken;
          if NextIs(ArgTerm,True) then begin
            if i > LoopDepth then
              AddLoopBound(i)
            else
              Dec(LDepth);
            SingleValue[i-1] := False;
          end else begin
            Expression(0);
            if NextIs(':') then begin
              SingleValue[i-1] := False;
              { assignment to range }
              IsAssignment := True;
              Expression(0);
              if i > LoopDepth then begin
                PushLoop; { Only add loop variables where needed }
                Inc(Loops);
              end else
                Dec(LDepth);
            end;
          end;
          CheckFor(ArgTerm);
          if SingleValue[i-1] then
            Inc(nSingles)
          else
            Inc(ExtraDrops,nSingles);
        end;
        { For each array index, there are three possibilities:
          1: Loop variable - generate push of loop variable;
          2: expression inside loops - value needs rotating to top of stack
          3: expression outside some loops - must generate Dup operation
          Initial state is all single values on the stack }
        DupOffset := nSingles-1;
        nSingles := 0;
        for i := 1 to t.Func.Arguments do
          if SingleValue[i-1] then begin
            { Single value - duplicate to top of stack }
            t1.ty := toDup;
            t1.Offset := DupOffset;
            AddToken(t1);
            if nSingles >= ExtraDrops then begin
              { 2: expression inside loops - discard deeper value }
              t1.ty := toDrop;
              t1.Offset := DupOffset+1;
              AddToken(t1);
            end;
            Inc(nSingles);
          end else begin
            { 1: Loop variable - generate push of loop variable; }
            Inc(DupOffset); { increase stack jump needed }
            t1.ty := toLoopVar;
            t1.Offset := LDepth;
            AddToken(t1);
            Inc(LDepth);
          end;
        if not IsAssignment then
          IsAssignment := NextIs(':')
        else
          CheckFor(':');
      end else begin
        if (t.Func.Arguments > 0) and NextIs(':') then begin
          { only start loops if essential }
          for i := LoopDepth+1 to t.Func.Arguments do
            AddLoopBound(i);
          AddImplicitLoopVariables(LoopDepth-t.Func.Arguments+1,LoopDepth);
          IsAssignment := True;
        end else begin
          IsAssignment := NextIs(':') or LetSeen;
          if not IsAssignment then
            AddImplicitLoopVariables(LoopDepth-t.Func.Arguments+1,LoopDepth);
        end;
      end;
      if IsAssignment then begin
        t.ty := toArrayStore;
        t.Fudge := {$IFDEF FPC}@{$ENDIF}t.Arr.Store;
        AssignmentBody(ExtraDrops);
      end else if LetSeen then
        LetError
      else begin
        t.Arr.Compile(Self);
        ExpressionTail(0);
      end;
    end;
  var
    IsLocal: Boolean;
  begin
    LetSeen := NextIs('l');
    IsAssignment := LetSeen;
    Loops := 0;
    { Read possible variable or array element }
    t := FLex.NextToken(True); { Read Variable to set }
    case t.ty of
      { May be an assignemnt }
      toUnknown:  if LetSeen then begin
                    t.Variable := FLex.MakeVariable(FLex.Token);
                    NextIs(':'); { Optional := for Let }
                  end else
                    raise EExprFormatError.CreateFmt('Unknown variable or function: %s', [Flex.Token]);
      toVariable: begin
                    IsLocal := LocalVariables.IndexOf(t.Variable.Name) >= 0;
                    if not (xNonLocalAssign in FLex.Options) and not IsLocal and (LetSeen or NextIs(':')) then
                      raise EExprFormatError.Create('Can''t write to non local variable ' + t.Variable.Name);
                    IsAssignment := NextIs(':');
                    if IsAssignment then
                      AssignmentBody(0)
                    else begin
                      t.Variable.Compile(Self);
                      ExpressionTail(0);
                    end;
                  end;
      toFunction: if t.Func is TArray then
                    ArrayAssignment
                  else if LetSeen then
                    LetError
                  else begin
                    FLex.PutBackToken(t);
                    Expression(0);
                  end;
    else { definately not an assignment }
      if LetSeen then
        LetError
      else begin
        FLex.PutBackToken(t);
        Expression(0);
      end;
    end;
  end;
var
  tok: TToken;
begin
  FLex.Text := s;
  ClearTokList;
  LoopDepth := 0;
  if Length(s) > 0 then
    try
      try
        repeat
          Statement(True);
        until NextIs('e');
        tok.ty := toSpecial;
        tok.C := 'e';
        AddToken(tok);
        CheckStack;
      finally
        FLex.RemoveVariables(LocalVariables);
      end;
    except
      ClearTokList;
      raise;
    end;
  if LoopDepth <> 0 then
    raise EExprError.Create('Loop mismatch');
end;

function TExpr.MakeVariables(Text: string): TVariableList;
{ Create all variables not otherwise defined in text }
var
  tok: TToken;
begin
  FLex.Text := Text;
  ClearTokList;
  Result := TVariableList.Create;
  Result.CaseSensitive := FLex.CaseSensitive;
  try
    try
      repeat
        tok := FLex.NextToken(True);
        if (tok.ty = toSpecial) and (tok.C = 'l') then begin
          { Let - skip variable declaration }
          tok := FLex.NextToken(True);
          if tok.ty = toUnknown then
            LocalVariables.Add(FLex.MakeVariable(FLex.Token))
        end else if tok.ty = toUnknown then
          Result.Add(FLex.MakeVariable(FLex.Token))
      until (tok.ty = toSpecial) and (tok.C = 'e')
    finally
      FLex.RemoveVariables(LocalVariables);
    end;
  except
    FLex.RemoveVariables(Result);
    FreeAndNil(Result);
    ClearTokList;
    raise;
  end;
end;

constructor TLex.Create;
begin
  inherited Create;
  IdentifierChars := ['a'..'z','A'..'Z', '_', '0'..'9', '''', '"', '.'];
  CommentChars := ['#', '/', '{']; { Note / must be followed by / or * to be a comment; curly brackets must match }
  ReturnUnknown := False;
  FFunctions := TCaseStringList.Create;
  FFunctions.Duplicates := dupError;
  FFunctions.Sorted := True;
  FVariables := TVariableList.Create;
  FVariables.OwnsObjects := False;
  FVariables.CaseSensitive := FCaseSensitive;
  FOptions := [xGlobalVariables, xGlobalFunctions, xFundamentalConstants,
               xBinary, xOctal, xCHex, xPascalHex];
end;

destructor TLex.Destroy;
var
  i: Integer;
begin
  FLocalFunctions.Free;
  FFunctions.Free;
  FVariables.Free;
  inherited Destroy;
end;

function TLex.FindVariableFromAddress(PValue: PDouble): TVariable;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to FVariables.Count - 1 do
    if @TVariable(FVariables.Objects[i]).Values[0] = PValue then begin
      Result := TVariable(FVariables.Objects[i]);
      Break;
    end;
end;

procedure TLex.SetCaseSensitive(AValue: Boolean);
begin
  if FCaseSensitive = AValue then Exit;
  FCaseSensitive := AValue;
  FFunctions.CaseSensitive := FCaseSensitive;
  FVariables.CaseSensitive := FCaseSensitive;
end;

procedure TLex.SetText(s: string);
begin
  FString := s;
  p := 1;
end;

procedure TLex.PutBackToken(t:TToken);
begin
  if FPutBack then
    raise EExprError.CreateFmt('TLex: Token put back twice (%s)', [Token]);
  FPutBackToken := t;
  FPutBack := True;
end;

function TLex.NextToken(LocalReturnUnknown: Boolean):TToken;
var
  l, start, base, i, ep: Integer;
  o: TObject;
  digits: set of char;
  OperatorChars: Boolean;
  function FindFunction(s:string):Boolean;
  var
    i: Integer;
    FuncName: string;
  begin
    Result := True;
    if FFunctions.Find(s, i) then
      o := FFunctions.Objects[i]
    else if (xImplicitLoopVariables in FOptions) and ImplicitLoopVariables.Find(s, i, CaseSensitive) then
      o := ImplicitLoopVariables.Objects[i]
    else if (xGlobalFunctions in FOptions) and GlobalFunctions.Find(s, i, CaseSensitive) then
      o := GlobalFunctions.Objects[i]
    else if (xFundamentalConstants in FOptions) and FundamentalConstants.Find(s, i, CaseSensitive) then
      o := FundamentalConstants.Objects[i]
    else if (xLoadedFunctions in FOptions) and (s[1] in ['"', '''']) then begin
      FuncName := ExtractFileName(Copy(s,2,Length(s)-2));
      if Length(ExtractFileExt(FuncName)) > 0 then
        FuncName := Copy(FuncName, 1, Length(FuncName)-Length(ExtractFileExt(FuncName)));
      if not LoadedFunctions.Find(FuncName, i, CaseSensitive) then begin
        o := TLoadedFunction.Create(FuncName);
        LoadedFunctions.AddObject(FuncName, o);
      end else
        o := LoadedFunctions.Objects[i];
    end else if s = '**' then
      o := OperatorPower
    else
      Result := False;
  end;
var
  IsComment, SlashStar: Boolean;
  CommentEnd: set of Char;
begin
  if FPutBack then begin
    Result := FPutBackToken;
    FPutBack := False;
    Exit;
  end;
  l := Length(FString);
  repeat
    while (p <= l) and (FString[p] in [#0..' ']) do Inc(p);
    if p > l then begin
      Result.ty := toSpecial;
      Result.C := 'e';
      Token := '';
      Exit;
    end;
    IsComment := FString[p] in CommentChars;
    CommentEnd := [#10,#13];
    if IsComment and (FString[p] = '/') then begin
      { / is a special case - must be // or /* }
      IsComment := (p < l) and (FString[p+1] in ['/', '*']);
      SlashStar := IsComment and (FString[p+1] = '*');
    end;
    if IsComment then begin
      if FString[p] = '{' then
        CommentEnd := CommentEnd + ['}'];
      while (p <= l) and not (FString[p] in CommentEnd) do
        if SlashStar and (p < l) and (FString[p] = '*') and (FString[p+1] = '/') then begin
          Inc(p,2);
          Break;
        end else
          Inc(p);
    end;
  until not IsComment;
  start := p;
  p := p + 1;
  if FString[start] in ['(',')',',',';',':'] then begin
    Token := FString[start];
    Result.ty := toSpecial;
    Result.C := FString[start];
  end else if FString[start] in ['0'..'9','.','$'] then begin
    { Number }
    digits := ['0'..'9', '.'];
    Base := 10;
    if (xPascalHex in FOptions) and (FString[start] = '$') then begin
      Base := 16;
      digits := ['0'..'9', 'a'..'f'];
    end else if (p <= l) and (FString[start] = '0') and (FString[p] in ['x', 'o', 'b']) then begin
      case FString[p] of
       'b': if xBinary in FOptions then begin
              Base := 2; digits := ['0'..'1'];
            end;
       'o': if xOctal in FOptions then begin
              Base := 8; digits := ['0'..'7'];
            end;
       'x': if xCHex in FOptions then begin
              Base := 16; digits := ['0'..'9', 'a'..'f'];
            end;
      end;
      if Base <> 10 then Inc(p);
    end;
    Result.Value := 0;
    while (p <= l) and (FString[p] in digits) do begin
      if FString[p] = '.' then
        digits := digits - ['.']
      else
        if FString[p] >= 'a' then
          Result.Value := Result.Value*Base + Ord(FString[p]) - Ord('a') + 10
        else
          Result.Value := Result.Value*Base + Ord(FString[p]) - Ord('0');
      Inc(p);
    end;
    if (p <= l) and (Base = 10) and (FString[p] in ['e', 'E', 'd', 'D']) then begin
      ep := p;
      Inc(p);
      if (p <= l) and (FString[p] in ['+','-']) then Inc(p);
      while (p <= l) and (FString[p] in  ['0'..'9']) do Inc(p);
    end else
      ep := -1;
    Token := Copy(FString, Start, p-Start);
    if ep > 0 then
      Token[ep-Start+1] := 'e'; { Allow multiple forms of exponent }
    if Base = 10 then
      try
        Result.Value := StrToFloat(Token);
        Result.ty := toNumber;
      except
        Result.ty := toUnknown;
        raise EExprFormatError.CreateFmt('Bad number: %s', [Token]);
      end
    else
      Result.ty := toNumber;
  end else begin
    OperatorChars := False;
    if FString[start] in ['a'..'z','A'..'Z', '_'] then
      while (p <= l) and (FString[p] in IdentifierChars) do Inc(p)
    else if FString[start] in ['"', ''''] then begin
      while (p <= l) and (FString[p] <> FString[start]) do Inc(p);
      if p > l then
        raise EExprFormatError.Create('Unterminated string');
      if p-Start = 1 then
        raise EExprFormatError.Create('Empty string');
      Inc(p);
    end else begin
      { >>, <<, <>, !=, <=, >= ** are special case }
      if (p <= l) and (FString[start] in ['<','>', '!', '*']) and FindFunction(Copy(FString, Start, 2)) then
        Inc(p);
      OperatorChars := True;
    end;
    if not (xLoadedFunctions in FOptions) and (FString[start] in ['"', '''']) then
      Token := Copy(FString, Start+1, p-Start-2)
    else
      Token := Copy(FString, Start, p-Start);
    if Token = 'let' then begin { Variable declaration }
      Result.ty := toSpecial;
      Result.c := 'l';
    end else if FVariables.Find(Token, i) then begin
      Result.ty := toVariable;
      Result.Variable := TVariable(FVariables.Objects[i]);
    end else if FindFunction(Token) then begin
      if o is TOperator then begin
        Result.ty := toOperator;
        Result.Op := TOperator(o);
      end else begin
        Result.ty := toFunction;
        Result.Func := TFunction(o);
      end;
    end else if (xGlobalVariables in FOptions) and GlobalVariables.Find(Token, i, CaseSensitive) then begin
      Result.ty := toVariable;
      Result.Variable := TVariable(GlobalVariables.Objects[i]);
    end else if OperatorChars then
      raise EExprFormatError.Create('Unexpected character:'+Token)
    else begin
      Result.ty := toUnknown;
      if not (ReturnUnknown or LocalReturnUnknown) or (Token[1] in ['"', '''']) then
        raise EExprFormatError.CreateFmt('Unknown variable or function: %s', [Token]);
    end;
  end;
end;

function TLex.AddFunction(Name: string; f: TFunction; Local: Boolean = False):TFunction;
var
  i: Integer;
begin
  if FFunctions.Find(Name, i) then begin
    FFunctions.Delete(i);
    if Assigned(FLocalFunctions) and FLocalFunctions.Find(Name, i) then
      FLocalFunctions.Delete(i);
  end;
  FFunctions.AddObject(Name, f);
  if Local then begin
    if not Assigned(FLocalFunctions) then
      FLocalFunctions := TOwnedCaseStringList.Create;
    FLocalFunctions.AddObject(Name, f);
  end;
  Result := f;
end;

function TLex.AddFunction(f: TFunction; Local: Boolean):TFunction;
begin
  AddFunction(f.Name,f,Local);
  Result := f;
end;

procedure TLex.RemoveVariable(v: TVariable);
var
  i: Integer;
begin
  if FVariables.Find(v.Name, i) then
    FVariables.Delete(i);
end;

procedure TLex.RemoveVariables(AVariables: TVariableList);
var
  i: Integer;
begin
  for i := AVariables.Count-1 downto 0 do
    RemoveVariable(AVariables.Objects[i]);
end;

procedure TLex.AddVariable(v: TVariable);
begin
  RemoveVariable(v);
  FVariables.Add(v);
end;

function TLex.MakeVariable(Name: string): TVariable;
var
  i: Integer;
begin
  if FVariables.Find(Name, i) then
    Result := TVariable(Variables.Objects[i])
  else begin
    Result := TVariable.Create(Name);
    AddVariable(Result);
  end;
end;

procedure TVariable.SetAllValues;
var
  i: Integer;
begin
  if nWorkers >= Length(Values) then
    if ThreadIndex <> 1 then
      raise Exception.Create('Value array extended in worker')
    else
      SetLength(Values, nWorkers+1);
  for i := 0 to nWorkers do
    Values[i] := Value;
end;

procedure TVariable.SetValue(AValue: Double);
begin
  Values[ThreadIndex-1] := AValue;
end;

function TVariable.GetValue: Double;
begin
  Result := Values[ThreadIndex-1]
end;

constructor TVariable.Create(s:String);
begin
  inherited Create;
  Name := s;
  SetLength(Values, nWorkers+1);
end;

function TVariable.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
  OldLength, i: Integer;
begin
  { This assumes nWorkers will not change for the life of the compiled Expr }
  if Length(Values) <> nWorkers+1 then begin
    OldLength := Length(Values);
    SetLength(Values, nWorkers+1);
    for i := OldLength to High(Values) do
      Values[i] := Values[0];
  end;
  if nWorkers > 0 then begin
    Token.ty := toThreadVariable;
    Token.Variable := Self;
  end else begin
    Token.ty := toPVariable;
    Token.PValue := @Values[0];
  end;
  Result := Evaluator.AddToken(Token);
end;

function TFunction.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  if nArgs < 0 then
    nArgs := Arguments;
  Token.ty := toType(Ord(toFunction)+nArgs);
  Token.Funcv := {$IFDEF FPC}@{$ENDIF}Value;
  Result := Evaluator.AddToken(Token);
end;

constructor TFunction.Create(s:String; n:Integer; AProperties: TOperatorProperties);
begin
  inherited Create;
  Arguments := n;
  Name := s;
  Properties := AProperties;
end;

constructor TConstant.Create(v:Double; s:String);
begin
  inherited Create(s, 0);
  FValue := v;
end;

function TConstant.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toNumber;
  Token.Value := FValue;
  Result := Evaluator.AddToken(Token);
end;

function TConstant.Value(var a:array of Double):Double;
begin
  Result := FValue;
end;

{ TMonadicFunction }

function TMonadicFunction.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toMonadicFunction;
  Token.Func1 := Func;
  Token.MonFunc := Self;
  Result := Evaluator.AddToken(Token);
end;

constructor TMonadicFunction.Create(f:TFunction1; s:String);
begin
  inherited Create(s, 1);
  Func := f;
end;

function TMonadicFunction.Value(var a:array of Double):Double;
begin
  Result := Func(a[0]);
end;

{ TDiadicFunction }

function TDiadicFunction.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toDiadicFunction;
  Token.Func2 := Func;
  Token.DiFunc := Self;
  Result := Evaluator.AddToken(Token);
end;

constructor TDiadicFunction.Create(f:TFunction2; s:String; AProperties: TOperatorProperties = []);
begin
  inherited Create(s, 2, AProperties);
  Func := f;
end;

function TDiadicFunction.Value(var a:array of Double):Double;
begin
  Result := Func(a[1], a[0]);
end;

{ TOperator }

constructor TOperator.Create(f:TFunction2; s:String; APriority: Integer; AProperties: TOperatorProperties = []);
begin
  inherited Create(f, s, AProperties);
  Priority := APriority;
end;

function TOperator.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := a[High(a)];
  for i := High(a)-1 downto 0 do
    Result := Func(Result,a[i]);
end;

procedure Op(var o:TOperator; f: TFunction2; c:string; Priority:Integer; AProperties: TOperatorProperties = []);
begin
  o := TOperator.Create(f, c, Priority, AProperties);
  GlobalFunctions.AddObject(c, o);
end;

procedure RegisterDiadicFunction(var o:TDiadicFunction; f: TFunction2; c:string);
begin
  o := TDiadicFunction.Create(f, c);
  GlobalFunctions.AddObject(c, o);
end;

procedure RegisterMonadicFunction(var o:TMonadicFunction; f: TFunction1; c:string);
begin
  o := TMonadicFunction.Create(f, c);
  GlobalFunctions.AddObject(c, o);
end;

procedure F0(c:string; v:Double);
begin
  FundamentalConstants.AddObject(c, TConstant.Create(v, c));
end;

procedure FreeStringList(var l:TCaseStringList);
var
  i: Integer;
begin
  if l <> nil then begin
    for i := l.Count-1 downto 0 do begin
      l.Objects[i].Free;
      l.Delete(i);
    end;
    l.Free;
  end;
  l := nil;
end;

{$IFDEF VER80}
{$F+}
{$ENDIF}

procedure DeleteFunctions;
begin
  FreeStringList(GlobalFunctions);
  FreeStringList(GlobalVariables);
  FreeStringList(FundamentalConstants);
  FreeStringList(LoadedFunctions);
  FreeStringList(ImplicitLoopVariables);
  MonadicMinus.Free;
end;

function _Plus(a, b:Double):Double;
begin
  Result := a + b;
end;

function _Minus(a, b:Double):Double;
begin
  Result := a - b;
end;

function _Power(a, b: Double):Double;
begin
  Result := Power(a,b);
end;

function _Times(a, b: Double):Double;
begin
  Result := a * b;
end;

function _Modulus(a, b: Double):Double;
begin
  Result := Frac(a/b)*b;
end;

function _ShiftRight(a, b: Double):Double;
begin
  Result := Round(a) shr Round(b);
end;

function _ShiftLeft(a, b: Double):Double;
begin
  Result := Round(a) shl Round(b);
end;

function _GreaterThan(a, b: Double):Double;
begin
  if a > b then
   Result := 1
  else
   Result := 0;
end;

function _LessThan(a, b: Double):Double;
begin
  if a < b then
    Result := 1
  else
    Result := 0;
end;

function _Equals(a, b: Double):Double;
begin
  if SameValue(a ,b) then
    Result := 1
  else
    Result := 0;
end;

function _LessThanOrEqual(a, b: Double):Double;
begin
  if a <= b then
    Result := 1
  else
    Result := 0;
end;

function _GreaterThanOrEqual(a, b: Double):Double;
begin
  if a >= b then
    Result := 1
  else
    Result := 0;
end;

function _NotEqual(a, b: Double):Double;
begin
  if not SameValue(a ,b) then
    Result := 1
  else
    Result := 0;
end;

function _And(a, b: Double):Double;
begin
  Result := Round(a) and Round(b);
end;

function _Or(a, b: Double):Double;
begin
  Result := Round(a) or Round(b);
end;

function _MonadicMinus(a:Double):Double;
begin
  Result := -a;
end;

function _Abs(a:Double):Double;
begin
  Result := Abs(a);
end;

function _Sin(a:Double):Double;
begin
  Result := Sin(a);
end;

function _Cos(a:Double):Double;
begin
  Result := Cos(a);
end;

function _Tan(a:Double):Double;
begin
  Result := Sin(a)/Cos(a);
end;

function _Acos(a:Double):Double;
begin
  if a = 0 then
    Result := Pi/2
  else
    Result := ArcTan (sqrt (1-sqr (a)) /a);
end;

function _Asin(a:Double):Double;
var
  b: Double;
begin
  b := Sqrt(1-Sqr(a));
  if b = 0 then
    Result := Pi/2
  else
    Result := ArcTan (a/b);
end;

function _Atan(a:Double):Double;
begin
  Result := ArcTan(a);
end;

function _Atan2(a, b:Double):Double;
begin
  if b <> 0 then
    Result := ArcTan2(a, b)
  else if a > 0 then
    Result := Pi/2
  else
    Result := -Pi/2;
end;

function _Mod(a, b:Double):Double;
begin
  Result := Frac(Abs(a)/b)*b;
  if a < 0 then Result := b - Result;
end;

function _Lor(a, b:Double):Double; { Normalized Lorentzian }
begin
  Result := b / (2*Pi) / (Sqr(a)+sqr(b/2));
end;

function _Gau(a, b:Double):Double;  { Normalized Gaussian }
begin
  Result := 2*SQRT(Ln(2)) / b;
  a := Sqr(Result*a);
  Result := Result/SQRT(Pi)*EXP(-a);
end;

function _Exp(a:Double):Double;
begin
  if a < -11355.137111933024 then
    Result := 0
  else
    Result := Exp(a);
end;

function _Ln(a:Double):Double;
begin
  Result := Ln(a);
end;

function _Log10(a:Double):Double;
begin
  Result := Ln(a)/ln(10);
end;

function _Sqrt(a:Double):Double;
begin
  Result := Sqrt(a);
end;

function _Sqr(a:Double):Double;
begin
  Result := Sqr(a);
end;

function _Ceil(a:Double):Double;
begin
  Result := Ceil(a);
end;

function _Floor(a:Double):Double;
begin
  Result := Floor(a);
end;

function _Round(a:Double):Double;
begin
  Result := Round(a);
end;

function _Int(a:Double):Double;
begin
  Result := Int(a);
end;

function _Frac(a:Double):Double;
begin
  Result := Frac(a);
end;

function _Factorial(a:Double):Double;
begin
  if a <= 0 then begin
    Result := 1;
    Exit;
  end;
  Result := a;
  while a > 2 do begin
    a := a - 1;
    Result := Result * a;
  end;
end;

function _Poly(var a: array of Double): Double;
var
  i, nTerms: Integer;
  x: Double;
begin
  nTerms := High(a);
  x := a[nTerms];
  Result := a[0];
  for i := 1 to nTerms-1 do
    Result := Result*x + a[i];
end;

function _Max(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := a[0];
  for i := 1 to High(a) do
    if a[i] > Result then
      Result := a[i];
end;

function _Min(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := a[0];
  for i := 1 to High(a) do
    if a[i] < Result then
      Result := a[i];
end;

function _Voigt(var a: array of Double): Double;
begin
  Result := Voigt(a[2], a[1], a[0]);
end;

{ TNadicFunction }

constructor TNadicFunction.Create(f: TFunctionv; NArgs: Integer; s: String; AProperties: TOperatorProperties);
begin
  inherited Create(s, NArgs, AProperties);
  Func := f;
end;

{$IFDEF FPC}
type
  TPointerFunctionv = function (a: Pdouble; n: Integer):Double;
{$ENDIF}

function TNadicFunction.Value(var a: array of Double): Double;
{$IFDEF FPC}
var
  fudge: TPointerFunctionv;
{$ENDIF}
begin
  {$IFDEF FPC}
  fudge := TPointerFunctionv({$IFNDEF FPC}@{$ENDIF}Func); { Fudged conversion of PDouble to array of Double }
  Result := fudge(@a[0],Arguments); { This defeats the bound checking }
  {$ELSE}
  Result := Func(Slice(a,Arguments)); { This defeats the bound checking }
  {$ENDIF}
end;

{ TVaradicFunction }

function TVaradicFunction.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
begin
  if nArgs < 0 then
    raise EExprError.Create('TVaradicFunction.Compile');
  Result := inherited Compile(Evaluator, nArgs);
end;

constructor TVaradicFunction.Create(f: TFunctionv; MinArgs: Integer; s: String; AProperties: TOperatorProperties);
begin
  inherited Create(s, -MinArgs, AProperties);
  Func := f;
end;

function TVaradicFunction.Value(var a: array of Double): Double;
begin
  Result := Func(a);
end;

{ TLoadedFunction }

constructor TLoadedFunction.Create(s: String);
var
  Aname: string;
begin
  if s[1] = '"' then
    Aname := Copy(s,2,Length(s)-2)
  else
    Aname := s;
  if Length(ExtractFileExt(Aname)) = 0 then
  {$IFDEF MSWINDOWS}
    Aname := Concat(Aname, '.dll');
  {$ENDIF}
  {$IFDEF LINUX}
    Aname := Aname + '.so';
  {$ENDIF}
  {$IFDEF LINUX}
  if Pos('lib', ExtractFileName(Aname)) <> 1 then
    { Add lib to name }
    Aname := ExtractFilePath(Aname) + 'lib' + ExtractFileName(Aname);
  {$ENDIF}
  hLibrary := LoadLibrary(PChar(Aname));
  if hLibrary <> TLibHandle(0) then begin
    {$IFNDEF FPC}@{$ENDIF}Func := {$IFNDEF FPC}@{$ENDIF}TFunctionv(GetProcAddress(hLibrary, 'CustomFunction'));
    if @Func <> nil then begin
      {$IFNDEF FPC}@{$ENDIF}StatusFunc := {$IFNDEF FPC}@{$ENDIF}TStatusProc(GetProcAddress(hLibrary, 'CustomFunctionStatus'));
      inherited Create(Func, 1, Aname);
      Exit;
    end else
      FreeLibrary(hLibrary);
  end;
  hLibrary := TLibHandle(0);
  raise EExprFormatError.CreateFmt('TLoadedFunction: Can''t load %s', [Copy(s,2,Length(s)-2)]);
end;

destructor TLoadedFunction.Destroy;
begin
  if hLibrary <> TLibHandle(0) then FreeLibrary(hLibrary);
  hLibrary := TLibHandle(0);
  inherited Destroy;
end;

{ TOwnedCaseStringList }

constructor TOwnedCaseStringList.Create;
begin
  inherited;
  OwnsObjects := True;
end;

procedure TOwnedCaseStringList.Delete(Index: Integer);
begin
  {$IFNDEF FPC}
  if OwnsObjects then
    Objects[Index].Free;
  {$ENDIF}
  inherited;
end;

destructor TOwnedCaseStringList.Destroy;
var
  i: Integer;
begin
  if OwnsObjects then
    for i := Count-1 downto 0 do
      Delete(i);
  inherited Destroy;
end;

procedure TOwnedCaseStringList.SetCapacity(NewCapacity: Integer);
var
  i: Integer;
begin
  if OwnsObjects then
    for i := NewCapacity to Count - 1 do
      Objects[i].Free;
  inherited;
end;

{ TVariableList }

procedure TVariableList.Add(Variable: TVariable);
begin
  AddObject(Variable.Name, Variable);
end;

procedure TVariableList.CheckThreadVars;
var
  i: Integer;
begin
  { Call to ensure all variable in list are set up for multithreaded use }
  for i := 0 to Count - 1 do
    (Objects[i] as TVariable).SetAllValues;
end;

constructor TVariableList.Create;
begin
  inherited;
  Duplicates := dupError;
  Sorted := True;
  CaseSensitive := True;
end;

function TVariableList.GetObjects(Index: Integer): TVariable;
begin
  Result := TVariable(inherited GetObject(Index));
end;

procedure TVariableList.SetObjects(Index: Integer; const Value: TVariable);
begin
  inherited PutObject(Index, Value);
end;

function TVariableList.VariableValues: string;
var
  i: Integer;
begin
  Result := '';
  for i := 0 to Count - 1 do begin
    if i > 0 then
      Result := Result + ', ';
    Result := Result + Objects[i].Name + '=' + FloatToStr(Objects[i].Value);
  end;
end;

{ TExprTree }

procedure TExprNode.AddNode(ANode: TExprNode);
begin
  SetLength(FChildren, nChildren+1);
  FChildren[nChildren-1] := ANode;
  Changed;
end;

function TExprNode.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
begin
  Result := Base.Compile(Evaluator);
end;

function TExprNode.Compile(EvaluatorOptions: TEvaluatorOptions): TEvaluator;
var
  Token: TToken;
begin
  Result := TEvaluator.Create;
  Result.EvaluatorOptions := EvaluatorOptions;
  Compile(Result);
  Token.ty := toSpecial;
  Token.C := 'e';
  Result.AddToken(Token);
  Result.CheckStack;
end;

function TExprNode.Compare(Node: TExprNode; ConstantsHigh: Boolean): Integer;
{ 1 for Self in greater, 0 for equal }
begin
  if Node is TIntegerNode then
    if ConstantsHigh then
      Result := -1
    else
      Result := 1
  else
    Result := CompareStr(SortText(ConstantsHigh), Node.SortText(ConstantsHigh));
end;

constructor TExprNode.Create(ABase: TTokenNode);
begin
  Base := ABase;
  {$IFOPT R+}
  Inc(ExprNodesCreated);
  {$ENDIF}
  Simplified := [siNone];
end;

procedure TExprNode.DeleteNode(AIndex: Integer);
var
  i: Integer;
begin
  Children[AIndex].Free;
  for i := AIndex to nChildren - 2 do
    FChildren[i] := FChildren[i+1];
  SetLength(FChildren, nChildren-1);
  Changed;
end;

function TExprNode.EvalDependsOn: String;
var
  i,s,e: Integer;
  d, n: string;
begin
  Result := ' ';
  for i := 0 to nChildren - 1 do begin
    d := Children[i].DependsOn;
    s := 1;
    e := 1;
    while e < Length(d) do begin
      Inc(e);
      if d[e] = ' ' then begin
        n := Copy(d,s,e-s)+' ';
        if Pos(' '+n,Result) <=0 then
          Result := Result + n;
        s := e+1;
        e := s;
      end;
    end;
  end;
  if Length(Result) > 0 then
    Delete(Result,1,1);
end;

function TExprNode.DependsOn: String;
begin
  if not FDependsOnValid then begin
    FDependsOn := EvalDependsOn;
    FDependsOnValid := True;
  end;
  Result := FDependsOn;
end;

function GCD(m, n: Longint):Longint;
var
  r: Longint;
begin
  repeat
    r := m mod n;
    if r <> 0 then begin
      m := n;
      n := r;
    end;
  until r = 0;
  Result := n;
end;

function TExprNode.nChildren: Integer;
begin
  Result := Length(FChildren);
end;

function TExprNode.nNodes: Integer;
var
  i: Integer;
begin
  Result := 1;
  for i := 0 to nChildren - 1 do
    Inc(Result, Children[i].nNodes);
end;

function TExprNode.EqualTo(AValue: Integer): Boolean;
begin
  if not Assigned(Self) then
    Result := AValue = 0
  else
    Result := (Self is TIntegerNode) and (TIntegerNode(Self).Value = AValue);
end;

function TExprNode.GetChildren(AIndex: Integer): TExprNode;
begin
  Result := FChildren[AIndex];
end;

destructor TExprNode.Destroy;
var
  i: Integer;
begin
  for i := 0 to nChildren - 1 do
    FreeAndNil(FChildren[i]);
  inherited;
  {$IFOPT R+}
  Inc(ExprNodesDestroyed);
  {$ENDIF}
end;

function TExprNode.Duplicate: TExprNode;
var
  i: Integer;
begin
  Result := Base.CreateExprNode;
  SetLength(Result.FChildren, nChildren);
  for i := 0 to nChildren-1 do
    Result.FChildren[i] := Children[i].Duplicate;
  Duplicate.Simplified := Simplified;
end;

function TExprNode.IsConstant(out Value: Double):Boolean;
begin
  Value := 0;
  Result := False;
end;

function TExprNode.MoveChild(Index: Integer): TExprNode;
begin
  Result := Children[Index];
  Children[Index] := nil;
  Changed;
end;

procedure TExprNode.SetChildren(AIndex: Integer; const Value: TExprNode);
begin
  FChildren[AIndex] := Value;
  Changed;
end;

{$IFDEF CHECKSIMPLIFY}
procedure TExprNode.SetInitialText;
begin
  try
    InitialValue := RealValue;
  except
    SetNan(InitialValue);
  end;
  InitialText := Text(True);
end;
{$ENDIF}

function TExprNode.Simplify(Actions: TSimplifyActions): TExprNode;
begin
  {$IFDEF CHECKSIMPLIFY}
  SetInitialText;
  {$ENDIF}
  Result := Self;
end;

function TExprNode.SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions): TExprNode;
begin
  Result := nil;
end;

function TExprNode.SortText(ConstantsHigh:Boolean): string;
begin
  if ConstantsHigh then begin
    if not FSortTextHighValid then begin
      FSortTextHigh := EvalSortText(True);
      FSortTextHighValid := True;
    end;
    Result := FSortTextHigh;
  end else begin
    if not FSortTextLowValid then begin
      FSortTextLow := EvalSortText(False);
      FSortTextLowValid := True;
    end;
    Result := FSortTextLow;
  end;
end;

function TExprNode.EvalSortText(ConstantsHigh: Boolean): string;
begin
  Result := DependsOn;
  if ConstantsHigh and (Result = '') then
    Result := '~ ';
  Result := Result + Text;
end;

function IsIntegerValue(Value: Double):Boolean;
begin
  { Assumption here that integer values are all reasonably small }
  Result := (Abs(Value) <= High(Integer)) and (Value = Round(Value));
end;

{ TIntegerNode }

constructor TIntegerNode.Create(AValue: Integer);
begin
  Value := AValue;
  inherited Create(nil);
end;

function TIntegerNode.Compare(Node: TExprNode; ConstantsHigh: Boolean): Integer;
var
  V: Double;
begin
  if Node.IsConstant(V) then
    Result := Sign(Value-V)
  else if ConstantsHigh then
    Result := 1
  else
    Result := -1
end;

function TIntegerNode.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toNumber;
  Token.Value := Value;
  Result := Evaluator.AddToken(Token);
end;

function TIntegerNode.Duplicate: TExprNode;
begin
  Result := TIntegerNode.Create(Value);
end;

function TIntegerNode.RealValue: Double;
begin
  Result := Value;
end;

function TIntegerNode.IsConstant(out AValue: Double):Boolean;
begin
  AValue := Value;
  Result := True;
end;

function TIntegerNode.Text(Readable:Boolean): string;
begin
  Result := IntToStr(Value);
end;

{ TRealNode }

constructor TRealNode.Create(AValue: Double);
begin
  Value := AValue;
  inherited Create(nil);
end;

function TRealNode.Compare(Node: TExprNode; ConstantsHigh: Boolean): Integer;
var
  V: Double;
begin
  if Node.IsConstant(V) then
    Result := Sign(Value-V)
  else if ConstantsHigh then
    Result := 1
  else
    Result := -1
end;

function TRealNode.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toNumber;
  Token.Value := Value;
  Result := Evaluator.AddToken(Token);
end;

function TVariable.CreateExprNode: TExprNode;
begin
  Result := TVariableNode.Create(Self);
end;

function TRealNode.Duplicate: TExprNode;
begin
  Result := TRealNode.Create(Value);
end;

function TRealNode.RealValue: Double;
begin
  Result := Value
end;

function TRealNode.IsConstant(out AValue: Double):Boolean;
begin
  AValue := Value;
  Result := True;
end;

function TRealNode.Text(Readable:Boolean): string;
begin
  Result := FloatToStr(Value);
end;

{ TVariableNode }

function TVariableNode.EvalDependsOn: String;
begin
  Result := Text + ' ';
end;

function TVariableNode.EvalSortText(ConstantsHigh:Boolean): string;
begin
  Result := Text; { This puts simple variables before expressions }
end;

function TVariableNode.RealValue: Double;
begin
  Result := (Base as TVariable).Value;
end;

procedure TVariableNode.SetAllValues;
begin
  (Base as TVariable).SetAllValues;
end;

procedure TVariableNode.SetValue(const AValue: Double);
begin
  (Base as TVariable).Value := AValue;
end;

function TVariableNode.Text(Readable:Boolean): string;
begin
  Result := (Base as TVariable).Name;
end;

{ TFunctionNode }

function TFunctionNode.RealValue: Double;
var
  i: Integer;
  Values: array of Double;
begin
  SetLength(Values, nChildren);
  for i := 0 to nChildren - 1 do
    Values[nChildren-1-i] := Children[i].RealValue;
  Result := (Base as TFunction).Value(Values);
end;

function TFunctionNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  F: TFunction;
  i, j: Integer;
  AllConstant: Boolean;
  AllInteger: Boolean;
  Values: array of Double;
  V: Double;
  Node: TExprNode;
  ConstantsAtEnd, Sorted, HasChanged: Boolean;
begin
  {$IFDEF CHECKSIMPLIFY}
  Result := inherited Simplify(Actions);
  {$ENDIF}
  { Simplify each child node, and check for all constant }
  SetLength(Values, nChildren);
  AllConstant := True;
  AllInteger := True;
  for i := 0 to nChildren - 1 do begin
    Children[i] := CheckSimplify(Children[i],Actions);
    AllConstant := AllConstant and Children[i].IsConstant(Values[nChildren-1-i]);
    AllInteger := AllConstant and (Children[i] is TIntegerNode);
  end;
  F := Base as TFunction;
  if AllConstant and not (opNoEvaluate in F.Properties) then begin
    V := F.Value(Values);
    if IsNanOrInf(V) then
      Result := nil { e.g. sqrt(-1) }
    else if AllInteger and IsIntegerValue(V) then
      Result := TIntegerNode.Create(Round(V))
    else if not AllInteger then
      Result := TRealNode.Create(V)
    else
      Result := nil;
    if Assigned(Result) then begin
      Free;
      Exit;
    end;
  end;
  Result := Self;
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}

  { Merge subtrees: +(+(a,b),c) -> +(a,b,c) }
  if opAssociative in F.Properties then begin
    i := 0;
    while i < nChildren do begin
      if Children[i].Base = Self.Base then begin
        { Bubble up }
        Node := Children[i];
        SetLength(FChildren, nChildren+Node.nChildren-1);
        { Move up nodes here }
        for j := nChildren - 1 downto i+Node.nChildren do
          Children[j] := Children[j-Node.nChildren+1];
        { Copy up nodes from below }
        for j := 0 to Node.nChildren - 1 do begin
          Children[i+j] := Node.Children[j];
          Node.Children[j] := nil;
        end;
        Inc(i,Node.nChildren);
        Node.Free;
      end else
        Inc(i);
    end;
  end;
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}

  if opCommutative in F.Properties then begin
    { Sort nodes - constants at end (+) or beginning (*) }
    ConstantsAtEnd := F = OperatorPlus;
    repeat
      Sorted := True;
      for i := 0 to nChildren - 2 do
        if Children[i].Compare(Children[i+1],ConstantsAtEnd) > 0 then begin
          Node := Children[i];
          Children[i] := Children[i+1];
          Children[i+1] := Node;
          Sorted := False;
        end;
    until Sorted;
  end;
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}

  HasChanged := False;
  if opAssociative in F.Properties then begin
    { Combine adjacent constants }
    i := 0;
    while i < nChildren-1 do begin
      Node := SimplifyPair(Children[i],Children[i+1],Actions);
      if Assigned(Node) then begin
        { Merge 2 nodes }
        HasChanged := True;
        Children[i].Free;
        Children[i] := Node;
        DeleteNode(i+1);
      end else
        Inc(i);
    end;
  end;
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}

  if (opDiscardSingleArg in F.Properties) and (nChildren = 1) then begin
    { +a -> a }
    Result := MoveChild(0);
    Free;
    Exit;
  end;

  if HasChanged then
    Result := CheckSimplify(Result,Actions);
end;

function TFunctionNode.SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions): TExprNode;
var
  VV: array [0..1] of Double;
  V: Double;
begin
  Result := inherited SimplifyPair(Node1,Node2);
  if Assigned(Result) then
    Exit;
  if opAssociative in (Base as TFunction).Properties then begin
    { Combine adjacent constants }
      if Node1.IsConstant(VV[1]) and Node2.IsConstant(VV[0]) then begin
        V := (Base as TFunction).Value(VV);
        if (Node1 is TIntegerNode) and (Node2 is TIntegerNode) then begin
          if IsIntegerValue(V) then
            Result := TIntegerNode.Create(Round(V));
        end else
          Result := TRealNode.Create(V);
      end;
  end;
end;

function TFunctionNode.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  i: Integer;
begin
  if (opAssociative in (Base as TFunction).Properties) and (nChildren > 2) then begin
    Result := Children[0].Compile(Evaluator);
    for i := 1 to nChildren - 1 do begin
      Children[i].Compile(Evaluator);
      Result := inherited Compile(Evaluator);
    end;
  end else begin
    for i := 0 to nChildren - 1 do
      Children[i].Compile(Evaluator);
    Result := inherited Compile(Evaluator);
  end;
end;

function TFunctionNode.Text(Readable:Boolean): string;
var
  F: TFunction;
  i: Integer;
begin
  F := Base as TFunction;
  Result := F.Name;
  if nChildren = 0 then
    Exit;
  Result := Result + '(';
  for I := 0 to nChildren-1 do
    if Assigned(Children[i]) then
      Result := Result + Children[i].Text(Readable) + ','
    else
      Result := Result + '?,';
  Result[Length(Result)] := ')';
end;

function TFunction.CreateExprNode: TExprNode;
begin
  Result := TFunctionNode.Create(Self);
end;

type

  TMinusNode = class(TOperatorNode)
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  end;

  TOperatorMinus = class (TOperator)
  public
    function CreateExprNode:TExprNode; override;
  end;

  TPlusNode = class(TOperatorNode)
  protected
    function SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions):TExprNode; override;
  public
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  end;

  TOperatorPlus = class (TOperator)
  public
    function CreateExprNode:TExprNode; override;
  end;

  TDivideNode = class(TOperatorNode)
  protected
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  public
    function LookForCancelations(Actions: TSimplifyActions):TExprNode;
  end;

  { TOperatorDivide }

  TOperatorDivide = class (TOperator)
  public
    function CreateExprNode:TExprNode; override;
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  { TTimesNode }

  TTimesNode = class(TOperatorNode)
  protected
    function SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions):TExprNode; override;
  public
    function EvalSortText(ConstantsHigh:Boolean = False): string; override;
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
    function Text(Readable:Boolean = False): string; override;
  end;

  TOperatorTimes = class (TOperator)
  public
    function CreateExprNode:TExprNode; override;
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

  TPowerNode = class(TOperatorNode)
  protected
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  public
    function EvalSortText(ConstantsHigh:Boolean = False): string; override;
  end;

  TOperatorPower = class (TOperator)
  public
    function CreateExprNode:TExprNode; override;
  end;

  TSqrtNode = class(TFunctionNode)
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  end;

  TSqrtFunction = class(TMonadicFunction)
    function CreateExprNode:TExprNode; override;
    function Compile(Evaluator: TEvaluator; nArgs: Integer = -1):Integer; override;
  end;

{ TOperatorMinus }

function TOperatorMinus.CreateExprNode: TExprNode;
begin
  Result := TMinusNode.Create(Self);
end;

{ TMinusNode }

function TMinusNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  Node: TExprNode;
begin
  {$IFDEF CHECKSIMPLIFY}
  InitialValue := RealValue;
  {$ENDIF}
  { -(a,b) -> +(a,-(b)) }
  Node := OperatorTimes.CreateExprNode;
  Node.AddNode(TIntegerNode.Create(-1));
  Node.AddNode(MoveChild(1));
  Result := OperatorPlus.CreateExprNode;
  Result.AddNode(MoveChild(0));
  Result.AddNode(Node);
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}
  Result := CheckSimplify(Result,Actions);
  Free;
end;

type

  TMonadicMinusNode = class(TFunctionNode)
  public
    function EvalSortText(ConstantsHigh:Boolean = False): string; override;
    function Simplify(Actions: TSimplifyActions):TExprNode; override;
  end;

  TMonadicMinusFunction = class (TMonadicFunction)
  public
    function CreateExprNode:TExprNode; override;
  end;

{ TMonadicFunctionMinuw }

function TMonadicMinusFunction.CreateExprNode: TExprNode;
begin
  Result := TMonadicMinusNode.Create(Self);
end;

{ TMonadicMinusNode }

function TMonadicMinusNode.Simplify(Actions: TSimplifyActions): TExprNode;
begin
  Result := CheckSimplify(OperatorTimes.CreateExprNode(TIntegerNode.Create(-1),MoveChild(0)),Actions);
  Free;
end;

function TMonadicMinusNode.EvalSortText(ConstantsHigh:Boolean): string;
begin
  Result := Children[0].Text + ' -';
end;

{ TOperatorPlus }

function TOperatorPlus.CreateExprNode: TExprNode;
begin
  Result := TPlusNode.Create(Self);
end;

function OppositeSign(Num, Den: TExprNode):Boolean;
  { Note limited functionality }
  function SimpleOppositeSign(Num, Den: TExprNode): Boolean;
  var
    v1, v2: Double;
  begin
    if (Num is TTimesNode) and (Num.nChildren = 2) then begin
      Result :=  Num.Children[0].EqualTo(-1) and (Num.Children[1].Compare(Den,False) = 0);
      if not Result and (Den is TTimesNode) and (Den.nChildren = 2) and (Num.Children[1].Compare(Den.Children[1],False) = 0) then
        Result := SimpleOppositeSign(Num.Children[0], Den.Children[0]);
    end else
      Result := Num.IsConstant(v1) and Den.IsConstant(v2) and (v1 = -v2);
  end;
var
  i: Integer;
begin
  {$IFDEF CHECKSIMPLIFY}
  Num.SetInitialText;
  Den.SetInitialText;
  {$ENDIF}
  Result := SimpleOppositeSign(Num,Den) or SimpleOppositeSign(Den,Num);
  if not Result then begin
    Result := (Num is TPlusNode) and (Den is TPlusNode) and (Num.nChildren = Den.nChildren);
    if Result then begin
      for i := 0 to Num.nChildren - 1 do
        if Result then
          Result := Result and OppositeSign(Num.Children[i],Den.Children[i])
    end;
  end;
end;

{ TPlusNode }

function TPlusNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  i, G: Integer;
  Node: TExprNode;
  HasChanged, AllNeg: Boolean;
  function Expandable(Node: TExprNode):Boolean;
  begin
    Result := (Node is TTimesNode)
              and (Node.nChildren = 2)
              and (Node.Children[0] is TIntegerNode)
              and (Node.Children[1] is TPlusNode)
  end;
  function LeadingConstant(i: Integer):Integer;
  begin
    Result := 1;
    if Children[i] is TIntegerNode then
      Result := TIntegerNode(Children[i]).Value
    else if Children[i] is TTimesNode then
      if Children[i].Children[0] is TIntegerNode then
        Result := TIntegerNode(Children[i].Children[0]).Value;
    if Result >= 0 then
      AllNeg := False
    else
      Result := -Result;
  end;

begin
  Result := inherited Simplify(Actions);
  if Result <> Self then
    Exit;
  if nChildren > 1 then begin
    if not (siExpand in Actions) then begin
      { (2a+2) -> 2(a+1) }
      AllNeg := True;
      G := LeadingConstant(0);
      for i := 1 to nChildren - 1 do
        if G > 1 then
          G := GCD(G, LeadingConstant(i))
        else
          Break;
      if G > 1 then begin
        if AllNeg then
          G := -G;
        for i := 0 to nChildren - 1 do
          Children[i] := OperatorDivide.CreateExprNode(Children[i],TIntegerNode.Create(G));
        Result := OperatorTimes.CreateExprNode(TIntegerNode.Create(G),Self);
        {$IFDEF CHECKSIMPLIFY}
        CheckUnchanged(Result);
        {$ENDIF}
        Result := CheckSimplify(Result,Actions);
        Exit;
      end;
    end;
    { +(... *(2,+(a,b...) ...) -> +(... *(2,b...), ... *(2,a) ) }
    { May want to supress this one }
    HasChanged := False;
    i := 0;
    while i < nChildren do
      if Expandable(Children[i]) then begin
        //WriteLn;
        //WriteLn('Exp');
        //WriteLn(Children[i].Text(True));
        Node := OperatorTimes.CreateExprNode;
        Node.AddNode(Children[i].Children[0].Duplicate); { Integer }
        Node.AddNode(Children[i].Children[1].MoveChild(0)); { a }
        Node := CheckSimplify(Node,Actions);
        AddNode(Node);
        //WriteLn(Node.Text(True));
        Children[i].Children[1].Children[0] := TIntegerNode.Create(0);
        Children[i].Changed;
        //WriteLn(Children[i].Text(True));
        //WriteLn;
        {$IFDEF CHECKSIMPLIFY}
        CheckUnchanged(Result);
        {$ENDIF}
        Children[i] := CheckSimplify(Children[i],Actions);
        HasChanged := True;
      end else
       Inc(i);
    if HasChanged then begin
      {$IFDEF CHECKSIMPLIFY}
      CheckUnchanged(Result);
      {$ENDIF}
      Result := CheckSimplify(Result,Actions);
      Exit;
    end;
  end;
end;

function TPlusNode.SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions): TExprNode;
var
  nStart1, nStart2, nNodes1, nNodes2, Base1, Base2: Integer;
  i, j, V: Integer;
  FirstNode, Match:Boolean;
  procedure StartN(out nStart, nNodes: Integer; Node: TExprNode);
  begin
    if Node is TTimesNode then begin
      nNodes := Node.nChildren;
      nStart := 0;
    end else begin
      nNodes := 1;
      nStart := -1;
    end;
  end;
  function NodeN(n, nStart: Integer; Node: TExprNode):TExprNode;
  begin
    if nStart < 0 then
      Result := Node
    else
      Result := Node.Children[n];
  end;
  procedure ZapNode(n: Integer; var nStart, nNodes:Integer; var Node: TExprNode; Replacement: TExprNode = nil);
  var
    NewNode: TExprNode;
  begin
    if nStart < 0 then begin
      { Single node - not product }
      if not FirstNode then
        Node.Free;
      Node := Replacement;
      nNodes := Ord(Assigned(Replacement));
    end else begin
      if FirstNode then
        Node := Node.Duplicate;
      if not Assigned(Replacement) then begin
        Dec(nNodes);
        if nNodes = 1 then begin
          NewNode := Node.MoveChild(1-n);
          Node.Free;
          Node := NewNode;
          nStart := -1;
        end else
          Node.DeleteNode(n);
      end else begin
        Node.Children[n].Free;
        Node.Children[n] := Replacement;
      end
    end;
  end;
  function ExpandDivide(Node1,Node2:TExprNode; out Node:TExprNode):Boolean;
  begin
    if (Node1 is TDivideNode) and not(Node2 is TDivideNode) then begin
      Node := EB([Node2, '*', Node1.Children[1], '+', Node1.Children[0], '/', Node1.Children[1]],Actions);
      Result := True;
    end else
      Result := False
  end;
  function LoadPower(Expr: TExprNode; out Base: TExprNode):Integer;
  begin
    if (Expr is TPowerNode) and SimplifiesToInt(Result, [Expr.Children[1]]) then
      Base := Expr.Children[0]
    else begin
      Base := Expr;
      Result := 1;
    end;
  end;
  var
    BaseNode1, BaseNode2: TExprNode;
    Pow1, Pow2: Integer;
    RelSign: Integer;
  function IsCombinable(Node1, Node2: TExprNode):Boolean;
  { Match: Node1 = a^n, Node2 = a^m }
  begin
    Pow1 := LoadPower(Node1, BaseNode1);
    Pow2 := LoadPower(Node2, BaseNode2);
    Result := BaseNode1.Compare(BaseNode2,False) = 0;
    if not Result then begin
      Result := OppositeSign(BaseNode1, BaseNode2);
      RelSign := -1;
    end else
      RelSign := 1;
  end;
  function IsMinusOnePower(Expr:TExprNode; out Power: TExprNode):Boolean;
  begin
    Result := (Expr is TPowerNode) and Expr.Children[0].EqualTo(-1);
    if Result then
      Power := Expr.Children[1];
  end;
begin
  Result := inherited SimplifyPair(Node1,Node2,Actions);
  if Assigned(Result) then
    Exit;
  if Node2.EqualTo(0) then begin
    { Ignore + 0 }
    Result := Node1.Duplicate;
    Exit;
  end;
  if (Node2 is TMonadicMinusNode) and (Node1.Text = Node2.Children[0].Text) then begin
    Result := TIntegerNode.Create(0);
    Exit;
  end;
  { +(*(3,j1),j1) -> +(4,j2) and similar }
//  WriteLn('+SP1:',Node1.Text(True));
//  WriteLn('+SP2:',Node2.Text(True));
//  LoadN(N1, nSub1, V1, Node1);
//  LoadN(N2, nSub2, V2, Node2);
//  if nSub1 = nSub2 then
//    if nSub1 = 1 then begin
//      if N1.Compare(N2,False) = 0 then begin
//        Result := OperatorTimes.CreateExprNode;
//        Result.AddNode(TIntegerNode.Create(V1+V2));
//        Result.AddNode(N1.Duplicate);
//        Result := CheckSimplify(Result);
//        Exit;
//      end
//    end else
//    begin
//      FirstDiff := 0;
//      for i := 1 to High(Node1.Children) do
//        if Node1.Children[i].Compare(Node2.Children[i],False) <> 0 then begin
//          if FirstDiff = 0 then
//            FirstDiff := i
//          else
//            FirstDiff := -1;
//        end;
//      if FirstDiff >= 0 then begin
//        Result := Node1.Duplicate;
//        if FirstDiff = 0 then
//          (Result.Children[0] as TIntegerNode).Value := V1+V2
//        else begin
//          { V1*a*(b+1) + V2*a*(b+c) -> a*(V1*(b+1)+V2*(b+c)) ; may want to disable this one }
//          Result.Children[FirstDiff] := EN([EN([V1,'*',Result.Children[FirstDiff]]), '+', EN([V2,'*',Node2.MoveChild(FirstDiff)])]);
//          (Result.Children[0] as TIntegerNode).Value := 1;
//        end;
//        Result := CheckSimplify(Result);
//        Exit;
//      end;
//    end;
  StartN(nStart1,nNodes1,Node1);
  StartN(nStart2,nNodes2,Node2);
  if siExpand in Actions then begin
    { Restricted subset }
    Base1 := 0;
    if NodeN(0,nStart1,Node1) is TIntegerNode then
      Inc(Base1);
    Base2 := 0;
    if NodeN(0,nStart2,Node2) is TIntegerNode then
      Inc(Base2);
    if (nNodes1-Base1) = (nNodes2-Base2) then begin
      Match := True;
      for i := 0 to nNodes1-Base1 - 1 do
        if NodeN(i+Base1,nStart1,Node1).Compare(NodeN(i+Base2,nStart2,Node2),False) <> 0 then begin
          Match := False;
          Break;
        end;
      if Match then begin
        //Writeln('+Ex1:',Node1.Text(True));
        //Writeln('+Ex2:',Node2.Text(True));
        if Base1 > 0 then
          V := (NodeN(0,nStart1,Node1) as TIntegerNode).Value
        else
          V := 1;
        if Base2 > 0 then
          V := V + (NodeN(0,nStart2,Node2) as TIntegerNode).Value
        else
          Inc(V);
        Result := Node1.Duplicate;
        if Base1 > 0 then begin
          (Result.Children[0] as TIntegerNode).Value := V;
          Result.Changed;
        end else
          Result := EN([V, '*', Result],Actions);
        //Writeln('+Ex:',Result.Text(True));
        Exit;
      end;
    end
  end else begin
    FirstNode := True;
    i := 0;
    while i < nNodes1 do begin
      j := 0;
      while j < nNodes2 do begin
        if not(NodeN(i,nStart1,Node1) is TIntegerNode)
           and IsCombinable(NodeN(i,nStart1,Node1), NodeN(j,nStart2,Node2)) then begin
          { ith node in Node1 has common factor with jth node in Node2 }
          if FirstNode then begin
            //WriteLn('+SP1:',Node1.Text(True));
            //WriteLn('+SP2:',Node2.Text(True));
          end;
          { accumulate common factors }
          if (Pow1 = Pow2) and (RelSign > 0) then begin
            Result := EN([BaseNode1.Duplicate, '^', Pow1, '*', Result],Actions,1);
            ZapNode(i,nStart1,nNodes1,Node1);
            ZapNode(j,nStart2,nNodes2,Node2);
          end else if Abs(Pow1) >= Abs(Pow2) then begin
            Result := EN([BaseNode2.Duplicate, '^', Pow2, '*', Result],Actions,1);
            BaseNode1 := EB([BaseNode1, '^', Pow1-Pow2]);
            if Odd(Pow2) and (RelSign < 0) then
              BaseNode1 := MonadicMinus.CreateExprNode(BaseNode1);
            ZapNode(i,nStart1,nNodes1,Node1, BaseNode1);
            ZapNode(j,nStart2,nNodes2,Node2);
          end else begin
            Result := EN([BaseNode1.Duplicate, '^', Pow1, '*', Result],Actions,1);
            BaseNode2 := EB([BaseNode2, '^', Pow2-Pow1]);
            if Odd(Pow1) and (RelSign < 0) then
              BaseNode2 := MonadicMinus.CreateExprNode(BaseNode2);
            ZapNode(i,nStart1,nNodes1,Node1);
            ZapNode(j,nStart2,nNodes2,Node2, BaseNode2);
          end;
          FirstNode := False;
          Dec(i);
          Break;
        end;
        Inc(j);
      end;
      Inc(i);
    end;
    if Assigned(Result) then begin
      //WriteLn('+SP-:',Result.Text(True));
      //if Assigned(Node1) then WriteLn('+SP1-:',Node1.Text(True));
      //if Assigned(Node2) then WriteLn('+SP2-:',Node2.Text(True));
      Result := EN([Node1, '+', Node2, '*', Result],Actions,1);
      //WriteLn('+SP:',Result.Text(True));
      Exit;
    end;
  end;
  if ExpandDivide(Node1,Node2,Result) or ExpandDivide(Node2,Node1,Result) then
    Exit;
  if (Node1 is TDivideNode) and (Node2 is TDivideNode) then begin
    { +( /(a,b), /(c,d) ) -> (a*d+b*c)/(b*d) }
    //WriteLn('D/:',Text(True));
    if Node1.Children[1].Compare(Node2.Children[1],True) = 0 then
      { +( /(a,b), /(c,b) ) -> (a*b+b*c)/(b*b) -> (a+c)/b }
      Result := EB([Node1.Children[0], '+', Node2.Children[0], '/', Node1.Children[1]],Actions)
    else
      Result := EN([EB([Node1.Children[0], '*', Node2.Children[1]],Actions),
                   '+', EB([Node2.Children[0], '*', Node1.Children[1]],Actions),
                   '/', EB([Node1.Children[1], '*', Node2.Children[1]],Actions)],Actions);
    //WriteLn('D/:',Result.Text(True));
    Exit;
  end;
  if IsMinusOnePower(Node1, BaseNode1) and IsMinusOnePower(Node2, BaseNode2)
     and SimplifiesToInt(Pow1, [BaseNode2, '-', BaseNode1]) then begin
    { (-1)^a + (-1)^b = (-1)^a(1+(-1)^(b-a)) }
    if Pow1 >= 0 then
      Result := EB([-1, '^', Pow1, '+', 1, '*', Node1])
    else
      Result := EB([-1, '^', -Pow1, '+', 1, '*', Node2]);
    Exit;
  end;
end;

{ TOperatorDivide }

function _Divide(a, b: Double):Double;
begin
  Result := a / b;
end;

function _ZeroDivide(a, b: Double):Double;
begin
  if (a = 0) and (b = 0) then
    Result := 0
  else
    Result := a / b;
end;

function TOperatorDivide.CreateExprNode: TExprNode;
begin
  Result := TDivideNode.Create(Self);
end;

function TOperatorDivide.Compile(Evaluator: TEvaluator; nArgs: Integer): Integer;
var
  Token: TToken;
begin
  if evoZeroOverZero in Evaluator.EvaluatorOptions then begin
    Token.ty := toDiadicFunction;
    Token.Func2 := {$IFDEF FPC}@{$ENDIF}_ZeroDivide;
    Token.DiFunc := Self;
    Result := Evaluator.AddToken(Token);
  end else
    Result := inherited Compile(Evaluator, nArgs);
end;

{ TDivideNode }

function TDivideNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  Int1: TIntegerNode;
  Node: TExprNode;
begin
  Result := inherited Simplify(Actions);
  if Result <> Self then
    Exit;
  if Children[0].EqualTo(0) then begin
    Result := MoveChild(0);
    Free;
    Exit;
  end else if Children[1] is TIntegerNode then begin
    Int1 := TIntegerNode(Children[1]);
//    if Children[0] is TIntegerNode then begin
//      Int0 := TIntegerNode(Children[0]);
//      g := GCD(Abs(Int0.Value), Abs(Int1.Value));
//      if Int1.Value < 0 then
//        g := -g;
//      if g <> 1 then begin
//        Int0.Value := Int0.Value div g;
//        Int1.Value := Int1.Value div g;
//      end;
//    end else if Int1.Value = -1 then begin
//      Result := MonadicMinus.CreateExprNode;
//      Result.AddNode(MoveChild(0));
//      {$IFDEF CHECKSIMPLIFY}
//      CheckUnchanged(Result);
//      {$ENDIF}
//      Free;
//      Exit;
//    end;
    if Int1.Value = 1 then begin
      Result := MoveChild(0);
      {$IFDEF CHECKSIMPLIFY}
      CheckUnchanged(Result);
      {$ENDIF}
      Free;
      Exit;
    end
//    else if (Children[0] is TTimesNode) and (Children[0].Children[0] is TIntegerNode) then begin
//      Int0 := TIntegerNode(Children[0].Children[0]);
//      g := GCD(Abs(Int0.Value), Abs(Int1.Value));
//      if g <> 1 then begin
//        Int0.Value := Int0.Value div g;
//        Int1.Value := Int1.Value div g;
//        {$IFDEF CHECKSIMPLIFY}
//        CheckUnchanged(Result);
//        {$ENDIF}
//        Result := CheckSimplify(Result,Actions);
//      end;
//      Exit;
//    end;
  end;
  {$IFDEF ALLSQRT}
  if Children[0] is TSqrtNode then begin
    { /( sqrt(a), b) -> sqrt( /(a,*(b,b)) ) }
    Children[1] := OperatorTimes.CreateExprNode(Children[1], Children[1].Duplicate);
    Node := Children[0];
    Children[0] := Node.Children[0];
    Node.Children[0] := Self;
    Result := CheckSimplify(Node);
    Exit;
  end;
  {$ENDIF}
  { /( /(a,b), c) -> /(a, *(b,c)) }
  if Children[0] is TDivideNode then begin
    Node := OperatorTimes.CreateExprNode;
    Node.AddNode(Children[0].MoveChild(1));
    Node.AddNode(MoveChild(1));
    Children[1] := CheckSimplify(Node,Actions);
    Node := Children[0];
    Children[0] := Node.MoveChild(0);
    Node.Free;
  end;
  { /(a, /(b,c)) -> /(*(a,c),b) }
  if Children[1] is TDivideNode then begin
    Node := OperatorTimes.CreateExprNode;
    Node.AddNode(MoveChild(0));
    Node.AddNode(Children[1].MoveChild(1));
    Children[0] := CheckSimplify(Node,Actions);
    Node := Children[1];
    Children[1] := Node.MoveChild(0);
    Node.Free;
    Result := CheckSimplify(Self,Actions);
    Exit;
  end;
  {$IFDEF CHECKSIMPLIFY}
  CheckUnchanged(Result);
  {$ENDIF}
  Result := LookForCancelations(Actions);
end;

function CancelOut(var Num, Den: TExprNode; Actions:TSimplifyActions):Boolean;
  procedure Replace(var E:TExprNode; W: TExprNode);
  begin
    E.Free;
    E := W;
    Result := True;
  end;
  procedure ReplaceIn(var E:TExprNode; W, T: TExprNode);
  var
    i: Integer;
  begin
    if not (E is TTimesNode) then
      Replace(E, W)
    else begin
      for i := 0 to E.nChildren - 1 do
        if E.Children[i] = T then begin
          E.Children[i] := W;
          T.Free;
          Exit;
        end;
      raise EExprError.Create('Factorial Cancelation Error');
    end;
  end;
  procedure FactorialCancel(FacNum, FacDen: TExprNode; var Num, Den: TExprNode; Diff, PowerNum, PowerDen:Integer);
  var
    i: Integer;
    Prod, Fac: TExprNode;
  begin
    //WriteLn('Num Before:',Num.Text(True));
    //WriteLn('Den Before:',Den.Text(True));
    //WriteLn(' Cancelling ', FacNum.Text(True), ' ', FacDen.Text(True));
    Prod := nil;
    for i := 1 to Diff do
      Prod := EN([Prod, '*', EB([FacDen.Children[0],'+',i],Actions)],Actions,1);
    Prod := EN([Prod, '^', Min(PowerNum,PowerDen)],Actions,1);
    if PowerNum >= PowerDen then
      Fac := EB([FacNum, '^', PowerNum-PowerDen],Actions)
    else
      Fac := EB([FacDen, '^', PowerDen-PowerNum],Actions);
    ReplaceIn(Num,Prod,FacNum);
    if PowerNum >= PowerDen then begin
      Num := EN([Num, '*', Fac],Actions);
      ReplaceIn(Den,TIntegerNode.Create(1),FacDen);
    end else
      ReplaceIn(Den,Fac,FacDen);
    //WriteLn('Num After:',Num.Text(True));
    //WriteLn('Den After:',Den.Text(True));
    //WriteLn;
  end;
  function LoadFac(Expr: TExprNode; out Fac: TExprNode; out Power: Integer): Boolean;
  begin
    Result := Expr.Base = FactorialFunction;
    if Result then begin
      Fac := Expr;
      Power := 1;
    end else begin
      Result := (Expr is TPowerNode) and (Expr.Children[1] is TIntegerNode)
                and LoadFac(Expr.Children[0],Fac,Power);
      if Result then
        Power := TIntegerNode(Expr.Children[1]).Value;
      Result := Result and (Power > 0);
    end;
  end;
  function FacCanCancel(Num, Den: TExprNode; out Fac: TExprNode; out Power, Diff: Integer):Boolean;
  var
    i:Integer;
  begin
    Result := LoadFac(Num, Fac, Power);
    if Result then
      Result := SimplifiesToInt(Diff, [Fac.Children[0],'-',Den.Children[0]])
                and (Abs(Diff) < 14)
    else if Num is TTimesNode then
      with Num do
        for i := 0 to nChildren-1 do
          Result := Result or FacCanCancel(Children[i],Den,Fac,Power,Diff);
  end;
  function LoadPower(Expr: TExprNode; out Base: TExprNode):Integer;
  begin
    Result := Ord(Expr is TPowerNode);
    if Result = 0 then
      Base := Expr
    else
      Base := Expr.Children[0];
  end;
var
  FacNum, FacDen, DenNode: TExprNode;
  V, PowerNum, PowerDen: Integer;
  i, g: Integer;
  Int0, Int1: TIntegerNode;
begin
  Result := False;
  if Den is TIntegerNode then begin
    Int1 := TIntegerNode(Den);
    if Num is TIntegerNode then begin
      Int0 := TIntegerNode(Num);
      g := GCD(Abs(Int0.Value), Abs(Int1.Value));
      if Int1.Value < 0 then
        g := -g;
      if g <> 1 then begin
        Int0.Value := Int0.Value div g;
        Int1.Value := Int1.Value div g;
        Result := True;
      end;
      Exit;
    end else if Int1.Value < 0 then begin
      Num := EN([Int1.Value, '*', Num],Actions);
      Int1.Value := 1;
      Result := True;
    end;
  end;
  {$IFDEF CHECKSIMPLIFY}
  Num.SetInitialText;
  Den.SetInitialText;
  {$ENDIF}
  if Num.Compare(Den,False) = 0 then begin
    Replace(Num,TIntegerNode.Create(1));
    Replace(Den,TIntegerNode.Create(1));
    Result := True;
  end else if OppositeSign(Num,Den) then begin
    Replace(Num,TIntegerNode.Create(-1));
    Replace(Den,TIntegerNode.Create(1));
    Result := True;
  end else if (Num is TSqrtNode) and (Den is TSqrtNode) then begin
    if not (siSplitDivide in Actions) then begin
      { Combine sqrt(a)/sqrt(b) -> sqrt(a/b) }
      Replace(Num, SqrtFunction.CreateExprNode(OperatorDivide.CreateExprNode(Num.MoveChild(0),Den.MoveChild(0))));
      Replace(Den, TIntegerNode.Create(1));
      Result := True;
    end else begin
      Result := CancelOut(Num.FChildren[0], Den.FChildren[0], Actions);
      if Result then begin
        Num.Changed;
        Den.Changed;
      end;
    end;
  end else if LoadFac(Den,FacDen,PowerDen)
          and FacCanCancel(Num, FacDen, FacNum, PowerNum, V)
          then begin
    { (a+m)!^p/a!^q = p >= q: [(a+m)(a+m-1)..(a+1)]^q.(a+m)!^(p-q)
                      p < q: [(a+m)(a+m-1)..(a+1)]^p/a!^(q-p) }
    if V > 0 then
      FactorialCancel(FacNum, FacDen, Num, Den, V, PowerNum, PowerDen)
    else
      FactorialCancel(FacDen, FacNum, Den, Num, -V, PowerDen, PowerNum);
    Result := True;
  end else if LoadFac(Den,FacDen,PowerDen)
              and (Num is TPlusNode) then begin
    { Look for cancelations like:
      (a! + (a+1)!)/a! = (a!/a! + (a+1)!/a!) = (1+a)
      May need to invoke at a different level }
    Result := True;
    for i := 0 to Num.nChildren - 1 do
      Result := Result and FacCanCancel(Num.Children[i],FacDen,FacNum,PowerNum,V);
    if Result then begin
      //WriteLn;
      for i := 0 to Num.nChildren - 1 do begin
        Num.Children[i] := CheckSimplify(OperatorDivide.CreateExprNode(Num.Children[i],Den.Duplicate),Actions);
        //WriteLn('DivFac:',Num.Children[i].Text(True));
      end;
      //WriteLn;
      Replace(Den,TIntegerNode.Create(1));
    end;
  end else if (LoadPower(Den, FacDen)+LoadPower(Num, FacNum) > 0) { or with complete evaluation } then begin
    if FacDen.Compare(FacNum,False) = 0 then
      { a^m/a^n = a^(m-n) }
      DenNode := TIntegerNode.Create(1)
    else if OppositeSign(FacNum, FacDen) then
      { a^m/(-a)^n = (-1)^n a^(m-n) }
      DenNode := TIntegerNode.Create(-1)
    else
      DenNode := nil;
    if Assigned(DenNode) then begin
      if Num = FacNum then begin
        { a/a^n }
        Replace(Num,DenNode);
        Den.Children[1] := EN([Den.Children[1], '-', 1]);
      end else begin
        if Den = FacDen { a^n/a } then
          Num.Children[1] := EN([Num.Children[1], '-', 1])
        else
          Num.Children[1] := EN([Num.Children[1], '-', Den.MoveChild(1)]);
        if SimplifiesToInt(i, [Num.Children[1]]) and (i < 0) then begin
          Den.Free;
          Den := Num;
          Den.Children[1] := MonadicMinus.CreateExprNode(Den.Children[1]);
          Num := DenNode;
        end else
          Replace(Den,DenNode);
      end;
    end;
  end;
end;

function TDivideNode.LookForCancelations(Actions: TSimplifyActions): TExprNode;
var
  Numerator, Denominator: TExprNode;
  NumIndex, NumEnd, DenIndex, DenStart, i, G: Integer;
  NumDep, DenDep: string;
  VI: TIntegerNode;
  function LeadingConstant(i: Integer):TIntegerNode;
  begin
    Result := nil;
    if Children[0].Children[i] is TIntegerNode then
      Result := TIntegerNode(Children[0].Children[i])
    else if Children[0].Children[i] is TTimesNode then
      if Children[0].Children[i].Children[0] is TIntegerNode then
        Result := TIntegerNode(Children[0].Children[i].Children[0])
  end;
begin
  Result := Self;
  if Children[1] is TTimesNode then begin
    Denominator := Children[1];
    DenStart := 0;
  end else begin
    Denominator := Self;
    DenStart := 1;
  end;
  { (2*a+4*b)/2 -> (1*a+2*b)/1 }
  if (Denominator.Children[DenStart] is TIntegerNode) and (Children[0] is TPlusNode) and Assigned(LeadingConstant(0)) then begin
    G := Abs(LeadingConstant(0).Value);
    for i := 1 to Children[0].nChildren - 1 do begin
      VI := LeadingConstant(i);
      if Assigned(VI) and (VI.Value <> 0) then
        G := GCD(G, Abs(VI.Value))
      else
        G := 1;
    end;
    VI := TIntegerNode(Denominator.Children[DenStart]);
    if (G > 1) and (GCD(G,Abs(VI.Value)) > 1) then begin
      G := GCD(G,Abs(VI.Value))*Sign(VI.Value);
      VI.Value := VI.Value div G;
      for i := 0 to Children[0].nChildren - 1 do begin
        with LeadingConstant(i) do
          Value := Value div G;
        Children[0].Children[i].Changed;
      end;
      Children[0].Changed;
      {$IFDEF CHECKSIMPLIFY}
      CheckUnchanged(Result);
      {$ENDIF}
      Result := CheckSimplify(Result,Actions);
      Exit;
    end;
  end;
  { /(*(a,b), *(c,d)) -> Look for cancellations like a/c, a/d ... }
  if Children[0] is TTimesNode then begin
    Numerator := Children[0];
    NumEnd := Numerator.nChildren - 1;
  end else begin
    Numerator := Self;
    NumEnd := 0;
  end;
  { Combine Sqrt() nodes on top and bottom }
  DenIndex := DenStart;
  while (DenIndex < Denominator.nChildren) and not (Denominator.Children[DenIndex] is TSqrtNode) do
    Inc(DenIndex);
  NumIndex := 0;
  while (NumIndex <= NumEnd) and not (Numerator.Children[NumIndex] is TSqrtNode) do
    Inc(NumIndex);
  if (DenIndex < Denominator.nChildren) and (NumIndex <= NumEnd) then begin
    CancelOut(Numerator.FChildren[NumIndex],Denominator.FChildren[DenIndex],Actions);
    Numerator.Changed;
    Denominator.Changed;
    {$IFDEF CHECKSIMPLIFY}
    CheckUnchanged(Result);
    {$ENDIF}
    Changed;
    Result := CheckSimplify(Result,Actions);
    Exit;
  end;
  { Look for cancellations among all pairs that depend on the same things }
  for NumIndex := 0 to NumEnd do begin
    NumDep := Numerator.Children[NumIndex].DependsOn;
    for DenIndex := DenStart to Denominator.nChildren - 1 do begin
      DenDep := Denominator.Children[DenIndex].DependsOn;
      if NumDep = DenDep then begin
        //WriteLn(Output,'Num Before:',Numerator.Text(True));
        //WriteLn(Output,'Den Before:',Denominator.Text(True));
        //WriteLn(Output, Numerator.FChildren[NumIndex].Text(True));
        //WriteLn(Output, Denominator.FChildren[DenIndex].Text(True));
        if CancelOut(Numerator.FChildren[NumIndex],Denominator.FChildren[DenIndex],Actions) then begin
          //WriteLn(Output,'After:',Result.Text(True));
          Numerator.Changed;
          Denominator.Changed;
          Changed;
          {$IFDEF CHECKSIMPLIFY}
          CheckUnchanged(Result);
          {$ENDIF}
          Result := CheckSimplify(Result,Actions);
          Exit;
        end;
      end;
    end;
  end;
end;

{ TOperatorTimes }

function _ZeroTimes(a, b: Double):Double;
begin
  Result := a * b;
  { Force 0 * Inf = 0, rather than Nan }
  if IsNan(Result) then
    if a = 0 then begin
      if IsInfinite(b) then
        Result := 0;
    end else if b = 0 then begin
      if IsInfinite(a) then
        Result := 0;
    end;
end;

function TOperatorTimes.Compile(Evaluator: TEvaluator; nArgs: Integer): Integer;
var
  Token: TToken;
begin
  if evoZeroOverZero in Evaluator.EvaluatorOptions then begin
    Token.ty := toDiadicFunction;
    Token.Func2 := {$IFDEF FPC}@{$ENDIF}_ZeroTimes;
    Token.DiFunc := Self;
    Result := Evaluator.AddToken(Token);
  end else
    Result := inherited Compile(Evaluator, nArgs);
end;

function TOperatorTimes.CreateExprNode: TExprNode;
begin
  Result := TTimesNode.Create(Self);
end;

{ TTimesNode }

function TTimesNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  i, j: Integer;
  Int: TIntegerNode;
  DivNode, Node: TExprNode;
begin
  Result := inherited Simplify(Actions);
  if Result <> Self then
    Exit;
  if Children[0] is TIntegerNode then begin
    Int := TIntegerNode(Children[0]);
    if Int.Value = 0 then begin
      Result := MoveChild(0);
      Free;
      Exit;
    end else if (Int.Value = -1) and (nChildren >= 2) and (Children[1] is TPlusNode) then begin
      for i := 0 to Children[1].nChildren - 1 do
        Children[1].Children[i] := MonadicMinus.CreateExprNode(Children[1].Children[i]);
      Children[1].Changed;
      Int.Value := 1;
      Result := CheckSimplify(Result,Actions);
      Exit;
    end;
  end;
  if siExpand in Actions then begin
    for i := 0 to nChildren-1 do
      if Children[i] is TPlusNode then begin
        Result := MoveChild(i);
        Children[i] := TIntegerNode.Create(1);
        Node := CheckSimplify(Self,Actions);
        for j := 0 to Result.nChildren - 1 do
          Result.Children[j] := OperatorTimes.CreateExprNode(Result.Children[j],Node.Duplicate);
        Node.Free;
        Result := CheckSimplify(Result,Actions);
        Exit;
      end;
  end;
  { Combine all sqrt() nodes }
  if not (siNoCombineSqrt in Actions) then
    for i := 0 to nChildren-2 do
      if Children[i] is TSqrtNode then
        for j := i+1 to nChildren-1 do
          if Children[j] is TSqrtNode then begin
            Node := SimplifyPair(Children[i],Children[j],Actions);
            { Merge 2 nodes }
            Children[i].Free;
            Children[i] := Node;
            DeleteNode(j);
            Result := CheckSimplify(Result,Actions);
            Exit;
          end;

  DivNode := nil;
  { *(... /(a,b) ...) -> /(*(...),b) }
  for i := 0 to nChildren - 1 do
    if Children[i] is TDivideNode then
      if not Assigned(DivNode) then begin
        DivNode := MoveChild(i);
        Children[i] := DivNode.Children[0];
        DivNode.Children[0] := Self;
      end else
        if DivNode.Children[1] is TTimesNode then begin
          DivNode.Children[1].AddNode(Children[i].MoveChild(1));
          Children[i].Children[1] := TIntegerNode.Create(1);
        end else begin
          Node := OperatorTimes.CreateExprNode;
          Node.AddNode(DivNode.Children[1]);
          Node.AddNode(Children[i].MoveChild(1));
          DivNode.Children[1] := Node;
          { Remove divide node }
          Node := Children[i];
          Children[i] := Node.MoveChild(0);
          Node.Free;
        end;
  if Assigned(DivNode) then
    Result := CheckSimplify(DivNode,Actions);
end;

function TTimesNode.Text(Readable: Boolean): string;
begin
  Result:=inherited Text(Readable);
  if (Length(Result) > 3) and (Copy(Result,1,3) = '-1*') then
    System.Delete(Result,2,2);
end;

function TTimesNode.SimplifyPair(Node1, Node2: TExprNode; Actions: TSimplifyActions): TExprNode;
  procedure LoadPower(out Node, Power: TExprNode; TopNode: TExprNode);
  begin
    if TopNode is TPowerNode then begin
      Node := TopNode.Children[0];
      Power := TopNode.Children[1];
    end else begin
      Node := TopNode;
      Power := nil;
    end;
  end;
  procedure CombineFactorials(Big, Small: TExprNode; Diff: Integer);
  var
    i: Integer;
  begin
    Result := EB([Small, '^', 2],Actions - [siExpand]); { Expand will cause infinite recursion }
    for i := 0 to Diff-1 do
      Result := EN([EB([Big.Children[0], '-', i],Actions), '*', Result],Actions  - [siExpand]);
  end;
var
  NodeA, PowerA, NodeB, PowerB, Sum, Difference: TExprNode;
  Diff: Integer;
begin
  Result := inherited SimplifyPair(Node1,Node2,Actions);
  if Assigned(Result) then
    Exit;
  if Node1.EqualTo(1) then begin
    Result := Node2.Duplicate;
    Exit;
  end;
  if not (siNoCombineSqrt in Actions) and (Node1 is TSqrtNode) and (Node2 is TSqrtNode) then begin
    { Sqrt(a).Sqrt(b) -> Sqrt(a*b) }
    Result := SqrtFunction.CreateExprNode(
                OperatorTimes.CreateExprNode(Node1.MoveChild(0),Node2.MoveChild(0)));
    Result := CheckSimplify(Result,Actions);
    Exit;
  end;
  LoadPower(NodeA, PowerA, Node1);
  LoadPower(NodeB, PowerB, Node2);
  if (NodeA.Compare(NodeB,False) = 0)
     and (not (siExpand in Actions)
          or (NodeA.nChildren = 0)) then begin
    { a^n.a^m -> a^(n+m); if siExpand only combine basic tokens }
    Sum := EB([PowerA, '+', PowerB],Actions,1);
    if (NodeA.Base = SqrtMinusOne) and SimplifiesToInt(Diff,[Sum]) then begin
      { Special case for sqrt(-1) }
      if Odd(Diff) then begin
        if Odd(Diff div 2) then
          Result := EB([NodeA, '*', -1])
        else
          Result := NodeA.Duplicate
      end else
        if Odd(Diff div 2) then
          Result := TIntegerNode.Create(-1)
        else
          Result := TIntegerNode.Create(1);
      Sum.Free;
    end else if NodeA.EqualTo(-1) and (siMinusOneInteger in Actions) then begin
      { (-1)^n(-1)^m = (-1)^(n+m) = (-1)(n-m) so pick simplest }
      if not SimplifiesToInt(Diff, [Sum]) then begin
        Difference := EB([PowerA, '-', PowerB],Actions,1);
        if SimplifiesToInt(Diff, [Difference]) or (Difference.nNodes < Sum.nNodes) then begin
          Sum.Free;
          Sum := Difference;
        end else
          Difference.Free;
      end;
      Result := EN([NodeA.Duplicate,'^',Sum],Actions);
    end else
      Result := EN([NodeA.Duplicate,'^',Sum],Actions);
    Exit;
  end;
  if (NodeA.Base = FactorialFunction) and (NodeB.Base = FactorialFunction)
     and (PowerA <> nil) and (PowerB = nil)
     and SimplifiesToInt(Diff, [NodeA.Children[0], '-', NodeB.Children[0]]) and (Diff < 0)
     then begin
    { (a!^n.(a+c)! = a!^(n-1).(a+c)!^2 }
    { Other simplifications possible }
    CombineFactorials(Node2, NodeA, -Diff);
    Node1 := EN([NodeA.Duplicate, '^', EB([PowerA, '-', 1])]);
    Result := EN([Node1, '*', Result]);
    Exit;
  end;

  if (Node1.Base = FactorialFunction) and (Node2.Base = FactorialFunction)
     and SimplifiesToInt(Diff, [Node1.Children[0], '-', Node2.Children[0]])
     then
    { (a+1)!.a! = (a+1).a!^2 }
    if Diff < 0 then
      CombineFactorials(Node2,Node1,-Diff)
    else
      CombineFactorials(Node1,Node2,Diff)
end;

function TTimesNode.EvalSortText(ConstantsHigh: Boolean): string;
var
  i:Integer;
begin
  if Children[0] is TIntegerNode then begin
    { want 2*a and 3*a to be adjacent }
    Result := DependsOn;
    if ConstantsHigh and (Result = '') then
      Result := '~ ';
    for i := 1 to nChildren-1 do begin
      if Children[i] is TVariableNode then
       Result := Result + ' '; { Force simple expressions earlier }
      Result := Result + Children[i].Text + ' ';
    end;
    Result := Result + '* ' + Children[0].Text;
  end else
    Result := inherited EvalSortText(ConstantsHigh);
end;

{ TTokenNode }

function TTokenNode.CreateExprNode(Node: TExprNode): TExprNode;
begin
  Result := Self.CreateExprNode;
  Result.AddNode(Node);
end;

function TTokenNode.CreateExprNode(Node1, Node2: TExprNode): TExprNode;
begin
  Result := Self.CreateExprNode;
  Result.AddNode(Node1);
  Result.AddNode(Node2);
end;

{ TSqrtFunction }

function _ImagSqrt(a: Double):Double;
begin
  Result := Sqrt(-a);
end;

function TSqrtFunction.Compile(Evaluator: TEvaluator; nArgs: Integer): Integer;
var
  Token: TToken;
begin
  if evoPureImag in Evaluator.EvaluatorOptions then begin
    Token.ty := toMonadicFunction;
    Token.Func1 := {$IFDEF FPC}@{$ENDIF}_ImagSqrt;
    Token.MonFunc := Self;
    Result := Evaluator.AddToken(Token);
    Inc(Evaluator.nSqrts);
  end else
    Result := inherited Compile(Evaluator, nArgs);
end;

function TSqrtFunction.CreateExprNode: TExprNode;
begin
  Result := TSqrtNode.Create(Self);
end;

{ TPowerNode }

function TPowerNode.Simplify(Actions: TSimplifyActions): TExprNode;
var
  Int1: TIntegerNode;
  Node: TExprNode;
  i: Integer;
  procedure TakeOut(NodeIndex: Integer; Multiply: Boolean);
  begin
    if Multiply then
      Result := OperatorTimes.CreateExprNode(Result, Base.CreateExprNode(Children[0].MoveChild(NodeIndex),Int1.Duplicate))
    else
      Result := OperatorDivide.CreateExprNode(Result, Base.CreateExprNode(Children[0].MoveChild(NodeIndex),Int1.Duplicate));
    Children[0].Children[NodeIndex] := TIntegerNode.Create(1);
    Children[0].Changed;
  end;
begin
  Result := inherited Simplify(Actions);
  if Result <> Self then
    Exit;
  if Children[1] is TIntegerNode then begin
    Int1 := TIntegerNode(Children[1]);
    if Int1.Value = 0 then begin
      Result := TIntegerNode.Create(1);
      Free;
      Exit;
    end;
    if Int1.Value = 1 then begin
      Result := MoveChild(0);
      Free;
      Exit;
    end;
    { Take constants outside }
    if Children[0] is TTimesNode then
      for i := 0 to nChildren-1 do
        TakeOut(i,True)
    else if Children[0] is TDivideNode then begin
      TakeOut(0,True);
      TakeOut(1,False);
    end else if (Children[0] is TSqrtNode) and not Odd(Int1.Value) then begin
      Node := Children[0].MoveChild(0);
      Children[0].Free;
      Children[0] := Node;
      Int1.Value := Int1.Value div 2;
      Result := CheckSimplify(Result,Actions);
      Exit;
    end else if Children[0] is TPowerNode then begin
      { (a^2)^3 -> a^6 }
      Result := MoveChild(0);
      Result.Children[1] := OperatorTimes.CreateExprNode(MoveChild(1), Result.Children[1]);
      Free;
    end else if (siExpand in Actions) and (Children[0].nChildren <> 0) then begin
      Result := OperatorTimes.CreateExprNode(MoveChild(0));
      for i := 2 to Int1.Value do
        Result.AddNode(Result.Children[0].Duplicate);
      Free;
    end;
    if Result <> Self then
      Result := CheckSimplify(Result,Actions);
  end;
end;

function TPowerNode.EvalSortText(ConstantsHigh:Boolean): string;
begin
  { Force powers of same object to be together }
  Result := Children[0].SortText + '  ^ ' + Children[1].SortText;
end;

{ TOperatorPower }

function TOperatorPower.CreateExprNode: TExprNode;
begin
  Result := TPowerNode.Create(Self);
end;

{ TOperatorNode }

function TOperatorNode.Text(Readable:Boolean): string;
var
  i: Integer;
  ct, op: string;
  
  function ChildText(Node:TExprNode):string;
  var
    BracketsNeeded: Boolean;
  begin
    BracketsNeeded := False;
    if Assigned(Node) then begin
      if Node is TOperatorNode then
        BracketsNeeded := (Base as TOperator).Priority >= (Node.Base as TOperator).Priority;
      Result := Node.Text(True);
    end else
      Result := '?';
    if BracketsNeeded then
      Result := '('+Result+')';
  end;
begin
  if Readable and (nChildren > 0) then begin
    Result := ChildText(Children[0]);
    for i := 1 to nChildren-1 do begin
      if Assigned(Children[i]) then
        ct := ChildText(Children[i])
      else
        ct := '?';
      op := (Base as TFunction).Name;
      //if (op = '+') and (ct[1] = '-') and not (Children[i] is TPowerNode) then
      //  { +-2*a -> -2*a }
      //  op := '';
      if (op = '+') and (ct[1] = '-') and ((Children[i].nChildren = 0) or (Children[i] is TTimesNode and (Children[i].Children[0].nChildren = 0)) ) then
        { +-2*a -> -2*a }
        op := '';
      Result := Result + op + ct;
    end;
  end else
    Result := inherited Text(Readable);
end;

{ TSqrtNode }

function TSqrtNode.Simplify(Actions: TSimplifyActions): TExprNode;
  procedure LookForSquares(var Node: TExprNode; Multiply: Boolean);
  var
    i, v, f, f2: Integer;
    PreviousResult: TExprNode;
    procedure AddNode(ANode: TExprNode; DeleteOld: Boolean = True);
    begin
      if Multiply then
        Result := OperatorTimes.CreateExprNode(Result,ANode)
      else
        Result := OperatorDivide.CreateExprNode(Result,ANode);
      if DeleteOld then begin
        Node.Free;
        Node := TIntegerNode.Create(1);
      end;
    end;
  begin
    if Node is TIntegerNode then begin
      v := TIntegerNode(Node).Value;
      if (v > 1) and (Sqr(Round(Sqrt(v))) = v) then
        AddNode(TIntegerNode.Create(Round(Sqrt(v))))
      else begin
        f := 2;
        f2 := Sqr(f);
        while f2 < v do begin
          if Abs(v) mod f2 = 0 then begin
            AddNode(TIntegerNode.Create(f),False);
            v := v div f2;
          end else
            Inc(f);
          f2 := Sqr(f);
        end;
        TIntegerNode(Node).Value := v;
      end;
    end else if (Node is TPowerNode) and (Node.Children[1] is TIntegerNode) then begin
      v := TIntegerNode(Node.Children[1]).Value;
      if not Odd(v) then begin
        TIntegerNode(Node.Children[1]).Value := v div 2;
        Node.Changed;
        AddNode(Node.Duplicate);
      end;
    end else if Node is TDivideNode then begin
      PreviousResult := Result;
      LookForSquares(Node.FChildren[0],Multiply);
      LookForSquares(Node.FChildren[1],not Multiply);
      if PreviousResult <> Result then
        Node.Changed;
    end else if Node is TTimesNode then begin
      PreviousResult := Result;
      for i := 0 to Node.nChildren - 1 do
        LookForSquares(Node.FChildren[i],Multiply);
      if PreviousResult <> Result then
        Node.Changed;
    end;
  end;
var
  i: Integer;
begin
  Result := inherited Simplify(Actions);
  if Result <> Self then
    Exit;
  if (Children[0] is TDivideNode) and ((siSplitDivide in Actions) or Children[0].Children[0].EqualTo(1)) then begin
    { Sqrt(a/b) -> Sqrt(a)/Sqrt(b) }
    Result := MoveChild(0);
    Result.Children[0] := SqrtFunction.CreateExprNode(Result.Children[0]);
    Result.Children[1] := SqrtFunction.CreateExprNode(Result.Children[1]);
    Result.Changed;
    Free;
  end else if (siNoCombineSqrt in Actions) and (Children[0] is TTimesNode) then begin
    Result := MoveChild(0);
    with Result do begin
      { Sqrt(a*b) -> Sqrt(a)*Sqrt(b) }
      for i := 0 to nChildren-1 do
        Children[i] := SqrtFunction.CreateExprNode(Children[i]);
    end;
    Free;
  end else
    LookForSquares(FChildren[0], True);
  if Result <> Self then
    Result := CheckSimplify(Result,Actions);
end;

{ TEvaluator }

function TEvaluator.Value(Resume: Boolean): Double;
var
  p1: PDouble;
  i: Integer;
  State: TEvaluatorState;
begin
  if FTokSize = 0 then
    raise EExprError.Create('TExpr: No Expression');
  SetLength(ThreadStates, nWorkers+1);
  ThreadStates[ThreadIndex-1] := @State;
  if Resume then begin
    if not Suspended then
      raise EExprError.Create('TExpr: Not suspended');
    State := SavedState;
  end else
    with State do begin
      t := @FTokList[0];
      p := @Stack[MaxStackDepth];
      Inc(p);
      LoopDepth := 0;
    end;
  FSuspended := False;
  with State do
    try
      while t^.ty <> toSpecial do begin
        case t^.ty of
          toPVariable:   begin Dec(p); p^ := t^.PValue^; end;
          toThreadVariable:
                         begin
                           Dec(p);
                           p^ := t^.Variable.Values[ThreadIndex-1];
                         end;
          toStore:       t^.PValue^ := p^;
          toThreadStore: t^.Variable.Values[ThreadIndex-1] := p^;
          toDrop:        if t^.Offset = 0 then
                           Inc(p)
                         else begin
                           Inc(p, t^.Offset);
                           p1 := p;
                           Dec(p1);
                           for i := 1 to t^.Offset do begin
                             p^ := p1^;
                             Dec(p);
                             Dec(p1);
                           end;
                           Inc(p);
                         end;
          toDup:         begin p1 := p; Dec(p); Inc(p1,t^.Offset); p^ := p1^; end;
          toNumber:      begin Dec(p); p^ := t^.Value; end;
          toMonadicFunction: p^ := t^.Func1(p^);
          toDiadicFunction:  begin
                               p1 := p;
                               Inc(p);
                               p^ := t^.Func2(p^, p1^);
                             end;
          toArrayLow..toArrayLowMax:
                          begin Dec(p); p^ := t^.Arr.LowerBound(Ord(t^.ty)-Ord(toArrayLow)); end;
          toArrayHigh..toArrayHighMax:
                          begin Dec(p); p^ := t^.Arr.UpperBound(Ord(t^.ty)-Ord(toArrayHigh)); end;
          toArrayStore:   begin
                            p1 := p;
                            Inc(p, t^.Func.Arguments);
                            t^.StoreProc(p1,t^.Func.Arguments);
                            p^ := p1^; { Result is value stored }
                          end;
          toLoopVar:      begin Dec(p); p^ := LoopVars[t^.Offset]; end;
          toIterationStart:
                          t^.IFunc.StartIteration(Self);
          { toIterationStart normally followed by toStartLoop }
          toStartLoop:    begin
                            Inc(LoopDepth);
                            { Start End on stack; replaced by default loop result (0) }
                            LoopEnds[LoopDepth] := p^ + 0.5;
                            Inc(p); { Pop end }
                            LoopVars[LoopDepth] := p^-1; { Start - 1 so toEndLoop test can be used }
                            p^ := 0; { Default loop result }
                            Inc(t,t^.Offset); { Jump toEndLoop }
                            Continue;
                          end;
          toIterationNext: { Normally just before toEndLoop }
                          begin
                            Inc(nValues[LoopDepth]);
                            p^ := t^.IFunc.NextIteration(Self,p^); { TODO: generalize for <> 1 argument }
                          end;
          toEndLoop:      begin
                            LoopVars[LoopDepth] := LoopVars[LoopDepth] + 1;
                            if LoopVars[LoopDepth] < LoopEnds[LoopDepth] then begin
                              Dec(t,t^.Offset); { Jump to just after toStartLoop }
                              Inc(p); { Discard loop body result }
                              Continue;
                            end else
                              Dec(LoopDepth);
                          end;
    //      toFunction:
    //        begin
    //          p1 := p;
    //          Inc(p, t^.Func.Arguments - 1);
    //          {$IFDEF FPC}
    //          fudge := TFunctionCaller(@t^.Func.Value); { Fudged conversion of PDouble to array of Double }
    //          p^ := fudge(p1,t^.Func.Arguments - 1);
    //          {$ELSE}
    //          fudge.Real := t^.Func.Value;
    //          p^ := fudge.Fake(p1,t^.Func.Arguments - 1);
    //          // Original - fails with bound checking: p^ := t^.Func.Value(p1^);
    //          {$ENDIF}
    //        end;
        else
          { toFunction, toFunction1... }
          p1 := p;
          Inc(p, Ord(t^.ty)-Ord(toFunction)-1);
          p^ := t^.FuncvF(p1,Ord(t^.ty)-Ord(toFunction)-1);
        end;
        Inc(t);
      end;
    except
      on EAbort do begin
        { To suspend execution, raise EAbort in the function;
          on resume, same function call is re-executed }
        FSuspended := True;
        { Set up stack for re-execution of the same statement }
        case t^.ty of
          toMonadicFunction: ; { Had no effect }
          toDiadicFunction, toArrayStore: Dec(p);
        else
          i := Ord(t^.ty)-Ord(toFunction)-1;
          if i >= -1 then
            Dec(p,i)
          else
            raise EExprError.Create('TExpr: Can''t suspend');
        end;
        SavedState := State;
        raise;
      end;
    end;
  Result := State.p^;
end;

function TEvaluator.AddToken(Token: TToken):Integer;
begin
  SetLength(ThreadStates, nWorkers+1);
  Result := FTokSize;
  Inc(FTokSize);
  if FTokSize > Length(FTokList) then begin
    SetLength(FTokList,FTokSize+10);
  end;
  FTokList[FTokSize-1] := Token;
end;

procedure TEvaluator.ClearTokList;
begin
  FTokList := nil;
  FTokSize := 0;
  FSuspended := False;
end;

destructor TEvaluator.Destroy;
begin
  ClearTokList;
  inherited Destroy;
end;

procedure TEvaluator.CheckStack;
var
  sp, n: Integer;
  LastType: toType;
  t: ^TToken;
begin
  { Dry run of GetValue, to ensure no stack problems }
  t := @FTokList[0];
  sp := MaxStackDepth + 1;
  while t^.ty <> toSpecial do begin
    case t^.ty of
      toArrayLow..toArrayLowMax, toArrayHigh..toArrayHighMax, toLoopVar,
      toPVariable, toThreadVariable, toNumber:
                           Dec(sp);
      toDup:               if sp + t^.Offset >= MaxStackDepth+1 then
                             raise EExprError.Create('TExpr: Stack Problem in Dup')
                           else
                             Dec(sp);
      toDrop:              if sp + t^.Offset >= MaxStackDepth+1 then
                             raise EExprError.Create('TExpr: Stack Problem in drop')
                           else
                             Inc(sp);
      toArrayStore:        Inc(sp, t^.Func.Arguments);
      toDiadicFunction:    Inc(sp);
      toStartLoop:         Inc(sp,2);
      toMonadicFunction, toStore, toThreadStore, toIterationStart, toIterationNext,
      toEndLoop:           { Skip }
    else
      { toFunction, toFunction1, ... }
      n := Ord(t^.ty) - Ord(toFunction);
      if n < 0 then
        raise EExprError.Create('TExpr: Bad Token Type');
      Inc(sp, n-1);
    end;
    LastType := t^.ty;
    if ((LastType = toDrop) and (sp > MaxStackDepth+1))
       or ((LastType = toStartLoop) and ((sp < 0) or (sp > MaxStackDepth+1)))
       or (not (LastType in [toDrop,toStartLoop]) and ((sp < 0) or (sp > MaxStackDepth))) then
      raise EExprError.Create('TExpr: Stack Problem');
    Inc(t);
  end;
  if sp <> MaxStackDepth then
    raise EExprError.Create('TExpr: Stack Imbalance ');
end;

{ TImplicitLoopVariable }

function TImplicitLoopVariable.Compile(Evaluator: TEvaluator; nArgs: Integer):Integer;
var
  Token: TToken;
begin
  Token.ty := toLoopVar;
  Token.Offset := Offset;
  Result := Evaluator.AddToken(Token);
end;

constructor TImplicitLoopVariable.Create(s: string; AOffset: Integer);
begin
  inherited Create(s,0);
  Offset := AOffset;
end;

{ TIteratingFunction }

function TIteratingFunction.Compile(Evaluator: TEvaluator; nArgs: Integer): Integer;
var
  Token: TToken;
begin
  if nArgs < 0 then begin
    { At loop end in iteration }
    Token.ty := toIterationNext;
    Token.IFunc := Self;
    Result := Evaluator.AddToken(Token);
  end else
    Result := inherited Compile(Evaluator, nArgs); { Normal call }
end;

procedure TIteratingFunction.StartIteration(Evaluator: TEvaluator);
begin
  with Evaluator.ThreadStates[ThreadIndex]^ do begin
    { Note that LoopDepth is set for current loop, not new loop }
    nValues[LoopDepth+1] := 0;
    Sums[LoopDepth+1] := 0;
  end;
end;

type

  TSimpleIteratingFunction = class(TIteratingFunction)
  public
    constructor Create(s:String; n:Integer; AProperties: TOperatorProperties = []);
  end;

  TSumFunction = class(TSimpleIteratingFunction)
  public
    function NextIteration(Evaluator: TEvaluator; var a:array of Double): Double; override;
    function Value(var a:array of Double):Double; override;
  end;

  TMeanFunction = class(TSimpleIteratingFunction)
  public
    function NextIteration(Evaluator: TEvaluator; var a:array of Double): Double; override;
    function Value(var a:array of Double):Double; override;
  end;

  TMaxFunction = class(TSimpleIteratingFunction)
  public
    function NextIteration(Evaluator: TEvaluator; var a:array of Double): Double; override;
    function Value(var a:array of Double):Double; override;
  end;

  TMinFunction = class(TSimpleIteratingFunction)
  public
    function NextIteration(Evaluator: TEvaluator; var a:array of Double): Double; override;
    function Value(var a:array of Double):Double; override;
  end;

  { TSimpleIteratingFunction }

constructor TSimpleIteratingFunction.Create(s: String; n: Integer; AProperties: TOperatorProperties);
begin
  inherited Create(s, n, AProperties);
end;

{ TSumFunction }

function TSumFunction.NextIteration(Evaluator: TEvaluator; var a: array of Double): Double;
begin
  with Evaluator.ThreadStates[ThreadIndex]^ do begin
    Sums[LoopDepth] := Sums[LoopDepth] + a[0];
    Result := Sums[LoopDepth];
  end;
end;

function TSumFunction.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to High(a) do
    Result := Result + a[i];
end;

{ TMeanFunction }

function TMeanFunction.NextIteration(Evaluator: TEvaluator; var a: array of Double): Double;
begin
  with Evaluator.ThreadStates[ThreadIndex]^ do begin
    Sums[LoopDepth] := Sums[LoopDepth] + a[0];
    Result := Sums[LoopDepth]/nValues[LoopDepth];
  end;
end;

function TMeanFunction.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to High(a) do
    Result := Result + a[i];
  Result := Result/Length(a);
end;

{ TMaxFunction }

function TMaxFunction.NextIteration(Evaluator: TEvaluator; var a: array of Double): Double;
begin
  with Evaluator.ThreadStates[ThreadIndex]^ do begin
    if (nValues[LoopDepth] = 1) or (a[0] > Sums[LoopDepth]) then
      Sums[LoopDepth] := a[0];
    Result := Sums[LoopDepth];
  end;
end;

function TMaxFunction.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := a[0];
  for i := 1 to High(a) do
    Result := Max(a[i],Result);
end;

{ TMinFunction }

function TMinFunction.NextIteration(Evaluator: TEvaluator; var a: array of Double): Double;
begin
  with Evaluator.ThreadStates[ThreadIndex]^ do begin
    if (nValues[LoopDepth] = 1) or (a[0] < Sums[LoopDepth]) then
      Sums[LoopDepth] := a[0];
    Result := Sums[LoopDepth];
  end;
end;

function TMinFunction.Value(var a: array of Double): Double;
var
  i: Integer;
begin
  Result := a[0];
  for i := 1 to High(a) do
    Result := Min(a[i],Result);
end;

procedure MakeImplicitLoopVariablesAndFunctions;
var
  i: Integer;
  n: Char;
begin
  ImplicitLoopVariables := TCaseStringList.Create;
  ImplicitLoopVariables.Duplicates := dupError;
  ImplicitLoopVariables.Sorted := True;
  for i := 1 to MaxLoopDepth do begin
    n := Char(Ord('i')+i-1);
    ImplicitLoopVariables.AddObject(n,TImplicitLoopVariable.Create(n,i));
  end;
  ImplicitLoopVariables.AddObject('sum',TSumFunction.Create('sum',1));
  ImplicitLoopVariables.AddObject('mean',TMeanFunction.Create('mean',1));
  ImplicitLoopVariables.AddObject('max',TMaxFunction.Create('max',1));
  ImplicitLoopVariables.AddObject('min',TMinFunction.Create('min',1));
end;

{ TVectorVariable }

constructor TVectorVariable.Create(s: String; Size: Integer; IsBoundChecked: Boolean);
begin
  inherited Create(s,1);
  SetLength(Values,Size);
  CheckBounds := IsBoundChecked;
end;

function TVectorVariable.LowerBound(Bound: Integer): Integer;
begin
  Result := 0;
end;

procedure TVectorVariable.Store(var a: array of Double);
begin
  if (a[1] > -0.5) and (a[1] < Length(Values)-0.5) then
    Values[Round(a[1])] := a[0]
  else if CheckBounds then
    raise EExprError.CreateFmt('Write to %s[%g] bound = %d',[Name,a[1], Length(Values)]);
end;

function TVectorVariable.UpperBound(Bound: Integer): Integer;
begin
  Result := High(Values);
end;

function TVectorVariable.Value(var a: array of Double): Double;
begin
  if (a[0] > -0.5) and (a[0] < Length(Values)-0.5) then
    Result := Values[Round(a[0])]
  else if CheckBounds then
    raise EExprError.CreateFmt('Read from %s[%g] bound = %d',[Name,a[0], Length(Values)])
  else
    Result := 0;
end;

initialization

MakeImplicitLoopVariablesAndFunctions;

GlobalFunctions := TCaseStringList.Create;
GlobalFunctions.Duplicates := dupError;
GlobalFunctions.Sorted := True;
GlobalFunctions.CaseSensitive := True;

GlobalVariables := TCaseStringList.Create;
GlobalVariables.Duplicates := dupError;
GlobalVariables.Sorted := True;
GlobalVariables.CaseSensitive := True;

FundamentalConstants := TCaseStringList.Create;
FundamentalConstants.Duplicates := dupError;
FundamentalConstants.Sorted := True;
FundamentalConstants.CaseSensitive := True;

LoadedFunctions := TCaseStringList.Create;
LoadedFunctions.Duplicates := dupError;
LoadedFunctions.Sorted := True;
LoadedFunctions.CaseSensitive := True;

OperatorMinus := TOperatorMinus.Create({$IFDEF FPC}@{$ENDIF}_Minus, '-', 10);
GlobalFunctions.AddObject('-',OperatorMinus);

OperatorPlus := TOperatorPlus.Create({$IFDEF FPC}@{$ENDIF}_Plus, '+', 10, [opAssociative, opCommutative, opDiscardSingleArg]);
GlobalFunctions.AddObject('+', OperatorPlus);

OperatorDivide := TOperatorDivide.Create({$IFDEF FPC}@{$ENDIF}_Divide, '/',  20 );
GlobalFunctions.AddObject('/', OperatorDivide);

OperatorTimes := TOperatorTimes.Create({$IFDEF FPC}@{$ENDIF}_Times, '*',  20, [opAssociative, opCommutative, opDiscardSingleArg]);
GlobalFunctions.AddObject('*', OperatorTimes);

OperatorPower := TOperatorPower.Create({$IFDEF FPC}@{$ENDIF}_Power, '^',  25);
GlobalFunctions.AddObject('^', OperatorPower);

//OperatorPowerAlt := TOperatorPower.Create({$IFDEF FPC}@{$ENDIF}_Power, '**',  25);
//GlobalFunctions.AddObject('**', OperatorPowerAlt);

Op(OperatorModulus, {$IFDEF FPC}@{$ENDIF}_Modulus, '%',  20);
Op(OperatorShiftRight, {$IFDEF FPC}@{$ENDIF}_ShiftRight, '>>',  8);
Op(OperatorShiftLeft, {$IFDEF FPC}@{$ENDIF}_ShiftLeft, '<<',  8);
Op(OperatorGreaterThan, {$IFDEF FPC}@{$ENDIF}_GreaterThan, '>',  5);
Op(OperatorLessThan, {$IFDEF FPC}@{$ENDIF}_LessThan, '<',  5);
Op(OperatorEquals, {$IFDEF FPC}@{$ENDIF}_Equals, '=',  5);
Op(OperatorLessThanOrEqual, {$IFDEF FPC}@{$ENDIF}_LessThanOrEqual, '<=',  5);
Op(OperatorGreaterThanOrEqual, {$IFDEF FPC}@{$ENDIF}_GreaterThanOrEqual, '>=',  5);
Op(OperatorNotEqual, {$IFDEF FPC}@{$ENDIF}_NotEqual, '<>',  5);
Op(OperatorNotEqualAlt, {$IFDEF FPC}@{$ENDIF}_NotEqual, '!=',  5);
Op(OperatorAnd, {$IFDEF FPC}@{$ENDIF}_And, '&',  3, [opAssociative, opCommutative, opDiscardSingleArg]);
Op(OperatorOr, {$IFDEF FPC}@{$ENDIF}_Or, '|',  3, [opAssociative, opCommutative, opDiscardSingleArg]);

MonadicMinus := TMonadicMinusFunction.Create({$IFDEF FPC}@{$ENDIF}_MonadicMinus, '-');

ConstantPi := TConstant.Create(Pi, 'pi');
GlobalFunctions.AddObject('pi', ConstantPi);

PolyFunction := TVaradicFunction.Create({$IFDEF FPC}@{$ENDIF}_Poly, 2, 'poly');
GlobalFunctions.AddObject('poly', PolyFunction);

MinFunction := TVaradicFunction.Create({$IFDEF FPC}@{$ENDIF}_Min, 2, 'min', [opAssociative, opCommutative]);
GlobalFunctions.AddObject('min', MinFunction);

MaxFunction := TVaradicFunction.Create({$IFDEF FPC}@{$ENDIF}_Max, 2, 'max', [opAssociative, opCommutative]);
GlobalFunctions.AddObject('max', MaxFunction);

VoigtFunction := TNadicFunction.Create({$IFDEF FPC}@{$ENDIF}_Voigt, 3, 'voigt');
GlobalFunctions.AddObject('voigt', VoigtFunction);

SqrtFunction := TSqrtFunction.Create({$IFDEF FPC}@{$ENDIF}_Sqrt, 'sqrt');
GlobalFunctions.AddObject('sqrt', SqrtFunction);

RegisterMonadicFunction(AbsFunction, {$IFDEF FPC}@{$ENDIF}_Abs, 'abs');
RegisterMonadicFunction(SinFunction, {$IFDEF FPC}@{$ENDIF}_Sin, 'sin');
RegisterMonadicFunction(CosFunction, {$IFDEF FPC}@{$ENDIF}_Cos, 'cos');
RegisterMonadicFunction(TanFunction, {$IFDEF FPC}@{$ENDIF}_Tan, 'tan');
RegisterMonadicFunction(AcosFunction, {$IFDEF FPC}@{$ENDIF}_Acos, 'acos');
RegisterMonadicFunction(AsinFunction, {$IFDEF FPC}@{$ENDIF}_Asin, 'asin');
RegisterMonadicFunction(AtanFunction, {$IFDEF FPC}@{$ENDIF}_Atan, 'atan');
RegisterMonadicFunction(AcosFunctionAlt, {$IFDEF FPC}@{$ENDIF}_Acos, 'arccos');
RegisterMonadicFunction(AsinFunctionAlt, {$IFDEF FPC}@{$ENDIF}_Asin, 'arcsin');
RegisterMonadicFunction(AtanFunctionAlt, {$IFDEF FPC}@{$ENDIF}_Atan, 'arctan');
RegisterMonadicFunction(ExpFunction, {$IFDEF FPC}@{$ENDIF}_Exp, 'exp');
RegisterMonadicFunction(LnFunction, {$IFDEF FPC}@{$ENDIF}_Ln, 'ln');
RegisterMonadicFunction(Log10Function, {$IFDEF FPC}@{$ENDIF}_Log10, 'log10');
RegisterMonadicFunction(SqrFunction, {$IFDEF FPC}@{$ENDIF}_Sqr, 'sqr');
RegisterMonadicFunction(AirFunction, {$IFDEF FPC}@{$ENDIF}Air, 'air');
RegisterMonadicFunction(nQuartzFunction, {$IFDEF FPC}@{$ENDIF}nQuartz, 'nquartz');
RegisterMonadicFunction(VacFunction, {$IFDEF FPC}@{$ENDIF}Vac, 'vac');
RegisterMonadicFunction(CeilFunction, {$IFDEF FPC}@{$ENDIF}_Ceil, 'ceil');
RegisterMonadicFunction(FloorFunction, {$IFDEF FPC}@{$ENDIF}_Floor, 'floor');
RegisterMonadicFunction(RoundFunction, {$IFDEF FPC}@{$ENDIF}_Round, 'round');
RegisterMonadicFunction(IntFunction, {$IFDEF FPC}@{$ENDIF}_Int, 'int');
RegisterMonadicFunction(FracFunction, {$IFDEF FPC}@{$ENDIF}_Frac, 'frac');
RegisterMonadicFunction(FactorialFunction, {$IFDEF FPC}@{$ENDIF}_Factorial, 'factorial');

F0('a0', BohrRadius);
F0('c', SpeedOfLight);
F0('ee', ElectronCharge);
F0('ep', 1/(4.0e-7*Pi*SpeedOfLight*SpeedOfLight)); { permittivity of a vacuum }
F0('ge', ge); { g value for electron }
F0('gg', GravitationalConstant);
F0('f', Avagadro*ElectronCharge); { Farady Constant }
F0('k', BoltzmannConstant); { Boltzmann Constant }
F0('l', Avagadro); { Avagadro's Number }
F0('r', Avagadro*BoltzmannConstant); { Gas Constant }
F0('h', Planck); { Planck constant }
F0('hc', hCross); { hcross }
F0('me', ElectronMass); { mass of electron }
F0('mn', ElectronMass/MeOverMp*MnOverMp); { mass of neutron }
F0('mp', ElectronMass/MeOverMp); { mass of proton }
F0('mu0', 4.0e-7*Pi); { permeability of a vacuum }
F0('mub', BohrMagneton); { Bohr magneton }
F0('mun', 5.05078343e-27); { nuclear magneton }
{ F0('pi', Pi); Pi - moved to Global Functions }
F0('rydberg', 10973731.568525); { Rydberg constant }
F0('u', 1.66053886e-27); { atomic mass unit }

RegisterDiadicFunction(Atan2Function, {$IFDEF FPC}@{$ENDIF}_Atan2, 'atan2');
RegisterDiadicFunction(ModFunction, {$IFDEF FPC}@{$ENDIF}_Mod, 'mod');
RegisterDiadicFunction(LorFunction, {$IFDEF FPC}@{$ENDIF}_Lor, 'lor');
RegisterDiadicFunction(GauFunction, {$IFDEF FPC}@{$ENDIF}_Gau, 'gau');

SqrtMinusOne := TFunction.Create('i',0,[opNoEvaluate]); { Not included in global functions }

finalization
  DeleteFunctions;
  SqrtMinusOne.Free;
end.

