unit LSQSpline;
{$I utils.inc}{ Copyright (C) 2007 C M Western; see UCompat.pas for license}
interface

uses MathSup;

procedure l2appr (
            const tau:    array of Double;
            const gtau:   array of Double;
            const weight: array of Double;
            const t: array of Double {1..n+k};
            const k: Integer;
            out bcoef: TDoubleArray);

procedure l2knts (
            const break: array of Double {1..1};
            const k: Integer;
            out t: TDoubleArray;
            out n: Integer);

procedure bsplpp (
            const t: array of Double;
            const bcoef: array of Double;
            const n: Integer;
            const k: Integer;
            out break: TDoubleArray;
            out coef: TDoubleArrayArray);

implementation

uses Math;

procedure bchfac (
            var w: TDoubleArrayArray {1..nrow,1..nbands};
            const nbands: Integer;
            const nrow: Integer);
label
  L9, L15, L20;
var
  i: Integer;
  imax: Integer;
  j: Integer;
  jmax: Integer;
  n: Integer;
  ratio: Double;
  diag: array of Double;

//   from  * a practical guide to splines *  by c. de boor
// onstructs cholesky factorization
//                      c  =  l * d * l-transpose
//   with l unit lower triangular and d diagonal, for given matrix c of
//   order  n r o w , in case  c  is (symmetric) positive semidefinite
//   and  b a n d e d , having  n b a n d s  diagonals at and below the
//   main diagonal.
// 
// ******  i n p u t  ******
//   nrow.....is the order of the matrix  c .
//   nbands.....indicates its bandwidth, i.e.,
//           c(i,j) = 0 for i-j .ge. nbands .
//   w.....workarray of size (nbands,nrow)  containing the  nbands  diago-
//         nals in its rows, with the main diagonal in row  1 . precisely,
//         w(i,j)  contains  c(i+j-1,j), i=1,...,nbands, j=1,...,nrow.
//           for example, the interesting entries of a seven diagonal sym-
//         metric matrix  c  of order  9  would be stored in  w  as
// 
// 
// 
// 
// 
// 
//         all other entries of  w  not identified in this way with an en-
//         try of  c  are never referenced .
//   diag.....is a work array of length  nrow .
// 
// ******  o u t p u t  ******
//   w.....contains the cholesky factorization  c = l*d*l-transp, with
//         w(1,i) containing  1/d(i,i)
//         and  w(i,j)  containing  l(i-1+j,j), i=2,...,nbands.
// 
// ******  m e t h o d  ******
//    gauss elimination, adapted to the symmetry and bandedness of  c , is
//    used .
//      near zero pivots are handled in a special way. the diagonal ele-
//   ment c(n,n) = w(1,n) is saved initially in  diag(n), all n. at the n-
//   th elimination step, the current pivot element, viz.  w(1,n), is com-
//   pared with its original value, diag(n). if, as the result of prior
//   elimination steps, this element has been reduced by about a word
//   length, (i.e., if w(1,n)+diag(n) .le. diag(n)), then the pivot is de-
//   clared to be zero, and the entire n-th row is declared to be linearly
//   dependent on the preceding rows. this has the effect of producing
//    x(n) = 0  when solving  c*x = b  for  x, regardless of  b. justific-
//   ation for this is as follows. in contemplated applications of this
//   program, the given equations are the normal equations for some least-
//   squares approximation problem, diag(n) = c(n,n) gives the norm-square
//   of the n-th basis function, and, at this point,  w(1,n)  contains the
//   norm-square of the error in the least-squares approximation to the n-
//   th basis function by linear combinations of the first n-1 . having
//   w(1,n)+diag(n) .le. diag(n) signifies that the n-th function is lin-
//   early dependent to machine accuracy on the first n-1 functions, there
//   fore can safely be left out from the basis of approximating functions
//      the solution of a linear system
//                        c*x = b
//    is effected by the succession of the following  t w o  calls:
//      call bchfac ( w, nbands, nrow, diag )       , to get factorization
//      call bchslv ( w, nbands, nrow, b )          , to solve for x.
// 


begin
  if nrow > 1 then
    goto L9;
  if w[1, 1] > 0.0 then
    w[1, 1] := 1 / w[1, 1];
  Exit;
//                                         store diagonal of  c  in  diag.
L9:
  SetLength(diag,nrow+1);
  for n := 1 to nrow do begin
    diag[n] := w[n, 1];
  end;
//                                                         factorization .
  for n := 1 to nrow do begin
    if w[n, 1] + diag[n] > diag[n] then
      goto L15;
    for j := 1 to nbands do begin
      w[n, j] := 0;
    end;
    goto L20;
  L15:
    w[n, 1] := 1 / w[n, 1];
    imax := min(nbands - 1, nrow - n);
    if imax < 1 then
      goto L20;
    jmax := imax;
    for i := 1 to imax do begin
      ratio := w[n, i + 1] * w[n, 1];
      for j := 1 to jmax do begin
        w[n + i, j] := w[n + i, j] - w[n, j + i] * ratio;
      end;
      jmax := jmax - 1;
      w[n, i + 1] := ratio;
    end;
  L20:

  end;

end;

procedure bchslv (
            const w: TDoubleArrayArray {1..nrow,1..nbands};
            const nbands: Integer;
            const nrow: Integer;
            var b: array of Double {1..nrow});
label
  L21, L30, L39, L40;
var
  j: Integer;
  jmax: Integer;
  n: Integer;
  nbndm1: Integer;

//   from  * a practical guide to splines *  by c. de boor
//   solves the linear system     c*x = b   of order  n r o w  for  x
//   provided  w  contains the cholesky factorization for the banded (sym-
//   metric) positive definite matrix  c  as constructed in the subroutine
//     b c h f a c  (quo vide).
// 
// ******  i n p u t  ******
//   nrow.....is the order of the matrix  c .
//   nbands.....indicates the bandwidth of  c .
//   w.....contains the cholesky factorization for  c , as output from
//         subroutine bchfac  (quo vide).
//   b.....the vector of length  n r o w  containing the right side.
// 
// ******  o u t p u t  ******
//   b.....the vector of length  n r o w  containing the solution.
// 
// ******  m e t h o d  ******
//   with the factorization  c = l*d*l-transpose  available, where  l  is
//   unit lower triangular and  d  is diagonal, the triangular system
//   l*y = b  is solved for  y (forward substitution), y is stored in  b,
//   the vector  d**(-1)*y is computed and stored in  b, then the triang-
//   ular system  l-transpose*x = d**(-1)*y is solved for  x (backsubstit-
//   ution).


begin
  if nrow > 1 then
    goto L21;
  b[1] := b[1] * w[1, 1];
  Exit;
// 
//                  forward substitution. solve l*y = b for y, store in b.
L21:
  nbndm1 := nbands - 1;
  for n := 1 to nrow do begin
    jmax := min(nbndm1, nrow - n);
    if jmax < 1 then
      goto L30;
    for j := 1 to jmax do begin
      b[j + n] := b[j + n] - w[n, j + 1] * b[n];
    end;
  L30:

  end;
//
//      backsubstitution. solve l-transp.x = d**(-1)*y  for x, store in b.
  n := nrow;
L39:
  b[n] := b[n] * w[n, 1];
  jmax := min(nbndm1, nrow - n);
  if jmax < 1 then
    goto L40;
  for j := 1 to jmax do begin
    b[n] := b[n] - w[n, j + 1] * b[j + n];
  end;
L40:
  n := n - 1;
  if n > 0 then
    goto L39;

end;

threadvar { TODO: Move this block into l2appr? }
  j: Integer;
  deltal: array of Double;
  deltar: array of Double;

procedure bsplvb (
            const t: array of Double;
            const jhigh: Integer;
            const index: Integer;
            const x: Double;
            const left: Integer;
            var biatx: TDoubleArray);
label
  L10, L20, L99;
var
  i: Integer;
  jp1: Integer;
  saved: Double;
  term: Double;

//   from  * a practical guide to splines *  by c. de boor
// calculates the value of all possibly nonzero b-splines at  x  of order
// 
//                jout  =  max( jhigh , (j+1)*(index-1) )
// 
//   with knot sequence  t .
// 
// ******  i n p u t  ******
//   t.....knot sequence, of length  left + jout  , assumed to be nonde-
//         creasing.  a s s u m p t i o n . . . .
//                        t(left)  .lt.  t(left + 1)   .
//    d i v i s i o n  b y  z e r o  will result if  t(left) = t(left+1)
//   jhigh,
//   index.....integers which determine the order  jout = max(jhigh,
//         (j+1)*(index-1))  of the b-splines whose values at  x  are to
//         be returned.  index  is used to avoid recalculations when seve-
//         ral columns of the triangular array of b-spline values are nee-
//         ded (e.g., in  bsplpp  or in  bsplvd ). precisely,
//                      if  index = 1 ,
//         the calculation starts from scratch and the entire triangular
//         array of b-spline values of orders 1,2,...,jhigh  is generated
//         order by order , i.e., column by column .
//                      if  index = 2 ,
//         only the b-spline values of order  j+1, j+2, ..., jout  are ge-
//         nerated, the assumption being that  biatx , j , deltal , deltar
//         are, on entry, as they were on exit at the previous call.
//            in particular, if  jhigh = 0, then  jout = j+1, i.e., just
//         the next column of b-spline values is generated.
// 
//   x.....the point at which the b-splines are to be evaluated.
//   left.....an integer chosen (usually) so that
//                   t(left) .le. x .le. t(left+1)  .
// 
// ******  o u t p u t  ******
//   biatx.....array of length  jout , with  biatx(i)  containing the val-
//         ue at  x  of the polynomial of order  jout  which agrees with
//         the b-spline  b(left-jout+i,jout,t)  on the interval (t(left),
//         t(left+1)) .
// 
// ******  m e t h o d  ******
//   the recurrence relation
//
//                        x - t(i)              t(i+j+1) - x
//      b(i,j+1)(x)  =  -----------b(i,j)(x) + ---------------b(i+1,j)(x)
//                      t(i+j)-t(i)            t(i+j+1)-t(i+1)
// 
//   is used (repeatedly) to generate the (j+1)-vector  b(left-j,j+1)(x),
//   ...,b(left,j+1)(x)  from the j-vector  b(left-j+1,j)(x),...,
//   b(left,j)(x), storing the new values in  biatx  over the old. the
//   facts that
//             b(i,1) = 1  if  t(i) .le. x .lt. t(i+1)
//   and that
//             b(i,j)(x) = 0  unless  t(i) .le. x .lt. t(i+j)
//   are used. the particular organization of the calculations follows al-
//   gorithm  (8)  in chapter x of the text.
// 



//      real biatx(jhigh),t(1),x,   deltal(20),deltar(20),saved,term
//      dimension biatx(jout), t(left+jout)
//   current fortran standard makes it impossible to specify the length of
//   t  and of  biatx  precisely without the introduction of otherwise
//   superfluous additional arguments.

begin
  case index of
    1: goto L10; 2: goto L20; 
  end;

L10:
  j := 1;
  biatx[1] := 1.0;
  if j >= jhigh then
    goto L99;

L20:
  jp1 := j + 1;
  if High(deltar) <= j then begin
    SetLength(deltar,j+1);
    SetLength(deltal,j+1);
  end;
  deltar[j] := t[left + j] - x;
  deltal[j] := x - t[left + 1 - j];
  saved := 0.0;
  for i := 1 to j do begin
    term := biatx[i] / (deltar[i] + deltal[jp1 - i]);
    biatx[i] := saved + deltar[i] * term;
    saved := deltal[jp1 - i] * term;
  end;
  biatx[jp1] := saved;
  j := jp1;
  if j < jhigh then
    goto L20;

L99:

end;

procedure l2appr (
            const tau:    array of Double;
            const gtau:   array of Double;
            const weight: array of Double;
            const t: array of Double {1..n+k};
            const k: Integer;
            out bcoef: TDoubleArray);
label
  L10, L15;
//const
//  kmax: Integer = 20;
//  ntmax: Integer = 200;
var
  i: Integer;
  ntau: Integer;
  n: Integer;
  j: Integer;
  jj: Integer;
  left: Integer;
  leftmk: Integer;
  ll: Integer;
  mm: Integer;
  biatx: TDoubleArray {[1..kmax]};
  dw: Double;
  q: TDoubleArrayArray;

//   from  * a practical guide to splines *  by c. de boor
//   to be called in main program  l 2 m a i n .
// alls subprograms  bsplvb, bchfac/slv
//
// onstructs the (weighted discrete) l2-approximation by splines of order
//   k  with knot sequence  t(1), ..., t(n+k)  to given data points
//   ( tau(i), gtau(i) ), i=1,...,ntau. the b-spline coefficients
//   b c o e f   of the approximating spline are determined from the
//   normal equations using cholesky's method.
// 
// ******  i n p u t  ******
//   t(1), ..., t(n+k)  the knot sequence
//   n.....the dimension of the space of splines of order k with knots t.
//   k.....the order
// 
//   w a r n i n g  . . .  the restriction   k .le. kmax (= 20)   is impo-
//         sed by the arbitrary dimension statement for  biatx  below, but
//         is  n o w h e r e   c h e c k e d   for.
// 
// ******  w o r k  a r r a y s  ******
//   q....a work array of size (at least) k*n. its first  k  rows are used
//        for the  k  lower diagonals of the gramian matrix  c .
//   diag.....a work array of length  n  used in bchfac .
// 
// ******  i n p u t  via  c o m m o n  /data/  ******
//   ntau.....number of data points
//   (tau(i),gtau(i)), i=1,...,ntau     are the  ntau  data points to be
//         fitted .
//   weight(i), i=1,...,ntau    are the corresponding weights .
//
// ******  o u t p u t  ******
//   bcoef(1), ..., bcoef(n)  the b-spline coeffs. of the l2-appr.
// 
// ******  m e t h o d  ******
//   the b-spline coefficients of the l2-appr. are determined as the sol-
//   ution of the normal equations
//      sum ( (b(i),b(j))*bcoef(j) : j=1,...,n)  = (b(i),g),
//                                                i = 1, ..., n .
//   here,  b(i)  denotes the i-th b-spline,  g  denotes the function to
//   be approximated, and the  i n n e r   p r o d u c t  of two funct-
//   ions  f  and  g  is given by
//       (f,g)  :=  sum ( f(tau(i))*g(tau(i))*weight(i) : i=1,...,ntau) .
//   the arrays  t a u  and  w e i g h t  are given in common block
//    d a t a , as is the array  g t a u  containing the sequence
//   g(tau(i)), i=1,...,ntau.
//   the relevant function values of the b-splines  b(i), i=1,...,n, are
//   supplied by the subprogram  b s p l v b .
//      the coeff.matrix  c , with
//            c(i,j)  :=  (b(i), b(j)), i,j=1,...,n,
//   of the normal equations is symmetric and (2*k-1)-banded, therefore
//   can be specified by giving its k bands at or below the diagonal. for
//   i=1,...,n,  we store
//    (b(i),b(j))  =  c(i,j)  in  q(i-j+1,j), j=i,...,min0(i+k-1,n)
//   and the right side
//    (b(i), g )  in  bcoef(i) .
//   since b-spline values are most efficiently generated by finding sim-
//   ultaneously the value of  e v e r y  nonzero b-spline at one point,
//   the entries of  c  (i.e., of  q ), are generated by computing, for
//   each ll, all the terms involving  tau(ll)  simultaneously and adding
//   them to all relevant entries.
// 




// 
begin
  ntau := High(tau);
  n := High(t)-k;
  SetLength(bcoef,n+1);
  SetLength(q,n+1,k+1);
  SetLength(biatx,k+1);
  for j := 1 to n do begin
    bcoef[j] := 0.0;
    for i := 1 to k do begin
      q[j, i] := 0.0;
    end;
  end;
  left := k;
  leftmk := 0;
  for ll := 1 to ntau do begin
//                    locate  l e f t  s.t. tau(ll) in (t(left),t(left+1))
  L10:
    if left = n then
      goto L15;
    if tau[ll] < t[left + 1] then
      goto L15;
    left := left + 1;
    leftmk := leftmk + 1;
    goto L10;
  L15:
    bsplvb(t, k, 1, tau[ll], left, biatx);
//         biatx(mm) contains the value of b(left-k+mm) at tau(ll).
//         hence, with  dw := biatx(mm)*weight(ll), the number dw*gtau(ll)
//         is a summand in the inner product
//            (b(left-k+mm), g)  which goes into  bcoef(left-k+mm)
//         and the number biatx(jj)*dw is a summand in the inner product
//            (b(left-k+jj), b(left-k+mm)), into  q(jj-mm+1,left-k+mm)
//         since  (left-k+jj) - (left-k+mm) + 1  =  jj - mm + 1 .
    for mm := 1 to k do begin
      dw := biatx[mm] * weight[ll];
      j := leftmk + mm;
      bcoef[j] := dw * gtau[ll] + bcoef[j];
      i := 1;
      for jj := mm to k do begin
        q[j, i] := biatx[jj] * dw + q[j, i];
        i := i + 1;
      end;
    end;
  end;
//
//              construct cholesky factorization for  c  in  q , then use
//              it to solve the normal equations
//                     c*x  =  bcoef
//              for  x , and store  x  in  bcoef .
  bchfac(q, k, n);
  bchslv(q, k, n, bcoef);

end;

procedure bsplpp (
            const t: array of Double;
            const bcoef: array of Double;
            const n: Integer;
            const k: Integer;
            out break: TDoubleArray;
            out coef: TDoubleArrayArray);
label
  L50, L9;
var
  i: Integer;
  j: Integer;
  jp1: Integer;
  kmj: Integer;
  left: Integer;
  lsofar: Integer;
  biatx: TDoubleArray;
  diff: Double;
  factor: Double;
  sum: Double;
  scrtch: TDoubleArrayArray;

//   from  * a practical guide to splines *  by c. de Boor (7 may 92)
// calls  bsplvb
// 
// converts the b-representation  t, bcoef, n, k  of some spline into its
//   pp-representation  break, coef, l, k .
// 
// ******  i n p u t  ******
//   t.....knot sequence, of length  n+k
//   bcoef.....b-spline coefficient sequence, of length  n
//   n.....length of  bcoef  and  dimension of spline space  spline(k,t)
//   k.....order of the spline
// 
//
// ******  w o r k   a r e a  ******
//   scrtch......of size  (k,k) , needed to contain bcoeffs of a piece of
//         the spline and its  k-1  derivatives
// 
// ******  o u t p u t  ******
//   break.....breakpoint sequence, of length  l+1, contains (in increas-
//         ing order) the distinct points in the sequence  t(k),...,t(n+1)
//   coef.....array of size (k,l), with  coef(i,j) = (i-1)st derivative of
//         spline at break(j) from the right
//   l.....number of polynomial pieces which make up the spline in the in-
//         terval  (t(k), t(n+1))
//
// ******  m e t h o d  ******
//      for each breakpoint interval, the  k  relevant b-coeffs of the
//   spline are found and then differenced repeatedly to get the b-coeffs
//   of all the derivatives of the spline on that interval. the spline and
//   its first  k-1  derivatives are then evaluated at the left end point
//   of that interval, using  bsplvb  repeatedly to obtain the values of
//   all b-splines of the appropriate order at that point.
// 



// 
begin
  SetLength(biatx,k+1);
  SetLength(scrtch,k+1,k+1);
  SetLength(coef,n+1,k+1); { TODO: These two could be too big }
  SetLength(break,n+1);
  lsofar := 0;
  break[1] := t[k];
  for left := k to n do begin
//                                 find the next nontrivial knot interval.
    if t[left + 1] = t[left] then
      goto L50;
    lsofar := lsofar + 1;
    break[lsofar+1] := t[left + 1];
    if k > 1 then
      goto L9;
    coef[lsofar,1] := bcoef[left];
    goto L50;
//         store the k b-spline coeff.s relevant to current knot interval
//                              in  scrtch(.,1) .
  L9:
    for i := 1 to k do begin
      scrtch[1,i] := bcoef[left - k + i];
    end;
// 
//         for j=1,...,k-1, compute the  k-j  b-spline coeff.s relevant to
//         current knot interval for the j-th derivative by differencing
//         those for the (j-1)st derivative, and store in scrtch(.,j+1) .
    for jp1 := 2 to k do begin
      j := jp1 - 1;
      kmj := k - j;
      for i := 1 to kmj do begin
        diff := t[left + i] - t[left + i - kmj];
        if diff > 0.0 then
          scrtch[jp1,i] := (scrtch[j, i + 1] - scrtch[j, i]) / diff;

      end;
    end;
//
//         for  j = 0, ..., k-1, find the values at  t(left)  of the  j+1
//         b-splines of order  j+1  whose support contains the current
//         knot interval from those of order  j  (in  biatx ), then comb-
//         ine with the b-spline coeff.s (in scrtch(.,k-j) ) found earlier
//         to compute the (k-j-1)st derivative at  t(left)  of the given
//         spline.
//            note. if the repeated calls to  bsplvb  are thought to gene-
//         rate too much overhead, then replace the first call by
//            biatx(1) = 1.
//         and the subsequent call by the statement
//            j = jp1 - 1
//         followed by a direct copy of the lines
//            deltar(j) = t(left+j) - x
//                   ......
//            biatx(j+1) = saved
//         from  bsplvb . deltal(kmax)  and  deltar(kmax)  would have to
//         appear in a dimension statement, of course.
//
    bsplvb(t,1,1,t[left],left,biatx);
    coef[lsofar,k] := scrtch[k, 1];
    for jp1 := 2 to k do begin
      bsplvb(t,jp1,2,t[left],left,biatx);
      kmj := k + 1 - jp1;
      sum := 0;
      for i := 1 to jp1 do begin
        sum := biatx[i] * scrtch[kmj, i] + sum;
      end;
      coef[lsofar,kmj] := sum;
    end;
  L50:

  end;
  factor := 1;
  for i := 2 to k do begin
    factor := factor * (k + 1 - i);
    for j := 1 to lsofar do begin
      coef[j,i] := coef[j, i] * factor;
    end;
  end;
  SetLength(coef, lsofar+1, k+1);
  SetLength(break, lsofar+2);
end;

procedure l2knts (
            const break: array of Double {1..1};
            const k: Integer;
            out t: TDoubleArray;
            out n: Integer);
var
  i: Integer;
  km1: Integer;
  l: Integer;

//   from  * a practical guide to splines *  by c. de boor
//   to be called in main program  l 2 m a i n .
// onverts the breakpoint sequence  b r e a k   into a corresponding knot
//   sequence  t  to allow the repr. of a pp function of order  k  with
//   k-2 continuous derivatives as a spline of order  k  with knot
//   sequence  t . this means that
//   t(1), ..., t(n+k) =  break(1) k times, then break(i), i=2,...,l, each
//                        once, then break(l+1) k times .
//   therefore,  n = k-1 + l.
//
// ******  i n p u t  ******
//   k     the order
//   l     the number of polynomial pieces
//   break(1), ...,break(l+1)  the breakpoint sequence
//
// ******  o u t p u t  ******
//   t(1),...,t(n+k)   the knot sequence
//   n     the dimension of the corresp. spline space of order  k .
//


//      dimension break(l+1),t(n+k)
begin
  l := High(break)-1;
  km1 := k - 1;
  n := km1 + l;
  SetLength(t,n+k+1);
  for i := 1 to km1 do begin
    t[i] := break[1];
  end;
  for i := 1 to l do begin
    t[km1 + i] := break[i];
  end;
  for i := 1 to k do begin
    t[n + i] := break[l + 1];
  end;

end;

initialization
finalization
  SetLength(deltal,0); { TODO: Should be local variables, so this is not needed }
  SetLength(deltar,0);
end.
