unit LapackLSQ;
{$I utils.inc}{ Copyright (C) 2010 C M Western; see UCompat.pas for license}
interface

uses MatrixVec, MathSup;

{ Interface to LAPACK least squares routines }

function dgels(
{ Computes the least squares solution to an over-determined system
  of linear equations, A X=B or A**H X=B,  or the minimum norm
  solution of an under-determined system, where A is a general
  rectangular matrix of full rank,  using a QR or LQ factorization
  of A. }
  A: TMatrix; { On entry, the M-by-N matrix A.
                On exit,
                  if M >= N, A is overwritten by details of its QR
                             factorization as returned by DGEQRF;
                  if M <  N, A is overwritten by details of its LQ
                             factorization as returned by DGELQF. }
  B :TColumnVector; { Can also be TMatrix;
                      On entry, the matrix B of right hand side vectors, stored
                      columnwise; B is M-by-NRHS if TRANS = 'N', or N-by-NRHS if TRANS = 'T'.
                      On exit, if INFO = 0, B is overwritten by the solution vectors, stored columnwise:
                      if TRANS = 'N' and m >= n, rows 1 to n of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements N+1 to M in that column;
                      if TRANS = 'N' and m < n, rows 1 to N of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m >= n, rows 1 to M of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m < n, rows 1 to M of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements M+1 to N in that column. }
  Trans: AnsiChar = 'N' { = 'N': the linear system involves A;
                          = 'T': the linear system involves A**T. }
  ): Integer;
{ Result: = 0:  successful exit
          < 0:  if INFO = -i, the i-th argument had an illegal value
          > 0:  if INFO =  i, the i-th diagonal element of the
                 triangular factor of A is zero, so that A does not have
                 full rank; the least squares solution could not be computed. }

function dgelsd(
{ Computes the least squares solution to an over-determined system
  of linear equations, A X=B or A**H X=B,  or the minimum norm
  solution of an under-determined system, using a divide and conquer
  method, where A is a general rectangular matrix of full rank,
  using a QR or LQ factorization of A.}
  A: TMatrix; { On entry, the M-by-N matrix A. On exit, A has been destroyed. }
  B :TColumnVector  { Can also be TMatrix;
                      On entry, the M-by-NRHS right hand side matrix B.
                      On exit, B is overwritten by the N-by-NRHS solution matrix X.
                      If m >= n and RANK = n, the residual sum-of-squares for the
                      solution in the i-th column is given by the sum of squares of elements n+1:m in that column. }
): Integer;
{ Result : = 0:  successful exit
           < 0:  if INFO = -i, the i-th argument had an illegal value.
           > 0:  the algorithm for computing the SVD failed to converge;
                 if INFO = i, i off-diagonal elements of an intermediate
                 bidiagonal form did not converge to zero. }

function dtrtrs(uplo, trans, diag: AnsiChar; A: TMatrix; B: TColumnVector):Integer;
{ Solves a triangular system of linear equations AX=B, A**T X=B or A**H X=B }

procedure PolyFit(const x, y: array of Double; var ThePoly: array of Double);
function MaxPolyErr(const x, y, ThePoly: array of Double):Double;

implementation

uses SysUtils, Math, Lapack;

function dtrtrs(uplo, trans, diag: AnsiChar; A: TMatrix; B: TColumnVector):Integer;
var
  N, Nrhs, LDA: Integer;
begin
  Nrhs := 1;
  N := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    N := TMatrix(B).RowCount;
  end;
  LDA := A.RowCount;
  dtrtrs_(uplo, trans, diag, N, Nrhs, A.FirstItem^, LDA, B.FirstItem^, N, Result, 1, 1, 1);
end;

function dgels(
  A: TMatrix; { On entry, the M-by-N matrix A.
                On exit,
                  if M >= N, A is overwritten by details of its QR
                             factorization as returned by DGEQRF;
                  if M <  N, A is overwritten by details of its LQ
                             factorization as returned by DGELQF. }
  B :TColumnVector; { Can also be TMatrix;
                      On entry, the matrix B of right hand side vectors, stored
                      columnwise; B is M-by-NRHS if TRANS = 'N', or N-by-NRHS if TRANS = 'T'.
                      On exit, if INFO = 0, B is overwritten by the solution vectors, stored columnwise:
                      if TRANS = 'N' and m >= n, rows 1 to n of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements N+1 to M in that column;
                      if TRANS = 'N' and m < n, rows 1 to N of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m >= n, rows 1 to M of B contain the minimum norm solution vectors;
                      if TRANS = 'T' and m < n, rows 1 to M of B contain the least squares solution vectors;
                      the residual sum of squares for the solution in each column is given by the sum of squares of elements M+1 to N in that column. }
  Trans: AnsiChar = 'N' { = 'N': the linear system involves A;
                          = 'T': the linear system involves A**T. }
  ): Integer;
var
  M, N, Nrhs, LDB: Integer;
  AWork: Double;
  Work: array of Double;
  Lwork: Integer;
begin
  M := A.RowCount;
  N := A.ColCount;
  Nrhs := 1;
  LDB := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    LDB := TMatrix(B).RowCount;
  end;
  Lwork := -1;
  dgels_(Trans, M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, AWork, Lwork, Result, 1);
  if Result = 0 then begin
    Lwork := round(AWork);
    SetLength(Work, Lwork);
    dgels_(Trans, M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, Work[0], Lwork, Result, 1);
  end;
end;

function dgelsd(
  A: TMatrix; { On entry, the M-by-N matrix A. On exit, A has been destroyed. }
  B :TColumnVector  { Can also be TMatrix;
                      On entry, the M-by-NRHS right hand side matrix B.
                      On exit, B is overwritten by the N-by-NRHS solution matrix X.
                      If m >= n and RANK = n, the residual sum-of-squares for the
                      solution in the i-th column is given by the sum of squares of elements n+1:m in that column. }
): Integer;
{ Result : = 0:  successful exit
           < 0:  if INFO = -i, the i-th argument had an illegal value.
           > 0:  the algorithm for computing the SVD failed to converge;
                 if INFO = i, i off-diagonal elements of an intermediate
                bidiagonal form did not converge to zero. }
var
  M, N, Nrhs, LDB: Integer;
  AWork, RCOND: Double;
  Work, S: array of Double;
  Lwork, Rank: Integer;
  Iwork: array of Integer;
begin
  M := A.RowCount;
  N := A.ColCount;
  Nrhs := 1;
  LDB := B.Size;
  if B is TMatrix then begin
    NRhs := TMatrix(B).ColCount;
    LDB := TMatrix(B).RowCount;
  end;
  Lwork := -1;
  SetLength(S, Min(M,N));
  RCOND := 0;
  SetLength(Iwork, Min(N,M)*(11+3*100)+100); { TODO: This is ony a guess }
  dgelsd_(M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, S[0], RCOND, Rank, AWork, Lwork, Iwork[0], Result);
  if Result = 0 then begin
    Lwork := round(AWork);
    SetLength(Work, Lwork);
    dgelsd_(M, N, Nrhs, A.FirstItem^, M, B.FirstItem^, LDB, S[0], RCOND, Rank, Work[0], Lwork, Iwork[0], Result);
  end;
end;

procedure PolyFit(const x, y: array of Double; var ThePoly: array of Double);
var
  i, j: Integer;
  B: TColumnVector;
  A: TMatrix;
begin
  B := TColumnVector.Create(Length(x));
  A := TMatrix.Create(Length(x), Length(ThePoly));
  try
    for i := 1 to Length(x) do begin
      A[i,1] := 1;
      B[i] := y[i-1];
      for j := 2 to Length(ThePoly) do
        A[i,j] := A[i,j-1] * x[i-1];
    end;
    i := dgels(A,B);
    if i <> 0 then
      raise Exception.Create('PolyFit:'+IntToStr(i));
    for i := 0 to High(ThePoly) do
      ThePoly[i] := B[i+1];
  finally
    A.Free;
    B.Free;
  end;
end;

function MaxPolyErr(const x, y, ThePoly: array of Double):Double;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to High(x) do
    Result := Max(Result, Abs(y[i]-Poly(x[i],ThePoly)));
end;

end.

