unit MatrixVec; { Unit Matrix is part of FPC run time library }
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes, UCompat, Vector, MathSup;

type

  TShowName = function (Index: Integer): string of object;

  TMatrix = class;

  TColumnVector = class(TVector) { As TVector, but first index is 1 }
  public
    constructor Create; overload;
    constructor Create(ASize: SizeInt); overload;
    constructor Create(AVector: TColumnVector); overload;
    procedure PreMultiply(A: TMatrix; Transpose: Boolean = False);
    procedure CrossProduct(V: TColumnVector);
  end;

  { TMatrix }

  TMatrix = class (TColumnVector) { Dimensions 1..RowCount, 1..ColCount }
  private
    FRowCount: Longint;
    FColCount: Longint;
    function GetValid(ARow, ACol:Longint):Boolean;
    procedure SetValid(ARow, ACol:Longint; IsValid:Boolean);
    function GetValue(ARow, ACol:Longint):Double;
    procedure SetValue(ARow, ACol:Longint; v:Double); inline;
    function GetPointer(ARow, ACol:Longint):PDouble; inline;
    procedure SetRowCount(n: Longint); overload;
    procedure SetColCount(n: Longint); overload;
    procedure SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
  public
    constructor Create(Rows, Cols: Longint); overload;
    constructor Create(AMatrix: TMatrix); overload;
    procedure SetColCountAndClear(n: Longint; DoClear: Boolean); overload;
    procedure SetRowCountAndClear(n: Longint; DoClear: Boolean); overload;
    procedure SwapColumns(Col1, Col2: Longint);
    procedure MoveColumn(Src, Dst: Integer);
    procedure MoveRow(Src, Dst: Integer);
    function Index(ARow, ACol: Longint): Integer; inline;
    function GetCell(ARow, ACol:Longint; out v:Double):Boolean;
    function DefaultLabel(AIndex: Integer): string;
    procedure Resize(Rows, Cols: Integer; DoClear:Boolean); { Resize preserving content; add NaN's if Clear }
    procedure ForceReal; virtual;

    procedure Assign(Source: TVector); override;
    procedure AssignAndNil(var Source: TVector); virtual;
    procedure CopyTo(var A: TDoubleArrayArray);

    procedure PostMultiply(M: TMatrix);
    function Transpose: TMatrix;
    procedure TransposeInPlace;
    procedure AddDiagonal(v: Double);
    procedure Write(var F: Text; const Title: string; RowLabel: TShowName = nil; ColLabel: TShowName = nil; Width: Integer = 10; Fixed: Integer = -1); virtual;
    function CheckOrthogonal(var F: Text; const Title: string; LeftTranspose: Boolean; Threshold: Double; Print: Boolean = True; RowLabel: TShowName = nil; ColLabel: TShowName = nil):Double;
    function CheckUnit(var F: Text; Title: string; Threshold: Double; Print: Boolean = True; RowLabel: TShowName = nil; ColLabel: TShowName = nil):Double;
    procedure MatrixError(ARow, ACol: LongInt);

    property Item[ARow, ACol: Longint]:PDouble read GetPointer;
    property RowCount: Longint read FRowCount write SetRowCount;
    property ColCount: Longint read FColCount write SetColCount;
    property Valid[ARow, ACol:Longint]:Boolean read GetValid write SetValid;
    property Value[ARow, ACol:Longint]:Double read GetValue write SetValue; default;
  end;

implementation

uses Math, SysUtils, Lapack;

constructor TMatrix.Create(Rows, Cols: Longint);
begin
  FRowCount := Rows;
  FColCount := Cols;
  inherited Create(SizeInt(Rows)*Cols);
end;

function TMatrix.Index(ARow, ACol: Longint): Integer; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) then
    MatrixError(ARow, ACol);
  {$ENDIF}
  Result := SizeInt(ACol-1)*RowCount+ARow;
end;

procedure TMatrix.MoveColumn(Src, Dst: Integer);
var
  w: array of Double;
begin
  if Src = Dst then
    Exit;
  SetLength(w,RowCount);
  System.Move(Item[1,Src]^,w[0],RowCount*SizeOf(Double));
  if Src < Dst then
    { Move columns down }
    System.Move(Item[1,Src+1]^,Item[1,Src]^,SizeInt(Dst-Src)*RowCount*SizeOf(Double))
  else
    { Move columns up }
    System.Move(Item[1,Dst]^,Item[1,Dst+1]^,SizeInt(Src-Dst)*RowCount*SizeOf(Double));
  System.Move(w[0],Item[1,Dst]^,RowCount*SizeOf(Double));
end;

procedure TMatrix.MoveRow(Src, Dst: Integer);
var
  c: Integer;
  w: Double;
  r: Integer;
begin
  if Src = Dst then
    Exit;
  for c := 1 to ColCount do begin
    w := Value[Src,c];
    if Src < Dst then
      for r := Src+1 to Dst do
        Value[r,c] := Value[r-1,c]
    else
      for r := Dst to Src - 1 do
        Value[r+1,c] := Value[r,c];
    Value[Dst,c] := w;
  end;
end;

procedure TMatrix.CopyTo(var A: TDoubleArrayArray);
var
  i, j: Integer;
begin
  SetLength(A, RowCount);
  for i := 0 to RowCount - 1 do begin
    SetLength(A[i],ColCount);
    for j := 0 to ColCount - 1 do
      A[i][j] := Value[i+1,j+1];
  end;
end;

constructor TMatrix.Create(AMatrix: TMatrix);
begin
  Create(0,0);
  Assign(AMatrix);
end;

function TMatrix.DefaultLabel(AIndex: Integer): string;
begin
  Result := IntToStr(AIndex);
end;

procedure TMatrix.ForceReal;
begin
  { No action here .... }
end;

function TMatrix.GetCell(ARow, ACol:Longint; out v:Double):Boolean;
var
  p: PDouble;
begin
  p := Item[ARow, ACol];
  Result := not IsNanOrInf(p^);
  System.Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
end;

function TMatrix.GetPointer(ARow, ACol:Longint):PDouble; {$IFDEF FPCINLINE} inline; {$ENDIF}
begin
  Result := inherited Item[Index(ARow, ACol)];
end;

procedure TMatrix.SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
var
  p: PDouble;
begin
  p := GetPointer(ARow, ACol);
  { Note: To Set a specific invalid value, set IsValid := True
    and v to the required value }
  if IsValid then
    p^ := v
    // May need to use: System.Move(v, p^, SizeOf(Double))
    // if compiler generates floating point load/store
  else
    SetNan(p^);
end;

function TMatrix.GetValid(ARow, ACol:Longint):Boolean;
var
  v: Double;
begin
  Result := GetCell(ARow, ACol, v);
end;

procedure TMatrix.SetValid(ARow, ACol:Longint; IsValid:Boolean);
begin
  SetCell(ARow, ACol, 0, IsValid);
end;

function TMatrix.GetValue(ARow, ACol:Longint):Double;
begin
  {$IFOPT R+}
  if (ARow <= 0) or (ACol <= 0) or (ARow > RowCount) then
    MatrixError(ARow, ACol);
  {$ENDIF}
  Result := PDouble(inherited Item[SizeInt(ACol-1)*RowCount+ARow])^;
end;

procedure TMatrix.SetValue(ARow, ACol:Longint; v:Double); {$IFDEF FPCINLINE} inline; {$ENDIF}
begin;
  SetCell(ARow, ACol, v, True);
end;

procedure TMatrix.SetRowCountAndClear(n: Longint; DoClear: Boolean);
var
  c, OldRows: Longint;
begin
  { Also preserves current content }
  if n <> RowCount then begin
    OldRows := RowCount;
    FRowCount := n;
    if (RowCount < OldRows) and (RowCount > 0) then
      for c := 2 to ColCount do
        System.Move(inherited Item[SizeInt(c-1)*OldRows+1]^, inherited Item[SizeInt(c-1)*RowCount+1]^, OldRows*Sizeof(Double));
    SetSizeAndClear(ColCount * RowCount, DoClear);
    if RowCount > OldRows then begin
      { Matrix is Expanding - Shift Each row into position }
      for c := ColCount downto 2 do
        System.Move(inherited Item[SizeInt(c-1)*OldRows+1]^, Item[1,c]^, OldRows*Sizeof(Double));
      { Clear resulting gaps }
      for c := 1 to ColCount do
        ClearRange(Index(OldRows+1, c), RowCount-OldRows);
    end;
  end;
end;

procedure TMatrix.SetRowCount(n: Longint);
begin
  SetRowCountAndClear(n, False);  { This way current content is preserved on shrinking }
end;

procedure TMatrix.SetColCount(n: Longint);
begin
  if n <> ColCount then begin
    FColCount := n;
    Size := ColCount * RowCount;
  end;
end;

procedure TMatrix.SetColCountAndClear(n: Longint; DoClear:Boolean);
begin
  SetSizeAndClear(SizeInt(n) * RowCount, DoClear);
  FColCount := n;
end;

procedure TMatrix.AddDiagonal(v: Double);
var
  p: PDouble;
  I: Integer;
begin
  p := FirstItem;
  for I := 1 to Min(RowCount,ColCount) do begin
    P^ := P^ + v;
    Inc(p,(RowCount+1));
  end;
end;

procedure TMatrix.Assign(Source: TVector);
var
  s: TMatrix;
begin
  if Source is TMatrix then begin
    s := TMatrix(Source);
    FRowCount := s.RowCount;
    FColCount := s.ColCount;
    Size := SizeInt(RowCount)*ColCount;
    System.Move(s.FirstItem^, FirstItem^, SizeInt(RowCount)*ColCount*Sizeof(Double));
  end else
    inherited Assign(Source);
end;

procedure TMatrix.AssignAndNil(var Source: TVector);
begin
  if Source is TMatrix then begin
    FRowCount := TMatrix(Source).RowCount;
    FColCount := TMatrix(Source).ColCount;
  end;
  inherited AssignAndNil(Source);
end;

procedure TMatrix.SwapColumns(Col1, Col2: Longint);
var
  i: Longint;
  v: Double;
begin
  if Col1 <> Col2 then begin
    {TODO: Speed up}
    for i := 1 to RowCount do begin
      System.Move(Item[i, Col1]^, v, sizeof(Double));
      System.Move(Item[i, Col2]^, Item[i, Col1]^, sizeof(Double));
      System.Move(v, Item[i, Col2]^, sizeof(Double));
    end;
  end;
end;

function TMatrix.Transpose: TMatrix;
var
  i, j: Integer;
begin
  Result := TMatrix.Create(ColCount,RowCount);
  for i := 1 to RowCount do
    for j := 1 to ColCount do
      Result.Value[j,i] := Value[i,j];
end;

procedure TMatrix.TransposeInPlace;
var
  M: TVector;
begin
  { Potentially inefficient as making a copy of the entire matrix }
  M := Transpose;
  AssignAndNil(M);
end;

procedure TMatrix.Write(var F: Text; const Title: string; RowLabel: TShowName; ColLabel: TShowName; Width: Integer; Fixed: Integer);
var
  r, c, lt: Integer;
  s: string;
begin
  if not Assigned(ColLabel) then
    ColLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
  if not Assigned(RowLabel) then
    RowLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
  lt := Length(Title);
  for r := 1 to RowCount do
    lt := Max(lt, Length(RowLabel(r)));
  System.Write(F, Title:lt);
  if Fixed < 0 then
    System.Write(F, '  ')
  else
    System.Write(F, ' ':Width-Fixed-1);
  for c := 1 to ColCount do begin
    s := ColLabel(c);
    System.Write(F, s, ' ':Width-Length(s));
  end;
  WriteLn(F);
  for r := 1 to RowCount do begin
    System.Write(F, RowLabel(r):lt, ' ');
    for c := 1 to ColCount do
      System.Write(F, DTosMA(Value[r,c],Width,Fixed));
    WriteLn(F)
  end;
end;

function TMatrix.CheckOrthogonal(var F: Text; const Title: string; LeftTranspose: Boolean; Threshold: Double; Print: Boolean; RowLabel: TShowName; ColLabel: TShowName):Double;
{ Print Matrix * Matrix^T if maxium error above Threshold }
var
  M, T: TMatrix;
begin
  if LeftTranspose then begin
    M := Transpose;
    M.PostMultiply(Self);
  end else begin
    M := TMatrix.Create(Self);
    T := Transpose;
    M.PostMultiply(T);
    T.Free;
  end;
  Result := M.CheckUnit(F, Title+' Orthogonality Check', Threshold, Print, RowLabel, ColLabel);
  M.Free;
end;

function TMatrix.CheckUnit(var F: Text; Title: string; Threshold: Double; Print: Boolean; RowLabel: TShowName; ColLabel: TShowName):Double;
var
  i, j, im, jm: Integer;
  V: Double;
begin
  Result := 0;
  im := 1;
  jm := 1;
  for i := 1 to RowCount do
    for j := 1 to ColCount do begin
      if i = j then
        V := Abs(1-Value[i,i])
      else
        V := Abs(Value[i,j]);
      if V > Result then begin
        Result := V;
        im := i;
        jm := j;
      end;
    end;
  if not Print then
    Exit;
  if not Assigned(ColLabel) then
    ColLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
  if not Assigned(RowLabel) then
    RowLabel := {$IFDEF FPC}@{$ENDIF}DefaultLabel;
  Title := Title + ': Maxiumum Error = '+FloatToStr(Result)+' at '+RowLabel(im)+', '+ColLabel(jm);
  WriteLn(F, Title);
  if Result > Threshold*RowCount then
    Write(F, '  ', RowLabel, ColLabel);
end;

procedure TMatrix.MatrixError(ARow, ACol: LongInt);
begin
  raise Exception.CreateFmt('%s[%d,%d]: Bound check failure (%d)', [ClassName, ARow, ACol, RowCount]);
end;

{$IFNDEF DESIGNPACKAGE}
{$DEFINE BLAS}
{$ENDIF}
{ Note that ATLAS seems to be slower than blas in this application }

procedure TMatrix.PostMultiply(M: TMatrix);
{ Overwrite matrix with matrix * M }
var
  W: array of Double;
{$IFDEF BLAS}
  mm, k, N: Integer;
  alpha, beta: Double;
  A: PDouble;
  No: AnsiChar;
{$ELSE}
  WM: TMatrix;
  Sum: Double;
  i, ii, iii: Integer;
  PW, PWM, PM, PE: PDouble;
{$ENDIF}
begin
  if ColCount <> M.RowCount then
    raise Exception.Create('TMatrix.PostMultiply:Size mismatch');
  {$IFDEF BLAS}
  if Size = 0 then
    Exit;
  alpha := 1;
  beta := 0;
  No := 'N';
  mm := RowCount;
  k := ColCount;
  N  := M.ColCount;
  if ColCount = N then begin
    { No change in size }
    Size := Size * 2; { Allocate workspace }
    A := PDouble(PAnsiChar(FirstItem) + (Size div 2)*SizeOf(Double));
    System.Move(FirstItem^, A^, (Size div 2)*SizeOf(Double));
    dgemm_(No, No, mm, N, N, alpha, A^, mm, M.FirstItem^, N, beta, FirstItem^, mm, 1, 1);
    Size := Size div 2;
  end else begin
    SetLength(W,Size);
    System.Move(FirstItem^, w[0], Size*SizeOf(Double));
    ColCount := N;
    dgemm_(No, No, mm, N, k, alpha, W[0], mm, M.FirstItem^, k, beta, FirstItem^, mm, 1, 1);
  end;
  {$ELSE}
  if ColCount <> M.ColCount then
    raise Exception.Create('TMatrix.PostMultiply:Size change');
  {$DEFINE ALT2}
  {$IFDEF ALT2}
  { Closest to blas code, but still slower, at least on g77/FPC }
  WM := TMatrix.Create(RowCount, ColCount);
  WM.Assign(Self);
  Zero;
  PM := M.FirstItem;
  for ii := 1 to ColCount do begin
    for iii := 1 to ColCount do begin
      {$IFDEF SIMPLE}
      Sum := M[iii,ii];
      {$ELSE}
      Sum := PM^;
      Inc(PM);
      {$ENDIF}
      if Sum <> 0 then begin
        {$IFDEF SIMPLE}
        for i := 1 to RowCount do
          Value[i,ii] := Value[i,ii] + Sum * WM[i,iii];
        {$ELSE}
        PW := Item[1,ii];
        PWM := WM.Item[1,iii];
        PE := PDouble(PAnsiChar(PWM) + RowCount*SizeOf(Double));
        while Cardinal(PWM) < Cardinal(PE) do begin
          PW^ := PW^ + Sum*PWM^;
          Inc(PW); Inc(PWM);
        end;
        {$ENDIF}
      end;
    end;
  end;
  WM.Free;
  {$ELSE}
  {$IFDEF ALT}
  SetLength(W, ColCount);
  for i := 1 to RowCount do begin
    for ii := 1 to ColCount do
      W[ii-1] := Value[i,ii];
    for ii := 1 to ColCount do begin
      Sum := 0;
      PW := @W[0];
      PM := M.Item[1,ii];
      for iii := 1 to ColCount do begin
        {$IFDEF SIMPLE}
        Sum := Sum + W[iii-1] * M[iii,ii];
        {$ELSE}
        Sum := Sum + PW^*PM^;
        Inc(PW); Inc(PM);
        {$ENDIF}
      end;
      Value[i,ii] := Sum;
    end;
  end;
  {$ELSE}
  SetLength(W, ColCount+1);
  for i := 1 to RowCount do begin
    for ii := 1 to ColCount do begin
      Sum := 0;
      for iii := 1 to ColCount do
        Sum := Sum + Value[i, iii] * M[iii,ii];
      W[ii] := Sum;
    end;
    for ii := 1 to ColCount do
      Value[i,ii] := W[ii];
  end;
  {$ENDIF}
  {$ENDIF}
  {$ENDIF}
end;

procedure TMatrix.Resize(Rows, Cols: Integer; DoClear: Boolean);
begin
  SetRowCountAndClear(Rows, DoClear);
  SetColCountAndClear(Cols, DoClear);
end;

{ TColumnVector }

constructor TColumnVector.Create(ASize: SizeInt);
begin
  inherited Create(ASize);
  Offset := 1;
end;

constructor TColumnVector.Create(AVector: TColumnVector);
begin
  Create(0);
  Assign(AVector);
end;

constructor TColumnVector.Create;
begin
  Create(0);
end;

procedure TColumnVector.CrossProduct(V: TColumnVector);
var
  I: Integer;
  W: array [0..2] of Double;
begin
  if (Size <> 3) or (V.Size <> 3) then
    raise Exception.Create('TColumnVector.CrossProduct');
  System.Move(FirstItem^,W[0],Size*SizeOf(Double));
  for i := 0 to 2 do
    Value[i+1] := W[(i+1) mod 3] * V[(i+2) mod 3 + 1] - W[(i+2) mod 3] * V[(i+1) mod 3 + 1]
end;

procedure TColumnVector.PreMultiply(A: TMatrix; Transpose: Boolean);
var
  W: array of Double;
  i, j: Integer;
  Sum: Double;
begin
  SetLength(W, Size);
  if Size > 0 then
    System.Move(FirstItem^,W[0],Size*SizeOf(Double));
  if Transpose then begin
    if A.RowCount <> Size then
      raise Exception.Create('TColumnVector.PreMultiply');
    Size := A.ColCount;
  end else begin
    if A.ColCount <> Size then
      raise Exception.Create('TColumnVector.PreMultiply - Size mismatch');
    Size := A.RowCount;
  end;
  for i := 1 to Size do begin
    Sum := 0;
    if Transpose then
      for j := 1 to Length(W) do
        Sum := Sum + A[j,i]*W[j-1]
    else
      for j := 1 to Length(W) do
        Sum := Sum + A[i,j]*W[j-1];
    Value[i] := Sum;
  end;
end;

end.
