unit PiecewisePoly;
{$I utils.inc}{ Copyright (C) 2007 C M Western; see UCompat.pas for license}

{ Piecewise Polynomial routines, based on those in:
  "A practical guide to splines by C. de Boor"
}

interface

uses MathSup;

type
  TPiecewisePolynomial = class
    knots: TDoubleArray;     { Boundaries of each piece }
    coef: TDoubleArrayArray; { Coefficients of polynomial;
                               coef[i] gives the array of coefficents at knot i;
                               order of polynomial is given by Length(coef[i]); }
    function Value(x:Double; Derivative: Integer = 0):Double;
    function FindKnot(x: Double): Integer; overload;
    function FindKnot(x: Double; Start: Integer): Integer; overload;
  end;

implementation

{ TPiecewisePolynomial }

threadvar
  xLast: Integer;

function TPiecewisePolynomial.FindKnot(x: Double): Integer;
begin
  { As an optimization, remember last starting point for this thread }
  { First check for sanity, though }
  if xLast < 0 then
    xLast := 0
  else if xLast > High(knots) then
    xLast := High(knots);
  Result := FindKnot(x,xLast);
  xLast := Result;
end;

function TPiecewisePolynomial.FindKnot(x: Double; Start: Integer): Integer;
var
  iStep, mid: Integer;
begin
  if Start = High(knots) then begin
    Result := Start;
    Dec(Start);
  end else
    Result := Start + 1;
  { Initial trial interval: Start .. Result }
  if x > knots[Start] then begin
    if (x <= knots[Result]) or (Result = High(knots)) then begin
      Result := Start;
      Exit;
    end;
    { Above interval; increase until included }
    iStep := 1;
    repeat
      Result := Result + iStep;
      if Result >= High(Knots) then begin
        Result := High(Knots);
        Break;
      end;
      Inc(iStep,iStep);
    until x <= knots[Result];
  end else begin
    { Below interval; decrease lower limit until included }
    iStep := -1;
    repeat
      Start := Start + iStep;
      if Start <= 0 then begin
        Start := 0;
        Break;
      end;
      Inc(iStep,iStep);
    until x > knots[Result];
  end;
  { Now within interval }
  while Result - Start > 1 do begin
    mid := (Result + Start) div 2;
    if x <= knots[mid] then
      Result := mid
    else
      Start := mid;
  end;
  Result := Start;
end;

function TPiecewisePolynomial.Value(x: Double; Derivative: Integer): Double;
var
  knot: TDoubleArray;
  i, m, f: Integer;
  { Based on ppvalu }
begin
  Result := 0;
  i := FindKnot(x);
  x := x - knots[i];
  knot := coef[i];
  m := High(knot);
  f := m - Derivative + 1;
  while f > 0 do begin
    Result := Result / f * x + knot[m];
    Dec(f);
    Dec(m);
  end;
end;

end.
