unit SplineControl;
{$I utils.inc}{ Copyright (C) 2007 C M Western; see UCompat.pas for license}
interface

uses MatrixVec, PiecewisePoly, MathSup;

function MakeSmoothSpline(xv, yv, wv: TColumnVector; out ActualError: Double; ExpectedError: Double):TPiecewisePolynomial; overload;
function MakeSmoothSpline(x, y, w: TDoubleArray; out ActualError: Double; ExpectedError: Double):TPiecewisePolynomial; overload;

implementation

uses SmoothSpline, LSQSpline, ShellSort;

type
  TDataSorter = class (TShellSorter)
    x, y, w: TDoubleArray;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

{ TODO: Convert above routines so that offset by one is not needed }
function MakeSmoothSpline(x, y, w: TDoubleArray; out ActualError: Double; ExpectedError: Double):TPiecewisePolynomial;
var
  Sorter: TDataSorter;
  a: TDoubleArrayArray;
  i, j, nObs: Integer;

  procedure DoLSQSpline;
  var
    nKnots, i, j, n: Integer;
    bcoef, t: TDoubleArray;
    SplineOrder: Integer;
  begin
    SplineOrder := 4;
    nKnots := Round(-ExpectedError);
    SetLength(Result.knots, nKnots+1);
    { Prepare equally spaced knots }
    for i := 1 to nKnots do
      Result.knots[i] := x[1] + (x[nObs]-x[1])*(i-1)/(nKnots-1);
    l2knts(Result.knots, SplineOrder, t, n);
    l2appr(x, y, w, t, SplineOrder, bcoef);
    bsplpp(t, bcoef, n, SplineOrder, Result.knots, Result.coef);
    { Trim matrices - silly }
    for i := 0 to nKnots-2 do begin
      Result.knots[i] := Result.knots[i+1];
      for j := 0 to SplineOrder-1 do
        Result.coef[i,j] := Result.coef[i+1,j+1];
      SetLength(Result.coef[i], SplineOrder);
    end;
    Result.knots[nKnots-1] := Result.knots[nKnots]; { coef is one point smaller }
    SetLength(Result.knots,nKnots);
    SetLength(Result.coef,nKnots-1);
  end;
begin
  nObs := High(x);
  { Make sure values are sorted }
  Sorter := TDataSorter.Create;
  try
    Sorter.x := x;
    Sorter.y := y;
    Sorter.w := w;
    Sorter.Sort(nObs);
  finally
    Sorter.Free;
  end;
  Result := TPiecewisePolynomial.Create;
  try
    if ExpectedError < 0 then begin
      { Least Squares spline }
      DoLSQSpline;
    end else begin
      { Smoothing Spline }
      ActualError := Smooth(x, y, w, ExpectedError, a);
      Result.knots := x;
      { Coefficient matrix must be transposed for use }
      SetLength(Result.coef,nObs,4);
      for i := 0 to nObs-1 do begin
        Result.knots[i] := Result.knots[i+1];
        for j := 0 to 3 do
          Result.coef[i,j] := a[j+1,i+1]
      end;
      SetLength(Result.knots,nObs);
    end;
  except
    Result.Free;
  end;
end;

function MakeSmoothSpline(xv, yv, wv: TColumnVector; out ActualError: Double; ExpectedError: Double):TPiecewisePolynomial;
var
  x, y, w: TDoubleArray;
  i, nObs: Integer;
begin
  { Copy in data }
  nObs := xv.Size;
  SetLength(x, nObs+1);
  SetLength(y, nObs+1);
  SetLength(w, nObs+1);
  for i := 1 to nObs do begin
    x[i] := xv[i];
    y[i] := yv[i];
    if Assigned(wv) then
      w[i] := wv[i]
    else
      w[i] := i;
  end;
  Result := MakeSmoothSpline(x,y,w, ActualError, ExpectedError);
end;

{ TDataSorter }

function TDataSorter.InOrder(I, J: Integer): Boolean;
begin
  Result := x[I+1] <= x[J+1];
end;

procedure TDataSorter.Swap(I, J: Integer);
begin
  SwapValues(x[I+1],x[J+1],SizeOf(Double));
  SwapValues(y[I+1],y[J+1],SizeOf(Double));
  SwapValues(w[I+1],w[J+1],SizeOf(Double));
end;

end.
