unit StackDump;
{$I utils.inc}{ Copyright (C) 2005, 2015 C M Western; see UCompat.pas for license}
interface

uses Classes, SysUtils;

function StackWatch(Enabled: Boolean):Boolean;
procedure DumpCurrent(var f: Text);

var
  FallBackDir: string;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFNDEF CONSOLE} Dialogs, Forms, {$ENDIF}
     Worker, StackTrace, UCompat;

type
  TStackWatcher = class
  {$IF CompilerVersion < 19.0}
  private
    PreviousExceptObjProc: Pointer;
    PreviousRaiseExceptionProc: Pointer;
  {$IFEND}
  public
    S: TCallStack;
    StackFilled: Boolean;
    SavedLastError: Integer;
    SavedThreadName: string;
    {$IFNDEF CONSOLE}
    LogFile: Text;
    LogFileName: ShortString;
    FileNameSet, FileIsOpen: Boolean;
    procedure OnException(Sender: TObject; E: Exception);
    {$ENDIF}
    constructor Create;
    destructor Destroy; override;
  end;

var
  StackWatcher:TStackWatcher;

{$IF CompilerVersion < 19.0}
type
  TExceptObjProc = function (P: PExceptionRecord): Exception;

function InterceptGetExceptionObject(P: PExceptionRecord): Exception;
begin
  Result := TExceptObjProc(StackWatcher.PreviousExceptObjProc)(P);
  if not StackWatcher.StackFilled then begin
    StackWatcher.SavedLastError := GetLastError;
    FillCallStack(StackWatcher.S, 0);
    StackWatcher.SavedThreadName := ThreadName;
    StackWatcher.StackFilled := True;
  end;
end;

type
  TRaiseException = procedure (dwExceptionCode, dwExceptionFlags, nNumberOfArguments: DWORD; lpArguments: PDWORD); stdcall;

procedure InterceptRaiseException(dwExceptionCode, dwExceptionFlags, nNumberOfArguments: DWORD; lpArguments: PDWORD); stdcall;
begin
  if (dwExceptionCode <> $0EEDFAE2 { cDelphiTerminate } ) and not StackWatcher.StackFilled then begin
    StackWatcher.SavedLastError := GetLastError;
    StackWatcher.StackFilled := True;
    FillCallStack(StackWatcher.S, 0);
  end;
  TRaiseException(StackWatcher.PreviousRaiseExceptionProc)(dwExceptionCode, dwExceptionFlags, nNumberOfArguments, lpArguments);
end;
{$IFEND}

{ TStackWatcher }

constructor TStackWatcher.Create;
var
  p: Integer;
begin
  {$IF CompilerVersion < 19.0}
    PreviousExceptObjProc := ExceptObjProc;
    ExceptObjProc := @InterceptGetExceptionObject;
    PreviousRaiseExceptionProc := {$IFDEF WIN64}@{$ENDIF}RaiseExceptionProc;
    RaiseExceptionProc := @InterceptRaiseException;
  {$IFEND}
  {$IFNDEF CONSOLE}
  Application.OnException := OnException;
  LogFileName := Application.ExeName;
  FileNameSet := not FileIsReadOnly(ExtractFilePath(LogFileName)); { Can't write to directory }
//  ShowMessage(Format('%s %d %s', [Application.ExeName, Ord(FileNameSet), LogFileName]));
  if FileNameSet then begin
    p := LastDelimiter('.'+ PathDelim + DriveDelim, LogFileName);
    if (p <= 0) or (LogFileName[p] <> '.') then
      p := Length(LogFileName);
    LogFileName := Copy(LogFileName, 1, p-1) + '.log';
    FileNameSet := not (FileExists(LogFileName) and FileIsReadOnly(LogFileName));
  end;
  if not FileNameSet then
    LogFileName := ExtractFileName(Application.ExeName) + '.log';
  FileIsOpen := False;
  {$ENDIF}
  StackFilled := False;
end;

destructor TStackWatcher.Destroy;
begin
  {$IFNDEF CONSOLE}
  Application.OnException := nil;
  if FileIsOpen then
    CloseFile(LogFile);
  {$ENDIF}
  {$IF CompilerVersion < 19.0}
    ExceptObjProc := PreviousExceptObjProc;
    RaiseExceptionProc := PreviousRaiseExceptionProc;
  {$IFEND}
  inherited;
end;

{$IFNDEF CONSOLE}
procedure TStackWatcher.OnException(Sender: TObject; E: Exception);
var
  Msg: string;
  Tries: Integer;
  function IO_OK:Boolean;
  begin
    { Note: Need to clear IOResult always }
    Result := (IOResult = 0) and FileIsOpen;
    if not Result then begin
      FileNameSet := False;
      FileIsOpen := False;
    end;
  end;
begin
  if DebugHook <> 0 then
    Exit; { Don't show dialog box, as already caught by debugger }
  Msg := E.Message;
  {$I-}
  if IsExpectedException(E.ClassType) then
    { Don't dump stack }
  else
  if (E is EFCreateError) or (E is EFOpenError) then begin
// Exclude this, as it leads to system error message being shown twice
//    LastError := SavedLastError;
//    if LastError <> 0 then
//      Msg := Msg + ': ' + SysErrorMessage(LastError);
    { No stack dump for simple file system errors }
  end else if {$IF CompilerVersion < 19.0} StackFilled {$ELSE} (E.StackInfo <> nil) {$IFEND} and (LogFileName <> '') then begin
    if not FileNameSet and not FileIsReadOnly(GetCurrentDir) then begin
      LogFileName := GetCurrentDir + PathDelim + LogFileName;
      FileNameSet := True;
    end;
    if FileNameSet then begin
      Tries := 0;
      while not FileIsOpen and (Tries < 2) do begin
        Inc(Tries);
        AssignFile(LogFile, LogFileName);
        if FileExists(LogFileName) then
          Append(LogFile)
        else
          Rewrite(LogFile);
        FileIsOpen := IOResult = 0;
        if not FileIsOpen and (FallBackDir <> '') then
          LogFileName := FallBackDir+PathDelim+ExtractFileName(LogFileName);
      end;

      Writeln(LogFile, '');
      Writeln(LogFile, '********* Unexpected exception detected - ' + DateTimeToStr(Now) + ' *********');
      Writeln(LogFile, VersionInfo);
      if SavedThreadName <> '' then
        WriteLn(LogFile, 'Thread:', SavedThreadName);
      Writeln(LogFile, 'Exception code: ' + E.ClassName);
      if IO_OK then
        Writeln(LogFile, 'Exception address: ' + TextualDebugInfoForAddress(Cardinal(ExceptAddr)));
      Writeln(LogFile, #13#10'Call stack (oldest call at bottom):');
      if IO_OK then
        {$IF CompilerVersion < 19.0}
          Writeln(LogFile, CallStackTextualRepresentation(S, ''));
        {$ELSE}
          Writeln(LogFile, E.StackTrace);
        {$IFEND}
      Writeln(LogFile, '*****************************************************************');
      Writeln(LogFile, '');
      Flush(LogFile);
      if IO_OK then
        Msg := Msg + ' - Logged to ' + LogFileName
      else begin
        Msg := Msg + ' - not logged to ' + LogFileName;
        LogFileName := 'stack.log';
      end;
    end else
      Msg := Msg + ' - log not saved to ' + GetCurrentDir;
    StackFilled := False;
  end else
    Msg := Msg + ' (No Stack info)';
  {$I+}
  Application.MessageBox(PChar(Msg), PChar(Application.Title), MB_OK + MB_ICONSTOP);
end;
{$ENDIF}

{$IF CompilerVersion >= 19.0}
function GetExceptionStackInfoProc(P: System.PExceptionRecord): Pointer;
var
  C: PCallStack;
begin
  New(C);
  FillCallStack(C^, 0, P.ExceptionAddress);
  Result := C;
end;

function GetStackInfoStringProc(Info: Pointer): string;
begin
  Result := CallStackTextualRepresentation(PCallStack(Info)^, '');
end;

procedure CleanUpStackInfoProc(Info: Pointer);
begin
  FreeMem(Info);
end;
{$IFEND}

function StackWatch(Enabled: Boolean):Boolean;
begin
  Result := Assigned(StackWatcher);
  if Enabled and not Assigned(StackWatcher) then begin
    StackWatcher := TStackWatcher.Create;
    {$IF CompilerVersion >= 19.0}
      Exception.GetExceptionStackInfoProc := GetExceptionStackInfoProc;
      Exception.GetStackInfoStringProc := GetStackInfoStringProc;
      Exception.CleanUpStackInfoProc := CleanUpStackInfoProc;
    {$IFEND}
  end else if not Enabled and Assigned(StackWatcher) then begin
    FreeAndNil(StackWatcher);
    {$IF CompilerVersion >= 19.0}
      Exception.GetExceptionStackInfoProc := nil;
      Exception.GetStackInfoStringProc := nil;
      Exception.CleanUpStackInfoProc := nil;
    {$IFEND}
  end;
end;

procedure DumpCurrent(var f: Text);
begin
  if not Assigned(StackWatcher) then
    Exit;
  if StackWatcher.StackFilled then begin
    WriteLn(f,CallStackTextualRepresentation(StackWatcher.S,''));
    StackWatcher.StackFilled := False;
  end;
end;

initialization
  FallBackDir := ApplicationSettingsDir;
finalization
  StackWatch(False);
end.
