unit TextStrm;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Classes;

type

  TBufferedStream = class(TStream)
  private
    IntBuffer: array of AnsiChar;
    ReadPointer: Integer;
    BufferEnd: Integer;
    FBufferSize: Integer;
    FStream: TStream;
    procedure SetBufferSize(const Value: Integer);
  public
    constructor Create(Stream:TStream);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
    function RestOfLine: AnsiString;
    property BufferSize: Integer read FBufferSize write SetBufferSize;
  end;

  TItemSeparator = (isTab, isComma);

  TPendingChar = (pNone, pCharSaved, pNewline, pEOF, pRecord);

  { TTextStream }

  TTextStream = class(TBufferedStream)
  private
    FColumn: Integer;
    pending: TPendingChar;
    SavedChar: AnsiChar;
    FLine:Longint;
    FItemRead: Boolean;  { Flag if a free format item has been read, but
                          not the field terminator }
  public
    NULLisEOF:Boolean;
    Token:AnsiString;
    ItemSeparators: set of TItemSeparator;
    FlowToNextLine: Boolean;
    FlowToNextRecord: Boolean; { / forces default to be read until nextline called;
                                 NextRecord makes sure a / has been read }
    Comments: Boolean; { If true, '/' '#' and ';' mark end of record }
    CommentChars: set of Char;
    AllowDefaults: Boolean;
    IncludeSpaces: Boolean; { If true, then spaces are included in tokens }
    NoParseQuotes: Boolean; { If true, do not parse quoted strings }
    RetainQuotes:  Boolean; { If true, include quotes in parsed tokens }
    StreamName: string;
    constructor Create(Stream:TStream);
    function Write(const Buffer; Count: Longint): Longint;  override;
    function Seek(const Offset: Int64; Origin: TSeekOrigin): Int64; override;
    procedure BadToken(msg:string);
  { Write }
    function PutChar(c:Char):TTextStream;
    function PutFloat(v:Extended):TTextStream;
    function PutInteger(i:Longint):TTextStream;
    function PutString(s:AnsiString):TTextStream;
    function PutPChar(p:PAnsiChar):TTextStream;
    function Space:TTextStream;
    function Tab:TTextStream;
    function Newline:TTextStream;
    function Format(const FormatStr: string; const Args: array of const): TTextStream;
  { Read }
    function Read(var Buffer; Count: Longint): Longint; override;
    function Getch:AnsiChar;
    function GetFreeItem:Boolean;
    function NextRecord:Boolean;
    function NextLine:Boolean; { Note that token is set to the (trimmed) text read }
    function RealToken:Double;
    function IsRealToken(out Value:Double):Boolean;
    function IsRealTokenOrBlank(out Value:Double):Boolean;
    function IntegerToken:Integer;
    function IsIntegerToken(var Value:Integer):Boolean;
    function Int64Token:Int64;
    function BooleanToken: Boolean;
    function CardinalToken:Cardinal;
    function IsCardinalToken(var Value:Cardinal):Boolean;
    function AtEndOfLine:Boolean;
    function AtEndOfFile:Boolean;
    function NextInteger(Default: Integer): Integer;
    function NextInt64(Default: Int64): Integer;
    function NextDouble(Default: Double): Double;
    function NextString(Default: string): string;
    function NextIntegerVar(var i: Integer):Boolean;
    function NextInt64Var(var i: Int64):Boolean;
    function NextDoubleVar(var d: Double):Boolean;
    function NextStringVar(var s: string):Boolean;
    function RestOfLine: AnsiString;
    function RestOfFile: AnsiString;
    procedure SkipItems(n: Integer);
    property Line: Longint read FLine write FLine;
    property Column: Integer read FColumn;
  end;

  TFreeFormatStream = class(TTextStream)
  public
    constructor Create(Stream:TStream);
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} Math, Huge, StrUtils,
     SysUtils, MathSup, PropHelp;

function IsEOLChar(ch:AnsiChar):Boolean;
begin
  Result := (ch = #10) or (ch = #13);
end;

procedure TTextStream.BadToken(msg:string);
var
  Name: string;
begin
  Name := StreamName;
  if Name = '' then
    Name := 'data file';
  raise Exception.CreateFmt('Error reading %s on line %d column %d: %s (%s)', [StreamName, Line, Column, msg, Token]);
end;
                                      
function TTextStream.AtEndOfLine:Boolean;
begin
  Result := pending in [pEOF, pNewline, pRecord];
end;

function TTextStream.AtEndOfFile:Boolean;
begin
  Result := pending = pEOF;
end;

function TTextStream.NextRecord:Boolean;
begin
  if pending <> pEOF then begin
    if (pending <> pRecord) and not FlowToNextRecord then begin
      repeat GetFreeItem until pending in [pEOF, pRecord];
    end;
    pending := pNewline;
    Nextline;
  end;
  FItemRead := False;
  Result := pending = pEOF;
end;

function TTextStream.Nextline:Boolean;
var
  ch:AnsiChar;
begin
  if not(pending in [pEOF, pRecord]) then begin
    if pending = pNone then
      SavedChar := Getch;
    pending := pNewline;
    Token := '';
    repeat
      ch := Getch;
      if pending = pEOF then Break;
      if IsEOLChar(SavedChar) then
        { Either LF or CR forces end of line; accept either order,
          but two LF's implies two newlines }
        if IsEOLChar(ch) and (ch <> SavedChar) then
          pending := pNone
        else
          pending := pCharSaved
      else
        Token := Token + ch;
      SavedChar := ch;
    until (pending <> pNewline);
  end;
  Token := Trim(Token);
  FItemRead := False;
  Result := pending = pEOF;
  if not Result then begin
    FLine := Line + 1;
    FColumn := 0;
  end;
end;

function TTextStream.Read(var Buffer; Count: Longint): Longint;
var
  P: PAnsiChar;
begin
  Result := 0;
  if Count = 0 then
    Exit;
  P := @Buffer;
  if pending = pCharSaved then begin
    pending := pNone;
    P^ := SavedChar;
    Inc(P);
    Inc(Result);
  end;
  Result := Result + inherited Read(P^,Count-Result);
end;

function TTextStream.Getch:AnsiChar;
begin
  if pending = pEOF then
    Result := #0
  else if pending = pCharSaved then begin
    pending := pNone;
    Result := SavedChar;
    Inc(FColumn);
  end else if inherited Read(Result, 1) <> 1 then begin
    pending := pEOF;
    Result := #0;
  end else if (Result = #0) and NULLisEOF then
    pending := pEOF
  else
    Inc(FColumn);
end;

function TTextStream.RestOfLine: AnsiString;
begin
  Result := '';
  case pending of
    pNewline:   if FlowToNextLine then NextLine else Exit;
    pEOF:       Exit;
   // pRecord:    if FlowToNextRecord then NextLine else Exit;
  end;
  if pending = pCharSaved then begin
    inherited Seek(-1,soCurrent);
    pending := pNone;
  end;
  Result := inherited RestOfLine;
  Inc(FColumn,Length(Result));
  SavedChar := Getch;
  if SavedChar <> #0 then
    pending := pNewLine;
end;

procedure TTextStream.SkipItems(n: Integer);
begin
  while (n > 0) and GetFreeItem do
    Dec(n);
end;

function TTextStream.GetFreeItem:Boolean; { Returns True if got something }
var
  Quote: AnsiChar;
  procedure AddChar;
  begin
    if Length(Token) < 256 then
      Token := Token + SavedChar
    else
      BadToken('Token too long');
    if pending = pCharSaved then
      pending := pNone;
    SavedChar := Getch;
  end;
begin
  Token := '';
  Quote := #0;
  Result := True;
  if (pending = pNone) or (pending = pCharSaved)
     or ((pending = pNewline) and FlowToNextLine)
     or ((pending = pRecord) and FlowToNextRecord)
      then begin
    if pending = pNone then begin
      SavedChar := Getch;
      if pending <> pEOF then
        pending := pCharSaved;
      Dec(FColumn);
    end else if Pending = pNewline then
      NextLine;
    while (pending <> pEOF) and Result do begin
      if not IncludeSpaces then
        while SavedChar = ' ' do SavedChar := Getch;
      if ( (SavedChar = #9) and (isTab in ItemSeparators) )
         or ((SavedChar = ',') and (isComma in ItemSeparators)) then begin
         if FItemRead then
           pending := pNone { Absorb separator of previous item }
         else begin
           Result := False; { Defaulted Item }
           Pending := pCharSaved;
           Dec(FColumn);
         end;
        FItemRead := False; { Either way no longer in a field }
      end else if IsEOLChar(SavedChar) and FlowToNextLine then begin
        pending := pNewline;
        NextLine;
      end else if SavedChar <> #9 then
        Break;
      SavedChar := Getch;
    end;
    if (pending <> pEOF) and Result then begin
      repeat
        if Comments and (Quote = #0) and (SavedChar in CommentChars) then
          if FlowToNextRecord then
            pending := pNewline
          else
            pending := pRecord
        else
          case SavedChar of
            #0:  if NULLisEOF then pending := pEOF else Break;
            ' ': if (Quote = #0) and not IncludeSpaces then begin
                   pending := pCharSaved;
                   Dec(FColumn);
                   Break;
                 end else
                   AddChar;
            #9:  if (Quote = #0) and not (IncludeSpaces and not (isTab in ItemSeparators)) then begin
                   pending := pCharSaved;
                   Dec(FColumn);
                   Break;
                 end else AddChar;
            ',': if (Quote = #0) and (isComma in ItemSeparators) then begin
                   pending := pCharSaved;
                   Dec(FColumn);
                   Break;
                 end else AddChar;
            #10,#13:
                 begin
                   pending := pNewline;
                   if (Quote <> #0) and (Length(Token) > 0) and (Token[Length(Token)] = '\') then
                     NextLine
                   else
                     Break;
                 end;
            '''', '"':
                 if NoParseQuotes then
                   AddChar
                 else if (Quote = #0) and ((Length(Token) = 0) or RetainQuotes) then begin
                   Quote := SavedChar;
                   if pending = pCharSaved then
                     pending := pNone;
                   if RetainQuotes then
                     AddChar
                   else
                     SavedChar := Getch;
                 end else if Quote = SavedChar then begin
                   if RetainQuotes then
                     AddChar
                   else
                     SavedChar := Getch;
                   if Pending = pEOF then
                     Break
                   else if SavedChar = Quote then
                     AddChar { Doubled quote character - treat as single quote in string }
                   else if not RetainQuotes then begin
                     Pending := pCharSaved;
                     Dec(FColumn);
                     Break;
                   end;
                 end else
                   AddChar; { Quote in middle of Token - either unquoted string or other sort of quote }
          else
            AddChar;
          end;
      until AtEndOfLine;
      FItemRead := True;
    end;
  end;
  Result := Length(Token) > 0;
end;

function TTextStream.IsRealToken(out Value: Double): Boolean;
begin
  Result := TryStrToFloatEx(Token,Value);
end;

function TTextStream.IsRealTokenOrBlank(out Value: Double): Boolean;
begin
  if Token = '' then begin
    Result := True;
    SetNan(Value);
  end else
    Result := IsRealToken(Value);
end;

function TTextStream.RealToken:Double;
begin
  if AnsiCompareText(Token, 'NAN') = 0 then
    SetNan(Result)
  else if (Token[1] = '*') and (Token = StringOfChar('*',Length(Token))) then
    SetNan(Result)
  else if not IsRealToken(Result) then begin
    BadToken('Bad Floating Point Number');
    Result := 0;
  end;
end;

function TTextStream.IsIntegerToken(var Value: Integer): Boolean;
begin
  Result := TryStrToInt(Token,Value);
end;

function TTextStream.Int64Token: Int64;
begin
  if not TryStrToInt64(Token,Result) then begin
    BadToken('Bad Integer');
    Result := 0;
  end;
end;

function TTextStream.BooleanToken: Boolean;
begin
  if not TryStrToBoolean(Token,Result) then begin
    BadToken('Bad True/False value');
    Result := False;
  end;
end;

function TTextStream.IntegerToken:Integer;
begin
  if not TryStrToInt(Token,Result) then begin
    BadToken('Bad Integer');
    Result := 0;
  end;
end;

function TTextStream.IsCardinalToken(var Value: Cardinal): Boolean;
var
  Code: Integer;
begin
  Val(Token,Value,Code);
  Result := Code = 0;
end;

function TTextStream.CardinalToken: Cardinal;
begin
  if not IsCardinalToken(Result) then begin
    BadToken('Bad Cardinal');
    Result := 0;
  end;
end;

function TTextStream.NextInteger(Default: Integer): Integer;
begin
  Result := Default;
  if GetFreeItem then
    Result := IntegerToken;
end;

function TTextStream.NextInt64(Default: Int64): Integer;
begin
  Result := Default;
  if GetFreeItem then
    Result := Int64Token;
end;

function TTextStream.NextDouble(Default: Double): Double;
begin
  Result := Default;
  if GetFreeItem then
    Result := RealToken;
end;

function TTextStream.NextString(Default: string): string;
begin
  Result := Default;
  if GetFreeItem then
    Result := Token;
end;

function TTextStream.NextIntegerVar(var i: Integer):Boolean;
begin
  Result := GetFreeItem;
  if Result then
    i := IntegerToken
  else if not AllowDefaults then
    raise Exception.Create('No integer');
end;

function TTextStream.NextInt64Var(var i: Int64): Boolean;
begin
  Result := GetFreeItem;
  if Result then
    i := Int64Token
  else if not AllowDefaults then
    raise Exception.Create('No int64');
end;

function TTextStream.NextDoubleVar(var d: Double):Boolean;
begin
  Result := GetFreeItem;
  if Result then
    d := RealToken
  else if not AllowDefaults then
    raise Exception.Create('No double');
end;

function TTextStream.NextStringVar(var s: string): Boolean;
begin
  Result := GetFreeItem;
  if Result then
    s := Token
  else if not AllowDefaults then
    raise Exception.Create('No string');
end;

constructor TTextStream.Create(Stream:TStream);
begin
  inherited Create(Stream);
  CommentChars := ['/', ';', '#'];
  pending := pNone;
  Token := '';
  FLine := 1;
  FColumn := 0;
  NULLisEOF := True;
  ItemSeparators := [isTab];
  FlowToNextRecord := True;
end;

constructor TFreeFormatStream.Create(Stream:TStream);
begin
  inherited Create(Stream);
  ItemSeparators := [isComma];
  Comments := True;
  FlowToNextLine := True;
  FlowToNextRecord := False;
  AllowDefaults := True;
end;

function TTextStream.Write(const Buffer; Count: Longint): Longint;
begin
  if pending = pCharSaved then
    Seek(0, soFromCurrent);
  pending := pNone;
  Result := inherited Write(Buffer, Count);
end;

function TTextStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
begin
  if (pending = pCharSaved) and (Origin = soCurrent) then
    Result := inherited Seek(Offset-1, Origin)
  else
    Result := inherited Seek(Offset, Origin);
  pending := pNone;
  FItemRead := False;
end;

function TTextStream.PutChar(c:Char):TTextStream;
begin
  {$IFDEF UNICODE}
  if Ord(c) > 128 then begin
    PutString(c); { Force conversion }
    Exit;
  end;
  {$ENDIF}
  WriteBuffer(c, 1);
  Result := Self;
end;

function TTextStream.PutFloat(v:Extended):TTextStream;
begin
  PutString(DtosM(v,-1,-1){FloatToStr(v)});
  Result := Self;
end;

function TTextStream.PutInteger(i:Longint):TTextStream;
begin
  PutString(IntToStr(i));
  Result := Self;
end;

function TTextStream.PutString(s:AnsiString):TTextStream;
begin
  if Length(s) > 0 then
    WriteBuffer(s[1], Length(s));
  Result := Self;
end;

function TTextStream.Format(const FormatStr: string; const Args: array of const): TTextStream;
begin
  Result := PutString(SysUtils.Format(FormatStr, Args));
end;

function TTextStream.PutPChar(p:PAnsiChar):TTextStream;
begin
  WriteBuffer(p[0], StrLen(p));
  Result := Self;
end;

function TTextStream.Space:TTextStream;
begin
  PutChar(' ');
  Result := Self;
end;

function TTextStream.Tab:TTextStream;
begin
  PutChar(#9);
  Result := Self;
end;

function TTextStream.Newline:TTextStream;
begin
  Result := PutString(sLineBreak);
end;

function TTextStream.RestOfFile: AnsiString;
var
  p: Integer;
begin
  { Return remainder of file as one large string }
  Result := '';
  if pending = pEOF then
    Exit;
  repeat
    p := Length(Result);
    SetLength(Result, p+10000);
    SetLength(Result, Read((PAnsiChar(Result)+p)^,10000)+p);
  until Length(Result) <> p+10000;
end;


{ TBufferedStream }

constructor TBufferedStream.Create(Stream: TStream);
begin
  inherited Create;
  FStream := Stream;
  ReadPointer := High(Integer);
  BufferEnd := 0;
  if not ((Stream is TCustomMemoryStream) or (Stream is THugeMemoryStream)) then
    BufferSize := 60000
  else
    BufferSize := 0;
end;

function TBufferedStream.Read(var Buffer; Count: Integer): Longint;
var
  n: Integer;
  P: PAnsiChar;
begin
  if BufferSize = 0 then
    Result := FStream.Read(Buffer, Count)
  else begin
    Result := 0;
    P := @Buffer;
    while Result < Count do begin
      if ReadPointer < BufferEnd then begin
        n := Min(Count-Result,BufferEnd-ReadPointer);
        Move(IntBuffer[ReadPointer],P^,n);
        Inc(Result,n);
        Inc(ReadPointer,n);
        Inc(P,n);
      end else begin
        BufferEnd := FStream.Read(IntBuffer[0],BufferSize);
        if BufferEnd = 0 then
          Exit; { End of file }
        ReadPointer := 0;
      end;
    end;
  end;
end;

function TBufferedStream.RestOfLine: AnsiString;
var
  S, T, Len: Integer;
begin
  if BufferSize = 0 then begin
    { The slow method }
    Len := 1;
    SetLength(Result,100); { Guess }
    while Read(Result[Len],1) = 1 do begin
      if Result[Len] in [#10,#13] then begin
        Seek(-1,soCurrent);
        Break;
      end;
      Inc(Len);
      if Len > Length(Result) then
        SetLength(Result,Len+100); { Guess }
    end;
    SetLength(Result,Len-1);
  end else begin
    T := 0;
    repeat
      S := ReadPointer;
      Len := S;
      while (Len < BufferEnd) and not (IntBuffer[Len] in [#10,#13]) do
        Inc(Len);
      ReadPointer := Len;
      if Len-S > 0 then begin
        SetLength(Result,T+Len-S);
        Move(IntBuffer[S],Result[T+1],Len-S);
      end else
        SetLength(Result,T);
      if Len >= BufferEnd then begin
        if Read(Len,1) <> 1 then { Trigger bufer fill }
          Exit; { End of file }
        Dec(ReadPointer);
      end else
        Exit; { End of line char found }
      T := Length(Result);
    until False;
  end;
end;

function TBufferedStream.Seek(const Offset: Int64; Origin: TSeekOrigin): Int64;
var
  NewPointer: Int64;
begin
  if FBufferSize <> 0 then
    if Origin = soCurrent then begin
      NewPointer := ReadPointer + Offset;
      if (NewPointer >= 0) and (NewPointer < BufferEnd) then begin
        { just move around within buffer }
        ReadPointer := NewPointer;
        Result := FStream.Seek(0,soCurrent)-(BufferEnd-ReadPointer);
        Exit;
      end;
    end;
  ReadPointer := BufferEnd;
  Result := FStream.Seek(Offset, Origin);
end;

procedure TBufferedStream.SetBufferSize(const Value: Integer);
begin
  FBufferSize := Value;
  SetLength(IntBuffer,FBufferSize);
end;

function TBufferedStream.Write(const Buffer; Count: Integer): Longint;
begin
  Result := FStream.Write(Buffer, Count);
end;

end.
