unit UCompat;
{ Various routines to help with compatibility between different operating
  systems and compilers }
{$I utils.inc}
{
Putils - utility routines for pgopger
Copyright (C) 2005-15 C M Western
School of Chemistry, University of Bristol, Bristol, UK
For further information please see http://pgopher.chm.bris.ac.uk/
or contact help-pgopher@bristol.ac.uk

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
interface

uses
 {$IFDEF MSWINDOWS} Windows, {$ENDIF}
 {$IFDEF UNIX} unixtype, {$ENDIF}
 Classes, IniFiles, SysUtils;

var
  VersionInfo: ShortString;

type
  EReportException = class(Exception);

function IsExpectedException(ExceptionType: TClass):Boolean;
procedure AddExpectedException(ExceptionType: ExceptClass);

procedure WriteNewline(Stream: TStream);

{$IFDEF UNIX}
const
  LineEnd: string = #10;
  NewLineLength = 1;
type
  TMRWlock_ = pthread_rwlock_t;
{$ENDIF}

{$IFDEF LINUX}
// Not connected at the moment: function CopyFile(Source, Dest:PChar; bFailIfExists: Boolean):Boolean;
function timeGetTime:Cardinal;
{$ENDIF}

{$IFDEF DARWINOFF}
type
  TMRWlock_ = TMultiReadExclusiveWriteSynchronizer;
{$ENDIF}

{$IFDEF FPC}
function StripHotKey(const s: string):string;
{$IFNDEF MSWINDOWS}
  procedure Yield;
  function MulDiv(A, B, C:Integer):Integer;
{$ENDIF}
function RoundTo(const Value: Double; Places: Integer):Double;
type
  TTextIOFunc = function (Var t:TTextRec):Integer;
procedure SetInOutRes(Value:Word);
{$ELSE}
const
  AllFilesMask: string = '*.*';
{$ENDIF}

{$IFDEF MSWINDOWS}
type
  TSRWLock = Pointer;
  TMRWlock_ = TSRWLock;

const
  LineEnd: string = #13#10;
  NewLineLength = 2;

{$IFDEF CPU16}
const
  MinusOne:Cardinal = $ffff;
{$ELSE}
const
  MinusOne:LongInt = -1;
{$ENDIF}

{$ENDIF}

procedure hmemcpy(p1, p2: Pointer; L: Longint);

function AddressOfFirstChar(const s:string):PChar;

function ApplicationSettingsDir: string;
function GetIniFile(const Name: string; WithBorland:Boolean{ = False}): TCustomIniFile;
procedure AllowCtrlC; { Allow ^C to kill process }

{$IFNDEF FPC}
type
  TLibHandle = THandle;
  PtrInt = {$IFDEF WIN64}NativeInt{$ELSE}Integer{$ENDIF};
  PtrUInt = {$IFDEF WIN64}NativeUInt{$ELSE}Cardinal{$ENDIF};
  Qword = UInt64;
  SizeUInt = {$IF CompilerVersion >= 23.0}NativeUInt {$ELSE} Cardinal {$IFEND};
  SizeInt =  {$IF CompilerVersion >= 23.0} NativeInt {$ELSE}  Integer {$IFEND};
{$IFDEF MSWINDOWS}
  TThreadID = THandle;
procedure InitCriticalSection(var lpCriticalSection: TRTLCriticalSection); stdcall;
procedure DoneCriticalSection(var lpCriticalSection: TRTLCriticalSection); stdcall;
function ParamCount: Integer;
function ParamStr(Index: Integer): string;
{$ENDIF}
procedure DoDirSeparators(var FileName: string);
function NtoLE(i:Integer):Integer;
function PadRight(const s: string; L: Integer):string;
{$ENDIF}

type
  TMRWLock = class
  private
    FlockSet: Boolean;
    FLock: TMRWlock_;
  public
    destructor Destroy; override;
    procedure SetUpLock;
    procedure BeginWrite;
    procedure BeginRead;
    procedure EndRead;
    procedure EndWrite;
  end;

procedure InitializeMRWlock(var Lock: TMRWlock_);
procedure FinalizeMRWlock(var Lock);
procedure BeginMRWread(var Lock: TMRWlock_);
procedure EndMRWread(var Lock: TMRWlock_);
procedure BeginMRWwrite(var Lock: TMRWlock_);
procedure EndMRWwrite(var Lock: TMRWlock_);

implementation
{.$DEFINE LIBC_SYSCALLS}
uses
  {$IFDEF MSWINDOWS}
    Registry, ShlObj,
  {$ELSE}
    {$IFDEF FPC} Unix, BaseUnix, {$ENDIF}
    {$IFDEF LIBC_SYSCALLS} Libc, {$ENDIF}
    dl,
  {$ENDIF}
  Worker, Math;

{$IFDEF MSWINDOWS}
procedure hmemcpy(p1, p2: Pointer; L: Longint);
begin
  CopyMemory(p1, p2, L);
end;
{$ENDIF}

{$IFDEF UNIX}

{$IFDEF FPC}

function MulDiv(A, B, C:Integer):Integer;
begin
  Result := (A * B) div C;
end;

procedure Yield;
begin
  ThreadSwitch;
end;

{$ENDIF}

procedure hmemcpy(p1, p2: Pointer; L: Longint);
begin
  Move(p2^, p1^, L);
end;
{$ENDIF}

function AddressOfFirstChar(const s:string):PChar;
begin
{$IFNDEF VER80}
  Result := PChar(s);
{$ELSE}
  Result := @s[1];
{$ENDIF}
end;

{$IFDEF LINUX}

{$IFDEF cpVERSION}
var
  CopyDosCommand: TDosCommand;
{$ENDIF}

function timeGetTime:Cardinal;
var
  tz:timeval;
begin
  fpgettimeofday(@tz,nil);
  { Ignore overflow }
  Result := Cardinal(tz.tv_sec*1000) + tz.tv_usec div 1000;
end;

function CopyFile(Source, Dest:PChar; bFailIfExists: Boolean):Boolean;
//var
  // ExitCode: LongWord;
  //SourceStream, DestStream: TFileStream;
  //Count: Int64;
  //st, dt: {$IFDEF LIBC_SYSCALLS}TStat{$ELSE}Stat{$ENDIF};
  //u: utimbuf;
begin
  Result := False;
  if bFailIfExists and FileExists(Dest) then begin
    {$IFDEF FPC}
    fpseterrno(ESysEEXIST);
    {$ELSE}
    SetLastError(EEXIST);
    {$ENDIF}
    Exit;
  end;
  {$IFDEF cpVERSION}
  if CopyDosCommand = nil then begin
    CopyDosCommand := TDosCommand.Create(Application);
    with CopyDosCommand.Arguments do begin
      Add('cp');
      Add('-p');
      Add(''); { Source }
      Add(''); { Dest }
    end;
  end;
  with CopyDosCommand do begin
    Arguments[2] := Source;
    Arguments[3] := Dest;
    Execute;
    ExitCode := WaitFor;
    if Lines.Count > 0 then
      raise Exception.Create(Lines[0])
    else if ExitCode <> 0 then
      SetLastError(ExitCode) { May not be meaningful }
    else
      Result := True;
  end;
  {$ENDIF}
  {$IFDEF Hmm}
  SourceStream := TFileStream.Create(Source, fmOpenRead);
  try
    DestStream := TFileStream.Create(Dest, fmCreate);
    Count := DestStream.CopyFrom(SourceStream, 0);
    {$IFNDEF LIBC_SYSCALLS}
    if (FPFStat(SourceStream.Handle,st) = -1)
       or (FPFStat(DestStream.Handle, dt) = -1) then
    {$ELSE}
    if (Libc.fpfstat(SourceStream.Handle, st) = -1)
       or (Libc.fpstat(DestStream.Handle, dt) = -1) then
    {$ENDIF}
      RaiseLastOSError;
    if Count <> dt.st_size then
      raise Exception.Create('Wrong number of bytes copied to '+Dest);
    {$IFNDEF LIBC_SYSCALLS}
    if Fpfchmod(DestStream.Handle, st.st_mode) = -1 then
    {$ELSE}
    if Libc.fchmod(DestStream.Handle, st.st_mode) = -1 then
    {$ENDIF}
      RaiseLastOSError;
    { Try and copy owner and group, but ignore errors }
    if st.st_gid <> dt.st_gid then
    {$IFNDEF LIBC_SYSCALLS}
      FpfChown(DestStream.Handle, uid_t(-1), st.st_gid);
    {$ELSE}
      Libc.fchown(DestStream.Handle, uid_t(-1), st.st_gid);
    {$ENDIF}
    {$IFNDEF LIBC_SYSCALLS}
    if (st.st_uid <> dt.st_uid) and (FpGetuid = 0) then
      FpfChown(DestStream.Handle, st.st_uid, gid_t(-1));
    {$ELSE}
    if (st.st_uid <> dt.st_uid) and (Libc.getuid = 0) then
      Libc.fchown(DestStream.Handle, st.st_uid, gid_t(-1));
    {$ENDIF}
    FreeAndNil(DestStream);
    u.actime := st.st_atime;
    u.modtime := st.st_mtime;
    {$IFNDEF LIBC_SYSCALLS}
    if FpUtime(Dest, @u) = -1 then
    {$ELSE}
    if Libc.utime(Dest, u) = -1 then
    {$ENDIF}
      RaiseLastOSError;
    Result := True;
  finally
    DestStream.Free;
  end;
  SourceStream.Free;
  {$ENDIF}
end;
{$ENDIF}

{$IFDEF UNIX}
type
  TMyMemIniFile = class(TMemIniFile)
  public
    destructor Destroy; override;
  end;

destructor TMyMemIniFile.Destroy;
begin
  UpdateFile;
  inherited;
end;
{$ENDIF}

var
  ApplicationSettingsDirCalled: Boolean = False;
  ApplicationSettingsDirResult: string;

function ApplicationSettingsDir: string;
 { per user space for wrtting application settings and other data }
begin
  if ApplicationSettingsDirCalled then begin
    Result := ApplicationSettingsDirResult;
    Exit;
  end;
  {$IFDEF MSWINDOWS}
    SetLength(Result,MAX_PATH+1);
    if SHGetSpecialFolderPath(0, @Result[1], CSIDL_APPDATA, False)
       or SHGetSpecialFolderPath(0, @Result[1], CSIDL_PERSONAL, False) then
      SetLength(Result, Strlen(PChar(@Result[1])))
    else
      Result := '';
  {$ENDIF}
  {$IFDEF UNIX}
    Result := GetEnvironmentVariable('HOME');
  {$ENDIF}
  ApplicationSettingsDirCalled := True;
  ApplicationSettingsDirResult := Result;
end;

function GetIniFile(const Name: string; WithBorland:Boolean{ = False}): TCustomIniFile;
var
  KeyName: string;
begin
{$IFDEF MSWINDOWS}
  {$IFDEF VER80} {$DEFINE USEINIFILE} {$ENDIF}
  {$IFDEF USEINIFILE}
    KeyName := Name + '.ini';
    Result := TIniFile.Create(KeyName);
  {$ELSE}
    if WithBorland then
      KeyName := 'SOFTWARE\BORLAND\DELPHI\' + Name
    else
      KeyName := 'SOFTWARE\' + Name;
    Result := TRegistryIniFile.Create(KeyName);
  {$ENDIF}
{$ELSE}
  {$IFDEF UNIX}
    KeyName := ApplicationSettingsDir;
    if WithBorland then
      KeyName := KeyName + {$IFDEF KYLIX}'/.borland/'{$ELSE}'/.lazarus/'{$ENDIF} + Name
    else
      KeyName := KeyName + '/.' + Name;
    Result := TMyMemIniFile.Create(KeyName);
  {$ENDIF}
{$ENDIF}
end;

procedure AllowCtrlC; { Allow ^C to kill process }
begin
  {$IFDEF LINUX}
    {$IFNDEF LIBC_SYSCALLS}
    FpSignal(SIGINT, TSignalHandler(SIG_DFL));
    {$ELSE}
    Libc.signal(SIGINT, TSignalHandler(SIG_DFL));
    {$ENDIF}
  {$ENDIF}
end;

{$IFDEF FPC}
function StripHotKey(const s: string):string;
var
  p: Integer;
begin
  Result := s;
  repeat
    p := Pos('&',Result);
    if p > 0 then
      Delete(Result,p,1);
  until p <= 0;
end;

function RoundTo(const Value: Double; Places: Integer):Double;
var
  Scale: Double;
begin
  Scale := IntPower(10, Places);
  Result := Value/Scale;
  if Abs(Result) < High(Int64) then
    Result := Round(Result); { Avoid overflow }
  Result := Result * Scale;
end;

procedure SetInOutRes(Value:Word);
begin
  InOutRes := Value;
end;
{$ENDIF}

{$IFNDEF FPC}
function NtoLE(i:Integer):Integer;
begin
  Result := i;
end;

function PadRight(const s: string; L: Integer):string;
begin
  if L > Length(s) then
    Result := s+StringOfChar(' ',L-Length(s))
  else
    Result := s;
end;

procedure DoDirSeparators(var FileName: string);
var
  i: Integer;
  OtherDelim: Char;
begin
  if PathDelim = '\' then
    OtherDelim := '/'
  else
    OtherDelim := '\';
  for i := 1 to Length(FileName) - 1 do
    if FileName[i] = OtherDelim then
      FileName[i] := PathDelim;
end;

{$IFDEF MSWINDOWS}
procedure InitCriticalSection; external kernel32 name 'InitializeCriticalSection';
procedure DoneCriticalSection; external kernel32 name 'DeleteCriticalSection';

{ Standard delphi RTL version does not allow quotes in arguments }
var
  SavedParams: array of string;

function ParamCount: Integer;
var
  P, nSlashes: Integer;
  s, a: string;
  InQuote: Boolean;
begin
  if Length(SavedParams) > 0 then
    Result := High(SavedParams)
  else begin
    { Setup cache }
    Result := -1;
    SetLength(SavedParams, 1);
    SavedParams[0] := System.ParamStr(0);
    s := GetCommandLine;
    P := 1;
    repeat
      a := '';
      InQuote := False;
      while (P <= Length(s)) and (s[P] <= ' ') do
        Inc(P);
      if P > Length(s) then
        Exit;
      repeat
        case s[P] of
          '\': begin
                 { Weird microsoft quoting rules: see "Parsing C Command-Line Arguments"
                   https://msdn.microsoft.com/en-us/library/ms880421.aspx }
                 nSlashes := 0;
                 repeat
                   Inc(nSlashes);
                   Inc(P);
                 until (P > Length(s)) or (s[P] <> '\');
                 if (P > Length(s)) or (s[P] <> '"') then begin
                   a := a + StringOfChar('\', nSlashes); { Leave as is }
                   Dec(P); { Re-cosider terminal characret }
                 end else begin
                   { \\\\\\" }
                   a := a + StringOfChar('\', nSlashes div 2);
                   if Odd(nSlashes) then
                     a := a + '"'
                   else
                     InQuote := not InQuote;
                 end;
               end;
          '"': InQuote := not InQuote;
          #0..' ': if InQuote then
                     a := a + s[P]
                   else
                     Break;
        else
          a := a + s[P];
        end;
        Inc(P);
      until P > Length(s);
      Inc(Result);
      SetLength(SavedParams, Result+1);
      if Result > 0 then
        SavedParams[Result] := a;
    until False;
  end;
end;

function ParamStr(Index: Integer): string;
begin
  if Index > ParamCount then
    Result := ''
  else
    Result := SavedParams[Index];
end;

{$ENDIF}
{$ENDIF}

procedure WriteNewline(Stream: TStream);
var
  s: string[2];
begin
  s := LineEnd;
  Stream.WriteBuffer(s[1], NewLineLength);
end;

{ TODO: May want to trap runerror or add an exit proc,
        as some thread failures call exit, rather than raise an exception }

var
  ExpectedExceptions: array of ExceptClass;

function IsExpectedException(ExceptionType: TClass):Boolean;
var
  i: Integer;
begin
  Result := True;
  for i := 0 to High(ExpectedExceptions) do
    if ExpectedExceptions[i] = ExceptionType then
      Exit;
  Result := false;
end;

procedure AddExpectedException(ExceptionType: ExceptClass);
begin
  if IsExpectedException(ExceptionType) then
    Exit;
  SetLength(ExpectedExceptions, Length(ExpectedExceptions)+1);
  ExpectedExceptions[High(ExpectedExceptions)] := ExceptionType;
end;

{$IFDEF MSWINDOWS}
var
  InitializeSRWLock,
  AcquireSRWLockShared, ReleaseSRWLockShared,
  AcquireSRWLockExclusive, ReleaseSRWLockExclusive: function (var P: TSRWLock): NativeUInt; stdcall;
  Kernel: HMODULE;
{$ENDIF}

{$IFNDEF UNIX}
procedure InitializeMRWlock(var Lock: TMRWlock_);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock <> nil then
    InitializeSRWLock(Lock)
  else
  {$ENDIF}
    Lock := TMultiReadExclusiveWriteSynchronizer.Create;
end;

procedure BeginMRWread(var Lock: TMRWlock_);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock <> nil then
    AcquireSRWLockShared(Lock)
  else
  {$ENDIF}
    TMultiReadExclusiveWriteSynchronizer(Lock).BeginRead;
end;

procedure EndMRWread(var Lock: TMRWlock_);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock <> nil then
    ReleaseSRWLockShared(Lock)
  else
  {$ENDIF}
    TMultiReadExclusiveWriteSynchronizer(Lock).EndRead;
end;

procedure BeginMRWwrite(var Lock: TMRWlock_);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock <> nil then
    AcquireSRWLockExclusive(Lock)
  else
  {$ENDIF}
    TMultiReadExclusiveWriteSynchronizer(Lock).BeginWrite;
end;

procedure EndMRWwrite(var Lock: TMRWlock_);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock <> nil then
    ReleaseSRWLockExclusive(Lock)
  else
  {$ENDIF}
    TMultiReadExclusiveWriteSynchronizer(Lock).EndWrite;
end;

procedure FinalizeMRWlock(var Lock);
begin
  {$IFDEF MSWINDOWS}
  if @InitializeSRWLock = nil then
  {$ENDIF}
    FreeAndNil(Lock)
end;
{$ENDIF}

{$IFDEF UNIX}
type
  ppthread_rwlock_t = ^pthread_rwlock_t;
  ppthread_rwlockattr_t = ^pthread_rwlockattr_t;

var
  threadLib : Pointer = nil;
  pthread_rwlock_init: function(rwlock: ppthread_rwlock_t; attr: ppthread_rwlockattr_t):longint;cdecl;
  pthread_rwlock_destroy: function(rwlock: ppthread_rwlock_t):longint;cdecl;
  pthread_rwlock_rdlock: function(rwlock: ppthread_rwlock_t):longint;cdecl;
  pthread_rwlock_tryrdlock: function(rwlock: ppthread_rwlock_t):longint;cdecl;
  pthread_rwlock_unlock: function(rwlock: ppthread_rwlock_t):longint;cdecl;
  pthread_rwlock_wrlock: function(rwlock: ppthread_rwlock_t):longint;cdecl;
  pthread_rwlock_trywrlock: function(rwlock: ppthread_rwlock_t):longint;cdecl;

//procedure sched_yield; cdecl; external 'c' name 'sched_yield';

procedure InitializeMRWlock(var Lock: TMRWlock_);
begin
  pthread_rwlock_init(@Lock, nil);
end;

procedure FinalizeMRWlock(var Lock);
begin
  pthread_rwlock_destroy(@Lock);
end;

procedure BeginMRWread(var Lock: TMRWlock_);
begin
  //while pthread_rwlock_tryrdlock(@Lock) <> 0 do
  //  sched_yield();
  pthread_rwlock_rdlock(@Lock);
end;

procedure EndMRWread(var Lock: TMRWlock_);
begin
  pthread_rwlock_unlock(@Lock);
end;

procedure BeginMRWwrite(var Lock: TMRWlock_);
begin
  //while pthread_rwlock_trywrlock(@Lock) <> 0 do
  //  sched_yield();
  pthread_rwlock_wrlock(@Lock);
end;

procedure EndMRWwrite(var Lock: TMRWlock_);
begin
  pthread_rwlock_unlock(@Lock);
end;
{$ENDIF}

{ TMRWLock }

procedure TMRWLock.BeginRead;
begin
  if FlockSet then
    BeginMRWread(FLock);
end;

procedure TMRWLock.BeginWrite;
begin
  if FlockSet then
    BeginMRWwrite(FLock);
end;

destructor TMRWLock.Destroy;
begin
  if FlockSet then begin
    FinalizeMRWlock(Flock);
    FlockSet := False;
  end;
  inherited;
end;

procedure TMRWLock.EndRead;
begin
  if FlockSet then
    EndMRWread(FLock);
end;

procedure TMRWLock.EndWrite;
begin
  if FlockSet then
    EndMRWwrite(FLock);
end;

procedure TMRWLock.SetUpLock;
begin
  if (nWorkers > 0) and not FlockSet then begin
    InitializeMRWlock(FLock);
    FlockSet := True;
  end;
end;

initialization
  AddExpectedException(EReportException);
  {$IFDEF UNIX}
    {$IFDEF LINUX}
    threadLib := dlopen('libpthread.so.0', RTLD_LAZY );
    {$ENDIF}
    {$IFDEF DARWIN}
    threadlib := dlopen('libSystem.'+SharedSuffix, RTLD_LAZY);
    {$ENDIF}
    if threadLib <> nil then begin
      Pointer(pthread_rwlock_init) := dlsym(threadLib, 'pthread_rwlock_init');
      Pointer(pthread_rwlock_destroy) := dlsym(threadLib, 'pthread_rwlock_destroy');
      Pointer(pthread_rwlock_rdlock) := dlsym(threadLib, 'pthread_rwlock_rdlock');
      Pointer(pthread_rwlock_tryrdlock) := dlsym(threadLib, 'pthread_rwlock_tryrdlock');
      Pointer(pthread_rwlock_unlock) := dlsym(threadLib, 'pthread_rwlock_unlock');
      Pointer(pthread_rwlock_wrlock) := dlsym(threadLib, 'pthread_rwlock_wrlock');
      Pointer(pthread_rwlock_trywrlock) := dlsym(threadLib, 'pthread_rwlock_trywrlock');
    end
  {$ENDIF}
  {$IFDEF MSWINDOWS}
    Kernel := GetModuleHandle(Windows.kernel32);
    if Kernel <> 0 then begin
      InitializeSRWLock := GetProcAddress(Kernel, 'InitializeSRWLock');
      AcquireSRWLockShared := GetProcAddress(Kernel, 'AcquireSRWLockShared');
      ReleaseSRWLockShared := GetProcAddress(Kernel, 'ReleaseSRWLockShared');
      AcquireSRWLockExclusive := GetProcAddress(Kernel, 'AcquireSRWLockExclusive');
      ReleaseSRWLockExclusive := GetProcAddress(Kernel, 'ReleaseSRWLockExclusive');
    end;
  {$ENDIF}
finalization
  {$IFDEF LINUX}
    if threadLib <> nil then
      dlclose(threadLib);
  {$ENDIF}
end.


