unit Worker;
{$I utils.inc}{ Copyright (C) 2006 C M Western; see UCompat.pas for license}
{ Worker threads }
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Classes, SyncObjs, SysUtils, UCompat, TextAssign, Math;

type

  { TWorkerThread }

  TWorkerThread = class (TThread)
  private
    FNumber: Integer;
    FBusy: Boolean;
    FGoEvent: TEvent;
    PLog: PTextRec;
    FDone: TEvent;
    FExceptionMask: TFPUExceptionMask;
    procedure SetBusy(const Value: Boolean);
  protected
    procedure Execute; override;
    procedure DoWork; virtual; abstract;
    procedure DoTerminate; override;
  public
    LastException: string;
    constructor Create(CreateSuspended: Boolean; ANumber: Integer; var ALog: TextFile);
    destructor Destroy; override;
    procedure StartWork;
    procedure Discard(InDestroy:Boolean = False);
    property Number: Integer read FNumber;
    property Busy: Boolean read FBusy write SetBusy;
    property Done: TEvent read FDone;
  end;
  TWorkerClass = class of TWorkerThread;

procedure StartWorkers(nWorkers: Integer; WorkerClass: TWorkerClass);
function FindWorker:TWorkerThread;
function FindWorkerWait:TWorkerThread;
procedure WaitNoBusyWorkers;
function nWorkers:Integer;
procedure AtThreadEnd(Proc: TProcedure);

function SimpleWait(var Event: TEvent; Timeout: Cardinal; ReturnOnAbort: Boolean = True): TWaitResult;

type
  TWaitFunction = function (var Event: TEvent; Timeout: Cardinal; ReturnOnAbort: Boolean = True): TWaitResult;

var
  Aborted: Boolean = False;
  DoWait: TWaitFunction = {$IFDEF FPC}@{$ENDIF}SimpleWait; { Wait for event, with timeout in ms; variable so a different function can be used
                                         for the main thread of a gui app to process messages while doing so }
  ThreadState: ShortString = '__';

procedure SetThreadName(AThreadName: String; Dummy: Boolean = False);
procedure AbortCalculation;
procedure SaveThreadStatus;
procedure ShowThreadStatus(var F: Text);
procedure SetThreadState(c: AnsiChar);

procedure ConsiderException(var F: Text; E: Exception);

threadvar
  Log: Text; { Log output }
  ThreadIndex: Integer;

function ThreadName: string;

{$IFDEF UNIX}
const
  INFINITE = Cardinal(-1);
{$ENDIF}

implementation

uses {$IFDEF VCL_} StackDump, StackTrace, {$ENDIF}
     {$IFDEF UNIX} unixtype, dl, {$ENDIF}
     Sysinfo, Profile, Clock;

{$IFDEF DELPHI}
  {$IF CompilerVersion >= 19.0}
  {$DEFINE NEWDELPHIEX}
  {$IFEND}
{$ENDIF}

procedure ConsiderException(var F: Text; E: Exception);
begin
  if not IsExpectedException(E.ClassType) then begin
    WriteLn(F, '**** Exception: ',E.Message,' in ', ThreadName);
    {$IFDEF FPC}
      DumpExceptionBackTrace(F);
    {$ENDIF}
    {$IFDEF NEWDELPHIEX}
      WriteLn(F, E.StackTrace);
    {$ENDIF}
    WriteLn(F);
  end;
end;

var
  Workers: array of TWorkerThread;
  FnWorkers: Integer = 0;
  nFreeWorkers: Integer = 0;
  AnyWorkerFree: TEvent;
  AllWorkersFree: TEvent;
  BusyLock: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

var
  BusyLockWait: string = 'BusyLock Wait';
  StartWorker: string = 'Start Worker';

function nWorkers: Integer;
begin
  Result := FnWorkers;
end;

{$IFDEF UNIX}
// function sem_timedwait(var __sem: sem_t; var __abs_timeout: TTimeSpec):cint;cdecl;external libthreads;
{$ENDIF}

var
  Initialized: Boolean = False;

procedure InitWorker;
begin
  if not Initialized then begin
    AnyWorkerFree := TEvent.Create(nil,True,False,'');
    AllWorkersFree := TEvent.Create(nil,True,False,'');
    InitCriticalSection(BusyLock);
  end;
  Initialized := True;
end;

{ TWorkerThread }

constructor TWorkerThread.Create(CreateSuspended: Boolean; ANumber: Integer; var ALog: TextFile);
begin
  InitWorker;
  { FPC note: Stack size for worker threads should be the same size as the main thread }
  inherited Create(CreateSuspended{$IFDEF FPC},StackLength{$ENDIF});
  FExceptionMask := GetExceptionMask;
  FNumber := ANumber;
  PLog := PTextRec(@ALog);
  FGoEvent := TEvent.Create(nil,False,False,'');
  FBusy := True;
  FDone := TEvent.Create(nil,True,True,'');
  EnterCriticalSection(BusyLock);
  try
    InterLockedIncrement(FnWorkers);
    SetBusy(True);
  finally
    LeaveCriticalsection(BusyLock);
  end;
end;

destructor TWorkerThread.Destroy;
begin
  Discard(True);
  inherited;
  FGoEvent.Free;
  FDone.Free;
  InterLockedDecrement(FnWorkers);
  SetBusy(Busy); { Make sure Anyworker/AllWorkers in correct state }
end;

procedure MakeBigStack(n: Integer);
var
  a: array [0..50] of Double;
begin
  a[0] := 1;
  a[High(a)] := 1;
  if n > 0 then
    MakeBigStack(n-1);
end;
{$IFDEF OLD}
procedure MakeBigStack(n: Integer);
var
//  Eigenvals: TColumnVector;
//  Matrix: TMatrix;
//  i: Integer;
  a: array [0..10000] of Double;
begin
  a[0] := 1;
  a[High(a)] := n;
  n := n div 2;
//  Eigenvals := TColumnVector.Create(n);
//  Matrix := TMatrix.Create(n, n);
//  Matrix.Zero;
//  for i := 1 to n do begin
//    Matrix[i,i] := i;
//  end;
//  Eigen(Eigenvals, Matrix, True);
//  Eigenvals.Free;
//  Matrix.Free;
end;
{$ENDIF}

procedure TWorkerThread.Execute;
begin
  SetThreadName('Worker '+IntToStr(Number));
  AssignTextFile(Log,TextFile(PLog^));
  Rewrite(Log);
  SetExceptionMask(FExceptionMask);  { Copy exception mask from main thread }
  { TODO: This may no longer be required }
  try
    MakeBigStack(200);
  except
    on E:Exception do
      WriteLn(Log, E.Message);
    { ??? Needed to kickstart the stack expansion mechanism }
  end;
  repeat
    Busy := False;
    SetThreadState('I');
    FGoEvent.WaitFor(INFINITE);
    SetThreadState(' ');
    if not Terminated then begin
      LastException := '';
      try
        DoWork;
      except
        on E:Exception do begin
          LastException := E.Message;
          ConsiderException(Log, E);
        end;
      end;
    end;
  until Terminated;
  // WriteLn(Log, Clocks, ' Going: ', FnWorkers, ' ', nFreeWorkers, ' ', Terminated);
  CloseFile(Log);
  SetThreadName('');
end;

var
  ThreadEndProcs: array of TProcedure;

procedure AtThreadEnd(Proc: TProcedure);
begin
  SetLength(ThreadEndProcs, Length(ThreadEndProcs)+1);
  ThreadEndProcs[High(ThreadEndProcs)] := Proc;
end;

procedure TWorkerThread.DoTerminate;
var
  i: Integer;
begin
  for i := 0 to High(ThreadEndProcs) do
    ThreadEndProcs[i];
  SetThreadName('');
  inherited DoTerminate;
end;

procedure TWorkerThread.SetBusy(const Value: Boolean);
var
  i: Integer;
begin
  SetThreadState('B');
  TStart(BusyLockWait);
  EnterCriticalSection(BusyLock);
  TEnd(BusyLockWait);
  try
    SetThreadState('b');
    if FBusy <> Value then
      if Value then begin
        FBusy := True;
        Dec(nFreeWorkers);
        Done.ResetEvent;
        if nFreeWorkers < 0 then
          raise Exception.Create('ChangeBusy-');
      end else begin
        FBusy := False;
        Inc(nFreeWorkers);
        Done.SetEvent;
        if nFreeWorkers > nWorkers then
          raise Exception.Create('ChangeBusy+');
      end;
    if nFreeWorkers = 0 then
      AnyWorkerFree.ResetEvent
    else
      AnyWorkerFree.SetEvent;
    if nFreeWorkers >= nWorkers then
      AllWorkersFree.SetEvent
    else
      AllWorkersFree.ResetEvent;
    //sem_getvalue(PSemaphore(AllWorkersFree.Handle)^, i1);
    //sem_getvalue(PSemaphore(AnyWorkerFree.Handle)^, i2);

    // Can only do this if writing to file:
    // WriteLn(log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers, ' ', FBusy, ' ',ThreadName, ' All', i1, ' Any', i2);
  finally
    LeaveCriticalSection(BusyLock);
  end;
  SetThreadState(' ');
  // WriteLn(log, Clocks, ' ', FnWorkers, ' ', nFreeWorkers, ' ', FBusy, ' ',ThreadName, ' All', i1, ' Any', i2);
end;

procedure TWorkerThread.StartWork;
begin
  TStart(StartWorker);
  SetThreadState('S');
  FGoEvent.SetEvent;
  SetThreadState(' ');
  TEnd(StartWorker);
end;

procedure TWorkerThread.Discard(InDestroy:Boolean);
begin
  FreeOnTerminate := not InDestroy;
  EnterCriticalSection(BusyLock);
  try
    Terminate;
    if not Busy then begin
      Busy := True;
      StartWork; { If waiting (as it probably will be), stop the wait }
    end;
    FDone.SetEvent; { In case anyone is waiting }
  finally
    LeaveCriticalsection(BusyLock);
  end;
end;

procedure StartWorkers(nWorkers: Integer; WorkerClass: TWorkerClass);
{ Assumed called from main thread }
var
  OldnWorkers, I: Integer;
begin
  if nWorkers < 0 then begin
    if nCPUs = 1 then
      nWorkers := 0
    else
      nWorkers := 2*nCPUs;
  end else if nWorkers < 2 then
    nWorkers := 0; { Only one worker is not useful }
  OldnWorkers := Length(Workers);
  if nWorkers > OldnWorkers then begin
    SetLength(Workers, nWorkers);
    SetLength(ThreadState, Max(255,nWorkers));
    for I := OldnWorkers to nWorkers - 1 do begin
        Workers[I] := WorkerClass.Create(True, I, Log);
        if Assigned(Workers[I].FatalException) then
          raise Workers[I].FatalException;
        Workers[I].Resume;
        WaitNoBusyWorkers; { Forces timing information in same order as worker number }
    end;
  end else if nWorkers < OldnWorkers then begin
    for I := OldnWorkers - 1 downto nWorkers do
      Workers[I].Discard;
    AllWorkersFree.WaitFor(INFINITE);
    EnterCriticalsection(BusyLock); { This ensures all the threads have left the BusyLock section, }
    LeaveCriticalsection(BusyLock); { so must be exiting SetBusy at the end of the Destroy method }
    Sleep(1); { Allow a little more time for the threads to really finish - probably only an issue for leak detection }
    SetLength(ThreadState, Max(255,nWorkers));
    SetLength(Workers, nWorkers);
  end;
end;

function FindWorker:TWorkerThread;
{ Assumed called from main thread }
var
  I: Integer;
begin
  if GetCurrentThreadID <> MainThreadID then
    raise Exception.Create('FindWorker called from worker');
  Result := nil;
  if Aborted then
    Exit;
  for I := 0 to High(Workers) do
    if not Workers[I].Busy then begin
      Result := Workers[I];
      Result.Busy := True;
      Exit;
    end;
end;

function SimpleWait(var Event: TEvent; Timeout: Cardinal; ReturnOnAbort: Boolean): TWaitResult;
var
  TheEvent: TEvent;
begin
  TheEvent := Event; { In case Event is assigned to a different value }
  if not Assigned(TheEvent) then
    Result := wrAbandoned
  else begin
    Timeout := INFINITE;
    Result := TheEvent.WaitFor(Timeout);
    if Result = wrError then
      RaiseLastOSError;
  end;
end;

function FindWorkerWait:TWorkerThread;
var
  W: TWaitResult;
begin
  Result := nil;
  while nFreeWorkers = 0 do begin
    SetThreadState('W');
    W := DoWait(AnyWorkerFree, 1000);
    case W of
      wrSignaled: Break;
      wrTimeout:  Continue; { TODO: On repeat, could check progress dialog is up }
    else
      WriteLn(Log, Ord(W));
      Exit; { Abort or other issue }
    end;
  end;
  SetThreadState(' ');
  Result := FindWorker;
  if not Assigned(Result) and not Aborted then begin
    { Don't try again, as must be called from main thread }
    raise Exception.Create('No worker found');
    // Result := FindWorkerWait;
  end;
end;

procedure WaitNoBusyWorkers;
begin
  while nFreeWorkers < nWorkers do
    DoWait(AllWorkersFree, 1000, False);
end;

{$IFDEF DCC}
{ Help for Delphi Debugger }
type
  TThreadNameInfo = record
    FType: LongWord;     // must be 0x1000
    FName: PChar;        // pointer to name (in user address space)
    FThreadID: LongWord; // thread ID (-1 indicates caller thread)
    FFlags: LongWord;    // reserved for future use, must be zero
  end;
{$ENDIF}
{$IFDEF UNIX}
var
  {$IFDEF LINUX}
    pthread_setname_np : function(__thread:pthread_t; __name:pointer):longint;cdecl;
  {$ENDIF}
  {$IFDEF DARWIN}
    pthread_setname_np : function(name:PAnsiChar):longint;cdecl;
  {$ENDIF}
  threadLib : Pointer = nil;
{$ENDIF}

threadvar
  MyThreadName: string;

function ThreadName: string;
begin
  Result := MyThreadName;
end;

procedure SetThreadState(c: AnsiChar);
begin
  if (ThreadIndex > 0) and (ThreadIndex < 256) then
    ThreadState[ThreadIndex] := c;
end;

procedure SetThreadName(AThreadName: String; Dummy: Boolean = False);
var
{$IFDEF DCC}
  ThreadNameInfo: TThreadNameInfo;
{$ENDIF}
  T: AnsiString;
begin
  if AThreadName <> '' then begin
    T := AThreadName;
  {$IFDEF DCC}
      ThreadNameInfo.FType := $1000;
      ThreadNameInfo.FName := PChar(T);
      ThreadNameInfo.FThreadID := $FFFFFFFF;
      ThreadNameInfo.FFlags := 0;
      try
        RaiseException( $406D1388, 0, sizeof(ThreadNameInfo) div sizeof(LongWord), @ThreadNameInfo );
      except
      end;
  {$ENDIF}
  {$IFDEF LINUX}
    if Length(T) > 15 then
      SetLength(T,15); { Must be no more than 16 chars, including trminating null }
    if pthread_setname_np <> nil then
      pthread_setname_np(GetCurrentThreadId, PAnsiChar(T));
  {$ENDIF}
  {$IFDEF DARWIN}
    if not Dummy and (pthread_setname_np <> nil) then
      pthread_setname_np(PAnsiChar(T));
  {$ENDIF}
  end;
  {$IFDEF FPC}
  TThread.NameThreadForDebugging(AThreadName);
  {$ENDIF}
  MyThreadName := AThreadName;
  if ThreadName <> '' then begin
    if Initialized then
      EnterCriticalSection(BusyLock);
    try
      ThreadIndex := Length(Marks);
      SetLength(Marks, Length(Marks)+1);
      SetLength(ThreadNames, Length(Marks));
      ThreadNames[ThreadIndex] := ThreadName;
      GrowMarks(0);
    finally
      if Initialized then
        LeaveCriticalsection(BusyLock);
    end;
  end;
end;

var
  SavedThreadStatus: array of Boolean;

procedure SaveThreadStatus;
var
  i: Integer;
begin
  SetLength(SavedThreadStatus, nWorkers);
  for i := 0 to nWorkers - 1 do
    SavedThreadStatus[i] := Workers[i].Busy;
end;

procedure ShowThreadStatus(var F: Text);
var
  i: Integer;
begin
  for i := 0 to Min(nWorkers, Length(SavedThreadStatus)) - 1 do
    Writeln(Log, ThreadNames[i+2], ' ', SavedThreadStatus[i]);
end;

procedure AbortCalculation;
{$IFDEF DumpAbortStack}
{$IFNDEF FPC}
var
  St: TCallStack;
{$ENDIF}
{$ENDIF}
begin
  {$IFDEF DumpAbortStack}
    SaveThreadStatus;
    {$IFDEF FPC}
    Dump_Stack(Log, get_frame);
    {$ELSE}
    FillCallStack(St,0);
    WriteLn(Log,CallStackTextualRepresentation(St,''));
    {$ENDIF}
    ShowThreadStatus(Log);
  {$ENDIF}
  WaitNoBusyWorkers;
  Abort;
end;

initialization
  SetThreadName('Totals', True); { Allocate index 0 for totals }
  SetThreadName('Main');   { So main thread is index 1 }
  AddExpectedException(EReportException);
  {$IFDEF UNIX}
    {$IFDEF LINUX}
    threadLib := dlopen('libpthread.so.0', RTLD_LAZY );
    {$ENDIF}
    {$IFDEF DARWIN}
    threadLib := dlopen('libSystem.'+SharedSuffix, RTLD_LAZY);
    {$ENDIF}
    if threadLib <> nil then
      Pointer(pthread_setname_np) := dlsym(threadLib, 'pthread_setname_np');
  {$ENDIF}
finalization
  if Initialized then begin
    StartWorkers(0,nil);
    FreeAndNil(AnyWorkerFree);
    FreeAndNil(AllWorkersFree);
    DoneCriticalSection(BusyLock);
  end;
  SetThreadName ('');
  SavedThreadStatus := nil;
  {$IFDEF UNIX}
    pthread_setname_np := nil;
    if threadLib <> nil then
      dlclose(threadLib);
  {$ENDIF}
end.

