unit FixForm;
{$I utils.inc}
interface

{ Derive forms from this to get more conventional MS Windows behaviour,
  especially under Vista; See:
  http://www.installationexcellence.com/articles/VistaWithDelphi/Index.html
  http://www.installationexcellence.com/articles/VistaWithDelphi/Original/Index.html
  For Delphi 2006 (and probably below) this requires Forms.pas in the VCL to be patched to avoid problems with ShowModal.
  In addition Define OLD below
}

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     Classes, Controls, Forms, Menus;

type

  { TFixForm }

  TFixForm = class(TForm)
  private
    FFixThisForm: Boolean;
  protected
    {$IFDEF VCL_}
    FCurrentChildrenOnTop: Boolean;
    {$IFDEF OLD}
    procedure WMActivate(var Message: TWMActivate); message WM_ACTIVATE;
    {$ENDIF}
    procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
    procedure CreateParams(var Params: TCreateParams); override;
    {$ELSE}
    procedure CreateWnd; override;
    {$ENDIF}
    procedure UpdateActions; override;
    procedure ApplyFix; virtual;
    property FixThisForm: Boolean read FFixThisForm;
  public
    { Used to display a shortcut key in a menu, but not handle it via the standard
      shortcut mechanism. Appropriate for basic printing keys like a..z 0..9, which
      would otherwise be unusable in, say, text fields. Currently requires manual
      handling of character in KeyPress or related events }
    procedure AddShortCut(Item: TMenuItem; ShortCut: string);
  end;

var
  FixAllForms: Boolean = True;

type
  TWindowManagerMode = (wmmStandard, wmmOnTop, wmmIndependent);

function GetAllOnTaskBarMode: Boolean;
procedure SetAllOnTaskBarMode(OnTaskBar: Boolean);

function GetChildrenOnTop: Boolean;
procedure SetChildrenOnTop(OnTop: Boolean);

implementation

var
  AllOnTaskBarMode: Boolean = False;
  PendingAllOnTaskBarMode: Boolean;
  { ChildrenOnTop fails on Carbon as popup menus on child forms are behind form }
  ChildrenOnTop: Boolean = {$IFDEF DARWIN} False {$ELSE} True {$ENDIF};
  PendingChildrenOnTop: Boolean;

function GetAllOnTaskBarMode: Boolean;
begin
  Result := AllOnTaskBarMode;
end;

procedure SetAllOnTaskBarMode(OnTaskBar: Boolean);
begin
  PendingAllOnTaskBarMode := OnTaskBar;
end;

function GetChildrenOnTop: Boolean;
begin
  Result := ChildrenOnTop;
end;

procedure SetChildrenOnTop(OnTop: Boolean);
begin
  {$IFNDEF DARWIN}
  PendingChildrenOnTop := OnTop;
  {$ENDIF}
end;

{ TFixForm }

{$IFDEF VCL_}
procedure TFixForm.CreateParams(var Params: TCreateParams);
begin
  FCurrentChildrenOnTop := PendingChildrenOnTop;
  if Assigned(Application.MainForm) and (Self <> Application.MainForm) then
    FFixThisForm := FixAllForms
  else
    FFixThisForm := False; { No action needed on main form }
  if FixThisForm then
    if PendingAllOnTaskBarMode then
      BorderIcons := BorderIcons + [biMinimize]
    else
      BorderIcons := BorderIcons - [biMinimize];
  inherited;
  if FixThisForm then begin
    if PendingAllOnTaskBarMode then
      Params.ExStyle := Params.ExStyle or WS_EX_APPWINDOW;
    { Must switch ownership of every form to MainForm, so that the
      application minimizes corrrectly }
    if not FCurrentChildrenOnTop then
      Params.WndParent := Application.Handle;
  end;
end;

procedure TFixForm.WMSysCommand(var Message: TWMSysCommand);
var
  i: Integer;
begin
  inherited;
  if (Application.MainForm = Self) and not (AllOnTaskBarMode or ChildrenOnTop)
     and ( ((Message.CmdType and $FFF0) = SC_MINIMIZE)
           or ((Message.CmdType and $FFF0) = SC_RESTORE) ) then
    for i := 0 to Screen.CustomFormCount - 1 do
      if (Screen.CustomForms[i] is TFixForm) and (Screen.CustomForms[i] <> Self) then
        with TFixForm(Screen.CustomForms[i]) do
          if FixThisForm then
            case (Message.CmdType and $FFF0) of
              SC_MINIMIZE: if Visible then WindowState := wsMinimized;
              SC_RESTORE:  if Visible then WindowState := wsNormal;
            end;
end;

{$IFDEF OLD}
var
  OriginalStyle: LongInt;

procedure AdjustHiddenWindow;
begin
  if OriginalStyle = 0 then
    OriginalStyle := GetWindowLong(Application.Handle,GWL_EXSTYLE);
  if WindowManagerMode <> wmmStandard then
    AdjustWindow(Application.Handle, OriginalStyle and not WS_EX_APPWINDOW or WS_EX_TOOLWINDOW)
  else
    AdjustWindow(Application.Handle, OriginalStyle)
end;

procedure TFixForm.WMActivate(var Message: TWMActivate);
begin
 if (Message.Active = WA_ACTIVE) and not IsWindowEnabled(Handle) then begin
   SetActiveWindow(Application.Handle);
   Message.Result := 0;
 end else
   inherited;
end;
{$ENDIF}

type TOpenWindow = class(TWinControl);

procedure ApplyChange;
var
  i: Integer;
begin
  for i := 0 to Screen.CustomFormCount - 1 do
    if Screen.CustomForms[i] is TFixForm then
       TFixForm(Screen.CustomForms[i]).ApplyFix;
end;

procedure TFixForm.ApplyFix;
  procedure AdjustWindow(NewStyle: Longint);
  var
    OldStyle: Longint;
    WasVisible: Boolean;
    OldState: TWindowState;
  begin
    OldStyle := GetWindowLong(Handle, GWL_EXSTYLE);
    if OldStyle <> NewStyle then begin
      WasVisible := Visible;
      if Visible then
         Hide;
      OldState := WindowState;
      SetWindowLong(Handle, GWL_EXSTYLE, NewStyle);
      if WasVisible then
        Show;
      if not AllOnTaskBarMode and (OldState = wsMinimized) then
        WindowState := wsNormal;
    end;
  end;
begin
  if FixThisForm and HandleAllocated then
    if FCurrentChildrenOnTop <> ChildrenOnTop then
      RecreateWnd
    else if AllOnTaskBarMode then begin
      { Window on task bar; can minimize }
      AdjustWindow(GetWindowLong(Handle, GWL_EXSTYLE) or WS_EX_APPWINDOW);
      BorderIcons := BorderIcons + [biMinimize];
    end else begin
      { Window not on task bar; can't minimize }
      AdjustWindow(GetWindowLong(Handle, GWL_EXSTYLE) and not WS_EX_APPWINDOW);
      BorderIcons := BorderIcons - [biMinimize];
    end;
end;

{$ELSE}

procedure ApplyChange;
var
  i: Integer;
begin
  if AllOnTaskBarMode then
    Application.TaskBarBehavior := tbMultiButton
  else
    Application.TaskBarBehavior := tbSingleButton;
  for i := 0 to Screen.CustomFormCount - 1 do
    if Screen.CustomForms[i] is TFixForm then
       TFixForm(Screen.CustomForms[i]).ApplyFix;
end;

procedure TFixForm.CreateWnd;
begin
  FFixThisForm := FixAllForms;
  ApplyFix;
  inherited CreateWnd;
end;

procedure TFixForm.ApplyFix;
begin
  if FixThisForm and Assigned(Application.MainForm) and (Self <> Application.MainForm) then begin
    if AllOnTaskBarMode then
      BorderIcons := BorderIcons + [biMinimize]
    else
      BorderIcons := BorderIcons - [biMinimize];
    if ChildrenOnTop then
      PopupParent := Application.MainForm
    else
      PopupParent := nil;
    if HandleAllocated then
      RecreateWnd(Self);
  end;
end;

{$ENDIF}

procedure TFixForm.UpdateActions;
begin
  inherited;
  { Update is deferred to here as VCL seems to be confused about recreated window handles }
  if (PendingAllOnTaskBarMode <> AllOnTaskBarMode) or (PendingChildrenOnTop <> ChildrenOnTop) then begin
    AllOnTaskBarMode := PendingAllOnTaskBarMode;
    ChildrenOnTop := PendingChildrenOnTop;
    if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then begin
      ApplyChange;
    end;
  end;
end;

procedure TFixForm.AddShortCut(Item: TMenuItem; ShortCut: string);
begin
  Item.Caption := Item.Caption + #9 + Shortcut;
end;

initialization
  PendingAllOnTaskBarMode := AllOnTaskBarMode;
  PendingChildrenOnTop := ChildrenOnTop;
end.
