unit GraphI;
{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}
interface

{ Graph plotting for command line applications }

uses {$IFNDEF CONSOLE} ScrlCtrl,{$ENDIF}
     Types, Graphics, Classes, ExpCan, GStyle, Gmarks;

const
  MaxPens = 16;
  AxisPen = MaxPens;
  DefPolySize = 200;
  MaxPlot = 32767; { To avoid problems in Win95/8 in 32 bit programs, limit
                     graphics calls to 16 bit }

type

  TGraphOption = (grFixedCharSize, grPensHaveWidth, grColouredPens, grPensHaveStyle,
                  grIsotropic, grForceSymmetricalMarks, goThumbTracking, grNoAreaSelection,
                  grUnlimitedSegmentLength);

  TGraphOptions = set of TGraphOption;

  { TGraphI }

  TGraphI = class({$IFNDEF CONSOLE}TScrollableControl{$ELSE}TComponent{$ENDIF})
  private
    FBold: Boolean;
    FXmm: Double;
    FYmm: Double;
    FGOptions: TGraphOptions;
    FErrorBarSize: Single;

    CurrentMarkSize: Single;
    MarkLineStyle: TLineStyle;
    FMarkStyles: array [0..MaxPens-1] of TMarkStyle;

    FLineStyles: array [0..MaxPens] of TLineStyle;

    CurrentPenSize: Double;           { Current pen size as fraction of font size }
    PenSize: Integer;                 { Current Pen Size is device units }
    PenStyle: Integer;                { Current Pen Style }
    Pattern: array[0..MAXPATLENGTH-1] of Integer; { Current Pattern }
    PatternLength: Integer;           { Number of segments in pattern }
    PenStep: Integer;                 { where in current pattern? }
    SegLeft: Double;                  { length left in current part pattern }

    { Line Info }
    absX: Integer;        { Current pen position }
    absY: Integer;
    DRAW_FLAG: Boolean;   { Started drawing line flag }
    SETDOT: Boolean;      { Dot next at current Point }
    SYMBOL: Boolean;      { If set, plot a character at each Point }
    SymbolMark: Integer;  { Mark to plot at each Point }
    CurSeg: array of TPoint; { Holds current line segment; flushed by flush_polyline }
    NextSeg: Integer;

    procedure SetBold(const AValue: Boolean);

    function GetLineStyle: TLineStyle;
    procedure SetLineStyle(const AValue: TLineStyle);
    procedure SetLineType(ipen: Integer);

    procedure SetTextDeciAngle(const AValue: Integer);
    procedure SetTextSize(const AValue: Integer);
    procedure SetXmm(const Value: Double);
    procedure SetYmm(const Value: Double);
    procedure SetPensize(p:Double);
  protected
    FCurrentCanvas: TExportableCanvas;

    FDefaultMarkStyles: array [0..MaxPens-1] of TMarkStyle;
    FDefaultLineStyles: array [0..MaxPens] of TLineStyle;

    FXPixels, FYPixels: Integer;  { Current plot range in pixels }
    DPIX, DPIY: Integer;          { pixels per inch - may be out by a scale factor for metafiles }
    FXOffset, FYOffset: Integer;  { Location (in device co-ordinates) of top left of plot }
    YPerX:Double;            (* Multipy to convert an x displacement to a y displacement with the same physical distance *)

    FJustify: TAlignment;
    FTextDeciAngle:Integer; { rotation of chars/(10*degrees) }
    FTextSize: Integer;

    FontSize:Double;         { Base character size in current pixels; may be negative! }
    CurFontHeight: Integer;  { Height in screen or printer pixels }
    CurFont10Width: Integer; { 10 * average char width }
    CurFontLeading, CurFontInternalLeading, CurFontAscent: Integer;

    mp: TMarkParameters;    { Current settings for a mark }
    FLinetype: Integer;     { Current pen number } { TODO: Make private }
    CLineStyle: TLineStyle;

    procedure DoChangeFont; virtual;
    procedure DoOptionChanged(Sender: TObject);
    procedure HintDisplacement(Displacement:Integer; var v, h:Integer; AllowZero:Boolean);
    procedure PenInvalid;
    procedure NeedPen(AStyle: TLineStyle);
    procedure EndDraw;
    procedure FlushPolyline;
    procedure PreparePoints(var p: array of TPoint; np:Integer);
    procedure SetScaledLinestyle(ALinestyle:Integer; linescale:Double);
    procedure SetJustify(mode:TAlignment); virtual;
    procedure SetCurrentMarkSize(Size: Single);
    procedure SetCurrentCanvas(ACanvas: TExportableCanvas); virtual;
    procedure InitGraphics; virtual;
    procedure LoadMP;
    procedure SetGOptions(const AValue: TGraphOptions);
    function GetMarkStyle(i: Integer): TMarkStyle;
    procedure SetMarkStyle(i: Integer; const Value: TMarkStyle);
    function GetLineStyles(i:Integer):TLineStyle;
    procedure SetLineStyles(i:Integer; AStyle:TLineStyle);
    procedure CLineStyleChanged(Sender:TObject); virtual;
  public
    {$IFDEF CONSOLE}
    Width: Integer; { Used to give an estimate of pixel size }
    Font: TFont;
    {$ENDIF}

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Start; virtual;
    procedure Finish; virtual;

    procedure Setmm(x, y: Double);
    procedure LogToDev(var x, y: Integer); overload; { Convert from logical to device pixels; mainly sign conversion }
    procedure DevToLog(var x, y: Integer); overload; { Convert from device to logical pixels; mainly sign conversion }
    procedure LogToDev(var APoint: TPoint); overload;
    procedure DevToLog(var APoint: TPoint); overload;

    function IsRealDevice:Boolean; virtual;
    function IsPrinter:Boolean; virtual;
    function IsClipboard:Boolean; virtual;
    procedure StartPlotting; virtual; { All plotting calls should be within this pair of functions }
    procedure StopPlotting; virtual;
    procedure OptionChanged; virtual; { Options changed, but basic plot has not }
    procedure NeedLinePen;
    procedure SetToDefaultStyle(Pen: Integer; AStyle: TLineStyle);

    procedure MT(x:Integer; y:Integer);
    procedure LT(x:Integer; y:Integer);
    procedure Move(x:Integer; y:Integer);
    procedure Vector(x:Integer; y:Integer);
    procedure LineStyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
    procedure Box(AStyle: TBrushStyle; Density: Double; const Rect: TRect);
    procedure ArrowI(x0, y0, x1, y1:Integer; ArrowType: TArrowType);

    function AvCharWidth:Double;
    function CellHeight:Double; virtual;
    procedure TextI(x:Integer; y:Integer; str:string); virtual;
    function  TextExtent(const str:string):TSize; virtual;

    procedure Mark(x:Integer; y:Integer; Number:Integer);
    procedure MarkT(x:Integer; y:Integer; AStyle: TMarkStyle);
    procedure MarkE(x:Integer; y:Integer; AType: TMarkType);

    property CurrentCanvas:TExportableCanvas read FCurrentCanvas write SetCurrentCanvas;
    property Xmm:Double read FXmm write SetXmm;
    property Ymm:Double read FYmm write SetYmm;
    property XPixels:Integer read FXPixels; { Current pixel range }
    property YPixels:Integer read FYPixels;
    property GOptions:TGraphOptions read FGOptions write SetGOptions default [grForceSymmetricalMarks];
    property LineStyles[i:Integer]:TLineStyle read GetLineStyles write SetLineStyles;
    property MarkStyles[i:Integer]:TMarkStyle read GetMarkStyle write SetMarkStyle;
    property ErrorBarSize: Single read FErrorBarSize write FErrorBarSize;

    property LineStyle:TLineStyle read GetLineStyle write SetLineStyle;
    property LineType:Integer read FLinetype write SetLineType; { Pen number: -1 for custom }
    property Justify: TAlignment read FJustify write SetJustify;
    property TextDeciAngle:Integer read FTextDeciAngle write SetTextDeciAngle; { In 1/10ths of degrees }
    property TextSize:Integer { Text Size scale factor - % } read FTextSize write SetTextSize;
    property Bold: Boolean read FBold write SetBold;
  published
    {$IFNDEF CONSOLE}
    property Font;
    {$ENDIF}
  end;

const
  DefaultPenColour: array [0..Maxpens-1] of TColor = (
    clBlack,
    clMaroon,
    clGreen,
    clYellow,
    clNavy,
    clPurple,
    clTeal,
    clRed,
    clLime,
    clBlue,
    clFuchsia,
    clGray,
    clDkGray,
    clOlive,
    clAqua,
    clWhite
   );

function SmallX:Integer;
function SmallY:Integer;

implementation

uses {$IFNDEF CONSOLE} WCompat,{$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Math, Huge, UCompat;

function SmallX:Integer;
begin
  Result := {$IFDEF MSWINDOWS} GetSystemMetrics(SM_CXDOUBLECLK) {$ELSE} 3 {$ENDIF};
end;

function SmallY:Integer;
begin
  Result := {$IFDEF MSWINDOWS} GetSystemMetrics(SM_CYDOUBLECLK) {$ELSE} 3 {$ENDIF};
end;

{ TGraphI }

procedure TGraphI.LogToDev(var x, y: Integer); { Convert from logical to device pixels; }
begin
  y := YPixels - y;
  Inc(x, FXOffset);
  Inc(y, FYOffset);
end;

procedure TGraphI.LogToDev(var APoint: TPoint);
begin
  LogToDev(APoint.x, APoint.y);
end;

procedure TGraphI.DevToLog(var x, y: Integer); { Convert from device to logical pixels; }
begin
  Dec(x, FXOffset);
  Dec(y, FYOffset);
  y := YPixels - y
end;

procedure TGraphI.DevToLog(var APoint: TPoint);
begin
  DevToLog(APoint.x, APoint.y);
end;

function TGraphI.IsClipboard: Boolean;
begin
  Result := False;
end;

function TGraphI.IsPrinter: Boolean;
begin
  Result := True;
end;

function TGraphI.IsRealDevice: Boolean;
begin
  Result := False;
end;

procedure TGraphI.SetTextDeciAngle(const AValue: Integer);
begin
  if FTextDeciAngle = AValue then exit;
  FTextDeciAngle := AValue;
  DoChangeFont;
end;

procedure TGraphI.SetTextSize(const AValue: Integer);
begin
  if AValue <> TextSize then begin
    FTextSize := AValue;
    DoChangeFont;
  end;
end;

procedure TGraphI.SetCurrentCanvas(ACanvas: TExportableCanvas);
begin
  FreeAndNil(FCurrentCanvas);
  FCurrentCanvas := ACanvas;
  if not Assigned(ACanvas) then
    Exit;
  DPIX := ACanvas.DPI;
  DPIY := ACanvas.DPI;
  YPerX := 1;
  FXPixels := Round(Xmm * DPIX / 25.4);
  FYPixels := Round(Ymm * DPIY / 25.4);
  if grIsotropic in GOptions then begin
    { Shrink one axis to maintain Xmm/Ymm ratio }
    if XPixels * YPerX / YPixels > Xmm/Ymm then
      FXPixels := Round(YPixels * Xmm/Ymm / YPerX)
    else
      FYPixels := Round(XPixels * Ymm/Xmm * YPerX);
  end;
  FXOffset := 0;
  FYOffset := 0;
  FontSize := Font.Size*DPIY/72;
  FTextSize := 100;
  FBold := False;
end;

procedure TGraphI.SetGOptions(const AValue: TGraphOptions);
begin
  if AValue <> FGOptions then begin
    FGOptions := AValue;
    OptionChanged;
  end;
end;

procedure TGraphI.SetBold(const AValue: Boolean);
begin
  if FBold = AValue then exit;
  FBold := AValue;
  DoChangeFont;
end;

function TGraphI.GetLineStyle: TLineStyle;
begin
  Result := ClineStyle
end;

procedure TGraphI.SetLineStyle(const AValue: TLineStyle);
begin
  EndDraw;
  ClineStyle.Assign(AValue);
  FLinetype := -1;
end;

function TGraphI.GetLineStyles(i: Integer): TLineStyle;
begin
  Result := FLineStyles[i];
end;

procedure TGraphI.SetLineStyles(i: Integer; AStyle: TLineStyle);
begin
  FLineStyles[i].Assign(AStyle);
end;

procedure TGraphI.CLineStyleChanged(Sender: TObject);
begin
  EndDraw;
  FLinetype := -1; { Flag that the current line style no longer corresponds to one of the numbered styles }
end;

procedure TGraphI.SetLineType(ipen: Integer);
begin
  EndDraw;
  if (ipen > MAXPENS-1) or (ipen < 0) then ipen := ipen mod MAXPENS;
  if ipen < 0 then ipen := ipen + MAXPENS;
  ClineStyle.Assign(Linestyles[ipen]);
  FLinetype := ipen;
end;

procedure TGraphI.SetPensize(p: Double);
begin
  CurrentPenSize := p;
  if Font.Size = 0 then
    p := FontSize
  else
    p := Abs(p / Font.Size * FontSize);
  PenSize := Round(p);
  if not IsRealDevice and (p < Abs(FontSize) / 20) then
    HintDisplacement(Round(Abs(FontSize) / 20), mp.BlobV, mp.BlobH, False)
  else
    HintDisplacement(Round(p/2), mp.BlobV, mp.BlobH, IsRealDevice);
  if CurrentCanvas <> nil then begin
     with CurrentCanvas.Pen do begin
       Color := mp.rgbColour;
       Mode := pmCopy;
       Style := Graphics.psSolid;
       Width := PenSize;
     end;
     with CurrentCanvas.Brush do begin
       Color := mp.rgbColour;
       Style := bsSolid;
     end;
  end;
end;

procedure TGraphI.SetCurrentMarkSize(Size: Single);
var
  f: Double;
begin
  if CurrentMarkSize <> Size then begin
    CurrentMarkSize := Size;
    if Font.Size = 0 then
      f := Size
    else
      f := Size / Font.Size * FontSize;
    HintDisplacement(Round(Abs(f)), mp.v, mp.h, False);
  end;
end;

function TGraphI.GetMarkStyle(i: Integer): TMarkStyle;
begin
  Result := FMarkStyles[i];
end;

procedure TGraphI.SetMarkStyle(i: Integer; const Value: TMarkStyle);
begin
  FMarkStyles[i].Assign(Value);
end;

procedure TGraphI.PreparePoints(var p: array of TPoint; np: Integer);
var
  i: Integer;
begin
  for i := 0 to np-1 do begin
    LogToDev(p[i].x, p[i].y);
  end;
end;

procedure TGraphI.InitGraphics;
begin
  DRAW_FLAG := False;
  FLinetype := 0;
  LoadMP;
  DoChangeFont;
end;

procedure TGraphI.LoadMP;
begin
  mp.IsRealDevice := IsRealDevice;
  mp.ForceSymmetricalMarks := grForceSymmetricalMarks in GOptions;
end;

procedure TGraphI.DoChangeFont;
begin
  CurrentMarkSize := -1;
  if Assigned(CurrentCanvas) then
    CurrentCanvas.Font.Height := Round(FontSize*TextSize/100);
end;

procedure TGraphI.DoOptionChanged(Sender: TObject);
begin
  OptionChanged;
end;

procedure TGraphI.HintDisplacement(Displacement: Integer; var v, h: Integer; AllowZero: Boolean);
begin
  v := Abs(Displacement);
  if (v = 0) and (not AllowZero or IsPrinter) then v := 1;
  h := MulDiv(v, DPIX, DPIY);
end;

procedure TGraphI.PenInvalid;
begin
  CurrentPenSize := -1;
  PenStyle := MAXSTYLE+10;
end;

procedure TGraphI.NeedPen(AStyle: TLineStyle);
var
  newsize: Double;
  NewColour: TColor;
  NewPenStyle: Integer;
begin
  if grColouredPens in GOptions then
    NewColour := ColorToRGB(AStyle.Colour)
  else NewColour := 0;
  if grPensHaveWidth in GOptions then
    NewSize := AStyle.Width
  else NewSize := 0;
  if grPensHaveStyle in GOptions then
    NewPenStyle := Ord(AStyle.Pattern)
  else
    NewPenStyle := 0;
  if penstyle <> NewPenStyle then
    SetScaledLinestyle(NewPenStyle, abs(FontSize*2));
  if (NewColour <> mp.rgbColour) or
     (NewSize <> CurrentPenSize) then begin
    mp.rgbColour := NewColour;
    SetPensize(NewSize);
  end;
end;

procedure TGraphI.EndDraw;
begin
  if DRAW_FLAG and (nextseg > 1) then begin
    FlushPolyline;
    nextseg := 0;
  end;
  DRAW_FLAG := False;
  segleft := pattern[0];
  if penstep >=0 then penstep := 0;
end;

procedure TGraphI.FlushPolyline;
begin
  if nextseg > 1 then begin
    StartPlotting;
    NeedLinePen;
    CurrentCanvas.Polyline(CurSeg, NextSeg);
    StopPlotting;
  end;
end;

procedure TGraphI.SetScaledLinestyle(ALinestyle: Integer; linescale: Double);
{ select a line style/type and scale/pitch }
var
  i: Integer;
  p:TPat;
begin
  penstyle := ALinestyle;
  if (penstyle <= 0) or (penstyle >= MAXSTYLE) then begin
    { Simple solid line or dots at end of segments }
    SETDOT := ALinestyle = MAXSTYLE;
    if not SETDOT then penstyle := 0;
    SYMBOL := false;
    penstep := -1;
  end else begin
    p := pats[penstyle];
    PatternLength := p.length;
    linescale := linescale / p.scale;
    for i := 0 to PatternLength-1 do
      pattern[i] := Round(p.segments[i] * linescale);
    SETDOT := False;
    penstep := 0;
    segleft := p.segments[0];
  end;
end;

constructor TGraphI.Create(AOwner: TComponent);
var
  i: Integer;
begin
  inherited Create(AOwner);
  SetLength(CurSeg, DefPolySize);
  {$IFDEF CONSOLE}
  Font := TFont.Create;
  Font.Size := 8;
  {$ENDIF}
  SymbolMark := 3;
  FErrorBarSize := 2;
  FGOptions := [grForceSymmetricalMarks];
  FTextDeciAngle := 0;
  YPerX := 1;
  MarkLineStyle := TLineStyle.Create;
  MarkLineStyle.Width := 0;
  MarkLineStyle.Pattern := lpSolid;
  for i := 0 to MaxPens do begin
    FLineStyles[i] := TLineStyle.Create;
    FDefaultLineStyles[i] := TLineStyle.Create;
    if i < MaxPens then begin
      FMarkStyles[i] := TMarkStyle.Create;
      FDefaultMarkStyles[i] := TMarkStyle.Create;
    end
  end;
  CLineStyle := TLineStyle.Create;
  CLineStyle.OnChange := {$IFDEF FPC}@{$ENDIF}CLineStyleChanged;
  mp.PreparePoints := {$IFDEF FPC}@{$ENDIF}PreparePoints;
  for i := 0 to MaxPens-1 do begin
    with FLineStyles[i] do begin
      Width := 0;
      Colour := DefaultPenColour[i];
      Pattern := TLinePattern(i mod MAXSTYLE);
    end;
    FLineStyles[i].OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
    FDefaultLineStyles[i].Assign(LineStyles[i]);
    FMarkStyles[i].Colour := DefaultPenColour[i];
    FMarkStyles[i].MarkType := TMarkType(i mod nMarkTypes);
    FMarkStyles[i].DotAtCentre := True;
    FMarkStyles[i].OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
    FDefaultMarkStyles[i].Assign(FMarkStyles[i]);
  end;
  with FLineStyles[AxisPen] do begin
    Width := 0;
    Colour := clBlack;
    Pattern := lpSolid;
    OnChange := {$IFDEF FPC}@{$ENDIF}DoOptionChanged;
  end;
  PenStep := -1;
  DPIX := 72;
  DPIY := 72;
  FXmm := 250;
  FYmm := 170;
end;

destructor TGraphI.Destroy;
var
  I: Integer;
begin
  inherited Destroy;
  MarkLineStyle.Free;
  CLineStyle.Free;
  for i := 0 to MaxPens do begin
    FLineStyles[i].Free;
    FLineStyles[i] := nil;
    FDefaultLineStyles[i].Free;
    FDefaultLineStyles[i] := nil;
    if i < MaxPens then begin
      FMarkStyles[i].Free;
      FMarkStyles[i] := nil;
      FDefaultMarkStyles[i].Free;
      FDefaultMarkStyles[i] := nil;
    end;
  end;
  {$IFDEF CONSOLE}
  FreeAndNil(Font);
  {$ENDIF}
end;

procedure TGraphI.Start;
begin
  InitGraphics;
end;

procedure TGraphI.Finish;
begin

end;

procedure TGraphI.SetXmm(const Value: Double);
begin
  Setmm(Value, FYmm);
end;

procedure TGraphI.SetYmm(const Value: Double);
begin
  Setmm(FXmm, Value);
end;

procedure TGraphI.Setmm(x, y: Double);
begin
  if (x <> Xmm) or (y <> Ymm) then begin
    FXmm := x;
    FYmm := y;
    OptionChanged;
  end;
end;

procedure TGraphI.StartPlotting;
begin

end;

procedure TGraphI.StopPlotting;
begin

end;

procedure TGraphI.OptionChanged;
begin

end;

procedure TGraphI.NeedLinePen;
begin
  NeedPen(Linestyle);
end;

procedure TGraphI.SetToDefaultStyle(Pen: Integer; AStyle: TLineStyle);
begin
  AStyle.Assign(FDefaultLineStyles[Pen]);
end;

procedure TGraphI.MT(x: Integer; y: Integer);
begin
  FlushPolyline;
  NextSeg := 0;
  LT(x, y);
end;

procedure TGraphI.LT(x: Integer; y: Integer);
begin
  if nextseg > High(curseg) then
    if grUnlimitedSegmentLength in GOptions then
      SetLength(CurSeg, Length(CurSeg)+DefPolySize)
    else begin
      FlushPolyline;
      curseg[0] := curseg[nextseg-1];
      nextseg := 1;
    end;
  LogToDev(x,y);
  CurSeg[NextSeg].x := x;
  CurSeg[NextSeg].y := y;
  Inc(NextSeg);
end;

procedure TGraphI.Move(x: Integer; y: Integer);
begin
  EndDraw;
  absX := x;
  absY := y;
  if (SYMBOL) then begin
    StartPlotting;
    Mark(x, y, SymbolMark);
    StopPlotting;
  end else
    MT(absX, absY);
end;

procedure TGraphI.Vector(x: Integer; y: Integer);
var
  length, x1, y1, f:Double;
  procedure plot_blob(x:Integer; y:Integer);    { Plot a dot }
  begin
    Dot(CurrentCanvas, x, y, mp);
  end;
begin
  StartPlotting;
  NeedLinePen;
  if (SETDOT) then begin
    if not DRAW_FLAG then plot_blob(absX, absY);
    plot_blob(x, y);
  end else begin
    if penstep<0 then
      LT(x, y)
    else begin
      length := sqrt(1.*(x-absX)*(x-absX)+1.*(y-absY)*(y-absY));
      x1 := absX;
      y1 := absY;
      while length > 0 do
        if length > segleft then begin
          if segleft > 0 then begin
            f := segleft / length;
            x1 := x1 + f*(x-x1);
            y1 := y1 + f*(y-y1);
            if (penstep and 1) <> 0 then
              MT(Round(x1), Round(y1))
            else
              LT(Round(x1), Round(y1));
            length := length - segleft;
          end else if pattern[penstep] < 0 then
            plot_blob(absX, absY); (* First dot in line *)
          repeat    (* onto next segment of pattern *)
            Inc(penstep);
            if PenStep >= PatternLength then
              PenStep := 0;
            SegLeft := Pattern[PenStep];
            if SegLeft < 0 then
              plot_blob(Round(x1), Round(y1));
          until segleft > 0;
        end else begin
          if (PenStep and 1) <> 0 then
            MT(x, y)
          else
            LT(x, y);
          SegLeft := SegLeft - length;
          length := -1;
        end;
    end;
    if SYMBOL then Mark(x, y, SymbolMark);
  end;
  absX := x;
  absY := y;
  DRAW_FLAG := TRUE;
  StopPlotting;
end;

procedure TGraphI.LineStyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
begin
  EndDraw;
  ClineStyle.Colour := Colour;
  ClineStyle.Width := AWidth;
  ClineStyle.Pattern := APattern;
end;

procedure TGraphI.Box(AStyle: TBrushStyle; Density: Double; const Rect: TRect);
var
  p: TRect;
  function ScaleColour(Value: Integer): Integer;
  begin
    { Interpolate colour:
       - Density = 1 => full colour => value unchanged
       - Density = 0 => white => 255 }
    Result := Round((Value-255)*Density+255);
    { Sanity check - probably never happens }
    if Result < 0 then
      Result := 0
    else if Result > 255 then
      Result := 255;
  end;
begin
  StartPlotting;
  EndDraw;
  NeedLinePen;
  if (Density >= 0) and (Density < 1) then
    CurrentCanvas.Brush.Color := RGB(ScaleColour(GetRValue(mp.rgbColour)),
                                     ScaleColour(GetGValue(mp.rgbColour)),
                                     ScaleColour(GetBValue(mp.rgbColour)));
  p := Rect;
  LogToDev(p.Left, p.Top);
  LogToDev(p.Right, p.Bottom);
  if AStyle <> bsSolid then begin
    CurrentCanvas.Brush.Style := AStyle;
    {$IFDEF MSWINDOWS} {$IFNDEF CONSOLE}
    { TRANSPARENT seems to be ignored, so set sensible background colour }
    if CurrentCanvas.RealCanvas <> nil then
      SetBkColor(CurrentCanvas.RealCanvas.Handle, ColorToRGB(Color));
    {$ENDIF} {$ENDIF}
  end;
  CurrentCanvas.FillRect(p);
  { Restore Brush }
  CurrentCanvas.Brush.Color := mp.rgbColour;
  CurrentCanvas.Brush.Style := bsSolid;
  StopPlotting;
end;

procedure TGraphI.ArrowI(x0, y0, x1, y1: Integer; ArrowType: TArrowType);
begin
  StartPlotting;
  EndDraw;
  NeedLinePen;
  Arrow(CurrentCanvas,x0,y0,x1,y1,ArrowType,mp);
  StopPlotting;
end;

function TGraphI.AvCharWidth: Double;
begin
  Result := CellHeight; { This Sets the CurFont10Width and other variables }
  Result := Result * CurFont10Width / 10 /(CurFontHeight + CurFontLeading) / YPerX;
  {$IFDEF FIX_CHARACTER_BOX_SIZE_ON_ZOOM}
  if IsRealDevice then
    Result := Result / YSCALE * XSCALE
  {$ENDIF}
end;

function TGraphI.CellHeight: Double;
begin
  Result := Abs(FontSize*TextSize/100)*11/8;
  CurFontHeight := Round(Result);
  CurFontLeading := 0;
  CurFontInternalLeading := 0;
  CurFontAscent := CurFontHeight;
  CurFont10Width := Round(Result*5); { Guess }
end;

function TGraphI.TextExtent(const str: string): TSize;
begin
  { Ruugh estimate }
  Result.cx := Round(Length(Str)*AvCharWidth);
  Result.cy := Round(CellHeight);
end;

procedure TGraphI.TextI(x: Integer; y: Integer; str: string);
begin
  StartPlotting;
  EndDraw;
  CurrentCanvas.Brush.Style := bsClear;
  LogToDev(x,y);
  CurrentCanvas.RotatedTextOut(x, y, str, TextDeciAngle, FJustify);
  PenInvalid;
  StopPlotting;
end;

procedure TGraphI.Mark(x: Integer; y: Integer; Number: Integer);
begin
  Number := Number mod MaxPens;
  if Number < 0 then Inc(Number, MaxPens);
  MarkT(x, y, MarkStyles[Number]);
end;

procedure TGraphI.MarkT(x: Integer; y: Integer; AStyle: TMarkStyle);
begin
  StartPlotting;
  EndDraw;
  MarkLineStyle.Colour := AStyle.Colour;
  mp.DotAtCentre := AStyle.DotAtCentre;
  NeedPen(MarkLineStyle);
  SetCurrentMarkSize(AStyle.Size);
  Gmarks.Mark(CurrentCanvas, x, y, AStyle.MarkType, mp);
  StopPlotting;
end;

procedure TGraphI.MarkE(x: Integer; y: Integer; AType: TMarkType);
begin
  StartPlotting;
  EndDraw;
  NeedLinePen;
  SetCurrentMarkSize(ErrorBarSize);
  Gmarks.Mark(CurrentCanvas, x, y, AType, mp);
  StopPlotting;
end;

procedure TGraphI.SetJustify(mode: TAlignment);
begin
  FJustify := mode;
end;

end.
