unit GraphM;
{$I utils.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}
interface

{ Graph plotting for command line applications }

uses {$IFDEF VCL_} Types, {$ENDIF}
     Graphics, Classes, ExpCan, GraphI, GStyle, Gmarks, MathSup;

type

  RecordType = (tyEnd, tyMap, tyPspace, tyMapLin, tyPoint, tyJoin, tyClipPoint, tyClipJoin,
                tyTextJ, tyTitlen, tyAxis, tyScales, tyAxes, tyTextAngle, tySymbol, tyRectangle,
                tyPolyStart, tyPolyPoint, tyPolyEnd, tyStyleT, tySymbolT, tyErrorBarX, tyErrorBarY, tyNoYaxis,
                tyXaxLab, tyYaxLab, tyRectangleEx, tyDefScale, tyGraphic, tyArrow);

  { TGraphM }

  TGraphM = class(TGraphI)
  private
    FRecording: Boolean; { If recording calls }

    CMarkStyle: TMarkStyle;
    FTextStream: TStream;

    Xlast:Double;   (* Last argument to Point/Join *)
    Ylast:Double;

    FXaxPos, FYaxPos: TDPoint; { Last axis position }

    FPolygonPoints: array of TPoint;
    NPolygonPoints: Integer;
    XLastPoly, YLastPoly: Double; { Last argument to PolyPoint }

    function GetAxisStyle: TLineStyle;
    procedure SetAxisStyle(const AValue: TLineStyle);
    procedure SetRecording(const AValue: Boolean);
    procedure SetTextStream(const AValue: TStream);
  protected
    FXmax: Double; { User co-ordinate limits }
    FYmax: Double;
    FXmin: Double;
    FYmin: Double;
    Xp1: Double;     { Fraction of available area to use for user co-ordinate limits }
    Xp2: Double;
    Yp1: Double;
    Yp2: Double;
    Xof: Double;     { For conversion to TGraphI co-ordinates }
    Yof: Double;
    Xs: Double;
    Ys: Double;

    ClipFile: TextFile;

    RStream:TStream; { For recording calls }
    procedure PutReals(t:RecordType; d:array of Double);
    procedure PutType(t:RecordType);
    procedure PutString(s:string);
    procedure PlayBack(s:TStream);
    procedure InitGraphics; override;

    { Helpers }
    procedure Titlen(n:Integer; str:string);
    procedure RecomputeScale;
    procedure Clamp(var x1:Double; var y1:Double; AOutCode:Integer);
    procedure CLineStyleChanged(Sender:TObject); override;
  public

    property Recording:Boolean read FRecording write SetRecording;
    property TextStream: TStream read FTextStream write SetTextStream; { For recording values as text }

    property AxisStyle:TLineStyle read GetAxisStyle write SetAxisStyle;
    property Xmin: Double read FXmin;    { User co-ordinate limits }
    property Xmax: Double read FXmax;
    property Ymin: Double read FYmin;
    property Ymax: Double read FYmax;

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function ProcX(x:Double):Integer;
    function ProcY(y:Double):Integer;
    procedure DefaultScale;
    procedure Map(xl:Double; xu:Double; yl:Double; yu:Double);
    procedure MapLin(xl:Double; xu:Double; yl:Double; yu:Double);
    procedure MapArg(out xl, xu, yl, yu:Double);
    procedure Pspace(xl:Double; xu:Double; yl:Double; yu:Double);
    procedure PspaceArg(var xl, xu, yl, yu:Double);
    procedure NoYaxis;
    function PixelSize(x: Boolean): Double; virtual;

    procedure Axis(xl1:Double; xu1:Double; yl:Double; xaxis:Boolean; s:string);
    procedure Axes(xaxis:string; yaxis:string);
    procedure Scales(xaxis:string; yaxis:string);
    procedure ScaledAxis(xle, xue, xl1:Double; xu1:Double; yl:Double; xaxis:Boolean; s:string);
    procedure XaxLab(xaxis:string); { Must be called *after* call to axis/axes/scales }
    procedure YaxLab(yaxis:string);

    function  OutCode(x:Double; y:Double):Integer;
    procedure ClpPnt(var x1:Double; var y1:Double; x2:Double; y2:Double; var AOutCode:Integer);
    procedure ClipPoint(x:Double; y:Double);
    procedure ClipJoin(x1:Double; y1:Double);
    procedure Point(x:Double; y:Double);
    procedure Join(x:Double; y:Double);
    procedure Symbol(x:Double; y:Double; n:Integer);
    procedure SymbolT(x:Double; y:Double; AStyle: TMarkStyle);
    procedure ErrorBarX(x1, x2, y:Double);
    procedure ErrorBarY(x, y1, y2:Double);
    procedure Arrow(x0, y0, x1, y1:Double; ArrowType: TArrowType);

    procedure Rectangle(x1, x2, y1, y2:Double);
    procedure RectangleEx(AStyle: TBrushStyle; Density: Double; x1, x2, y1, y2:Double);
    procedure PolyStart;
    procedure PolyPoint(x1:Double; y1:Double);
    procedure PolyEnd;
    procedure DrawGraphic(x1, x2, y1, y2:Double; Graphic: TGraphic); virtual;

    { Character sizes; can be approximate }
    procedure Chlen(out Acx:Double; out Acy:Double);
    function  Cx: Double;
    function  Cy: Double;

    procedure Style(AStyle:Integer);
    procedure StyleT(AStyle:TLineStyle);
    procedure StyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);

    procedure TextAngle(ang:Double);
    procedure Text(x:Double; y:Double; s:string);
    procedure TextJ(x:Double; y:Double; mode:TAlignment; s:string);
    procedure Title1(title:string);
    procedure Title2(title:string);

   end;

procedure AxisInterval(range:Double; var nticks:Integer; out ds:Double; out se:Double; out ne:Integer; out nd:Integer);

implementation

uses SysUtils, Huge, TextAssign, Math;

{ TGraphM }

constructor TGraphM.Create(AOwner: TComponent);
begin
  Xp2 := 1;
  Yp2 := 1;
  inherited Create(AOwner);
  CMarkStyle := TMarkStyle.Create;
end;

destructor TGraphM.Destroy;
begin
  Recording := False;
  FreeAndNil(RStream);
  TextStream := nil;
  CMarkStyle.Free;
  inherited Destroy;
end;

function TGraphM.ProcX(x: Double): Integer;
begin
  x := (x-Xmin)*Xs + Xof;
  if abs(x) > MaxPlot then
    if x < 0 then
      Result := -MaxPlot
    else
      Result := MaxPlot
  else
    Result := Integer(Round(x));
end;

function TGraphM.ProcY(y: Double): Integer;
begin
  y := (y-Ymin)*Ys + Yof;
  if abs(y) > MaxPlot then
    if y < 0 then
      Result := -MaxPlot
    else
      Result := MaxPlot
  else
    Result := Integer(Round(y));
end;

procedure TGraphM.DefaultScale;
var
  xc: Double;
begin
  if Recording then
    PutType(tyDefScale);
  FXmin := 0;
  FYmin := 0;
  FXmax := 1;
  FYmax := 1;
  xc := AvCharWidth/XPixels;
  Xp2 := 1-3.5*xc;
  Xp1 := xc*8;
  if Xp1 >= Xp2 then begin
    { Plot too small, or font too large }
    Xp1 := 0.1;
    Xp2 := 0.9;
  end;
  Yp1 := 2*CellHeight/YPixels;
  Yp2 := 1-Yp1;
  if Yp1 >= Yp2 then begin
    { Plot too small, or font too large }
    Yp1 := 0.1;
    Yp2 := 0.9;
  end;
  RecomputeScale;
end;

procedure TGraphM.RecomputeScale;
begin
  Xof := Xp1*Xpixels;
  Yof := Yp1*Ypixels;
  try
    if Xmax <> Xmin then
      Xs := (Xp2-Xp1)*Xpixels/(Xmax-Xmin);
  except
    Xs := 1;
  end;
  try
    if Ymax <> Ymin then
      Ys := (Yp2-Yp1)*Ypixels/(Ymax-Ymin);
  except
    Ys := 1;
  end;
end;

procedure TGraphM.Map(xl: Double; xu: Double; yl: Double; yu: Double);
begin
  if Recording then
    PutReals(tyMap, [xl, xu, yl, yu]);
  FXmin := xl;
  FYmin := yl;
  FXmax := xu;
  FYmax := yu;
  RecomputeScale;
end;

procedure TGraphM.MapLin(xl: Double; xu: Double; yl: Double; yu: Double);
var
  t, fac, dy, dx: Double;
begin
  if Recording then
    PutReals(tyMapLin, [xl, xu, yl, yu]);
  { MapLin ensures that at least the range of x and y specified is
    displayed and sets map such that a displacement in an arbitrary
    direction will cause the same movement to be displayed. }
  if (Xp2 = Xp1) or (Yp2 = Yp1) then Exit;
  t := abs( (Yp2 - Yp1) * YPixels / YPerX
            / (Xp2 - Xp1) / XPixels );
  if xu <> xl then
    fac := abs( (yu-yl) / (xu-xl) )
  else
    fac := t+1; { Force too much y }
  if fac > t then begin
    { too much y - extend x }
    dx := (abs(yu-yl)/t - abs(xu-xl))/2;
    if xl > xu then dx := -dx;
    xl := xl - dx;
    xu := xu + dx;
  end else if fac < t then begin
    { too much x - extend y }
    dy := (t*abs(xu-xl) - abs(yu-yl))/2;
    if yl > yu then dy := -dy;
    yl := yl - dy;
    yu := yu + dy;
  end;
  FXmin := xl;
  FYmin := yl;
  FXmax := xu;
  FYmax := yu;
  RecomputeScale;
end;

procedure TGraphM.MapArg(out xl, xu, yl, yu: Double);
begin
  xl := Xmin;
  yl := Ymin;
  xu := Xmax;
  yu := Ymax;
end;

procedure TGraphM.Pspace(xl: Double; xu: Double; yl: Double; yu: Double);
begin
  if Recording then
    PutReals(tyPspace, [xl, xu, yl, yu]);
  if xl <> xu then begin
    Xp1 := xl;
    Xp2 := xu;
  end else
    DefaultScale;
  if yl <> yu then begin
    Yp1 := yl;
    Yp2 := yu;
  end else
    DefaultScale;
  RecomputeScale;
end;

procedure TGraphM.PspaceArg(var xl, xu, yl, yu: Double);
begin
  xl := Xp1;
  yl := Yp1;
  xu := Xp2;
  yu := Yp2;
end;

procedure TGraphM.NoYaxis;
begin
  if Recording then
    PutType(tyNoYaxis);
  Xp1 := 1 - Xp2;
  RecomputeScale;
end;

function TGraphM.PixelSize(x: Boolean): Double;
begin
  Result := 1;
  if x then begin
    if Xs <> 0 then Result := XPixels / Xs;
  end else begin
    if Ys <> 0 then Result := YPixels / Ys;
  end;
  Result := Abs(Result);
end;

procedure TGraphM.SetRecording(const AValue: Boolean);
begin
  if FRecording=AValue then exit;
  FRecording:=AValue;
  if not Assigned(RStream) then
    RStream := THugeMemoryStream.Create;
  if FRecording then
    RStream.Position := 0;
end;

procedure TGraphM.SetTextStream(const AValue: TStream);
begin
  if Assigned(TextStream) then
      Close(ClipFile);
  FTextStream := AValue;
  if Assigned(TextStream) then begin
    AssignStream(ClipFile,TextStream);
    ReWrite(ClipFile);
  end;
end;

function TGraphM.GetAxisStyle: TLineStyle;
begin
  Result := LineStyles[AxisPen];
end;

procedure TGraphM.SetAxisStyle(const AValue: TLineStyle);
begin
  LineStyles[AxisPen] := AValue;
end;

procedure TGraphM.PutReals(t: RecordType; d: array of Double);
begin
  PutType(t);
  RStream.WriteBuffer(d, SizeOf(d));
end;

procedure TGraphM.PutType(t: RecordType);
begin
  RStream.WriteBuffer(t, SizeOf(t));
end;

procedure TGraphM.PutString(s: string);
var
  b: Byte;
begin
  if Length(s) > 255 then
    b := 255
  else
    b := Length(s);
  RStream.WriteBuffer(b, 1);
  if b > 0 then
    RStream.WriteBuffer(s[1], b*SizeOf(Char));
end;

procedure TGraphM.PlayBack(s: TStream);
var
  t: RecordType;
  d:array [0..4] of Double;
  Graphic: TGraphic;
  mode: TAlignment;
  buffer, buffer2:string;
  n:Integer;
  xaxis:Boolean;
  AWidth: Single;
  Colour: TColor;
  APattern: TLinePattern;
  Size: Single;
  MarkType: TMarkType;
  DotAtCentre: Boolean;
  BrushStyle: TBrushStyle;
  ArrowType: TArrowType;
  procedure ReadString(var str:string);
  var
    b:Byte;
  begin
    s.ReadBuffer(b, 1);
    SetLength(str, b);
    if b > 0 then s.ReadBuffer(str[1], b*SizeOf(Char));
  end;
begin
   if s = nil then Exit;
   s.Position := 0;
   while (s.Read(t, SizeOf(t))=SizeOf(t)) and (t <> tyEnd ) do begin
    case t of
      tyMap:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          Map(d[0], d[1], d[2], d[3]);
        end;
      tyPspace:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          Pspace(d[0], d[1], d[2], d[3]);
        end;
      tyMapLin:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          MapLin(d[0], d[1], d[2], d[3]);
        end;
      tyDefScale: DefaultScale;
      tyPoint:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          Point(d[0], d[1]);
        end;
      tySymbol:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          s.ReadBuffer(n, SizeOf(Integer));
          Symbol(d[0], d[1], n);
        end;
      tyJoin:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          Join(d[0], d[1]);
        end;
      tyClipPoint:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          ClipPoint(d[0], d[1]);
        end;
      tyClipJoin:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          ClipJoin(d[0], d[1]);
        end;
      tyTextJ:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          s.ReadBuffer(mode, SizeOf(TAlignment));
          ReadString(Buffer);
          Justify := mode;
          TextI(ProcX(d[0]), ProcY(d[1]), buffer);
        end;
      tyTitlen:
        begin
          s.ReadBuffer(n, SizeOf(Integer));
          ReadString(Buffer);
          Titlen(n, buffer);
        end;
      tyAxis:
        begin
          s.ReadBuffer(d, 3*SizeOf(Double));
          s.ReadBuffer(xaxis, SizeOf(Boolean));
          ReadString(Buffer);
          Axis(d[0], d[1], d[2], xaxis, buffer);
        end;
      tyScales:
        begin
          ReadString(Buffer);
          ReadString(Buffer2);
          Scales(buffer, buffer2);
        end;
      tyXaxLab:
        begin
          ReadString(Buffer);
          XaxLab(buffer);
        end;
      tyYaxLab:
        begin
          ReadString(Buffer);
          YaxLab(buffer);
        end;
      tyAxes:
        begin
          ReadString(Buffer);
          ReadString(Buffer2);
          Axes(buffer, buffer2);
        end;
      tyTextAngle:
        begin
          s.ReadBuffer(d, SizeOf(Double));
          TextAngle(d[0]);
        end;
      tyRectangleEx:
        begin
          s.ReadBuffer(d, 5*SizeOf(Double));
          s.ReadBuffer(BrushStyle, SizeOf(BrushStyle));
          RectangleEx(BrushStyle, d[0], d[1], d[2], d[3], d[4]);
        end;
      tyRectangle:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          Rectangle(d[0], d[1], d[2], d[3]);
        end;
      tyGraphic:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          s.ReadBuffer(Graphic, Sizeof(Graphic));
          DrawGraphic(d[0], d[1], d[2], d[3], Graphic);
        end;
      tyPolyStart: PolyStart;
      tyPolyEnd:   PolyEnd;
      tyPolyPoint:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          PolyPoint(d[0], d[1]);
        end;
      tyStyleT:
        begin
          s.ReadBuffer(AWidth, SizeOf(AWidth));
          s.ReadBuffer(Colour, SizeOf(Colour));
          s.ReadBuffer(APattern, SizeOf(APattern));
          StyleP(Colour, AWidth, APattern);
        end;
      tySymbolT:
        begin
          s.ReadBuffer(d, 2*SizeOf(Double));
          s.ReadBuffer(Size, SizeOf(Size));
          s.ReadBuffer(Colour, SizeOf(Colour));
          s.ReadBuffer(MarkType, SizeOf(MarkType));
          s.ReadBuffer(DotAtCentre, SizeOf(DotAtCentre));
          CMarkStyle.Size := Size;
          CMarkStyle.Colour := Colour;
          CMarkStyle.MarkType := MarkType;
          CMarkStyle.DotAtCentre := DotAtCentre;
          SymbolT(d[0], d[1], CMarkStyle);
        end;
      tyErrorBarX:
        begin
          s.ReadBuffer(d, 3*SizeOf(Double));
          ErrorBarX(d[0], d[1], d[2]);
        end;
      tyErrorBarY:
        begin
          s.ReadBuffer(d, 3*SizeOf(Double));
          ErrorBarY(d[0], d[1], d[2]);
        end;
      tyNoYaxis: NoYaxis;
      tyArrow:
        begin
          s.ReadBuffer(d, 4*SizeOf(Double));
          s.ReadBuffer(ArrowType,SizeOf(ArrowType));
          Arrow(d[0], d[1], d[2], d[3], ArrowType);
        end;
    else
      raise Exception.Createfmt('TGraphD:Bad Record %d at %d', [Ord(t), s.Position]);
    end;
   end;
end;

procedure TGraphM.InitGraphics;
begin
  inherited;
  DefaultScale;
end;

function AxPos(a:Double; b:Double):Double;
begin
  if ((a > 0) and (b > 0)) or ((a < 0) and (b < 0)) then begin
    Result := a;
    if b < a then Result := b;
  end else
    Result := 0;
end;

procedure TGraphM.Axis(
{ Draw an x or y axis. For an x axis the parameters are: }
    xl1:Double; { The start of the axis }
    xu1:Double; { The end of the axis }
    yl:Double;  { The y co-ordinate of the axis }
    xaxis:Boolean; { false for a y axis }
    s:string    { The label }
);
{ For a y axis, swap x and y above and xaxis := false }
var
  cxl, cyl, xl, xu, ds, se, t, xc: Double;
  nticks, ne, nd: Integer;
  tt:Int64;
  TickType: TMarkType;
  SavedStyleNumber: Integer;
  SavedLineStyle: TLineStyle;
  SavedTextStream: TStream; { Current text stream file }
  Sample: string;
begin
  if IsNanOrInf(xl1) or IsNanOrInf(xu1) or IsNanOrInf(yl) then
    Exit;
  if Recording then begin
    PutReals(tyAxis, [xl1, xu1, yl]);
    RStream.WriteBuffer(xaxis, SizeOf(xaxis));
    PutString(s);
    Exit;
  end;
  SavedStyleNumber := FLinetype;
  SavedLineStyle := nil;
  if SavedStyleNumber < 0 then begin
    SavedLineStyle := TLineStyle.Create;
    SavedLineStyle.Assign(LineStyle);
  end;
  SavedTextStream := TextStream;
  FTextStream := nil; { Turn off Text stream for axis }
  try
    LineStyle := AxisStyle;
    (* Determine steps for annotation *)
    Chlen(cxl, cyl);
    xl := xl1;
    xu := xu1;
    if xu < xl then begin
      xl := xu1;
      xu := xl1;
    end;
    (* Draw Axis line *)
    (* Nticks = Approximate number of major ticks wanted *)
    nticks := 1;
    if xaxis then begin
      Point(xl, yl);
      Join(xu, yl);
      if cxl <> 0 then begin
        nticks := Round((xu-xl)/cxl/8);
        repeat
          nticks := nticks div 2;
          AxisInterval(xu-xl, nticks, ds, se, ne, nd);
          if nticks <= 2 then
            Break;
          Sample := Format('%.*f', [nd, Max(Abs(xu),Abs(xl))/se]);
          t := ds*se*5/cxl; { Approimate characters per major label }
        until Length(Sample) < t*0.75;
      end;
    end else begin
      Point(yl, xl);
      Join(yl, xu);
      if cyl > 0 then
        nticks := Round((xu-xl)/cyl/8);
      AxisInterval(xu-xl, nticks, ds, se, ne, nd);
    end;
    SetCurrentMarkSize(Font.Size / 3);
    (* ds = Minor tick interval *)
    (* Convert to non-exponential range *)
    xl := xl/se;
    xu := xu/se;
    { Give up if range too small }
    if Max(Abs(xl),Abs(xu))*1e-15 > Abs(ds) then Exit;
    xc := Trunc(xl/ds)*ds;
    if abs(xc-ds-xl) < ds/1000 then
      xc := xc - ds;      { Allow for point just outside range }
    (* draw ticks *)
    while xc <= (xu + ds/1000) do begin
        if xc >= xl then begin
            t := xc*se;
            tt := Round(abs(xc)/ds);
            if tt mod 5 = 0 then
              { A major (labelled) tick mark }
              if xaxis then begin
                TextJ(xc*se, yl-cyl, taCenter, Format('%.*f', [nd, xc]));
                TickType := mtXMajor;
              end else begin
                TextJ(yl-cxl/6, xc*se-cyl/3, taRightJustify, Format('%.*f', [nd, xc]));
                TickType := mtYMajor;
              end
            else
              if xaxis then
                TickType := mtXMinor
              else
                TickType := mtYMinor;
            if xaxis then
              Gmarks.Mark(CurrentCanvas, ProcX(t), ProcY(yl), TickType, mp)
            else
              Gmarks.Mark(CurrentCanvas, ProcX(yl), ProcY(t), TickType, mp);
        end;
        xc := xc + ds;
    end;
    (* write * 10 ^ ? *)
    if ne <> 0 then begin
       if xaxis then
        TextJ(xu1+3.5*cxl, yl-2*cyl, taRightJustify, Format('/1e%d', [ne]))
       else begin
        TextAngle(90);
        TextJ(yl-4.49*cxl, xu1, taRightJustify, Format('/1e%d', [ne]));
        TextAngle(0);
       end;
    end;
    if xaxis then begin
      FXaxPos.X := (xl1+xu1)/2;
      FXaxPos.Y := yl-1.9*cyl;
      XaxLab(s);
    end else begin
      FYaxPos.X := yl-5.99*cxl;
      FYaxPos.Y := (xl1+xu1)/2;
      YaxLab(s);
    end;
  finally
    FTextStream := SavedTextStream;
    if SavedStyleNumber < 0 then begin
      LineStyle := SavedLineStyle;
      SavedLineStyle.Free;
    end else
      LineType := SavedStyleNumber;
  end;
end;

procedure TGraphM.Arrow(x0, y0, x1, y1: Double; ArrowType: TArrowType);
begin
  if Recording then begin
    PutReals(tyArrow, [x0, y0, x1, y1]);
    RStream.WriteBuffer(ArrowType, SizeOf(ArrowType));
  end else
    ArrowI(ProcX(x0), ProcY(y0), ProcX(x1), ProcY(y1), ArrowType);
end;

procedure TGraphM.Axes(xaxis: string; yaxis: string);
begin
  if Recording then begin
    PutType(tyAxes);
    PutString(xaxis);
    PutString(yaxis);
  end else begin
    Axis(Xmin, Xmax, AxPos(Ymin, Ymax), true, xaxis);
    Axis(Ymin, Ymax, AxPos(Xmin, Xmax), false, yaxis);
  end;
end;

procedure TGraphM.Scales(xaxis: string; yaxis: string);
begin
  if Recording then begin
    PutType(tyScales);
    PutString(xaxis);
    PutString(yaxis);
  end else begin
    Axis(Xmin, Xmax, Ymin, true, xaxis);
    Axis(Ymin, Ymax, Xmin, false, yaxis);
  end;
end;

procedure TGraphM.ScaledAxis(xle, xue, xl1: Double; xu1: Double; yl: Double;
  xaxis: Boolean; s: string);
var
  xi, xx, yi, yx: Double;
begin
  if xl1 = xu1 then
    Exit;
  MapArg(xi, xx, yi, yx);
  try
    if xaxis then
      Map(xle+(xue-xle)*(xi-xl1)/(xu1-xl1), xue+(xue-xle)*(xx-xu1)/(xu1-xl1), yi, yx)
    else
      Map(xi, xx, xle+(xue-xle)*(yi-xl1)/(xu1-xl1), xue+(xue-xle)*(yx-xu1)/(xu1-xl1));
    Axis(xle, xue, yl, xaxis, s);
  finally
    Map(xi, xx, yi, yx);
  end;
end;

procedure TGraphM.XaxLab(xaxis: string);
begin
  if Length(xaxis) = 0 then Exit;
  if Recording then begin
    PutType(tyXaxLab);
    PutString(xaxis);
  end else
    TextJ(FXaxPos.X, FXaxPos.Y, taCenter, xaxis);
end;

procedure TGraphM.YaxLab(yaxis: string);
begin
  if Length(yaxis) = 0 then Exit;
  if Recording then begin
    PutType(tyYaxLab);
    PutString(yaxis);
  end else begin
    TextAngle(90);
    TextJ(FYaxPos.X, FYaxPos.Y, taCenter, yaxis);
    TextAngle(0);
  end;
end;

procedure TGraphM.Clamp(var x1: Double; var y1: Double; AOutCode: Integer);
begin
  { Internal routine to force point to boundary }
  if (AOutCode and 1) <> 0 then
      x1 := Xmin
  else if (AOutCode and 4) <> 0 then
      x1 := Xmax;
  if (AOutCode and 2) <> 0 then
      y1 := Ymax
  else if (AOutCode and 8) <> 0 then
      y1 := Ymin;
end;

procedure TGraphM.ClpPnt(var x1: Double; var y1: Double; x2: Double; y2: Double; var AOutCode: Integer);
begin
    { Internal routine to clip start point of line (x1, y1)
      NB: no need to test for divide by zero here;
      eliminated by pre-testing }
    if (AOutCode and 5) <> 0 then
      if ((AOutCode and 1) <> 0) xor (Xmin > Xmax) then begin
          y1 := y1 + (y2-y1)*(Xmin-x1)/(x2-x1);
          x1 := Xmin;
      end else { if (out and 4) <> 0 then } begin
          y1 := y1 + (y2-y1)*(Xmax-x1)/(x2-x1);
          x1 := Xmax;
      end
    else if (AOutCode and 10) <> 0 then
      if ((AOutCode and 2) <> 0) xor (Ymin > Ymax) then begin
          x1 := x1 + (x2-x1)*(Ymax-y1)/(y2-y1);
          y1 := Ymax;
      end else { if (out and 8) <> 0 then } begin
          x1 := x1 + (x2-x1)*(Ymin-y1)/(y2-y1);
          y1 := Ymin;
      end
    else
      Exit;
    AOutCode := OutCode(x1, y1);
    { Note that only one clip is done; may need to repeat }
end;

function TGraphM.OutCode(x: Double; y: Double): Integer;
begin
    (* Return bitmap indicating where x, y is with respect
       to the region specified by map:
       Result has bits set as follows:
       Bit 0: (1) to left
           1: (2) above
           2: (4) to right
           3: (8) below
     *)
    Result := 0;
    if Xmin < Xmax then begin
        if x < Xmin then Result := 1;
        if x > Xmax then Result := Result + 4;
    end else begin
        if x < Xmax then Result := 1;
        if x > Xmin then Result := Result + 4;
    end;
    if Ymin < Ymax then begin
        if y < Ymin then Result := Result + 8;
        if y > Ymax then Result := Result + 2;
    end else begin
        if y < Ymax then Result := Result + 8;
        if y > Ymin then Result := Result + 2;
    end;
end;

procedure TGraphM.ClipPoint(x: Double; y: Double);
begin
  if Recording then
    PutReals(tyClipPoint, [x, y])
  else begin
    Xlast := x;  Ylast := y;
    if OutCode(x,y) = 0 then
      Point(x, y);
  end;
end;

procedure TGraphM.ClipJoin(x1: Double; y1: Double);
var
  x, y: Double;
  orgout1, out1, out2: Integer;
begin
  if Recording then begin
    PutReals(tyClipJoin, [x1, y1]);
    Exit;
  end;
  { Join previous Point to x, y; clips line to edge of map area
     Clip adapted from R Wilton, Programmers Guide to PC and PS/2 Video Systems,  p 217 }
  if IsNanOrInf(XLast) or IsNanOrInf(YLast) then
    ClipPoint(x1, y1)
  else if IsNanOrInf(x1) or IsNanOrInf(y1) then begin
    SetNan(XLast);
    Exit;
  end else begin
    x := x1;
    y := y1;
    orgout1 := OutCode(Xlast, Ylast);
    out1 := orgout1;
    out2 := OutCode(x, y);
    while (out1 and out2) = 0 do begin
      { Some of line visible }
      if (out1 or out2) <> 0 then begin
          { Clipping Required }
          if out1 <> 0 then begin
            clppnt(Xlast, Ylast, x, y, out1);
            Continue;
          end;
          if out2 <> 0 then clppnt(x, y, Xlast, Ylast, out2);
          Continue;
      end;
      if orgout1 <> 0 then Point(Xlast, Ylast);
      Join(x, y);
      Break;
    end;
    Xlast := x1;
    Ylast := y1;
  end;
end;

procedure TGraphM.Chlen(out Acx: Double; out Acy: Double);
begin
  Acy := cy;
  Acx := cx;
end;

function TGraphM.Cx: Double;
begin
  if Ys = 0 then
    Result := 0
  else
    Result := AvCharWidth / Xs;
end;

function TGraphM.Cy: Double;
begin
  if Ys = 0 then
    Result := 0
  else
    Result := CellHeight / Ys;
end;

procedure TGraphM.Point(x: Double; y: Double);
begin
  if Assigned(TextStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y);
  end;
  if Recording then
    PutReals(tyPoint, [x, y])
  else begin
    Xlast := x;  Ylast := y;
    Move(ProcX(x), ProcY(y));
  end;
end;

procedure TGraphM.Join(x: Double; y: Double);
begin
  if Assigned(TextStream) then
    WriteLn(ClipFile, x, #9, y);
  if Recording then
    PutReals(tyJoin, [x, y])
  else begin
    Xlast := x;  Ylast := y;
    Vector(ProcX(x), ProcY(y));
  end;
end;

procedure TGraphM.Symbol(x: Double; y: Double; n: Integer);
begin
  if Assigned(TextStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y, #9, n);
  end;
  if Recording then begin
    PutReals(tySymbol, [x, y]);
    RStream.WriteBuffer(n, SizeOf(n));
  end else if OutCode(x,y) = 0 then
    Mark(ProcX(x), ProcY(y), n);
end;

procedure TGraphM.SymbolT(x: Double; y: Double; AStyle: TMarkStyle);
var
  Size: Single;
  Colour: TColor;
  MarkType: TMarkType;
  DotAtCentre: Boolean;
begin
  if Assigned(TextStream) then begin
    WriteLn(ClipFile);
    WriteLn(ClipFile, x, #9, y, #9, AStyle.Colour, #9, Ord(AStyle.MarkType)+100*Ord(AStyle.DotAtCentre), #9, AStyle.Size);
  end;
  if Recording then begin
    PutReals(tySymbolT, [x, y]);
    Size := AStyle.Size;
    Colour := AStyle.Colour;
    MarkType := AStyle.MarkType;
    DotAtCentre := AStyle.DotAtCentre;
    RStream.WriteBuffer(Size, SizeOf(Size));
    RStream.WriteBuffer(Colour, SizeOf(Colour));
    RStream.WriteBuffer(MarkType, SizeOf(MarkType));
    RStream.WriteBuffer(DotAtCentre, SizeOf(DotAtCentre));
  end else if OutCode(x,y) = 0 then
    MarkT(ProcX(x), ProcY(y), AStyle);
end;

procedure TGraphM.ErrorBarX(x1, x2, y: Double);
begin
  if Recording then
    PutReals(tyErrorBarX, [x1, x2, y])
  else begin
    ClipPoint(x1, y);
    ClipJoin(x2, y);
    MarkE(ProcX(x1), ProcY(y), mtVBar);
    MarkE(ProcX(x2), ProcY(y), mtVBar);
  end;
end;

procedure TGraphM.ErrorBarY(x, y1, y2: Double);
begin
  if Recording then
    PutReals(tyErrorBarY, [x, y1, y2])
  else begin
    ClipPoint(x, y1);
    ClipJoin(x, y2);
    MarkE(ProcX(x), ProcY(y1), mtHBar);
    MarkE(ProcX(x), ProcY(y2), mtHBar);
  end;
end;

procedure TGraphM.Rectangle(x1, x2, y1, y2: Double);
begin
  if Recording then
    PutReals(tyRectangle, [x1, x2, y1, y2])
  else
    RectangleEx(bsSolid, -1, x1, x2, y1, y2);
end;

procedure TGraphM.RectangleEx(AStyle: TBrushStyle; Density: Double; x1, x2, y1, y2: Double);
var
  p: Array [0..1] of TPoint;
  function Check(var v1, v2:Double; vmin, vmax:Double):Boolean;
  var
    t:Double;
  begin
    if (v1 < v2) <> (vmin < vmax) then begin
      t := v1;
      v1 := v2;
      v2 := t;
    end;
    if vmin < vmax then
      Result := (v2 >= vmin) and (v1 <= vmax)
    else
      Result := (v2 <= vmin) and (v1 >= vmax);
  end;
begin
  if Recording then begin
    PutReals(tyRectangleEx, [Density, x1, x2, y1, y2]);
    RStream.WriteBuffer(AStyle, SizeOf(AStyle));
  end else if Check(x1, x2, Xmin, Xmax) and Check(y1, y2, Ymin, Ymax) then begin
    p[0].x := ProcX(x1);
    p[1].x := ProcX(x2);
    p[0].y := ProcY(y1);
    p[1].y := ProcY(y2);
    Box(AStyle, Density, TRect(p));
  end;
end;

procedure TGraphM.PolyStart;
begin
  if Assigned(TextStream) then
    WriteLn(ClipFile);
  if Recording then
    PutType(tyPolyStart)
  else begin
    if NPolygonPoints <> 0 then
      raise Exception.CreateFmt('PolyStart with %d points unshown', [NPolygonPoints]);
  end;
end;

procedure TGraphM.PolyPoint(x1: Double; y1: Double);
var
  x, y: Double;
  orgout1, out1, out2: Integer;
  procedure AddPoint(x, y: Double);
  var
    ix, iy: Integer;
  begin
    ix := ProcX(x);
    iy := ProcY(y);
    if (NPolygonPoints = 0)
       or (ix <> FPolygonPoints[NPolygonPoints-1].x)
       or (iy <> FPolygonPoints[NPolygonPoints-1].y) then begin
      LogToDev(ix, iy); { TODO : Shift this into TGraphB }
      if High(FPolygonPoints) <= NPolygonPoints then
        SetLength(FPolygonPoints, NPolygonPoints+300);
      FPolygonPoints[NPolygonPoints].x := ix;
      FPolygonPoints[NPolygonPoints].y := iy;
      Inc(NPolygonPoints);
    end;
  end;
begin
  if Assigned(TextStream) then
    WriteLn(ClipFile, x1, #9, y1);
  if Recording then begin
    PutReals(tyPolyPoint, [x1, y1]);
    Exit;
  end;
  x := x1;
  y := y1;
  out2 := OutCode(x1, y1);
  if NPolygonPoints > 0 then begin
    out1 := OutCode(XLastPoly, YLastPoly);
    orgout1 := out1;
    while (out1 and out2) = 0 do begin
      { Some of line visible }
      if (out1 or out2) <> 0 then begin
          { Clipping Required }
          if out1 <> 0 then begin
            clppnt(XlastPoly, YlastPoly, x, y, out1);
            Continue;
          end;
          if out2 <> 0 then clppnt(x, y, XlastPoly, YlastPoly, out2);
          Continue;
      end;
      if orgout1 <> 0 then AddPoint(XlastPoly, YlastPoly);
      AddPoint(x, y);
      Break;
    end;
  end;
  if (out2 <> 0) or (NPolygonPoints = 0) then begin
    Clamp(x, y, out2);
    AddPoint(x, y);
  end;
  XLastPoly := x1;
  YLastPoly := y1;
end;

procedure TGraphM.PolyEnd;
var
  InPlot:Boolean;
begin
  if Recording then
    PutType(tyPolyEnd)
  else begin
    if NPolygonPoints = 0 then
      Exit;
    InPlot := CurrentCanvas <> nil;
    if not InPlot then
      StartPlotting;
    NeedLinePen;
    CurrentCanvas.Pen.Style := psClear;
    SetLength(FPolygonPoints, NPolygonPoints);
    CurrentCanvas.Polygon(FPolygonPoints, Length(FPolygonPoints));
    PenInvalid;
    if not InPlot then
      StopPlotting;
    NPolygonPoints := 0;
  end;
end;

procedure TGraphM.DrawGraphic(x1, x2, y1, y2: Double; Graphic: TGraphic);
begin
  if Recording then begin
    PutReals(tyGraphic, [x1, x2, y1, y2]);
    RStream.WriteBuffer(Graphic, SizeOf(Graphic));
  end;
  { Not implemented in the absence of a real device }
end;

procedure TGraphM.CLineStyleChanged(Sender: TObject);
var
  AWidth: Single;
  Colour: TColor;
  APattern: TLinePattern;
begin
  inherited;
  if (Recording) then begin
    PutType(tyStyleT);
    AWidth := CLineStyle.Width;
    Colour := CLineStyle.Colour;
    APattern := CLineStyle.Pattern;
    RStream.WriteBuffer(AWidth, SizeOf(AWidth));
    RStream.WriteBuffer(Colour, SizeOf(Colour));
    RStream.WriteBuffer(APattern, SizeOf(APattern));
  end;
end;

procedure TGraphM.Style(AStyle: Integer);
begin
  Linetype := AStyle;
end;

procedure TGraphM.StyleT(AStyle: TLineStyle);
begin
  Linestyle := AStyle;
end;

procedure TGraphM.StyleP(Colour: TColor; AWidth: Single; APattern: TLinePattern);
begin
  EndDraw;
  CLineStyle.Colour := Colour;
  CLineStyle.Width := AWidth;
  CLineStyle.Pattern := APattern;
  StyleT(CLineStyle);
end;

procedure TGraphM.TextAngle(ang: Double);
begin
  if Recording then
    PutReals(tyTextAngle, [ang])
  else
    TextDeciAngle := Round(ang*10);
end;

procedure TGraphM.Text(x: Double; y: Double; s: string);
var
  p: Integer;
  SavedTextStream: TStream; { Current text stream file }
begin
  { Expand escape sequences, signified by #1 }
  while Length(s) > 0 do begin
    p := Pos(#1, s);
    if p = 0 then begin
      TextJ(x, y, taLeftJustify, s);
      Break;
    end else if p > 0 then begin
      TextJ(x, y, taLeftJustify, Copy(s, 1, p-1));
      Delete(s,1,p-1);
      x := x + (p-1)*cx;
    end;
    if Length(s) < 3 then Break; { Bad escape sequence }
    if Ord(s[3]) = 255 then
      CMarkStyle.Assign(MarkStyles[Ord(s[2])])
    else begin
      CMarkStyle.Colour := Ord(s[3]);
      CMarkStyle.MarkType := TMarkType(Ord(FirstMark) + (Ord(s[2]) and $7f));
      CMarkStyle.DotAtCentre := Ord(s[2]) < 128;
    end;
    CMarkStyle.Size := 2;
    SavedTextStream := TextStream; { Disable text output here }
    try
      TextStream := nil;
      MarkT(ProcX(x+cx), ProcY(y+cy/2), CMarkStyle);
    finally
      TextStream := SavedTextStream;
    end;
    Delete(s,1,3);
    x := x + 2.5*cx;
  end;
end;

procedure TGraphM.TextJ(x: Double; y: Double; mode: TAlignment; s: string);
begin
  if Recording then begin
    PutReals(tyTextJ, [x, y]);
    RStream.WriteBuffer(mode, SizeOf(mode));
    PutString(s);
  end else begin
    Justify := mode;
    TextI(ProcX(x), ProcY(y), s);
  end;
end;

procedure TGraphM.Titlen(n: Integer; str: string);
begin
  if (Recording) then begin
    PutType(tyTitlen);
    RStream.WriteBuffer(n, SizeOf(n));
    PutString(str);
  end else
    if n = 1 then begin
      Bold := True;
      TextJ((Xmin+Xmax)/2, Ymax+n*cy, taCenter, str);
      Bold := False;
    end else
      Text(Xmin+1.5*cx, Ymax+n*cy, str);
end;

procedure TGraphM.Title1(title: string);
begin
  Titlen(1, title);
end;

procedure TGraphM.Title2(title: string);
begin
  Titlen(0, title);
end;

procedure  AxisInterval(
(*
  Given an axis length, range, work out the minor tick Integererval, ds
  The major ticks are assumed to be at Intervals of 5*ds, and there
  will be at least nticks of these. Where exponential notation must
  be used, dividing by se := 10**ne will reduce d to the range 1 - 1000
  This routine will not give good results if the range is much smaller
  than the absolute values.
 *)
    (* Input: *)
    range:Double; (* Axis length *)
    (* Input/output *)
    var nticks:Integer; (* Minimum number of major ticks (Forced to be >= 2) *)
    (* Outputs *)
    out ds:Double;  (* Minor tick Interval divided by se (so Interval is actually ds*se) *)
    out se:Double;  (* Scale factor required to reduce d to where exponential notation is not required *)
    out ne:Integer; (* se := 10**ne, i.e. power of 10 needed *)
    out nd:Integer  (* Number of places after the decimal Point required (>=0) *)
);
var
  d, t, s:Double;
  ll:Integer;
begin
  (* Range (to be scaled) *)
  d := abs(range);
  if d = 0 then d := 1;
  (* First Reduce length (d) to approx engineering notation
     i.e 1 - 1000, though 0.1 - 1 and 1e3 - 1e5 are allowed to stand *)
  ne := 0; (* power of 10 *)
  se := 1; (* scale factor (10**ne) *)
  while True do begin
    if (d >= 1e5) or ( (d >= 1e3) and (ne <> 0) ) then begin
      t := d / 1000;
      if t >= d then Exit; (* Only true if d is a nan *)
      d := t;
      se := se * 1000;
      ne := ne + 3;
    end else if (d < 1e-3) or ( (d < 1 ) and (ne <> 0) ) then begin
      d := d * 1000;
      ne := ne - 3;
      se := se / 1000;
    end else
      Break;
  end;
  (* Now reduce to 1 <= d < 10 *)
  s := 1;  (* scale factor *)
  nd := 0; (* power of 10 *)
  while True do begin
    if d < 1 then begin
      d := d * 10;
      s := s / 10;
      nd := nd + 1;
    end else if d >= 10 then begin
      d := d / 10;
      s := s * 10;
      nd := nd - 1;
    end else
      Break;
  end;
  (* Now decide major tick spacing
     decrease ds in a 10 5 2 1 sequence) and (adjust format
     until a reasonable number is obtained *)
  ds := 1;
  ll := 0; (* Current position in 10 5 2 1 ... sequence *)
  if nticks <= 2 then begin
    (* Fix v small plots *)
    nticks := 2;
    if d/ds > 2 then begin
      ds := ds * 2;
      ll := 2;
    end;
    if d/ds > 2 then begin
      ds := ds * 2.5;
      ll := 1;
    end;
  end;
  while d/ds < nticks do begin
    ll := ll + 1;
    case ll of
      1:     (* convert 1 to 0.5 *)
        begin
          ds := ds/2;
          nd := nd + 1;
        end;
      2:     (* convert 0.5 to 0.2 *)
          ds := ds*0.4;
      3:     (* convert 0.2 to 0.1 *)
          ds := ds/2;
    end;
    ll := ll mod 3;
  end;

  ds := ds * s / 5;    (* Minor tick spacing *)

  (* Calculate mark positions *)
  if nd < 0 then nd := 0;
  if range < 0 then ds := -ds;
end;

end.
