unit PHtmlHelpViewer;

interface

procedure ShowHtmlSearch;

implementation

uses SysUtils, Classes, HelpIntfs, Windows, Messages, Forms;

type

 THtmlHelpViewer = class(TInterfacedObject, ICustomHelpViewer, IExtendedHelpViewer)
  private
    FHelpManager: IHelpManager;
    FViewerID: Integer;
    FLastHelpWindow: HWND;
  public
   constructor Create;
   destructor Destroy; override;
   procedure DoHelp(uCommand: cardinal; dwData: longint = 0; Extra: string = ''; HelpFile: string = '');
   procedure InternalShutDown;
   { ICustomHelpViewer - required of all Help Viewers }
   function GetViewerName : String;
   function UnderstandsKeyword(const HelpString: String): Integer;
   function GetHelpStrings(const HelpString: String): TStringList;
   function CanShowTableOfContents: Boolean;
   procedure ShowTableOfContents;
   procedure ShowHelp(const HelpString: String);
   procedure NotifyID(const ViewerID: Integer);
   procedure SoftShutDown;
   procedure ShutDown;
   { IExtendedHelpViewer - Topic and Context }
   function UnderstandsTopic(const Topic: String): Boolean;
   procedure DisplayTopic(const Topic: String);
   function UnderstandsContext(const ContextID: Integer;
                               const HelpFileName: String): Boolean;
   procedure DisplayHelpByContext(const ContextID: Integer;
                                  const HelpFileName: String);
   { ISpecialWinHelpViewer - Winhelp-specific messages }
   // function CallWinHelp(Handle: LongInt; const HelpFileName: String;
   //                     Command: Word; Data: LongInt) : Boolean;
   property ViewerID : Integer read FViewerID;
   property HelpManager : IHelpManager read FHelpManager write FHelpManager;
   { Extensions }
   procedure ShowSearch;
  end;


{ global instance of THtmlHelpViewer which HelpIntfs can talk to. }
var
  HelpViewer: THtmlHelpViewer;
  HelpViewerIntf: ICustomHelpViewer;

{ THtmlHelpViewer }

function THtmlHelpViewer.CanShowTableOfContents: Boolean;
begin
  Result := true;
end;

constructor THtmlHelpViewer.Create;
begin
  inherited Create;
  HelpViewerIntf := Self;
end;

destructor THtmlHelpViewer.Destroy;
begin
  HelpViewer := nil;
  inherited;
end;

procedure THtmlHelpViewer.DisplayHelpByContext(const ContextID: Integer; const HelpFileName: String);
begin
  DoHelp(HH_HELP_CONTEXT, ContextID, '', HelpFileName);
end;

procedure THtmlHelpViewer.DisplayTopic(const Topic: String);
begin
  DoHelp(HH_DISPLAY_TOPIC, 0, ' ::/'+Topic);
end;

procedure THtmlHelpViewer.DoHelp(uCommand: cardinal; dwData: Integer; Extra: string; HelpFile: string);
var
  FileName : String;
begin
  FileName := HelpFile;
  if (FileName = '') and Assigned(FHelpManager) then
    FileName := HelpManager.GetHelpFile;
  if FileName = '' then
    FileName := Application.CurrentHelpFile;
  if FileName <> '' then begin
    if Extra <> '' then
      FileName := FileName + Extra;
    try
      FLastHelpWindow := HtmlHelpA({HelpManager.GetHandle} 0, PChar(FileName), uCommand, dwData);
    except
      raise Exception.Create('Help system not loaded');
    end;
    if Windows.IsWindow(FLastHelpWindow) then
      Windows.SetWindowPos(FLastHelpWindow, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
  end;
end;

function THtmlHelpViewer.GetHelpStrings(const HelpString: String): TStringList;
begin
  Result := TStringList.Create;
  Result.Add(GetViewerName + ': ' + HelpString);
end;

function THtmlHelpViewer.GetViewerName: String;
begin
  Result := 'HtmlHelp';
end;

procedure THtmlHelpViewer.InternalShutDown;
begin
  SoftShutDown;
  if Assigned(FHelpManager) then begin
    HelpManager.Release(ViewerID);
    FHelpManager := nil;
  end;
end;

procedure THtmlHelpViewer.NotifyID(const ViewerID: Integer);
begin
  FViewerID := ViewerID;
end;

procedure THtmlHelpViewer.ShowHelp(const HelpString: String);
var
  link: THH_AKLink;
begin
  link.cbStruct := SizeOf(link);
  link.fReserved := False;
  link.pszKeywords := PChar(HelpString);
  link.pszUrl := nil;
  link.pszMsgText := nil;
  link.pszMsgTitle := nil;
  link.pszWindow := nil;
  link.fIndexOnFail := True;
  DoHelp(HH_ALINK_LOOKUP, LongInt(@link));
end;

procedure THtmlHelpViewer.ShowSearch;
var
  Query: THH_Fts_Query;
begin
  { According to http://support.microsoft.com/kb/241381 this is simply broken }
  Query.cbStruct := SizeOf(Query);
  Query.fUniCodeStrings := FALSE;
  Query.pszSearchQuery := '';
  Query.iProximity := 0;
  Query.fStemmedSearch := FALSE;
  Query.fTitleOnly := FALSE;
  Query.fExecute := FALSE;
  Query.pszWindow := nil;
  DoHelp(HH_DISPLAY_SEARCH, Integer(@Query));
end;

procedure THtmlHelpViewer.ShowTableOfContents;
begin
  DoHelp(HH_DISPLAY_TOC);
end;

procedure THtmlHelpViewer.ShutDown;
begin
  SoftShutDown;
  if Assigned(FHelpManager) then HelpManager := nil;
end;

procedure THtmlHelpViewer.SoftShutDown;
begin
  if Windows.IsWindow(FLastHelpWindow) then
    Windows.SendMessage(FLastHelpWindow, WM_CLOSE, 0, 0);
end;

function THtmlHelpViewer.UnderstandsContext(const ContextID: Integer;
  const HelpFileName: String): Boolean;
begin
  Result := True;
end;

function THtmlHelpViewer.UnderstandsKeyword(const HelpString: String): Integer;
begin
  Result := 1; { Assume OK }
end;

function THtmlHelpViewer.UnderstandsTopic(const Topic: String): Boolean;
begin
  Result := true;
end;

procedure ShowHtmlSearch;
begin
  if Assigned(HelpViewer) then
    HelpViewer.ShowSearch;
end;

initialization
  HelpViewer := THtmlHelpViewer.Create;
  HelpIntfs.RegisterViewer(HelpViewerIntf, HelpViewer.FHelpManager);

finalization
  if Assigned(HelpViewer.FHelpManager) then
    HelpViewer.InternalShutDown;
  HelpViewerIntf := nil;

end.
