unit Mixture;
{ Generic Types required for overall description of the spectrum of a mixture of compounds }

{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics, Contnrs, Classes, SysUtils, pgDOM, Vector, MatrixVec, IntVec, Life,
     MNode, Units, QuantumNumber, DRintensity, Expr, MatrixElements, MNodeColour,
     Interlock, MathSup, QuantumMath, TypInfo, ComplexMatrix, StaticField, Math,
     MNodeExpr, BigSet, Exporter, UCompat;

const
  { Various special index values }
  TemperatureIndex = 5;
  GaussianIndex = 6;
  LorentzianIndex = 7;
  FoffsetIndex = 8;
  EfieldIndex = 16;
  BfieldIndex = 17;
  ShowSumIndex = 8;
  ShowPartsIndex = 9;
  ShowFortratIndex = 10;
  DefaultSmallE = 2e-18; { cm-1 }
  LSP = 3; { First unused parameter index in TState }
  LSI = 4; { First unused setting index in TState }
  LMI = 7; { First unused setting index in TMolecule }
  LPI = 4; { First unused setting index in TPerturbation }

type

  TEigenvalues = class;
  TBasisState = class;

  TLine = record
    Frequency: Double;
    Strength:  Double;
    Width:     Single;
    EigenvaluesU, EigenvaluesL: TEigenvalues; { nil for EigenvaluesU implies a population distribution }
    IndexU, IndexL: Integer;  { Index into Eigenvalues }
    Colour:    TColor;
    CachedName:string; { Not necessarily set; if blank call MakeTxName }
  end;

  TOnCheckLine = function (const Line: TLine): Boolean of Object;

  TFitType = (ftLine, ftContour, ftOverlaySum);

  TFitDetails = record
    FitType: TFitType;
    Parameters: TParameters;
    Correlation: TDoubleArrayArray;
  end;

  { TMNodeWithVariables }

  TMNodeWithVariables = class(TMNodeWithParameters)
  protected
    VariableList: TVariableList;
    LastSuffix: string;
  public
    destructor Destroy; override;
    { Following two are shared between TState and TNucleus }
    function DoSymbolicRotationalMatrixElement(AIndex:Integer; RankWanted: Integer):TMatrixElements; virtual;
    function SymbolicRotationalMatrixElement(AIndex:Integer; RankWanted: Integer):TMatrixElements; virtual;
    procedure DoCheckMatrixElementsAvailable(var AMatrixElements, ASymmetrizedMatrixElements: TMatrixElements; RankWanted: Integer);
    function ParameterVariable(AIndex: Integer; Suffix: string = ''):TSortedVariableNode; virtual;
    function nParameterVariables: Integer; virtual; { Normally nParameters, but Vib has a special case to allow for subsidiary parameters }
  end;

  TQnoInfoMNode = class(TMNodeWithVariables)
  public
    function QnoInfo(QnoType: TQnoType): TQnoInfo; virtual;
    procedure LevelListBySymmetry(Sender: TObject);
    procedure LevelListByEnergy(Sender: TObject);
    procedure PartitionFunctionTable(Sender: TObject);
    procedure AddHandlers; override;
  end;

  TMixtureMNode = class(TQnoInfoMNode)
  public
    function BoltzmannConstant: Double; { k in units of cm-1/K or MHz/K }
    function Precision: Integer;
    function QuantumNumberToStr(AQuantumNumber: TQuantumNumber): string;
    function StrToQuantumNumber(s: string):TQuantumNumber;
    function QuantumNumberFormat: TQuantumNumberFormat;
  end;

  TStreamableMNode = class(TMixtureMNode)
  private
    FDirectoryName: string;
  public
    property DirectoryName: string read FDirectoryName;
    procedure LoadFromFile(AFileName: string); virtual;
    procedure LoadFromStream(Stream: TStream; AFileName: string = '<Stream>');
    procedure MergeFromStream(Stream: TStream; AFileName: string = '<Stream>'); virtual;
    procedure SaveToFile(AFileName: string); virtual;
  end;

  TUnitsMNode = class(TStreamableMNode)
  protected
    procedure SetIntensityUnits(const Index: Integer; const AValue: TIntensityUnits);
    function GetIntensityUnits(const Index: Integer): TIntensityUnits;
    procedure SetUnits(const Index: Integer; const AValue: TResonanceUnits);
    function GetUnits(const Index: Integer): TResonanceUnits;
  public
    function UnitsName: string;
    function UnitsTypeName: string;
  end;

  TStretchableMNode = class(TUnitsMNode)
  { Used as a common base class for TSimulation and TPlot }
  published
    property Ymin: Double            index 3 read GetParameter write SetParameter; { Default is unset - auto }
    property Ymax: Double            index 4 read GetParameter write SetParameter; { Default is unset - auto }
    property AutoMin:   Boolean      index 6 read GetFlag write SetFlag;
    property AutoMax:   Boolean      index 7 read GetFlag write SetFlag;
  end;

  TRootMNode = class(TUnitsMNode)
  protected
    FReadVersion: Integer; { Only valid during read; version set in input file }
  public
    destructor Destroy; override;
    procedure BeginPaste;
    property ReadVersion: Integer read FReadVersion; { Only valid during read; version set in input file }
    procedure PrepareForFit(FitType: TFitType; OtherNode: TRootMNode); virtual; abstract;
  end;

  TBasisOrder = (boDefault, boDiagonal, boStateDiagonal); { How to order the basis states }

  TMixture = class;     { Note logical hierarchy }
   TSimulation = class;
   TSpecies = class;
    TMolecule = class;
     TManifold = class;
      TState = class;
      TPerturbation = class;

  TQuantumNumberReader = function (var a: Integer; const si: string):Boolean of object;

  TObsType = (obLine, obContour, obIntensity, obELower, obEUpper, obParameter, obExpression, obCDLower, obCDUpper);
  TObsFlag = (ofCalculatedSet, ofRegenerateLabel, ofCD);
  TObsFlags = set of TObsFlag;

  TObservation = record
    Frequency:        Double; { Or intensity - see IntensityValue below }
    Weight:           Double;
    Calculated:       Double;
    ExField:          TStaticFields;
    Correlation:      TDoubleArray; { Correlation with following observations }
    ManifoldU, ManifoldL: TManifold;
    MNode:            TMNode;
    JU, JL:           TQuantumNumber;
    SymU, SymL:       Shortint;
    IndexU, IndexL:   Smallint; { State number - 0 to search; For obParameter ParameterIndex, StateIndex }
    AIndexU, AIndexL: Smallint; { Actual state number used }
    Name:             string;
    FileName:         string;  { Name of source file }
    LineNumber:       Integer;
    Colour:           TColor;  { Colour as set in input file }
    RemoveSpins:      Byte;    { Ignore this number of spins, starting with highest numbered nucleus }
    ObsType:          TObsType;{ Type of observation - never obContour }
    Flags:            TObsFlags;
    QuantumNumberFormat: TQuantumNumberFormat;
    IntensityUnits:   TIntensityUnits; { Only if ObsType = obIntensity }
  end;
  TObservationArray = array of TObservation;
  PObservation = ^TObservation;

  TLineList = class;

  TSpectrum = class
  private
    FColor: TColor;
    FUnits: TResonanceUnits;
    FHeadingPrinted: Boolean;
    FIsDistribution: Boolean;
  protected
    SS: TShower;
    FnLines: Integer;
  public
    FMin: Double;
    FMax: Double;
    Temperature: Double;
    ITotal: Double; { TODO: Check/automate usage - this should depend on temperature }
    nPrinted: Integer;
    constructor Create(AUnits: TResonanceUnits; AIsDistribution: Boolean);
    procedure SetShower(var F: Text; PrintLevel: TPrintLevel);
    procedure Clear; virtual;
    procedure AddLine(Frequency, Strength, Width: Double; AColour: TColor; AUnits: TResonanceUnits); overload; virtual;
    procedure AddLine(const Line: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); overload; virtual; abstract;
    procedure AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer = 0; Increment: Integer = 1); overload; virtual;
    procedure AddIn(LineList: TLineList); overload;
    procedure ShowFooter;
    procedure ShowLine(var Line: TLine; Threshold: Double; LineFreqUnits, DisplayFreqUnits: TResonanceUnits; LinFormat: Boolean; var Problem: Exception);
    property nLines: Integer read FnLines;
    property Colour: TColor read FColor write FColor;
    property Units: TResonanceUnits read FUnits write FUnits;
    property IsDistribution: Boolean read FIsDistribution;
  end;

  TLineList = class(TSpectrum)
  private
    FLines: array of TLine;
    function GetLine(AIndex: Integer): TLine;
  public
    Selected: Boolean;
    procedure AddLine(const ALine: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception); overload; override;
    procedure Clear; override;
    procedure Delete(n: Integer);
    procedure DeleteSelected(UpStateSel, LoStateSel: Integer);
    function LargestIntensity: Double;
    {$IFNDEF CONSOLE}
    procedure FindLines(Centre, Range: Double; AUnits: TResonanceUnits; AColour: TColor = clHide; OnCheckLine: TOnCheckLine = nil);
    {$ENDIF}
    function FindPeak(Left, Right: Double; out Position: Double; AUnits: TResonanceUnits; AColour: TColor):Boolean;
    procedure ShowLine(LineNumber: Integer; Threshold:Double; LineFreqUnits, DispFreqUnits: TResonanceUnits; LinFormat: Boolean; var Problem: Exception); overload;
    property Line[AIndex: Integer]: TLine read GetLine;
  end;

  { TAxisSwitch }

  TAxisSwitch = class(TMNode)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    class function CanBeChildOf(MNode: TMNode): Boolean; override;
  published
    property Chi: Double   index 1 read GetParameter write SetParameter;
    property Theta: Double index 2 read GetParameter write SetParameter;
    property Phi: Double   index 3 read GetParameter write SetParameter;
  end;

  { TTransitionFunction - for generating custom transition matrix elements }

  TTransitionFunction = class(TMNodeExprVars)
  public
    procedure MatrixElement(var Value: Complex; ABra, AKet: TBasisState; SignA, SignB: Integer; Intensity: Boolean); virtual; abstract;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Signed:  Boolean index 3 read GetFlag write SetFlag;
  end;

  { TTransitionMoment }

  TVibronicallyAllowed = (vaMaybe, vaYes, vaNo);

  TTransitionMoment = class(TMNodePair)
  private
    FVibronicallyAllowed: TVibronicallyAllowed;
    FMatrixElements, FNoHyperfineMatrixElements, FSymmetrizedMatrixElements: TMatrixElements; { Matrix elements in symbolic form }
    FAxisSwitcher: TAxisSwitch;
    FnPerturbations: Integer;
    FMatrixElementsEmpty,
    FMatrixElementsImplemented: Boolean;
    FParameterVariable: TSortedVariable;
    FActiveTransitionFunction: TTransitionFunction;
  protected
    FRank: TQuantumNumber;
    FMagnetic: Boolean;
    FNuclearMagneton: Boolean;
  public
    TMlist: array [0..1] of TTransitionMoment;
    function BraOwner: TMNode; override;
    function KetOwner: TMNode; override;
    class function BraKetClass: TMNodeClass; override;
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    function nChildTypes: Integer; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; State1Name, State2Name: string); overload;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function Allowed(Sym1, Sym2: Integer): Boolean;
    function CheckProjections(K1, K2: TQuantumNumber):Boolean; virtual;
    function MatrixElement(ABra, AKet: TBasisState; Intensity: Boolean): Complex;
    function CheckMatrixElementsAvailable: Boolean; virtual;
    procedure ShowSymbolicMatrixElements(var Log: Text; const Indent: string = '');
    function SymbolicMatrixElements(Intensity: Boolean; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean = False): TMatrixElements;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    property ParameterVariable: TSortedVariable read FParameterVariable;
    property MatrixElements: TMatrixElements read FMatrixElements;
    property SymmetrizedMatrixElements: TMatrixElements read FSymmetrizedMatrixElements;
    property Magnetic: Boolean read FMagnetic;
    property NuclearMagneton: Boolean read FNuclearMagneton;
    property VibronicallyAllowed: TVibronicallyAllowed read FVibronicallyAllowed write FVibronicallyAllowed ;
    property Rank: TQuantumNumber read FRank;
    property AxisSwitcher: TAxisSwitch read FAxisSwitcher;
    property nPerturbations: Integer read FnPerturbations;
    property ActiveTransitionFunction: TTransitionFunction read FActiveTransitionFunction;

  end;

  TTransitionMomentClass = class of TTransitionMoment;

  TSingleTransitionMoment = class(TTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property Strength: Double  index 1 read GetParameter write SetParameter;
  end;

  TNuclearMagneticMoment = class(TTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property gN: Double index 1 read GetParameter write SetParameter;
    property Nucleus:  Integer index 1 read GetSetting write SetSetting;
  end;

  TBaseRotationalTransitionMoment = class(TSingleTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer; override;
  end;

  TRotationalTransitionMoment = class(TSingleTransitionMoment)
  public
    constructor Create(AOwner: TComponent); overload; override;
    function PairName: string; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
  end;

  { TSphericalTransitionMoment }

  TSphericalTransitionMoment = class(TBaseRotationalTransitionMoment)
  private
    FActualComponent: TQuantumNumber;
  public
    constructor Create(AOwner: TComponent); override;
    function PairName: string; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function FixupProperties:Exception; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    { Same as component, unless component = 'auto' in which case the symmetry allowed value }
    property ActualComponent: TQuantumNumber read FActualComponent;
  published
    property Rank:      Integer index 1 read GetSetting write SetSetting;
    property Component: Integer index 2 read GetSetting write SetSetting;
  end;

  { TMagneticTransitionMoment }

  TMagneticTransitionMoment = class(TSphericalTransitionMoment)
  public
    function PairName: string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    constructor Create(AOwner: TComponent); overload; override;
  end;

  { TOrbitalTransitionMoment }

  TOrbitalTransitionMoment = class(TMagneticTransitionMoment)
    function PairName: string; override;
  end;

  { TSpinTransitionMoment }

  TSpinTransitionMoment = class(TMagneticTransitionMoment)
  public
    function PairName: string; override;
    constructor Create(AOwner: TComponent); overload; override;
  end;

  TabcAxis = (aAxis=0, bAxis, cAxis);
  TxyzAxis = (zAxis=0, yAxis, xAxis);

  TCartesianTransitionMoment = class(TBaseRotationalTransitionMoment)
  public
    function PairName: string; override;
  private
    function GetAxis(const Index: Integer): TabcAxis;
    procedure SetAxis(const Index: Integer; const AValue: TabcAxis);
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties:Exception; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Axis: TabcAxis index 1 read GetAxis write SetAxis;
  end;

  TTransitionMoments = class(TMNodePair)
  private
    FnRanks: Integer;
    function GetTransitionMoment(AIndex: Integer): TTransitionMoment;
  public
    DisableRankCheck: Boolean; { TODO: Fix this fudge for vibrational modes }
    TMlist: array [0..1] of array of TTransitionMoment;  { Linked list of individual transition moments for each state }
    class function BraKetClass: TMNodeClass; override;
    constructor Create(AOwner: TComponent); overload; override;
    constructor Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string); overload;
    destructor Destroy; override;
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = False); virtual;
    function MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false):Boolean; overload; virtual;
    function MakeLine(out Line: TLine; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; Simulation: TSimulation):Boolean; overload;
    procedure AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = false);
    function OriginalTransitionMatrix(Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment; Intensity, Electric, Magnetic: Boolean; Simulation: TSimulation; NilIfEmpty: Boolean = False): TComplexMatrix;
    function TransitionMatrix(Ev1, Ev2: TEigenvalues; Wait, Cache: Boolean; TransitionMomentSel: TTransitionMoment; Simulation: TSimulation; APrintLevel: TPrintLevel = plDefault): TComplexMatrix;
    function Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean; virtual;
    function PairName: string; override;
    function MaxRank: TQuantumNumber; virtual;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    { TODO: Add checks in loader for duplicate or erroneous TransitionMoment(s) }
    property nRanks: Integer read FnRanks;
    property TransitionMoment[AIndex: Integer]: TTransitionMoment read GetTransitionMoment;
  published
    property Colour: TColor index 1 read GetColour write SetColour;
  end;

  { TNucleus }

  TNucleus = class(TMNodeWithVariables)
  private
    FState: TState;
    FMatrixElements, FSymmetrizedMatrixElements: TMatrixElements;
    FMatrixElementsByRank: array [1..2] of TMatrixElements;
    FMatrixElementsEmpty: Boolean;
  protected
    function GetName: TComponentName; override;
    procedure SetName(const Value: TComponentName); override;
    procedure SetFloating(AIndex: Integer; const Value: Boolean); override;
    procedure SetIncrement(AIndex: Integer; const Value: Double); override;
    procedure SetComponentIndex(const Value: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ClearCache(ClearChildren: Boolean); override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    procedure ConvertToPerturbations; virtual;
    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; virtual;
    function SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements; override;
    procedure CheckMatrixElementsAvailable;
    function ParameterVariable(AIndex: Integer; Suffix: string = ''):TSortedVariableNode; override;
    function Molecule: TMolecule;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HasIdentical: Boolean; virtual;
    function MatchingNucleus: TNucleus;
    procedure GlobalShift(Shift:Integer); virtual;
    procedure GlobalMoveUp(Sender: TObject);
    procedure GlobalMoveDown(Sender: TObject);
    procedure AddHandlers; override;
    procedure AddNodeParameters(var FittingParameters: TParameters); override;
    property State: TState read FState;
    property MatrixElements: TMatrixElements read FMatrixElements;
    property SymmetrizedMatrixElements: TMatrixElements read FSymmetrizedMatrixElements;
    property MatrixElementsEmpty: Boolean read FMatrixElementsEmpty;
    property Spin: Integer   index 1 read GetSetting write SetSetting;
    property AsNext: Boolean index 2 read GetFlag write SetFlag;
  end;

  TNucleusClass = class of TNucleus;

  { TNucleusWithSpin }

  TNucleusWithSpin = class(TNucleus)
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Spin;
    property AsNext;
  end;

  TSpinSpin = class(TMNodePair)
  public
    class function BraKetClass: TMNodeClass; override;
    function PairName: string; override;
  end;

  TSpinSpinClass = class of TSpinSpin;

  THyperfineBasis = array of TQuantumNumber;

  { TPartitionFunction - for generating the partition function as a function of temperture }

  TPartitionFunction = class(TMNode)
  private
    procedure SetActive(const Index: Integer; const Value: Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    function CalculatePartitionFunction(Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double; var SpinSpeciesAbundance: TDoubleArray;
                                        EMax: Double = MaxDouble; Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false;
                                        RemoveSpins: Byte = 0):Double; virtual; abstract;
    function EnergyZero(EstimatedOrigin: Double):Double; virtual;
    function HelpOnSetting(AIndex: Integer): string; override;
  published
    property Active: Boolean index 1 read GetFlag write SetActive default True;
  end;

  { TPopulationFunction - for generating custom populations }

  TPopulationFunction = class(TMNodeExprVars)
  public
    procedure Population(var Result: Double; Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer); virtual; abstract;
  end;

  { TMolecule }

  TMolecule = class(TMixtureMNode) { Isotopomer }
  private
    FLongestManifoldName: string;
    FnStates, FnManifolds: Integer;
    FManifolds: TMNodeArray;
    FnTransitionMoments: Integer;
    FHalfSpin, FMaxRank: TQuantumNumber;
    FActualAssumedOrigin: Double;
    FLastPartitionFunction, FLastEmin: TDoubleArray;
    FLastTemperature, FLastTvib, FLastTspin: Double;
    FLastField: TStaticFields;
    FLastProblem: string;
    FAllSinglet: Boolean;
    FSymLength: Integer;
    FActivePartitionFunction, FDefaultPartitionFunction: TPartitionFunction;
    FSpinSpeciesAbundance: TDoubleArrayArray;
    function GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
    function GetMixture: TMixture;
    function GetManifold(AIndex: Integer):TManifold;
    function GetActivePartitionFunction: TPartitionFunction;
  protected
    FnSpinSpecies: Integer;
    TransitionsMomentsCacheLock, PartitionFunctionCacheLock: TMRWLock;
    procedure SetnNuclei(const Index, Value: Integer); virtual;
    procedure SetModified(const Value: Boolean); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function CheckManifolds: Exception;
  public
    Nuclei: array of TNuclearInfo;
    HyperfineBasisStates: array of array of THyperfineBasis;
      { First index is number of nuclear spins, to allow for partial ignoring of spins }
      { Second index is for the basis }
      { Third index is nuclear spin }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    function nChildTypes: Integer; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure ScaleParameters(UFrom, UTo: TResonanceUnits); override;
    procedure ClearCache(ClearChildren: Boolean); override;
    function FixupProperties:Exception; override;
    procedure PrecomputeSymNames; virtual;
    procedure SetUpTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray); virtual;
    procedure SetTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray; ABra, AKet: TBasisState; SignA, SignB: Integer); virtual;
    procedure SetUpPopulationFunctionVariables(PopulationFunction: TPopulationFunction; var ExtraVariables: TVariableArray); virtual;
    procedure SetPopulationFunctionVariables(PopulationFunction: TPopulationFunction; var ExtraVariables: TVariableArray; AState: TBasisState); virtual;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function GetTransitionMoments(AIndex: Integer): TTransitionMoments;
    property HalfSpin: TQuantumNumber read FHalfSpin;
    property AllSinglet: Boolean read FAllSinglet;
    function Species: TSpecies;
    function EinsteinA(const Line: TLine; AUnits: TResonanceUnits): Double;
    function AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception;
                            Simulation: TSimulation;
                            J1Sel: Integer = Low(Integer); Sym1Sel: Integer = -1; J2Sel: Integer = Low(Integer); Sym2Sel: Integer = -1;
                            Manifold1Sel: TManifold = nil; Manifold2Sel: TManifold = nil;
                            DeltaJ: TQuantumNumber = -10000;
                            TransitionMomentsSel: TTransitionMoments = nil; TransitionMomentSel: TTransitionMoment = nil;
                            State1Sel: TState = nil; State2Sel: TState = nil; IncludePerturbations: Boolean = True; IncludeZero: Boolean = False; RemoveSpins: Byte = 0):Boolean; virtual;
    function PartitionFunction(Simulation: TSimulation; var Overflow: Exception; out Emin: Double; RemoveSpins: Byte = 0):Double;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; virtual; abstract;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual; abstract;
    function JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; virtual;
    function ConvertDefaultSymmetry(J: TQuantumNumber; DefaultSymmetry: Integer): Integer;
    function LastSym: Integer; virtual; abstract;
    function SymName(Sym: Integer): string; virtual;
    function JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
    function JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string; virtual;
    function BothSymNames(J: TQuantumNumber; Sym: Integer): string;
    function TryStrToJAdjustedSymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
    function TryStrToAnySymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
    function TryStrToSimpleQuantumNumber(const str: string; var Value:TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; var SpecialQN: Integer; QnoIndex: Integer):Boolean; virtual;
    procedure FixSimpleFormat(var s: string); virtual;
    function FixSym(J: TQuantumNumber; Sym:Integer; InputPgopherVersion: Integer):Integer; virtual;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function StartHyperfineCycle(Nuc: TNucleus; nBase, Rank: Integer): TIntegerArray;
    function NextHyperfineCycle(Nuc: TNucleus; nBase, Rank: Integer; var Signature: TIntegerArray): Boolean; overload;
    function StartHyperfineTransitionCycle(TransitionMoment: TTransitionMoment; nBase: Integer): TIntegerArray;
    function NextHyperfineCycle(TransitionMoment: TTransitionMoment; nBase: Integer; var Signature: TIntegerArray): Boolean; overload;
    function HyperfineMatrixElementType(Signature: TIntegerArray; nBase: Integer; IsBra: Boolean): string;
    function HyperfineCouplingMatrixElement(Nuc: TNucleus; nBase, Rank: Integer; Signature: TIntegerArray; eJ:TExprNode): TExprNode;
    function HyperfineTransitionMatrixElement(TransitionMoment: TTransitionMoment; nBase: Integer; Signature: TIntegerArray; eJ:TExprNode): TExprNode;
    property LongestManifoldName: string read FLongestManifoldName; { For formatting only }
    function ManifoldWithName(AName: string): TManifold;
    property ActualAssumedOrigin: Double read FActualAssumedOrigin;
    property nManifolds: Integer read FnManifolds;
    property nStates: Integer read FnStates;
    property Manifold[AIndex: Integer]: TManifold read GetManifold;
    property MaxRank: TQuantumNumber read FMaxRank;
    property Mixture: TMixture read GetMixture;
    function FindTransitionMoments(ManifoldA, ManifoldB: TManifold):TTransitionMoments;
    property TransitionMoments[AIndex, BIndex:Integer]: TTransitionMoments read GetTransitionMomentsFunction;
    property nSpinSpecies: Integer read FnSpinSpecies;
    function SpinSpeciesAbundance(Simulation: TSimulation; Index: Integer; var Overflow: Exception; RemoveSpins: Byte):Double;
    property nTransitionMoments: Integer read FnTransitionMoments;
    property SymLength: Integer read FSymLength;
    property ActivePartitionFunction: TPartitionFunction read GetActivePartitionFunction;
  published
    property Abundance:     Double  index 1 read GetParameter write SetParameter;
    property AssumedOrigin: Double  index 2 read GetParameter write SetParameter;
    property nNuclei:    Integer index 1 read GetSetting write SetnNuclei;
    property Jmin:       Integer index 2 read GetSetting write SetSetting;
    property Jmax:       Integer index 3 read GetSetting write SetSetting;
    property Colour:     TColor  index 4 read GetColour write SetColour;
    property JAdjustSym: Boolean index 5 read GetFlag write SetFlag;
    property BlockMatrix:Boolean index 6 read GetFlag write SetFlag;
  end;

  TMoleculeClass = class of TMolecule;

  { TBasisState }

  TBasisState = class(TComponent)
  private
    FCacheLock: {$IFDEF FPC}Cardinal{$ELSE}LongInt{$ENDIF};
    FCachedStateLabel: string;
  public
    State: TState;
    SubBasisIndex: Integer; { For help with eigenvalue searches }
    HyperfineState: THyperfineBasis;
    F, J, N, S, M: TQuantumNumber; { Note than N is not guaranteed, particularly for linear molecules }
                                   { S is included here to allow it to be excluded from calculation if required }
    Sym: Integer; { Note that this F and M are included here, as they will be different from the owning TEigenvalues for non zero external fields }
    Phase: Integer;  { Where wavefunction is of the form A +- B, sign of combination; 0 if single component }
    iPower: Integer; { Multiply wavefunction by i^iPower; helps to avoid use of complex wavefunctions }
    StatisticalWeight: SmallInt; { The statistical weight from nuclei not included in the basis }
    StatWtGroup: SmallInt;       { The index of the statistical weight used }
    constructor Create(AOwner: TComponent; AF, AM: TQuantumNumber; ASym: Integer);
    procedure CopyFrom(OtherState: TBasisState; Sign: Integer); virtual;
    function NextBasis(var LoopState: TBasisState; TransitionMoment: TTransitionMoment; Sign: Integer):Boolean; virtual;
    function TrackIndex:Integer; virtual; { A number >= 0 for tracking related states, mainly in energy level plots. }
    function BasisLabel: string; virtual; { Basis State Label }
    function BasisLabelKey: string; virtual; { Explanation of Basis State Label }
    function StateLabel(QuantumNumberFormat: TQuantumNumberFormat): string; virtual; { State name that goes with the given basis state
                                            (assuming match between basis state and final state is OK) }
    function CachedStateLabel(QuantumNumberFormat: TQuantumNumberFormat): string;
    function StateLabelKey: string; virtual; { Explanation of State Name }
    function Qno(QnoType: TQnoType): TQuantumNumber; virtual;
    { Note MNode parameter is required for these two because of special case in Symmetric tops }
    function MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode; Symmetrized: Boolean): TIntegerArray; virtual; abstract;
    procedure SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer; MNode: TMNode); virtual; abstract;
    function MakeHyperfineSignature(Ket: TBasisState; nBase: Integer): TIntegerArray;
    function SymbolicVariableValues: string; virtual;
    function HyperfineStateLabel(WithBlanks: Boolean = True): string;
    function HyperfineStateLabelKey(WithBlanks: Boolean = True; ExtraChar: Char = ' '): string;
    procedure SetHyperfineVariables;
    function HyperfineVariableValues: string;
    function TxName(Upper, Lower: TBasisState): string; virtual;    { Make transition label such as R(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; virtual; { String explaining transition label }
    function HyperfineTxNameKey(WithBlanks: Boolean):string;
  end;

  { Type of matrix elements required }
  TMatElemType = (meHrot {Normal Hamiltonian}, meWidth {Operator for lifetime}, meFloatedDerivatives, meAllDerivatives);
  TMatElemTypes = set of TMatElemType;
  TDArray = array of Double;
  TComplexMatrices = array of TComplexMatrix;

  TCachedTransitionMatrix = class(TInterlockedComponent)
  public
    Eigenvalues, OtherEigenvalues:TEigenvalues;
    TransitionMatrix: TComplexMatrix;
    PrintLevel: TPrintLevel;
    TransitionMoments: TTransitionMoments;
    Simulation: TSimulation;
    IsEmpty: Boolean;
    destructor Destroy; override;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
  end;

  { TEigenvalues }

  TDerivative = record
    MNode: TMNode;
    Index: Integer;
    Values: TColumnVector;
  end;

  TDerivatives = array of TDerivative;

  TEigenvalues = class(TInterlockedComponent)
  private
    FE: TColumnVector;
    FWidths: TColumnVector;
    FC: TComplexMatrix;
    FDominantBasisIndex: TIntColumnVector; { Dominant basis state for each eigenvalue }
    CachedTransitionMatrices: array of TCachedTransitionMatrix;
    FHasFields: Boolean;
    FDerivativesCalculated: Boolean;
    function GetDominantBasisState(AIndex: Integer): TBasisState; inline;
  protected
    FSize: Integer;
    function GetEnergy(AIndex: Integer): Double; virtual;
  public
    Manifold: TManifold;
    J: TQuantumNumber; { M is HasFields set }
    Sym: Integer;
    LastSubBasisIndex: Integer;
    SelectedState: TState;    { Eigenvalues calculated with this State only in basis }
    NoPerturbations: Boolean; { Eigenvalues calculated excluding perturbations }
    RemovedSpins:    Byte;    { Number of spins not included in calculation }
    Basis: array of TBasisState;
    PrintLevel: TPrintLevel;
    ExField: TStaticFields;   { External fields; note that if either is not zero, J is interpreted as MJ and Sym is <0 }
    Derivatives: TDerivatives; { Derivatives with respect to parameters }
    MultipleSpinSpecies: Boolean; { True if Multiple Statistical Weight groups may be used in the calculation }
    function Degeneracy(AIndex: Integer): Integer; virtual;
    function StatisticalWeight(Index: Integer):Integer; virtual;
    function SpinSpecies(Index: Integer):Integer; virtual;
    procedure MatrixElements(out Matrices: TComplexMatrices; BlockMatrix: Boolean; MatElemType: TMatElemType; IncludePerturbations: Boolean = True);
    function AddPerturbation(Matrix: TComplexMatrix; SinglePerturbation: TPerturbation; MatElemType: TMatElemType; var Connections: array of TBigSet; SmallE: Double):Integer;
    procedure CalculateDerivatives(MatElemTypes: TMatElemTypes);
    procedure DoEigenSearch(Co: TComplexMatrix; Offset: Integer = 0);
    procedure DoLimitedEigenSearch(Co: TComplexMatrix; Offset: Integer = 0);
    function DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
    procedure Diagonalize(var Matrices: TComplexMatrices; EigenvectorsWanted, EigenSearch, LimitSearch: Boolean);
    procedure Fixup;
    procedure SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer = 0);
    procedure SortFrom(Start: Integer; Emax: Double);
    procedure Multiply(M: TComplexMatrix);
    procedure TransformTransitionMatrix(M: TComplexMatrix; OtherEv: TEigenvalues);
    procedure MakeHamiltonian(out Matrices: TComplexMatrices; BlockMatrix: Boolean);
    procedure MakeBasis(BasisOrder: TBasisOrder);
    function Make(Wait:Boolean; MatElemTypes: TMatElemTypes):Boolean; overload;
    procedure DoMake(RequiredAttributes: Pointer); override;
    function HasCorrectAttributes(Attributes: Pointer):Boolean; override;
    procedure MakeWidths(APrintLevel: TPrintLevel);
    procedure ShowBasisMatrix(M: TComplexMatrix; const Title: string; Offset: Integer; MaxElements: Integer = 0);
    procedure ShowTransitionMatrix(Ev: TEigenValues; M: TComplexMatrix; const Title: string; Offset: Integer; ShowBasis: Boolean = True; MaxElements: Integer = 0);
    function FindDerivatives(const Parameter: TParameter):Integer;

    function FieldLabel: string;
    procedure ShowEigenvalues;
    function Transform(M: TComplexMatrix): TColumnVector;

    { Helpers mainly to allow for no basis states }
    function Qno(AIndex: Integer; QnoType: TQnoType):TQuantumNumber; virtual;
    function TrackIndex(AIndex: Integer):Integer; virtual;
    function DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat):string; virtual;
    function StateLabelKey(AIndex: Integer):string; virtual;
    function JDeAdjustedSym(AIndex: Integer):Integer; virtual;
    function TxName(AIndex: Integer; OtherEv: TEigenvalues; OtherIndex: Integer):string; virtual;


    constructor Create(AManifold: TManifold; AJ: TQuantumNumber; ASym: Integer; const AExField: TStaticFields;
                       IncludePerturbations: Boolean = True;
                       AState: TState = nil; RemoveSpins: Byte = 0);
    destructor Destroy; override;
    // property Basis[AIndex: Integer]: TBasisState read GetBasisState;
    property Size: Integer read FSize; { -1 implies basis not set up yet }
    property DominantBasis[AIndex: Integer]: TBasisState read GetDominantBasisState;
    property E: TColumnVector read FE;
    property C: TComplexMatrix read FC;
    property Widths: TColumnVector read FWidths;
    property Energy[AIndex:Integer]:Double read GetEnergy; { Same as E[AIndex], but allows override }
    property HasFields: Boolean read FHasFields;
  end;

  TEigenvalueArray = array of TEigenvalues;

  { TPerturbation }

  TPerturbation = class(TMNodePair)
  private
    FMatrixElements, FNoHyperfineMatrixElements, FSymmetrizedMatrixElements: TMatrixElements; { Matrix elements in symbolic form }
    FMatrixElementsEmpty: Boolean;
    FParameterVariable: TSortedVariable;
  public
    NuclearRank: TQuantumNumber; { Rank of nuclear spin operator }
    PerturbationList: array [0..1] of TPerturbation; { Linked list of perturbations involving given Bra or Ket; 0 is Bra, 1 is Ket }
    property MatrixElements: TMatrixElements read FMatrixElements;
    property SymmetrizedMatrixElements: TMatrixElements read FSymmetrizedMatrixElements;
    property ParameterVariable: TSortedVariable read FParameterVariable;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    class function BraKetClass: TMNodeClass; override;
    function Connects(State1, State2: TState): Boolean;
    function FixupProperties:Exception; override;
    function MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Complex;
    function SymbolicMatrixElements(WithHyperfine: Boolean):TMatrixElements; virtual;
    procedure ShowSymbolicMatrixElements(var Log: Text; const Indent: string = '');
    function CheckMatrixElementsAvailable: Boolean; virtual;
    function PairName: string; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    procedure PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function IsAntiSym: Boolean; virtual;
  published
    property Value:  Double index 1 read GetParameter write SetParameter;
    { Index, starting from 1, of nuclear spin involved in perturbation
      0 (default) for those not involving a nuclear spin }
    property Nucleus:Integer   index LPI-3 read GetSetting write SetSetting;
    property SymSelect:Integer index LPI-2 read GetSetting write SetSetting;
    property ScalePrev:Integer index LPI-1  read GetSetting write SetSetting;
  end;

  TMatrixElementFunction = function (A, B: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer):Complex of object;
  TSymCheck = (MustEqual, MayCancel, NoCheck, FullSum);

  { TObservationsReaderBase }

  TObservationsReaderBase = class
  public
    BranchRank: TQuantumNumber;
    RemoveSpins: Integer;
    DefaultSymmetry: Integer; { Default symmetry number; if not J adjusted held as 1000*(Symmetry+1) }
    function ReadQuantumNumber(var a: Integer; const si: string): Boolean; virtual; abstract;
    procedure ReportError(es: string); virtual; abstract;
    function NextItem: string; virtual; abstract;
  end;

  { TState }

  TState = class(TQnoInfoMNode)
  private
    FMolecule: TMolecule;
    FMixture: TMixture;
    FNuclei: TMNodeArray;
    FMatrixElements, FSymmetrizedMatrixElements: TMatrixElements;
    FMatrixElementsEmpty, FNucleiDone: Boolean;
  protected
    FFullRotationalSymmetry: Boolean;
    AllowComplex: Boolean;
  public
    PerturbationList: array [0..1] of TPerturbation; { Linked list of perturbations involving this state; 0 is Bra, 1 is Ket }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; virtual;
    function AddBasis(Eigenvalues: TEigenvalues; F, M: TQuantumNumber; Sym: Integer): Integer; virtual; { Returns Statistical Weight }
    function AddOneJ(Eigenvalues: TEigenvalues; J, F, M: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber): Integer; virtual; abstract;
    class function SpinSpinType: TSpinSpinClass; virtual; abstract;
    class function NucleusType: TNucleusClass; virtual;
    procedure AdjustNuclei(FixNumber: Boolean);
    function SpinSpin(Nuc1, Nuc2: Integer):TSpinSpin;
    function FixupProperties:Exception; override;
    procedure ClearCache(ClearChildren: Boolean); override;
    function HasWidth: Boolean; virtual;

    { Functions working on symmetrized basis functions }
    function MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Complex;
    function SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer; MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck = MustEqual): Complex;
    function SymmetrizedHyperfineMatrixElement(A, B: TBasisState; Rank: TQuantumNumber;
                                    MatrixElementFunction: TMatrixElementFunction; MNode: TMNode): Complex;

    { Note that transition moment matrix elements can be implemented in TState or TransitionMoment descendents }
    function TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex; virtual;
    function SymbolicTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean = False): TMatrixElements;
    function PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Complex;

    { Functions working on unsymmetrized basis functions }
    // function RotSpinMatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Double; virtual; abstract;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; virtual; abstract;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TTransitionMoment; F: TQuantumNumber; IncludeStrength:Boolean):Complex; virtual; abstract;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex; virtual; abstract;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMnode; RankWanted: TQuantumNumber): Complex; virtual; abstract;
    function SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements; override;
    function SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean = False):TMatrixElements; virtual;
    function ConvertToSymmetrized(const MatrixElements: TMatrixElements): TMatrixElements; virtual;

    procedure CheckedMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; const AMatrixElements: TMatrixElements; Numeric: Complex; MNode: TMNode; IncludeJ: Boolean; Symmetrized: Boolean);
    function CheckedRotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; Flag: Integer):Complex;
    function CheckedRotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
    function CheckedRotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Complex;
    function CheckedHyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMnode; RankWanted: TQuantumNumber): Complex;
    procedure CheckMatrixElementsAvailable(WithNuclei: Boolean);
    function AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber):Exception; virtual;

    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; virtual;
    function ReadHyperfineBranch(Text: string; var Obs: TObservation; var HyperfineU, HyperfineL: THyperfineBasis; Reader: TObservationsReaderBase; HasFields:Boolean): Boolean;
    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; virtual;
    function MatrixElementType(const TheMatrixElement: TMatrixElement; const H: string; ShowPhases: Boolean = False):string; virtual; abstract;
    procedure ConvertToPerturbations(Nuclei: Boolean); virtual;
    function Manifold: TManifold;
    function FindColour: TColor;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;

    procedure SymmetryTable(var F: Text); virtual;
    function EstimatedStateOrigin: Double; virtual; { Default is just Origin, but ideally choose lowest energy level }
                                                    { Used to stop Boltzman factors blowing up at low temperatures }
    function Nucleus(Index:Integer): TNucleus;
    property Molecule: TMolecule read FMolecule;
    property Mixture: TMixture read FMixture;
    property FullRotationalSymmetry: Boolean read FFullRotationalSymmetry; { If point group symmetry used is not a subgroup of the rotational Hamiltonian }

    property MatrixElements: TMatrixElements read FMatrixElements;
    property SymmetrizedMatrixElements: TMatrixElements read FSymmetrizedMatrixElements;
    property MatrixElementsEmpty: Boolean read FMatrixElementsEmpty;
  published
    property Origin: Double index 1 read GetParameter write SetParameter;
    property Width:  Double index 2 read GetParameter write SetParameter;
    property Colour: TColor index 1 read GetColour write SetColour;
    property RveSelect: Integer index 2 read GetSetting write SetSetting; { Select overall symmetry }
    property S: TQuantumNumber index 3 read GetSetting write SetSetting; { Spin }
  end;


  { TEigenValueCache }

  TEigenValueCache = class(TMRWLock)
  private
    Entries: array [0..31] of TEigenvalueArray;
  public
    procedure Clear;
    function Hash(J: TQuantumNumber; Sym: Integer):Integer; inline;
    procedure Add(Ev: TEigenvalues; InsertAt: Integer);
  end;

  { TManifold }

  TManifold = class(TMixtureMNode)
  private
    FEigenValueCache: TEigenValueCache;
    FHasWidth: Boolean;
    FnPerturbations: Integer;
    FnStates: Integer;
    FLowestEstimatedOrigin: Double;
    FMaxSpin: TQuantumNumber;
    FPerturbations: TMNodeArray;
    FStates: TMNodeArray;
    PNumbers: array of Integer;
    FActivePopulation: TPopulationFunction;
    function GetState(AIndex: Integer): TState;
    function GetPerturbation(AIndex: Integer): TPerturbation; inline;
    function FindEigenvalues(StartFrom: Integer; out LastSearched: Integer; J: TQuantumNumber; Sym: Integer;
                             const AExField: TStaticFields;
                             SelState: TState = nil; IncludePerturbations: Boolean = True;
                             RemoveSpins: Byte = 0; HasWriteLock: Boolean = False): TEigenvalues;
    procedure UpdateCounts(AComponent: TComponent; Operation: TOperation);
  protected
    procedure SetModified(const Value: Boolean); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    property HasWidth: Boolean read FHasWidth;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function HalfSpin(RemovedSpins: Byte = 0): TQuantumNumber; virtual;
    function GetChildTypes(AIndex: Integer): TMNodeClass; override;
    function nChildTypes: Integer; override;

    { These provide the implementation of the selective population stuff }
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function PopulationIndex(J: TQuantumNumber; Sym, EvNo: Integer; AutoCreate: Boolean = False):Integer;
    procedure DeleteParameter(AIndex: Integer); override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure DecodeParameterName(s: string; out J: TQuantumNumber; out Sym, EvNo: Integer; AQuantumNumberFormat: TQuantumNumberFormat);
    function EncodeParameterName(J: TQuantumNumber; Sym, EvNo: Integer):string;
    function nPopulations: Integer;
    function Population(Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double; virtual;
    procedure AddAllLevels(AState: TState); virtual;
    procedure AddNodeParameters(var FittingParameters: TParameters); override;

    procedure ClearCache(ClearChildren: Boolean); override;
    function Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                         const AExField: TStaticFields;
                         Wait: Boolean = True;
                         PrintLevel: TPrintLevel = plDefault; SelState: TState = nil; IncludePerturbations: Boolean = True;
                         RemoveSpins: Byte = 0;
                         CreateIfNotFound: Boolean = True
                         ): TEigenvalues; virtual;
    function bf(T, E: Double): Double;
    function FindJmin(Default: Integer; RemovedSpins: Byte = 0):Integer; virtual;
    function FindJmax(Default: Integer; RemovedSpins: Byte = 0):Integer; virtual;
    function FindColour: TColor;
    function FirstSym:Integer; { Could in principle be virtual, and depend on J, but not actually used }
    function NextSym(Sym: Integer):Integer; { Could in principle be virtual, and depend on J, but not actually used }
    function FixupProperties:Exception; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function nSym: Integer; virtual; abstract; { Number of different symmetries }
    function FindStateLabel(out IndexF: Smallint; out SymF: Shortint; out AlternateSym: ShortInt; FW: TQuantumNumber; OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; const ExField: TStaticFields; SymW: Integer = -1; JW: TQuantumNumber = -1; NW: TQuantumNumber = -1; StateW: TState = nil; RemoveSpins: Byte = 0; HyperfineW: THyperfineBasis = nil; ShowPossible: Boolean = False): string;
    function FindIndex(out EigenvaluesF: TEigenvalues; out IndexF: Smallint; out SymF: Shortint; out AlternateSym: ShortInt; FW: TQuantumNumber; OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; const ExField: TStaticFields; SymW: Integer = -1; JW: TQuantumNumber = -1; NW: TQuantumNumber = -1; StateW: TState = nil; RemoveSpins: Byte = 0; HyperfineW: THyperfineBasis = nil; EigenvaluesWanted: Boolean = True; ShowPossible: Boolean = False): string;
    function Mixture: TMixture;
    function Molecule: TMolecule;
    function PageName: string; override;
    procedure ConvertToPerturbations;
    function StateWithName(AName: string): TState;
    property nStates: Integer read FnStates;
    property State[AIndex: Integer]: TState read GetState;
    property LowestEstimatedOrigin: double read FLowestEstimatedOrigin;
    property nPerturbations: Integer read FnPerturbations;
    property Perturbation[AIndex: Integer]: TPerturbation read GetPerturbation; default;
    property MaxSpin: TQuantumNumber read FMaxSpin;
  published
    property Jmin:             Integer index 1 read GetSetting write SetSetting;
    property Jmax:             Integer index 2 read GetSetting write SetSetting;
    property Initial:          Boolean index 3 read GetFlag write SetFlag;
    property Colour:           TColor  index 4 read GetColour write SetColour;
    property EigenSearch:      Boolean index 5 read GetFlag write SetFlag;
    property LimitSearch:      Boolean index 6 read GetFlag write SetFlag;
    property AutoQConverge:    Boolean index 7 read GetFlag write SetFlag;
    property UsePopParams:     Boolean index 8 read GetFlag write SetFlag;
    { TODO: property ShowStateNumbers: Boolean index 7 read GetFlag write SetFlag; }
  end;

  TSpecies = class(TQnoInfoMNode)
  private
    function GetIsotopomer(AIndex: Integer): TMolecule;
  public
    constructor Create(AOwner: TComponent); override;
    function AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean; virtual;
    function FindMolecule(AName: string): TMolecule;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function IsSettingColour(AIndex: Integer): Boolean; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    procedure SetModified(const Value: Boolean); override;
    property Isotopomer[AIndex: Integer]: TMolecule read GetIsotopomer;
  published
    property Concentration: Double index 1 read GetParameter write SetParameter;
    property Jmin:      Integer index 1 read GetSetting write SetSetting;
    property Jmax:      Integer index 2 read GetSetting write SetSetting;
    property Colour:    TColor  index 3 read GetColour write SetColour;
  end;

  { TSimulation }

  TColours = array of TColor;
  TGraphStyle = (gsNoOffset,   { All spectra on one scale }
                 gsOffset,     { Separate simulation and experiment }
                 gsOffsetAll,  { Separate all spectra }
                 gsInverted,   { Show simulation inverted below experiment }
                 gsInvertedAll { Separate all traces, with simulated traces inverted }
                );

  TSimulation = class(TStretchableMNode)
  private
    FSettingOverridden, FParameterOverridden: array of Boolean; { Track variables that have been set }
    FBaseNode: TSimulation;
    FWritten: Boolean;
    procedure SetLifeModel(const Index: Integer; const AValue: TLifeModel);
    function GetLifeModel(const Index: Integer): TLifeModel;
    procedure SetFortratQno(const Index: Integer; const Value: TQnoType);
    function GetFortratQno(const Index: Integer): TQnoType;
    function GetExField: TStaticFields;
  protected
    function GetDefaultSettings(AIndex: Integer): string; override;
  public
    HasPlotted: Boolean;
    GraphStyle: TGraphStyle;

    constructor Create(AOwner: TComponent); override;
    procedure ResetOverride;
    function Mixture: TMixture;
    procedure ChangePlotUnits(NewPlotUnits: TResonanceUnits);
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function ParameterOverridden(AIndex: Integer):Boolean;
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
    function SettingToStr(AIndex: Integer): string; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure AddHandlers; override;
    procedure TemperatureTable(Sender: TObject);
    function MakeLineList(var Problem: Exception): TLineList;
    procedure UpdateBaseNode(ABaseNode: TSimulation);
    procedure WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
    procedure PrettyPrintHeader(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;

    function LineIntensity(const Line: TLine; AUnits: TResonanceUnits; var Problem: Exception; RemoveSpins: Byte = 0): Double;
    function HasWidths:Boolean;
    function HasFields:Boolean;
    property ExField: TStaticFields read GetExField;

  published
    property IntensityUnits: TIntensityUnits index 1 read GetIntensityUnits write SetIntensityUnits;
    property LifeModel:     TLifeModel index 2 read GetLifeModel write SetLifeModel;
    property PlotUnits:     TResonanceUnits index 3 read GetUnits write SetUnits;
    property nDF:           Integer index 4 read GetSetting write SetSetting;
    property WidthMult:     Integer index 5 read GetSetting write SetSetting;     { Multiple of linewidth to extend convolution over }
    { In Base: property AutoMin:       Boolean index 6 read GetFlag write SetFlag; // if True (default) Ymin is determined from plotted spectrum }
    { In Base: property AutoMax:       Boolean index 7 read GetFlag write SetFlag; // if True (default) Ymax is determined from plotted spectrum }

    property ShowSum:       Boolean  index ShowSumIndex read GetFlag write SetFlag;
    property ShowParts:     Boolean  index ShowPartsIndex read GetFlag write SetFlag;
    property ShowFortrat:   Boolean  index ShowFortratIndex read GetFlag write SetFlag;
    property UseUpper:      Boolean  index ShowFortratIndex+1 read GetFlag write SetFlag;
    property ShowSymmetry:  Boolean  index 12 read GetFlag write SetFlag; { ... in Fortrat plots }
    property ShowDeltaJ:    Boolean  index 13 read GetFlag write SetFlag; { ... in Fortrat plots }
    property ScaleMarkSize: Boolean  index 14 read GetFlag write SetFlag; { ... in Fortrat plots }
    property UseSymmetry:   Boolean  index 15 read GetFlag write SetFlag; { ... in Fortrat plots }
    property UseStateNumber:Boolean  index 16 read GetFlag write SetFlag; { ... in Fortrat plots }
    property FortratQno:    TQnoType index 17 read GetFortratQno write SetFortratQno; { Quantum number to use in Fortrat plots }


    property Fmin:        Double index  1 read GetParameter write SetParameter;
    property Fmax:        Double index  2 read GetParameter write SetParameter;
    { In Base: property Ymin:        Double index  3 read GetParameter write SetParameter; }
    { In Base: property Ymax:        Double index  4 read GetParameter write SetParameter; }
    property Temperature: Double index TemperatureIndex read GetParameter write SetParameter;
    property Gaussian:    Double index GaussianIndex read GetParameter write SetParameter;
    property Lorentzian:  Double index LorentzianIndex read GetParameter write SetParameter;
    property Foffset:     Double index FoffsetIndex read GetParameter write SetParameter;
    property SMargin:     Double index  9 read GetParameter write SetParameter;
    { Relative Size of peak to ignore when convoluting for contour fitting }
    property OThreshold:  Double index 10 read GetParameter write SetParameter;
    property RefWidth:    Double index 11 read GetParameter write SetParameter;
    property Tvib:        Double index 12 read GetParameter write SetParameter;
    property MinI:        Double index 13 read GetParameter write SetParameter; { as a fraction of maximum intensity; For Fortrat plots; TODO: Use elsewhere }
    property Saturation:  Double index 14 read GetParameter write SetParameter;
    property Tspin:       Double index 15 read GetParameter write SetParameter;
    property EField:      Double index EFieldIndex read GetParameter write SetParameter;
    property BField:      Double index BFieldIndex read GetParameter write SetParameter;
    { TODO: property Angle:       Double index 16 read GetParameter write SetParameter; }
    property Doppler:     Double index 18 read GetParameter write SetParameter;
    property PlotSplit:   Double index 19 read GetParameter write SetParameter;
  end;

  TCorrelationEntry = record
    Row, Col: string;
    Value: Double;
  end;

  TUnpackedCorrelations = array of TCorrelationEntry;

  TCorrelations = class(TMNode)
  public
    FitDetails: TFitDetails;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Clear;
    function nEntries: Integer;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    function EntryName(i: Integer; Quote: Boolean):string;
    function ReadData(UnpackedCorrelations: TUnpackedCorrelations; FirstPass: Integer; CheckUnread: Boolean; OtherNode:TRootMNode):Integer;
    procedure Resync(Sender: TObject);
    procedure AdjustSize;
    procedure AddHandlers; override;
    procedure UnpackCorrelations(var UnpackedCorrelations: TUnpackedCorrelations);
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
  end;

  { TMixture }

  TMixture = class(TRootMNode)
  private
    FnMolecules: Integer;
    FnSpecies: Integer;
    FnSimulations: Integer;
    FSavedParameters: array of TFitDetails;
    FUndoPosition: Integer; { Index within the above of current set of parameters }
    FLongestMoleculeName: string; { For formatting }
    FLongestManifoldName: string; { For formatting }
    FLongestSymLength: Integer; { For formatting }
    FLargestIndex: Integer;     { For formatting }
    FLoadedFormSettings: TDOMDocument; { required between loading and applying to forms only }
    FLineList: TLineList;
    FMatrixElementsAvailable: Boolean;
    function GetQuantumNumberFormat(AIndex: Integer): TQuantumNumberFormat;
    function GetSimulation(AIndex: Integer): TSimulation;
    function GetSpecies(AIndex: Integer): TSpecies;
    procedure SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
    function GetBasisOrder(const Index: Integer): TBasisOrder;
    procedure SetPrintLevel(const Index: Integer; const AValue: TPrintLevel);
    function GetPrintLevel(const Index: Integer): TPrintLevel;
    procedure LoadCurrentSavedParameters;
    procedure SetQuantumNumberFormat(const Index: Integer; const Value: TQuantumNumberFormat);
    procedure SetLineList(const Value: TLineList);
    procedure SetEstUnc(const Index: Integer; const Value: Boolean);
  protected
    procedure SetModified(const Value: Boolean); override;
    procedure SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
    function GetEUnits(const Index: Integer): TEnergyUnits;
  public
    FittingParameters: TParameters;
    FCorrelations: TCorrelations;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure FillParameters; { Set up parameters array }
    procedure PushParameters(FitType: TFitType; SetCurrent: Boolean);
    procedure DiscardFit;
    procedure UndoFit;
    procedure ClearUndo;
    function CanUndoFit: Boolean;
    procedure RedoFit;
    function CanRedoFit: Boolean;
    procedure CycleUnits; override;
    procedure ClearCache(ClearChildren: Boolean); override;
    procedure CheckMatrixElementsAvailable;
    function FixupProperties:Exception; override;
    function SubPropInfo(const Token: string; out ErrMsg: string; ImplicitDot:Boolean = True): TValueInfo; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    procedure PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string = ''); override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    procedure PreCalculatePartitionFunctions(Simulation: TSimulation; var Overflow: Exception);
    procedure ConvertToPerturbations;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    function WriteProperties(Parent: TDOMNode; AllParameters: Boolean = True; WithChildren: Boolean = True): TDOMElement; override;
    procedure ClearFormSettings;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); override;
    function AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean;
    function Correlations: TCorrelations;
    procedure PrepareForFit(FitType: TFitType; OtherNode: TRootMNode); override;

    property Species[AIndex: Integer]: TSpecies read GetSpecies; default;
    function FindMolecule(AName: string): TMolecule;
    property nMolecules: Integer read FnMolecules;
    property nSpecies: Integer read FnSpecies;
    property LongestMoleculeName: string read FLongestMoleculeName;        { For formatting }
    property LongestManifoldName: string read FLongestManifoldName;        { For formatting }
    property LongestSymLength: Integer read FLongestSymLength;             { For formatting }
    property LargestIndex: Integer read FLargestIndex write FLargestIndex; { For formatting }

    property Simulations[AIndex: Integer]: TSimulation read GetSimulation;
    property nSimulations: Integer read FnSimulations;
    property LoadedFormSettings: TDOMDocument read FLoadedFormSettings;
    property LineList: TLineList read FLineList write SetLineList;
  published
    property Units:               TEnergyUnits index 1 read GetEUnits write SetEUnits;
    property PrintLevel:          TPrintLevel index 2 read GetPrintLevel write SetPrintLevel;
    property Precision:           Integer index 3 read GetSetting write SetSetting;     { Precision (of printout) }
    property BasisOrder:          TBasisOrder index 4 read GetBasisOrder write SetBasisOrder;
    property QuantumNumberFormat: TQuantumNumberFormat index 5 read GetQuantumNumberFormat write SetQuantumNumberFormat; { How to display quantum numbers }
    property FitCycles:           Integer index 6 read GetSetting write SetSetting;
    property FitDetails:          Boolean index 7 read GetFlag write SetFlag;
    property AutoReplot:          Boolean index 8 read GetFlag write SetFlag; { Turn off if replot is rather slow }
    property AllowComplex:        Boolean index 9 read GetFlag write SetFlag; { Set to use complex wavefunctions if required }
    property NoLineList:          Boolean index 10 read GetFlag write SetFlag; { Disable use of linelist }
    property ShowEstUnc:          Boolean index 11 read GetFlag write SetEstUnc; { Show estimated uncertainties in line positions }
    property ShowDerivatives:     Boolean index 12 read GetFlag write SetFlag; { Show derivatives in line listings }
    property SVDlimit:            Integer index 13 read GetSetting write SetSetting;
    property ShowSubBasis:        Boolean index 14 read GetFlag write SetFlag;
    property AllowZeroI:          Boolean index 15 read GetFlag write SetFlag;
    property ScaledParameters:    Boolean index 16 read GetFlag write SetFlag;

    property SVDThresh:   Double index 1 read GetParameter write SetParameter;
    property ScaleChanges:Double index 2 read GetParameter write SetParameter;
    property SmallE:      Double index 3 read GetParameter write SetParameter;

  end;

function NotYet:Exception;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
function MakeTxName(var Line: TLine): string;
function AxisName(AAxis: TabcAxis): string;
function LineListWanted(PrintLevel: TPrintLevel):Boolean;
function EstErr(EvU, EvL: TEigenvalues; IndexU, IndexL: Integer):Double;
procedure ShowMatrixElements(var f: Text; MNode: TMNode; State: TState; First, Last: Integer; Zero: Boolean; const Indent: string);

var
 IndividualSymbolicMatrixElementTime: string = 'Work out individual symbolic matrix elements';
 SymbolicMatrixElementTime: string = 'Work out overall symbolic matrix elements';

implementation

uses {$IFNDEF CONSOLE} Fitter, LineFrm, Main, Dialogs, SettingsForm, {$ENDIF}
     {$IFDEF EXPLICITTEST} ExplicitTest, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF} { For TRTLCriticalSection }
     CODATA, Progress, nj, EigLA, CEig, pgXMLWrite, pgXMLRead, Huge,
     Profile, ShellSort, Global, Constrain, PartFunc, OverlaySumFitter,
     Worker, Clock, Obs, OscFields, LevelLister, PopFunc,
     VibPart, TranFunc, PGWork, Lapack, OAsymT, SpectrumPlot;

function NotYet:Exception;
begin
  Result := Exception.Create('Not Yet Implemented');
end;

function LineListWanted(PrintLevel: TPrintLevel):Boolean;
begin
  Result := (PrintLevel >= plLineList) or (PrintLevel in [plCSV, plText, plBinary]);
end;

{ Profiling Variables }
var
 AddMatrixToSpectrumTime: string = 'AddMatrixToSpectrum';
 EigenvaluesTime: string = 'Make and Diagonalize Hamiltonian';
 MakeHamiltonianTime: string = 'Make Hamiltonian Matrix';
 PerturbationMatrixElementTime: string = 'Perturbation Matrix Elements';
 CountBlocksTime: string = 'Counting Blocks';
 BlockMatrixTime: string = 'Blocking Matrix';
 BlockMoveTime: string =   'Blocking Matrix - Values';
 EigenSearchTime: string = 'Identifying eigenvalues';
 EigenvaluesSearchTime: string = 'Eigenvalues Search';
 MakeLineListTime: string = 'Make Line List';
 AddInLinesTime: string = 'AddInLines';
 TransitionMatrixTime: string = 'Make and Transform Transition Matrix';
 MakeTransitionMatrixTime: string = 'Make Transition Matrix';
 MatrixMultiplyTime: string = 'Matrix Multiply';
 PostMultiplyTime: string = 'Post Multiply';
 TxNameTime: string = 'Transition Name';
 FixupPropertiesTime: string = 'Fixup Properties';
 MakeBasisTime: string = 'Make Basis';
const
  ColourHelpString = 'Colour - set to "None" to take value from elsewhere.';

threadvar
  SymSelNames: array of string; { Must be in static variable }

function SymSettingsNames(MNode: TMnode; out LowestValue: Integer): PPChar;
var
  Molecule: TMolecule;
  i: Integer;
begin
  LowestValue := -1;
  Molecule := TMolecule(MNode.AncestorWithClass(TMolecule));
  if Assigned(Molecule) then begin
    Molecule.PrecomputeSymNames;
    SetLength(SymSelNames, Molecule.LastSym+3);
    for i := 0 to Molecule.LastSym do
      if Molecule.JAdjustSym then
        SymSelNames[i+1] := Molecule.JAdjustedSymName(0,i)
      else
        SymSelNames[i+1] := Molecule.SymName(i);
  end else
    SetLength(SymSelNames, 1);
  SymSelNames[0] := 'all';
  Result := PPChar(@SymSelNames[0]);
end;

{ TLine }

function MakeTxName(var Line: TLine): string;
var
  Mixture: TMixture;
begin
  TStart(TxNameTime);
  with Line do
    if (Length(CachedName) = 0) and Assigned(EigenvaluesL) then begin
      if Assigned(EigenvaluesU) then begin
        Result := EigenvaluesU.TxName(IndexU, EigenvaluesL, IndexL);
        if Length(Result) > 0 then
          Result := Result + ' : ';
      end else
        Result := '';
      Mixture := EigenvaluesL.Manifold.Mixture;
      if Mixture.nMolecules > 1 then
        Result := Result + EigenvaluesL.Manifold.Molecule.Name + ' ';
      if Assigned(EigenvaluesU) then
        Result := Result + EigenvaluesU.DominantLabel(IndexU, Mixture.QuantumNumberFormat) + ' - ';
      Result := Result + EigenvaluesL.DominantLabel(IndexL, Mixture.QuantumNumberFormat);
      CachedName := Result;
    end else
      Result := CachedName;
  TEnd(TxNameTime);
end;

procedure ConvertLine(var Line: TLine; UFrom, UTo: TResonanceUnits);
begin
  if UFrom <> UTo then begin
    Line.Width := ConvertWidth(Line.Width, Line.Frequency, UFrom, UTo);
    Line.Frequency := Convert(Line.Frequency, UFrom, UTo);
  end;
end;

{ TMNodeWithVariables }

destructor TMNodeWithVariables.Destroy;
begin
  FreeAndNil(VariableList);
  inherited Destroy;
end;

function TMNodeWithVariables.DoSymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer):TMatrixElements;
begin
  TStart(IndividualSymbolicMatrixElementTime);
  Result := SymbolicRotationalMatrixElement(AIndex, RankWanted);
  TEnd(IndividualSymbolicMatrixElementTime);
end;

function TMNodeWithVariables.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
begin
  raise NotYet;
end;

procedure TMNodeWithVariables.DoCheckMatrixElementsAvailable(var AMatrixElements, ASymmetrizedMatrixElements: TMatrixElements; RankWanted: Integer);
var
  ME, MS: TMatrixElements;
  M: array of TMatrixElements;
  i: Integer;
  Worker: TPgopherWorker;
  Problem: Exception;
begin
  MustBeSerial;
  Problem := nil;
  SetLength(M, nParameterVariables+1);
  Worker := TPgopherWorker(1); { Assume OK to start with }
  for i := 1 to nParameterVariables do
    if ParameterVariable(i).Value <> 0 then begin
      ParameterVariable(i).SetAllValues;
      if nWorkers > 0 then begin
        { Parallel algorithm }
        Worker := FindWorkerWait as TPgopherWorker;
        if Assigned(Worker) and not Assigned(Problem) then
          Worker.SymbolicMatrixElement(@M[i], Self, i, RankWanted, Problem)
        else begin
          { Something has failed }
          if Assigned(Worker) then
            Worker.Skip;
          Worker := nil; { Keep as error flag }
          Break;
        end;
      end else
        M[i] := DoSymbolicRotationalMatrixElement(i, RankWanted);
    end;
  { All variables done; wait for complete }
  if nWorkers > 0 then begin
    WaitNoBusyWorkers;
    if Assigned(Problem) then
      raise Problem
    else if not Assigned(Worker) then
      AbortCalculation;
  end;
  { Construct local copy in ME for locking purposes }
  ME := nil;
  for i := 1 to nParameterVariables do begin
    TStart(SymbolicMatrixElementTime);
//DumpMatrixELements(M[i], 'p:'+ParameterVariable(i).Text(True));
    MergeInMatrixElements(ME, M[i], ParameterVariable(i), RankWanted <= 0);
    TEnd(SymbolicMatrixElementTime);
  end;
  TStart(SymbolicMatrixElementTime);
  CompileMatrixElements(ME, nil);
  TEnd(SymbolicMatrixElementTime);
  AMatrixElements := ME;
  if RankWanted > 0 then
    Exit;
  ME := nil;
  for i := 1 to nParameterVariables do begin
    TStart(SymbolicMatrixElementTime);
//DumpMatrixELements(M[i], 'B');
    MS := (AncestorWithClass(TState) as TState).ConvertToSymmetrized(M[i]);
//DumpMatrixELements(MS, '  ');
    MergeInMatrixElements(ME, MS, ParameterVariable(i), False);
//DumpMatrixELements(ME, 'A');
    TEnd(SymbolicMatrixElementTime);
  end;
  CompileMatrixElements(ME, nil);
  ASymmetrizedMatrixElements := ME;
end;

function TMNodeWithVariables.ParameterVariable(AIndex: Integer; Suffix: string): TSortedVariableNode;
var
  i: Integer;
begin
  { Note that this implementation relies on FixupProperties regenerating the
    symbolic expressions whenever a parameter value changes }
  if Suffix <> LastSuffix then
    FreeAndNil(VariableList);
  if not Assigned(VariableList) then begin
    MustBeSerial;
    VariableList := TVariableList.Create;
    for i := 1 to nParameters do
      VariableList.Add(TSortedVariable.Create(ParameterNames[i]+Suffix,i));
    LastSuffix := Suffix;
  end;
  i := VariableList.IndexOf(ParameterNames[AIndex]+Suffix);
  Result := (VariableList.Objects[i] as TSortedVariable).Template;
  Result.Value := Parameters[AIndex];
end;

function TMNodeWithVariables.nParameterVariables: Integer;
begin
  { Normally nParameters, but Vib has a special case to allow for subsidiary parameters }
  Result := nParameters;
end;

{ TManifold }

function TManifold.GetState(AIndex: Integer): TState;
begin
  {$IFOPT R+}
  Result := FStates[AIndex] as TState;
  {$ELSE}
  Result := TState(FStates[AIndex]);
  {$ENDIF}
end;


function TManifold.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  if AIndex = inherited nChildTypes then
    Result := TCustomPopulationFunction
  else
    Result := inherited GetChildTypes(AIndex);
end;

function TManifold.nChildTypes: Integer;
begin
  Result := inherited nChildTypes + 1;
end;

function TManifold.GetPerturbation(AIndex: Integer): TPerturbation;
begin
  {$IFOPT R+}
  Result := FPerturbations[AIndex] as TPerturbation;
  {$ELSE}
  Result := TPerturbation(FPerturbations[AIndex]);
  {$ENDIF}
end;

function TManifold.StateWithName(AName: string): TState;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TState) and (TState(Items[i]).Name = AName) then begin
      Result := TState(Items[i]);
      Exit;
    end;
end;

procedure TManifold.UpdateCounts(AComponent: TComponent; Operation: TOperation);
  procedure AddPerturbation(P: TPerturbation; bk: Integer);
  var
    State: TState;
  begin
    if bk = 0 then
      State := P.Bra as TState
    else
      State := P.Ket as TState;
    if not Assigned(State) then
      Exit; { Still being fixed up }
    P.PerturbationList[bk] := State.PerturbationList[bk];
    State.PerturbationList[bk] := P;
  end;
var
  i: Integer;
begin
  FnStates := AdjustNodeIndexes(TState,FStates,AComponent,Operation);
  FnPerturbations := AdjustNodeIndexes(TPerturbation,FPerturbations,AComponent,Operation);
  for i := 0 to nStates-1 do
    with State[i] do begin
      PerturbationList[0] := nil;
      PerturbationList[1] := nil;
  end;
  for i := 0 to nPerturbations-1 do begin
    AddPerturbation(Perturbation[i], 0);
    AddPerturbation(Perturbation[i], 1);
  end;
end;

constructor TManifold.Create(AOwner: TComponent);
begin
  FEigenValueCache := TEigenValueCache.Create;
  SavedProperties := SavedProperties + ',AutoQConverge=True,Colour=None,Jmin=-1,Jmax=-1,EigenSearch=True';
  inherited Create(AOwner);
end;

destructor TManifold.Destroy;
begin
  ClearCache(False);
  FreeAndNil(FEigenValueCache);
  inherited Destroy;
end;

function TManifold.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
begin
  Result := State[0].HalfSpin(RemovedSpins);
end;

procedure TManifold.ClearCache(ClearChildren: Boolean);
begin
  inherited ClearCache(ClearChildren);
  PerformMNodeNotification(Self, mnCacheCleared, 0);
  if Assigned(FEigenValueCache) then
    try
      FEigenValueCache.BeginWrite;
      FEigenValueCache.Clear;
    finally
      FEigenValueCache.EndWrite;
    end
end;

function TManifold.FindColour: TColor;
begin
  Result := Colour;
  if Result = clNone then
    Result := Molecule.Colour;
  if (Result = clNone) and (Molecule.Owner is TSpecies) then
    Result := Molecule.Species.Colour;
end;

function TManifold.FindEigenvalues(
  StartFrom: Integer; out LastSearched: Integer;
  J: TQuantumNumber; Sym: Integer;
  const AExField: TStaticFields;
  SelState: TState; IncludePerturbations: Boolean;
  RemoveSpins: Byte; HasWriteLock: Boolean): TEigenvalues;
var
  Hash, i: Integer;
  Entry: TEigenvalueArray;
begin
  { Have we already calculated it ? }
  TStart(EigenvaluesSearchTime);
  Hash := FEigenValueCache.Hash(J,Sym);
  try
    if not HasWriteLock then
      FEigenValueCache.BeginRead;
    Entry := FEigenValueCache.Entries[Hash];
    LastSearched := High(Entry);
    for i := StartFrom to LastSearched do begin
      Result := Entry[i];
      if Result = nil then begin
        LastSearched := i-1;
        Break;
      end;
      if (Result.J = J) and (Result.Sym = Sym)
         and (Result.SelectedState = SelState)
         and (Result.NoPerturbations <> IncludePerturbations)
         and (Result.RemovedSpins = RemoveSpins)
         and (Result.ExField.E = AExField.E)
         and (Result.ExField.B = AExField.B)
         and (Result.ExField.Angle = AExField.Angle)
        then begin
        Exit;
      end;
    end;
    Entry := nil;
  finally
    if not HasWriteLock then
      FEigenValueCache.EndRead;
    TEnd(EigenvaluesSearchTime);
  end;
  Result := nil;
end;

function TManifold.Eigenvalues(J: TQuantumNumber; Sym: Integer; MatElemTypes: TMatElemTypes;
                               const AExField: TStaticFields; Wait: Boolean; PrintLevel: TPrintLevel;
                               SelState: TState; IncludePerturbations: Boolean; RemoveSpins: Byte;
                               CreateIfNotFound: Boolean): TEigenvalues;
var
  LastSearched: Integer;
begin
  if PrintLevel = plDefault then
    PrintLevel := Mixture.PrintLevel;
  with Mixture do begin
    if not ShowDerivatives and (meAllDerivatives in MatElemTypes) then
      MatElemTypes := MatElemTypes - [meAllDerivatives];
    if not ShowEstUnc and (meFloatedDerivatives in MatElemTypes) then
      MatElemTypes := MatElemTypes - [meFloatedDerivatives];
  end;
  { Turn off selection if it makes no difference }
  if nStates = 1 then
    SelState := nil;
  if nPerturbations = 0 then
    IncludePerturbations := True;
  if (Molecule.nNuclei = 0) and (RemoveSpins > 0) then
    RemoveSpins := 1; { Electron spin is the only thing that can be removed }
  if IsNonZero(AExField) then
    Sym := -1;
  Result := FindEigenvalues(0, LastSearched, J, Sym, AExField, SelState, IncludePerturbations, RemoveSpins);
  if not Assigned(Result) then
    if CreateIfNotFound then
      try
        FEigenValueCache.BeginWrite;
        { Check that it hasn't been created by another thread }
        Result := FindEigenvalues(LastSearched+1, LastSearched, J, Sym, AExField, SelState, IncludePerturbations, RemoveSpins, True);
        if not Assigned(Result) then begin
          Result := TEigenvalues.Create(Self, J, Sym, AExField, IncludePerturbations, SelState, RemoveSpins);
          FEigenValueCache.Add(Result, LastSearched+1);
        end;
      finally
        FEigenValueCache.EndWrite;
      end
    else
      Exit;
  { We now have the eigenvalue structure, but not necessarily complete }
  Result.PrintLevel := PrintLevel;
  Result.Make(Wait, MatElemTypes);
end;

function TManifold.FindJmin(Default: Integer; RemovedSpins: Byte): Integer;
var
  HS: Integer;
begin
  Result := Jmin;
  if Result < 0 then
    Result := Molecule.Jmin;
  if (Result < 0) and (Molecule.Owner is TSpecies) then
    Result := Molecule.Species.Jmin;
  if Result < 0 then
    Result := Default;
  HS := HalfSpin(RemovedSpins);
  if Odd(HS+Result) then
    Dec(Result);
  if Result < HS then
    Result := HS
end;

function TManifold.FindJmax(Default: Integer; RemovedSpins: Byte): Integer;
var
  HS: Integer;
begin
  Result := Jmax;
  if Result < 0 then
    Result := Molecule.Jmax;
  if (Result < 0) and (Molecule.Owner is TSpecies) then
    Result := Molecule.Species.Jmax;
  if Result < 0 then
    Result := Default;
  HS := HalfSpin(RemovedSpins);
  if Odd(HS+Result) then
    Dec(Result);
end;

function TManifold.FirstSym:Integer;
begin
  Result := 0;
end;

function TManifold.NextSym(Sym: Integer): Integer;
begin
  Result := Sym+1;
  if Result >= nSym then
    Result := -1;
end;

function TManifold.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

procedure TManifold.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  UpdateCounts(AComponent,Operation);
end;

procedure TManifold.ConvertToPerturbations;
var
  i: Integer;
begin
  for i := 0 to nStates-1 do
    State[i].ConvertToPerturbations(True);
end;

function TManifold.Mixture: TMixture;
begin
  Result := RootNode as TMixture;
end;

procedure TManifold.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache(True);
end;

function TManifold.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  if IsSettingQuantum(AIndex) and (Settings[AIndex] < 0) then
    Exit;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TManifold.FixupProperties:Exception;
var
  i: Integer;
  LastPert, Pert: TPerturbation;
begin
  FActivePopulation := nil;
  FEigenValueCache.SetUpLock;
  UpdateCounts(nil,opInsert);
  Result := inherited FixupProperties;
  FHasWidth := False;
  FMaxSpin := 0;
  for i := 0 to nStates-1 do
    with State[i] do begin
      if HasWidth then
        FHasWidth := True;
      if i = 0 then
        FLowestEstimatedOrigin := EstimatedStateOrigin
      else
        FLowestEstimatedOrigin := Math.Min(FLowestEstimatedOrigin, EstimatedStateOrigin);
      FMaxSpin := Max(FMaxSpin, S);
    end;
  LastPert := nil;
  if not Assigned(Result) then
    for i := 0 to ComponentCount - 1 do
      if Components[i] is TPerturbation then begin
        Pert := Components[i] as TPerturbation;
        if Pert.ScalePrev <> 0 then
          if LastPert = nil then
            Result := EReportException.Create(UniqueName+': First perturbation scaled to previous')
          else begin
            Pert.Value := Pert.ScalePrev*LastPert.Value;
            Pert.Floating[1] := False;
          end;
        LastPert := Pert;
      end else if not Assigned(FActivePopulation) and (Components[i] is TPopulationFunction)
                  and TPopulationFunction(Components[i]).FullyActive then
        FActivePopulation := TPopulationFunction(Components[i]);
end;

function TManifold.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TManifold.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

const
  ManifoldSettingNames: array[1..8] of string = (
'Minimum J in calculation - set negative to take from molecule or manifold.',
'Maximum J in calculation - set negative to take from molecule or manifold.',
'True to include population in this state when calculating spectra.',
'Colour for spectra - set to "None" to get colour from elsewhere.',
'Identify state by looking for largest coefficient in eigenvector.',
'Set to assume the energy ordering within an individual state does not change, but the eigenvectors are used to select the state.',
'Partition function (Q) sum extends until converged (if true) or Jmax (if false).',
'Use numerical populations specified as parameters, rather than Boltzmann equation'
);

function TManifold.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(ManifoldSettingNames))
     and (AIndex <= High(ManifoldSettingNames)) then
    Result := ManifoldSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TManifold.PageName: string;
begin
  Result := 'Manifold'; { These are currently not overridden }
end;

function TManifold.FindIndex(out EigenvaluesF: TEigenvalues;
  out IndexF: Smallint; out SymF: Shortint; out AlternateSym: ShortInt; FW: TQuantumNumber;
  OtherQno: array of TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat;
  const ExField: TStaticFields;
  SymW:{-1} Integer; JW {-1}, NW {-1}: TQuantumNumber; StateW{nil}: TState;
  RemoveSpins:{0} Byte; HyperfineW:{[]} THyperfineBasis;
  EigenvaluesWanted:{True} Boolean; ShowPossible:{False} Boolean): string;
var
  nMatches: Integer;
  Sym, Index: Integer;
  Ev: TEigenvalues;
  function MatchOther(BasisState: TBasisState): Boolean;
  var
    i: Integer;
  begin
    Result := True;
    i := 0;
    while Result and (i <= High(OtherQno)) do begin
      Result := (OtherQno[i] = Low(TQuantumNumber)) or (OtherQno[i] = BasisState.Qno(TQnoType(Ord(qtA)+i)));
      Inc(i);
    end;
    i := 0;
    while Result and (i <= High(HyperfineW)) do begin
      Result := (HyperfineW[i] = Low(TQuantumNumber)) or (BasisState.HyperfineState[i] = HyperfineW[i]);
      Inc(i);
    end;
  end;
  function ShowHyperfine(A: THyperfineBasis):string;
  var
    i: Integer;
    Show: Boolean;
    FF, Fn: TQuantumNumber;
  begin
    Show := False;
    for i := 0 to High(A) do begin
      Show := A[i] <> Low(TQuantumNumber);
      if Show then
        Break;
    end;
    if Show then begin
      Result := ')';
      FF := FW;
      for i := High(A) downto 0 do begin
        if A[i] <> Low(TQuantumNumber) then begin
          if Molecule.Nuclei[i].AsNext then
            Fn := A[i]
          else begin
            Fn := FF;
            Dec(FF, A[i]);
          end;
          Result := QuantumNumber.QuantumNumberToStr(Fn,AQuantumNumberFormat) + Result;
        end else
          Result := '?' + Result;
        if i <> 0 then
          Result := ' ' + Result;
      end;
      Result := ' Hyperfine=('+Result;
    end else
      Result := ''
  end;
var
  s, sy: string;
  KeyShown, HasFields, LocalEv: Boolean;
  SymEnd, i: Integer;
  QI: TQnoInfo;
begin
  AlternateSym := -1;
  nMatches := 0; { More than one match is an error }
  if ShowPossible then
    WriteLn(Log, 'Possible Matches:');
  KeyShown := not ShowPossible;
  EigenvaluesF := nil;
  HasFields := IsNonZero(ExField);
  if HasFields then begin
    Sym := -1;
    SymEnd := Sym;
  end else if SymW < 0 then begin
    { SymW = -1 to search all; more negative
      to start search at higher symmetry numbers }
    Sym := -SymW-1;
    SymEnd := Molecule.LastSym;
  end else begin
    Sym := SymW;
    SymEnd := SymW;
  end;
  repeat
    Ev := Eigenvalues(FW, Sym, [meHrot], ExField, True, Mixture.PrintLevel, nil, True, RemoveSpins, EigenvaluesWanted);
    LocalEv := not Assigned(Ev);
    if LocalEv then
      Ev := TEigenvalues.Create(Self, FW, Sym, ExField, True, nil, RemoveSpins);
    try
      Ev.MakeBasis(Mixture.BasisOrder);
      for Index := 1 to Ev.Size do
        { Look for matches }
        with Ev.DominantBasis[index] do begin
          if ShowPossible then begin
            if not KeyShown then
              WriteLn(Log, StateLabelKey);
            KeyShown := True;
            Writeln(Log, StateLabel(AQuantumNumberFormat));
          end;
          if ((JW < 0) or (J = JW))
             and ((NW < 0) or (N = NW))
             and (not Assigned(StateW) or (State = StateW))
             and MatchOther(Ev.DominantBasis[index]) then begin
            Inc(nMatches);
            IndexF := Index;
            if nMatches = 1 then
              SymF := Sym { Record first match }
            else if (nMatches = 2) and (SymF <> Sym) then
              AlternateSym := Sym
            else
              AlternateSym := -1; { Only save AlternateSym if there are two matches with two different symmetries }
            Result := StateLabel(AQuantumNumberFormat);
            if EigenvaluesWanted then
              EigenvaluesF := Ev;
          end;
        end;
    finally
      if LocalEv then
        Ev.Free;
    end;
    Inc(Sym);
  until Sym > SymEnd;
  if nMatches <> 1 then begin
    EigenvaluesF := nil;
    IndexF := -1-nMatches; {No unique result found}
    if nMatches=0 then
      SymF := -1
    else
      SymF := -1 - SymF; { Return first match }
    if Assigned(StateW) then
      s := StateW.UniqueName
    else
      s := UniqueName;
    if SymW >= 0 then
      sy := Molecule.SymName(SymW)
    else
      sy := 'Sym=any';
    Result := Format(' %d matches for %s Jtot=%s %s', [nMatches, s, QuantumNumber.QuantumNumberToStr(FW,AQuantumNumberFormat), sy]);
    if JW >= 0 then
      Result := Result + Format(' J=%s', [QuantumNumber.QuantumNumberToStr(JW,AQuantumNumberFormat)]);
    if NW >= 0 then
      Result := Result + Format(' N=%s', [QuantumNumber.QuantumNumberToStr(NW,AQuantumNumberFormat)]);
    for i := 0 to High(OtherQno) do begin
      QI := Molecule.QnoInfo(TQnoType(Ord(qtA)+i));
      Result := Result + ' ' + QI.Name + '=';
      if OtherQno[i] = Low(TQuantumNumber) then
        Result := Result + '?'
      else if QI.StepSize = 2 then
        Result := Result + QuantumNumber.QuantumNumberToStr(OtherQno[i],AQuantumNumberFormat)
      else
        Result := Result + IntToStr(OtherQno[i]);
    end;
    Result := Result + ShowHyperfine(HyperfineW);
  end;
end;

{ Given some information about a state, try and find a complete state label and symmetry
  for identifying that state. Notes:
  1. The hyperfine basis excludes the total angular momentum
  2. An index is returned, but does not use the full eigenvalue matching;
     Use FindIndex for this puprpose }
function TManifold.FindStateLabel(
  out IndexF: Smallint; out SymF: ShortInt; out AlternateSym: ShortInt;
  FW: TQuantumNumber; OtherQno: array of TQuantumNumber;
  AQuantumNumberFormat: TQuantumNumberFormat;
  const ExField: TStaticFields;
  SymW:{-1}Integer; JW {-1}, NW {-1}: TQuantumNumber; StateW:{nil}TState;
  RemoveSpins:{0}Byte; HyperfineW: THyperfineBasis;
  ShowPossible:{False} Boolean): string;
var
  Unused: TEigenvalues;
begin
  Result := FindIndex(Unused,IndexF, SymF, AlternateSym, FW, OtherQno, AQuantumNumberFormat, ExField,
                      SymW, JW, NW, StateW, RemoveSpins, HyperfineW, False, ShowPossible);
end;

function TManifold.EncodeParameterName(J: TQuantumNumber; Sym, EvNo: Integer): string;
var
  Jstr: string;
begin
  if (QuantumNumberFormat = qftJ) and Odd(J) then
    Jstr := IntToStr(J)+'/2' { "." too confusing in parameter names }
  else
    Jstr := QuantumNumberToStr(J);
  Result := Format('%s %s %d', [Jstr, Molecule.JAdjustedSymName(J,Sym), EvNo]);
end;

function TManifold.nPopulations: Integer;
begin
  Result := nParameters;
end;

procedure TManifold.DeleteParameter(AIndex: Integer);
begin
  inherited;
  if AIndex <= nParameters then
    Move(PNumbers[AIndex+1], PNumbers[AIndex], (nParameters+1-AIndex)*SizeOf(PNumbers[AIndex]));
  SetLength(PNumbers, nParameters+1);
end;

function TManifold.PopulationIndex(J: TQuantumNumber; Sym, EvNo: Integer; AutoCreate: Boolean = False): Integer;
var
  PNumber, t: LongInt;
  f, l, n: Integer;
begin
  { Find index corresponding to this parameter, or create it if it doesn't exist }
  { Calculate sorting index for speed }
  if EvNo >= 32768 then
    raise Exception.Create('PopulationIndex: EvNo too large');
  if Sym >= 0 then
    PNumber := (J*(Molecule.LastSym+3)+Molecule.JAdjustedSym(J,Sym))*32768+EvNo
  else begin
    { Two special "symmetries", corresponding to M >=0 and M < 0 }
    n := Molecule.LastSym+1;
    if J < 0 then
      Inc(n);
    PNumber := (Abs(J)*(Molecule.LastSym+3)+n)*32768+EvNo;
  end;
  Result := -1;
  if nParameters > 0 then begin
    { Binary search for index }
    f := 1;
    l := nParameters;
    while f <= l do begin
      n := (f + l) div 2;
      t := PNumber - PNumbers[n];
      if t = 0 then begin
        Result := n;
        Break;
      end else if t < 0 then
        l := n-1 { In first half }
      else
        f := n+1;
    end;
    if Result < 0 then
      if t < 0 then
        Result := -n
      else
        Result := -(n+1);
  end;
  if (Result <= 0) and AutoCreate then begin
    SetLength(PNumbers, nParameters+2);
    Result := -Result;
    AddParameter(EncodeParameterName(J,Sym,EvNo), 0, Result);
    if Result < nParameters then
      Move(PNumbers[Result], PNumbers[Result+1], (nParameters-Result)*SizeOf(PNumbers[Result]));
    PNumbers[Result] := PNumber;
  end;
end;

var
  BoltzMax: Double; { = Ln(MaxDouble)-2 }

procedure TManifold.AddAllLevels(AState: TState);
var
  J, JUp: TQuantumNumber;
  Sym, l, Ind, LastSym: Integer;
  Ev: TEigenvalues;
  Pops: array of Double;
  Syms: array of Integer;
  Overflow: Exception;
  Simulation: TSimulation;
  HasFields: Boolean;
begin
  Simulation := Mixture.Simulations[0];
  HasFields := Simulation.HasFields;
  Overflow := nil;
  J := FindJmin(HalfSpin);
  JUp := FindJmax(J+40);
  if HasFields then
    J := -JUp;
  while J <= JUp do begin
    if HasFields then begin
      Sym := -1;
      LastSym := -1;
    end else begin
      Sym := 0;
      LastSym := nSym-1;
    end;
    while Sym <= LastSym do begin
      Ev := Eigenvalues(J, Molecule.JAdjustedSym(J, Sym), [meHrot], Simulation.ExField);
      { Save populations and required states here, as adding
        custom populations will discard eigenvalues }
      SetLength(Pops, Ev.Size);
      SetLength(Syms, Ev.Size);
      for l := 1 to Ev.Size do begin
        Pops[l-1] := Population(Ev, l, Simulation, Overflow, 0);
        if not Assigned(AState) or (AState = Ev.DominantBasis[l].State) then
          Syms[l-1] := Molecule.JAdjustedSym(J, Sym)
        else
          Syms[l-1] := -2; { Not selected state }
      end;
      for l := 1 to Ev.Size do
        if Syms[l-1] >= -1 then begin
          Ind := PopulationIndex(J,Syms[l-1],l,True);
          Parameters[Ind] := Pops[l-1];
        end;
      Inc(Sym);
    end;
    Inc(J, 2);
  end;
  if Assigned(Overflow) then
    raise Overflow;
end;

procedure TManifold.AddNodeParameters(var FittingParameters: TParameters);
var
  i, n: Integer;
begin
  n := Length(FittingParameters);
  inherited;
  { Adjust parameter names to include more information }
  for i := n to High(FittingParameters) do
    if FittingParameters[i].MNode = Self then
      FittingParameters[i].Name := UniqueName + ' ' + HelpOnParameter(FittingParameters[i].Index);
end;

function TManifold.bf(T, E: Double): Double;
begin
  Result := -E/T/BoltzmannConstant;
  if Result > BoltzMax then
    Result := Infinity
  else
    Result := Exp(Result);
end;

function TManifold.Population(Eigenvalues: TEigenvalues; i: Integer; Simulation: TSimulation; var Overflow: Exception; IncludeTspin: Integer): Double;
var
  EBottom: Double;
  AState: TState;
  PIndex: Integer;
begin
  EBottom := Molecule.ActualAssumedOrigin;
  if Simulation.Tvib > 0 then begin
    { Vibrational temperature calc }
    AState := Eigenvalues.DominantBasis[i].State;
    Result := bf(Simulation.Tvib, AState.EstimatedStateOrigin-EBottom);
    EBottom := AState.EstimatedStateOrigin;
  end else
    Result := 1;
  if Simulation.Tspin >= 0 then
    case IncludeTspin of
      0: Result := Result * Molecule.SpinSpeciesAbundance(Simulation, EigenValues.SpinSpecies(i),Overflow, EigenValues.RemovedSpins);
      1: begin
           { Calculating at Tspin }
           Result := Result * bf(Simulation.Tspin, Eigenvalues.E[i] - EBottom);
           Exit; { Skip normal population calculation }
         end;
    { else skip use of SpinSpeciesAbundance }
    end;
  if not (UsePopParams or (Simulation.Temperature <= 0)) then
    { Normal Boltzman calculation }
    Result := Result * bf(Simulation.Temperature, Eigenvalues.E[i] - EBottom)
  else
    with Eigenvalues do begin
      { TODO: This may need speeding up - optimize J range in AddManifoldToSpectrum, for example }
      PIndex := PopulationIndex(J,Sym,i);
      if PIndex > 0 then
        Result := Result * Parameters[PIndex]
      else
        Result := 0;
    end;
  if Assigned(FActivePopulation) then
    FActivePopulation.Population(Result, Eigenvalues, i, Simulation, Overflow, IncludeTSpin);
end;

procedure TManifold.DecodeParameterName(s: string; out J: TQuantumNumber;
  out Sym, EvNo: Integer; AQuantumNumberFormat: TQuantumNumberFormat);
var
  p: Integer;
begin
  p := Pos(' ',s);
  if (p > 1) and TryStrToQuantumNumber(Copy(s,1,p-1),J,AQuantumNumberFormat) then begin
    Delete(s,1,p);
    p := Pos(' ',s);
    if Molecule.TryStrToAnySymName(J,Trim(Copy(s,1,p-1)),Sym)
       and TryStrToInt(Copy(s,p+1,255),EvNo) then
      Exit;
  end;
  raise EReportException.Create('Bad population element:'+s);
end;

function TManifold.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMNode;
  Attr: TDOMAttr;
  J: TQuantumNumber;
  Sym, Index, EvNo: Integer;
  QnFmt: TQuantumNumberFormat;
begin
  Result := inherited ReadProperties(Node, False);
  if Mixture.ReadVersion < 5002528 then
    QnFmt := qft2J
  else
    QnFmt := QuantumNumberFormat;
  Child := Node.FirstChild;
  if Assigned(Child) then
    Molecule.FixupProperties.Free; { Required for parameter name decoding to work }
  { Read (and create) dynamic parameters containing population }
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if (Child is TDOMElement) and (CompareText(Child.NodeName, 'Parameter') = 0) then begin
      Attr := (Child as TDOMElement).GetAttributeNode('Name');
      if Assigned(Attr) then begin
        DecodeParameterName(Attr.NodeValue,J,Sym,EvNo,QnFmt);
        Index := PopulationIndex(J,Sym,EvNo,True);
        Result := Result + ReadFloatableDOMNode(Index,'0',Child as TDOMElement);
        Node.RemoveChild(Child);
      end;
    end;
    Child := NextChild;
  end;
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

function TManifold.HelpOnParameter(AIndex: Integer): string;
var
  Eig: TEigenvalues;
  J: TQuantumNumber;
  Sym, EvNo, LastSearched: Integer;
begin
  Result := '';
  if (AIndex=0) or (AIndex > nParameters) then
    Exit;
  if Mixture.Simulations[0].HasFields then
    Result := 'Population for M='
  else
    Result := 'Population for J=';
  Result := Result + ParameterNames[AIndex] + ' level';
  DecodeParameterName(ParameterNames[AIndex],J,Sym,EvNo,QuantumNumberFormat);
  Eig := FindEigenvalues(0,LastSearched,J,Sym,Mixture.Simulations[0].ExField);
  if Assigned(Eig) then
    if Eig.Size < EvNo then
      Result := Result + ' (non existent level)'
    else
      Result := Result + ':' + Eig.GetDominantBasisState(EvNo).StateLabel(Mixture.QuantumNumberFormat)
                       + ' E= ' + FloatToStr(Eig.E[EvNo]);
end;

{ TDistribution }

{ Special purpose transition moment that shows a population
  distribution only }

type
  TDistribution = class(TTransitionMoments)
  public
    procedure AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = False); override;
    function MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false):Boolean; override;
    function Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean; override;
    function MaxRank: TQuantumNumber; override;
  end;

var
  PopulationConnection: TDistribution;

procedure TDistribution.AddToSpectrum(Spectrum: TSpectrum; Ev1,
  Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment;
  Simulation: TSimulation;
  Wait: Boolean; PrintLevel: TPrintLevel;
  const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean);
var
  i, ColourI: Integer;
  te, wt, str: Double;
  Line: TLine;
begin
  if not Wait then
    Exit;
  if Ev1 <> Ev2 then
    raise Exception.Create('TDistribution.AddToSpectrum');
  TStart(AddMatrixToSpectrumTime);
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := nil;
  for i := 1 to Ev1.Size do begin
    { For each state }
    wt := Ev1.StatisticalWeight(i) * Ev1.Manifold.Molecule.Abundance * Ev1.Manifold.Molecule.Species.Concentration;
    Line.IndexL := i;
    Line.IndexU := -1;
    ColourI := Colour;
    if ColourI = clNone then
      ColourI := Ev1.DominantBasis[i].State.FindColour;
    te := Ev1.E[i];
    Str := wt;
    Line.Frequency := te;
    Line.Strength := Str;
    Line.Width := 0;
    if Assigned(Ev1.Widths) then
      Line.Width := Line.Width + Ev1.Widths[i];
    if ColourI = clNone then
      ColourI := clBlack;
    Line.Colour := ColourI;
    if LineListWanted(PrintLevel) then begin
      Line.CachedName := '';
      Spectrum.ShowLine(Line, Ev1.Manifold.Mixture.Simulations[0].OThreshold, Ev1.Manifold.Mixture.Units, Ev1.Manifold.Mixture.Simulations[0].PlotUnits, False, Problem);
      Line.CachedName := '';
    end;
    Spectrum.AddLine(Line, Ev1.Manifold.Mixture.Simulations[0], Ev1.Manifold.Mixture.Units, Problem);
  end;
  TEnd(AddMatrixToSpectrumTime);
end;

function TDistribution.Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean;
begin
  Result := (J1 = J2) and (Sym1 = Sym2);
end;

function TDistribution.MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean = false): Boolean;
begin
  raise NotYet;
end;

function TDistribution.MaxRank: TQuantumNumber;
begin
  Result := 0;
end;

{ TPartitionFunction }

constructor TPartitionFunction.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Active=True';
  inherited Create(AOwner);
  Name := 'Default Partition Function';
end;

function TPartitionFunction.EnergyZero(EstimatedOrigin: Double): Double;
begin
  Result := EstimatedOrigin;
end;

function TPartitionFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'If true, use this to generate the partition function'
  else
    Result:=inherited HelpOnSetting(AIndex);
end;

procedure TPartitionFunction.SetActive(const Index: Integer; const Value: Boolean);
var
  Molecule: TMolecule;
  I: Integer;
begin
  if Value and Assigned(Owner) then begin
    Molecule := Owner as TMolecule;
    for I := 0 to Molecule.ComponentCount-1 do
      if (Molecule.Components[i] is TPartitionFunction) and (Molecule.Components[I] <> Self) then
        TPartitionFunction(Molecule.Components[i]).Active := False;
    Molecule.FActivePartitionFunction := Self;
  end;
  SetFlag(Index, Value);
end;

{ TMolecule }

constructor TMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None,JAdjustSym=True,Abundance=1,Jmin=-1,Jmax=-1,BlockMatrix=True,AssumedOrigin=1e99';
  DimensionlessParameters := DimensionlessParameters + ',Abundance';
  FLastPartitionFunction := nil;
  FnSpinSpecies := 1;
  inherited Create(AOwner);
end;

destructor TMolecule.Destroy;
begin
  FreeAndNil(TransitionsMomentsCacheLock);
  FreeAndNil(PartitionFunctionCacheLock);
  FreeAndNil(FDefaultPartitionFunction);
  FActivePartitionFunction := nil;
  inherited Destroy;
end;

function TMolecule.EinsteinA(const Line: TLine; AUnits: TResonanceUnits): Double;
begin
  { TODO: May need adjustment in the presence of a field; needs documentation either way }
  Result := Line.Strength * IntensityFactor(Line.Frequency, AUnits, HzperMolecule)
            / Line.EigenvaluesU.Degeneracy(Line.IndexU)
end;

function TMolecule.GetManifold(AIndex: Integer): TManifold;
begin
  {$IFOPT R+}
  Result := FManifolds[AIndex] as TManifold;
  {$ELSE}
  Result := TManifold(FManifolds[AIndex]);
  {$ENDIF}
end;

function TMolecule.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  CheckManifolds.Free;
end;

procedure TMolecule.ScaleParameters(UFrom, UTo: TResonanceUnits);
var
  AsOrg: Double;
begin
  AsOrg := AssumedOrigin;
  inherited ScaleParameters(UFrom, UTo);
  { Not essential, but reset to large (default) value to avoid possibly writing
    it to file and confusing previous versions }
  if AsOrg > 1e50 then
    AssumedOrigin := 1e99;
end;

function TMolecule.CheckManifolds: Exception;
var
  i, j: Integer;
begin
  Result := nil;
  FnTransitionMoments := 0;
  FMaxRank := 0;
  { Compute nManifolds and check for duplicate names }
  for i := 0 to ComponentCount-1 do begin
    if Items[i] is TManifold then begin
      for j := i+1 to ComponentCount-1 do
        if (Items[j] is TManifold) and (Items[i].Name = Items[j].Name) then
          Result := EReportException.Create('Duplicate Manifold Names:'+ Items[i].Name)
    end else if Items[i] is TTransitionMoments then begin
      Inc(FnTransitionMoments);
      Result := TTransitionMoments(Items[i]).FixupName;
      FMaxRank := Max(MaxRank, TTransitionMoments(Items[i]).MaxRank);
    end;
    if Assigned(Result) then
      Break;
  end;
end;

function TMolecule.FixupProperties:Exception;
var
  n, i, ii, j, nLower, nNuc, nEquiv, SumNuclearSpin: Integer;
  Carry, New: TQuantumNumber;
  HalfSpinDetermined, Reject: Boolean;
  LowerLim, UpperLim: array of TQuantumNumber;
begin
  Result := nil;
  FActivePartitionFunction := nil;
  FnStates := 0;
  if nWorkers > 0 then begin
    if not Assigned(TransitionsMomentsCacheLock) then
      TransitionsMomentsCacheLock := TMRWLock.Create;
    TransitionsMomentsCacheLock.SetUpLock;
    if not Assigned(PartitionFunctionCacheLock) then
      PartitionFunctionCacheLock := TMRWLock.Create;
    PartitionFunctionCacheLock.SetUpLock;
  end;
  Result := CheckManifolds;
  if not Assigned(Result) then
    Result := inherited FixupProperties;
  if nManifolds = 0 then Exit;
  nLower := 0;
  if AssumedOrigin >= 1e50 then
    FActualAssumedOrigin := Manifold[0].LowestEstimatedOrigin { Not necessarily valid, but avoid awkward numbers }
  else
    FActualAssumedOrigin := AssumedOrigin;
  FLongestManifoldName := '';
  HalfSpinDetermined := False;
  FAllSinglet := True;
  for i := 0 to nManifolds-1 do begin
    Inc(FnStates, Manifold[i].nStates);
    if Manifold[i].nStates > 0 then
      if not HalfSpinDetermined then begin
        HalfSpinDetermined := True;
        FHalfSpin := Manifold[i].HalfSpin;
      end else if (Result = nil) and (FHalfSpin <> Manifold[i].HalfSpin) then
        Result := EReportException.Create('Odd and even numbers of electrons in ' + Name);
    if (AssumedOrigin >= 1e50) and Manifold[i].Initial and ((nLower=0) or (Manifold[i].LowestEstimatedOrigin < FActualAssumedOrigin)) then
      FActualAssumedOrigin := Manifold[i].LowestEstimatedOrigin;
    if Manifold[i].Initial then
      Inc(nLower);
    if Length(Manifold[i].Name) > Length(LongestManifoldName) then
      FLongestManifoldName := Manifold[i].Name;
    for j := 0 to Manifold[i].nStates - 1 do
      FAllSinglet := FAllSinglet and (Manifold[i].State[j].S=0);
  end;
  if AssumedOrigin >= 1e50 then
    FActualAssumedOrigin := ActivePartitionFunction.EnergyZero(FActualAssumedOrigin);
  { Make up HyperfineBasisStates }
  { First Determine total number of spin combinations }
  n := 1;
  SumNuclearSpin := 0;
  SetLength(HyperfineBasisStates,nNuclei+1);
  SetLength(HyperfineBasisStates[0],1); { nNuclei = 0 -> Empty array }
  SetLength(LowerLim, nNuclei);
  SetLength(UpperLim, nNuclei);
  nEquiv := 0;
  for i := 0 to nNuclei-1 do begin
    n := n * (Nuclei[i].Spin+1);
    Inc(SumNuclearSpin,Nuclei[i].Spin);
    if Nuclei[i].AsNext then begin
      { Two coupled nuclei - must treat as pair }
      Inc(nEquiv);
      SetLength(HyperfineBasisStates[i+1],0); { Flag invalid as can't omit just one of the spins }
      if not Assigned(Result) then
        if i = nNuclei-1 then
          Result := EReportException.Create('Last nucleus for '+ Name + ' must not be set equivalent')
        else if Nuclei[i+1].AsNext then
          Result := EReportException.Create('More than two equivalent nuclei not implemented in '+ Name);
      { For I12 = I1 + I2, just record quantum number rather than offset }
      LowerLim[i] := Abs(Nuclei[i].Spin-Nuclei[i+1].Spin);
      UpperLim[i] := Nuclei[i].Spin+Nuclei[i+1].Spin;
    end else begin
      if (i > 0) and Nuclei[i-1].AsNext then
        { Two equivalent nuclei require special handling }
        UpperLim[i] := UpperLim[i-1]
      else
        UpperLim[i] := Nuclei[i].Spin;
      LowerLim[i] := -UpperLim[i];
      SetLength(HyperfineBasisStates[i+1],n,i+1);
    end;
  end;
  if (SumNuclearSpin = 0) and (nEquiv = 0) then
    { Only discard spin 0 nuclei if they are not equivalent }
    SetLength(HyperfineBasisStates,1)
  else
    { Now make up different combinations }
    for nNuc := 1 to nNuclei do
      if not Nuclei[nNuc-1].AsNext then begin
        for i := 0 to nNuc-1 do
          if (i > 0) and Nuclei[i-1].AsNext then
            HyperfineBasisStates[nNuc,0,i] := Max(LowerLim[i], -HyperfineBasisStates[nNuc,0,i-1])
          else
            HyperfineBasisStates[nNuc,0,i] := LowerLim[i];
        for i := 1 to High(HyperfineBasisStates[nNuc]) do begin
          ii := i-1;
          repeat
            Carry := 2;
            Reject := False;
            for j := nNuc-1 downto 0 do begin
              New := HyperfineBasisStates[nNuc,ii,j] + Carry;
              if New > UpperLim[j] then
                New := LowerLim[j]
              else
                Carry := 0;
              HyperfineBasisStates[nNuc,i,j] := New;
              Reject := Reject or (Nuclei[j].AsNext and (Abs(HyperfineBasisStates[nNuc,i,j+1]) > New));
            end;
            ii := i;
          until not Reject;
        end;
      end;
  if nNuclei > 0 then
    eFn(nNuclei-1);
  FSymLength := 1;
  for i := 0 to LastSym do begin
    FSymLength := Max(FSymLength, Length(JAdjustedSymName(0,i)));
    FSymLength := Max(FSymLength, Length(JAdjustedSymName(2,i)));
    FSymLength := Max(FSymLength, Length(SymName(i)));
  end;
end;

function TMolecule.PartitionFunction(Simulation: TSimulation; var Overflow: Exception; out Emin: Double; RemoveSpins: Byte):Double;
var
  LocalOverflow: Exception;
  LastSize, i: Integer;
  SameConditions, SameSpins: Boolean;
  procedure TestForSameConditions;
  begin
    SameConditions := (Simulation.Temperature = FLastTemperature) and (Simulation.Tvib = FLastTvib)
                      and (Simulation.Tspin = FLastTspin) and SameFields(Simulation.ExField, FLastField);
    SameSpins := SameConditions and (Length(FLastPartitionFunction) > RemoveSpins) and (FLastPartitionFunction[RemoveSpins] >= 0);
  end;
begin
  if RemoveSpins <> 0 then
    RemoveSpins := Min(RemoveSpins, nNuclei+1);
  { Note - relies on ClearCache being triggered on TPartitionFunction change }
  try
    if Assigned(PartitionFunctionCacheLock) then
      PartitionFunctionCacheLock.BeginRead;
    TestForSameConditions;
  finally
    if Assigned(PartitionFunctionCacheLock) then
      PartitionFunctionCacheLock.EndRead;
  end;
  if SameSpins then begin
    if (Length(FLastProblem) > 0) and not Assigned(Overflow) then
      Overflow := EReportException.Create(FLastProblem);
    { Last problem is shared between different values of SpinSpecies }
  end else
    try
      if Assigned(PartitionFunctionCacheLock) then
        PartitionFunctionCacheLock.BeginWrite;
      TestForSameConditions;
      if not SameSpins then begin
        { Not calculated in the measn time }
        if SameConditions then
          LastSize := Length(FLastPartitionFunction)
        else
          LastSize := 0;
        SetLength(FLastPartitionFunction, Max(RemoveSpins+1,LastSize));
        SetLength(FLastEmin, Length(FLastPartitionFunction));
        SetLength(FSpinSpeciesAbundance,Length(FLastPartitionFunction));
        for i := LastSize to RemoveSpins do
          FLastPartitionFunction[i] := -99; { Make sure unused entries not set }
        FLastTemperature := Simulation.Temperature;
        FLastTvib := Simulation.Tvib;
        FLastTspin := Simulation.Tspin;
        FLastField := Simulation.ExField;
        if (nSpinSpecies > 1) and (FLastTspin >= 0) then begin
          SetLength(FSpinSpeciesAbundance[RemoveSpins],nSpinSpecies)
        end else
          SetLength(FSpinSpeciesAbundance[RemoveSpins],0);
        FLastPartitionFunction[RemoveSpins] := ActivePartitionFunction.CalculatePartitionFunction(Self,Simulation,LocalOverflow,FLastEmin[RemoveSpins],FSpinSpeciesAbundance[RemoveSpins],MaxDouble,False,False,False,RemoveSpins);
        if (FLastPartitionFunction[RemoveSpins] = 0) and not Assigned(LocalOverflow) then
          LocalOverflow := EReportException.Create('Zero partition function - underflow in Boltzmann factor?');
        if Assigned(LocalOverflow) then begin
          FLastProblem := LocalOverflow.Message;
          if Assigned(Overflow) then
            LocalOverflow.Free
          else
            Overflow := LocalOverflow;
        end else
          FLastProblem := '';
      end;
    finally
      if Assigned(PartitionFunctionCacheLock) then
        PartitionFunctionCacheLock.EndWrite;
    end;
  Emin := FLastEmin[RemoveSpins];
  Result := FLastPartitionFunction[RemoveSpins];
end;

procedure TMolecule.PrecomputeSymNames;
begin
  { No action by default }
end;

function TMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  with Result do begin
    MaxValue := High(Integer);
    if QnoType = qtM then
      MinValue := -MaxValue
    else
      MinValue := 0; { TODO: Switch on F number }
    StepSize := 2;
    Name := QnoTypeNames[Ord(QnoType)];
  end;
end;

function TMolecule.StartHyperfineCycle(Nuc: TNucleus; nBase, Rank: Integer): TIntegerArray;
var
  i: Integer;
begin
  if Nuc = nil then begin
    SetLength(Result, nBase);
    Result[nBase-1] := 0;
  end else begin
    SetLength(Result, nBase+nNuclei);
    Result[nBase+nNuclei-1] := 0; { TODO: Additional handling for field }
    for i := nNuclei-2 downto -1 do { -1 assumes last element in nBase is J }
      if (i >= 0) and Nuclei[i].AsNext then
        raise NotYet
      else if i > Nuc.NodeIndex then
        Result[nBase+i] := 0
      else
        Result[nBase+i] := -Rank
  end;
end;

function TMolecule.NextHyperfineCycle(Nuc: TNucleus; nBase, Rank: Integer; var Signature: TIntegerArray): Boolean;
var
  i: Integer;
begin
  { TODO: If spin is zero, some possibilities should be excluded }
  Result := False;
  if Nuc <> nil then
    for i := -1 to nNuclei-2 do
      if (i >= 0) and Nuclei[i].AsNext then
        raise NotYet
      else if i <= Nuc.NodeIndex then begin
        Inc(Signature[nBase+i]);
        if Signature[nBase+i] <= Rank then
          Exit;
        Signature[nBase+i] := -Rank
      end;
  Result := True;
end;

function TMolecule.StartHyperfineTransitionCycle(TransitionMoment: TTransitionMoment; nBase: Integer): TIntegerArray;
var
  i, Rank: Integer;
begin
  SetLength(Result, nBase+nNuclei);
  Rank := TransitionMoment.Rank div 2;
  Result[nBase+nNuclei-1] := -Rank;
  for i := nNuclei-2 downto -1 do { -1 assumes last element in nBase is J }
    if (i >= 0) and Nuclei[i].AsNext then
      raise NotYet
    else
      Result[nBase+i] := -Rank;
end;

function TMolecule.NextHyperfineCycle(TransitionMoment: TTransitionMoment; nBase: Integer; var Signature: TIntegerArray): Boolean;
var
  i, Rank: Integer;
begin
  { TODO: If spin is zero, some possibilities should be excluded }
  Result := False;
  Rank := TransitionMoment.Rank div 2;
  for i := -1 to nNuclei-1 do
    if (i >= 0) and Nuclei[i].AsNext then
      raise NotYet
    else begin
      Inc(Signature[nBase+i]);
      if Signature[nBase+i] <= Rank then
        Exit;
      Signature[nBase+i] := -Rank
    end;
  Result := True;
end;

function TMolecule.HyperfineMatrixElementType(Signature: TIntegerArray; nBase: Integer; IsBra: Boolean): string;
  procedure Add(n:Integer);
  begin
    if IsBra then begin
      if n > 0 then
        Result := Result + '+';
      if n <> 0 then
        Result := Result + IntToStr(n);
    end;
    Result := Result + ',';
  end;
var
  i: Integer;
begin
  Result := 'J';
  if IsBra then
    Result := '<'+Result;
  Add(Signature[nBase-1]);
  for i := 0 to nNuclei-1 do
    if nBase + i < Length(Signature) then begin
      if (i >= 0) and Nuclei[i].AsNext then begin
        Result := Result + Format('I%d%d=', [i+1,i+2]);
        if IsBra then
          Result := Result + IntToStr(Signature[i] div 1000)
        else
          Result := Result + IntToStr(Signature[i] mod 1000);
        Result := Result + ',';
      end else begin
        Result := Result + 'F';
        if i < nNuclei-1 then
          Result := Result + IntToStr(i+1);
        Add(Signature[nBase+i]);
      end;
    end;
end;

function TMolecule.HyperfineCouplingMatrixElement(Nuc: TNucleus; nBase,
  Rank: Integer; Signature: TIntegerArray; eJ: TExprNode): TExprNode;
var
  i: Integer;
  eI, FF, FFp, FF1, FF1p, sJ: TExprNode;
begin
  Result := nil;
  eI := nil;
  FF := eFn(0);
  FFp := nil;
  FF1p := nil;
  try
    for i := nNuclei-1 downto 0 do begin
      { TODO: AsNext case }
      if i = 0 then
        FF1 := eJ
      else
        FF1 := eFn(i);
      { nBase-1 points to J }
      FF1p := Expr.EB([FF1, '+', Signature[i+nBase-1]]);
      if i >= Nuc.NodeIndex then begin
        if Signature[i+nBase] <> 0 then
          Exit;
        if i = Nuc.NodeIndex then begin
          eI := Halve(Nuc.Spin);
          Result := Expr.EB(['sixjU', FF, eI, FF1p, Rank, FF1, eI]);
          if Result.EqualTo(0) then begin
            FreeAndNil(Result);
            Exit;
          end;
          Result := EN([-1, '^', EB([FF1, '+', eI, '+', FF]), '*', Result]);
        end;
      end else begin
        eI := Halve(Nuclei[i].Spin);
        sJ := Expr.EB(['sixjU', FF1p, FFp, eI, FF, FF1, Rank]);
        if Result.EqualTo(0) then begin
          sJ.Free;
          FreeAndNil(Result);
          Exit;
        end;
        Result := EN([-1, '^', EB([FF1p, '+', eI, '+', FF, '+', Rank]), '*', Result, '*', sJ]);
        sJ := Expr.EN([Expr.EB([FFp, '*', 2, '+', 1]), '*', Expr.EB([FF, '*', 2, '+', 1])]);
        Result := Expr.EN([sJ, 'sqrt', '*', Result]);
      end;
      eI.Free;
      FFp.Free;
      FFp := FF1p;
      FF1p := nil;
      FF := FF1;
    end;
  finally
    FFp.Free;
    FF1p.Free;
  end;
end;

function TMolecule.HyperfineTransitionMatrixElement(
  TransitionMoment: TTransitionMoment; nBase: Integer;
  Signature: TIntegerArray; eJ: TExprNode): TExprNode;
var
  Nuc, Rank, NuclearDipole, dF: Integer;
  FF, FFp, FF1, FF1p, sJ, Ieff: TExprNode;
begin
  Result := nil;
  Ieff := nil;
  FF := eFn(0);
  FFp := Expr.EB([FF, '+', Signature[nNuclei+nBase-1]]);
  FF1p := nil;
  if TransitionMoment is TNuclearMagneticMoment then
    NuclearDipole := TNuclearMagneticMoment(TransitionMoment).Nucleus-1
  else
    NuclearDipole := -1;
  Rank := TransitionMoment.Rank div 2;
  try
    Nuc := nNuclei-1;
    if NuclearDipole > Nuc then
      Exit; { Removed spin }
    Result := TIntegerNode.Create(1);
    while Nuc >= 0 do begin
      { TODO: AsNext case }
      if Nuc = 0 then
        FF1 := eJ
      else
        FF1 := eFn(Nuc);
      dF := Signature[Nuc+nBase-1];
      FF1p := Expr.EB([FF1, '+', dF]);
      if (Nuc > 0) and Nuclei[Nuc-1].AsNext then begin
        if NuclearDipole >= 0 then
          raise NotYet;
        { Coupled spins; need to use total nuclear spin for pair }
        raise NotYet;
      end else
        Ieff := Halve(Nuclei[Nuc].Spin);
      if NuclearDipole = Nuc then
        if dF = 0 then begin
          sJ := Expr.EB(['sixjU', Ieff, FFp, FF1, FF, Ieff, Rank]);
          if sJ.EqualTo(0) then begin
            sJ.Free;
            FreeAndNil(Result);
          end else begin
            Result := Expr.EN([-1, '^', EB([FF1p, '+', Ieff, '+', FF, '+', Rank+1]), '*', sJ, '*', Result]);
            sJ := Expr.EN([Expr.EB([FFp, '*', 2, '+', 1]), '*', Expr.EB([FF, '*', 2, '+', 1])]);
            Result := Expr.EN([sJ, 'sqrt', '*', Result, '*', ReducedMatrixElementE(2, Ieff)]);
          end;
        end else
          FreeAndNil(Result)
      else if (NuclearDipole >= 0) and (NuclearDipole < Nuc) then begin
        if dF <> 0 then
          FreeAndNil(Result);
      end else begin
        sJ := Expr.EB(['sixjU', FF1p, FFp, Ieff, FF, FF1, Rank]);
        Result := Expr.EN([-1, '^', EB([FF1p, '+', Ieff, '+', FF, '+', Rank]), '*', sJ, '*', Result]);
        sJ := Expr.EN([Expr.EB([FFp, '*', 2, '+', 1]), '*', Expr.EB([FF, '*', 2, '+', 1])]);
        Result := Expr.EN([sJ, 'sqrt', '*', Result]);
      end;
      if not Assigned(Result) or Result.EqualTo(0) then begin
        FreeAndNil(Result);
        Exit;
      end;
      FreeAndNil(Ieff);
      FreeAndNil(FFp);
      FFp := FF1p;
      FF1p := nil;
      FF := FF1;
      Dec(Nuc);
    end;
  finally
    Ieff.Free;
    FFp.Free;
    FF1p.Free;
  end;
end;

function TMolecule.AddToSpectrum(Spectrum: TSpectrum; PrintLevel: TPrintLevel;
  const Excitation: TExcitation; var Problem: Exception;
  Simulation: TSimulation; J1Sel: Integer; Sym1Sel: Integer; J2Sel: Integer;
  Sym2Sel: Integer; Manifold1Sel: TManifold; Manifold2Sel: TManifold;
  DeltaJ: TQuantumNumber; TransitionMomentsSel: TTransitionMoments;
  TransitionMomentSel: TTransitionMoment; State1Sel: TState; State2Sel: TState;
  IncludePerturbations: Boolean; IncludeZero: Boolean; RemoveSpins: Byte
  ): Boolean;
var
  i, j, Sym1, Sym2: Integer;
  J1, J1Low, J1Up: TQuantumNumber;
  J2, J2Low, J2Loww, J2Up, J2Upp: TQuantumNumber;
  Prog1, Prog1Max, Prog2, Prog2Max: Integer; { for progress display }
  MaxDj: TQuantumNumber;
  Manifold1, Manifold2: TManifold;
  Connection: TTransitionMoments;
  Ev1, Ev2: TEigenvalues;
  nInitial, nConnections: Integer;
  HasFields: Boolean;
  PassMin, Pass: Integer;
  Status: string;
  MatElemTypes: TMatElemTypes;
begin
  if Abundance = 0 then Exit;
  nInitial := 0;
  nConnections := 0;
  HasFields := Simulation.HasFields;
  if not HasFields then begin
    if J1Sel < 0 then
      J1Sel := Low(Integer);
    if J2Sel < 0 then
      J2Sel := Low(Integer);
  end;
  MatElemTypes := [meHrot, meWidth, meAllDerivatives, meFloatedDerivatives];
  { If parallel processing, make a prepass to fill eigenvalue and transition matrix caches }
  if (nWorkers > 0) and (PrintLevel <= plLineList) then
    PassMin := 1
  else
    PassMin := 3;
  for Pass := PassMin to 3 do begin
    if (Pass = 3) and Assigned(TransitionMomentSel) then
      Break; { If transition moment specified, don't cache transition matrices, so third pass must be skipped }
    for i := 0 to nManifolds-1 do
      if (Manifold[i] = Manifold1Sel)
         or (not Assigned(Manifold1Sel) and Manifold[i].Initial ) then begin
        { For each manifold marked as initial }
        Manifold1 := Manifold[i];
        Inc(nInitial);
        for j := 0 to nManifolds-1 do
          if (Manifold[j] = Manifold2Sel)
             or (not Assigned(Manifold2Sel) and ((j >= i) or not Manifold[j].Initial)) then begin
            { For every other manifold, except those we have already started from }
            Manifold2 := Manifold[j];
            if Spectrum.IsDistribution then
              if Manifold1 = Manifold2 then begin
                Connection := PopulationConnection;
                { Ignore selection of "other" state }
                TransitionMomentSel := nil;
                J2Sel := J1Sel;
                Sym2Sel := Sym1Sel;
                State2Sel := State1Sel;
              end else
                Connection := nil { Only need one state for a population distribution }
            else
              Connection := TransitionMoments[i, j];
            if Assigned(Connection)
              and ((Connection = TransitionMomentsSel) or not Assigned(TransitionMomentsSel)) then begin
              Inc(nConnections);
              if Assigned(TransitionMomentSel) then
                MaxDj := TransitionMomentSel.Rank
              else
                MaxDj := Connection.MaxRank;
              if J1Sel <> Low(Integer) then begin
                J1Low := J1Sel;
                J1Up := J1Sel;
                if J2Sel <> Low(Integer) then begin
                  J2Low := J2Sel;
                  J2Up := J2Sel;
                end else if Abs(DeltaJ) < 100 then begin
                  J2Low := J1Low + DeltaJ;
                  J2Up := J2Low;
                  if not HasFields and (J2Low < 0) then Exit;
                end else begin
                  J2Low := J1Sel - MaxDj;
                  if not HasFields then
                    J2Low := SmallestProjection(J2Low);
                  J2Up := J1Sel + MaxDj;
                end;
              end else if J2Sel <> Low(Integer) then begin
                J2Low := J2Sel;
                J2Up := J2Sel;
                if Abs(DeltaJ) < 100 then begin
                  J1Low := J2Low - DeltaJ;
                  J1Up := J1Low;
                  if not HasFields and (J1Low < 0) then Exit;
                end else begin
                  J1Low := J2Sel - MaxDj;
                  if not HasFields then
                    J1Low := SmallestProjection(J2Low);
                  J1Up := J2Sel + MaxDj;
                end;
              end else begin
                J1Up := Manifold1.FindJmax(-1);
                if HasFields then
                  J1Low := -J1Up
                else
                  J1Low := Manifold1.FindJmin(HalfSpin);
                J2Up := Manifold2.FindJmax(J1Up + MaxDj);
                if HasFields then
                  J2Low := -J2Up
                else
                  J2Low := Manifold2.FindJmin(J1Low - MaxDj);
              end;
              J1 := J1Low;
              case Pass of
                1: Status := 'Energy Levels ';
                2: Status := 'Transition Matrices ';
                3: Status := '';
              end;
              PushProgress(Status + Manifold2.Name + ' - ' + Manifold1.Name, J1Up div 2);
              try
                while J1 <= J1Up do begin
                  Prog1 := (J1 - J1Low) div 2;
                  Prog1Max := (J1Up - J1Low + 2) div 2;
                  if HasFields then
                    Sym1 := 0
                  else if Sym1Sel >= 0 then begin
                    Sym1 := Sym1Sel;
                    if JAdjustSym then
                      Sym1 := JAdjustedSym(J1, Sym1);
                  end else begin
                    Sym1 := Manifold1.FirstSym;
                    Prog1 := Prog1*Manifold1.nSym + Sym1;
                    Prog1Max := Prog1Max*Manifold1.nSym;
                  end;
                  while Sym1 >= 0 do begin
                    Ev1 := Manifold1.Eigenvalues(J1, Sym1, MatElemTypes, Simulation.ExField, Pass >= 2, PrintLevel, State1Sel, IncludePerturbations, RemoveSpins);
                    if Ev1.Size > 0 then begin
                      if Abs(DeltaJ) < 100 then begin
                        J2 := J1 + DeltaJ;
                        J2Upp := J2;
                      end else begin
                        J2 := J1-MaxDj;
                        J2Upp := Min(J2Up,J1+MaxDj);
                      end;
                      if J2 < J2Low then J2 := J2Low;
                      J2Loww := J2;
                      while J2 <= J2Upp do begin
                        Prog2 := (J2 - J2Loww) div 2;
                        Prog2Max := (J2Upp - J2Loww + 2 ) div 2;
                        if HasFields then
                          Sym2 := 0
                        else if Sym2Sel >= 0 then begin
                          Sym2 := Sym2Sel;
                          if JAdjustSym then
                            Sym2 := JAdjustedSym(J2, Sym2);
                        end else begin
                          Sym2 := Manifold2.FirstSym;
                          Prog2 := Prog2*Manifold2.nSym + Sym2;
                          Prog2Max := Prog2Max*Manifold2.nSym;
                        end;
                        SetProgress((Prog1*Prog1Max+Prog2)*2,Prog1Max*Prog2Max*2);
                        while Sym2 >= 0 do begin
                          if HasFields or Connection.Allowed(J1, Sym1, J2, Sym2) then begin
                            SetProgress((Prog1*Prog1Max+Prog2)*2,Prog1Max*Prog2Max*2);
                            Ev2 := Manifold2.Eigenvalues(J2, Sym2, MatElemTypes, Simulation.ExField, Pass >= 2, PrintLevel, State2Sel, IncludePerturbations, RemoveSpins);
                            if (Ev2.Size > 0) and (Pass >= 2) then
                              Connection.AddToSpectrum(Spectrum, Ev1, Ev2, TransitionMomentSel, Simulation, (Pass = 3) or Assigned(TransitionMomentSel), PrintLevel, Excitation, Problem, IncludeZero);
                            SetProgress((Prog1*Prog1Max+Prog2)*2+1,Prog1Max*Prog2Max*2);
                          end;
                          Inc(Prog2);
                          if HasFields or (Sym2Sel >= 0) then
                            Break
                          else
                            Sym2 := Manifold2.NextSym(Sym2);
                        end;
                        Inc(J2, 2);
                      end;
                    end;
                    Inc(Prog1);
                    if HasFields or (Sym1Sel >= 0) then
                      Break
                    else
                      Sym1 := Manifold1.NextSym(Sym1);
                  end;
                  Inc(J1, 2);
                end;
              finally
                PopProgress;
              end;
            end;
          end;
      end;
  end;
  if (nManifolds > 0) and (nInitial = 0) then
      raise EReportException.Create('No Initial States set for '+ Name);
  Result := nConnections > 0;
end;

function TMolecule.ManifoldWithName(AName: string): TManifold;
var
  i: Integer;
begin
  Result := nil;
  { NB: can't use FindComponent as Name property has been overridden }
  for i := 0 to ComponentCount-1 do
    if (Items[i] is TManifold) and (TManifold(Items[i]).Name = AName) then begin
      Result := TManifold(Items[i]);
      Exit;
    end;
end;

function TMolecule.FindTransitionMoments(ManifoldA, ManifoldB: TManifold): TTransitionMoments;
var
  i: Integer;
  tms: TTransitionMoments;
begin
  Result := nil;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      tms := TTransitionMoments(Items[i]);
      if ((tms.Bra = ManifoldA) and (tms.Ket = ManifoldB))
         or ((tms.Bra = ManifoldB) and (tms.Ket = ManifoldA)) then begin
          if Assigned(Result) then
            raise EReportException.Create('Duplicate transition moments <' + ManifoldA.Name + '|' + Result.PairName + '|' + ManifoldB.Name + '> in '+Name);
          Result := TTransitionMoments(Items[i]);
      end;
    end;
end;

function TMolecule.GetTransitionMomentsFunction(AIndex, BIndex: Integer): TTransitionMoments;
begin
  { TODO: Speed up search? }
  Result := FindTransitionMoments(Manifold[AIndex], Manifold[BIndex]);
end;

function TMolecule.Species: TSpecies;
begin
  Result := Owner as TSpecies;
end;

function TMolecule.SpinSpeciesAbundance(Simulation: TSimulation; Index: Integer; var Overflow: Exception; RemoveSpins: Byte): Double;
var
  Emin: Double;
begin
  if nSpinSpecies > 1 then begin
    PartitionFunction(Simulation,Overflow,Emin); { Ensure caches filled }
    Result := FSpinSpeciesAbundance[Min(nNuclei+1,RemoveSpins)][Index-1];
  end else
    Result := 1;
end;

function TMolecule.GetMixture: TMixture;
begin
  Result := RootNode as TMixture;
end;

function TMolecule.GetTransitionMoments(AIndex: Integer): TTransitionMoments;
var
  i: Integer;
begin
  Result := nil;
  if nTransitionMoments < 1 then Exit;
  for i := 0 to ComponentCount-1 do
    if Items[i] is TTransitionMoments then begin
      Dec(AIndex);
      if AIndex < 0 then begin
        Result := TTransitionMoments(Items[i]);
        Exit;
      end;
    end;
end;

procedure TMolecule.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache(True);
end;

procedure TMolecule.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  FnManifolds := AdjustNodeIndexes(TManifold,FManifolds,AComponent,Operation);
end;

function TMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-' { Symmetry not used, typically for external field }
  else
    Result := IntToStr(Sym);
end;

function TMolecule.JAdjustedSymName(J: TQuantumNumber; Sym: Integer): string;
begin
  if JAdjustSym then
    Result := JAdjustedSymNameBase(J, Sym)
  else
    Result := SymName(Sym);
end;

function TMolecule.JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-' { Symmetry not used, typically for external field }
  else
    Result := IntToStr(Sym); { Default to number }
end;

function TMolecule.BothSymNames(J: TQuantumNumber; Sym: Integer): string;
var
  Other: string;
begin
  if JAdjustSym then begin
    Result := JAdjustedSymName(J, Sym);
    Other := SymName(Sym);
    if Result <> Other then
      Result := Result + ' (' + Other + ')';
  end else
    Result := SymName(Sym);
end;

function TMolecule.JDeAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
var
  i: Integer;
begin
  { Reverse effects of JAdjustedSym }
  { This can be overridden at the individual molecule level to make it more efficient }
  if JAdjustedSym(J, Sym) = Sym then
    Result := Sym { No adjustment - easy }
  else begin
    for i := 0 to LastSym do
      if (i <> Sym) and (JAdjustedSym(J, i) = Sym) then begin
        Result := i;
        Exit;
      end;
    raise Exception.CreateFmt('JDeAdjustedSym fails for %s: %d %d', [Name, J, Sym]);
  end;
end;

function TMolecule.ConvertDefaultSymmetry(J: TQuantumNumber;  DefaultSymmetry: Integer): Integer;
begin
  if DefaultSymmetry >= 1000 then
    Result := DefaultSymmetry div 1000 - 1
  else if DefaultSymmetry >= 0 then
    Result := JAdjustedSym(J, DefaultSymmetry)
  else
    Result := -1;
end;

procedure TMolecule.SetnNuclei(const Index, Value: Integer);
var
  i, j: Integer;
begin
  inherited SetSetting(Index, Value);
  SetLength(Nuclei,Value);
  { Make sure each state has the right number of nuclei }
  for i := 0 to ComponentCount-1 do
    if Items[i] is TManifold then
      with TManifold(Items[i]) do
        for j := 0 to ComponentCount-1 do
          if Items[j] is TState then
            TState(Items[j]).AdjustNuclei(True);
end;

function TMolecule.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  case AIndex of
    1: if nNuclei <= 0 then Exit;
    2,3: if Settings[AIndex] <= 0 then Exit; { Jmin, Jmax }
  end;
  Result := inherited SettingToPrettyStr(AIndex);
end;

procedure TMolecule.SetUpTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray);
begin
  { None }
end;

procedure TMolecule.SetTransitionFunctionVariables(TransitionFunction: TTransitionFunction; var ExtraVariables: TVariableArray;
                                                   ABra, AKet: TBasisState; SignA, SignB: Integer);
begin
  { None }
end;

procedure TMolecule.SetUpPopulationFunctionVariables(
  PopulationFunction: TPopulationFunction; var ExtraVariables: TVariableArray);
begin
  { None }
end;

procedure TMolecule.SetPopulationFunctionVariables(
  PopulationFunction: TPopulationFunction; var ExtraVariables: TVariableArray;
  AState: TBasisState);
begin
  { None }
end;

function TMolecule.GetActivePartitionFunction: TPartitionFunction;
var
  i: Integer;
begin
  if Assigned(FActivePartitionFunction) and not FActivePartitionFunction.Active then
    FActivePartitionFunction := nil;
  if not Assigned(FActivePartitionFunction) then
    for i := 0 to ComponentCount-1 do
      if (Components[i] is TPartitionFunction)
         and TPartitionFunction(Components[i]).Active then begin
        if Assigned(FActivePartitionFunction) then
          raise EReportException.Create('Multiple Active Partition Functions');
        FActivePartitionFunction := TPartitionFunction(Components[i]);
      end;
  if not Assigned(FActivePartitionFunction) then begin
    if not Assigned(FDefaultPartitionFunction) then
      FDefaultPartitionFunction := TDefaultPartitionFunction.Create(nil);
    FActivePartitionFunction := FDefaultPartitionFunction;
  end;
  Result := FActivePartitionFunction;
end;

function TMolecule.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 4;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TMolecule.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [2,3];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Number of nuclei to calculate hyperfine structure for.'
  else if AIndex = 2 then
    Result := 'Minimum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 3 then
    Result := 'Maximum J to use in calculation - set to -1 (default) to use value from species.'
  else if AIndex = 4 then
    Result := ColourHelpString
  else if AIndex = 5 then
    Result := 'If set, energy level plots and Fortrat diagrams will take account of the alternation in symmetry with J that is often found. For example in linear molecules an energy level will typically alternate between + and - parity with J (hence the e/f notation).'
  else if AIndex = 6 then
    Result := 'Select to split Hamiltonian matrix into unconnected blocks, if any exist.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMolecule.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Abundance of this isotope; default is 1'
  else if AIndex = 2 then
    Result := 'Energy origin for Boltzmann calculations; 1e99 implies automatic estimate. Currently '+FloatToStr(ActualAssumedOrigin)
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TMolecule.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  case nChildTypes - AIndex of
    3: Result := TTransitionMoments;
    2: Result := TInterpolatedPartitionFunction;
    1: Result := TVibrationalPartitionFunction;
  else
    Result := inherited GetChildTypes(AIndex);
  end;
end;

function TMolecule.nChildTypes: Integer;
begin
  Result := inherited nChildTypes + 3;
end;

procedure TMolecule.ClearCache(ClearChildren: Boolean);
begin
  inherited ClearCache(ClearChildren);
  FLastPartitionFunction := nil;
  FLastEmin := nil;
end;

function TMolecule.TryStrToAnySymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
var
  i: Integer;
begin
  for i := 0 to LastSym do
    if CompareText(s,SymName(i)) = 0 then begin
      Result := True;
      Sym := i;
      Exit;
    end;
  Result := TryStrToJAdjustedSymName(J, s, Sym);
  if not Result then
    if TryStrToInt(s, i) and (i <= LastSym) then begin
      Result := True;
      Sym := i;
    end;
end;

function TMolecule.TryStrToJAdjustedSymName(J: TQuantumNumber; const s: string; out Sym: Integer): Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to LastSym do
    if CompareText(JAdjustedSymNameBase(J,i), s) = 0 then begin
      Result := True;
      Sym := i;
      Exit;
    end;
  if s = '-' then begin
    Result := True;
    Sym := -1;
  end;
end;

function TMolecule.TryStrToSimpleQuantumNumber(const str: string; var Value: TQuantumNumber; AQuantumNumberFormat: TQuantumNumberFormat; var SpecialQN: Integer; QnoIndex: Integer): Boolean;
begin
  { Overridden by TLinearMolecule to handle F1e, etc}
  { Overridden by TSymmetricMolecule to allow default symmetry to be changed }
  { SpecialQN indicates PGOPHER extension to simple format; updated to true if used }
  Result := TryStrToQuantumNumber(str, Value, AQuantumNumberFormat);
end;

procedure TMolecule.FixSimpleFormat(var s: string);
begin
  { Convert s from state label format to simple line list format }
  { Linear molecules need a correction }
end;

function TMolecule.FixSym(J: TQuantumNumber; Sym: Integer; InputPgopherVersion: Integer): Integer;
begin
  { Fix symmetry as required for older pgopher version .lin files }
  Result := Sym;
end;

{ TSpecies }

function TSpecies.GetIsotopomer(AIndex: Integer): TMolecule;
begin
  Result := TMolecule(Items[AIndex]);
end;

constructor TSpecies.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None,Concentration=1,Jmin=-1,Jmax=50';
  DimensionlessParameters := 'Concentration,' + DimensionlessParameters;
  inherited Create(AOwner);
end;

function TSpecies.AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation):Boolean;
var
  i: Integer;
begin
  Result := False;
  for i := 0 to Count-1 do
    if Isotopomer[i].Abundance <> 0 then
      { Note that we want to call AddToSpectrum for every species }
      Result := Isotopomer[i].AddToSpectrum(Spectrum, (Owner as TMixture).PrintLevel, RandomExcitation, Problem, Simulation)
                or Result ;
end;

function TSpecies.FindMolecule(AName: string): TMolecule;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Count-1 do
    if CompareText(Isotopomer[i].Name, AName) = 0 then
      if Assigned(Result) then
        raise EReportException.CreateFmt('Duplicate Isotopomer name in %s: %s', [AName, Result.Name])
      else
        Result := Isotopomer[i];
end;

function TSpecies.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TSpecies.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TSpecies.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: Integer;
begin
  if Count = 0 then
    Result := NullQnoInfo
  else
    Result := Isotopomer[0].QnoInfo(QnoType);
  for i := 1 to Count - 1 do
    MergeQno(Result, Isotopomer[i].QnoInfo(QnoType), QnoType);
end;

procedure TSpecies.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache(True);
end;

function TSpecies.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  if IsSettingQuantum(AIndex) and (Settings[AIndex] < 0) then
    Exit;
  Result := inherited SettingToPrettyStr(Aindex);
end;

function TSpecies.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Minimum J to use in calculation - set to -1 to use lowest J.'
  else if AIndex = 2 then
    Result := 'Maximum J to use in calculation.'
  else if AIndex = 3 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSpecies.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative concentration of species'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TMixture }

function TMixture.GetSettingsNames(AIndex: Integer;
  out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = 5 then
    Result := @QuantumNumberFormatNames
  else if AIndex = 2 then
    Result := @PrintLevelNames;
end;

function TMixture.GetSimulation(AIndex: Integer): TSimulation;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    if Items[i] is TSimulation then begin
      if AIndex = 0 then begin
        Result := TSimulation(Items[i]);
        Exit;
      end;
      Dec(AIndex);
    end;
  Result := nil;
end;

function TMixture.GetSpecies(AIndex: Integer): TSpecies;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    if Items[i] is TSpecies then begin
      if AIndex = 0 then begin
        Result := TSpecies(Items[i]);
        Exit;
      end;
      Dec(AIndex);
    end;
  Result := nil;
end;

function TMixture.GetQuantumNumberFormat(AIndex: Integer): TQuantumNumberFormat;
begin
  Result := TQuantumNumberFormat(GetSetting(AIndex));
end;

constructor TMixture.Create(AOwner: TComponent);
begin
  FReadVersion:= DecodePgopherVersion(VersionInfo);
  SavedProperties := SavedProperties + 'Precision=4,PrintLevel=plMinimal,SVDThresh=1e-6';
  SavedProperties := SavedProperties + ',AutoReplot=True,QuantumNumberFormat=J,FitCycles=1';
  SavedProperties := SavedProperties + ',ScaleChanges=1,SVDlimit=5,SmallE='+FloatToStr(DefaultSmallE);
  SavedProperties := SavedProperties + ',ScaledParameters=True';
  DimensionlessParameters := DimensionlessParameters + ',ScaleChanges,SVDThresh,';
  inherited;
  ClearUndo;
  TSimulation.Create(Self); { Must have at least one simulation }
  FnSimulations := 1;
  Modified := False;
end;

destructor TMixture.Destroy;
begin
  SetLineList(nil);
  FreeAndNil(FLoadedFormSettings);
  inherited Destroy;
end;

procedure TMixture.PreCalculatePartitionFunctions(Simulation: TSimulation; var Overflow: Exception);
var
  i, j: Integer;
  Emin: Double;
begin
  for i := 0 to nSpecies-1 do
    for j := 0 to Species[i].Count-1 do
      Species[i].Isotopomer[j].PartitionFunction(Simulation, Overflow, Emin);
end;

procedure TMixture.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i: Integer;
begin
  WriteLn(F, Indent, VersionInfo);
  PrettyPrintHeader(F,PrintLevel,Indent);
  Indent := Indent + '  ';
  PrettyPrintSettings(F, PrintLevel, Indent);
  PrettyPrintParameters(F, PrintLevel, Indent);
  for i := 0 to nSpecies - 1 do
    Species[i].PrettyPrint(F, PrintLevel, Indent);
  if Assigned(FCorrelations) and (Correlations.nEntries > 1) then
    Correlations.PrettyPrint(F,PrintLevel,Indent);
  for i := 0 to nSimulations-1 do
    Simulations[i].PrettyPrint(F,PrintLevel,Indent);
  WriteLn(F, Indent, FundamentalConstantsSummary);
  { Do not print form parameters }
  WriteLn(F);
end;

procedure TMixture.ConvertToPerturbations;
var
  i, j, k: Integer;
begin
  for i := 0 to nSpecies-1 do
    with Species[i] do
      for j := 0 to Count-1 do
        with Isotopomer[j] do
          for k := 0 to nManifolds-1 do
            Manifold[k].ConvertToPerturbations;
  FixupProperties;
end;

function TMixture.Correlations: TCorrelations;
begin
  if not Assigned(FCorrelations) then begin
    { FCorrelations := } TCorrelations.Create(Self);
    FCorrelations.Name := 'Correlations';
  end;
  Result := FCorrelations
end;

function TMixture.FixupProperties:Exception;
var
  i, j: Integer;
  LManName: string;
begin
  TStart(FixupPropertiesTime);
  Result := inherited FixupProperties;
  FnMolecules := 0;
  FnSpecies := 0;
  FLongestMoleculeName := '';
  FLongestSymLength := 1;
  LManName := '';
  FnSimulations := 0;
  for i := 0 to Count - 1 do
    if Items[i] is TSpecies then
      with TSpecies(Items[i]) do begin
        Inc(FnSpecies);
        Inc(FnMolecules, Count);
        for j := 0 to Count-1 do
          with Isotopomer[j] do begin
            if Length(LongestMoleculeName) < Length(Name) then
              FLongestMoleculeName := Name;
            if Length(LongestManifoldName) > Length(LManName) then
              LManName := LongestManifoldName;
            if SymLength > FLongestSymLength then
              FLongestSymLength := SymLength;
          end;
      end else if Items[i] is TSimulation then
        Inc(FnSimulations);
  FLongestManifoldName := LManName;
  TEnd(FixupPropertiesTime);
end;

procedure TMixture.SetModified(const Value: Boolean);
begin
  inherited;
  if Value then
    ClearCache(False);
end;

procedure TMixture.SetQuantumNumberFormat(const Index: Integer; const Value: TQuantumNumberFormat);
begin
  SetSetting(Index, Ord(Value));
end;

procedure TMixture.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if ((AIndex = 4 { BasisOrder }) or (AIndex = 9 {AllowComplex}))
     and (Settings[AIndex] <> AValue) then
    ClearCache(True);
  inherited;
end;

function TMixture.SubPropInfo(const Token: string; out ErrMsg: string; ImplicitDot:Boolean): TValueInfo;
var
  AltMsg: string;
begin
  Result := inherited SubPropInfo(Token, ErrMsg, ImplicitDot);
  { Backward compatibility - treat properties of first Simulation object
    as properties of mixture }
  if Result.MNode = nil then
    Result := Simulations[0].SubPropInfo(Token, AltMsg);
  if (Result.MNode = nil) and (Pos('.',Token) < 0) then
    ErrMsg := AltMsg; { May be more meaningfull }
end;

function TMixture.FindMolecule(AName: string): TMolecule;
var
  Molecule: TMolecule;
  i: Integer;
begin
  Result := nil;
  for i := 0 to nSpecies-1 do begin
    Molecule := Species[i].FindMolecule(AName);
    if Assigned(Molecule) then begin
      if Assigned(Result) then
        raise EReportException.Create('Duplicate molecule name: ' + Result.Name);
      Result := Molecule;
    end;
  end;
end;

const
  MixtureSettingNames: array[1..16] of string = (
'Units for calculation - valid values are cm1, MHz, Kelvin and eV. The key change this makes is in calculating Boltzmann factors from temperature.',
'Controls the amount of printout in the Log window.',
'Precision of printout (decimal places after the point).',
'How to order the basis states making up a basis. Valid values are: boDefault, boDiagonal and boStateDiagonal.',
'How to display quantum numbers.',
'Number of least squares fit cycles to do on each press of a "fit" button.',
'True to show more info on fit',
'If false, automatic recalculation of the simulation is disabled in some circumstances. Use if the program is running slowly.',
'Set to use complex arithmetic if necessary.',
'Disable use of line lists; may be required for very large calculations.',
'Show estimated uncertainties of line positions in line lists.',
'Show derivatives in line listings.',
'Number of contributions to show for each SVD vector; 0 for all',
'Include sub-basis labels in quantum number display',
'Lines with zero calculated intensity are permitted in fitting',
'Scale parameters when fitting using derivatives'
);

function TMixture.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureSettingNames))
     and (AIndex <= High(MixtureSettingNames)) then
    Result := MixtureSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMixture.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and (MNodeClass = TCorrelations) then
    Result := not Assigned(FCorrelations);
end;

const
 MixtureParameterNames: array [1..3] of string = (
'Discard Singular Value Decomposition eigenvalues below this fraction of the largest eigenvalue; <0 to use QR factorisation.',
'Scale changes produced by fitting by this - normally 1',
'Matrix elements smaller than this are discarded'
);

function TMixture.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(MixtureParameterNames))
     and (AIndex <= High(MixtureParameterNames)) then
    Result := MixtureParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TMixture.SetBasisOrder(const Index: Integer; const AValue: TBasisOrder);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetBasisOrder(const Index: Integer): TBasisOrder;
begin
  Result := TBasisOrder(GetSetting(Index));
end;

procedure TMixture.SetPrintLevel(const Index: Integer; const AValue: TPrintLevel);
begin
  SetSetting(Index, Ord(AValue));
end;

function TMixture.GetPrintLevel(const Index: Integer): TPrintLevel;
begin
  Result := TPrintLevel(GetSetting(Index));
end;

function TMixture.AddToSpectrum(Spectrum: TSpectrum; var Problem: Exception; Simulation: TSimulation): Boolean;
var
  i: Integer;
begin
  Result := False;
  Problem := FixupProperties;
  if Assigned(Problem) then
    Exit;
  CheckMatrixElementsAvailable;
  Spectrum.SetShower(Log, PrintLevel);
  for i := 0 to nSpecies-1 do
    if Species[i].Concentration <> 0 then
      Result := Species[i].AddToSpectrum(Spectrum, Problem, Simulation)
                or Result;
  if LineListWanted(PrintLevel) then
    Spectrum.ShowFooter;
  if not Result then begin
    if not Assigned(Problem) and (nSpecies > 0) then
      Problem := EReportException.Create('No transition moments set');
  end;
end;

function TMixture.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  i: Integer;
begin
  if nSpecies = 0 then
    Result := NullQnoInfo
  else
    Result := Species[0].QnoInfo(QnoType);
  for i := 1 to nSpecies-1 do
    MergeQno(Result, Species[i].QnoInfo(QnoType), QnoType);
end;

function TMixture.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMNode;
  s: string;
  Simulation: TSimulation;
  Value, I, J: Integer;
  ReadSimulation: Boolean;
  CorrelationNode: TDOMDocument; { Correlation node must be read last, and there should only be one }
  OldSVDThresh: Boolean;
begin
  { Version info first written in 5.2.275 }
  FReadVersion :=  DecodePgopherVersion(ReadValue(Node, 'Version', ''));
  { PrintLevel was an integer before 5.2.823, and was written as such until 6.0.207; read it here,
    with the side effect of removing it from the Node }
  s := ReadValue(Node, 'PrintLevel', '');
  ClearFormSettings;
  CorrelationNode := nil;
  { On writing, the parameters and settings of the only simulation object
    are written to the Mixture object, so reverse this logic here }
  Simulation := Simulations[0];
  ReadSimulation := False;
  Result := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if not ReadSimulation and (CompareText(Child.NodeName, 'Simulation') = 0) then begin
      Result := Result + Simulation.ReadProperties(Child, CheckUnread);
      Node.RemoveChild(Child);
      ReadSimulation := True; { Flag first simulation read }
    end else if (CompareText(Child.NodeName, 'Polarization') = 0) then begin
      Simulation.ReadMNode(Child,Value);
      if CheckUnread then
        Result := Result + nUnreadProperties(Child);
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'Lineshape') = 0) then begin
      Simulation.ReadMNode(Child,Value);
      if CheckUnread then
        Result := Result + nUnreadProperties(Child);
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'FormSettings') = 0) then begin
      { Read form settings into LoadedFormSettings for loading to forms later if required }
      LoadedFormSettings.AppendChild(Child.CloneNode(True, LoadedFormSettings));
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'Correlations') = 0) then begin
      if Assigned(CorrelationNode) then
        raise EReportException.Create('Duplicate Correlations Object');
      CorrelationNode := TDOMDocument.Create;
      CorrelationNode.AppendChild(Child.CloneNode(True,CorrelationNode));
      Node.RemoveChild(Child);
    end else if (CompareText(Child.NodeName, 'Variables') = 0) then begin
      { Variables objects should be read early so global variables are defined when reading other nodes }
      ReadMNode(Child,Value);
      if CheckUnread then
        Result := Result + nUnreadProperties(Child);
      Node.RemoveChild(Child);
    end;
    Child := NextChild;
  end;
  { The interpretation of SVDThresh setting changed }
  OldSVDThresh := (ReadVersion < 8003390) and (FindFloatableValue(Node, 'SVDThresh') <> nil);
  if not ReadSimulation then
    { Compatibility: Some properties have moved into Simulations[0] }
    Simulation.TransferProperties(Node, False, True, False);
  Result := Result + inherited ReadProperties(Node, False);
  if OldSVDThresh and (SVDThresh > 0) then
        ScaledParameters := False; { This forces the old behaviour if SVDThresh was set in control file }
  { Now use earlier PrintLevel string }
  if (s <> '') then
    if TryStrToInt(s, Value) then begin
      Inc(Value,3); { Read as integer offset from plMinimal, offset by -3 }
      if Value <= 0 then
        PrintLevel := TPrintLevel(Min(Max(0,Ord(plMinimal) + Value),Ord(High(TPrintLevel))))
      else if Value = 1 then
        PrintLevel := plMinimal { Old plBasic, which no longer exists }
      else { skip over plCSV, plText, plBinary }
        PrintLevel := TPrintLevel(Min(Ord(plDetail) + Value-2,Ord(High(TPrintLevel))))
    end else
      SetSettingFromStr(IndexOfSetting('PrintLevel'),s);
  s := ReadValue(Node, 'DoubleQnos', '');
  if s <> '' then
    if StrToBool(s) then
      QuantumNumberFormat := qft2J
    else
      QuantumNumberFormat := qftJ;
  { ContourFile has moved into TOverlay }
  s := ReadValue(Node,'ContourFile','0');
  if s <> '0' then
    ShowInfo('ContourFile setting in Mixture must now be set in Overlay properties instead');
  { JAdjustSym has moved into TMolecule }
  s := ReadValue(Node,'JAdjustSym','True');
  if (s <> '') and not StrToBool(s) then
    for i := 0 to Count - 1 do
      if Items[i] is TSpecies then
        with TSpecies(Items[i]) do
          for j := 0 to Count-1 do
            if Items[j] is TMolecule then
              TMolecule(Items[j]).JAdjustSym := False;
  { Read only TCorrelations object }
  if Assigned(CorrelationNode) then begin
    Result := Result + Correlations.ReadProperties(CorrelationNode.FirstChild, CheckUnread);
    CorrelationNode.Free;
  end;
  if Assigned(FCorrelations) then
      FCorrelations.ComponentIndex := ComponentCount-1; { Correlations normally at end }
  if CheckUnread then
    Result := Result + nUnreadProperties(Node)
end;

function TMixture.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Child: TDOMNode;
  i: Integer;
begin
  { For backwards compatibility, write Simulation parameters as
    parameters to the mixture object if there is only one simulation
    object }
  for i := 0 to nSimulations - 1 do
    Simulations[i].FWritten := False;
  if nSimulations = 1 then
    Simulations[0].FWritten := True;
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  WriteValue(Result, 'Version', VersionInfo, '');
  Simulations[0].TransferProperties(Result, True, AllParameters, WithChildren);
  if Assigned(LoadedFormSettings) and WithChildren then begin
    Child := LoadedFormSettings.FirstChild;
    while Assigned(Child) do begin
      Result.AppendChild(Child.CloneNode(True,Parent.OwnerDocument));
      Child := Child.NextSibling;
    end;
  end;
  { PrintLevel was an integer before 5.2.823, and was written as such until 6.0.207 }
  { No longer used as plCSV etc now introduced at 6.0.207, so better compatibility
    with 6.0 using normal text, with pl prefix }
  WriteValue(Result,'PrintLevel','pl'+SettingToStr(IndexOfSetting('PrintLevel')),'plMinimal');
  //{ Compatibility - do not write correlation matrix if ShowEstUnc not set }
  //if not ShowEstUnc then begin
  //  Child := Result.FindNode('Correlations');
  //  if Assigned(Child) then
  //    Result.RemoveChild(Child);
  //end;
end;

procedure TMixture.ClearCache(ClearChildren: Boolean);
begin
  if ClearChildren then
    FLargestIndex := 1;
  PerformMNodeNotification(Self, mnCacheCleared, 0);
  LineList := nil;
  FMatrixElementsAvailable := False;
  inherited ClearCache(ClearChildren);
end;

procedure TMixture.ClearFormSettings;
begin
  FreeAndNil(FLoadedFormSettings);
  FLoadedFormSettings := TDOMDocument.Create;
end;

procedure TMixture.PrepareForFit(FitType: TFitType; OtherNode: TRootMNode);
begin
  FillParameters;
  if Assigned(OtherNode) then
    OtherNode.AddNodeParameters(FittingParameters);
  PushParameters(FitType, False);
  ClearStdDev;
  if Assigned(OtherNode) then
    OtherNode.ClearStdDev;
  if Assigned(FCorrelations) then
    Correlations.Clear;
end;

procedure TMixture.ClearUndo;
begin
  FUndoPosition := 0;
  SetLength(FSavedParameters, 0);
  ClearConstraintsAndExpressions;
  if Assigned(FCorrelations) then
    Correlations.Clear;
end;

procedure TMixture.DiscardFit;
begin
  Dec(FUndoPosition);
  if FUndoPosition < 0 then
    FUndoPosition := 0;
  SetLength(FSavedParameters, FUndoPosition);
end;

procedure TMixture.RedoFit;
begin
  if CanRedoFit and not BeginCalculation(True) then begin
    Inc(FUndoPosition);
    LoadCurrentSavedParameters;
  end;
end;

procedure TMixture.UndoFit;
begin
  if CanUndoFit and not BeginCalculation(True) then begin
    Dec(FUndoPosition);
    LoadCurrentSavedParameters;
  end;
end;

function TMixture.CanRedoFit: Boolean;
begin
  Result := FUndoPosition < High(FSavedParameters);
end;

function TMixture.CanUndoFit: Boolean;
begin
  Result := FUndoPosition > 0;
end;

procedure TMixture.CheckMatrixElementsAvailable;
  procedure HandleMNode(MNode:TComponent);
  var
    i: Integer;
  begin
    if MNode is TState then
      TState(MNode).CheckMatrixElementsAvailable(True)
    else if MNode is TPerturbation then
      TPerturbation(MNode).CheckMatrixElementsAvailable
    else if MNode is TTransitionMoment then
      TTransitionMoment(MNode).CheckMatrixElementsAvailable;
    for i := 0 to MNode.ComponentCount-1 do
      HandleMNode(MNode.Components[i]);
  end;
begin
  if FMatrixElementsAvailable then
    Exit;
  if ExprChecking then
    HandleMNode(Self);
  FMatrixElementsAvailable := True;
end;

procedure TMixture.FillParameters;
begin
  SetLength(FittingParameters, 0);
  AddNodeParameters(FittingParameters);
end;

procedure TMixture.PushParameters(FitType: TFitType; SetCurrent: Boolean);
begin
  if (Length(FSavedParameters) > 0) and (FSavedParameters[0].FitType <> FitType) then
    ClearUndo;
  SetLength(FSavedParameters, FUndoPosition+1);
  FSavedParameters[FUndoPosition].FitType := FitType;
  FSavedParameters[FUndoPosition].Parameters := Copy(FittingParameters);
  FSavedParameters[FUndoPosition].Correlation := Copy(Correlations.FitDetails.Correlation);
  if not SetCurrent then
    Inc(FUndoPosition);
end;

procedure TMixture.LoadCurrentSavedParameters;
var
  i: Integer;
  {$IFNDEF CONSOLE}
  DummyParameters: TParameters;
  {$ENDIF}
begin
  { Enter with BeginCalculation set }
  try
    Correlations.FitDetails.Correlation := Copy(FSavedParameters[FUndoPosition].Correlation);
    with FSavedParameters[FUndoPosition] do
      for i := 0 to High(Parameters) do
        with Parameters[i] do begin
          MNode.SetParameter(Index, Value);
          MNode.StdDev[Index] := StdDev;
        end;
    ApplyConstraints;
    {$IFNDEF CONSOLE}
    ResetProgress;
    with MainForm do
      case FSavedParameters[FUndoPosition].FitType of
        ftLine:       if not Split.Visible and Mixture.Simulations[0].HasPlotted then begin
                        EndCalculation; { Turned on in ConsiderReplot }
                        ConsiderRePlot;
                      end;
        ftContour:    FitContour(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowContour, DummyParameters, True, 0);
        ftOverlaySum: FitOverlaySum(Mixture, Overlays, {$IFDEF FPC}@{$ENDIF}ShowOverlaySum, DummyParameters, True, 0);
      end;
    {$ENDIF}
  finally
    EndCalculation;
  end;
end;

procedure TMixture.SetEstUnc(const Index: Integer; const Value: Boolean);
begin
  SetFlag(Index, Value);
  if Value then
    Correlations; { Ensure correlation object exists }
end;

procedure TMixture.SetEUnits(const Index: Integer; const AValue: TEnergyUnits);
begin
  SetUnits(Index, AValue);
  FDefaultParameters[IndexOfParameter('SmallE')] := Convert(DefaultSmallE,cm1,AValue);
end;

procedure TMixture.SetLineList(const Value: TLineList);
begin
  FLineList.Free;
  FLineList := Value;
end;

function TMixture.GetEUnits(const Index: Integer): TEnergyUnits;
begin
  Result := GetUnits(Index);
end;

procedure TMixture.CycleUnits;
var
  NewUnits: TEnergyUnits;
begin
  if Units = High(TEnergyUnits) then
    NewUnits := Low(TEnergyUnits)
  else
    NewUnits := Succ(Units);
  BeginGlobalUpdate;
  try
    ScaleParameters(Units,NewUnits);
    Units := NewUnits;
  finally
    EndGlobalUpdate(True);
  end;
end;

{ TState }

function TState.AddBasis(Eigenvalues: TEigenvalues; F, M: TQuantumNumber; Sym: Integer): Integer;
{ Note that Sym <> Eigenvalues.Sym for non zero external fields, where all fields used }
var
  F1, J, EffectiveS, Ieff: TQuantumNumber;
  i, ii, nuc, nNucleiEffective, nStates: Integer;
  LastStatWtGroup, StatWtGroup, StatWt: SmallInt;
begin
  Result := -1;
  StatWtGroup := -1;
  if RveSelect >=0 then
    if Molecule.JAdjustedSym(F,Sym) <> RveSelect then
      Exit;

  nNucleiEffective := Min(Max(Molecule.nNuclei-Eigenvalues.RemovedSpins,0),High(Molecule.HyperfineBasisStates));
  if Eigenvalues.RemovedSpins > Molecule.nNuclei then
    EffectiveS := 0
  else
    EffectiveS := S;
  for i := 0 to High(Molecule.HyperfineBasisStates[nNucleiEffective]) do begin
    F1 := F;
    nuc := nNucleiEffective-1;
    while nuc >= 0 do begin
      J := F1 - Molecule.HyperfineBasisStates[nNucleiEffective,i,nuc];
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        { This is the J I12 F case; J is correct, but need to use total
          nuclear spin I12 rather then an individual spin in checking validity.
          In this case HyperfineBasisStates contains the true quantum number
          rather than an offset }
        Eigenvalues.MultipleSpinSpecies := Eigenvalues.MultipleSpinSpecies or Nucleus(nuc).HasIdentical;
        Dec(nuc);
        Ieff := Molecule.HyperfineBasisStates[nNucleiEffective,i,nuc];
      end else
        Ieff := Molecule.Nuclei[nuc].Spin;
      if (J < 0) or not IsTriangle(J,Ieff,F1) then begin
        F1 := -1;
        Break;
      end;
      Dec(nuc);
      F1 := J;
    end;
    if F1 >= 0 then begin
      nStates := Eigenvalues.ComponentCount;
      StatWt := AddOneJ(Eigenvalues,F1,F,M,Sym,Molecule.HyperfineBasisStates[nNucleiEffective,i],EffectiveS);
      if Eigenvalues.ComponentCount > nStates then begin
        LastStatWtGroup := (Eigenvalues.Components[nStates] as TBasisState).StatWtGroup;
        for ii := nStates to Eigenvalues.ComponentCount-1 do begin
          (Eigenvalues.Components[ii] as TBasisState).F := F;
          (Eigenvalues.Components[ii] as TBasisState).M := M;
        end;
        if StatWtGroup < 0 then begin
          StatWtGroup := LastStatWtGroup;
          Result := StatWt;
        end else if not Eigenvalues.MultipleSpinSpecies and (StatWtGroup <> LastStatWtGroup) then
          raise Exception.CreateFmt('AddBasis: Inconsistent Statistical Weight groups (%d <> %d) in ', [StatWtGroup, LastStatWtGroup, Name]);
      end;
    end;
  end;
end;

class function TState.NucleusType: TNucleusClass;
begin
  Result := TNucleusWithSpin;
end;

procedure TState.AdjustNuclei(FixNumber: Boolean);
{ Fixup Nucleus links within the state;
  Optionally Add or remove Nuclei to agree with molecule count }
begin
  if FixNumber then
    AdjustNodeIndexes(NucleusType, FNuclei, nil, opInsert, Molecule.nNuclei)
  else
    AdjustNodeIndexes(NucleusType, FNuclei)
end;

function TState.AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber): Exception;
begin
  Result := EReportException.Create('auto not implemented for this molecule type');
end;

constructor TState.Create(AOwner: TComponent);
var
  i: Integer;
begin
  if Assigned(AOwner) and Assigned(AOwner.Owner) then begin
    FMolecule := AOwner.Owner as TMolecule;
    FMixture := Molecule.RootNode as TMixture;
  end;
  SavedProperties := SavedProperties + ',Colour=None,RveSelect=all';
  inherited;
  FFullRotationalSymmetry := True;
  { To avoid problems when adding states, make default spin
    the same as the first state in the same manifold }
  if Assigned(Owner) then
    for i := 0 to Owner.ComponentCount- 1 do
      if (Owner.Components[i] <> Self)
         and (Owner.Components[i] is TState) then begin
        S := TState(Owner.Components[i]).S;
        Break;
      end;
end;

destructor TState.Destroy;
begin
  ClearCache(False);
  inherited Destroy;
end;

function TState.FindColour: TColor;
begin
  Result := Colour;
  if Result = clNone then
    Result := Manifold.FindColour;
end;

function TState.FixupProperties: Exception;
begin
  AllowComplex := Molecule.Mixture.AllowComplex;
  ClearCache(True); { Clears symbolic matrix elements }
  AdjustNuclei(True);
  if Assigned(HyperfineVariableList) then
    HyperfineVariableList.CheckThreadVars;
  Result := inherited FixupProperties;
end;

function TState.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if AIndex = 2 then
    Result := SymSettingsNames(Self, LowestValue)
  else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TState.HasWidth: Boolean;
begin
  Result := Width <> 0;
end;

function IsNegative(const za, zb: Complex):Boolean;
begin
  Result := (za.re = -zb.re) and (za.im = -zb.im);
end;

function IsSameNegative(const za, zb: Complex):Boolean;
var
  z: Complex;
begin
  CMul(z,zb,-1);
  Result := MathSup.SameValue(za,z);
end;

function TState.SymmetrizedRotationalMatrixElement(A, B: TBasisState; MNode: TMnode; Flag: Integer;
  MatrixElementFunction: TMatrixElementFunction; SymCheck: TSymCheck): Complex;
  function MF(PhaseA, PhaseB, Mult: Integer):Complex;
  begin
    Result := MatrixElementFunction(A, B, PhaseA, PhaseB, MNode, Flag);
    if Mult > 0 then
      Exit;
    Result.re := - Result.re;
    Result.im := - Result.im;
  end;
var
  r: Complex;
  procedure FullCheck(Id: Integer);
  begin
    { Slower test here, but allow for slight rounding errors }
    if not MathSup.SameValue(r, Result)
       and not ((SymCheck = MayCancel) and IsSameNegative(r,Result)) then begin
       Write(Log, 'Symmetry mismatch ', Id, ':', r.re, ',', r.im, ' ', Result.re, ',', Result.im, ' ', A.BasisLabel, ' ', B.BasisLabel);
       if Assigned(MNode) then
         Write(Log, ' ', MNode.Name);
       WriteLn(Log);
       WriteLn(Log,MF(1,1,1).re);
       WriteLn(Log,MF(-1,-1,1).re);
       WriteLn(Log,MF(1,-1,1).re);
       WriteLn(Log,MF(-1,1,1).re);
       raise EReportException.Create('Symmetry mismatch ' + IntToStr(Id));
    end;
  end;
begin
  Result := MF(1, 1, 1);
  if A.Phase = 0 then begin
    if B.Phase <> 0 then begin
      CAdd(Result, MF(1, -1, B.Phase));
      CMul(Result, Result,1/Sqrt(Two));
    end;
  end else if B.Phase = 0 then begin
    CAdd(Result, MF(-1, 1, A.Phase));
    CMul(Result, Result, 1/Sqrt(Two));
  end else begin
    if not FullRotationalSymmetry and (SymCheck = MustEqual) then
      SymCheck := MayCancel;
    {$IFOPT R-}
    if SymCheck = MustEqual then
      SymCheck := NoCheck;
    {$ENDIF}
    // SymCheck := FullSum;
    // SymCheck := NoCheck;
    if SymCheck <> NoCheck then begin
      r := MF(-1, -1, A.Phase*B.Phase);
      if (SymCheck <> FullSum) and not IsEqual(r,Result) and not ((SymCheck = MayCancel) and IsNegative(r, Result)) then begin
        // MatrixElementFunction(A, B, 1, 1, MNode, Flag);
        // MatrixElementFunction(A, B, -1, -1, MNode, Flag);
        FullCheck(1);
      end;
    end;
    CAdd(Result, MF(1, -1, B.Phase));
    if SymCheck <> NoCheck then begin
      CAdd(r, MF(-1, 1, A.Phase));
      if (SymCheck <> FullSum) and not IsEqual(r,Result) and not ((SymCheck = MayCancel) and IsNegative(r,Result)) then
        FullCheck(2);
    end;
    if (SymCheck = MayCancel) { or (SymCheck = FullSum) } then begin
      CAdd(Result, r);
      CMul(Result, Result, 1/Two);
    end;
  end;
  CMuliPower(Result,B.iPower - A.iPower);
  if not AllowComplex then
    if Abs(Result.im) > Mixture.SmallE then
      if Result.re = 0 then begin
        { Purely complex matrix element; in most cases changing the phase of one
          of the wavefunctions by -i has no net effect }
        Result.re := Result.im;
        Result.im := 0;
      end else
        raise EReportException.Create('Complex matrix element')
    else
      Result.im := 0;
end;

procedure TState.SymmetryTable(var F: Text);
begin
  raise NotYet;
end;

function TState.SymmetrizedHyperfineMatrixElement(A, B: TBasisState;
  Rank: TQuantumNumber;
  MatrixElementFunction: TMatrixElementFunction;
  MNode: TMNode): Complex;
begin
  Result := z0;
  if Abs(A.J-B.J) > Rank then
    Exit;
  Result := SymmetrizedRotationalMatrixElement(A, B, MNode, Rank, MatrixElementFunction);
end;

function TState.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
begin
  Result := nil; { By default, can't convert }
end;

procedure TState.ConvertToPerturbations(Nuclei: Boolean);
var
  i: Integer;
begin
  { Ignore failures }
  for i := 2 { Do not convert origin } to nParameterVariables do
    if ParameterVariable(i).Value <> 0 then
      ConvertParameter(i, True);
  if Nuclei then
    for i := 0 to Molecule.nNuclei-1 do
      Nucleus(i).ConvertToPerturbations;
end;

function TState.ConvertToSymmetrized(const MatrixElements: TMatrixElements): TMatrixElements;
begin
  Result := nil; { empty array implies symmetrization not possible }
end;

function TState.Manifold: TManifold;
begin
  Result := Owner as TManifold;
end;

function TState.MatrixElement(A, B: TBasisState; MatElemType: TMatElemType): Complex;
{ Matrix element of normal Hamiltonian }
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  Ieffp, Ieff: TQuantumNumber; { Nuclear spin, possibly a sum where equivalent nuclei used }
  I1, I2: TQuantumNumber;
  nuc: Integer;
  Rank1Sum, Rank2Sum, Rank1Extra, Rank2Extra: Complex;
  SpinsMatch: Boolean; { Keep track of nuclear spin states matching }
  Nucleus1, Nucleus2: TNucleus;

  function CallHyperfineMatrixElement(Rank: TQuantumNumber; Nucleus: TNucleus): Complex;
  var
    sJ: Double;
  begin
    sJ := sixj(FF,Ieffp,FF1p,Rank,FF1,Ieff);
    if sJ <> 0 then
      CMul(Result,SymmetrizedHyperfineMatrixElement(A, B, Rank,
                                                    {$IFDEF FPC}@{$ENDIF}CheckedHyperfineMatrixElement,
                                                    Nucleus),
                  Minus(FF1+Ieffp+FF)*sJ)
    else
      Result := z0;
  end;

  function CallTwoHyperfineMatrixElements(Rank: TQuantumNumber): Complex;
  var
    z: Complex;
  begin
    CMul(Result, CallHyperfineMatrixElement(Rank, Nucleus1),
                 Minus(I1+I2+Ieff+Rank) * Sqrt((Ieffp+1)*(Ieff+1))
                 *sixj(I1, Ieffp, I2, Ieff, I1, Rank));
    CMul(z, CallHyperfineMatrixElement(Rank, Nucleus2),
            Minus(I1+I2+Ieffp+Rank) * Sqrt((Ieffp+1)*(Ieff+1))
            *sixj(I2, Ieffp, I1, Ieff, I2, Rank));
    CAdd(Result, z);
  end;
begin
  Result := z0;
  if A.State <> B.State then
    Exit;
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (MatElemType = meHrot) { TODO: Other matrix element types }
     and (Molecule.nNuclei > 0) and (Abs(A.J-B.J) <= 4) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    Rank1Sum := z0;
    Rank2Sum := z0;
    SpinsMatch := True;
    nuc := High(A.HyperfineState);
    while nuc >= 0 do begin
      SpinsMatch := SpinsMatch and (FF = FFp);
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      Rank1Extra := z0;
      Rank2Extra := z0;
      Nucleus1 := A.State.Nucleus(nuc);
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        Dec(nuc);
        Ieffp := A.HyperfineState[nuc];
        Ieff := B.HyperfineState[nuc];
        if Nucleus1.HasIdentical then
          Nucleus2 := Nucleus1
        else begin
          Nucleus2 := Nucleus1;
          Nucleus1 := A.State.Nucleus(nuc);
        end;
        I1 := Nucleus1.spin;
        I2 := Nucleus2.spin;
        if SpinsMatch then begin
          { This nucleus contributes }
          Rank1Extra := CallTwoHyperfineMatrixElements(2);
          if (I1 > 1) or (I2 > 1) then
            Rank2Extra := CallTwoHyperfineMatrixElements(4);
        end;
        { Check I12 for match }
        if Ieffp <> Ieff then begin
          { Change in I12 - other spin operators can't change this }
          SpinsMatch := False;
          Rank1Sum := z0;
          Rank2Sum := z0;
        end;
      end else begin
        Ieff := Molecule.Nuclei[nuc].spin;
        Ieffp := Ieff;
        if SpinsMatch then begin
          { This nucleus contributes }
          Rank1Extra := CallHyperfineMatrixElement(2, Nucleus1);
          if Ieff > 1 then
            Rank2Extra := CallHyperfineMatrixElement(4, Nucleus1);
        end;
      end;
      if not MathSup.IsZero(Rank1Sum) then
        CMul(Rank1Sum,Rank1Sum, Minus(FF1p+Ieff+FF+2)*Sqrt((FFp+1)*(FF+1))
                                * sixj(FF1p, FFp, Ieff, FF, FF1, 2));
      if not MathSup.IsZero(Rank2Sum) then
        CMul(Rank2Sum,Rank2Sum, Minus(FF1p+Ieff+FF+4)*Sqrt((FFp+1)*(FF+1))
                                * sixj(FF1p, FFp, Ieff, FF, FF1, 4));
      CAdd(Rank1Sum, Rank1Extra);
      CAdd(Rank2Sum, Rank2Extra);
      FFp := FF1p;
      FF := FF1;
      Dec(nuc);
    end;
    SpinsMatch := SpinsMatch and (FF = FFp);
    CAdd(Result, Rank1Sum, Rank2Sum);
  end else
    SpinsMatch := A.HyperfineState = B.HyperfineState;
  if SpinsMatch then begin
    CAdd(Result, SymmetrizedRotationalMatrixElement(A, B, nil, Ord(MatElemType), {$IFDEF FPC}@{$ENDIF}CheckedRotationalMatrixElement));
    if ExprChecking then
      CheckedMatrixElement(A, B, 0, 0, SymmetrizedMatrixElements, Result, nil, False, True);
  end;
end;

function TState.PerturbationMatrixElement(Perturbation: TPerturbation; A, B: TBasisState; MatElemType: TMatElemType): Complex;
var
  FFp, FF, FF1p, FF1: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
  MatrixElementFunction: TMatrixElementFunction;
  sJ: Double;
begin
  Result := z0;
  if MatElemType <> meHrot then Exit; { TODO: Perturbation meWidth }
  if Perturbation.SymSelect >= 0 then
    if Molecule.JAdjustedSym(A.J,A.Sym) <> Perturbation.SymSelect then
      Exit;

  MatrixElementFunction := {$IFDEF FPC}@{$ENDIF}CheckedRotationalPerturbationMatrixElement;
  { Handle multiple nuclear spin case here, by decomposing into simple calls
    for a single nucleus }
  if (Molecule.nNuclei > 0) and (Perturbation.Nucleus > 0) and (Abs(A.J-B.J) <= Perturbation.NuclearRank) then begin
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    for nuc := High(A.HyperfineState) downto 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if nuc >= Perturbation.Nucleus-1 then begin
        if FF <> FFp then
          Exit
        else if nuc = Perturbation.Nucleus-1 then begin
          sJ := sixj(FF,Molecule.Nuclei[nuc].spin,FF1p,Perturbation.NuclearRank,FF1,Molecule.Nuclei[nuc].spin);
          if sJ = 0 then
            Exit;
          CMul(Result, SymmetrizedHyperfineMatrixElement(A, B, Perturbation.NuclearRank, MatrixElementFunction, Perturbation),
                       Minus(FF1+Molecule.Nuclei[nuc].spin+FF)*sJ)
        end
      end else if nuc < Perturbation.Nucleus-1 then
        CMul(Result, Result, Minus(FF1p+Molecule.Nuclei[nuc].spin+FF+Perturbation.NuclearRank)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Molecule.Nuclei[nuc].spin, FF, FF1, Perturbation.NuclearRank));
      FFp := FF1p;
      FF := FF1;
    end;
  end else if (A.HyperfineState = B.HyperfineState) and (A.J = B.J) then
    Result := SymmetrizedRotationalMatrixElement(A, B, Perturbation, Ord(MatElemType), MatrixElementFunction);
  if ExprChecking then
    CheckedMatrixElement(A, B, 0, 0, Perturbation.SymmetrizedMatrixElements, Result, Perturbation, False, True);
end;

procedure ShowMatrixElements(var f: Text; MNode: TMNode; State: TState; First, Last: Integer; Zero: Boolean; const Indent: string);
var
  Elements, SymmetrizedElements: TMatrixElements;
  i, j: Integer;
begin
  for i := First to Last do
    if Zero or (MNode.Parameters[i] <> 0) then
      try
        Elements := (MNode as TMNodeWithVariables).SymbolicRotationalMatrixElement(i,-1);
        for j := 0 to High(Elements) do
          WriteLn(F, Indent, State.MatrixElementType(Elements[j], MNode.ParameterNames[i]), ' = ', Elements[j].Expr.Text(True));
        SymmetrizedElements := State.ConvertToSymmetrized(Elements);
        if Length(SymmetrizedElements) > 0 then begin
          WriteLn(F, Indent, '  Symmetrized: ');
          for j := 0 to High(SymmetrizedElements) do
            WriteLn(F, Indent, '  ', State.MatrixElementType(SymmetrizedElements[j], MNode.ParameterNames[i], True), ' = ', SymmetrizedElements[j].Expr.Text(True));
        end;
        DiscardMatrixElements(SymmetrizedElements);
        DiscardMatrixElements(Elements);
        WriteLn(Log);
      except
        WriteLn(F, Indent, 'Matrix elements for ', MNode.ParameterNames[i], ' not available');
      end;
end;

procedure TState.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  PrettyPrintHeader(F, PrintLevel, Indent);
  Indent := Indent + '  ';
  PrettyPrintSettings(F, PrintLevel, Indent);
  PrettyPrintParameters(F, PrintLevel, Indent);
  if PrintLevel >= plDetail then
    ShowMatrixElements(Log, Self, Self, 2, nParameters, False, Indent);
  PrettyPrintComponents(F, PrintLevel, Indent);
end;

function TState.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
begin
  raise EReportException.Create('Symbolic matrix elements not implemented for '+UniqueName+'.'+ParameterNames[AIndex]);
end;

function TState.SymbolicRotationalTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements;
begin
  raise NotYet;
end;

threadvar
  PSignature: TMatrixElement; { Minimize overhead }

procedure CleanSig;
begin
  PSignature.Signature := nil;
end;

var
  EvaluatorLock: TRTLCriticalSection;

procedure TState.CheckedMatrixElement(
  ABra, AKet: TBasisState;
  SignA, SignB: Integer;
  const AMatrixElements: TMatrixElements;
  Numeric: Complex; MNode: TMNode; IncludeJ: Boolean; Symmetrized: Boolean);
var
  i, Match: Integer;
  V: Double;
  Expr: TExprNode;
  NeedComplex: Boolean;
  AName, FailMsg: string;
  Evaluator: TEvaluator;
begin
  if MNode is TMNodePair then
    AName := TMNodePair(MNode).PairName
  else
    AName := 'Hrot';
  NeedComplex := False;
  if Abs(Numeric.im) > Mixture.SmallE then
    if Molecule.Mixture.AllowComplex then begin
      NeedComplex := True;
      Numeric.re := Numeric.im;
    end else if Abs(Numeric.re) > Mixture.SmallE then
      raise EReportException.Create('Comparing complex matrix element') // Should trigger check elsewhere normally anyway
    else
      Numeric.re := Numeric.im; // Ignore complex character for check below
  if Symmetrized then begin
    SignA := 1;
    SignB := 1;
    PSignature.BraPhase := ABra.Phase;
    PSignature.KetPhase := AKet.Phase;
    if Length(AMatrixElements) = 0 then
      Exit;
  end;
  PSignature.Signature := ABra.MakeSignature(SignA, AKet, SignB, MNode, Symmetrized);
  V := 0;
  Expr := nil;
  FailMsg := '';
  for i := 0 to High(AMatrixElements) do begin
    if SignaturesMatch(AMatrixElements[i].Signature, PSignature.Signature) then
      Match := i
    else
      Match := -1;
    if Symmetrized and (Match >= 0) and ((ABra.Phase <> AMatrixElements[i].BraPhase) or (AKet.Phase <> AMatrixElements[i].KetPhase)) then
      Match := -1;
    if Match >= 0 then begin
      Expr := AMatrixElements[i].Expr;
      ABra.SetSymbolicVariables(SignA, AKet, SignB, MNode);
      if NeedComplex then begin
        { TODO: Add complex to symbolic evaluation }
        if not Assigned(AMatrixElements[i].CEvaluator) then begin
          EnterCriticalsection(EvaluatorLock);
          try
            if not Assigned(AMatrixElements[i].CEvaluator) then begin
              AMatrixElements[i].CEvaluator := AMatrixElements[i].Expr.Compile([evoZeroOverZero, evoPureImag]);
              if AMatrixElements[i].CEvaluator.nSqrts = 0 then
                raise EReportException.Create('No sqrt calls in imaginary function');
            end;
          finally
            LeaveCriticalsection(EvaluatorLock);
          end;
        end;
        Evaluator := AMatrixElements[i].CEvaluator;
      end else
        Evaluator := AMatrixElements[i].Evaluator;
      try
        V := Evaluator.Value;
      except
        on E:Exception do
          FailMsg := E.Message;
      end;
      { Additional (very slow) check:
        if not SameValue(V, Expr.RealValue) then
          WriteLn(Log, Expr.Text, ' ', V, Expr.RealValue);
      }
      Break;
    end;
  end;
  if (FailMsg <> '') or not SameValue(Numeric.re, V) or IsNanOrInf(V) then begin
    Inc(nExprFailures);
    Write(Log, '<', ABra.State.UniqueName, ':', ABra.BasisLabel, ' F=', QuantumNumberToStr(ABra.F,qftJ), '(', SignA, ')|', AName, '|',
                    AKet.State.UniqueName, ':', AKet.BasisLabel, ' F=', QuantumNumberToStr(AKet.F,qftJ), '(', SignB, ')>');
    if NeedComplex then
      Write(Log, ' Imaginary');
    if Symmetrized then
      Write(Log, ' Symmetrized');
    Write(Log, ' SymbolicMatrixElement: Signature:', ABra.State.MatrixElementType(PSignature, AName, Symmetrized));
    WriteLn(Log, ' Numeric=', Numeric.re, ' Symbolic=', V, ', Diff = ', Numeric.re-V, ' ', FailMsg);
    if Assigned(Expr) then begin
      WriteLn(Log, Expr.Text(True), ' ', ABra.SymbolicVariableValues, ' ', Match);
      //WriteLn(Log, Expr.Text(False), ' ', Expr.RealValue);
    end;
    //ExprChecking := False;
    //PerturbationMatrixElement(MNode as TPerturbation, ABra, AKet, meHrot);
    //ExprChecking := True;
    // AMatrixElements[Match].Evaluator.Value; // Testing;
    // Testing: } ABra.MakeSignature(SignA, AKet, SignB);
    //if IsZero(Numeric) then
    //  Numeric := RotationalMatrixElement(ABra,AKet,SignA,SignB,nil,0);
      //Numeric := RotationalTransitionMatrixElement(ABra,AKet,SignA,SignB,MNode,0,True);
    //Numeric := TransitionMatrixElement(MNode as TTransitionMoment,ABra,AKet);
  end;
end;

function TState.CheckedRotationalPerturbationMatrixElement(
  ABra, AKet: TBasisState;
  SignA, SignB: Integer;
  MNode: TMNode; Flag: Integer): Complex;
var
  Perturbation: TPerturbation;
begin
  Result := RotationalPerturbationMatrixElement(ABra,AKet,SignA,SignB,MNode,Flag);
  if not ExprChecking then
    Exit;
  Perturbation := MNode as TPerturbation;
  if Perturbation.CheckMatrixElementsAvailable then
    CheckedMatrixElement(ABra, AKet, SignA, SignB, Perturbation.FNoHyperfineMatrixElements, Result, Perturbation, False, False);
end;

function TState.CheckedRotationalTransitionMatrixElement(ABra, AKet: TBasisState;
  SignA, SignB: Integer; TransitionMoment: TMNode;
  F: TQuantumNumber): Complex;
var
  P: TPerturbation;
  i: Integer;
  State: TBasisState;
  Z: Complex;
  TM: TTransitionMoment;
begin
  TM := TransitionMoment as TTransitionMoment; { Must be Mnode in parameter list }
  Result := RotationalTransitionMatrixElement(ABra,AKet,SignA,SignB,TM,F,True);
  if TM.nPerturbations > 0 then begin
    State := nil;
    while AKet.NextBasis(State, TM, SignB) do begin
      for i := 0 to TM.ComponentCount-1 do
        if TM.Components[i] is TPerturbation then begin
          P := TPerturbation(TM.Components[i]);
          Z := RotationalPerturbationMatrixElement(State, AKet, 1, SignB, P, Ord(meHrot));
          CMul(Z, RotationalTransitionMatrixElement(ABra,State,SignA,1,TM,F,False));
          CAdd(Result, Z);
        end;
    end;
    State := nil;
    while ABra.NextBasis(State, TM, SignA) do begin
      for i := 0 to TM.ComponentCount-1 do
        if TransitionMoment.Components[i] is TPerturbation then begin
          Z := RotationalPerturbationMatrixElement(ABra, State, SignA, 1, P, Ord(meHrot));
          CMul(Z, RotationalTransitionMatrixElement(State,AKet,1,SignB,TM,F,False));
          if P.IsAntiSym then
            CMul(Z, -1);
          CAdd(Result, Z);
        end;
    end;
  end;
  if Assigned(TM.ActiveTransitionFunction) and TM.ActiveTransitionFunction.Signed then
    TM.ActiveTransitionFunction.MatrixElement(Result, ABra, AKet, SignA, SignB, False);
  if not ExprChecking then
    Exit;
  if TM.CheckMatrixElementsAvailable then
    CheckedMatrixElement(ABra, AKet, SignA, SignB, TM.FNoHyperfineMatrixElements, Result, TransitionMoment, True, False);
  { TODO: Checking with intensity perturbations }
end;

function TState.CheckedHyperfineMatrixElement(ABra, AKet: TBasisState; Signp,
  Sign: Integer; MNode: TMnode; RankWanted: TQuantumNumber): Complex;
var
  Nuc: TNucleus;
begin
  Result := HyperfineMatrixElement(ABra, AKet, Signp, Sign, MNode, RankWanted);
  if not ExprChecking then
    Exit;
  Nuc := MNode as TNucleus;
  Nuc.CheckMatrixElementsAvailable;
  CheckedMatrixElement(ABra, AKet, Signp, Sign, Nuc.FMatrixElementsByRank[RankWanted div 2], Result, Self, False, False);
end;

function TState.CheckedRotationalMatrixElement(ABra, AKet: TBasisState;
  SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
begin
  Result := RotationalMatrixElement(ABra,AKet,SignA,SignB,MNode,Flag);
  if not ExprChecking then
    Exit;
  CheckMatrixElementsAvailable(True);
  if Flag = 0 then
    CheckedMatrixElement(ABra, AKet, SignA, SignB, MatrixElements, Result, Self, False, False);
end;

procedure TState.CheckMatrixElementsAvailable(WithNuclei: Boolean);
var
  i: Integer;
begin
  if (Length(MatrixElements) = 0) and not MatrixElementsEmpty then begin
    DoCheckMatrixElementsAvailable(FMatrixElements, FSymmetrizedMatrixElements, -1);
    FMatrixElementsEmpty := Length(FMatrixElements) = 0;
  end;
  if WithNuclei and not FNucleiDone then begin
    for i := 0 to Molecule.nNuclei-1 do begin
      Nucleus(i).CheckMatrixElementsAvailable;
//DumpMatrixElements(FSymmetrizedMatrixElements, 'B ');
//DumpMatrixElements(Nucleus(i).FSymmetrizedMatrixElements, '  ');
      MergeElements(FSymmetrizedMatrixElements, Nucleus(i).FSymmetrizedMatrixElements, 1, True, False { Must not simplify to avoid 0/0 issues });
//DumpMatrixElements(FSymmetrizedMatrixElements, 'A ');
    end;
    FNucleiDone := True;
  end;
end;

procedure TState.ClearCache(ClearChildren: Boolean);
begin
  inherited ClearCache(ClearChildren);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  FMatrixElementsEmpty := False;
  FNucleiDone := False;
end;

function TState.TransitionMatrixElement(TransitionMoment: TTransitionMoment; A, B: TBasisState): Complex;
{ Reduced matrix element of transition dipole }
{ Similar logic to TState.MatrixElement, but assuming transition moment does not
  act on any nuclear spins, except for nuclear magnetic dipole transitions }
var
  FFp, FF, FF1p, FF1, Ieff: TQuantumNumber; { Partial F quantum numbers }
  nuc: Integer;
  IsNuclearDipole: Boolean;
  Rank: TQuantumNumber;
begin
  Result := z0;
  Rank := TransitionMoment.Rank;
  if Abs(A.J-B.J) > Rank then
    Exit;
  Result := z1;
  if Molecule.nNuclei > 0 then begin
    IsNuclearDipole := TransitionMoment is TNuclearMagneticMoment;
    { Work from last nucleus }
    FFp := A.F;
    FF := B.F;
    nuc := High(A.HyperfineState);
    if IsNuclearDipole and ((TNuclearMagneticMoment(TransitionMoment)).Nucleus-1 > nuc) then begin
      Result := z0; { Removed nucleus }
      Exit;
    end;
    while nuc >= 0 do begin
      FF1p := FFp - A.HyperfineState[nuc];
      FF1 := FF - B.HyperfineState[nuc];
      if (nuc > 0) and Molecule.Nuclei[nuc-1].AsNext then begin
        if IsNuclearDipole then
          raise NotYet;
        { Coupled spins; need to use total nuclear spin for pair }
        Dec(nuc);
        Ieff := A.HyperfineState[nuc];
        if Ieff <> B.HyperfineState[nuc] then begin
          Result := z0; { Can't change hyperfine state }
          Exit;
        end;
      end else
        Ieff := Molecule.Nuclei[nuc].Spin;
      if IsNuclearDipole and ((TNuclearMagneticMoment(TransitionMoment)).Nucleus-1 = nuc) then begin
        if FF1p = FF1 then begin
          CMul(Result, Result, -TNuclearMagneticMoment(TransitionMoment).gN
                               * Minus(FF1+Ieff+FF+Rank)*Sqrt((FFp+1)*(FF+1))
                               * sixj(Ieff, FFp, FF1, FF, Ieff, Rank)
                               * ReducedMatrixElement(2,Ieff));
          Rank := 0; { Flag need to use identity operator for everything else }
        end else
          Result := z0;
      end else if IsNuclearDipole and (Rank = 0) then begin
        if FF1p <> FF1 then
          Result := z0;
      end else
        CMul(Result, Result, Minus(FF1p+Ieff+FF+Rank)*Sqrt((FFp+1)*(FF+1))
                             * sixj(FF1p, FFp, Ieff, FF, FF1, Rank));
      if MathSup.IsZero(Result) then
        Exit;
      FFp := FF1p;
      FF := FF1;
      Dec(nuc);
    end;
  end;
  CMul(Result, Result, SymmetrizedRotationalMatrixElement(A, B, TransitionMoment, -1, {$IFDEF FPC}@{$ENDIF}CheckedRotationalTransitionMatrixElement, MayCancel));
  { The following line can be added if a consistent sign is wanted for the reduced matrix elements;
    without it states differing in total angular momentum by an odd number will have a different
    sign to this matrix element depending on the order of calculation.
    Computes phase factor for smallest M }
  // CMul(Result, Result, Minus(Max(A.F,B.F)-SmallestProjection(A.F)));
  if ExprChecking then
      CheckedMatrixElement(A, B, 1, 1, TransitionMoment.SymmetrizedMatrixElements, Result, TransitionMoment, False, True);
end;

function TState.SymbolicTransitionMatrixElements(TransitionMoment: TTransitionMoment; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements;
begin
  Result := SymbolicRotationalTransitionMatrixElements(TransitionMoment, AddHyperfineCoupling, IncompleteOK);
end;

function TState.Nucleus(Index: Integer): TNucleus;
begin
  Result := FNuclei[Index] as TNucleus;
end;

function TState.HalfSpin(RemovedSpins: Byte): TQuantumNumber;
var
  i: Integer;
begin
  if RemovedSpins > Molecule.nNuclei then
    Result := 0
  else begin
    Result := S;
    for i := 0 to Molecule.nNuclei - 1 - RemovedSpins do
      Inc(Result,Molecule.Nuclei[i].Spin);
  end;
  Result := Result and 1;
end;

function TState.ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase): Boolean;
begin
  raise EReportException.Create('ReadBranchFormat not implemented for '+Name);
end;

function TState.ReadHyperfineBranch(Text: string; var Obs: TObservation; var HyperfineU, HyperfineL: THyperfineBasis; Reader: TObservationsReaderBase; HasFields:Boolean): Boolean;
  function ReadOneHyperfine(var J: TQuantumNumber; var Hyperfine: THyperfineBasis):Boolean;
  var
    p, i: Integer;
  begin
    Result := False;
    p := Molecule.nNuclei-Reader.RemoveSpins;
    if HasFields then
      Inc(p);
    SetLength(Hyperfine,p);
    for i := 0 to High(Hyperfine) do begin
      p := Pos(',',Text);
      if p <= 0 then
        p := Length(Text)+1;
      if Reader.ReadQuantumNumber(Hyperfine[i],Copy(Text,1,p-1)) then
        Exit; { Fails }
      if i >= Molecule.nNuclei - Reader.RemoveSpins then
        J := Hyperfine[i] { M }
      else if not Molecule.Nuclei[i].AsNext then begin
        Dec(Hyperfine[i],J);
        J := Hyperfine[i]+J;
      end;
      Delete(Text,1,p);
    end;
    if HasFields then
      SetLength(Hyperfine,Length(Hyperfine)-1);
    { TODO: May need modifying for hyperfine and field }
    Result := True;
  end;
begin
  Result := not (ReadOneHyperfine(Obs.JU,HyperfineU)
                 and ReadOneHyperfine(Obs.JL,HyperfineL)
                 and (Text = ''));
end;

function TState.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  Child, NextChild: TDOMNode;
  Index: Integer;
  NucName: string;
begin
  { Ensure Nuclei read in correct order, and discard any un-matched nuclei }
  Index := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do begin
    NextChild := Child.NextSibling;
    if Pos('Nucleus', Child.NodeName) = Length(Child.NodeName)-6 then begin
      NucName := ReadValue(Child, 'Name', '', '', False);
      if Index < Molecule.nNuclei then begin
        if (Molecule.Nuclei[Index].Name <> '') and (NucName <> Molecule.Nuclei[Index].Name) then
//          for i := 0 to Molecule.nNuclei-1 do
//            if Molecule.Nuclei[i].Name = NucName then
          raise EReportException.Create('Nucleus ' + IntToStr(Index) + ' out of order for ' + Owner.Name + ' - ' + Node.NodeName + '; Expected ' + Molecule.Nuclei[Index].Name + ' found ' + NucName);
        Inc(Index);
      end else
        Node.RemoveChild(Child); { Assume paste or merge from system with fewer nuclei }
    end;
    Child := NextChild;
  end;
  Result := inherited ReadProperties(Node, CheckUnread);
end;

function TState.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  case AIndex of
    2: if RveSelect < 0 then Exit;
    3: if S <= 0 then Exit;
  end;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TState.SpinSpin(Nuc1, Nuc2: Integer): TSpinSpin;
var
  i: Integer;
begin
  for i := 0 to ComponentCount - 1 do
    if Components[i] is TSpinSpin then
      with TSpinSpin(Components[i]) do
        if ((Bra.Name = Molecule.Nuclei[Nuc1].Name)
            and (Ket.Name = Molecule.Nuclei[Nuc2].Name))
           or
           ((Bra.Name = Molecule.Nuclei[Nuc2].Name)
            and (Ket.Name = Molecule.Nuclei[Nuc1].Name)) then begin
              Result := Self.Components[i] as TSpinSpin;
              Exit;
            end;
  Result := SpinSpinType.Create(Self);
  Result.SetBraKetName(Molecule.Nuclei[Nuc1].Name, Molecule.Nuclei[Nuc2].Name);
end;

function TState.EstimatedStateOrigin: Double;
begin
  Result := Origin;
end;

function TState.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TState.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = 3;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TState.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := ColourHelpString;
    2: Result := 'Only include given rovibronic symmetry - for normal use set to ''all''';
    3: Result := 'Electron Spin';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TState.HelpOnParameter(AIndex: Integer): string;
var
  E: Double;
begin
  if AIndex = 1 then begin
    Result := 'State Origin.';
    E := EstimatedStateOrigin;
    if E <> 0 then
      Result := Result + ' Estimated lowest level = '+FloatToStr(E);
  end else if AIndex = 2 then
    Result := 'Linewidth (rotation independent) for state.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TTransitionMoments }

function TTransitionMoments.GetTransitionMoment(AIndex: Integer): TTransitionMoment;
begin
  Result := Items[AIndex] as TTransitionMoment;
end;

constructor TTransitionMoments.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Colour=None';
  inherited Create(AOwner);
end;

constructor TTransitionMoments.Create(AOwner: TComponent; Manifold1Name, Manifold2Name: string);
begin
  Create(AOwner);
  FBraName := Manifold1Name;
  FKetName := Manifold2Name;
end;

destructor TTransitionMoments.Destroy;
begin
  inherited;
end;

function TTransitionMoments.FixupProperties:Exception;
var
  i: Integer;
  Swapped: Boolean;
  BraState, KetState: TMNode;
begin
  DisableRankCheck := False;
  Result := inherited FixupProperties;
  if Assigned(Result) then Exit;
  { Sort by Rank }
  repeat
    Swapped := False;
    for i := 0 to ComponentCount-2 do
      if TransitionMoment[i].Rank > TransitionMoment[i+1].Rank then begin
        TransitionMoment[i].ComponentIndex := i+1;
        Swapped := True;
      end;
  until not Swapped;
  if ComponentCount > 0 then
    FnRanks := 1
  else
    FnRanks := 0;
  for i := 0 to ComponentCount-2 do
    if TransitionMoment[i].Rank <> TransitionMoment[i+1].Rank then
      Inc(FnRanks);
  for i := 0 to ComponentCount-1 do
    TransitionMoment[i].FNodeIndex := i; { Repair NodeIndex }
  { Set up linked list of individual transition moments for each state }
  SetLength(TMlist[0], (Bra as TManifold).nStates);
  for i := 0 to High(TMlist[0]) do
    TMlist[0][i] := nil;
  SetLength(TMlist[1], (Ket as TManifold).nStates);
  for i := 0 to High(TMlist[1]) do
    TMlist[1][i] := nil;
  for i := 0 to ComponentCount-1 do begin
    BraState := TransitionMoment[i].Bra;
    KetState := TransitionMoment[i].Ket;
    if not Assigned(BraState) or not Assigned(KetState) then
      Break; { Still setting up }
    if (Bra <> Ket) and (BraState.Owner <> Bra) then begin
      KetState := BraState;
      BraState := TransitionMoment[i].Ket;
    end;
    TransitionMoment[i].TMlist[0] := TMList[0][BraState.NodeIndex];
    TMList[0][BraState.NodeIndex] := TransitionMoment[i];
    TransitionMoment[i].TMlist[1] := TMList[1][KetState.NodeIndex];
    TMList[1][KetState.NodeIndex] := TransitionMoment[i];
  end;
end;

function TTransitionMoments.MaxRank: Integer;
begin
  if ComponentCount > 0 then
    Result := TransitionMoment[ComponentCount-1].Rank
  else
    Result := -2;
end;

function TTransitionMoments.Allowed(J1: TQuantumNumber; Sym1: Integer; J2: TQuantumNumber; Sym2: Integer): Boolean;
{ Do any of the transition moments give rise to an allowed transition? }
var
  i: Integer;
  TM: TTransitionMoment;
begin
  Result := True;
  for i := 0 to ComponentCount-1 do begin
    TM := TransitionMoment[i];
    if (DisableRankCheck or IsTriangle(J1,J2,TM.Rank)) and TM.Allowed(Sym1, Sym2) then
      Exit;
  end;
  Result := False;
end;

procedure TTransitionMoments.AddToSpectrum(Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; ATransitionMoment: TTransitionMoment; Simulation: TSimulation; Wait: Boolean; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean);
var
  D: TMatrix;
begin
  D := TransitionMatrix(Ev1, Ev2, Wait, not Assigned(ATransitionMoment), ATransitionMoment, Simulation, PrintLevel);
  if Wait and Assigned(D) then
    try
      AddMatrixToSpectrum(D, Spectrum, Ev1, Ev2, PrintLevel, Excitation, Problem, IncludeZero);
    finally
      if Assigned(ATransitionMoment) then
        D.Free;
    end;
end;

procedure TTransitionMoments.AddMatrixToSpectrum(D: TMatrix; Spectrum: TSpectrum; Ev1, Ev2: TEigenvalues; PrintLevel: TPrintLevel; const Excitation: TExcitation; var Problem: Exception; IncludeZero: Boolean = false);
var
  i, f: Integer;
  Line, SLine: TLine;
  Swap: Boolean;
  procedure DoLine(var Line: TLine);
  begin
    if LineListWanted(PrintLevel) then
      Spectrum.ShowLine(Line, Mixture.Simulations[0].OThreshold, Mixture.Units, Mixture.Simulations[0].PlotUnits, False, Problem);
    Spectrum.AddLine(Line, Mixture.Simulations[0], Mixture.Units, Problem);
  end;
begin
  TStart(AddMatrixToSpectrumTime);
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := Ev2;
  for i := 1 to Ev1.Size do begin
    { For each lower state }
    for f := 1 to Ev2.Size do begin
      { For each upper state }
      Swap := MakeLine(Line,D,Ev1,Ev2,i,f,Excitation,IncludeZero);
      if not IncludeZero and (Line.Strength = 0) then Continue; { Zero intensity transitions are boring }
      if Swap then begin
        if Ev2.Manifold.Initial and (Ev1.Manifold = Ev2.Manifold) then
          Continue; { Avoid handling transitions twice if within same manifold }
        SLine := Line;
        SLine.EigenvaluesU := Ev1;
        SLine.EigenvaluesL := Ev2;
        Sline.IndexU := i;
        Sline.IndexL := f;
        DoLine(SLine);
      end else begin
        if (Ev1.Manifold = Ev2.Manifold) and (Line.Frequency <= 0) then
          Continue; { Discard zero frequency transitions within the same manifold }
        DoLine(Line);
      end;
    end;
  end;
  TEnd(AddMatrixToSpectrumTime);
end;

function TTransitionMoments.MakeLine(out Line: TLine; Ev1, Ev2: TEigenvalues; Index1, Index2: Integer; Simulation: TSimulation): Boolean;
var
  D: TMatrix;
begin
  D := TransitionMatrix(Ev1, Ev2, True, True, nil, Simulation);
  Result := MakeLine(Line, D, Ev1, Ev2, Index1, Index2, RandomExcitation, True);
end;

function TTransitionMoments.MakeLine(out Line: TLine; D: TMatrix; Ev1, Ev2: TEigenvalues; Index1,
  Index2: Integer; const Excitation: TExcitation; IncludeZero: Boolean): Boolean;
var
  ColourI: Integer;
  te, wt: Double;
begin
  Line.EigenvaluesL := Ev1;
  Line.EigenvaluesU := Ev2;
  { For Linear TVibratingMolecules only, statistical weights can be different for the two states;
    For cases like Pi <- Sigma, scale strength up by 2;
    Symmetric/Spherical top point groups may require additional handling }
  wt := Max(Ev1.StatisticalWeight(Index1),Ev2.StatisticalWeight(index2)) * (Owner as TMolecule).Abundance * (Owner.Owner as TSpecies).Concentration;
  Line.IndexL := Index1;
  ColourI := Colour;
  if ColourI = clNone then
    ColourI := Ev1.DominantBasis[Index1].State.FindColour;
  Line.IndexU := Index2;
  te := Ev2.E[Index2] - Ev1.E[Index1];
  Result := False;
  if Ev2.Manifold.Initial then begin
    if te < 0 then begin
      te := -te;
      Result := True;
    end;
  end else if (Ev2.Manifold.LowestEstimatedOrigin < Ev1.Manifold.LowestEstimatedOrigin) then begin
    te := -te; { Presumably Emission spectrum }
    Result := True;
  end;
  if not Assigned(D) then
    Line.Strength := 0
  else
    Line.Strength := D[Index1,Index2] * wt;
  Line.Frequency := te;
  if not IncludeZero and (Line.Strength = 0) then Exit; { Zero intensity transitions are boring }
  Line.Strength := Line.Strength*IntensityCorrection(Ev2.J, Ev1.J, Excitation);
  Line.Width := 0;
  if Assigned(Ev1.Widths) then
    Line.Width := Line.Width + Ev1.Widths[Index1];
  if Assigned(Ev2.Widths) then
    Line.Width := Line.Width + Ev2.Widths[Index2];
  Line.Colour := ColourI;
  if Line.Colour = clNone then
    Line.Colour := Ev2.DominantBasis[Index2].State.FindColour;
  if Line.Colour = clNone then
    Line.Colour := clBlack;
end;

function TTransitionMoments.OriginalTransitionMatrix(
            Ev1, Ev2: TEigenvalues; Rank: TQuantumNumber; ATransitionMoment: TTransitionMoment;
            Intensity, Electric, Magnetic: Boolean; Simulation: TSimulation; NilIfEmpty: Boolean
            ): TComplexMatrix;
{ Transition Moments in original Basis for the given rank only }
var
  State1: TState;
  bk: Integer;
  function IsWantedTM(TM: TTransitionMoment): Boolean;
  begin
    Result := (TM.Rank = Rank) and ((Electric and not TM.Magnetic) or (Magnetic and TM.Magnetic));
  end;

  function NextWantedTM(var TM: TTransitionMoment; var UseList2: Boolean): Boolean;
  begin
    repeat
      if Assigned(TM) then
        TM := TM.TMlist[bk];
      if not Assigned(TM) and UseList2 then begin
        bk := 1-bk;
        TM := TMlist[bk][State1.NodeIndex];
        UseList2 := False;
      end;
      Result := Assigned(TM);
    until not Result or IsWantedTM(TM);
  end;

  procedure MakeResult;
  begin
    if Assigned(Result) then
      Exit;
    Result := TComplexMatrix.Create(Ev1.Size, Ev2.Size);
    Result.Zero;
  end;

var
  tm: TTransitionMoment;
  State2: TState;
  b, bb, bkFirst, t, nTM: Integer;
  Sum, V: Complex;
  FirstTM: TTransitionMoment;
  UseList2, FirstUseList2: Boolean;
begin
  Result := nil;
  if not Ev1.HasFields and (Ev1.Size > 0) and (Ev2.Size > 0)
     and not((Ev1.SpinSpecies(1) = Ev2.SpinSpecies(1)) or Ev1.MultipleSpinSpecies or Ev2.MultipleSpinSpecies) then
    raise EreportException.CreateFmt('Inconsistent statistical weights for %s in %s', [Name, Molecule.UniqueName]);
  if not NilIfEmpty then
    MakeResult;
  if Assigned(ATransitionMoment) then begin
    if not IsWantedTM(ATransitionMoment) then
      Exit;
  end else begin
    nTM := 0;
    for t := 0 to Count-1 do
      if IsWantedTM(TransitionMoment[t]) then begin
        Inc(nTM);
        ATransitionMoment := TransitionMoment[t];
      end;
    case nTM of
      0: Exit;
      1: { ATransitionMoment already set };
    else
      ATransitionMoment := nil;
    end;
  end;
  TStart(MakeTransitionMatrixTime);
  try
    for b := 1 to Ev1.Size do begin
      State1 := Ev1.Basis[b].State;
      if Assigned(ATransitionMoment) then begin
        if not ((ATransitionMoment.Bra = State1) or (ATransitionMoment.Ket = State1)) then
          Continue;
      end else begin
        if Ev1.Manifold = Bra then
          bk := 0
        else
          bk := 1;
        FirstTM := TMlist[bk][State1.NodeIndex];
        FirstUseList2 := Bra = Ket;
        if not Assigned(FirstTM) or not IsWantedTM(FirstTM) then
          if not NextWantedTM(FirstTM, FirstUseList2) then
            Continue;
        bkFirst := bk;
      end;
      for bb := 1 to Ev2.Size do begin
        Sum := z0;
        State2 := Ev2.Basis[bb].State;
        if Assigned(ATransitionMoment) then
          tm := ATransitionMoment
        else begin
          tm := FirstTM;
          UseList2 := FirstUseList2 and (State1 <> State2); { Avoid double counting }
          bk := bkFirst;
        end;
        repeat
          if (tm.Bra = State1) and (tm.Ket = State2) then begin
            V := tm.MatrixElement(Ev1.Basis[b], Ev2.Basis[bb], Intensity);
            if Ev1.HasFields then { Can't apply to sum, as complex conjugate of matix sometimes evaluated }
              CMul(V, Mfactor(Ev1.Basis[b].F, Ev1.J, tm.Rank, Ev2.Basis[bb].F, Ev2.J, Simulation));
            Cadd(Sum, V)
          end else if (tm.Bra = State2) and (tm.Ket = State1) then begin
            { Work out complex conjugate }
            V := tm.MatrixElement(Ev2.Basis[bb], Ev1.Basis[b], Intensity);
            if Ev1.HasFields then
              CMul(V, Mfactor(Ev2.Basis[bb].F, Ev2.J, tm.Rank, Ev1.Basis[b].F, Ev1.J, Simulation));
            V.im := -V.im;
            Cadd(Sum, V);
          end;
        until Assigned(ATransitionMoment) or not NextWantedTM(tm, UseList2);
        if not IsZero(Sum) then begin
          if not Ev1.HasFields then
            CMul(Sum, Sqrt(One/(Rank+1))); { Sum over M }
          MakeResult;
          Result.AutoSetValue(b,bb,Sum);
        end;
      end;
    end;
  except
    Result.Free;
    raise;
  end;
  TEnd(MakeTransitionMatrixTime);
end;

function TTransitionMoments.TransitionMatrix(
  Ev1, Ev2: TEigenvalues; Wait, Cache: Boolean;
  TransitionMomentSel: TTransitionMoment;
  Simulation: TSimulation;
  APrintLevel: TPrintLevel): TComplexMatrix;
{ Transformed Transition Moments squared }
  function FindTransitionMatrix(WriteHeld:Boolean): TCachedTransitionMatrix;
  var
    i: Integer;
  begin
    if not WriteHeld and Assigned(Molecule.TransitionsMomentsCacheLock) then
      Molecule.TransitionsMomentsCacheLock.BeginRead;
    Result := nil;
    try
      for i := 0 to High(Ev1.CachedTransitionMatrices) do begin
        Result := Ev1.CachedTransitionMatrices[i];
        if (Result.OtherEigenvalues = Ev2) and (Result.TransitionMoments = Self) and (Result.Simulation = Simulation) then
          Exit;
      end;
      Result := nil;
    finally
      if not WriteHeld and Assigned(Molecule.TransitionsMomentsCacheLock) then
        Molecule.TransitionsMomentsCacheLock.EndRead;
    end;
  end;
var
  D: TComplexMatrix;
  i: Integer;
  Rank: Integer;
  DoneRank: array of Boolean;
  CacheEntry: TCachedTransitionMatrix;
begin
  if not Wait and (not Cache or Assigned(TransitionMomentSel)) then
    { Wait makes no sense if not caching }
    raise Exception.Create('TTransitionMoments.TransitionMatrix');
  if not Assigned(TransitionMomentSel) and Cache then begin
    CacheEntry := FindTransitionMatrix(False);
    if not Assigned(CacheEntry) then
      try
        if Assigned(Molecule.TransitionsMomentsCacheLock) then
          Molecule.TransitionsMomentsCacheLock.BeginWrite;
        { Check that it hasn't been created by another thread }
        CacheEntry := FindTransitionMatrix(True);
        if not Assigned(CacheEntry) then begin
          SetLength(Ev1.CachedTransitionMatrices, Length(Ev1.CachedTransitionMatrices)+1);
          CacheEntry := TCachedTransitionMatrix.Create(nil);
          Ev1.CachedTransitionMatrices[High(Ev1.CachedTransitionMatrices)] := CacheEntry;
          if APrintLevel = plDefault then
            APrintLevel := Mixture.PrintLevel;
          with CacheEntry do begin
            TransitionMoments := Self;
            EigenValues := Ev1;
            OtherEigenvalues := Ev2;
            PrintLevel := APrintLevel;
          end;
          CacheEntry.Simulation := Simulation;
        end;
      finally
        if Assigned(Molecule.TransitionsMomentsCacheLock) then
          Molecule.TransitionsMomentsCacheLock.EndWrite;
      end;
    { We now have the structure, but not necessarily complete }
    CacheEntry.Make(Wait,nil);
    if Wait then
      Result := CacheEntry.TransitionMatrix
    else
      Result := nil;
    Exit;
  end;
  { Check that required matrices are available }
  Ev1.Make(True, [meHrot, meWidth]);
  Ev2.Make(True, [meHrot, meWidth]);
  TStart(TransitionMatrixTime);
  Result := nil;
  SetLength(DoneRank, MaxRank + 1);
  for i := 0 to MaxRank do
    DoneRank[i] := False;
  try
    for i := 0 to Count-1 do
      { Each rank must be done separately }
      if (TransitionMoment[i] = TransitionMomentSel)
         or (not Assigned(TransitionMomentSel) and not DoneRank[TransitionMoment[i].Rank]) then begin
        Rank := TransitionMoment[i].Rank;
        DoneRank[Rank] := True;
        D := OriginalTransitionMatrix(Ev1, Ev2, Rank, TransitionMomentSel, True, True, True, Simulation, True);
        if APrintLevel >= plMatrices then
          Ev1.ShowTransitionMatrix(Ev2, D, Format('Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0);
        if Assigned(D) then begin
          try
            Ev1.TransformTransitionMatrix(D, Ev2);
            if APrintLevel >= plMatrices then
              Ev1.ShowTransitionMatrix(Ev2, D, Format('Transformed Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0, False);
            D.Sqr;
            if APrintLevel >= Succ(plMatrices) then
              Ev1.ShowTransitionMatrix(Ev2, D, Format('Squared Rank %s Transition Matrix for ', [Mixture.QuantumNumberToStr(Rank)]), 0, False);
          except
            D.Free;
            raise;
          end;
          if Assigned(Result) then begin
            Result.Add(D);
            D.Free;
          end else
            Result := D;
        end;
      end;
  except
    Result.Free;
    raise;
  end;
  TEnd(TransitionMatrixTime);
end;

function TTransitionMoments.Mixture: TMixture;
begin
  Result := Owner.Owner.Owner as TMixture;
end;

class function TTransitionMoments.BraKetClass: TMNodeClass;
begin
  Result := TManifold;
end;

function TTransitionMoments.Molecule: TMolecule;
begin
  Result := Owner as TMolecule;
end;

function TTransitionMoments.PairName: string;
begin
  Result := 'mu';
end;

function TTransitionMoments.IsSettingColour(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingColour(AIndex);
end;

function TTransitionMoments.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := ColourHelpString
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TAxisSwitch }

class function TAxisSwitch.CanBeChildOf(MNode: TMNode): Boolean;
begin
  Result := inherited CanBeChildOf(MNode);
  if not Result then
    Exit;
  Result := MNode.Owner.Owner is TAsymmetricMolecule;
  if not Result then
    Exit;
  { Only one axis switch object allowed }
  Result := MNode.ChildWithClass(TAxisSwitch) = nil;
end;

constructor TAxisSwitch.Create(AOwner: TComponent);
begin
  DimensionlessParameters := DimensionlessParameters + ',Chi,Theta,Phi';
  inherited;
end;

function TAxisSwitch.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex <= 3 then
    Result := 'Axis switching angle (degrees)'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TTransitionMoment }

constructor TTransitionMoment.Create(AOwner: TComponent);
begin
  FRank := 2;
  inherited Create(AOwner);
end;

constructor TTransitionMoment.Create(AOwner: TComponent; State1Name, State2Name: string);
begin
  Create(AOwner);
  FBraName := State1Name;
  FKetName := State2Name;
end;

destructor TTransitionMoment.Destroy;
begin
  if FMatrixElements <> FNoHyperfineMatrixElements then
    DiscardMatrixElements(FNoHyperfineMatrixElements);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  ParameterVariable.Free;
  inherited;
end;

function TTransitionMoment.FixupProperties: Exception;
var
  i: Integer;
begin
  if FMatrixElements <> FNoHyperfineMatrixElements then
    DiscardMatrixElements(FNoHyperfineMatrixElements);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  FMatrixElementsEmpty := False;
  FMatrixElementsImplemented := True;
  FVibronicallyAllowed := vaMaybe; { This variable otherwise handled at molecule level only }
  FAxisSwitcher := nil;
  FnPerturbations := 0;
  FActiveTransitionFunction := nil;
  if not Assigned(ParameterVariable) then
    FParameterVariable := TSortedVariable.Create(ParameterNames[1],500); { TODO: Name needs work }
  FParameterVariable.Value := GetParameter(1);
  FParameterVariable.SetAllValues;
  Result := inherited FixupProperties;
  if not Assigned(Result) and Odd(Rank) then
    Result := EReportException.Create('Transition Moment Rank must be integral in '+PairName);
  if not Assigned(Result) then
    for i := 0 to ComponentCount-1 do
      if Items[i] is TAxisSwitch then
        if AxisSwitcher <> nil then begin
          Result := EReportException.Create('Only one axis switching object allowed in '+PairName);
          Exit;
        end else
          FAxisSwitcher := Items[i] as TAxisSwitch
      else if Items[i] is TPerturbation then
        Inc(FnPerturbations)
      else if (Items[i] is TTransitionFunction) and TTransitionFunction(Items[i]).FullyActive then
        if (ActiveTransitionFunction <> nil) and not Assigned(Result) then
          Result := EReportException.Create('Only one active Transition Function object allowed in '+PairName)
        else
          FActiveTransitionFunction := TTransitionFunction(Items[i]);
end;

function TTransitionMoment.Allowed(Sym1, Sym2: Integer): Boolean;
begin
  Result := (Owner.Owner as TMolecule).Allowed(Self, Sym1, Sym2);
end;

function TTransitionMoment.MatrixElement(ABra, AKet: TBasisState; Intensity: Boolean): Complex;
begin
  Result := ABra.State.TransitionMatrixElement(Self, ABra, AKet);
//  if Magnetic and HasUnits((RootNode as TMixture).Simulations[0].IntensityUnits)  then
  if Magnetic then begin
    if Intensity then
      CMul(Result, Result, 1e21 * BohrMagneton);
    if NuclearMagneton then
      CMul(Result,MeOverMp);
  end;
  if Assigned(ActiveTransitionFunction) then
    ActiveTransitionFunction.MatrixElement(Result, ABra, AKet, 1, 1, Intensity);
end;

procedure TTransitionMoment.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  inherited;
  if (PrintLevel < plDetail) or (GetParameter(1) = 0) then
    Exit;
  try
    ShowSymbolicMatrixElements(Log, Indent);
  except
    Writeln(Log, Indent, 'Symbolic Matrix Elements not available for ',  Name);
  end;
end;

procedure TTransitionMoment.SetParameter(AIndex: Integer; const Value: Double);
begin
  inherited;
  if Assigned(ParameterVariable) and (AIndex = 1) then
    FParameterVariable.Value := GetParameter(AIndex);
end;

procedure TTransitionMoment.ShowSymbolicMatrixElements(var Log: Text; const Indent: string);
  procedure ShowBoth(Elements: TMatrixElements; const Prefix: string);
  var
    j: Integer;
    E: TExprNode;
  begin
    if Length(Elements) = 0 then
      Exit;
    Writeln(Log, Indent, Prefix, 'Reduced Matrix Elements of ',  Name);
    for j := 0 to High(Elements) do
      WriteLn(Log, Indent, (Bra as TState).MatrixElementType(Elements[j], PairName, Prefix <> ''), ' = ', Elements[j].Expr.Text(True));
    Writeln(Log, Indent, Prefix, 'Line Strengths from ',  Name);
    for j := 0 to High(Elements) do begin
      E := EB([Elements[j].Expr, '*', Elements[j].Expr]);
      WriteLn(Log, Indent, (Bra as TState).MatrixElementType(Elements[j], PairName, Prefix <> ''), ' = ', E.Text(True));
      E.Free;
    end;
  end;
var
  Elements, SymmetrizedElements: TMatrixElements;
begin
  TStart(IndividualSymbolicMatrixElementTime);
  Elements := SymbolicMatrixElements(True, True, True);
  TEnd(IndividualSymbolicMatrixElementTime);
  ShowBoth(Elements,'');
  SymmetrizedElements := (Bra as TState).ConvertToSymmetrized(Elements);
  if Length(SymmetrizedElements) > 0 then
    ShowBoth(SymmetrizedElements,'Symmetrized ');
  if Length(Elements) > 0 then
    WriteLn(Log);
  DiscardMatrixElements(Elements);
  DiscardMatrixElements(SymmetrizedElements);
end;

function TTransitionMoment.SymbolicMatrixElements(Intensity: Boolean; AddHyperfineCoupling: Boolean; IncompleteOK: Boolean): TMatrixElements;
begin
  Result := (Bra as TState).SymbolicTransitionMatrixElements(Self, AddHyperfineCoupling, IncompleteOK);
  { TODO -cSymbolicIntens: Need to add BohrMagneton as non symbolic? }
end;

function TTransitionMoment.BraOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Bra;
end;

function TTransitionMoment.KetOwner: TMNode;
begin
  Result := (Owner as TTransitionMoments).Ket;
end;

function TTransitionMoment.CheckMatrixElementsAvailable: Boolean;
begin
  Result := FMatrixElementsImplemented; { Default is that they can be implemented; Return false if not implementable }
  if not Result or FMatrixElementsEmpty or (Length(FMatrixElements) > 0) then
    Exit;
  MustBeSerial;
  try
    FMatrixElements := SymbolicMatrixElements(True, True);
  except
    Result := False;
    FMatrixElementsImplemented := False;
    Exit;
  end;
  CompileMatrixElements(FMatrixElements, ParameterVariable.Template);
  FMatrixElementsEmpty := Length(FMatrixElements) = 0;
  if FMatrixElementsEmpty then
    Exit;
  if (Owner.Owner as TMolecule).nNuclei > 0 then begin
    FNoHyperfineMatrixElements := SymbolicMatrixElements(True, False);
    CompileMatrixElements(FNoHyperfineMatrixElements, ParameterVariable.Template);
  end else
    FNoHyperfineMatrixElements := FMatrixElements;
  FSymmetrizedMatrixElements := (Bra as TState).ConvertToSymmetrized(MatrixElements);
  CompileMatrixElements(FSymmetrizedMatrixElements, nil);
end;

function TTransitionMoment.CheckProjections(K1, K2: TQuantumNumber): Boolean;
begin
  { Check change in given projections for consistency with transition moment;
    only return False if the matrix element is definately zero.
    Only used for optimization }
  Result := (AxisSwitcher <> nil) or (nPerturbations > 0) or (Abs(K1-K2) <= Rank) or (Abs(K1+K2) <= Rank);
end;

class function TTransitionMoment.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

function TTransitionMoment.GetChildTypes(AIndex: Integer): TMNodeClass;
begin
  if AIndex = inherited nChildTypes then
    Result := TCustomTransitionFunction
  else
    Result := inherited GetChildTypes(AIndex);
end;

function TTransitionMoment.nChildTypes: Integer;
begin
  Result := inherited nChildTypes + 1;
end;

{ TCartesianTransitionMoment }

const
  AxisNames: array [0..3] of PChar = ('a', 'b', 'c', nil);

function AxisName(AAxis: TabcAxis): string;
begin
  Result := AxisNames[Ord(AAxis)];
end;

function TCartesianTransitionMoment.GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = 1 then
    Result := @AxisNames[0];
end;

function TCartesianTransitionMoment.FixupProperties:Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) and (Rank <> 2) then
    Result := EReportException.CreateFmt('%s-%s: Transition Moment Rank Not 1: %d', [Name, AxisName(Axis), Rank]);
end;

function TCartesianTransitionMoment.GetAxis(const Index: Integer): TabcAxis;
begin
  Result := TabcAxis(GetSetting(Index));
end;

procedure TCartesianTransitionMoment.SetAxis(const Index: Integer; const AValue: TabcAxis);
begin
  SetSetting(Index, Ord(AValue));
end;

constructor TCartesianTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Axis=a';
  inherited;
end;

function TCartesianTransitionMoment.PairName: string;
begin
  Result := AxisName(Axis);
end;

function TCartesianTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Axis of transtion moment - a, b or c.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

{ TSphericalTransitionMoment }

function TSphericalTransitionMoment.PairName: string;
begin
  if Component <= -9999 then
    Result := Format('T(%s)', [TMixture(RootNode).QuantumNumberToStr(Rank)])
  else
    Result := Format('T(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

procedure TSphericalTransitionMoment.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  inherited SetSetting(AIndex, AValue);
  FRank := Rank;
end;

procedure TSphericalTransitionMoment.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if (AIndex = 2) and (Str = 'auto') then
    Component := -9999
  else
    inherited;
end;

function TSphericalTransitionMoment.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = 2) and (Component <= -9999) then
    Result := 'auto'
  else
    Result := inherited SettingToStr(AIndex);
end;

constructor TSphericalTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Rank=2,Component=auto';
  inherited;
end;

function TSphericalTransitionMoment.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result) and (Component <= -9999) then
    if Rank = 0 then
      FActualComponent := 0 { No Choice! }
    else
      Result := (Bra as TState).AllowedSphericalComponent(Self, Ket as TState, FActualComponent)
  else
    FActualComponent := Component;
  if not Assigned(Result) and (Odd(FActualComponent) or (Abs(FActualComponent) > Rank)) then
    Result := EReportException.Create('Invalid Transition Moment component in '+UniqueName);
end;

function TSphericalTransitionMoment.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  if ((RootNode as TMixture).ReadVersion < 5002300)
     and not Assigned((Node as TDOMElement).GetAttributeNode('Component')) then
    { Back compatibility - default changes from "0" to "auto" for component }
    (Node as TDOMElement).SetAttribute('Component', '0');
  Result:=inherited ReadProperties(Node, CheckUnread);
end;

function TSphericalTransitionMoment.HelpOnSetting(AIndex: Integer): string;
var
  E: Exception;
begin
  if AIndex = 1  then
    Result := 'Rank of transition - ' + GlobalMixture.QuantumNumberToStr(2) + ' for a normal electric dipole transition.'
  else if AIndex = 2 then
    if (Component <= -9999) and Assigned(Bra) and Assigned(Ket) then begin
      try
        E := (Bra as TState).AllowedSphericalComponent(Self, Ket as TState, FActualComponent)
      except
        on Ex: Exception do
          E := EReportException.Create(Ex.Message);
      end;
      if Assigned(E) then begin
        Result := 'Problem with auto setting: ' + E.Message
      end else
        Result := 'Current value = ' + (Owner.Owner as TMolecule).QuantumNumberToStr(FActualComponent);
    end else
      Result := 'Projection quantum number of transition moment - auto works in many cases'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSphericalTransitionMoment.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [1,2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

{ TEigenvalues }

function TEigenvalues.FieldLabel: string;
begin
  Result := '';
  if ExField.E <> 0 then
    Result := Format('Electric Field = %g', [ExField.E]);
  if ExField.B <> 0 then
    if Result <> '' then begin
      Result := Result + ', ';
    Result := Result + Format('Magnetic Field = %g', [ExField.B]);
    if ExField.E <> 0  then
      Result := Result + Format(', Angle = %g', [ExField.Angle]);
  end
end;

function TEigenvalues.FindDerivatives(const Parameter: TParameter): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to High(Derivatives) do
    with Derivatives[i] do
      if (MNode = Parameter.MNode) and (Index = Parameter.Index) then begin
        Result := i;
        Exit;
      end;
end;

procedure TEigenvalues.Fixup;
var
  i: Integer;
  MemNeeded: Double;
begin
  MemNeeded := ComponentCount;
  MemNeeded := Sqr(MemNeeded)*SizeOf(Double);
  if Manifold.Mixture.AllowComplex then
    MemNeeded := MemNeeded*2;
  if MemNeeded > High(PtrInt) then
    raise EReportException.Create('Basis too large:'+IntToStr(ComponentCount));
  { Make up Basis array }
  SetLength(Basis, ComponentCount+1); { Element 0 is unused }
  for i := 1 to ComponentCount do
    Basis[i] := Components[i-1] as TBasisState;
  Manifold.Mixture.LargestIndex := Max(Manifold.Mixture.LargestIndex,Size);
  FSize := ComponentCount;
end;

procedure TEigenvalues.MakeBasis(BasisOrder: TBasisOrder);
var
  StatWtGroup: SmallInt;
  procedure DoAddBasis(State: TState; F, M: TQuantumNumber; Sym: Integer);
  var
    LastCount: Integer;
    LastStatWtGroup: SmallInt;
  begin
    LastCount := ComponentCount;
    State.AddBasis(Self, F, M, Sym);
    if ComponentCount > LastCount then begin
      LastStatWtGroup := (Components[LastCount] as TBasisState).StatWtGroup;
      if StatWtGroup < 0 then
        StatWtGroup := LastStatWtGroup
      else if LastStatWtGroup <> StatWtGroup then
        if not HasFields then
          raise Exception.CreateFmt('Inconsistent Statistical Weight groups for %s : %d <> %d', [State.UniqueName, StatWtGroup, LastStatWtGroup]);
    end;
  end;
  procedure AddBasis(State: TState);
  var
    i: Integer;
    F: TQuantumNumber;
  begin
    if HasFields or (Sym < 0) then begin
      F := Max(Abs(J),Manifold.FindJmin(0));
      while F <= Manifold.FindJmax(F+4) do begin
        for i := 0 to Manifold.nSym-1 do { TODO -cField: May not need every symmetry with Field in place }
          DoAddBasis(State, F, J, i);
        Inc(F,2);
      end;
    end else
      DoAddBasis(State, J, 0, Sym);
  end;
var
  i: Integer;
  V: TDArray;
begin
  if Size >= 0 then
    Exit; { Already done }
  TStart(MakeBasisTime);
  StatWtGroup := -1;
  if Assigned(Manifold) then begin { Not Making dummy Eigenvalues object }
    if Assigned(SelectedState) then
      { If asked, just use specified state }
      AddBasis(SelectedState)
    else
      { Make up Basis }
      for i := 0 to Manifold.nStates-1 do
        AddBasis(Manifold.State[i]);
    Fixup;
  end;
  if BasisOrder <> boDefault then begin
    V := DiagonalMatrixElements(meHrot, 0, not NoPerturbations);
    SortBasisOn(V, BasisOrder = boStateDiagonal);
  end;
  TEnd(MakeBasisTime);
end;

constructor TEigenvalues.Create(AManifold: TManifold; AJ: TQuantumNumber;
  ASym: Integer; const AExField: TStaticFields; IncludePerturbations: Boolean;
  AState: TState; RemoveSpins: Byte);
begin
  inherited Create(nil);
  Manifold := AManifold;
  J := AJ; { M if HasFields }
  SelectedState := AState;
  LastSubBasisIndex := 0;
  RemovedSpins := RemoveSpins;
  NoPerturbations := not IncludePerturbations;
  ExField := AExField;
  FHasFields := IsNonZero(ExField);
  if Assigned(Manifold) then begin { Not Making dummy Eigenvalues object }
    if HasFields then
      Sym := -1 { Do not use symmetry }
    else begin
      Sym := ASym;
      if (Sym < 0) or (Sym >= Manifold.nSym) then
        raise Exception.CreateFmt('%s: Invalid symmetry number %d', [Manifold.Name, Sym]);
    end;
    if Assigned(AState) and (AState.Manifold <> Manifold) then
        raise Exception.Create('State does not belong to manifold');
    FSize := -1; { Flag basis not set up }
  end;
end;

function TEigenvalues.Degeneracy(AIndex: Integer): Integer;
begin
  Result := StatisticalWeight(AIndex);
  if not HasFields then
    Result := Result * (J+1);
end;

destructor TEigenvalues.Destroy;
var
  i: Integer;
begin
  for i := 0 to High(CachedTransitionMatrices) do
    FreeAndNil(CachedTransitionMatrices[i]);
  FreeAndNil(FE);
  FreeAndNil(FC);
  FreeAndNil(FWidths);
  FreeAndNil(FDominantBasisIndex);
  for i := 0 to High(Derivatives) do
    FreeAndNil(Derivatives[i].Values);
  inherited Destroy;
end;

function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType; Base: Integer = 0; IncludePerturbations: Boolean = False): TDArray;
var
  i, p: Integer;
begin
  SetLength(Result, Size-Base);
  for i := Base+1 to Size do begin
    Result[i-1-Base] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType).re; { TODO: Check for non-zero im }
    if IncludePerturbations then
      for p := 0 to Manifold.nPerturbations-1 do
        if Manifold.Perturbation[p].Connects(Basis[i].State, Basis[i].State) then
          Result[i-1-Base] := Result[i-1-Base] + Manifold.Perturbation[p].MatrixElement(Basis[i], Basis[i], MatElemType).re;
  end;
end;

{
function TEigenvalues.DiagonalMatrixElements(MatElemType: TMatElemType): array of Double;
var
  i: Integer;
begin
  SetLength(Result, Size+1);
  Result := TColumnVector.Create(Size);
  try
    for i := 1 to Size do
      Result[i] := Basis[i].State.MatrixElement(Basis[i], Basis[i], MatElemType);
  except
    Result.Free;
    raise;
  end;
end;
}
procedure TEigenvalues.MatrixElements(out Matrices: TComplexMatrices; BlockMatrix: Boolean; MatElemType: TMatElemType; IncludePerturbations: Boolean);
{ Note mostly duplicated to EigBlock }
var
  {$IFDEF SMALLSET}
  { Track non-zero matrix elements to allow matrix to be blocked if possible }
  Mask: Int64;
  nSet: Integer;
  Connections: array of Int64;
  {$ENDIF}
  { Track non-zero matrix elements to allow matrix to be blocked if possible }
  Mask: TBigSet;
  Connections: array of TBigSet;
  nBlocks: Integer;

  procedure FactorMatrix(var Matrices: TComplexMatrices; const Connections: array of TBigSet);
    { Given the Hamiltonian Matrix and a bitmap of non-zero elements, rearrange elements to block-diagonalize }
  var
    b, bb, bbb, bNext, i, k, CIndex, BlockSize: Integer;
    CurMask, Mask, bMask: TBigSet;
    Matrix, Block: TComplexMatrix;
    Locations: array of Integer;
    V: Complex;
  begin
    Matrix := Matrices[0];
    try
      { Now split matrix }
      SetLength(Matrices,nBlocks);
      for i := 0 to nBlocks - 1 do
        Matrices[i] := nil;
      b := Size;
      CIndex := Size-1;
      Mask.Size := Connections[Size].Size;
      bMask.Size := Mask.Size;
      SetLength(Locations, Size); { Note not all entries used }
      repeat
        BlockSize := 0;
        bNext := 0;
        Mask.Clear;
        Mask.SetBitn(b-1);
        CurMask := Connections[b];
        for bb := b downto 1 do begin
          if CurMask.AndNonZero(Mask) then begin
            { Rearrange basis by block }
            Basis[bb].ComponentIndex := CIndex;
            Dec(CIndex);
            Inc(BlockSize);
          end else if (bNext = 0) and (Connections[bb].Size <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
          Mask.RotateRight;
        end;
        TStart(BlockMoveTime);
        { Create next block matrix }
        Block := TComplexMatrix.Create(BlockSize,BlockSize);
        Matrices[nBlocks-1] := Block;
        Dec(nBlocks);
        Block.Mode := Matrix.Mode;
        { Copy elements }
        Mask.Clear;
        Mask.SetBitn(b-1);
        i := BlockSize;
        Locations[1] := -1; { Flag not set up }
        for bb := b downto 1 do begin
          if CurMask.AndNonZero(Mask) then begin
            { This basis state contributes }
            { On first hit, make array of contributing states
              to avoid multiple set operations }
            if Locations[1] < 0 then begin
              bMask.CopySet(Mask);
              k := i;
              for bbb := bb downto 1 do begin
                { For each other basis state }
                if CurMask.AndNonZero(bMask) then begin
                  Locations[k] := bbb;
                  Dec(k);
                end;
                bMask.RotateRight;
              end;
            end;
            Block[i,i] := Matrix[bb,Locations[i]];
            for k := 1 to i-1 do begin
              V := Matrix[bb,Locations[k]];
              Block[i,k] := V;
              Block[k,i] := V;
            end;
//            bMask.CopySet(Mask);
//            k := i;
//            for bbb := bb downto 1 do begin
//              { For each other basis state }
//              if CurMask.AndNonZero(bMask) then begin
//                Block[i,k] := Matrix[bb,bbb];
//                if i <> k then
//                  Block[k,i] := Matrix[bbb,bb];
//                Dec(k);
//              end;
//              bMask.RotateRight;
//            end;
            Dec(i);
          end;
          Mask.RotateRight;
        end;
        TEnd(BlockMoveTime);
        b := bNext;
      until b = 0;
      { Sanity check }
      if nBlocks <> 0 then
        raise Exception.Create('MatrixElements.FactorMatrix:n');
      bb := 0;
      for b := 0 to High(Matrices) do
        Inc(bb,Matrices[b].RowCount);
      if bb <> Size then begin
        raise Exception.Create('MatrixElements.FactorMatrix:RC');
        //WriteLn(Log, bb);
        //Clear(Mask);
        //for b := 1 to High(Connections) do
        //  if High(Connections[b]) > 0 then
        //    OrIn(Mask,Connections[b]);
        //MatrixElements(Matrices, True, meHrot);
        //for b := 1 to High(Connections) do begin
        //  Write(Log, b:4, ' ');
        //  for bb := Size downto 1 do
        //    Write(Log, ((Connections[b][0] shr (bb-1)) and 1));
        //  WriteLn(Log);
        //end;
      end;
    finally
      Matrix.Free;
    end;
    Fixup;
  end;

  function CountBlocks:Integer;
  { Count how many blocks to factor into }
  var
    b, bb, bNext: Integer;
    MaskChanged: Boolean;
    CurMask: TBigSet;
  begin
    TStart(CountBlocksTime);
    { Work backwards from last row, as this will have most connections set }
    b := Size;
    Result := 0;
    repeat
      CurMask := Connections[b];
      Inc(Result);
      repeat
        bNext := 0;
        MaskChanged := False;
        for bb := b-1 downto 1 do begin
          if CurMask.AndNonZero(Connections[bb]) then begin
            MaskChanged := CurMask.OrInChanged(Connections[bb]) or MaskChanged;
            Connections[bb].Size := 0;
          end else if (bNext = 0) and (Connections[bb].Size <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
        end;
      until not MaskChanged;
      b := bNext;
    until bNext = 0;
    nBlocks  := Result;
    TEnd(CountBlocksTime);
  end;

  {$IFDEF SMALLSET}
  function CountBlocks:Integer;
  { Count how many blocks to factor into }
  var
    b, bb, bNext: Integer;
    MaskChanged: Boolean;
    CurMask: Int64;
  begin
    { Work backwards from last row, as this will have most connections set }
    b := Size;
    Result := 0;
    repeat
      CurMask := Connections[b];
      Inc(Result);
      repeat
        bNext := 0;
        MaskChanged := False;
        for bb := b-1 downto 1 do begin
          if (CurMask and Connections[bb]) <> 0 then begin
            if (CurMask or Connections[bb]) <> CurMask then begin
              MaskChanged := True;
              CurMask := CurMask or Connections[bb];
            end;
            { Flag this basis state has been copied }
            Connections[bb] := 0
          end else if (bNext = 0) and (Connections[bb] <> 0) then
            bNext := bb; { Highest unused basis state not involved in the current matrix }
        end;
      until not MaskChanged;
      Connections[b] := CurMask;
      b := bNext;
    until bNext = 0;
    Connections[0] := Result;
  end;
  {$ENDIF}
var
  b, bb, EBField: Integer;
  Current: TState;
  Conv, SmallE: Double;
  TransitionMoments: TTransitionMoments;
  Matrix, EMat: TComplexMatrix;
  Sum: Complex;
  OK: Boolean;
begin
  SmallE := Manifold.Mixture.SmallE;
  BlockMatrix := BlockMatrix and (Size > 1);
  SetLength(Matrices,1);
  Matrices[0] := nil;
  if Size = 0 then Exit;
  Matrices[0] := TComplexMatrix.Create(Size, Size);
  Matrix := Matrices[0];
  Matrix.Zero;
  if BlockMatrix then begin
    SetLength(Connections,Size+1);
    Mask.Size := Size;
    Mask.SetBitn(0);
  end;
  OK := False;
  try
    for b := 1 to Size do begin
      Current := Basis[b].State;
      if BlockMatrix then begin
        Connections[b].CopySet(Mask); { Diagonal elements set }
        Mask.RotateLeft;
      end;
      for bb := 1 to b do
        if (Current = Basis[bb].State) and (not HasFields or ((Basis[b].F = Basis[bb].F) and (Basis[b].Sym = Basis[bb].Sym))) then begin
          Sum := Current.MatrixElement(Basis[b], Basis[bb], MatElemType);
          if not BlockMatrix or not MathSup.IsZero(Sum, SmallE) then begin
            if BlockMatrix and (b <> bb) then
              Connections[b].OrIn(Connections[bb]);
            Matrix.AutoSetValue(b,bb,Sum);
            if b <> bb then begin
              Sum.im := -Sum.im;
              Matrix[bb,b] := Sum;
            end;
          end;
        end
    end;
    { Shift perturbations to separate loop for timing purposes }
    if IncludePerturbations and (Manifold.nPerturbations > 0) then
      AddPerturbation(Matrix, nil, MatElemType, Connections, SmallE);
    //  TStart(PerturbationMatrixElementTime);
    //  for b := 1 to Size do begin
    //    Current := Basis[b].State;
    //    if (Current.PerturbationList[0] <> nil) or (Current.PerturbationList[1] <> nil) then
    //      for bb := 1 to b do begin
    //        Sum := z0;
    //        nSet := 0;
    //        if not HasFields or ((Basis[b].F = Basis[bb].F) and (Basis[b].Sym = Basis[bb].Sym)) then begin
    //          //for p := 0 to Manifold.nPerturbations-1 do
    //          //  if Manifold.Perturbation[p].Connects(Current, Basis[bb].State) then begin
    //          //    CAdd(Sum, Manifold.Perturbation[p].MatrixElement(Basis[b], Basis[bb], MatElemType));
    //          //    Inc(nSet);
    //          //  end;
    //          { Scan Bra list, then Ket list }
    //          for bk := 0 to 1 do begin
    //            Perturbation := Current.PerturbationList[bk];
    //            while Assigned(Perturbation) do begin
    //              if ((bk = 0) or (Perturbation.Bra <> Perturbation.Ket)) { Avoid applying digonal elements twice }
    //                 and Perturbation.Connects(Current, Basis[bb].State) then begin
    //                CAdd(Sum, Perturbation.MatrixElement(Basis[b], Basis[bb], MatElemType));
    //                Inc(nSet);
    //              end;
    //              Perturbation := Perturbation.PerturbationList[bk];
    //            end;
    //          end;
    //        end;
    //        if (nSet > 0) and not MathSup.IsZero(Sum, SmallE) then begin
    //          if BlockMatrix then
    //            Connections[b].OrIn(Connections[bb]);
    //          CAdd(Sum, Matrix[b,bb]);
    //          Matrix.AutoSetValue(b,bb,Sum);
    //          if b <> bb then begin
    //            Sum.im := -Sum.im;
    //            Matrix[bb,b] := Sum;
    //          end;
    //        end;
    //      end;
    //  end;
    //  TEnd(PerturbationMatrixElementTime);
    //end;
    if HasFields then begin
      { Transition moments diagonal in manifold }
      TransitionMoments := Manifold.Molecule.TransitionMoments[Manifold.NodeIndex,Manifold.NodeIndex];
      if Assigned(TransitionMoments) then begin
        for EBField := 0 to 1 do begin
          if EBField = 0 then
            { This factor is correct for electric fields - Debye = 1e-21/c Cm}
            Conv := ExField.E * ConversionFactor(MHz,Manifold.Mixture.Units)*1e-27/SpeedOfLight/Planck
          else
            { function returns BohrMagneton, so convert from J to required energy unit }
            Conv := ExField.B * ConversionFactor(MHz,Manifold.Mixture.Units) * 1e-6 * BohrMagneton / Planck;
          { TODO -cField: Caching likely to help a lot, as part matrices field independent }
          { TODO -cField: Check/Handle Rank <> 1 case; Sum over ranks }
          if Conv <> 0 then begin
            EMat := TransitionMoments.OriginalTransitionMatrix(Self,Self,TransitionMoments.MaxRank,nil,False,EBField=0,EBField=1,nil,True);
            if Assigned(EMat) then
              for b := 1 to Size do
                for bb := 1 to b do begin
                  CMul(Sum, EMat[b,bb], Conv);
                  if not IsZero(Sum, SmallE) then begin
                    CAdd(Sum, Matrix[b,bb]);
                    {$IFOPT R+}
                    if not MathSup.SameValue(EMat[b,bb],EMat[bb,b]) then
                      WriteLn(Log, 'Transition Matrix not symmetric:', EMat[b,bb].Re, ' ', EMat[bb,b].Re);
                    {$ENDIF}
                    if BlockMatrix then
                      Connections[b].OrIn(Connections[bb]);
                    Matrix.AutoSetValue(b,bb,Sum);
                    if b <> bb then begin
                      Sum.im := - Sum.im;
                      Matrix[bb,b] := Sum;
                    end;
                  end;
                end;
            EMat.Free;
          end;
        end;
      end;
    end;
    if BlockMatrix and (CountBlocks <> 1) then begin
      TStart(BlockMatrixTime);
      FactorMatrix(Matrices, Connections);
      TEnd(BlockMatrixTime);
    end;
    OK := True;
  finally
    { Don't use except to preserve original call stack }
    if not OK then
      for b := 0 to High(Matrices) do
        FreeAndNil(Matrices[b]);
  end;
end;

function TEigenvalues.AddPerturbation(Matrix: TComplexMatrix; SinglePerturbation: TPerturbation; MatElemType: TMatElemType; var Connections: array of TBigSet; SmallE: Double):Integer;
var
  b, bb, bk, nSet: Integer;
  Current: TState;
  Sum: Complex;
  Perturbation: TPerturbation;
begin
  Result := 0;
  if SinglePerturbation = nil then
    case Manifold.nPerturbations of
      0: Exit;
      1: SinglePerturbation := Manifold.Perturbation[0];
    end;
  TStart(PerturbationMatrixElementTime);
  for b := 1 to Size do begin
    Current := Basis[b].State;
    if Assigned(SinglePerturbation) then begin
      if not ((SinglePerturbation.Bra = Current) or (SinglePerturbation.Ket = Current)) then
        Continue;
    end else if (Current.PerturbationList[0] = nil) and (Current.PerturbationList[1] = nil) then
      Continue;
    for bb := 1 to b do begin
      Sum := z0;
      nSet := 0;
      if not HasFields or ((Basis[b].F = Basis[bb].F) and (Basis[b].Sym = Basis[bb].Sym)) then
        if Assigned(SinglePerturbation) then begin
          if not SinglePerturbation.Connects(Current, Basis[bb].State) then
            Continue;
          Sum := SinglePerturbation.MatrixElement(Basis[b], Basis[bb], MatElemType);
          nSet := 1;
        end else
          for bk := 0 to 1 do begin
            Perturbation := Current.PerturbationList[bk];
            while Assigned(Perturbation) do begin
              if ((bk = 0) or (Perturbation.Bra <> Perturbation.Ket)) { Avoid applying digonal elements twice }
                 and Perturbation.Connects(Current, Basis[bb].State) then begin
                CAdd(Sum, Perturbation.MatrixElement(Basis[b], Basis[bb], MatElemType));
                Inc(nSet);
              end;
              Perturbation := Perturbation.PerturbationList[bk];
            end;
          end;
      if (nSet > 0) and not MathSup.IsZero(Sum, SmallE) then begin
        Inc(Result);
        if Length(Connections) > 0 then
          Connections[b].OrIn(Connections[bb]);
        CAdd(Sum, Matrix[b,bb]);
        Matrix.AutoSetValue(b,bb,Sum);
        if b <> bb then begin
          Sum.im := -Sum.im;
          Matrix[bb,b] := Sum;
        end;
      end;
    end;
  end;
  TEnd(PerturbationMatrixElementTime);
end;

procedure TEigenvalues.ShowBasisMatrix(M: TComplexMatrix; const Title: string; Offset: Integer; MaxElements: Integer);
begin
  ShowTransitionMatrix(Self, M, Title, Offset, True, MaxElements);
end;

procedure TEigenvalues.ShowTransitionMatrix(Ev: TEigenValues; M: TComplexMatrix; const Title: string; Offset: Integer; ShowBasis: Boolean; MaxElements: Integer);
var
  b, bb, sw: Integer;
  LWidths: array of Integer;
  s: string;
  IsComplex: Boolean;
  procedure ShowState(Ev: TEigenValues);
  var
    s: string;
  begin
    if Ev.HasFields then
      s := ', ' + Ev.FieldLabel
    else
      s := '';
    Write(Log, Ev.Basis[1].State.Manifold.Name,', J = ', Ev.Manifold.Mixture.QuantumNumberToStr(Ev.J), ', Symmetry = ', Manifold.Molecule.BothSymNames(Ev.J, Ev.Sym), s);
  end;
  function SLabel(i: Integer; Ev: TEigenValues): string;
  begin
    if ShowBasis then
      Result := Ev.Basis[i+Offset].BasisLabel
    else
      Result := Ev.DominantBasis[i+Offset].StateLabel(Ev.Manifold.Mixture.QuantumNumberFormat);
  end;
begin
  { Print Hamiltonian or other Matrix, including basis state labels }
  if Size = 0 then Exit;
  Write(Log, Title,' ');
  if Ev <> Self then Write(Log, '<');
  ShowState(Self);
  if Ev <> Self then begin
    Write(Log, '| |');
    ShowState(Ev);
    Write(Log, '>');
  end;
  if M = nil then begin
    WriteLn(Log, ' (Empty)');
    Exit;
  end;
  IsComplex := IsComplexMatrix(M);
  if MaxElements = 0 then
    MaxElements := Max(M.ColCount, M.RowCount);
  if MaxElements < Max(M.ColCount, M.RowCount) then
    Write(Log, ' (first ',MaxElements,' rows and columns only)');
  WriteLn(Log);
  sw := 4; { Work out width of Row labels }
  for b := 1 to M.RowCount do
    sw := Max(sw, Length(SLabel(b,Self))+2);
  Write(Log, ' ':sw);
  SetLength(LWidths, Ev.Size+1); { Index 0 is not used }
  for b := 1 to Min(M.ColCount,MaxElements) do begin
    s := '|'+SLabel(b,Ev)+'>';
    LWidths[b] := Length(s);
    Write(Log, s);
    if IsComplex then
      Write(Log, ' i', ' ':Length(s)-2);
  end;
  WriteLn(Log);
  for b := 1 to Min(M.RowCount,MaxElements) do begin
    Write(Log, '<'+SLabel(b,Self)+'|':sw);
    for bb := 1 to Min(M.ColCount,MaxElements) do begin
      Write(Log, DtosMA(M.Re[b,bb],LWidths[bb],-1));
      if M.Mode <> cmReal then
        Write(Log, DtosMA(M.Im[b,bb],LWidths[bb],-1));
    end;
    WriteLn(Log);
  end;
  WriteLn(Log);
end;

procedure TEigenvalues.Multiply(M: TComplexMatrix);
{ Multiply matrix M by eigenvalues
  ( Actually Pre multiply by transpose of eigenvector matrix )
    so requires match of Rows of M with size of Basis }
var
  i, ii, iii: Integer;
  N: Integer;
  A: PDouble;
  No, Yes: AnsiChar;
  alpha, beta: Double;
  zAlpha, zBeta, Sum: Complex;
  W: TComplexArray;
begin
  if M.RowCount <> Size then
    raise Exception.Create('TEigenvalues.Multiply: Size mismatch');
  N := Size;
  {$DEFINE BLAS}
  {$IFDEF BLAS}
    if (C.Mode = cmReal) and (M.Mode = cmReal) then begin
      M.Size := M.Size * 2; { Allocate workspace }
      A := PDouble(PAnsiChar(M.FirstItem) + (M.Size div 2)*SizeOf(Double));
      System.Move(M.FirstItem^, A^, (M.Size div 2)*SizeOf(Double));
      No := 'N';
      Yes := 'C';
      alpha := 1;
      beta := 0;
      i := M.ColCount;
      dgemm_(Yes, No, N, i, N, alpha, C.FirstItem^, N, A^, N, beta, M.FirstItem^, N, 1, 1);
      M.Size := M.Size div 2;
    end else if C.Mode = cmReal then begin
      { Real(C) * Complex(M) - BLAS routine not available }
      { This ordering is significantly quicker, as transition matrices are typically
        banded and thus not very dense }
      SetLength(W, Size+1); { Index 0 is not used }
      for i := 1 to M.ColCount do begin
        for ii := 1 to M.RowCount do begin
          W[ii] := M[ii,i];
          M[ii,i] := z0;
        end;
        for iii := 1 to M.RowCount do begin
          Sum := W[iii]; { = original M[iii,i] }
          if not IsZero(Sum) then
            for ii := 1 to M.RowCount do begin
              CMul(zAlpha, Sum, C[iii,ii]);
              CAdd(zAlpha, M[ii,i]);
              M[ii,i] := zAlpha;
            end;
        end;
      end;
    end else begin
      if (C.Mode <> cmNComplex) then
        MustBeSerial;
      C.Mode := cmNComplex;
      M.Mode := cmNComplex;
      M.Size := M.Size * 2; { Allocate workspace }
      A := PDouble(PAnsiChar(M.FirstItem) + (M.Size div 2)*SizeOf(Double));
      System.Move(M.FirstItem^, A^, (M.Size div 2)*SizeOf(Double));
      No := 'N';
      Yes := 'C';
      zAlpha := z1;
      zBeta := z0;
      i := M.ColCount;
      zgemm_(Yes, No, N, i, N, zAlpha, PComplex(C.FirstItem)^, N, PComplex(A)^, N, zBeta, PComplex(M.FirstItem)^, N, 1, 1);
      M.Size := M.Size div 2;
    end;
  {$ELSE}
  //W: array of Double;
  //Sum: Double;
  //PM, PC: PDouble;
  {$DEFINE ALT}
  {$IFDEF ALT}
  { This ordering is significantly quicker, as transition matrices are typically
    banded and thus not very dense }
  SetLength(W, Size+1); { Index 0 is not used }
  for i := 1 to M.ColCount do begin
    for ii := 1 to M.RowCount do begin
      W[ii] := M[ii,i];
      M[ii,i] := 0;
    end;
    for iii := 1 to M.RowCount do begin
      Sum := W[iii]; { = original M[iii,i] }
      if Sum <> 0 then begin
        PM := M.Item[1,i];
        PC := C.Item[iii,1];
        for ii := 1 to M.RowCount do begin
          {$IFDEF SIMPLE}
          M[ii,i] := M[ii,i] + Sum*C[iii,ii];
          {$ELSE}
          PM^ := PM^ + Sum*PC^;
          Inc(PM);
          Inc(PC,N);
          {$ENDIF}
        end;
      end;
    end;
  end;
  {$ELSE}
  SetLength(W, Size+1); { Index 0 is not used }
  for i := 1 to M.ColCount do begin
    for ii := 1 to Size do begin
      Sum := 0;
      for iii := 1 to Size do
        Sum := Sum + C[iii,ii] * M[iii,i];
      W[ii] := Sum;
    end;
    for ii := 1 to Size do
      M[ii,i] := W[ii];
  end;
  {$ENDIF}
  {$ENDIF}
end;

procedure TEigenvalues.TransformTransitionMatrix(M: TComplexMatrix; OtherEv: TEigenvalues);
{ Transform Matrix M using eigenvectors - overwrites M }
begin
  TStart(MatrixMultiplyTime);
  Multiply(M);
  TEnd(MatrixMultiplyTime);
  TStart(PostMultiplyTime);
  M.PostMultiply(OtherEv.C);
  TEnd(PostMultiplyTime);
end;

function TEigenvalues.Qno(AIndex: Integer; QnoType: TQnoType): TQuantumNumber;
begin
  Result := DominantBasis[AIndex].Qno(QnoType);
end;

function TEigenvalues.TrackIndex(AIndex: Integer): Integer;
begin
  Result := DominantBasis[AIndex].TrackIndex;
end;

function TEigenvalues.JDeAdjustedSym(AIndex: Integer): Integer;
var
  DBasis: TBasisState;
begin
  DBasis := DominantBasis[AIndex];
  Result := Manifold.Molecule.JDeAdjustedSym(DBasis.F,DBasis.Sym);
end;

function TEigenvalues.TxName(AIndex: Integer; OtherEv: TEigenvalues; OtherIndex: Integer): string;
var
  B: TBasisState;
begin
  B := DominantBasis[AIndex];
  Result := B.TxName(B, OtherEv.DominantBasis[OtherIndex]);
end;

function TEigenvalues.Transform(M: TComplexMatrix): TColumnVector;
{ Return diagonal elements of M transformed by eigenvectors }
{ M is overwritten }
var
  i, ii: Integer;
  Sum: Double;
  Csum, T: Complex;
begin
  Multiply(M);
  Result := TColumnVector.Create(Size);
  try
    if M.Mode <> cmReal then begin
      if C.Mode = cmReal then
        raise NotYet; { if M is Complex, C should be also }
      for i := 1 to Size do begin
        Csum := z0;
        for ii := 1 to Size do begin
          CMul(T, M[i,ii], C[ii,i]);
          CAdd(Csum,T);
        end;
        Result[i] := Sqrt(SqrMod(Csum));
      end;
    end else
      for i := 1 to Size do begin
        Sum := 0;
        for ii := 1 to Size do
          Sum := Sum + M.Re[i, ii] * C.Re[ii, i];
        Result[i] := Sum;
      end;
  except
    Result.Free;
    raise;
  end;
end;

procedure TEigenvalues.SortFrom(Start: Integer; Emax: Double);
var
  V: TDarray;
  i: Integer;
begin
  Fixup;
  if (Size-Start <= 1) and (Emax >= 1e99) then
    Exit { 1 element must be sorted }
  else begin
    V := DiagonalMatrixElements(meHrot, Start, True);
    SortBasisOn(V, False, Start);
    if Emax < 1e99 then
      for i := 0 to High(V) do
        if V[i] > Emax then begin
          { Delete unwanted basis states }
          while ComponentCount > Start+i do
            Components[i+Start].Free;
          Fixup;
          Exit;
        end;
  end;
end;

function TEigenvalues.StatisticalWeight(Index: Integer): Integer;
begin
  { This allows a basis with mixed nuclear statistical weights;
    only normally required where HasFields }
  Result := DominantBasis[Index].StatisticalWeight;
end;

function TEigenvalues.SpinSpecies(Index: Integer): Integer;
begin
  { This allows a basis with mixed nuclear statistical weights;
    only normally required where HasFields }
  Result := DominantBasis[Index].StatWtGroup;
end;

type
  TBasisShellSorter = class(TShellSorter)
    ByState: Boolean;
    Eigenvalues: TEigenvalues;
    Base: Integer;
    V: TDArray;
    function InOrder(I,J:Integer):Boolean; override;
    procedure Swap(I,J: Integer); override;
  end;

function TBasisShellSorter.InOrder(I,J:Integer):Boolean;
var
  Diff: Integer;
begin
  if ByState then begin
    Diff := Eigenvalues.Basis[I+Base+1].SubBasisIndex - Eigenvalues.Basis[J+Base+1].SubBasisIndex;
    if Diff <> 0 then begin
      Result := Diff <= 0;
      Exit;
    end;
  end;
  Result := V[I] <= V[J];
end;

procedure TBasisShellSorter.Swap(I,J: Integer);
begin
  SwapValues(V[I], V[J], SizeOf(Double));
  SwapValues(Eigenvalues.Basis[I+Base+1], Eigenvalues.Basis[J+Base+1], SizeOf(TBasisState));
end;

procedure TEigenvalues.SortBasisOn(V: TDarray; ByState: Boolean; Base: Integer);
{ Sort the basis states starting from Base;
  Given V[b] for each basis state b, shuffle so that V[b] is in ascending order }
var
  Sorter: TBasisShellSorter;
  i: Integer;
begin
  Sorter := TBasisShellSorter.Create;
  Sorter.ByState := ByState;
  Sorter.Eigenvalues := Self;
  Sorter.Base := Base;
  Sorter.V := V;
  Sorter.Sort(Size-Base);
  for i := Base to Size-1 do
    Basis[i+1].ComponentIndex := i;
  Sorter.Free;
end;

procedure TEigenvalues.DoLimitedEigenSearch(Co: TComplexMatrix; Offset: Integer);
  { Must be called with Offset = 0 first }
var
  SubBasisIndex, nBasisStates, nContr: Integer;
  Ind: array of Integer;
  Contr: array of Double;
  IContr: array of Integer;
  Sum: Double;
  i, ii, k, N: Integer;
begin
  { Leave individual basis states in original order,
    but make sure states are correct }
  if (LastSubBasisIndex <= 1) or (Size <= 1) then begin
    FreeAndNil(FDominantBasisIndex); { Leave as is }
    Exit;
  end;
  if Assigned(FDominantBasisIndex) and (Offset = 0) then
    Exit; { Probably never called }
  if not Assigned(FDominantBasisIndex) then
    FDominantBasisIndex := TIntColumnVector.Create(Size);
  N := Co.RowCount;
  { For each sub-basis:
    - find out ouw many basis states it contains
    - select exactly that many on the basis of largest coefficients }
  SetLength(Ind, N);   { Index into basis for this BasisStateIndex }
  SetLength(Contr, N);
  SetLength(IContr, N);
  for SubBasisIndex := 0 to LastSubBasisIndex-1 do begin
    { Ind is Index into basis for this sub-basis }
    { Logically: SetLength(Ind, N); }
    nBasisStates := 0;
    for k := 1 to N do
      if Basis[k+Offset].SubBasisIndex = SubBasisIndex then begin
        Ind[nBasisStates] := k;
        Inc(nBasisStates);
      end else if (Basis[k+Offset].SubBasisIndex < 0) or (Basis[k+Offset].SubBasisIndex >= LastSubBasisIndex) then
        raise Exception.Create(Basis[k].BasisLabel); { Should never happen }
    if nBasisStates = 0 then
      Continue;
    { Ind is Index into basis for this sub-basis }
    { Logically, but causes unecessary calls to realloc:
      SetLength(Ind, nBasisStates); SetLength(Contr, nBasisStates); SetLength(IContr, nBasisStates); }
    nContr := 0; { Contr and IContr keep track of the most siginificant contributions }
    for k := 1 to N do
      if FDominantBasisIndex[k+Offset] = 0 then begin
        { Work out how much eigenvalue k has from this particular sub-basis }
        Sum := 0;
        for i := 0 to nBasisStates-1 do
          Sum := Sum + Co.SqrMod[Ind[i],k];
        { Insert into sorted list of the nBasisStates largest contributions }
        { Find insert position in list }
        i := 0;
        while (i < nContr) and (Sum > Contr[i]) do
          Inc(i);
        if nContr < nBasisStates then begin
          { Not enough numbers in list yet, so grow list
            - Some values to shift up }
          for ii := nContr downto i+1 do begin
            Contr[ii] := Contr[ii-1];
            IContr[ii] := IContr[ii-1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
          Inc(nContr);
        end else if i > 0 then begin
          Dec(i);
          { Replace a current entry, shifting lower values down }
          for ii := 0 to i-1 do begin
            Contr[ii] := Contr[ii+1];
            IContr[ii] := IContr[ii+1];
          end;
          Contr[i] := Sum;
          IContr[i] := k;
        end;
      end;
    if nContr <> nBasisStates then
      raise Exception.CreateFmt('DoLimitedEigenSearch %d %d', [nContr, nBasisStates]);
    { TODO: Alternate algorithm to handle case with coefficients^2 for a given
      sub basis arranged as below. Sub basis has two states, |1> and |2>, and there
      are 6 eigenvalues with significant contributions from the pair.
        |A>  0.48|1>
        |B>  0.36|1>
        |C>  0.21|2>
        |D>  0.28|2>
        |E>  0.21|2>
        |F>  0.28|2>
     Just picking the two eigenvalues with the largest contribution from the given sub basis
     picks A and B, but clearly only one of these should be picked - here B will be assigned 2 despite having no contribution
     from this component of the sub basis. }
    { TODO: From  custpop_anti.pgo - would benefit from assigning sub-basis with > 50% contribution first }
    //   8	{1} v4=1a v4=1a 10  3 +1 E' 2696.49	 .73752	{2} v4=1a v2=2s 10   2	  -.67432	{1} v4=1a v4=1a 10   3 1	.03675	{0} v4=1a v4=1a 10   1 -1
    //  10	{2} v4=1a v2=2s 10  2 E' 	  2704.68	-.63312	{0} v4=1a v4=1a 10   1 -1	 .55529	{1} v4=1a v4=1a 10   3 1	.53927	{2} v4=1a v2=2s 10   2
    //  11	{0} v4=1a v4=1a 10  1 -1 E' 2743.51	 .77317	{0} v4=1a v4=1a 10   1 -1	 .48676	{1} v4=1a v4=1a 10   3 1	.40652	{2} v4=1a v2=2s 10   2
    Sort(IContr,nBasisStates);
    for i := 0 to nBasisStates-1 do
      FDominantBasisIndex[IContr[i]+Offset] := Ind[i]+Offset;
  end;
end;

procedure TEigenvalues.DoEigenSearch(Co: TComplexMatrix; Offset: Integer);
  { Must be called with Offset = 0 first }
var
  i, ii, iimax, N: Integer;
  cmax, v: Double;
  Taken: array of Boolean;
begin
  if Assigned(FDominantBasisIndex) and (Offset = 0) then
    Exit; { Probably never called }
  if not Assigned(FDominantBasisIndex) then
    FDominantBasisIndex := TIntColumnVector.Create(Size);
  N := Co.RowCount; { Size of this block }
  SetLength(Taken,N+1);
  FillChar(Taken[0], SizeOf(Boolean)*(N+1),0);
  { Now scan eigenvalues for largest coefficient of each basis state }
  { Any coefficients > 1/Sqrt(2) are definite assignments }
  { Note that the assignment will be unstable for coefficients that are exactly 1/sqrt(2);
    loop below has bias towards assigning in basis order, but doesn't
    in the case that the coefficents**2 are just less than 0.5 }
  for i := 1 to N do
    for ii := 1 to N do
      if not Taken[ii] then begin
        v := Co.SqrMod[ii,i];
        if v >= 0.5 then begin
          FDominantBasisIndex[i+Offset] := ii + Offset;
          Taken[ii] := True;
          Break;
        end;
      end;
  { For any states not assigned above }
  for i := 1 to N do
    if FDominantBasisIndex[i+Offset] = 0 then begin
      cmax := 0;
      iimax := 0;
      for ii := 1 to N do
        { Find largest remaining coefficient }
        if not Taken[ii] then begin
          v := Co.SqrMod[ii,i];
          if v > cmax then begin
            iimax := ii;
            cmax := v;
          end;
        end;
      if iimax = 0 then begin
        { Horrible problem - all C zero or Nan; just take first }
        for ii := 1 to N do
          if not Taken[ii] then begin
            iimax := ii;
            Break;
          end;
      end;
      FDominantBasisIndex[i+Offset] := iimax + Offset;
      Taken[iimax] := True;
    end;
end;

function TEigenvalues.GetDominantBasisState(AIndex: Integer): TBasisState;
begin
  if Assigned(FDominantBasisIndex) then
    AIndex := FDominantBasisIndex[AIndex];
  Result := Basis[AIndex];
end;

function TEigenvalues.GetEnergy(AIndex: Integer): Double;
begin
  Result := E[AIndex];
end;

function TEigenvalues.HasCorrectAttributes(Attributes: Pointer): Boolean;
var
  MatElemTypes: TMatElemTypes;
begin
  Result := Size = 0;
  if not Result then begin
    if Size < 0 then
      Exit; { Basis not set up }
    MatElemTypes := TMatElemTypes({$IFDEF FPC}Cardinal{$ELSE}Byte{$ENDIF}(Attributes));
    Result := Assigned(E);
    if Result and Manifold.HasWidth and (meWidth in MatElemTypes) then
      Result := Assigned(Widths);
    if Result and (MatElemTypes <> []) then
      Result := Assigned(C);
    if Result and (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes) then
      Result := FDerivativesCalculated;
  end;
end;

procedure TEigenvalues.ShowEigenvalues;
var
  b, bb, sw: Integer;
  s: string;
  BWidths: array of Integer;
begin
  if HasFields then
    s := ', ' + FieldLabel
  else
    s := '';
  WriteLn(Log, 'Eigenvalues for ' + Manifold.Name + ', J = ', Manifold.Mixture.QuantumNumberToStr(J), ', Symmetry = ', Manifold.Molecule.BothSymNames(J,Sym), s);
  sw := 1; { Maximum width of state label }
  for b := 1 to Size do
    sw := Max(sw, Length(DominantBasis[b].StateLabel(Manifold.Mixture.QuantumNumberFormat)));
  Write(Log, ' ':18+sw);
  if Assigned(Widths) then Write(Log, ' ':10);
  SetLength(BWidths, Size+1);
  for b := 1 to Size do begin
    s := Basis[b].BasisLabel;
    BWidths[b] := Max(10, Length(s)+2);
    Write(Log, '|'+s+'>':10);
    if C.Mode <> cmReal then
      Write(Log, '  i', ' ':7);
  end;
  WriteLn(Log);
  for b := 1 to Size do begin
    Write(Log, b:3, ' ', DominantBasis[b].StateLabel(Manifold.Mixture.QuantumNumberFormat):sw+1, E[b]:13:3);
    if Assigned(Widths) then Write(Log, Widths[b]:10:5);
    for bb := 1 to Size do
      if C.Mode <> cmReal then begin
        Write(Log, C[bb,b].re:BWidths[bb]:6);
        Write(Log, C[bb,b].im:10:6);
      end else
        Write(Log, C.Re[bb,b]:BWidths[bb]:6);
    WriteLn(Log);
  end;
  WriteLn(Log);
end;

procedure TEigenvalues.CalculateDerivatives(MatElemTypes: TMatElemTypes);
var
  nDerivatives: Integer;
  Matrices: TComplexMatrices;
  procedure HandleParameters(MNode: TMNode);
  var
    I: Integer;
  begin
    for I := 1 to MNode.nParameters do
      if ((meAllDerivatives in MatElemTypes) and MNode.DerivativeRequired(I))
         or ((meFloatedDerivatives in MatElemTypes) and MNode.Floating[I]) then
        try
          if nDerivatives >= High(Derivatives) then
            SetLength(Derivatives, nDerivatives + MNode.nParameters);
          MNode.ModifyParameter(I,1,False, False);
          Manifold.CheckProperties;
          MatrixElements(Matrices, False, meHrot, not NoPerturbations);
          Derivatives[nDerivatives].Values := Transform(Matrices[0]);
          Derivatives[nDerivatives].Index := I;
          Derivatives[nDerivatives].MNode := MNode;
          Inc(nDerivatives);
        finally
          MNode.ModifyParameter(I,0,False, False);
          Matrices[0].Free;
        end;
    for I := 0 to MNode.Count - 1 do
      HandleParameters(MNode.Components[I] as TMNode);
  end;
begin
  if FDerivativesCalculated then
    Exit;
  nDerivatives := 0;
  Manifold.StartDerivatives;
  try
    HandleParameters(Manifold);
    SetLength(Derivatives,nDerivatives);
    FDerivativesCalculated := True;
  finally
    Manifold.StopDerivatives;
  end;
end;

procedure TEigenvalues.DoMake(RequiredAttributes: Pointer);
var
  CalculateVectors, CalculateWidths: Boolean;
  MatElemTypes: TMatElemTypes;
  Matrices: TComplexMatrices;
  i, Base: Integer;
begin
  if Size < 0 then
    MakeBasis(Manifold.Mixture.BasisOrder);
  if Size = 0 then
    Exit;
  MatElemTypes := TMatElemTypes({$IFDEF FPC}Cardinal{$ELSE}Byte{$ENDIF}(RequiredAttributes));
  CalculateWidths := Manifold.HasWidth and (meWidth in MatElemTypes) and not Assigned(Widths);
  CalculateVectors := ((meHrot in MatElemTypes) or CalculateWidths or (meAllDerivatives in MatElemTypes) or (meFloatedDerivatives in MatElemTypes)) and not Assigned(C);
  TStart(EigenvaluesTime);
  if not Assigned(E) or CalculateVectors then begin
    if Assigned(E) then
      WriteLn(Log,'**********', ThreadName); { TODO: Remove }
    try
      TStart(MakeHamiltonianTime);
      MakeHamiltonian(Matrices, Manifold.Molecule.BlockMatrix);
      TEnd(MakeHamiltonianTime);
      if PrintLevel >= plHamiltonian then begin
        WriteLn(Log);
        if Length(Matrices) = 1 then
          ShowBasisMatrix(Matrices[0], 'Hamiltonian Matrix for ', 0)
        else begin
          Base := 0;
          for i := 0 to High(Matrices) do begin
            ShowBasisMatrix(Matrices[i], 'Block ' + IntToStr(i+1) + ' of Hamiltonian Matrix for ', Base);
            Inc(Base,Matrices[i].RowCount);
          end;
        end;
      end;
      Diagonalize(Matrices,
                  (meHrot in MatElemTypes) or CalculateWidths or (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes)
                  or Manifold.EigenSearch or (PrintLevel >= plEigenVectors),
                  Manifold.EigenSearch, Manifold.LimitSearch);
    finally
      for i := 0 to High(Matrices) do
        Matrices[i].Free;
    end;
    {$IFDEF EXPLICITTEST}
    DoExplicitTest(Result);
    {$ENDIF}
  end;
  if (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes) then
    CalculateDerivatives(MatElemTypes);
  if (MatElemTypes <> []) or (PrintLevel >= plEigenVectors) then begin
    if CalculateWidths then
      MakeWidths(PrintLevel);
    if PrintLevel >= plEigenVectors then
      ShowEigenvalues;
  end;
  TEnd(EigenvaluesTime);
end;

function TEigenvalues.DominantLabel(AIndex: Integer; QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := DominantBasis[AIndex].StateLabel(QuantumNumberFormat);
end;

function TEigenvalues.StateLabelKey(AIndex: Integer): string;
begin
  Result := DominantBasis[AIndex].StateLabelKey;
end;

function TEigenvalues.Make(Wait: Boolean; MatElemTypes: TMatElemTypes):Boolean;
begin
  if (meFloatedDerivatives in MatElemTypes) or (meAllDerivatives in MatElemTypes) then
    Wait := True; { TODO: parallelize; currently fails as works by changing parameters }
  Result := Make(Wait, Pointer({$IFDEF FPC}PtrInt(Integer{$ELSE}Byte({$ENDIF}(MatElemTypes))));
end;

procedure TEigenvalues.MakeHamiltonian(out Matrices: TComplexMatrices; BlockMatrix: Boolean);
begin
  MatrixElements(Matrices, BlockMatrix, meHrot, not NoPerturbations);
end;

procedure TEigenvalues.MakeWidths(APrintLevel: TPrintLevel);
var
  ComplexMatrices: TComplexMatrices;
begin
  { TODO -cField: How should this handle complex eigenvectors? }
  MatrixElements(ComplexMatrices, False, meWidth);
  try
    if APrintLevel >= plMatrices then
      ShowBasisMatrix(ComplexMatrices[0], 'Width Matrix for ', 0);
    FWidths := Transform(ComplexMatrices[0]);
  finally
    ComplexMatrices[0].Free;
  end;
end;

procedure TEigenvalues.Diagonalize(var Matrices: TComplexMatrices; EigenvectorsWanted, EigenSearch, LimitSearch: Boolean);
var
  TempE, TE: TColumnVector;
  i, k, kk, Base, Block, BlockSize, BlockNo: Integer;
  BlockStarts, BlockP, BlockEnds, Ind, BlockSource: array of Integer;
  Emin: Double;
  TI: TIntColumnVector;
begin
  { Do do not overwrite E if it already exists }
  if not Assigned(E) then begin
    FE := TColumnVector.Create(Size);
    TempE := E;
  end else
    TempE := TColumnVector.Create(Size);
  try
    if Length(Matrices) = 1 then begin
      FC := Matrices[0];
      Matrices[0] := nil;
      CEigen(TempE, C, EigenvectorsWanted);
      if EigenSearch then begin
        TStart(EigenSearchTime);
        if LimitSearch then
          DoLimitedEigenSearch(C)
        else
          DoEigenSearch(C);
        TEnd(EigenSearchTime);
      end;
    end else begin
      Base := 0;
      SetLength(BlockStarts, Length(Matrices));
      SetLength(BlockEnds, Length(Matrices));
      for i := 0 to High(Matrices) do begin
        BlockStarts[i] := Base+1;
        BlockSize := Matrices[i].RowCount;
        BlockEnds[i] := Base+BlockSize+1;
        CEigen(TempE, Matrices[i], EigenvectorsWanted, Base);
        if EigenSearch then begin
          TStart(EigenSearchTime);
          if LimitSearch then
            DoLimitedEigenSearch(Matrices[i],Base)
          else
            DoEigenSearch(Matrices[i],Base);
          TEnd(EigenSearchTime);
        end;
        Inc(Base, BlockSize);
      end;
      { Now merge sort eigenvalues, recording indices as we go }
      SetLength(Ind, Size);
      SetLength(BlockSource, Size);
      BlockP := Copy(BlockStarts);
      i := 0;
      while i < Size do begin
        BlockNo := -1;
        { Find next lowest energy by searching individual blocks }
        for Block := 0 to High(Matrices) do
          if (BlockP[Block] < BlockEnds[Block])
             and ((BlockNo < 0) or (TempE[BlockP[Block]] < Emin)) then begin
            BlockNo := Block;
            Emin := TempE[BlockP[Block]];
          end;
        Ind[i] := BlockP[BlockNo];
        BlockSource[i] := BlockNo;
        Inc(BlockP[BlockNo]);
        Inc(i);
      end;
      { Permute energies }
      TE := TColumnVector.Create(Size);
      for i := 1 to Size do
        TE[i] := TempE[Ind[i-1]];
      TempE.AssignAndNil(TVector(TE));
      if Assigned(FDominantBasisIndex) then begin
        { Permute Basis indices }
        TI := TIntColumnVector.Create(Size);
        for i := 1 to Size do
          TI[i] := FDominantBasisIndex[Ind[i-1]];
        FDominantBasisIndex.AssignAndNil(TGHandle(TI));
      end else begin
        { No permutation of individual matrices }
        FDominantBasisIndex := TIntColumnVector.Create(Size);
        for i := 1 to Size do
          FDominantBasisIndex[i] := Ind[i-1];
      end;
      if EigenvectorsWanted then begin
        { Copy eigenvectors into position }
        FC := TComplexMatrix.Create(Size,Size,Matrices[0].Mode);
        FC.Zero;
        for i := 1 to Size do begin
          Block := BlockSource[i-1];
          Base := BlockStarts[Block]-1;
          k := Ind[i-1]-Base;
          for kk := 1 to Matrices[Block].RowCount do
            C[kk+Base,i] :=  Matrices[Block][kk,k];
        end;
      end;
    end;
  finally
    for i := 0 to High(Matrices) do
      FreeAndNil(Matrices[i]);
    if E <> TempE then
      TempE.Free;
  end;
end;

{ TBasisState }

function TBasisState.CachedStateLabel(QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := '';
  if QuantumNumberFormat = qftJ then begin
    { Only cache most common label style to avoid synchronization problems }
    if (FCachedStateLabel = '') and (InterLockedIncrement(FCacheLock) = 1) then
      FCachedStateLabel := StateLabel(QuantumNumberFormat);
    Result := FCachedStateLabel;
  end;
  if Result = '' then
    Result := StateLabel(QuantumNumberFormat);
end;

procedure TBasisState.CopyFrom(OtherState: TBasisState; Sign: Integer);
begin
  State := OtherState.State;
  SubBasisIndex := OtherState.SubBasisIndex;
  HyperfineState := OtherState.HyperfineState;
  F := OtherState.F;
  J := OtherState.J;
  N := OtherState.N;
  S := OtherState.S;
  M := OtherState.M;
  Sym := OtherState.Sym;
  Phase := OtherState.Phase;
  iPower := OtherState.iPower;
  StatisticalWeight := OtherState.StatisticalWeight;
  StatWtGroup := OtherState.StatWtGroup;
end;

constructor TBasisState.Create(AOwner: TComponent; AF, AM: TQuantumNumber; ASym: Integer);
begin
  inherited Create(AOwner);
  F := AF;
  Sym := ASym;
  M := AM;
  if Assigned(AOwner) and (AOwner is TEigenvalues) then
    with TEigenvalues(AOwner) do
      FSize := ComponentCount;
end;

function TBasisState.BasisLabelKey: string;
begin
  Result := 'Name  J +-';
  if State.Mixture.ShowSubBasis then
    Result := '{Sub} '+Result;
end;

function TBasisState.BasisLabel: string;
begin
  if State.Mixture.ShowSubBasis then
    Result := '{'+IntToStr(SubBasisIndex)+'} '
  else
    Result := '';
  if State.Molecule.nManifolds > 1 then
    Result := Result + State.Manifold.Name + ' ';
  Result := Result + State.Name
            + Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(J)]);
  if Phase > 0 then
    Result := Result + ' +'
  else if Phase < 0 then
    Result := Result + ' -'
  else
    Result := Result + '  ';
end;

function TBasisState.StateLabel(QuantumNumberFormat: TQuantumNumberFormat): string;
begin
  if State.Manifold.Molecule.nManifolds > 1 then
    Result := State.Manifold.Name + ' '
  else
    Result := '';
  Result := Result + State.Name;
  if State.Mixture.ShowSubBasis then
    Result := '{'+IntToStr(SubBasisIndex)+'} ' + Result;
end;

function TBasisState.StateLabelKey: string;
begin
  Result := 'Name';
  if State.Mixture.ShowSubBasis then
    Result := '{Sub} '+Result;
end;

function TBasisState.SymbolicVariableValues: string;
begin
  Result := '';
end;

function TBasisState.HyperfineStateLabel(WithBlanks: Boolean): string;
var
  FF, Fn: TQuantumNumber;
  i: Integer;
  procedure AddQno(Q: TQuantumNumber);
  begin
    if WithBlanks then
      Result := Format(' %2s', [State.Manifold.Mixture.QuantumNumberToStr(Q)]) + Result
    else begin
      if Length(Result) > 0 then
        Result := ',' + Result;
      Result := State.Manifold.Mixture.QuantumNumberToStr(Q) + Result;
    end;
  end;
begin
  Result := '';
  with Owner as TEigenvalues do
    if HasFields then
      AddQno(M);
  if Length(HyperfineState) > 0 then begin
    FF := F;
    for i := High(HyperfineState) downto 0 do begin
      if State.Nucleus(i).AsNext then
        Fn := HyperfineState[i]
      else begin
        Fn := FF;
        Dec(FF,HyperfineState[i]);
      end;
      AddQno(Fn);
    end;
  end;
end;

procedure TBasisState.SetHyperfineVariables;
var
  FF: TQuantumNumber;
  i, nNuclei: Integer;
begin
  nNuclei := State.Molecule.nNuclei;
  if nNuclei > 0 then begin
    FF := F;
    for i := nNuclei-1 downto 0 do
      if State.Nucleus(i).AsNext then
        { TODO: Fn := HyperfineState[i] }
      else begin
        if i = nNuclei-1 then
          eFn(0).Value := FF / 2
        else
          eFn(i+1).Value := FF/2;
        if i <= High(HyperfineState) then
          Dec(FF,HyperfineState[i]);
      end;
  end;
end;

function TBasisState.HyperfineVariableValues: string;
var
  i, nNuclei: Integer;
begin
  Result := '';
  nNuclei := State.Molecule.nNuclei;
  if nNuclei > 0 then
    for i := 0 to nNuclei-1 do
      Result := Result + ', ' + eFn(i).Text(True) + '=' + FloatToStr(eFn(i).Value);
end;

function TBasisState.HyperfineStateLabelKey(WithBlanks: Boolean; ExtraChar: Char): string;
var
  i: Integer;
  procedure AddQno(const Q: string);
  begin
    if not WithBlanks and (Result <> '') then
      Result := ',' + Result;
    if ExtraChar <> ' ' then
      Result := ExtraChar + Result;
    Result := Q + Result;
    if WithBlanks then
      Result := ' ' + Result;
 end;
begin
  Result := '';
  with Owner as TEigenvalues do
    if HasFields then
      AddQno('M');
  if Length(HyperfineState) > 0 then begin
    AddQno('F');
    for i := Length(HyperfineState)-2 downto 0 do begin
      if State.Nucleus(i).AsNext then
        AddQno(Format('I%d%d', [i+1,i+2]))
      else
        AddQno(Format('F%d', [i+1]));
     end;
  end;
end;

function TBasisState.HyperfineTxNameKey(WithBlanks: Boolean): string;
begin
  if (Length(HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then begin
    Result := HyperfineStateLabelKey(WithBlanks, '''');
    if WithBlanks then
      Result := Result + ' '
    else
      Result := Result + ',';
    Result := Result + HyperfineStateLabelKey(WithBlanks,'"')
  end else
    Result := '';
end;

function TBasisState.NextBasis(var LoopState: TBasisState; TransitionMoment: TTransitionMoment; Sign: Integer):Boolean;
begin
  raise NotYet;
end;

function TBasisState.TxName(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.TxNameKey(Upper, Lower: TBasisState): string;
begin
  Result := '';
end;

function TBasisState.TrackIndex: Integer;
begin
  Result := 0; { Default is all states in same class }
end;

function TBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
var
  i, StopAt: Integer;
begin
  case QnoType of
    qtJ: Result := J;
    qtN: Result := N;
    qtF: Result := F;
    qtM: Result := M;
    qtA, qtB: raise Exception.Create('Unimplemeted quantum number');
  else
    Result := F;
    StopAt := Ord(QnoType) - Ord(qtF);
    if (StopAt < Length(HyperfineState)) and State.Nucleus(StopAt).AsNext then
      Result := HyperfineState[StopAt]
    else
      for i := Length(HyperfineState)-2 downto StopAt do
        if not State.Nucleus(i).AsNext then
          Dec(Result,HyperfineState[i+1]);
  end;
end;

function TBasisState.MakeHyperfineSignature(Ket: TBasisState; nBase: Integer): TIntegerArray;
var
  i, nNuclei: Integer;
  FFBra, FFKet: TQuantumNumber;
begin
  nNuclei := State.Molecule.nNuclei;
  SetLength(Result, nBase+nNuclei);
  if Assigned(Owner) and (Owner as TEigenvalues).HasFields then
    raise NotYet;
  if nNuclei > 0 then begin
    FFBra := F;
    FFKet := Ket.F;
    for i := nNuclei-1 downto 0 do
      if State.Nucleus(i).AsNext then
        if i > High(HyperfineState) then
          Result[nBase+i] := 0
        else
          Result[nBase+i] := HyperfineState[i]*1000 + Ket.HyperfineState[i]
      else begin
        Result[nBase+i] := (FFBra - FFKet) div 2;
        if i <= High(HyperfineState) then begin
          Dec(FFBra,HyperfineState[i]);
          Dec(FFKet,Ket.HyperfineState[i]);
        end;
      end;
  end;
end;

{ TSpectrum }

procedure TSpectrum.Clear;
begin
  FnLines := 0;
  ITotal := 0;
end;

procedure TSpectrum.AddIn(LineList: TLineList);
var
  i: Integer;
begin
  for i := 0 to LineList.nLines - 1 do
    with LineList.Line[i] do
      AddLine(Frequency,Strength,Width,Colour,LineList.Units);
end;

procedure TSpectrum.AddIn(LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
var
  i, c: Integer;
begin
  { FirstLine and Increment allow the core loop to be parallelized }
  { Currently this is only implemented in TFitPlot, as it is only here the loop
    is normally particularly time consuming }
  { TODO: If many lines with state dependent widths, other other loops can also be slow; move to TSpectrumPlot }
  if FirstLine = 0 then
    PushProgress('Accumulating Spectrum', LineList.nLines div 100);
  try
    if (Increment = 1) and IsNormalized(Simulation.IntensityUnits) then
      Simulation.Mixture.PreCalculatePartitionFunctions(Simulation, Problem);
    c := 0;
    TStart(AddInLinesTime);
    i := FirstLine;
    while i < LineList.nLines do begin
      AddLine(LineList.Line[i], Simulation, LineList.Units, Problem);
      Dec(c);
      if (c < 0) and (FirstLine = 0) then begin
        SetProgress(i div 100);
        c := 100;
      end else if Aborted then
        Break;
      Inc(i,Increment);
    end;
    TEnd(AddInLinesTime);
  finally
    if FirstLine = 0 then
      PopProgress;
  end;
end;

constructor TSpectrum.Create(AUnits: TResonanceUnits; AIsDistribution: Boolean);
begin
  inherited Create;
  Colour := clNone;
  Clear;
  SetShower(Log, plNone);
  Units := AUnits;
  FIsDistribution := AIsDistribution;
end;

procedure TSpectrum.AddLine(Frequency, Strength, Width: Double; AColour: TColor; AUnits: TResonanceUnits);
var
  Line: TLine;
  var Problem: Exception;
begin
  Line.Frequency := Frequency;
  Line.Strength := Strength;
  Line.Width := Width;
  Line.EigenvaluesL := nil;
  Line.EigenvaluesU := nil;
  Line.Colour := AColour;
//  if AUnits <> Units then
//    ConvertLine(Line, AUnits, Units);
  Problem := nil;
  AddLine(Line, nil, AUnits, Problem);
  if Assigned(Problem) then
    raise Problem; { With current design, should never happen }
end;

procedure TSpectrum.SetShower(var F: Text; PrintLevel: TPrintLevel);
begin
  FHeadingPrinted := False;
  if @F = nil then
    SS.Setup(Log, PrintLevel) { Default output }
  else
    SS.Setup(F, PrintLevel);
end;

function EstErr(EvU, EvL: TEigenvalues; IndexU, IndexL: Integer):Double;
var
  Mixture: TMixture;
  VarCovar: TMatrix;
  Derivs: array of Double;
  i, j, nPar: Integer;
  v: Double;
begin
  Result := 0;
  if not Assigned(EvL) then
    Exit;
  Mixture := EvL.Manifold.Mixture;
  nPar := Mixture.Correlations.nEntries;
  { Make up variance-covariance matrix }
  VarCovar := TMatrix.Create(nPar, nPar);
  with Mixture.Correlations.FitDetails do
    for i := 0 to nPar - 1 do begin
      VarCovar[i+1,i+1] := Sqr(Parameters[i].StdDev);
      for j := 0 to i - 1 do begin
        VarCovar[i+1,j+1] := Mixture.Correlations.FitDetails.Correlation[i,j]*Parameters[i].StdDev*Parameters[j].StdDev;
        VarCovar[j+1,i+1] := VarCovar[i+1,j+1];
      end;
    end;
  { Now need dObs/dCalc, and Match }
  SetLength(Derivs, nPar);
  for i := 0 to nPar - 1 do begin
    v := 0;
    j := EvL.FindDerivatives(Mixture.Correlations.FitDetails.Parameters[i]);
    if j >= 0 then
      v := EvL.Derivatives[j].Values[IndexL];
    if Assigned(EvU) then begin
      v := -v;
      j := EvU.FindDerivatives(Mixture.Correlations.FitDetails.Parameters[i]);
      if j >= 0 then
        v := v + EvU.Derivatives[j].Values[IndexU];
    end;
    Derivs[i] := v;
  end;
  Result := 0;
  for i := 0 to nPar - 1 do
    for j := 0 to nPar - 1 do
      Result := Result + Derivs[i]*VarCovar[j+1,i+1]*Derivs[j];
  VarCovar.Free;
  Result := Sqrt(Result);
end;

procedure TSpectrum.ShowLine(var Line: TLine; Threshold: Double; LineFreqUnits, DisplayFreqUnits: TResonanceUnits; LinFormat: Boolean; var Problem: Exception);
var
  Intensity: Double;
  l: Integer;
  Ev1, Ev2: TEigenvalues;
  Mixture: TMixture;
  Molecule: TMolecule;
  Lab: string;
  CLine: TLine;
  procedure ShowQno(V: TQuantumNumber);
  begin
    with SS do
      if Separator = #0 then
        ShowInt(V)
      else
        ShowField(Mixture.QuantumNumberToStr(V));
  end;
  procedure ShowSym(J: TQuantumNumber; Sym: Integer);
  begin
    with SS do
      if Separator = #0 then
        ShowInt(Molecule.JAdjustedSym(J,Sym))
      else
        ShowField(Molecule.JAdjustedSymName(J,Sym));
  end;
  function SpaceQno(J: TQuantumNumber):string;
  var
    W: Integer;
  begin
    if Ev1.HasFields or (J >= 0) then
      Result := Mixture.QuantumNumberToStr(J)
    else
      Result := '-';
    if Mixture.QuantumNumberFormat = qftJ then
      W := 5
    else
      W := 3;
    if Length(Result) < W then
      Result := StringOfChar(' ', W-Length(Result)) + Result
    else if Result[1] <> ' ' then
      Result := ' ' + Result;
  end;
  procedure ShowDerivatives;
    procedure ShowDerivative(const Derivative: TDerivative; Value: Double);
    begin
      SS.ShowField(Derivative.MNode.FullParameterName(Derivative.Index));
      SS.ShowD(Value);
    end;
  var
    i: Integer;
  begin
    if not Mixture.ShowDerivatives or (Length(Ev1.Derivatives) <= 0) then
      Exit;
    if SS.Separator = ' ' then
      Write(SS.F^, ' : ');
    if IsDistribution then
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev1.Derivatives[i],Ev1.Derivatives[i].Values[Line.IndexL])
    else if Ev1.Manifold = Ev2.Manifold then
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev2.Derivatives[i], Ev2.Derivatives[i].Values[Line.IndexU]-Ev1.Derivatives[i].Values[Line.IndexL])
    else begin
      for i := 0 to High(Ev2.Derivatives) do
        ShowDerivative(Ev2.Derivatives[i], Ev2.Derivatives[i].Values[Line.IndexU]);
      for i := 0 to High(Ev1.Derivatives) do
        ShowDerivative(Ev1.Derivatives[i], -Ev1.Derivatives[i].Values[Line.IndexL]);
    end;
  end;
var
  SL, nL, Prec, i: Integer;
  JHeader, JorM: string;
  A, StdDev: Double;
begin
  Ev2 := Line.EigenvaluesU;
  Ev1 := Line.EigenvaluesL;
  if not Assigned(Ev1) or (not Assigned(Ev2) and not IsDistribution) then
    Exit;
  Mixture := Ev1.Manifold.Mixture;
  Molecule := Ev1.Manifold.Molecule;
  Intensity := Mixture.Simulations[0].LineIntensity(Line, LineFreqUnits, Problem);
  if not IsDistribution then
    A := Molecule.EinsteinA(Line, LineFreqUnits);
  ITotal := ITotal + Intensity;
  if Intensity < Threshold then
    Exit;
  Inc(nPrinted);
  Prec := Max(Mixture.Precision,3);
  CLine := Line;
  ConvertLine(CLine, LineFreqUnits, DisplayFreqUnits);
  l := Length(Mixture.LongestManifoldName);
  if SS.Separator = ' ' then
    nL := Length(IntToStr(Mixture.LargestIndex))+1;
  if not FHeadingPrinted then begin
    if Ev1.HasFields then
      JorM := 'M'
    else
      JorM := 'J';
    if SS.Separator = ' ' then begin
      Write(SS.F^, Copy('Molecule',1,Length(Mixture.LongestMoleculeName)):Length(Mixture.LongestMoleculeName), ' ');
      if Mixture.QuantumNumberFormat = qftJ then
        JHeader := '   '
      else
        JHeader := ' ';
      JHeader := JHeader+JorM;
      if IsDistribution then begin
        SL := Mixture.LongestSymLength;
        Write(SS.F^,'M':l, ' ', JHeader, ' ', 'S':SL, '#':nL, '    Position':10+Prec)
      end else begin
        SL := Max(Mixture.LongestSymLength,2);
        Write(SS.F^,Copy('Manifold',1,l):l, '''', JHeader, ''' ', 'S''':SL, '#''':nL,
                    Copy('Manifold',1,l):l, '"', JHeader, '" ', 'S"':SL, '#"':nL,
                    'Position':10+Prec);
      end;
      if Mixture.ShowEstUnc then
        Write(SS.F^, ' Est Unc ');
      if LinFormat then
        Write(SS.F^, ' Wt I      ')
      else if IsDistribution then
        Write(SS.F^, ' I       ','E':10+Prec,' S      ')
      else
        Write(SS.F^, ' I       ','Eupper':10+Prec,'Elower':10+Prec,' Spol     A      ');
      WriteLn(SS.F^, '  Width');
    end else
      with SS do begin
        if Separator <> ' ' then begin
          ShowField('Line list produced by '+VersionInfo);
          ShowInt(DecodePgopherVersion(VersionInfo));
          {$IFDEF CONSOLE}
          for i := 1 to Paramcount do
            ShowField(ParamStr(i));
          {$ENDIF}
          ShowEOL;
        end;
        ShowField('Molecule');
        if IsDistribution then begin
          ShowField('Manifold');
          ShowField(JorM);
          ShowField('Sym');
          ShowField('#');
        end else begin
          ShowField('Upper Manifold');
          ShowField(JorM+'''');
          ShowField('Sym''');
          ShowField('#''');
          ShowField('Lower Manifold');
          ShowField(JorM+'"');
          ShowField('Sym"');
          ShowField('#"');
        end;
        ShowField('Position');
        if Mixture.ShowEstUnc then
          ShowField('Est Unc');
        if LinFormat then
          ShowField('Wt');
        ShowField('Intensity');
        if not LinFormat then begin
          if IsDistribution then
            ShowField('E')
          else begin
            ShowField('Eupper');
            ShowField('Elower');
          end;
          ShowField('Spol');
          if not IsDistribution then
            ShowField('A');
        end;
        ShowField('Width');
        ShowField('Branch');
        ShowField('Label');
        ShowEOL;
      end;
  end;
  FHeadingPrinted := True;

  if Mixture.ShowEstUnc then
    if IsDistribution then
      StdDev := EstErr(nil, Ev1, -1, Line.IndexL)
    else
      StdDev := EstErr(Ev2, Ev1, Line.IndexU, Line.IndexL);

  if SS.Separator = ' ' then begin
    Write(SS.F^, Molecule.Name:Length(Mixture.LongestMoleculeName), ' ');
    SL := Mixture.LongestSymLength;
    if not IsDistribution then begin
      SL := Max(SL,2);
      Write(SS.F^,Ev2.Manifold.Name:l, SpaceQno(Ev2.J), ' ', Molecule.JAdjustedSymName(Ev2.J,Ev2.Sym):SL, AtLeast(Line.IndexU,nL), ' ');
    end;
    Write(SS.F^,Ev1.Manifold.Name:l, SpaceQno(Ev1.J), ' ', Molecule.JAdjustedSymName(Ev1.J,Ev1.Sym):SL, AtLeast(Line.IndexL,nL), ' ',
                AtLeast(CLine.Frequency,10+Prec,Prec));
    if Mixture.ShowEstUnc then
        Write(SS.F^, DtosMA(StdDev,9,-1));
    if LinFormat then
      Write(SS.F^, ' - ');
    Write(SS.F^, DtosMA(Intensity,9,-4));
    if not LinFormat then begin
      if not IsDistribution then
        Write(SS.F^, AtLeast(Ev2.Energy[Line.IndexU],10+Prec,Prec));
      Write(SS.F^, AtLeast(Ev1.Energy[Line.IndexL],10+Prec,Prec),
                   DtosMA(Line.Strength,9,-4));
      if not IsDistribution then
        Write(SS.F^, DtosMA(A,9,-4));
    end;
    Write(SS.F^, DtosMA(CLine.Width,9,-4), ' : ', MakeTxName(Line));
    ShowDerivatives;
    WriteLn(SS.F^);
  end else
    with SS do begin
      ShowField(Ev1.Manifold.Molecule.Name);
      if not IsDistribution then begin
        ShowField(Ev2.Manifold.Name);
        ShowQno(Ev2.J);
        ShowSym(Ev2.J,Ev2.Sym);
        ShowInt(Line.IndexU);
      end;
      ShowField(Ev1.Manifold.Name);
      ShowQno(Ev1.J);
      ShowSym(Ev1.J,Ev1.Sym);
      ShowInt(Line.IndexL);
      ShowD(CLine.Frequency);
      if Mixture.ShowEstUnc then
        ShowD(StdDev);
      if LinFormat then
        ShowField('-');
      ShowD(Intensity);
      if not LinFormat then begin
        if not IsDistribution then
          ShowD(Ev2.E[Line.IndexU]);
        ShowD(Ev1.E[Line.IndexL]);
        ShowD(Line.Strength);
        if not IsDistribution then
          ShowD(A);
      end;
      ShowD(CLine.Width);
      Lab := MakeTxName(Line);
      l := Pos(':',Lab);
      if l > 0 then begin
        ShowField(Trim(Copy(Lab,1,l-1)));
        ShowField(Trim(Copy(Lab,l+1,Length(Lab))));
      end else begin
        ShowField('');
        ShowField(Trim(Lab));
      end;
      ShowDerivatives;
      ShowEOL;
    end;
end;

procedure TSpectrum.ShowFooter;
begin
  with SS do
    if Separator = ' ' then
      WriteLn(F^, nPrinted, ' Lines above OThreshold, Total intensity = ', dtosm(ITotal,14,-1))
    else begin
      ShowInt(nPrinted);
      ShowD(ITotal);
      ShowEOL;
    end;
  SetShower(PText(nil)^, plNone);
end;

{ TLineList }

function TLineList.GetLine(AIndex: Integer): TLine;
begin
  Result := FLines[AIndex];
end;

procedure TLineList.AddLine(const ALine: TLine; Simulation: TSimulation; AUnits: TResonanceUnits; var Problem: Exception);
{ Note that problem is never touched here }
begin
  if nLines >= High(FLines) then begin
    { To avoid large array allocation while intensity fitting, allocate in small steps
      to start with, then with rather larger steps }
    if High(FLines) < 200 then
      SetLength(FLines, High(FLines)+100)
    else if High(FLines) < 10000 then
      SetLength(FLines, High(FLines)+10000)
    else
      SetLength(FLines, High(FLines)+50000)
  end;
  FLines[nLines] := ALine;
  if AUnits <> Units then
    ConvertLine(FLines[nLines], AUnits, Units);
  Inc(FnLines);
  if ALine.Frequency < FMin then
    FMin := ALine.Frequency;
  if ALine.Frequency > FMax then
    FMax := ALine.Frequency;
end;

procedure TLineList.Clear;
begin
  inherited;
  FMax := -1.0e99;
  FMin := 1.0e99;
end;

{$IFNDEF CONSOLE}
procedure TLineList.FindLines(Centre, Range: Double; AUnits: TResonanceUnits; AColour: TColor; OnCheckLine: TOnCheckLine);
var
  i: Integer;
  p1, p2: Double;
begin
  p1 := Convert(Centre-Range, AUnits, Units);
  p2 := Convert(Centre+Range, AUnits, Units);
  Centre := (p1+p2)/2;
  Range := Abs(p2-Centre);
  for i := 0 to nLines-1 do
    with FLines[i] do
      { clHide is used to flag two different things:
        - in the linelist itself, clHide lines are ignored
        - for matching colours, clHide implies match all lines }
      if (Colour <> clHide) and ((AColour = clHide) or (Colour = AColour)) then
        if Abs(Frequency-Centre) <= Range then begin
          if Assigned(OnCheckLine) and not OnCheckLine(FLines[i]) then
            Continue;
          LineForm.AddLine(FLines[i],Units);
        end;
end;
{$ENDIF}

function TLineList.FindPeak(Left, Right: Double; out Position: Double;
  AUnits: TResonanceUnits; AColour: TColor): Boolean;
var
  SumI, SumIX, Intensity: Double;
  i: Integer;
  Source: TMNode;
  Problem: Exception;
begin
  Problem := nil;
  Left := Convert(Left, AUnits, Units);
  Right := Convert(Right, AUnits, Units);
  InOrder(Left, Right);
  SumI := 0;
  SumIX :=0;
  Source := nil;
  for i := 0 to nLines-1 do
    with Flines[i] do
      if ((AColour = clHide) or (Colour = AColour))
         and (Frequency >= Left) and (Frequency <= Right) then begin
        if Assigned(EigenvaluesU) then
          Source := EigenvaluesU.Manifold.CommonAncestor(Source);
        if Assigned(EigenvaluesL) then
          Source := EigenvaluesL.Manifold.CommonAncestor(Source);
        Intensity := GlobalMixture.Simulations[0].LineIntensity(FLines[i],Units,Problem);
        SumI := SumI + Intensity;
        SumIX := SumIX + Intensity*Frequency;
      end;
  Problem.Free; { Ignore any propblems for now }
  Result := SumI > 0;
  if Result then begin
    Position := SumIX/SumI;
    { and consider assigning it to a calibration spectrum }
    if Assigned(Source) then
      PeakMeasured(Convert(Position,Units,AUnits),AUnits,'Calculated',Source,0);
  end;
end;

procedure TLineList.ShowLine(LineNumber: Integer; Threshold:Double; LineFreqUnits, DispFreqUnits: TResonanceUnits; LinFormat: Boolean; var Problem: Exception);
begin
  ShowLine(FLines[LineNumber], Threshold, LineFreqUnits, DispFreqUnits, LinFormat, Problem);
end;

function TLineList.LargestIntensity: Double;
var
  Mixture: TMixture;
  Intensity: Double;
  i: Integer;
  Problem: Exception;
begin
  Result := 0;
  Problem := nil;
  if nLines = 0 then
    Exit;
  if Assigned(Line[0].EigenvaluesU) then
    Mixture := Line[0].EigenvaluesU.Manifold.Mixture
  else
    Mixture := Line[0].EigenvaluesL.Manifold.Mixture;
  for i := 0 to nLines-1 do begin
    Intensity := Mixture.Simulations[0].LineIntensity(Line[i], Units, Problem);
    if Intensity > Result then
      Result := Intensity;
  end;
  Problem.Free;
end;

procedure TLineList.Delete(n: Integer);
var
  i: Integer;
begin
  if n = nLines-1 then
    SetLength(FLines, n) { Delete last line }
  else
    for i := n to nLines-2 do
      FLines[i] := Line[i+1];
  Dec(FnLines);
end;

procedure TLineList.DeleteSelected(UpStateSel, LoStateSel: Integer);
var
  w, i: Integer;
  Keep: Boolean;
begin
  if (UpStateSel <= 0) and (LoStateSel <= 0) then
    Exit; { Nothing to do }
  w := 0; { Write pointer }
  for i := 0 to nLines-1 do
    with FLines[i] do begin
      Keep := ((UpStateSel <= 0) or (IndexU = UpStateSel))
              and ((LoStateSel <= 0) or (IndexL = LoStateSel));
      if Keep then begin
        FLines[w] := FLines[i];
        Inc(w);
      end;
    end;
  FnLines := w;
  SetLength(FLines,w);
end;

{ TStreamableMNode }

procedure TStreamableMNode.LoadFromStream(Stream: TStream; AFileName: string);
begin
  DoDirSeparators(AFileName);
  MergeFromStream(Stream, AFileName);
  Name := ExtractFileName(AFileName);
  FDirectoryName := ExtractFilePath(ExpandFileName(AFileName));
  Modified := False;
end;

procedure TStreamableMNode.LoadFromFile(AFileName: string);
var
  Stream: TStream;
begin
  DoDirSeparators(AFileName);
  Stream := TFileStream.Create(AFilename, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream, AFileName);
  finally
    Stream.Free;
  end;
end;

procedure TStreamableMNode.MergeFromStream(Stream: TStream; AFileName: string);
var
  XMLDocument: TXMLDocument;
  MNodeClass: TMNodeClass;
  nUnread: Integer;
  E: Exception;
begin
  ReadXMLFile(XMLDocument, Stream, AFileName);
  try
    MNodeClass := ReadMNodeClass(XMLDocument.FirstChild.NodeName,nil);
    if MNodeClass = nil then
      { Ignore - message already presented }
      nUnread := 0
    else if MNodeClass.InheritsFrom(ClassType) then
      nUnread := ReadProperties(XMLDocument.FirstChild)
    else if MNodeClass.CanBeChildOf(Self) then
      nUnread := ReadChildProperties(XMLDocument)
    else begin
      nUnread := 1;
      WriteLn(Log, 'Unrecognized item - ', MNodeClass.ClassName);
    end;
    if nUnread > 0 then
      ShowInfo('Warning - '+IntToStr(nUnread)+' unread item(s); see log for details');
    E := FixupProperties;
    if Assigned(E) then
      ShowInfo('Error:'+E.Message);
    E.Free;
  finally
    XMLDocument.Free;
  end;
end;

procedure TStreamableMNode.SaveToFile(AFileName: string);
var
  XMLDocument: TXMLDocument;
  OldName: string;
begin
  FixupProperties; { Discard error }
  XMLDocument := TXMLDocument.Create;
  try
    OldName := FName;
    FName := ExtractFileName(AFileName);
    FDirectoryName := ExtractFilePath(AFileName);
    WriteProperties(XMLDocument);
    WriteXMLFile(XMLDocument, AFilename);
    Modified := False;
    if FName <> OldName then
      PerformMNodeNotification(Self, mnNameChange, -1);
  finally
    XMLDocument.Free;
  end;
end;

{ TUnitsMNode }

function TUnitsMNode.GetIntensityUnits(const Index: Integer): TIntensityUnits;
begin
  Result := TIntensityUnits(GetSetting(Index));
end;

procedure TUnitsMNode.SetIntensityUnits(const Index: Integer;
  const AValue: TIntensityUnits);
begin
  ModifySetting(Index, Ord(AValue), IsPopDist(AValue)<>IsPopDist(GetIntensityUnits(Index)));
end;

function TUnitsMNode.GetUnits(const Index: Integer): TResonanceUnits;
begin
  Result := TResonanceUnits(GetSetting(Index));
end;

procedure TUnitsMNode.SetUnits(const Index: Integer; const AValue: TResonanceUnits);
begin
  SetSetting(Index, Ord(AValue));
end;

function TUnitsMNode.UnitsName: string;
begin
  Result := NameOfUnit(GetUnits(1));
end;

function TUnitsMNode.UnitsTypeName: string;
begin
  Result := TypeOfUnit(GetUnits(1));
end;

{ TPerturbation }

destructor TPerturbation.Destroy;
begin
  if FNoHyperfineMatrixElements <> FMatrixElements then
    DiscardMatrixElements(FNoHyperfineMatrixElements);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  ParameterVariable.Free;
  inherited Destroy;
end;

function TPerturbation.IsAntiSym: Boolean;
begin
  raise NotYet; { Only used for Intensity perturbations }
end;

class function TPerturbation.BraKetClass: TMNodeClass;
begin
  Result := TState;
end;

function TPerturbation.Connects(State1, State2: TState): Boolean;
begin
  Result := ((State1 = Bra)and (State2 = Ket))
            or ((State2 = Bra)and (State1 = Ket));
end;

constructor TPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',SymSelect=all';
  inherited;
end;

function TPerturbation.FixupProperties: Exception;
begin
  if FNoHyperfineMatrixElements <> FMatrixElements then
    DiscardMatrixElements(FNoHyperfineMatrixElements);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  FMatrixElementsEmpty := False;
  if not Assigned(ParameterVariable) then
    FParameterVariable := TSortedVariable.Create(Name,500);
  FParameterVariable.Value := Value;
  FParameterVariable.SetAllValues;
  FParameterVariable.Name := Name;
  Result := inherited FixupProperties;
end;

function TPerturbation.CheckMatrixElementsAvailable: Boolean;
begin
  Result := True; { Default is that they can be implemented; Return false if not implementable }
  if FMatrixElementsEmpty or (Length(FMatrixElements) > 0) then
    Exit;
  MustBeSerial;
  FMatrixElements := SymbolicMatrixElements(True);
  FMatrixElementsEmpty := Length(FMatrixElements) = 0;
  if FMatrixElementsEmpty then
    Exit;
  CompileMatrixElements(FMatrixElements, ParameterVariable.Template);
  if (Owner.Owner as TMolecule).nNuclei > 0 then begin
    FNoHyperfineMatrixElements := SymbolicMatrixElements(False);
    CompileMatrixElements(FNoHyperfineMatrixElements, ParameterVariable.Template);
  end else
    FNoHyperfineMatrixElements := FMatrixElements;
  FSymmetrizedMatrixElements := (Bra as TState).ConvertToSymmetrized(FMatrixElements);
  CompileMatrixElements(FSymmetrizedMatrixElements, nil);
end;

function TPerturbation.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  if AIndex = LPI-2 then
    Result := SymSettingsNames(Self, LowestValue)
  else
    Result := inherited GetSettingsNames(AIndex, LowestValue);
end;

function TPerturbation.MatrixElement(State1, State2: TBasisState; MatElemType: TMatElemType): Complex;
begin
  Result := State1.State.PerturbationMatrixElement(Self, State1, State2, MatElemType)
end;

function TPerturbation.PairName: string;
begin
  Result := 'H';
end;

procedure TPerturbation.ShowSymbolicMatrixElements(var Log: Text; const Indent: string);
var
  Elements, SymmetrizedElements: TMatrixElements;
  j: Integer;
begin
  try
    TStart(IndividualSymbolicMatrixElementTime);
    Elements := SymbolicMatrixElements(True);
    TEnd(IndividualSymbolicMatrixElementTime);
    for j := 0 to High(Elements) do
      WriteLn(Log, Indent, (Bra as TState).MatrixElementType(Elements[j], PairName), ' = ', Elements[j].Expr.Text(True));
    WriteLn(Log);
    SymmetrizedElements := (Bra as TState).ConvertToSymmetrized(Elements);
    if Length(SymmetrizedElements) > 0 then begin
      Writeln(Log, Indent, 'Symmetrized Matrix Elements of ', Name);
      for j := 0 to High(SymmetrizedElements) do
        WriteLn(Log, Indent, (Bra as TState).MatrixElementType(SymmetrizedElements[j], PairName, True), ' = ', SymmetrizedElements[j].Expr.Text(True));
      WriteLn(Log);
    end;
    DiscardMatrixElements(SymmetrizedElements);
    DiscardMatrixElements(Elements);
  except
    WriteLn(Log, Indent, 'Symbolic matrix elements not available for ', PairName);
  end;
end;

procedure TPerturbation.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  inherited;
  if (PrintLevel < plDetail) or (Value = 0) then
    Exit;
  ShowSymbolicMatrixElements(F, Indent);
end;

procedure TPerturbation.PrettyPrintSettings(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  if Nucleus > 0 then
    WriteLn(F, Indent, 'Nucleus ', Nucleus);
  if ScalePrev <> 0 then
    WriteLn(F, Indent, 'Scale from previous ', FloatToStr(ScalePrev));
end;

function TPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    LPI-3: Result := 'Index, starting from 1, of nuclear spin involved in perturbation; 0 (default) for those not involving a nuclear spin.';
    LPI-2: Result := 'Symmetry select';
    LPI-1: Result := 'Scale factor with respect to preceding perturbation';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

function TPerturbation.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Size of perturbation.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure TPerturbation.SetParameter(AIndex: Integer; const Value: Double);
var
  i: Integer;
begin
  { Need to test for Nan/Inf, as this special case forces
    restoration of the default value }
  if not IsNanOrInf(Value) and (Parameters[AIndex] = Value) then
    Exit;
  inherited SetParameter(AIndex, Value);
  if Assigned(ParameterVariable) and (AIndex = 1) then
    FParameterVariable.Value := GetParameter(AIndex);
  if Assigned(Owner) then
    for i := ComponentIndex+1 to Owner.ComponentCount - 1 do
      if Owner.Components[i] is TPerturbation then begin
        if TPerturbation(Owner.Components[i]).ScalePrev <> 0 then
          TPerturbation(Owner.Components[i]).Value := Value*TPerturbation(Owner.Components[i]).ScalePrev;
        Break;
      end;
end;

function TPerturbation.SymbolicMatrixElements(WithHyperfine: Boolean): TMatrixElements;
begin
  raise NotYet;
end;

{ TMixtureMNode }

function TMixtureMNode.BoltzmannConstant: Double;
var
  Units: TEnergyUnits;
begin
  if RootNode is TMixture then
    Units := TMixture(RootNode).Units
  else
    Units := cm1;
  Result := BoltzmannConstants[Units];
end;

function TMixtureMNode.Precision: Integer;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).Precision
  else
    Result := 4; { TODO: Raise }
end;

function TMixtureMNode.QuantumNumberFormat: TQuantumNumberFormat;
begin
  if RootNode is TMixture then
    Result := TMixture(RootNode).QuantumNumberFormat
  else
    Result := GlobalMixture.QuantumNumberFormat;
end;

function TMixtureMNode.QuantumNumberToStr(AQuantumNumber: TQuantumNumber): string;
begin
  if Assigned(Self) then
    Result := QuantumNumber.QuantumNumberToStr(AQuantumNumber,QuantumNumberFormat)
  else { Take default; means that GlobalMixture.QuantumNumberToStr will always work }
    Result := QuantumNumber.QuantumNumberToStr(AQuantumNumber,qftJ);
end;

function TMixtureMNode.StrToQuantumNumber(s: string): TQuantumNumber;
var
  OK: Boolean;
begin
  OK := TryStrToQuantumNumber(s,Result,QuantumNumberFormat);
  if not OK then
    raise EReportException.Create('Bad quantum number format:'+s);
end;

{ TNucleus }

procedure TNucleus.AddNodeParameters(var FittingParameters: TParameters);
begin
  if not HasIdentical or not AsNext then
    inherited;
  { else Parameters constrained to be identical }
end;

constructor TNucleusWithSpin.Create(AOwner: TComponent);
begin
  { Default mechanism overwrites name, spin and AsNext so defeat during creation }
  SavedProperties := SavedProperties + ',Spin=0,AsNext=false';
  inherited Create(AOwner);
  { Spin and AsNext should be per molecule settings, but are actually
    in the per state object. Copy from the per molecule setting on
    initialisation }
  ModifySetting(1, Molecule.Nuclei[NodeIndex].Spin, False);
  ModifySetting(2, Ord(Molecule.Nuclei[NodeIndex].AsNext), False);
end;

constructor TNucleus.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if Owner = nil then Exit; { Documentation purposes only }
  FState := AOwner as TState;
  State.AdjustNuclei(False);
end;

destructor TNucleus.Destroy;
begin
  ClearCache(False);
  inherited Destroy;
end;

procedure TNucleus.ClearCache(ClearChildren: Boolean);
begin
  inherited ClearCache(ClearChildren);
  DiscardMatrixElements(FMatrixElements);
  DiscardMatrixElements(FSymmetrizedMatrixElements);
  DiscardMatrixElements(FMatrixElementsByRank[1]);
  DiscardMatrixElements(FMatrixElementsByRank[2]);
  FMatrixElementsEmpty := False;
end;

function TNucleus.MatchingNucleus: TNucleus;
begin
  Result := Self;
  if not HasIdentical then
    Exit;
  if Result.AsNext then begin
    { Same as next nucleus }
    if High(State.FNuclei) > NodeIndex then
      Result := State.Nucleus(NodeIndex+1);
  end else
    { Check for match to previous; if found return first nucleus in list }
    while (Result.NodeIndex > 0) and State.Nucleus(Result.NodeIndex-1).AsNext do
        Result := State.Nucleus(Result.NodeIndex-1)
end;

function TNucleus.Molecule: TMolecule;
begin
  Result := Owner.Owner.Owner as TMolecule;
end;

procedure TNucleus.ScaleParameters(UFrom, UTo: TResonanceUnits);
begin
  if not HasIdentical or not AsNext then
    inherited; { Test avoids parameter being scaled twice }
end;

procedure TNucleus.SetComponentIndex(const Value: Integer);
begin
  inherited;
  State.AdjustNuclei(False);
end;

procedure TNucleus.SetFloating(AIndex: Integer; const Value: Boolean);
var
  Other: TNucleus;
begin
  if Floating[Aindex] = Value then
    Exit;
  inherited SetFloating(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetFloating(AIndex, Value)
end;

procedure TNucleus.SetIncrement(AIndex: Integer; const Value: Double);
var
  Other: TNucleus;
begin
  { TODO -cStdDev: Is this logic required for StdDev also? }
  if Increment[Aindex] = Value then
    Exit;
  inherited SetIncrement(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetIncrement(AIndex, Value)
end;

procedure TNucleus.SetName(const Value: TComponentName);
begin
  if NodeIndex >= 0 then
    Molecule.Nuclei[NodeIndex].Name := Value;
  inherited;
end;

procedure TNucleus.SetParameter(AIndex: Integer; const Value: Double);
var
  Other: TNucleus;
begin
  { Need to test for Nan/Inf, as this special case forces
    restoration of the default value }
  if not IsNanOrInf(Value) and (Parameters[AIndex] = Value) then
    Exit;
  inherited SetParameter(AIndex, Value);
  Other := MatchingNucleus;
  if Other <> Self then
    Other.SetParameter(AIndex, GetParameter(AIndex)) { Use GetParameter to avoid infinte loop on setting default value }
end;

procedure TNucleus.ConvertToPerturbations;
var
  i: Integer;
begin
  for i := 1 to nParameters do
    if Parameters[i] <> 0 then
      ConvertParameter(i, True);
end;

function TNucleus.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
begin
  Result := nil;
end;

function TNucleus.SymbolicRotationalMatrixElement(AIndex: Integer; RankWanted: Integer): TMatrixElements;
begin
  raise EReportException.Create('Symbolic matrix elements not implemented for '+UniqueName+'.'+ParameterNames[AIndex]);
end;

procedure TNucleus.CheckMatrixElementsAvailable;
var
  Rank: Integer;
begin
  if (Length(MatrixElements) > 0) or MatrixElementsEmpty then
    Exit;
  DoCheckMatrixElementsAvailable(FMatrixElements, FSymmetrizedMatrixElements, -1);
  FMatrixElementsEmpty := Length(FMatrixElements) = 0;
  for Rank := 1 to 2 do
    DoCheckMatrixElementsAvailable(FMatrixElementsByRank[Rank], FSymmetrizedMatrixElements, Rank);
end;

function TNucleus.ParameterVariable(AIndex: Integer; Suffix: string): TSortedVariableNode;
begin
  Result := inherited ParameterVariable(AIndex, '_'+Name);
end;

function TNucleus.GetName: TComponentName;
begin
  Result := Molecule.Nuclei[NodeIndex].Name;
  if Result = '' then
    Result := Format('Nucleus%d',[NodeIndex+1]);
end;

procedure TNucleus.AddHandlers;
begin
  inherited;
  AddHandler('Global Shift Up', 'Shift values and names in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveUp);
  AddHandler('Global Shift Down', 'Shift values and names in all states', {$IFDEF FPC}@{$ENDIF}GlobalMoveDown);
end;

procedure TNucleus.GlobalMoveDown(Sender: TObject);
begin
  GlobalShift(1);
end;

procedure TNucleus.GlobalMoveUp(Sender: TObject);
begin
  GlobalShift(-1);
end;

procedure TNucleus.GlobalShift(Shift: Integer);
var
  s: string;
  Source, Target: TNucleus;
  i, j, MyIndex: Integer;
begin
  if (NodeIndex+Shift < 0) or (NodeIndex+Shift >= Molecule.nNuclei) then
    Exit;
  Target := State.Nucleus(NodeIndex+Shift);
  { Move global values once }
  s := Name;
  Name := Target.Name;
  Target.Name := s;
  AsNext := False;        { Probably not right, but probably safest option }
  Target.AsNext := False;
  MyIndex := NodeIndex;
  { Now move all other settings }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        with State[j] do begin
          Source := State[j].Nucleus(MyIndex);
          Target := State[j].Nucleus(MyIndex+Shift);
          if Shift > 0 then
            Source.ComponentIndex := Target.ComponentIndex
          else
            Target.ComponentIndex := Source.ComponentIndex;
        end;
  { Reset spin, so copied to global array }
  SetSetting(1,GetSetting(1));
  Target.SetSetting(1,Target.GetSetting(1));
end;

procedure TNucleus.SetSetting(AIndex: Integer; const AValue: Integer);
var
  i, j, MyIndex: Integer;
  Changed: Boolean;
begin
  inherited SetSetting(AIndex, AValue);
  { See notes in Create method; set per molecule settting also }
  if NodeIndex < 0 then
    Exit;
  Changed := False;
  case AIndex of
    1: begin Changed := Molecule.Nuclei[NodeIndex].Spin <> AValue; Molecule.Nuclei[NodeIndex].Spin := AValue; end;
    2: begin Changed := Molecule.Nuclei[NodeIndex].AsNext <> (AValue <> 0); Molecule.Nuclei[NodeIndex].AsNext := AValue <> 0; end;
  end;
  if not Changed then
    Exit;
  MyIndex := NodeIndex;
  { As GetSetting is not virtual, changes must be copied to all other nueclei }
  for i := 0 to Molecule.nManifolds - 1 do
    with Molecule.Manifold[i] do
      for j := 0 to nStates - 1 do
        State[j].Nucleus(MyIndex).ModifySetting(AIndex, AValue, True);
end;

function TNucleus.HelpOnSetting(AIndex: Integer): string;
begin
  case AIndex of
    1: Result := 'Nuclear spin';
    2: Result := 'Nucleus is equivalent to following nucleus';
  else
    Result := inherited HelpOnSetting(AIndex);
  end;
end;

procedure TNucleus.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
begin
  inherited PrettyPrint(F, PrintLevel, Indent);
  if PrintLevel >= plDetail then
    ShowMatrixElements(Log, Self, State, 1, nParameterVariables, False, Indent+'  ');
end;

function TNucleus.HasIdentical: Boolean;
begin
  Result := False; { By default, all nuclei must be different from any other }
end;

function TNucleus.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex = 1;
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

{ TMagneticTransitionMoment }

function TMagneticTransitionMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1  then
    Result := 'Rank of transition - ' + GlobalMixture.QuantumNumberToStr(2) + ' for a normal magnetic dipole transition.'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TMagneticTransitionMoment.PairName: string;
begin
  if Rank = 2 then
    if Component > -9999 then
      Result := Format('(%s)', [TMixture(RootNode).QuantumNumberToStr(Component)])
    else
      Result := ''
  else
    Result := Format('(%s,%s)', [TMixture(RootNode).QuantumNumberToStr(Rank), TMixture(RootNode).QuantumNumberToStr(Component)]);
end;

constructor TMagneticTransitionMoment.Create(AOwner: TComponent);
begin
  FMagnetic := True;
  inherited;
end;

{ TSpinTransitionMoment }

function TSpinTransitionMoment.PairName: string;
begin
  Result := 'gS' + inherited PairName;
end;

constructor TSpinTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=' + FloatToStr(ge);
  inherited;
end;

{ TOrbitalTransitionMoment }

function TOrbitalTransitionMoment.PairName: string;
begin
  Result := 'gL' + inherited PairName;
end;

{ TRotationalTransitionMoment }

constructor TRotationalTransitionMoment.Create(AOwner: TComponent);
begin
  FNuclearMagneton := True;
  FMagnetic := True;
  inherited;
end;

function TRotationalTransitionMoment.PairName: string;
begin
  Result := 'gr';
end;

function TRotationalTransitionMoment.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  if (RootNode as TMixture).ReadVersion < 5003335 then
    ShowInfo('Warning - change in definition of Strength of Rotational Transition Moments');
  Result := inherited ReadProperties(Node, CheckUnread);
end;

{ TSpinSpin }

class function TSpinSpin.BraKetClass: TMNodeClass;
begin
  Result := TNucleus;
end;

function TSpinSpin.PairName: string;
begin
  Result := 'II';
end;

{ TEigenValueCache }

procedure TEigenValueCache.Clear;
var
  i, j: Integer;
begin
  for i := 0 to High(Entries) do
    for j := 0 to High(Entries[i]) do
      FreeAndNil(Entries[i][j]);
end;

function TEigenValueCache.Hash(J: TQuantumNumber; Sym: Integer): Integer;
begin
  Result := Abs(J+Sym) mod Length(Entries);
end;

procedure TEigenValueCache.Add(Ev: TEigenvalues; InsertAt: Integer);
var
  Entry: Integer;
begin
  { Write Lock assumed held }
  Entry := Hash(Ev.J, Ev.Sym);
  if InsertAt > High(Entries[Entry]) then
    SetLength(Entries[Entry], InsertAt+10);
  Entries[Entry][InsertAt] := Ev;
end;

{ TSimulation }

procedure TSimulation.AddHandlers;
begin
  inherited;
  AddHandler('Temperature Table', 'Print Temperature Dependence of Centre Frequency of Plot', {$IFDEF FPC}@{$ENDIF}TemperatureTable);
end;

procedure TSimulation.ChangePlotUnits(NewPlotUnits: TResonanceUnits);
var
  MidPoint, d: Double; { Centre plot position - needed for scaling widths }
begin
  { Change plot units, scaling affected values also }
  if NewPlotUnits = PlotUnits then
    Exit;
  if IsWavelengthUnit(NewPlotUnits) and (Math.Min(FMin, FMax) <= 0) then begin
    d := Math.Max(Abs(Fmin), Abs(Fmax));
    FMin := d/10;
    FMax := d;
  end;
  MidPoint := (FMin + FMax) / 2;
  FMin := Convert(FMin,PlotUnits,NewPlotUnits);
  FMax := Convert(FMax,PlotUnits,NewPlotUnits);
  Gaussian := ConvertWidth(Gaussian, MidPoint, PlotUnits, NewPlotUnits);
  Lorentzian := ConvertWidth(Lorentzian, MidPoint, PlotUnits, NewPlotUnits);
  { Change sign of displacement if taking reciprocal }
  Foffset := IfThen(IsWavelengthUnit(PlotUnits) xor IsWavelengthUnit(NewPlotUnits), -1, 1)
             * ConvertWidth(Foffset, MidPoint, PlotUnits, NewPlotUnits);
  PlotUnits := NewPlotUnits;
end;

constructor TSimulation.Create(AOwner: TComponent);
var
  i: Integer;
begin
  SavedProperties := SavedProperties + ',OThreshold=1e-6,WidthMult=6,SMargin=-1,Temperature=300,FMin=0,FMax=50000,nDF=2000,ShowSum=True,Tvib=-1,Tspin=-1';
  SavedProperties := SavedProperties + ',ScaleMarkSize=True,ShowSymmetry=True,Name=Simulation';
  SavedProperties := SavedProperties + ',IntensityUnits=Normalized,FortratQno=J,AutoMin=True,AutoMax=True,PlotSplit=0.5';

  DimensionlessParameters := DimensionlessParameters + ',OThreshold,SMargin,Temperature,MinI,Tvib,Saturation,EField,BField,Tspin,Doppler,PlotSplit';
  { The next five are not dimensionless, but are only affected by PlotUnits }
  DimensionlessParameters := DimensionlessParameters + ',Foffset,Fmin,Fmax,Gaussian,Lorentzian,Ymin,Ymax';
  { ... in fact the only parameter with units is RefWidth! }

  inherited;
  HasPlotted := False;
  SetLength(FSettingOverridden, nSettings+1);
  SetLength(FParameterOverridden, nParameters+1);
  ResetOverride;
  if Assigned(AOwner) then
    for i := 0 to AOwner.ComponentCount - 1 do
      if AOwner.Components[i] is TSimulation then begin
        if AOwner.Components[i] <> Self then
          FBaseNode := AOwner.Components[i] as TSimulation;
        Break;
      end;
end;

function TSimulation.GetDefaultSettings(AIndex: Integer): string;
begin
  if Assigned(FBaseNode) then
    Result := 'Track'
  else
    Result := inherited GetDefaultSettings(AIndex);
end;

function TSimulation.GetExField: TStaticFields;
begin
  Result.E := EField;
  Result.B := BField;
  Result.Angle := 0;
end;

function TSimulation.GetFortratQno(const Index: Integer): TQnoType;
begin
  Result := TQnoType(GetSetting(Index));
end;

function TSimulation.GetLifeModel(const Index: Integer): TLifeModel;
begin
  Result := TLifeModel(GetSetting(Index));
end;

function TSimulation.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if SettingNames[AIndex] = 'FortratQno' then
    Result := @QnoTypeNames[0];
end;

const
 SimulationParameterNames: array [1..19] of string = (
'Left edge of plot range.',
'Right edge of plot range.',
'Bottom of plot range; updated automatically if AutoMin set.',
'Top of plot range; updated automatically if AutoMax set.',
'Rotational temeperature (Kelvin).',
'Gaussian contribution to linewidth (full width half maximum). If both this and a Lorentzian width (below) is set the result is a convolution of the two, a Voigt profile.',
'Lorentzian contribution to linewidth (full width half maximum). This and the Gaussian width can also be set from the main window.',
'Frequency offset to simulation.',
'Extra points to calculate off each end of required spectrum.',
'Ignore peaks smaller than this fraction of the maximum peak intensity.',
'Reference width in linewidth (predissociation) calculations.',
'Vibrational Temperature (Kelvin); set to -1 (default) to use rotational temperature for all Boltzmann factors.',
'Set the scaling of mark sizes in Fortrat plots.',
'Zero for normal calculation; positive values progesively switch strength to population only',
'Spin Temperature; set to -1 (default) to equilibrate different spinstates',
'Static Electric Field (V/m).',
'Static Magnetic Field (Tesla).',
'Doppler Splitting',
'Fraction of plot area to use for simulation where plots separated (-1..1)'
//'Angle between fields (degrees).'
);

function TSimulation.HasFields: Boolean;
begin
  Result := (EField <> 0) or (BField <> 0);
end;

function TSimulation.HasWidths: Boolean;
begin
  Result := (Gaussian <> 0) or (Lorentzian <> 0) or (Doppler <> 0);
end;

function TSimulation.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(SimulationParameterNames))
     and (AIndex <= High(SimulationParameterNames)) then
    Result := SimulationParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

const
  SimulationSettingNames: array[1..17] of string = (
'Units for intensity. For those with units, transition moments should be in Debye.',
'Model to use for state dependent lifetimes and linewidths. Valid values are lmNone, lmWidth, lmProduct, lmProductWidth, lmParent, lmParentWidth, lmGate, lmGateWidth.',
'Units for the horizontal scale for spectrum plots',
'Number of points (between Fmin and Fmax) to calculate the spectrum at. Note that if a peak width is less than 3*(Fmax-Fmin)./nDF, i.e. only a few points wide, then it is shown as a stick, rather than a peak.',
'Multiple of linewidth to extend convolution over.',
'True to set Ymin to minimum value in plot range',
'True to set Ymax to minimum value in plot range',
'Plot overall sum of individual spectra.',
'Plot individual spectra making up overall spectrum. The individual spectra are grouped by colour, so you will need to set colours to see something different to the sum. Colours can be set at the transition moment, state, molecule or species level.',
'Plot a Fortrat diagram, i.e. J against frequency, in the main window.',
'Set to use upper rather than lower state J, symmetry and state number in the Fortrat diagram.',
'Show symmetry in Fortrat plots.',
'Show change in J in Fortrat plots.',
'Scale Mark Size with intensity in Fortrat plots.',
'Set to show different symmetries in separate Fortrat plots.',
'Set to show different state numbers in separate Fortrat plots.',
'Select quantum number to use in Fortrat plots.'
);

function TSimulation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(SimulationSettingNames))
     and (AIndex <= High(SimulationSettingNames)) then
    Result := SimulationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TSimulation.LineIntensity(const Line: TLine; AUnits: TResonanceUnits; var Problem: Exception; RemoveSpins: Byte): Double;
var
  g: Integer;
  Str, Emin: Double;
begin
  if not Assigned(Line.EigenvaluesL) or not IncludesPopulation(IntensityUnits) then begin
    Result := Line.Strength;
    if IntensityUnits = Squared then
      Result := Sqr(Result);
    if (IntensityUnits = EinsteinA) and Assigned(Line.EigenvaluesU)  then
      Result := Result / Line.EigenvaluesU.Degeneracy(Line.IndexU);
    if (IntensityUnits = OscillatorStrength) and Assigned(Line.EigenvaluesL) then
      Result := Result / Line.EigenvaluesL.Degeneracy(Line.IndexL);
  end else begin
    if not IsPopDist(IntensityUnits) and Assigned(Line.EigenvaluesU) and Line.EigenvaluesU.Manifold.Initial then
      Result := Line.EigenvaluesU.Manifold.Population(Line.EigenvaluesU, Line.IndexU, Self, Problem, 0)
    else
      Result := 0;
    if Line.EigenvaluesL.Manifold.Initial then
      { Use population difference - essential for microwave transitions }
      Result := Abs(Line.EigenvaluesL.Manifold.Population(Line.EigenvaluesL,Line.IndexL,Self,Problem,0)-Result);
    if IsPopDist(IntensityUnits) then
      Str := Line.Strength * (Line.EigenvaluesL.J+1)
    else if Saturation > 0 then begin
      { Basic allowance for saturation (assuming z polarized light) }
      g := Line.EigenvaluesL.J;
      if Assigned(Line.EigenvaluesU) then
        g := Min(Line.EigenvaluesU.J, g);
      g := (g+1)*Line.EigenvaluesL.StatisticalWeight(Line.IndexL);
      Str := (1-Exp(-Line.Strength*Saturation/g)) * g;
    end else
      Str := Line.Strength;
    Result := Result * Str;
    if IntensityUnits = Squared then
      Result := Sqr(Result);
    if IsNormalized(IntensityUnits) then
      Result := Result / Line.EigenvaluesL.Manifold.Molecule.PartitionFunction(Self, Problem, Emin, RemoveSpins);
  end;
  { TODO: different scaling factor in IntensityFactor for Quadrupole transitions }
  Result := Result * IntensityFactor(Line.Frequency,AUnits,IntensityUnits);
end;

function TSimulation.MakeLineList(var Problem: Exception): TLineList;
begin
  TStart(MakeLineListTime);
  Mixture.LineList := nil;
  Result := TLineList.Create(Mixture.Units,IsPopDist(IntensityUnits));
  Result.Temperature := Temperature;
  try
    if not Mixture.AddToSpectrum(Result, Problem, Self) then
      FreeAndNil(Result);
    Mixture.LineList := Result;
  finally
    if Mixture.LineList = nil then
      Result.Free; { Exception raised }
  end;
  TEnd(MakeLineListTime);
end;

function TSimulation.Mixture: TMixture;
begin
  Result := Owner as TMixture;
end;

function TSimulation.ParameterOverridden(AIndex: Integer): Boolean;
begin
  Result := FParameterOverridden[AIndex];
end;

function TSimulation.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  Result := inherited ReadProperties(Node, CheckUnread);
  { EinsteinASum was introduced after 6.0.228; before that IntensityUnits = EinsteinA gave current EinsteinASum behaviour. }
  if ((RootNode as TMixture).ReadVersion <= 6000228) and (IntensityUnits = EinsteinA) then
    IntensityUnits := EinsteinASum;
end;

procedure TSimulation.ResetOverride;
begin
  FillChar(FSettingOverridden[0], Length(FSettingOverridden)*SizeOf(Boolean),0);
  FillChar(FParameterOverridden[0], Length(FParameterOverridden)*SizeOf(Boolean),0);
end;

procedure TSimulation.SetFortratQno(const Index: Integer; const Value: TQnoType);
begin
  SetSetting(Index, Ord(Value));
end;

procedure TSimulation.SetLifeModel(const Index: Integer; const AValue: TLifeModel);
begin
  SetSetting(Index, Ord(AValue));
end;

{ Changing frequency limits and so on does not count as a proper change
  Add mechanism to track individual changes }
procedure TSimulation.SetParameter(AIndex: Integer; const Value: Double);
begin
  if ((AIndex in [EfieldIndex, BfieldIndex]) and ((Parameters[AIndex] <> Value)) or IsNanOrInf(Value)) then
    Mixture.ClearCache(True);
  if Assigned(FBaseNode) then
    if IsNanOrInf(Value) then begin
      FParameterOverridden[AIndex] := False;
      inherited ModifyParameter(AIndex, FBaseNode.Parameters[AIndex], False);
      Exit;
    end else if GetParameter(AIndex) <> Value then
      FParameterOverridden[AIndex] := True;
  inherited ModifyParameter(AIndex, Value, False, True);
end;

procedure TSimulation.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if Assigned(FBaseNode) and (GetSetting(AIndex) <> AValue) then
    FSettingOverridden[AIndex] := True;
  inherited ModifySetting(AIndex, AValue, False, True);
end;

procedure TSimulation.SetSettingFromStr(AIndex: Integer; const s: string);
begin
  if Assigned(FBaseNode) and (s = 'Track') then begin
    FSettingOverridden[AIndex] := False;
    PerformMNodeNotification(Self, mnSettingChange, AIndex);
  end else
    inherited;
end;

procedure TSimulation.PrettyPrintHeader(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  C: double;
begin
  inherited PrettyPrintHeader(F, PrintLevel, Indent);
  C := IntensityFactor(1,PlotUnits,IntensityUnits);
  WriteLn(F, Indent, '  Intensity factor:', FloatToStr(C/Three), ' (', FloatToStr(C), ' from Spol)');
end;

function TSimulation.SettingToStr(AIndex: Integer): string;
begin
  if Assigned(FBaseNode) and not FSettingOverridden[AIndex] then
    Result := 'Track'
  else
    Result := inherited SettingToStr(AIndex);
end;

var
  TTstep: string = '100';

procedure TSimulation.TemperatureTable(Sender: TObject);
var
  p, Tstep, Tsave: Double;
  i, nPoints: Integer;
  SpectrumPlot: TSpectrumPlot;
  Problem: Exception;
  Treq: array of Double;
  num: string;
begin
  p := Fmin + (Fmax-Fmin)/(nDF-1)*(nDF div 2);
  {$IFNDEF CONSOLE}
  if not InputQuery('Temperature Dependence at '+FloatToStr(p)+' '+NameOfUnit(PlotUnits),
                    'Temperature step size, or list of temperatures:',TTstep) then
    Exit;
  {$ENDIF}
  TTstep := Trim(TTstep);
  if (Pos(' ', TTstep) > 0) or (Pos(',',TTstep) > 0) or (Pos(#9,TTstep) > 0) then begin
    num := '';
    TTstep := TTstep + ' '; { Ensure trailing separator }
    SetLength(Treq,0);
    for i := 1 to Length(TTstep) do
      if TTstep[i] in [' ',',',#9] then begin
        num := Trim(num);
        if num <> '' then begin
          SetLength(Treq, Length(Treq)+1);
          Treq[High(Treq)] := StrToFloatEx(num);
        end;
        num := '';
      end else
        num := num + TTstep[i];
    TTstep := Trim(TTstep);
  end else begin
    { Inteval only specified }
    Tstep := Abs(StrToFloat(TTstep));
    nPoints := Round(Temperature/Tstep+1);
    SetLength(Treq, nPoints);
    Treq[0] := Tstep/100;
    Treq[nPoints-1] := Temperature;
    for i := 1 to nPoints-2 do
      Treq[i] := i*Tstep;
  end;
  Tsave := Temperature;
  try
   PushProgress('Temperature Dependence', Round(Temperature));
   for i := 0 to High(Treq) do begin
     Temperature := Treq[i];
     try
       SpectrumPlot := TSpectrumPlot.Create(PlotUnits,IsPopDist(IntensityUnits));
       SpectrumPlot.MakeFromSimulation(Self,False,Problem);
       if i = 0 then begin
         WriteLn(Log, 'Temperature Dependence at '+FloatToStr(p)+' '+NameOfUnit(PlotUnits));
         WriteLn(Log, '    T/K   Intensity/(',NameOfUnit(IntensityUnits, SpectrumPlot.IntegratedUints, PlotUnits),')');
       end;
       SetProgress(Round(Temperature));
       if Assigned(Problem) then
         WriteLn(Log, FloatToStr(Temperature):7, ' ', Problem.Message)
       else begin
         SpectrumPlot.Convolute;
         WriteLn(Log, FloatToStr(Temperature):7, ' ', SpectrumPlot.Plot[nDF div 2]);
       end;
     finally
       SpectrumPlot.Free;
     end;
   end;
  finally
    Temperature := Tsave;
    PopProgress;
  end;
  ReShowLogForm;
end;

procedure TSimulation.UpdateBaseNode(ABaseNode: TSimulation);
var
  i: Integer;
begin
  FBaseNode := ABaseNode;
  if not Assigned(FBaseNode) then
    Exit;
  for i := 1 to nParameters do
    if not FParameterOverridden[i] then
      ModifyParameter(i, FBaseNode.GetParameter(i), False);
  for i := 0 to nSettings do
    if not FSettingOverridden[i] then
      ModifySetting(i, FBaseNode.GetSetting(i),False);
end;

procedure TSimulation.WriteFloatableValue(AIndex: Integer; Deflt: string; DOMNode: TDOMNode; DefaultCheck: Boolean);
begin
  if Assigned(FBaseNode) and not FParameterOverridden[AIndex] then
    Exit; { Skip - not overriden }
  { Compatibility - do not write Ymin/Ymax if autorange on }
  if AutoMin and (AIndex = IndexOfParameter('Ymin')) then
    Exit;
  if AutoMax and (AIndex = IndexOfParameter('Ymax')) then
    Exit;
  inherited;
end;

function TSimulation.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
begin
  if not FWritten then
    Result := inherited WriteProperties(Parent, AllParameters, WithChildren)
  else
    Result := nil;
end;

{ TCachedTransitionMatrix }

destructor TCachedTransitionMatrix.Destroy;
begin
  FreeAndNil(TransitionMatrix);
  inherited;
end;

procedure TCachedTransitionMatrix.DoMake(RequiredAttributes: Pointer);
begin
  EigenValues.Make(True,[meHrot]);
  OtherEigenValues.Make(True,[meHrot]);
  TransitionMatrix := TransitionMoments.TransitionMatrix(Eigenvalues,OtherEigenValues,True,False,nil,Simulation,PrintLevel);
  IsEmpty := not Assigned(TransitionMatrix);
end;

function TCachedTransitionMatrix.HasCorrectAttributes(Attributes: Pointer): Boolean;
begin
  Result := Assigned(TransitionMatrix) or IsEmpty;
end;

{ TBaseRotationalTransitionMoment }

constructor TBaseRotationalTransitionMoment.Create(AOwner: TComponent);
begin
  DimensionlessParameters := DimensionlessParameters + ',C,D';
  inherited;
end;

function TBaseRotationalTransitionMoment.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  cN, dN: TDOMNode;
  CustomTransitionFunction: TCustomTransitionFunction;
begin
  cN := FindFloatableValue(Node, 'C');
  dN := FindFloatableValue(Node, 'D');
  if (cN <> nil) or (dN <> nil) then begin
    {$IFNDEF CONSOLE}
      ShowInfo('Note that C and D parameters in TransitionMoment have been converted to Custom Transition Functions');
    {$ENDIF}
    CustomTransitionFunction := TCustomTransitionFunction.Create(Self);
    with CustomTransitionFunction do begin
      Comment := 'Value*Sqrt(1+CC*(J''*(J''+1)+J"*(J"+1))/2+D*(J''*(J''+1)+J"*(J"+1))^2/4)';
      Name := 'Custom Transition Function';
      CheckVariables.Free;
      if cN <> nil then
        ReadFloatableValue(Node, IndexOfParameter('CC'), 'C', '0');
      if dN <> nil then
        ReadFloatableValue(Node, IndexOfParameter('D'), 'D', '0');
    end;
  end;
  Result:=inherited ReadProperties(Node, CheckUnread);
end;

{ TSingleTransitionMoment }

constructor TSingleTransitionMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',Strength=1';
  DimensionlessParameters := DimensionlessParameters + ',Strength';
  inherited;
end;

function TSingleTransitionMoment.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Relative transition (dipole) moment - note that the relative intensity is proportional to the square of this value.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

{ TNuclearMagneticMoment }

constructor TNuclearMagneticMoment.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',gN=1,Nucleus=1';
  DimensionlessParameters := DimensionlessParameters + ',gN';
  FNuclearMagneton := True;
  FMagnetic := True;
  inherited;
end;

function TNuclearMagneticMoment.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Nuclear g factor in terms of Nuclear Magnetons.'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

function TNuclearMagneticMoment.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 1  then
    Result := 'Index of nucleus'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TNuclearMagneticMoment.PairName: string;
begin
  Result := 'gN(';
  with Owner.Owner as TMolecule do
    if (nNuclei >= Nucleus) and (Nucleus > 0) then
      Result := Result + Nuclei[Nucleus-1].Name
    else
      Result := Result + '?';
  Result := Result + ')';
end;

{ TRootMNode }

procedure TRootMNode.BeginPaste;
begin
  FReadVersion := DecodePgopherVersion(VersionInfo);
end;

destructor TRootMNode.Destroy;
begin
  PerformMNodeNotification(Self, mnDestroy, -1);
  inherited;
end;

{ TCorrelations }

procedure TCorrelations.AddHandlers;
begin
  inherited;
  AddHandler('Resync', 'Adjust matrix so only floated parameters are included', {$IFDEF FPC}@{$ENDIF}Resync);
end;

procedure TCorrelations.AdjustSize;
var
  i, j, n: Integer;
begin
  { Construct empty correlation array based on size of Parameters }
  n := Length(FitDetails.Parameters);
  SetLength(FitDetails.Correlation, n);
  for i := 0 to n - 1 do begin
    SetLength(FitDetails.Correlation[i],i);
    for j := 0 to i - 1 do
      FitDetails.Correlation[i,j] := 0;
  end;
end;

procedure TCorrelations.Clear;
begin
  FitDetails.Parameters := nil;
end;

procedure TCorrelations.UnpackCorrelations(var UnpackedCorrelations: TUnpackedCorrelations);
var
  i, j, n: Integer;
begin
  SetLength(UnpackedCorrelations, Sqr(nEntries) div 2 + 1); { Guess capacity }
  n := 0;
  for i := 0 to nEntries-1 do
    for j := 0 to i - 1 do
      if FitDetails.Correlation[i][j] <> 0 then
        with UnpackedCorrelations[n] do begin
          Row := EntryName(i,True);
          Col := EntryName(j,True);
          Value := FitDetails.Correlation[i][j];
          Inc(n);
        end;
  SetLength(UnpackedCorrelations, n);
end;

constructor TCorrelations.Create(AOwner: TComponent);
var
  Mixture: TMixture;
begin
  inherited;
  Mixture := RootNode as TMixture;
  if Assigned(Mixture.FCorrelations) then
    raise Exception.Create('Duplicate correlations object')
  else
    Mixture.FCorrelations := Self;
end;

destructor TCorrelations.Destroy;
begin
  if RootNode <> Self then
    (RootNode as TMixture).FCorrelations := nil;
  { otherwise being destroyed }
  inherited;
end;

function TCorrelations.EntryName(i: Integer; Quote: Boolean): string;
begin
  with FitDetails.Parameters[i] do begin
    Result := MNode.FullParameterName(Index,Quote);
    if not (MNode.RootNode is TMixture) then
      Result := Copy(MNode.RootNode.ClassName,2,255)+'.'+Result;
  end;
end;

function TCorrelations.nEntries: Integer;
begin
  { Allow for inconsistent value of result during update }
  with FitDetails do
    Result := Min(Length(Parameters),Length(Correlation));
end;

procedure TCorrelations.PrettyPrint(var F: Text; PrintLevel: TPrintLevel; Indent: string);
var
  i, j, l: Integer;
begin
  if nEntries <= 1 then
    Exit;
  inherited;
  l := 9;
  for i := 0 to nEntries-1 do
    l := Max(l,Length(EntryName(i,False)));
  Write(F,' ':l);
  for i := 0 to nEntries-1 do
    Write(F, ' ', EntryName(i,False));
  WriteLn(F);
  for i := 0 to nEntries-1 do begin
    Write(F, EntryName(i,False):l);
    for j := 0 to i-1 do
      Write(F, ' ', FitDetails.Correlation[i,j]:Length(EntryName(j,False)):6);
    WriteLn(F, ' ', 1.0:Length(EntryName(i,False)):6);
  end;
end;

function TCorrelations.ReadData(UnpackedCorrelations: TUnpackedCorrelations; FirstPass: Integer; CheckUnread: Boolean; OtherNode:TRootMNode): Integer;
var
  UnRead: TStringList; { Record unrecognised parameters, only one error message per parameter }
  OtherStart: string;

  function FindParameter(Name: string):Integer;
  var
    viInfo: TValueInfo;
    I: Integer;
    ErrMsg: string;
  begin
   viInfo := RootNode.SubPropInfo(Name, ErrMsg);
   if (not Assigned(viInfo.MNode) or (viInfo.viType <> viParameter)) and Assigned(OtherNode) and (Pos(OtherStart,Name)=1) then
     viInfo := OtherNode.SubPropInfo(Copy(Name,Length(OtherStart)+1,Length(Name)), ErrMsg);
   Result := -1;
   if not Assigned(viInfo.MNode) or (viInfo.viType <> viParameter) then begin
     if (Pos('Overlays.',Name) = 1) and not Assigned(UnRead) and not Assigned(OtherNode) then
       Exit; { Sliently discard entries from Overlay parameters, if no overlay }
     if not Assigned(UnRead) then begin
       UnRead := TStringList.Create;
       if CheckUnread then
         WriteLn(Log, 'Unrecognized parameter(s) in correlation matrix:');
     end;
     if UnRead.IndexOf(Name) < 0 then begin
       UnRead.Add(Name);
       if CheckUnread then
         WriteLn(Log, '  ', Name, ' ', ErrMsg);
     end;
   end else begin
     for I := High(FitDetails.Parameters) downto 0 do
       if (FitDetails.Parameters[i].Index = viInfo.Index) and (FitDetails.Parameters[i].MNode = viInfo.MNode) then begin
         Result := I;
         Break;
       end;
     if (Result < 0) and (FirstPass = 0) then begin
       Result := Length(FitDetails.Parameters);
       SetLength(FitDetails.Parameters, Result+1);
       FitDetails.Parameters[Result].MNode := viInfo.MNode;
       FitDetails.Parameters[Result].Index := viInfo.Index;
       FitDetails.Parameters[Result].StdDev := viInfo.MNode.StdDev[viInfo.Index];
     end;
   end;
  end;
var
  Pass, i, j, e:Integer;
begin
  if FirstPass = 0 then
    Clear;
  UnRead := nil;
  Result := 0;
  if Assigned(OtherNode) then
    OtherStart := Copy(OtherNode.ClassName,2,255)+'.'
  else
    OtherStart := '';
  for Pass := FirstPass to 1 do begin
    for e := 0 to High(UnpackedCorrelations) do
      with UnpackedCorrelations[e] do begin
        i := FindParameter(Row);
        if i >= 0 then begin
          j := FindParameter(Col);
          if (j >= 0) and (Pass = 1) then
            if i > j then
              FitDetails.Correlation[i,j] := Value
            else
              FitDetails.Correlation[j,i] := Value
        end;
      end;
    if Pass = 0 then
      AdjustSize;
  end;
  if Assigned(UnRead) then begin
    Result := UnRead.Count;
    Unread.Free;
  end;
end;

function TCorrelations.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
var
  Child, NextChild: TDOMNode;
  C: TUnpackedCorrelations;
  n: Integer;
begin
  Result := 0;
  n := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do
    if CompareText(Child.NodeName, 'Correlation') = 0 then begin
      Inc(n);
      Child := Child.NextSibling;
    end;
  SetLength(C, n);
  n := 0;
  Child := Node.FirstChild;
  while Assigned(Child) do
    if CompareText(Child.NodeName, 'Correlation') = 0 then
      with C[n] do begin
        Row := ReadValue(Child, 'Row', '');
        Col := ReadValue(Child, 'Col', '');
        Value := StrToFloatEx(ReadValue(Child, 'Value', '0'));
        Inc(n);
        NextChild := Child.NextSibling;
        if CheckUnread then
          Result := Result + nUnreadProperties(Child);
        Node.RemoveChild(Child);
        Child := NextChild;
      end;
  Result := Result + inherited ReadProperties(Node, False);
  Result := Result + ReadData(C, 0, CheckUnread, GlobalOverlays);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

procedure TCorrelations.Resync(Sender: TObject);
var
  Mixture: TMixture;
  Original: TUnpackedCorrelations;
begin
  UnpackCorrelations(Original);
  Mixture := RootNode as TMixture;
  Mixture.FillParameters;
  if GlobalOverlays <> nil then
    GlobalOverlays.AddNodeParameters(Mixture.FittingParameters);
  FitDetails.Parameters := Copy(Mixture.FittingParameters);
  AdjustSize;
  ReadData(Original,1,False,GlobalOverlays);
  PerformMNodeNotification(RootNode, mnParameterChange, 1); { Spurious change notification - triggers refresh of Correlation amtrix display }
end;

function TCorrelations.WriteProperties(Parent: TDOMNode; AllParameters: Boolean; WithChildren: Boolean): TDOMElement;
var
  Node: TDOMElement;
  i, j: Integer;
begin
  Result := inherited WriteProperties(Parent, AllParameters, WithChildren);
  for i := 0 to nEntries-1 do
    for j := 0 to i - 1 do
      if FitDetails.Correlation[i][j] <> 0 then begin
        Node := Parent.OwnerDocument.CreateElement('Correlation');
        WriteValue(Node, 'Row', EntryName(i,True), '');
        WriteValue(Node, 'Col', EntryName(j,True), '');
        WriteValue(Node, 'Value', FloatToStr(FitDetails.Correlation[i][j]), '');
        Result.AppendChild(Node);
      end;
end;

{ TQnoInfoMNode }

procedure TQnoInfoMNode.AddHandlers;
begin
  if not (Self is TSimulation) and (RootNode is TMixture) then begin
    AddHandler('Level List by symmetry', 'Print energy level list in log window sorted by symmetry and energy', {$IFDEF FPC}@{$ENDIF}LevelListBySymmetry);
    AddHandler('Level List by energy', 'Print energy level list in log window sorted by energy', {$IFDEF FPC}@{$ENDIF}LevelListByEnergy);
    AddHandler('Partition Function Table', 'Print a table of partition function against temperature', {$IFDEF FPC}@{$ENDIF}PartitionFunctionTable);
  end;
  inherited;
end;

function TQnoInfoMNode.QnoInfo(QnoType: TQnoType): TQnoInfo;
begin
  Result := (Owner as TQnoInfoMNode).QnoInfo(QnoType);
end;

var
  EmaxShow: string = '0';
  Tstep: string = '100';

procedure TQnoInfoMNode.LevelListByEnergy(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  if not CalculationInProgress and InputQuery('Energy Level List','Highest Energy to Show (0 for all)',EmaxShow) then begin
    if Trim(EmaxShow) = '' then
      EmaxShow := '0';
    ShowLevelList(Self, lltByEnergy, StrToFloatEx(EmaxShow));
    ReShowLogForm;
  end;
  {$ENDIF}
end;

procedure TQnoInfoMNode.LevelListBySymmetry(Sender: TObject);
begin
  {$IFNDEF CONSOLE}
  if not CalculationInProgress and InputQuery('Energy Level List','Highest Energy to Show (0 for all)',EmaxShow) then begin
    if Trim(EmaxShow) = '' then
      EmaxShow := '0';
    ShowLevelList(Self, lltBySymmetry, StrToFloatEx(EmaxShow));
    ReShowLogForm;
  end;
  {$ENDIF}
end;

procedure TQnoInfoMNode.PartitionFunctionTable(Sender: TObject);
var
  Emax: Double;
  Lbl: string;
begin
  Emax := 0;
  if not TryStrToFloatEx(EmaxShow, EMax) then
    EmaxShow := '0';
  if Emax = 0 then
    Lbl := 'None'
  else
    Lbl := EmaxShow;
  Lbl := 'Energy Selection Threshold: ' + Lbl + LineEnd + ' (Set by "Level List by ....")'+LineEnd;
  {$IFNDEF CONSOLE}
  if not CalculationInProgress and InputQuery('Partition Function Table',Lbl+'Temperature Step Size:',Tstep) then begin
    ShowLevelList(Self, lltQTable, Emax, StrToFloatEx(Tstep));
    ReShowLogForm;
  end;
  {$ENDIF}
end;

{ TTransitionFunction }

function TTransitionFunction.HelpOnSetting(AIndex: Integer): string;
begin
  if AIndex = 3 then
    Result := 'Apply to symmetrized wavefunctions'
  else
    Result := inherited HelpOnSetting(AIndex);
end;

initialization
  TSpecies.RegisterClass(TMixture);
  TTransitionMoments.RegisterClass(TMolecule);
  TSphericalTransitionMoment.RegisterClass(TTransitionMoments);
  TCartesianTransitionMoment.RegisterClass(TTransitionMoments);
  TSpinTransitionMoment.RegisterClass(TTransitionMoments);
  TOrbitalTransitionMoment.RegisterClass(TTransitionMoments);
  TRotationalTransitionMoment.RegisterClass(TTransitionMoments);
  TNuclearMagneticMoment.RegisterClass(TTransitionMoments);
  TSimulation.RegisterClass(TMixture);
  TMixture.RegisterClass(nil);
  TCorrelations.RegisterClass(TMixture);
  TAxisSwitch.RegisterClass(TSphericalTransitionMoment);
  TAxisSwitch.RegisterClass(TCartesianTransitionMoment);
  BoltzMax := Ln(MaxDouble)-2;
  PopulationConnection := TDistribution.Create(nil);
  AtThreadEnd(@CleanSig);
  InitCriticalSection(EvaluatorLock);
finalization
  DoneCriticalsection(EvaluatorLock);
  FreeAndNil(PopulationConnection);
  SetLength(SymSelNames,0);
end.

